/**************************************************************************
 *
 * Package:    gov.nih.nlm.mms.filters;
 * Interface:  ConceptsWithDefinitionsFilter.java
 *
 *************************************************************************/
import gov.nih.nlm.mms.filters.AbstractFilter;
import gov.nih.nlm.umls.meta.Concept;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Properties;

import javax.swing.JCheckBox;
import javax.swing.JLabel;
import javax.swing.JPanel;

/**
 * Sample filter that removes {@link Concept}s that do not contain any
 * definitions.
 * 
 * @author Brian Carlsen
 */
public class ConceptsWithDefinitionsFilter extends AbstractFilter {

	//
	// State and GUI Variables
	//

	/**
	 * Indicates whether or not the filter is active.
	 */
	private boolean is_active;

	/**
	 * {@link JCheckBox} representing the is_active state variable
	 */
	private JCheckBox jis_active;

	//
	// Constructors
	//

	/**
	 * Instantiates a {@link ConceptsWithDefinitionsFilter}.
	 */
	public ConceptsWithDefinitionsFilter() {
	}

	//
	// Filter Implementation
	//

	/**
	 * Indicates that this filter cannot run in efficient mode.
	 * 
	 * @return <code>false</code>
	 */
	public boolean isEfficientEnabled() {
		return false;
	}

	/**
	 * Returns the state of the filter.
	 * 
	 * @return the state of the filter
	 */
	public Properties getProperties() {
		Properties props = new Properties();
		props.put("is_active", String.valueOf(is_active));
		return props;
	}

	/**
	 * Sets the state of the filter.
	 * 
	 * @param props
	 *            the state of the filter
	 */
	public void setProperties(Properties props) {
		is_active = true;
		if (props.containsKey("is_active"))
			is_active = Boolean.valueOf(props.getProperty("is_active"))
					.booleanValue();
	}

	/**
	 * Resets filter to "default" state. In this case, it activates the filter.
	 */
	public void resetToDefaults() {
		is_active = true;
	}

	/**
	 * Returns the GUI configuration {@link JPanel}.
	 * 
	 * @return the GUI configuration {@link JPanel}
	 */
	public JPanel getPanel() {
		JPanel panel = new JPanel();
		jis_active = new JCheckBox();
		jis_active.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				fireDataChanged();
				is_active = jis_active.isSelected();
			}
		});

		panel.add(jis_active);
		panel.add(new JLabel("Check to activate filter"));
		return panel;
	}

	/**
	 * Sets the panel selections from the state variables.
	 */
	public void setPanelSelections() {
		jis_active.setSelected(is_active);
	}

	/**
	 * Returns a brief description of the filter for display above the GUI
	 * panel.
	 * 
	 * @return a brief description of the filter for display above the GUI panel
	 */
	public String getTabDescription() {
		return "Enabling this filter will remove all concepts without definitions.";
	}

	/**
	 * Returns the name.
	 * 
	 * @return the name
	 */
	public String getName() {
		return "Concepts Without Definitions Filter";
	}

	/**
	 * Empties the {@link Concept} if it has no definitions.
	 * 
	 * @param concept
	 *            the {@link Concept}
	 */
	public void applyFilter(Concept concept) {
		if (is_active && concept.getDefinitions().length == 0)
			concept.clear();
	}

	/**
	 * Returns configuration details to be written to the log.
	 * 
	 * @return configuration details to be written to the log
	 */
	public String getConfigurationDetails() {
		if (is_active)
			return "Concepts without definitions were removed";
		else
			return "";
	}

}
