/**************************************************************************
 *
 * File: OddEvenFilter.java
 * Author: Suresh Srinivasan
 *
 **************************************************************************/

import gov.nih.nlm.mms.ApplicationConfiguration;
import gov.nih.nlm.mms.filters.AbstractFilter;
import gov.nih.nlm.swing.SwingToolkit;
import gov.nih.nlm.umls.meta.Concept;

import java.util.Properties;

import java.awt.BorderLayout;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import javax.swing.ButtonGroup;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.border.CompoundBorder;
import javax.swing.border.EmptyBorder;
import javax.swing.border.EtchedBorder;

/**
 *
 * This sample filter removes all odd/even CUIs.
 * To use this filter in MetamorphoSys, do the following.
 * <ul>
 *  <li>Add the <code>mms.jar</code> file to your <code>CLASSPATH</code>.
 *  <li>Compile this file.
 *  <li>Create a <code>.jar</code> file from the resulting
 *      <code>.class</code> file.
 *  <li>Copy the resulting <code>.jar</code> file to the
 *      <code>METAMSYS/ext</code> directory.
 *  <li>Run Metamorphosys, and select <code>File->Import Filter</code>.
 * </ul>
 *
 * @author Suresh Srinivasan
 *
 */
public class OddEvenFilter extends AbstractFilter {

  //
  // Constants
  //

  // Default setting for filter
  private static final boolean DEFAULT_ODD = true;

  //
  // Fields
  //

  // which CUIs to exclude: odd or even?
  private boolean exclude_odd = DEFAULT_ODD;

  // Maintain references to GUI elements
  private JRadioButton odd, even;

  //
  // Constructors
  //

  /**
   * Constructs a filter.
   */
  public OddEvenFilter() {}

  /**
   * Method provides a {@link JPanel} that will be added to the main GUI
   * help menu to allow the user to consult a help window
   * that is particular to this filter.
   * @return {@link JPanel} to be added to the Help menu
   */
  public JPanel getHelpPanel() {
    return SwingToolkit.getPanelForHTML(
        this.getClass().getClassLoader().getResource(
        "NosNecHelp.html"));
  }

  /**
   * Returns the main configuration panel.  This panel contains
   * two radio buttons in a common group.  One configures the
   * filter to exclude odd CUIs and the other configures the filter
   * to exclude even CUIs.
   * @return the main configuration {@link JPanel}
   */
  public JPanel getPanel() {
    ButtonGroup g = new ButtonGroup();

    // Create even button
    even = new JRadioButton("Exclude even CUIs");
    even.setSelected(!exclude_odd);
    even.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        fireDataChanged();
        exclude_odd = false;
      }
    });
    g.add(even);

    // Create odd button
    odd = new JRadioButton("Exclude odd CUIs");
    odd.setSelected(exclude_odd);
    odd.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        fireDataChanged();
        exclude_odd = true;
      }
    });
    g.add(odd);

    // Layout the page
    JPanel p = new JPanel(new BorderLayout());
    p.setBorder(new CompoundBorder(new EtchedBorder(EtchedBorder.RAISED), new EmptyBorder(10, 10, 10, 10)));
    //	p.setLayout(new FlowLayout());
    p.setLayout(new GridLayout(0, 1));
    p.add(odd);
    p.add(even);
    JPanel p1 = new JPanel();
    p1.setLayout(new BorderLayout());
    p1.add(p, BorderLayout.NORTH);
    p1.add(new JPanel(), BorderLayout.CENTER);

    //	return p;
    return p1;

  }

  /**
   * Return the description to appear above the main configuration
   * panel the GUI.
   * @return the description
   */
  public String getTabDescription() {
	return "This is a sample filter provided to demonstrate the import filter feature.\n\nThis filter allows you to filter out odd numbered or even numbered CUIs.";
  }

  /**
   * Reset configuration to default, odd CUIs should be excluded.
   */
  public void resetToDefaults() {
    exclude_odd = DEFAULT_ODD;
  }

  /**
   * Do nothing.  This filter does not require access to
   * the configuration object.
   * @param c the application {@link ApplicationConfiguration}
   */
  public void setConfiguration(ApplicationConfiguration c) { }

  /**
   * Returns a {@link Properties} object with the configuration.
   * @return a {@link Properties} object with the configuration.
   */
  public Properties getProperties() {
    Properties p = new Properties();
    p.setProperty("exclude_odd", String.valueOf(exclude_odd));
    return p;
  }

  /**
   * Sets the configuration based on a {@link Properties} object.
   * This is the inverse operation of {@link #getProperties()}.
   * @param props a {@link Properties} object containing the configuration
   *              properties for this class.
   */
  public void setProperties(Properties props) {
    exclude_odd =
      Boolean.valueOf(props.getProperty("exclude_odd","true")).booleanValue();
  }

  /**
   * Apply the filter to the specified {@link Concept}. Depending upon
   * the configuration the CUI is emptied.
   * @param concept {@link Concept}
   */
  public void applyFilter(Concept concept) {
    int c = new Integer(concept.getCui().substring(1)).intValue();
    boolean cuiIsEven = ((c%2) == 0);
    if ((cuiIsEven && !exclude_odd) || (!cuiIsEven && exclude_odd)) {
      concept.clear();
    }
    return;
  }

  /**
   * Configure the panel based on the background fields.
   */
  public void setPanelSelections() {
    if (exclude_odd)
      odd.setSelected(true);
    else
      even.setSelected(true);
  }

  /**
   * Returns the filter name.
   * @return the filter name
   */
  public String getName() {
    return "OddEven (for Testing)";
  }

  /**
   * Return the configuration details, as they should appear
   * in the subset log file. Multi-line comments should be
   * separated by {@link System}.getProperty("line.separator") to
   * ensure platform independence.
   * @return the configuration details
   */
  public String getConfigurationDetails() {
    if (exclude_odd)
      return "CUI values ending with odd numbers have been removed";
    else
      return "CUI values ending with even numbers have been removed";
  }

  /**
   * Indicates that the filter can run in "efficient mode".
    * @return <code>true</code>
   */
  public boolean isEfficientEnabled() {
    return true;
  }
}
