/**************************************************************************
 *
 * Package:    gov.nih.nlm.mms;
 * Interface:  InstallationGUI.java
 *
 *************************************************************************/
package gov.nih.nlm.mms;

import gov.nih.nlm.swing.FontSizeManager;
import gov.nih.nlm.swing.GlassPaneListener;
import gov.nih.nlm.swing.SwingToolkit;
import gov.nih.nlm.util.SystemToolkit;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.*;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import javax.swing.*;
import javax.swing.border.CompoundBorder;
import javax.swing.border.EmptyBorder;
import javax.swing.border.EtchedBorder;
import javax.swing.border.TitledBorder;

/**
 * Validates byte counts and MD5s for the NLM files.
 *
 * @author Deborah Shapiro
 */
public class ChecksumValidator {

  /**
   * File containing the .nlm files.
   */
  private File file_dir;

  /**
   * Maps of files to byte counts and MD5 values.
   */
  private Map file_to_byte_count = new HashMap();

  private Map file_to_md5 = new HashMap();

  /**
   * Lists of files for each status.
   */
  private List passed_files = new ArrayList();

  private List missing_files = new ArrayList();

  private List byte_count_mismatch_files = new ArrayList();

  private List md5_mismatch_files = new ArrayList();

  /**
   * Validation log
   */
  private BufferedWriter log;

  /**
   * Charts progress of MD5 computations.
   */
  private MD5ProgressMonitor md5_pm = null;
  private ProgressMonitor byte_pm = null;

  /**
   * Indicates whether or not progress was canceled.
   */
  private boolean progress_canceled = false;

  //
  // Calling frame
  //
  protected GlassPaneListener gpl = null;

  protected final List main_files = new ArrayList();

  protected final List extra_files = new ArrayList();

  private Date start_date;

  private Date end_date;

  //
  // Constructor
  //

  /**
   * Instantiates a {@link ChecksumValidator}.
   *
   * @param dir
   *          the directory containing the NLM files
   * @param gpl
   *          {@link GlassPaneListener}
   */
  public ChecksumValidator(File dir, GlassPaneListener gpl) {
    file_dir = dir;
    this.gpl = gpl;
  };

  //
  // Methods
  //

  /**
   * Validates the user selections for directories based on file length checks
   * and MD5 checksums.
   *
   * @return <code>true</code> if validation passes <code>false</code>
   *         otherwise
   * @throws IOException
   * @throws NoSuchAlgorithmException
   */
  public boolean validate() throws IOException, NoSuchAlgorithmException {
    // Validate file lengths and MD5 checksums
    return validateFiles(false);
  }

  /**
   * Validates the user selections for directories based on file length checks.
   *
   * @return <code>true</code> if validation passes <code>false</code>
   *         otherwise
   * @throws IOException
   * @throws NoSuchAlgorithmException
   */
  public boolean validateByteCountsOnly()
      throws IOException, NoSuchAlgorithmException {
    // Validate file lengths only
    return validateFiles(true);
  }

  /**
   * Validates the user selections for directories based on file length checks.
   * Also checks MD5 checksums if <code>byteCountsOnly</code> is false.
   *
   * @param byteCountsOnly <code>true</code> if only file byte counts are to be
   *                       verified.  <code>false</code> if both byte counts and
   *                       MD5 checksums are to be validated.
   * @return <code>true</code> if validation passes <code>false</code>
   *         otherwise
   * @throws IOException
   * @throws NoSuchAlgorithmException
   */
  private boolean validateFiles(boolean byteCountsOnly)
      throws IOException, NoSuchAlgorithmException {
    //
    // Get release version
    //
    String release_version = "";
    File release_file = new File(file_dir, "release.dat");
    if (release_file.exists()) {
      // Open and load the properties file
      Properties props = new Properties();
      FileInputStream propInputStream = new FileInputStream(release_file);
      props.load(propInputStream);
      Enumeration keys = props.propertyNames();
      while (keys.hasMoreElements()) {
        String key = (String) keys.nextElement();
        if (key.equals("umls.release.name"))
          release_version = props.getProperty(key);
      }
    } else {
    	release_version = MMSToolkit.getReleaseVersion();
    }

    //
    // Get CHK file
    //
    File chk_file = new File(file_dir, release_version.toUpperCase() + ".CHK");
    if (!chk_file.exists()) {
      MMSToolkit
          .notifyUser(
              WelcomeGUI.getWelcomeGUI(),
              "Expected file is missing:\n\n"
                  + chk_file.getCanonicalPath()
                  + "\n\n"
                  + "This file is used to validate the size of the distribution files.\n"
                  + "Validation cannot proceed without this file.\n");
      return false;
    }

    //
    // get MD5 file
    //
    File md5_file = null;

    if (!byteCountsOnly) {
      //
      // Look up expected MD5 values
      //
      md5_file = new File(file_dir, release_version.toUpperCase() + ".MD5");
      if (!md5_file.exists()) {
        MMSToolkit
            .notifyUser(
                WelcomeGUI.getWelcomeGUI(),
                "Expected file is missing:\n\n"
                    + md5_file.getCanonicalPath()
                    + "\n\n"
                    + "This file is used to validate the content of the distribution files.\n"
                    + "Validation cannot proceed without this file.\n");
        return false;
      }
    }

    //
    // Cache files and expected CHK values
    //
    final String[][] chk_data = SystemToolkit.readFieldedFile(chk_file, "\t");
    for (int i = 0; i < chk_data.length; i++)
      file_to_byte_count.put(chk_data[i][2], chk_data[i][1]);

    if (!byteCountsOnly) {
      //
      // Cache files and expected MD5 values
      //
      final String[][] md5_data =
          SystemToolkit.readFieldedFile(md5_file, "=()");
      for (int i = 0; i < md5_data.length; i++)
        file_to_md5.put(md5_data[i][1], md5_data[i][3].substring(1));
    }

    populateFileNameLists();

    //
    // Perform CHK validation
    //
    start_date = new Date();
    initializeByteCountProgressMonitor();
    performByteCountValidation();

    if (!byteCountsOnly) {
      //
      // Perform MD5 checks
      //
      initializeMD5ProgressMonitor();
      performMD5Validation();
    }

    //
    // Close progress monitor
    //
    if (SwingToolkit.usingView()) {
      if (!byteCountsOnly)
        md5_pm.close();
      byte_pm.close();
    }

    end_date = new Date();

    //
    // Prepare report
    //
    final StringBuffer final_report = new StringBuffer(getReport());

    //
    // Open validation.log file
    //
    File log_file = null;
    boolean is_log_file_temp = false;
    try {
      log_file = new File(file_dir, "validation.log");
      if (!log_file.getParentFile().canWrite()) {
        log_file = File.createTempFile("tmp", ".txt");
        is_log_file_temp = true;
      }
      try {
        log = new BufferedWriter(new FileWriter(log_file));
      } catch (IOException ioe) {
        //
        // Treat this exception as if canWrite() returned true
        // for a DVD RW drive with a read-only DVD in it
        //
        log_file = File.createTempFile("tmp", ".txt");
        is_log_file_temp = true;
        log = new BufferedWriter(new FileWriter(log_file));        
      }
    } catch (Exception e) {
      e.printStackTrace();
      MMSToolkit.reportError(WelcomeGUI.getWelcomeGUI(),
          "Unexpected error trying to open the validation.log file.\n"
              + "See console for more details.", false);
    }

    //
    // Report validation results
    //
    if (progress_canceled) {
      MMSToolkit.notifyUser(WelcomeGUI.getWelcomeGUI(),
          "Validation of NLM files was cancelled.");
      return false;
    } else if (missing_files.isEmpty() && byte_count_mismatch_files.isEmpty()
        && md5_mismatch_files.isEmpty()) {
      if (!byteCountsOnly) {
        String msg = "All files passed all validation checks.";
        MMSToolkit.notifyUser(WelcomeGUI.getWelcomeGUI(), msg);
        log.write(msg);
        log.close();
      }
    } else {
      // Date sd = new Date();
      log.write(final_report.toString());
      log.close();

      if (byteCountsOnly) {
        // If checking byte counts only (as for installation) ask user whether
        //  to continue despite incorrect byte counts
      	String[] options = {"Continue", "Cancel"};
        String answer = (String) SwingToolkit.displayTextFile(WelcomeGUI.getWelcomeGUI(),
            "Validation Log", log_file, options, 20, 90);
        if (is_log_file_temp)
          log_file.delete();
        if (answer.equals("Cancel")) {
          return false;
        }
      }
      else {
        SwingToolkit.displayTextFile(WelcomeGUI.getWelcomeGUI(),
            "Validation Log", log_file);
        if (is_log_file_temp)
          log_file.delete();
        return false;
      }
    }

    return true;
  }

  /**
   * Validate the byte counts for files listed in VERSION.CHK.
   */
  private void performByteCountValidation() {
    int i = 0;
    for (; i < main_files.size(); i++) {
      performByteCountValidationForFile((String) main_files.get(i));
      byte_pm.setProgress(i);
      if (byte_pm.isCanceled()) { progress_canceled = true; return; }
    }

    for (int j = 0; j < extra_files.size(); j++) {
      byte_pm.setProgress(i+j);
      performByteCountValidationForFile((String) extra_files.get(j));
      if (byte_pm.isCanceled()) { progress_canceled = true; return; }
    }
    byte_pm.setProgress(byte_pm.getMaximum());
  }

  /**
   * Validate the md5s for files listed in VERSION.MD5.
   * 
   * @throws IOException
   * @throws NoSuchAlgorithmException
   */
  private void performMD5Validation() throws IOException, NoSuchAlgorithmException {
    for (int i = 0; i < main_files.size(); i++) {
      performMD5ValidationForFile((String) main_files.get(i));
      if (progress_canceled)
        return;
    }

    for (int i = 0; i < extra_files.size(); i++) {
      performMD5ValidationForFile((String) extra_files.get(i));
      if (progress_canceled)
        return;
    }
  }

  /**
   * Looks for a byte count mismatch on the given file. Compares the byte count
   * with the expected count from the VERSION.CHK file.
   * 
   * @param file
   *          {@link String}
   */
  private void performByteCountValidationForFile(String file) {

    final String str_count = (String) file_to_byte_count.get(file);
    final long byte_count = new Integer(str_count).longValue();
    final File local_file = new File(file_dir, file);
    if (!local_file.exists())
      addMissingFile(file);
    else if (byte_count != local_file.length())
      addByteCountMismatchFile(file);
    else
      addPassedFile(file);
    
  }

  /**
   * Looks for a md5 mismatch on the given file. Compares the md5 with the
   * expected md5 from the VERSION.MD5 file.
   * 
   * @param file
   *          {@link String}
   * @throws IOException
   * @throws NoSuchAlgorithmException
   */
  private void performMD5ValidationForFile(String file) throws IOException,
      NoSuchAlgorithmException {
    if (SwingToolkit.usingView()) {
      md5_pm.incrementOperation();
      if (main_files.contains(file)) {
        md5_pm.setNote("Validating " + file + " (0%)");
      }
    }

    final String md5 = (String) file_to_md5.get(file);
    final File local_file = new File(file_dir, file);

    if (local_file.exists()) {
      if (SwingToolkit.usingView())
        progress_canceled = md5_pm.isCanceled();

      final String local_md5 = SystemToolkit.md5(local_file, md5_pm);
      if (!local_md5.equals(md5))
        addMD5MismatchFile(file, md5, local_md5);
    }
  }

  /**
   * Sets up the {@link ProgressMonitor} and calculate the number of files that
   * will be validated and whose progress will be monitored.
   */
  private void initializeMD5ProgressMonitor() {

    //
    // Set up progress monitor if using view
    //
    if (SwingToolkit.usingView()) {

      md5_pm = new MD5ProgressMonitor();
      md5_pm.setMinimum(0);
      md5_pm.setMaximum(100);

      byte_pm = new ProgressMonitor(
          (Component)gpl,"Checking file lengths","",
          0,main_files.size() + extra_files.size());
      byte_pm.setMillisToDecideToPopup(1000);
      byte_pm.setMillisToPopup(50);
    } else {
      System.out.println("Validating NLM files ... ");
    }

  }

  /**
   * Sets up the {@link ProgressMonitor} for byte count validation
   */
  private void initializeByteCountProgressMonitor() {

    //
    // Set up progress monitor if using view
    //
    if (SwingToolkit.usingView()) {
      byte_pm = new ProgressMonitor(
          (Component)gpl,"Checking file lengths","",
          0,main_files.size() + extra_files.size());
      byte_pm.setMillisToDecideToPopup(1000);
      byte_pm.setMillisToPopup(50);
    } else {
      System.out.println("Validating NLM file byte counts ... ");
    }

  }

  /**
   * Populates the lists of "main" and "extra" files.
   */
  private void populateFileNameLists() {
    final List all_file_names = new ArrayList(file_to_byte_count.keySet());
    Collections.sort(all_file_names);
    for (int i = 0; i < all_file_names.size(); i++) {
      final String file_name = (String) all_file_names.get(i);
      
      // Skip the .sav file
      if (file_name.endsWith("mmsys.prop.sav")) continue;

      if (file_name.indexOf("-meta") != -1
          || file_name.indexOf("-otherks") != -1
          || file_name.equals("mmsys.zip"))
        main_files.add(file_name);
      else
        extra_files.add(file_name);
    }
  }

  /**
   * Returns written report of checksum validations. Indicates files that
   * passed, had byte count mismatches, had md5 mismatches, and files that were
   * missing.
   * 
   * @return {@link String} report
   */
  private String getReport() {
    final StringBuffer final_error_text = new StringBuffer();
    // Report results to user
    final_error_text.append("Validation Summary\n");
    final_error_text
        .append(passed_files.size() + "\tfiles passed validation\n");
    final_error_text.append(missing_files.size() + "\tfiles are missing\n");
    final_error_text.append(byte_count_mismatch_files.size()
        + "\tfiles have unexpected byte counts\n");
    final_error_text.append(md5_mismatch_files.size()
        + "\tfiles have unexpected content\n"
        + "       \t(the files may be corrupt or of the wrong version)\n\n");
    final_error_text.append("Validation started at:  " + start_date);
    final_error_text.append("\nValidation ended at:    " + end_date + "\n");

    //
    // Add missing file list
    //
    if (!missing_files.isEmpty()) {
      final_error_text.append("The following expected file(s) are missing:\n");
      Collections.sort(missing_files);
      for (int i = 0; i < missing_files.size(); i++)
        final_error_text.append((String) missing_files.get(i)).append("\n");

    }

    //
    // Add byte count mismatches
    //
    if (!byte_count_mismatch_files.isEmpty()) {
      final_error_text
          .append("\nThe following file(s) have unexpected byte counts:\n");
      Collections.sort(byte_count_mismatch_files);
      for (int i = 0; i < byte_count_mismatch_files.size(); i++)
        final_error_text.append((String) byte_count_mismatch_files.get(i))
            .append("\n");
    }

    //
    // Add MD5 mismatches
    //
    if (!md5_mismatch_files.isEmpty()) {
      final_error_text
          .append("\nThe following file(s) have unexpected content\n"
              + "(the files may be corrupt or of the wrong version):\n");
      Collections.sort(md5_mismatch_files);
      for (int i = 0; i < md5_mismatch_files.size(); i++)
        final_error_text.append((String) md5_mismatch_files.get(i))
            .append("\n");
    }

    //
    // List passed files
    //
    if (!passed_files.isEmpty()) {
      final_error_text
          .append("\nThe following file(s) passed all validation checks:\n");
      Collections.sort(passed_files);
      for (int i = 0; i < passed_files.size(); i++)
        final_error_text.append((String) passed_files.get(i)).append("\n");
    }

    return final_error_text.toString();

  }

  /**
   * Add given file name to the set of expected files that are not found.
   * 
   * @param file
   *          name of file
   */
  private void addMissingFile(String file) {
    if (!file.equals("mmsys.zip"))
      missing_files.add(file);
  }

  /**
   * Add given file name to the set of files that had an unexpected byte count.
   * 
   * @param file
   *          name of file
   */
  private void addByteCountMismatchFile(String file) {
    byte_count_mismatch_files.add(file);
    if (passed_files.contains(file))
      passed_files.remove(file);
  }

  /**
   * Add given file name to the set of files that had an unexpected md5 (message
   * digest) validation.
   * 
   * @param file
   *          name of file
   * @param expected_md5
   *          expected md5 string
   * @param actual_md5
   *          actual md5 string
   */
  private void addMD5MismatchFile(String file, String expected_md5,
      String actual_md5) {
    md5_mismatch_files.add(file + " (expected: " + expected_md5 + ", actual: "
        + actual_md5 + ")");
    if (passed_files.contains(file))
      passed_files.remove(file);
  }

  /**
   * Add given file name to the set of files that passed validation.
   * 
   * @param file
   *          name of file
   */
  private void addPassedFile(String file) {
    passed_files.add(file);
  }

  //
  // Inner class
  //
  /**
   * Progress monitor.
   */
  private class MD5ProgressMonitor extends ProgressMonitor implements
      ActionListener {

    //
    // fields
    //
    private String[] files = null;

    private JCheckBox[] boxes = null;

    private JLabel progress_label = new JLabel();

    private JProgressBar jprogress_bar = new JProgressBar();

    // Create an action object
    private Action cancelAction = new AbstractAction("Cancel") {
      {
        // Set action command to "Cancel"
        putValue(Action.ACTION_COMMAND_KEY, getValue(Action.NAME));
      }

      // This method is called when the action is invoked
      public void actionPerformed(ActionEvent e) {
        cancelActionPerformed(e);
      }
    };
    private JButton jcancel = new JButton(cancelAction);

    private boolean is_canceled = false;

    private JFrame frame = null;

    //
    // Index into progress list
    //
    private int operation = -1;

    //
    // Minimum and maximum values for a local operation.
    //
    private int maximum = 0;

    private int minimum = 0;

    /**
     * Instantiates the progress monitor with the specified file list.
     */
    public MD5ProgressMonitor() {
      super(null, null, null, 0, 0);
      setMillisToDecideToPopup(1000000000);
      frame = new JFrame();
      frame.setLocationRelativeTo(null);
      frame.setTitle("Validating NLM File");
      if (gpl != null)
        gpl.enableGlassPane();

      // get full list of files
      files = new String[main_files.size() + extra_files.size()];
      int z = 0;
      for (; z < main_files.size(); z++)
        files[z] = (String) main_files.get(z);
      for (; z < files.length; z++)
        files[z] = (String) extra_files.get(z - main_files.size());

      boxes = new JCheckBox[main_files.size() + 1];

      // Configure Layout
      frame.getContentPane().setLayout(new BorderLayout(20, 20));
      final JPanel operation_panel = new JPanel(new GridBagLayout());
      final GridBagConstraints constraints = new GridBagConstraints();
      constraints.fill = GridBagConstraints.BOTH;
      constraints.insets = new Insets(2, 2, 2, 2);

      // Configure Panel
      constraints.gridy = 0;
      constraints.gridwidth = 3;
      constraints.gridy++;
      JLabel operation_label = new JLabel(
          "<html><center>Validating File(s)<br>"
              + "MD5 calculations will be performed for:<br></center></html>");
      operation_panel.add(operation_label, constraints);
      constraints.gridy++;

      for (int i = 0; i < main_files.size(); i++) {
        boxes[i] = new JCheckBox(files[i]);
        boxes[i].setEnabled(true);
        final MouseListener[] mls = (MouseListener[]) boxes[i]
            .getListeners(MouseListener.class);
        for (int j = 0; j < mls.length; j++)
          boxes[i].removeMouseListener(mls[j]);
        operation_panel.add(boxes[i], constraints);
        constraints.gridy++;
      }
      boxes[main_files.size()] = new JCheckBox("All other files");
      boxes[main_files.size()].setEnabled(true);
      final MouseListener[] mls = (MouseListener[]) boxes[main_files.size()]
          .getListeners(MouseListener.class);
      for (int j = 0; j < mls.length; j++)
        boxes[main_files.size()].removeMouseListener(mls[j]);

      operation_panel.add(boxes[main_files.size()], constraints);
      constraints.gridy++;

      progress_label.setText("Operation is % Complete");
      operation_panel.add(progress_label, constraints);
      constraints.gridy++;
      operation_panel.add(jprogress_bar, constraints);
      constraints.gridy++;

      // Register keystroke
      operation_panel.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(
          KeyStroke.getKeyStroke("ESCAPE"), cancelAction.getValue(Action.NAME));
      // Register action
      operation_panel.getActionMap().put(cancelAction.getValue(Action.NAME),
          cancelAction);
      operation_panel.add(jcancel, constraints);

      // Initially invisible
      jprogress_bar.setVisible(true);
      jprogress_bar.setMinimum(0);
      jprogress_bar.setMaximum(100);

      // Compose Panels, pack and make visible
      final TitledBorder titledBorder = new TitledBorder(new EtchedBorder(
          EtchedBorder.RAISED), "");
      operation_panel.setBorder(new CompoundBorder(titledBorder,
          new EmptyBorder(10, 10, 10, 10)));
      frame.getContentPane().add(operation_panel, BorderLayout.CENTER);
      FontSizeManager.adjustFontSize(frame);
      frame.pack();
      frame.setVisible(true);

      // Configure window listener
      frame.addWindowListener(new WindowAdapter() {
        public void windowClosing(WindowEvent e) {
          is_canceled = true;
        }
      });
    }

    /**
     * Increments the operation counter. Causes the previous checkbox to become
     * checked and disabled. Indicates that the next checkbox is now being
     * processed.
     */
    public void incrementOperation() {
      if (operation < main_files.size()) {
        if (operation > -1) {
          boxes[operation].setEnabled(false);
        }
      }
      operation++;
      if (operation < boxes.length) {
        boxes[operation].setSelected(true);
        boxes[operation].setEnabled(true);
      }
      frame.setTitle("Validating...");
    }

    /**
     * Used by locally defined Action for cancel.
     *
     * @param e a "Cancel" action event
     */
    private void cancelActionPerformed(ActionEvent e) {
      actionPerformed(e);
    }

    /**
     * Called when "Cancel" button is clicked. Confirm with user, then exit
     * application.
     * 
     * @param e
     *          the {@link ActionEvent}
     */
    public void actionPerformed(ActionEvent e) {
      is_canceled = true;
    }

    /**
     * Indicates whether or not processing was cancelled.
     * 
     * @return <code>boolean</code> true if cancelled false otherwise
     */
    public boolean isCanceled() {
      return is_canceled;
    }

    /**
     * Closes the progress monitor.
     */
    public void close() {
      super.close();
      frame.dispose();
      if (gpl != null)
        gpl.disableGlassPane();
    }

    /**
     * Sets the progress note.
     * 
     * @param note
     *          the progress note
     */
    public void setNote(String note) {
      progress_label.setText(note);
    }

    /**
     * Computes the total progress for the current operation.
     * 
     * @param nv
     *          the progress
     */
    public void setProgress(int nv) {
      int pct_progress = 0;

      if (operation >= main_files.size()) {
        pct_progress = ((operation - main_files.size()) * 100)
            / extra_files.size();
        setNote("Validating other files (" + pct_progress + "%)");
      } else {
        pct_progress = ((nv - minimum) * 100) / (maximum - minimum);
        setNote("Validating " + files[operation] + " (" + pct_progress + "%)");
      }
      jprogress_bar.setValue(pct_progress);
    }

    /**
     * Sets max value for current operation.
     * 
     * @param max
     *          the max value
     */
    public void setMaximum(int max) {
      maximum = max;
    }

    /**
     * Set min value for current operation.
     * 
     * @param min
     *          the min value
     */
    public void setMinimum(int min) {
      minimum = min;
    }

  }

} // End of ChecksumValidator class
