/*****************************************************************************
 *
 * Object:  ConceptOutputStreamAction
 * Author:  Brian Carlsen, Deborah Shapiro
 *
 *****************************************************************************/

package gov.nih.nlm.mms;

import gov.nih.nlm.swing.ListDialog;

import java.awt.event.ActionEvent;
import java.io.IOException;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.JPanel;
import javax.swing.JTextField;

/**
 * Used to change the output stream. It is attached to the
 * {@link MetamorphoSysGUI} "Browse" button in the "Output Options" tab.
 * 
 * @author Brian Carlsen, Deborah Shapiro
 */
public class ConceptOutputStreamAction extends AbstractAction {

  //
  // Protected Fields
  //
  protected JTextField mmsys_output_stream_field = null;

  protected MetamorphoSysGUI gui;

  protected UserConfiguration user_configuration;

  protected JPanel content_panel;

  //
  // Constructor
  //

  /**
   * Instantiates the action.
   * 
   * @param gui
   *          the {@link MetamorphoSysGUI}
   * @param cisf
   *          the {@link JTextField} whose value is the output location
   * @param content_panel
   *          the {@link JPanel} that changes on the gui when the output stream
   *          changes
   */
  public ConceptOutputStreamAction(MetamorphoSysGUI gui, JTextField cisf,
      JPanel content_panel) {
    super();
    this.mmsys_output_stream_field = cisf;
    this.gui = gui;
    this.content_panel = content_panel;

    // configure action
    putValue(Action.NAME, "Browse..");
    putValue(Action.SHORT_DESCRIPTION, "Choose Output Format");
  }

  /**
   * Set the {@link ConceptOutputStream}.
   * 
   * @param e
   *          an {@link ActionEvent}
   */
  public void actionPerformed(ActionEvent e) {

    //
    // Get user configuration reference
    //
    user_configuration = gui.getUserConfiguration();

    try {

      //
      // Find all implementations in the classpath
      //
      String[] subinterfaces = MMSToolkit
          .getInterfaceImplementations("gov.nih.nlm.mms.MetamorphoSysOutputStream");
      String[] display_subinterfaces = new String[subinterfaces.length];
      int initial_selection = 0;

      //
      // Acquire names of output streams and initial selection
      //
      for (int i = 0; i < subinterfaces.length; i++) {
        MetamorphoSysOutputStream cis = (MetamorphoSysOutputStream) MMSToolkit
            .instantiateObject(subinterfaces[i]);
        display_subinterfaces[i] = cis.getName();
        if (display_subinterfaces[i].equals(user_configuration
            .getOutputStream().getName()))
          initial_selection = i;
      }

      //
      // Ask user for selection
      //
      String mmsys_output_stream_name = (String) ListDialog.showListSingleMode(
          MMSToolkit.getGui(), "", "Output File Format", display_subinterfaces,
          (Object) display_subinterfaces[initial_selection]);

      //
      // If the same name was chosen, do nothing
      //
      if (user_configuration.getOutputStream().getName().equals(
          mmsys_output_stream_name))
        return;

      //
      // Iterate through choices
      //
      for (int i = 0; i < display_subinterfaces.length; i++) {

        //
        // Find user selection
        //
        if (display_subinterfaces[i].equals(mmsys_output_stream_name)) {

          //
          // Get old stream, save its properties
          //
          MetamorphoSysOutputStream old_stream = user_configuration
              .getOutputStream();
          user_configuration.setProperties(old_stream);

          //
          // Update GUI
          //
          gui.changeOutputStream((MetamorphoSysOutputStream) MMSToolkit
              .instantiateObject(subinterfaces[i]), mmsys_output_stream_field,
              content_panel);

          //
          // Prepare undo stack
          //
          Object[] tuple = {
              new OutputStreamUndoHandler(gui, mmsys_output_stream_field,
                  content_panel), old_stream,
              user_configuration.getOutputStream() };
          gui.pushOnUndoStack(tuple);

          //
          // Finished
          //
          break;
        }
      }
    } catch (ClassNotFoundException cnfe) {
      //
      // this exception will never happen b/c the list to choose from consists
      // of classes that are known to exist
      //
      cnfe.printStackTrace();
    } catch (IOException ie) {
      ie.printStackTrace();
      MMSToolkit
          .reportError(MMSToolkit.getGui(),
              "Unexpected error while looking for installed output handlers.\n"
                  + "This is a serious error that cannot be recovered from.",
              false);
    }

  }
}