/*****************************************************************************
 *
 * Package: gov.nih.nlm.mms
 * Object:  ConceptQueue.java
 *
 ****************************************************************************/
package gov.nih.nlm.mms;

import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.impl.ConceptImpl;

import java.util.Date;

/**
 * First In First Out queue for {@link Concept} objects.  A producer
 * adds concepts until the queue is full and then blocks.  A consumer
 * reads concepts out (notifying the producer when room is available
 * for adding more) and blocks when there are no more concepts to read.
 * The producer wakes up the consumer when there is more data to read.
 *
 * @author Brian Carlsen
 */
public class ConceptQueue {

  //
  // Fields
  //
  private Concept[] concepts = null;
  private int next_consumer_index = 0;
  private int next_producer_index = 0;
  private boolean producer_finished = false;
  private boolean is_producer_waiting = false;
  private boolean is_consumer_waiting = false;
  private Object block = new Object();

  //
  // Statistics for dynamic load balancing
  //
  private long producer_last_request = new Date().getTime();
  private long producer_request_ct = 1;
  private long producer_total_request_time = 300;
  private long consumer_last_request = new Date().getTime();
  private long consumer_request_ct = 1;
  private long consumer_total_request_time = 300;


  //
  // Constructor
  //

  /**
   * Instantiates a {@link ConceptQueue} with the specified size.
   * @param size the size of the queue
   */
  public ConceptQueue(int size) {
    concepts = new Concept[size];
  }

  /**
   * Indicates whether or not the producer is finished adding {@link Concept}s.
   * @return <code>true</code> if so, <code>false</code> otherwise
   */
  public boolean isProducerFinished() { return producer_finished; }

  /**
   * Sets the flag to <code>true</codE> indicating that the producer is finished
   * adding {@link Concept}s.
   */
  public void setProducerFinished() {
    producer_finished = true;
    // Wake client if it is sleeping
    synchronized (block) {
      block.notifyAll();
    }
  }

  /**
   * Indicates whether or not the queue is full.
   * @return <code>true</code> if full, <codE>false</code> otherwise
   */
  public boolean isFull() {
    return next_consumer_index == next_producer_index && concepts[next_producer_index] != null;
  }

  /**
   * Indicates whether or not the queue is empty.
   * @return <code>true</code> if empty, <codE>false</code> otherwise
   */
  public boolean isEmpty() {
    return next_consumer_index == next_producer_index && concepts[next_consumer_index] == null;
  }

  /**
   * Adds a concept to the queue, blocks if the queue is full afterwards.
   * Wakes up a sleeping consumer if the queue is at least 50% full.
   * @param concept the {@link Concept} to add
   */
  public void put(Concept concept) {
    producer_total_request_time += (new Date()).getTime() - producer_last_request;
    producer_request_ct++;

    //
    // Block if we need to wait.  When waking up, we will be ready to add more
    //
    checkProducerWait();
    producer_last_request = new Date().getTime();

    //
    //
    //
    concepts[next_producer_index++] = concept;
    if (next_producer_index >= concepts.length) next_producer_index = 0;

    checkConsumerNotify();

  }

  /**
   * Removes a {@link Concept} from the queue, blocks if the queue is empty afterwards.
   * Wakes up a sleeping producer if the queue is at less than 50% full.
   * @return the {@link Concept}
   */
  public Concept get() {
    // System.out.println("pi: " + next_producer_index + ", ci: " + next_consumer_index);
    Concept concept = null;
    consumer_total_request_time += (new Date()).getTime() - consumer_last_request;
    consumer_request_ct ++;

    //
    // Block if we need to wait.  When waking up, we will be ready to return
    //
    checkConsumerWait();
    if (isEmpty() && isProducerFinished()) return null;
    consumer_last_request = new Date().getTime();

    //
    // Get concept, null reference
    //
    concept = concepts[next_consumer_index];
    concepts[next_consumer_index++] = null;
    if (next_consumer_index >= concepts.length) next_consumer_index = 0;

    checkProducerNotify();

    return concept;
  }

  /**
   * Check to see if producer should wait, if so, wait.
   */
  private void checkProducerWait() {
    try {
      synchronized (block) {
        if (isFull()) {
          is_producer_waiting = true;
          // System.out.println("Producer blocked");
          if (is_consumer_waiting)
            block.notifyAll();
          block.wait();
          // System.out.println("Producer unblocked");
          is_producer_waiting = false;
        }
      }
    } catch (InterruptedException ie) {}
  }

  /**
   * Check to see if producer should wait, if so, wait.
   */
  private void checkConsumerWait() {
    try {
      synchronized (block) {
        // System.out.println("Consumer blocked");
        if (isEmpty() && !isProducerFinished()) {
          is_consumer_waiting = true;
          if (is_producer_waiting)
            block.notifyAll();
          block.wait();
          // System.out.println("Consumer unblocked");
          is_consumer_waiting = false;
        }
      }
    }  catch (InterruptedException ie) {}
  }

  /**
   * Check to see if consumer should awaken.
   */
  private void checkConsumerNotify() {
    //
    // Get consumer/producer average time stats
    //
    double producer_avg = (producer_total_request_time / producer_request_ct);
    double consumer_avg = (consumer_total_request_time / consumer_request_ct);
    int perf_gap = (int)(concepts.length * (1 - (consumer_avg / (producer_avg+consumer_avg))));

    // System.out.println("Checking Consumer Notify");
    // System.out.println("pa, ca, gap: " + producer_avg + ", " + consumer_avg + ", "  + perf_gap);
    // System.out.println("Concepts waiting: " + getConceptWaitingCount());

    //
    // If consumer is waiting and producer is faster, wake consumer
    // OR if consumer is waiting and is faster, wake consumer if
    // the number of concepts waiting to be processed = gap
    //
    if (is_consumer_waiting &&
        (producer_avg < consumer_avg ||
         (consumer_avg < producer_avg && getConceptWaitingCount() >= perf_gap))) {
      synchronized (block) {
        block.notifyAll();
        // System.out.println("Consumer Notified");
      }
    }
  }


  /**
   * Check to see if producer should awaken.
   */
  private void checkProducerNotify() {

    //
    // Get consumer/producer average time stats
    //
    double producer_avg = (producer_total_request_time / producer_request_ct);
    double consumer_avg = (consumer_total_request_time / consumer_request_ct);
    int perf_gap = (int)(concepts.length * (producer_avg / (producer_avg+consumer_avg)));

    // System.out.println("Checking Producer Notify");
    // System.out.println("pa, ca, gap: " + producer_avg + ", " + consumer_avg + ", "  + perf_gap);
    // System.out.println("Concepts waiting: " + getConceptWaitingCount());
    //
    // If producer is waiting and consumer is faster, wake producer
    // OR if producer is waiting and is faster, wake producer if
    // the number of concepts waiting to be processed < gap
    //
    if (is_producer_waiting &&
        (consumer_avg < producer_avg ||
         (producer_avg < consumer_avg && getConceptWaitingCount() <= perf_gap))) {
      synchronized (block) {
        block.notifyAll();
        // System.out.println("Producer Notified");
      }
    }
  }

  /**
   * Returns the number of concepts waiting to be processed.
   * @return the number of concepts waiting to be processed
   */
  private int getConceptWaitingCount() {
    if (next_producer_index > next_consumer_index) {
      return next_producer_index - next_consumer_index;
    } else if (next_producer_index == next_consumer_index) {
      if (concepts[next_consumer_index] == null) return 0;
      else return concepts.length;
    } else
      return next_producer_index + (concepts.length - next_consumer_index);
  }

  /**
   * Self-qa test.
   * @param s command line arguments
   */
  public static void main (String[] s) {
    // Test the case where producer is faster
    final ConceptQueue queue = new ConceptQueue(100);
    final int producer_delay = 200;
    final int consumer_delay = 200;
    Thread producer = new Thread (
      new Runnable() {
        public void run() {
           for (int i = 0; i < 200; i ++) {
             try { Thread.sleep(producer_delay); } catch (InterruptedException e) {}
             // System.out.println("Produce C" + i);
             Concept c = new ConceptImpl();
             c.setCui("C"+i);
             queue.put(c);
           }
           queue.setProducerFinished();
        }
      }
    );

    Thread consumer = new Thread (
      new Runnable() {
        public void run() {
          while (true) {
            Concept c = queue.get();
            if (c == null) break;
            // System.out.println("Consume " + c.getCui());
            try {Thread.sleep(consumer_delay);} catch (InterruptedException ie) {}
          }

        }
      }
    );

   producer.start();
   consumer.start();

   try {
     producer.join();
     consumer.join();
   } catch (InterruptedException ie) { ie.printStackTrace(); }

 }
}
