/**************************************************************************
 *
 * Package:    gov.nih.nlm.mms;
 * Interface:  CopyToHardDriveGUI.java
 *
 *************************************************************************/
package gov.nih.nlm.mms;

import gov.nih.nlm.mms.browser.RRFBrowserGUI;
import gov.nih.nlm.swing.FontSizeManager;
import gov.nih.nlm.swing.GlassComponent;
import gov.nih.nlm.swing.GlassPaneListener;
import gov.nih.nlm.swing.SwingToolkit;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Date;

import javax.swing.*;
import javax.swing.border.CompoundBorder;
import javax.swing.border.EmptyBorder;
import javax.swing.border.EtchedBorder;
import javax.swing.text.SimpleAttributeSet;
import javax.swing.text.StyleConstants;

/**
 * This class is used for the <code>File->Copy to Hard Drive</code> menu of
 * the {@link WelcomeGUI}.
 * 
 * @author Brian Carlsen, Deborah Shapiro
 */
public class CopyToHardDriveAction extends AbstractAction {
  private static final String SEPARATOR = System.getProperty("file.separator");

  /**
   * Instantiates a {@link CopyToHardDriveAction}.
   */
  public CopyToHardDriveAction() {
    super();
    // configure action
    putValue(Action.NAME, "Copy To Hard Drive");
    putValue(Action.SHORT_DESCRIPTION, "Copy To Hard Drive");
  }

  /**
   * Opens a small dialog informing the user about the application.
   * 
   * @param e
   *          an {@link java.awt.event.ActionEvent}
   */
  public void actionPerformed(ActionEvent e) {
    CopyToHardDriveGUI copy_gui = new CopyToHardDriveGUI();
    copy_gui.setVisible(true);
  }

  //
  // Inner Class
  //
  /**
   * GUI displayed by CopyToHardDriveAction
   */
  private class CopyToHardDriveGUI extends JFrame implements GlassPaneListener {

    //
    // For enabling a glass pane
    //
    private GlassComponent gc = null;

    //
    // Indicates whether or not user cancelled the operation
    //
    private boolean is_cancelled = true;

    //
    // The log
    //
    private BufferedWriter log = null;

    //
    // Constructor
    //

    /**
     * Instantiates the {@link CopyToHardDriveGUI}. The constructor does not
     * make it visible.
     */
    public CopyToHardDriveGUI() {
      super();
      final InstallConfiguration ic = InstallConfiguration
          .getInstallConfiguration();

      //
      // Indicate to MetamorphoSysGUI to override config files with choices from
      // here
      //
      ic.setOverrideConfigurationFiles(true);

      //
      // Indicate that we are installing the full UMLS, not just Metathesaurus
      //
      ic.setInstallUMLS(true);

      //
      // disable welcome GUI
      //
      WelcomeGUI.getWelcomeGUI().enableGlassPane();

      //
      // Configure Copy Files to Hard Drive GUI
      //
      setTitle("Copy Files to Hard Drive");
      gc = new GlassComponent(this);
      setGlassPane(gc);
      getContentPane().setLayout(new BorderLayout());
      GridBagLayout gridbag = new GridBagLayout();
      JPanel choice_panel = new JPanel(gridbag);
      GridBagConstraints constraints = new GridBagConstraints();
      constraints.fill = GridBagConstraints.HORIZONTAL;
      constraints.insets = new Insets(2, 2, 2, 2);
      constraints.gridy = 0;
      constraints.gridx = 0;
      constraints.gridy++;

      //
      // Add header
      //
      constraints.gridy++;
      constraints.gridx = 2;
      constraints.gridwidth = 1;

      JTextPane text = new JTextPane();
      text.setEditable(false);
      text.setMinimumSize(new Dimension(200, 50));
      text.setPreferredSize(new Dimension(200, 50));
      text.setBackground(choice_panel.getBackground());
      SimpleAttributeSet set = new SimpleAttributeSet();
      StyleConstants.setAlignment(set, StyleConstants.ALIGN_CENTER);
      StyleConstants.setFontFamily(set, "Times");
      StyleConstants.setFontSize(set, 16);
      StyleConstants.setBold(set, true);
      text.setParagraphAttributes(set, true);
      text.setText("Copy Files To Hard Drive");

      choice_panel.add(text, constraints);
      constraints.gridy++;

      //
      // Add source directory: label, field, browse button
      //
      constraints.gridx = 0;
      choice_panel.add(new JLabel("Source"), constraints);
      constraints.gridx = 2;
      final JTextField source_field = new JTextField(40);
      source_field.setEditable(false);
      try {
        source_field.setText(new File("..").getCanonicalPath());
      } catch (Exception e) {
        // do nothing
      }
      source_field.setBackground(Color.white);
      gridbag.setConstraints(source_field, constraints);
      choice_panel.add(source_field);
      constraints.gridx = 3;
      final JButton source_browse = new JButton("Browse..");
      source_browse.setPreferredSize(new Dimension(100, 22));
      source_browse.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          try {
            //
            // Select a source directory
            //
            File selected_file = null;
            if (!source_field.getText().equals("")) {
              selected_file = new File(source_field.getText())
                  .getCanonicalFile();
            }
            File source_dir = MMSToolkit.chooseFile(
                "Select and Open the source directory", "Open..", ic
                    .getMetamorphoSysInputStream().getChooserType(),
                selected_file, selected_file, CopyToHardDriveGUI.this);

            //
            // If selected, configure IC
            //
            if (source_dir != null) {

              InstallConfiguration ic = InstallConfiguration
                  .getInstallConfiguration();
              source_field.setText(source_dir.getAbsolutePath());

              // also computes META, LEX, NET directories
              ic.setUmlsDirectory(source_dir);
            }
          } catch (Exception ex) {
            ex.printStackTrace();
            MMSToolkit.reportError(CopyToHardDriveGUI.this,
                "Unexpected error while setting source directory.\n"
                    + "See console for details.", false);
          }
        }
      });

      gridbag.setConstraints(source_browse, constraints);
      choice_panel.add(source_browse);

      //
      // Add source directory comment
      //
      constraints.gridy++;
      constraints.gridx = 2;
      constraints.insets = new Insets(0, 6, 10, 0);
      JLabel comment = new JLabel(
          "This directory contains .nlm files and mmsys.zip.");
      choice_panel.add(comment, constraints);
      constraints.insets = new Insets(2, 2, 2, 2);
      constraints.gridy++;

      //
      // Add destination directory: label, field, browse button
      //
      constraints.gridy++;
      constraints.gridx = 0;
      constraints.gridwidth = 1;
      choice_panel.add(new JLabel("Destination"), constraints);
      constraints.gridx = 2;
      final JTextField destination_field = new JTextField(40);
      destination_field.setEditable(false);
      destination_field.setBackground(Color.white);
      gridbag.setConstraints(destination_field, constraints);
      choice_panel.add(destination_field);
      constraints.gridx = 3;
      final JButton destination_browse = new JButton("Browse..");
      destination_browse.setPreferredSize(new Dimension(100, 22));
      destination_browse.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          try {
            File selected_file = null;

            if (!destination_field.getText().equals("")) {
              selected_file = new File(destination_field.getText())
                  .getCanonicalFile();
            } else {
              selected_file = new File(source_field.getText()).getParentFile();
            }

            File destination_dir = MMSToolkit.chooseFile(
                "Select and Open the destination directory", "Open..",
                JFileChooser.DIRECTORIES_ONLY, selected_file, selected_file,
                CopyToHardDriveGUI.this);
            if (destination_dir != null) {
              destination_field.setText(destination_dir.getAbsolutePath());
              ic.setUmlsDestinationDirectory(destination_dir);
            }
          } catch (Exception ex) {
            ex.printStackTrace();
            MMSToolkit.reportError(CopyToHardDriveGUI.this,
                "Unexpected error while setting destination directory.\n"
                    + "See console for details.", false);
          }
        }
      });

      gridbag.setConstraints(destination_browse, constraints);
      choice_panel.add(destination_browse);

      //
      // Add destination comment
      //
      constraints.gridy++;
      constraints.gridx = 2;
      constraints.insets = new Insets(0, 6, 0, 0);
      comment = new JLabel(
          "Select where to create the top-level UMLS directory.");
      choice_panel.add(comment, constraints);
      constraints.insets = new Insets(0, 6, 10, 0);
      constraints.gridy++;
      comment = new JLabel(
          "The directory name will correspond to the UMLS release, e.g.,  "
              + ic.getReleaseVersion().toUpperCase() + ".");
      choice_panel.add(comment, constraints);
      constraints.insets = new Insets(2, 2, 2, 2);
      constraints.gridy++;

      //
      // add Copy Button
      //
      constraints.gridx = 0;
      constraints.gridwidth = 1;
      constraints.weightx = 1;
      constraints.gridy++;
      JButton jinstall = new JButton("OK");
      jinstall.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {

          //
          // Indicate that "Cancel" was not pushed
          //
          is_cancelled = false;

          //
          // Disable the GUI
          //
          enableGlassPane();

          //
          // The installation should not happen in the Swing event handler
          // thread, so we put it off into a separate thread.
          //
          Thread t = new Thread(new Runnable() {
            public void run() {

              //
              // Get configuration information
              //
              InstallConfiguration ic = InstallConfiguration
                  .getInstallConfiguration();
              ic.setCopyToHardDriveChecked(true);

              File destination_dir = new File(destination_field.getText());

              //
              // Prepare variables
              //
              final String separator = System.getProperty("file.separator");
              final String unzip_cmd = System.getProperty("unzip.path");
              boolean mmsys_unzipped = false, nlm_copied = false;
              String mmsys_file = ic.getMetamorphoSysFile().getAbsolutePath();

              //
              // validate user selections
              //
              boolean validate_passed = validateChoices();
              if (!validate_passed) {
                disableGlassPane();
                return;
              }

              //
              // Open install.log file
              //
              File log_file = null;
              try {
                log_file = new File(ic.getUmlsDestinationDirectory()
                    + separator + ic.getReleaseVersion().toUpperCase(),
                    "install.log");
                log_file.getParentFile().mkdirs();
                log = new BufferedWriter(new FileWriter(log_file));

              } catch (Exception e) {
                e.printStackTrace();
                MMSToolkit.reportError(CopyToHardDriveGUI.this,
                    "Unexpected error trying to open the install.log file.\n"
                        + "See console for more details.", false);
              }
              try {
                log.write("Installation time:................................."
                    + new Date());
                log.newLine();
                log.write("Source directory:.................................."
                    + ic.getUmlsDirectory());
                log.newLine();
                log.write("Destination directory:............................."
                    + ic.getUmlsDestinationDirectory());
                log.newLine();

                //
                // Validate user selections
                //
                // Date sd = new Date();

                //
                // Copy .nlm files if indicated
                //
                if (ic.isCopyToHardDriveChecked()) {
                  log
                      .write("Copy NLM data format files (.nlm) to hard drive....");
                  Date start = new Date();

                  //
                  // Make the dest_dir/VERSION directory if does not exist
                  //
                  File copy_dest_dir = new File(destination_dir
                      .getAbsolutePath(), ic.getReleaseVersion().toUpperCase());
                  if (!copy_dest_dir.exists()) {
                    copy_dest_dir.mkdir();

                  }
                  File[] meta_files = ic.getMetathesaurusFiles();
                  boolean cancelled = false;
                  for (int i = 0; i < meta_files.length; i++) {
                    try {
                      cancelled = !SwingToolkit.copy(meta_files[i], new File(
                          copy_dest_dir.getAbsolutePath(), meta_files[i]
                              .getName()));
                      if (cancelled) {
                        log.write("NO (cancelled by user)");
                        log.newLine();
                        break;
                      }
                    } catch (IOException ioe) {
                      log.write("NO (error)");
                      log.newLine();
                      handleError(ioe, "Error copying "
                          + meta_files[i].getName() + ".");
                      return;
                    }
                  }

                  if (!cancelled) {
                    try {
                      cancelled = !SwingToolkit.copy(ic.getLexiconFile(),
                          new File(copy_dest_dir.getAbsolutePath(), ic
                              .getReleaseVersion().toLowerCase()
                              + "-otherks.nlm"));
                      if (cancelled) {
                        log.write("NO (cancelled by user");
                        log.newLine();
                      }
                    } catch (IOException ex) {
                      log.write("NO (error");
                      log.newLine();
                      handleError(ex, "Error copying " + ic.getLexiconFile());
                      return;
                    }

                    if (!cancelled) {
                      //
                      // Set directory containing .nlm files
                      //
                      ic.setUmlsDirectory(new File(destination_dir, ic
                          .getReleaseVersion().toUpperCase()));

                      nlm_copied = true;

                      Date end = new Date();
                      log.write("YES ("
                          + MMSToolkit.timeToString(MMSToolkit.timeDifference(
                              end, start)) + ")");
                      log.newLine();
                    }

                  }

                  if (cancelled) {
                    MMSToolkit.notifyUser(CopyToHardDriveGUI.this,
                        "Copying of .nlm files was cancelled.");
                    log.flush();
                    log.close();
                    disableGlassPane();
                    MMSToolkit.exitToWelcomeGUI(CopyToHardDriveGUI.this);
                    InstallConfiguration.getInstallConfiguration()
                        .clearProgressMonitor();
                    return;
                  }

                } else {
                  log
                      .write("Copy NLM data format files (.nlm) to hard drive....NO");
                  log.newLine();
                }

                //
                // unzip MMSYS if indicated
                //
                if (ic.isCopyToHardDriveChecked()) {
                  log
                      .write("Copy MetamorphoSys to hard drive...................");
                  Date start = new Date();
                  try {
                    // must use native unzip
                    mmsys_unzipped = SwingToolkit.unzip(unzip_cmd, mmsys_file,
                        destination_dir + separator
                            + ic.getReleaseVersion().toUpperCase(), "*",
                        CopyToHardDriveGUI.this);

                    if (!mmsys_unzipped) {
                      log.write("NO (cancelled by user)");
                      log.newLine();
                    } else {
                      ic.setMetamorphoSysConfigurationDirectory(new File(
                          new File(new File(destination_dir, ic
                              .getReleaseVersion().toUpperCase()), "MMSYS"),
                          "config"));

                      File chk_file = new File("..", ic.getReleaseVersion()
                          .toUpperCase()
                          + ".CHK");
                      if (chk_file.exists()) {
                        SwingToolkit.copy(chk_file, new File(destination_dir
                            + separator + ic.getReleaseVersion().toUpperCase(),
                            ic.getReleaseVersion().toUpperCase() + ".CHK"));
                      }
                      File md5_file = new File("..", ic.getReleaseVersion()
                          .toUpperCase()
                          + ".MD5");
                      if (md5_file.exists()) {
                        SwingToolkit.copy(md5_file, new File(destination_dir
                            + separator + ic.getReleaseVersion().toUpperCase(),
                            ic.getReleaseVersion().toUpperCase() + ".MD5"));
                      }

                      Date end = new Date();
                      log.write("YES ("
                          + MMSToolkit.timeToString(MMSToolkit.timeDifference(
                              end, start)) + ")");
                      log.newLine();
                    }
                  } catch (Exception e) {
                    log.write("NO (Error getting mmsys.zip path)");
                    log.newLine();
                    handleError(e, "Error copying MetamorphoSys.");
                    return;
                  }
                } else {
                  log
                      .write("Copy MetamorphoSys to hard drive...................NO");
                  log.newLine();
                }

                if (nlm_copied && mmsys_unzipped)
                  log
                      .write("\nApplication will close.  \nPlease run MetamorphoSys from "
                          + "newly installed installation at \n"
                          + ic.getUmlsDestinationDirectory() + ".");
                log.close();

                SwingToolkit.displayTextFile(CopyToHardDriveGUI.this,
                    "Copy to Hard Drive Log", log_file);

                if (nlm_copied && mmsys_unzipped)
                  System.exit(0);

                ic.clearProgressMonitor();
                dispose();
                WelcomeGUI.getWelcomeGUI().disableGlassPane();
              } catch (Exception e) {
                e.printStackTrace();
                MMSToolkit.reportError(CopyToHardDriveGUI.this,
                    "Unexpected error trying to open the install.log file.\n"
                        + "See console for more details.", false);
              }
            } // end public void run()

          }); // End of new Thread
          t.start();
        }
      }

      ); // end actionListener for install button

      //
      // Cancel button
      //
      // Create an action object
      Action cancelAction = new AbstractAction("Cancel") {
        {
          // Set action command to "Cancel"
          putValue(Action.ACTION_COMMAND_KEY, getValue(Action.NAME));
        }

        // This method is called when the action is invoked
        public void actionPerformed(ActionEvent e) {
          is_cancelled = true;
          MMSToolkit.exitToWelcomeGUI(CopyToHardDriveGUI.this);
          return;
        }
      };
      JButton jcancel = new JButton(cancelAction);

      //
      // Add OK, Cancel buttons
      //
      JPanel button_panel = new JPanel(new FlowLayout());
      // Register keystroke
      button_panel.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(
          KeyStroke.getKeyStroke("ESCAPE"), cancelAction.getValue(Action.NAME));
      // Register action
      button_panel.getActionMap().put(cancelAction.getValue(Action.NAME),
          cancelAction);
      jinstall.setPreferredSize(jcancel.getPreferredSize());
      button_panel.add(jinstall);
      button_panel.add(jcancel);

      //
      // Add ok and cancel button panel
      //
      constraints.gridx = 0;
      constraints.gridwidth = 4;
      constraints.weightx = 1;
      constraints.gridy++;
      choice_panel.add(button_panel, constraints);

      EtchedBorder etched_border = new EtchedBorder(EtchedBorder.RAISED);
      choice_panel.setBorder(new CompoundBorder(etched_border, new EmptyBorder(
          10, 10, 10, 10)));
      getContentPane().add(choice_panel, BorderLayout.CENTER);

      //
      // Prepare for display
      //
      FontSizeManager.adjustFontSize(this);
      this.pack();
      setLocationRelativeTo(null);

      //
      // Configure window listener
      //
      addWindowListener(new WindowAdapter() {
        public void windowClosing(WindowEvent e) {
          is_cancelled = true;
          MMSToolkit.exitToWelcomeGUI(CopyToHardDriveGUI.this);
          return;
        }
      });

    };

    /**
     * Indicates whether or not the installation was cancelled.
     * 
     * @return <code>true</code> if the GUI was cancelled, <code>false</code>
     *         otherwise
     */
    public boolean isCancelled() {
      return is_cancelled;
    }

    /**
     * Enable glass pane.
     */
    public void enableGlassPane() {
      gc.setVisible(true);
    }

    /**
     * Disable glass pane.
     */
    public void disableGlassPane() {
      gc.setVisible(false);
    }

    /**
     * Close this window and re-enable the {@link WelcomeGUI}.
     */
    public void dispose() {
      super.dispose();
    }

    //
    // Private Methods
    //
    /**
     * Validates the user selections for directories.
     * 
     * @return <code>true</code> if validation passes <code>false</code>
     *         otherwise
     */
    private boolean validateChoices() {

      //
      // validate all the fields
      //
      InstallConfiguration ic = InstallConfiguration.getInstallConfiguration();

      //
      // Create destination dir if not exists
      //
      if (ic.getUmlsDestinationDirectory() != null
          && !ic.getUmlsDestinationDirectory().exists()) {
        boolean confirm = MMSToolkit.confirmRequest(CopyToHardDriveGUI.this,
            "The destination directory does not exist.\n"
                + "Would you like to create it?");
        if (confirm) {
          ic.getUmlsDestinationDirectory().mkdirs();
        } else {
          MMSToolkit.notifyUser(CopyToHardDriveGUI.this,
              "Please select a destination directory that exists.");
          disableGlassPane();
          return false;
        }
      }

      //
      // Validate destination dir
      //
      if (ic.getUmlsDestinationDirectory() == null) {
        MMSToolkit.notifyUser(CopyToHardDriveGUI.this,
            "Please choose a destination directory.");
        disableGlassPane();
        return false;
      } else if (!ic.getUmlsDestinationDirectory().canWrite()) {
        MMSToolkit.notifyUser(CopyToHardDriveGUI.this,
            "The destination directory is write protected.\n"
                + "Please choose another destination directory.");
        disableGlassPane();
        return false;
      }

      String[] source_paths = new String[1];
      try {
        source_paths[0] = ic.getUmlsDirectory().getCanonicalPath();
      } catch (Exception e) {
        // do nothing
      }
      MetamorphoSysInputStream input_stream = new NLMFileMetamorphoSysInputStream();
      if (!input_stream.validateSourcePaths(source_paths)) {
        MMSToolkit.reportError(CopyToHardDriveGUI.this, input_stream
            .getValidationErrors()[0], false);
        return false;
      }

      return true;
    }

    /**
     * Handle error.
     * 
     * @param e
     *          the exception
     * @param msg
     *          the error message
     */
    private void handleError(Exception e, String msg) {
      MMSToolkit.printStackTrace(e, log);
      try {
        log.flush();
        log.close();
      } catch (Exception ex) {
        System.out.println("Unable to write to install.log.");
      }
      File log_file = new File(
          InstallConfiguration.getInstallConfiguration().getUmlsDestinationDirectory()
              + SEPARATOR
              + InstallConfiguration.getInstallConfiguration().getReleaseVersion().toUpperCase(),
          "install.log");
      SwingToolkit.displayTextFile(this, "UMLS Install Log", log_file);
      MMSToolkit.reportError(CopyToHardDriveGUI.this, msg, false);
      disableGlassPane();
      MMSToolkit.exitToWelcomeGUI(CopyToHardDriveGUI.this);
      InstallConfiguration.getInstallConfiguration().clearProgressMonitor();
      return;
    }

  } // End of CopyToHardDriveGUI class
}
