/**************************************************************************
 *
 * Package:     gov.nih.nlm.mms
 * Interface:   CuiSuiAuiList.java
 *
 **************************************************************************/
package gov.nih.nlm.mms;

import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.Concept;

import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

/**
 * This class is used during the first phase of {@link MetamorphoSys}
 * subsetting, in which a list of cuis to be kept is built up.
 * The class is designed to minimize the memory impact of keeping
 * track of the cuis.
 *
 * @author Deborah Shapiro, Brian Carlsen
 */
public class CuiSuiAuiList implements CuiList {

  //
  // Private Fields
  //

  private Set sui_full_cuis = new HashSet(14000);
  private Set full_cuis = new HashSet(14000);
  private Map sui_keep_only = new HashMap(14000);
  private Map sui_keep_all_except = new HashMap(14000);
  private Map aui_keep_only = new HashMap(14000);
  private Map aui_keep_all_except = new HashMap(14000);
  private Set user_keep_cui = new HashSet();
  private Set user_remove_cui = new HashSet();
  private Set aui_set = new HashSet(1000000);
  private String unfiltered_cui;
  private int[] unfiltered_suis = new int[0];
  private int[] unfiltered_auis = new int[0];

  //
  // Constructor
  //

  /**
   * Instantiates an empty {@link CuiSuiAuiList}
   */
  public CuiSuiAuiList() {}

  //
  // Methods
  //

  /**
   * Set the initial state of the {@link Concept} before filtering.
   * This is for later comparison purposes.
   * @param concept an unfiltered {@link Concept}
   */
  public void setUnfilteredConcept(Concept concept) {
    unfiltered_cui = concept.getCui();
    final Atom[] atoms = concept.getAtoms();
    unfiltered_suis = getSuiList(atoms);
    unfiltered_auis = getAuiList(atoms);
  }

  /**
   * Set the final state of the {@link Concept} after filtering.
   * Compare this against the unfiltered {@link Concept}, and determine what
   * was removed, what was kept, and how to track that information
   * in the least memory-intensive way.
   * @param concept a filtered {@link Concept}
   */
  public void setFilteredConcept(Concept concept) {

    final String filtered_cui = new String(concept.getCui());
    final Atom[] atoms = concept.getAtoms();

    // The filtered concept must match the unfiltered concept
    if (!filtered_cui.equals(unfiltered_cui))
      throw new IllegalArgumentException(
	  "Unfiltered CUI does not match filtered CUI.");

    if (unfiltered_auis.length > 0) {
      int[] filtered_auis = getAuiList(atoms);
      Arrays.sort(filtered_auis);
      Arrays.sort(unfiltered_auis);
      if (unfiltered_auis.length == filtered_auis.length) {
        full_cuis.add(filtered_cui);
      } else if(unfiltered_auis.length > filtered_auis.length*2) {
        // most of the uis have been filtered out
        aui_keep_only.put(filtered_cui, filtered_auis);
      } else {
        // few of the uis have been filtered out
        int [] aui_keep_except =
          new int[unfiltered_auis.length - filtered_auis.length];
        int i = 0;
        for (int u = 0, f = 0;
             (u < unfiltered_auis.length) && (f < filtered_auis.length);
             u++) {
          if (unfiltered_auis[u] != filtered_auis[f]) {
            aui_keep_except[i++] = unfiltered_auis[u];
          } else {
            // make sure that remainder of unfiltered_uis end up in
            // keep_except after filtered_suis are exhausted
            if(f<filtered_auis.length-1) f++;
          }
        }
        aui_keep_all_except.put(filtered_cui, aui_keep_except);
      }
      // put filtered auis into aui_set
      for (int m=0; m<filtered_auis.length; m++) {
      	aui_set.add(new Integer(filtered_auis[m]));
      }
    }


    // Get the SUIs from the CUI as an int array.
    final int[] filtered_suis = getSuiList(atoms);

    // Sort the filtered and unfiltered SUIs
    Arrays.sort(filtered_suis);
    Arrays.sort(unfiltered_suis);

    if (unfiltered_suis.length == filtered_suis.length &&
        !full_cuis.contains(filtered_cui)) {
      sui_full_cuis.add(filtered_cui);
    } else if(unfiltered_suis.length > filtered_suis.length*2) {
      // most of the uis have been filtered out
      sui_keep_only.put(filtered_cui, filtered_suis);
    } else if(unfiltered_suis.length != filtered_suis.length ) {
      // few of the uis have been filtered out
      int [] sui_keep_except =
	new int[unfiltered_suis.length - filtered_suis.length];
      int i = 0;
      for (int u = 0, f = 0;
	   (u < unfiltered_suis.length) && (f < filtered_suis.length);
	   u++) {
	if (unfiltered_suis[u] != filtered_suis[f]) {
	  sui_keep_except[i++] = unfiltered_suis[u];
	} else {
	  // make sure that remainder of unfiltered_uis end up in
	  // keep_except after filtered_suis are exhausted
	  if(f<filtered_suis.length-1) f++;
	}
      }
      sui_keep_all_except.put(filtered_cui, sui_keep_except);
    }

  }

  /**
   * Returns <code>true</code> if any part of the {@link Concept} is being kept.
   * @param cui a {@link String} cui value
   * @return <code>true</code> if any part of the {@link Concept} is being kept
   */
  public boolean inList(String cui) {
    final String ikey = cui;

    return !user_remove_cui.contains(ikey) &&
    (full_cuis.contains(ikey) ||
      aui_keep_only.containsKey(ikey) ||
      aui_keep_all_except.containsKey(ikey) ||
      sui_full_cuis.contains(ikey) ||
      sui_keep_only.containsKey(ikey) ||
      sui_keep_all_except.containsKey(ikey) ||
      user_keep_cui.contains(ikey));
  }

  /**
   * Returns <code>true</code> if the cui,sui,aui combination is being kept.
   * @param cui a {@link String} cui value
   * @param sui a {@link String} sui value
   * @param aui a {@link String} aui value
   * @return <code>true</code> if the cui,sui,aui combination is being kept
   */
  public boolean inList(String cui, String sui, String aui) {

    //
    // Verify parameters
    //
    if ((sui == null || sui.equals("")) && (aui == null || aui.equals("")))
      throw new IllegalArgumentException("Either sui or aui must be set.");
    
    //
    // null cui and sui indicates that this is to check only for aui
    //
    if ((cui == null || cui.equals("")) && (sui == null || sui.equals(""))) {
    	if (aui_set.contains(new Integer(aui.substring(1))))
    		return true;
    	else 
    		return false;
    }
    	
    //
    // Get CUI key
    //
    final String ikey = cui;

    //
    // Check for explicit user remove/add
    //
    if(user_remove_cui.contains(ikey))
      return false;

    if(user_keep_cui.contains(ikey))
      return true;

    //
    // Check if full CUI was kept
    //
    if(full_cuis.contains(ikey))
      return true;

    //
    // Look for matching SUI, AUI
    //
    int this_sui=0, this_aui=0;
    boolean return_value;

    //
    // Check AUI based lookup
    //
    if (aui != null && !aui.equals("")) {
      this_aui = Integer.parseInt(aui.substring(1));
      if (aui_keep_only.containsKey(ikey)) {
        int[] auis = (int [])aui_keep_only.get(ikey);
        return_value = Arrays.binarySearch(auis, this_aui) >= 0;
        return return_value;
      } else if(aui_keep_all_except.containsKey(ikey)) {
        int[] auis = (int [])aui_keep_all_except.get(ikey);
        return_value = Arrays.binarySearch(auis, this_aui) < 0;
        return return_value;
      }

    }

    //
    // Handle SUI based lookup
    //
    else if (sui != null && !sui.equals("")) {
      if(sui_full_cuis.contains(ikey))
        return true;
      this_sui = Integer.parseInt(sui.substring(1));
      if (sui_keep_only.containsKey(ikey)) {
        int[] suis = (int [])sui_keep_only.get(ikey);
        return_value = Arrays.binarySearch(suis, this_sui) >= 0;
        return return_value;
      } else if(sui_keep_all_except.containsKey(ikey)) {
        int[] suis = (int [])sui_keep_all_except.get(ikey);
        return_value = Arrays.binarySearch(suis, this_sui) < 0;
        return return_value;
      }
    }

    //
    // Nothing found, not in list
    //
    return false;
  }


  /**
   * Helper method to take a {@link Concept} and
   * obtain all of the SUIs in its constituent {@link Atom}s
   * and return them (without the leading S) as an <code>int []</code>.
   * @param atoms {@link Atom}[]
   * @return SUIs in the given {@link Concept}
   */
  private int[] getSuiList(Atom[] atoms) {
    final Set sui_set = new HashSet(atoms.length*2);
    for (int i = 0; i < atoms.length; i++) {
        if (atoms[i].getSui() != null && !atoms[i].getSui().equals("")) {
          sui_set.add(new String(atoms[i].getSui()));
        }
    }

    final int[] suis = new int[sui_set.size()];
    final Iterator iter = sui_set.iterator();
    int i = 0;
    while (iter.hasNext()) {
      final String next_string = (String)iter.next();
      suis[i++] = Integer.parseInt(next_string.substring(1));
    }
    return suis;

  }

  /**
   * Helper method to take a {@link Concept} and
   * obtain all of the AUIs in its constituent {@link Atom}s
   * and return them (without the leading A) as an <code>int []</code>.
   * @param atoms {@link Atom}[]
   * @return AUIs in the given {@link Concept}
   */
  private int[] getAuiList(Atom[] atoms) {
    final Set l_aui_set = new HashSet(atoms.length*2);
    for (int i = 0; i < atoms.length; i++) {
       if (atoms[i].getAui() != null && !atoms[i].getAui().equals("")) {
          l_aui_set.add(new String(atoms[i].getAui()));
       }
    }

    int[] auis = new int[l_aui_set.size()];
    final Iterator iter = l_aui_set.iterator();
    int i = 0;
    while (iter.hasNext()) {
      final String next_string = (String)iter.next();
      auis[i++] = Integer.parseInt(next_string.substring(1));
    }
    return auis;

  }
  /**
   * Enables a filter to specify a Cui should be added to the cui list.
   * @param cui {@link String}
   */
  public void addCui(String cui) {
    final String l_cui = new String(cui);
    user_keep_cui.add(l_cui);
    user_remove_cui.remove(l_cui);
  }

  /**
   * Indicates that cui was added by the user and not just left
   * through filtering process.
   * @param cui {@link String}
   * @return <code>boolean</code>
   */
  public boolean addedByUserOnly(String cui) {
    final String ikey = cui;
    return user_keep_cui.contains(ikey) &&
           !full_cuis.contains(ikey) &&
           !aui_keep_all_except.containsKey(ikey) &&
           !aui_keep_only.containsKey(ikey) &&
           !sui_full_cuis.contains(ikey) &&
           !sui_keep_all_except.containsKey(ikey) &&
           !sui_keep_only.containsKey(ikey);
  }

  /**
   * Enables a filter to specify a Cui that should be removed from the
   * cui list.
   * @param cui {@link String}
   */
  public void removeCui(String cui) {
    final String l_cui = new String(cui);
    user_keep_cui.remove(l_cui);
    user_remove_cui.add(l_cui);
  }
  
  
  /**
   * Clears all references to the aui_set so that it can be
   * garbage collected.
   */
  public void clearAuiSet() {
  	aui_set.clear();
  	aui_set = null;
  }
  
  /** Clears all refereces to the cui_list components so {@link CuiSuiAuiList}
   * can be garbage collected.
   */
  public void clear() {
  	full_cuis.clear();
  	full_cuis = null;
  	aui_keep_all_except.clear();
  	aui_keep_all_except = null;
  	aui_keep_only.clear();
  	aui_keep_only = null;
  	sui_full_cuis.clear();
  	sui_full_cuis = null;
  	sui_keep_all_except.clear();
  	sui_keep_all_except = null;
  	sui_keep_only.clear();
  	sui_keep_only = null;	
  }
  
  /**
   * Dumps size of the {@link CuiSuiAuiList} for debugging purposes.
   * @return dump string
   */
  protected String dump() {
    return
    "full_cuis " + full_cuis.size() + "\n" +
    "aui_keep_all_except " + aui_keep_all_except.size() + "\n" +
    "aui_keep_only " + aui_keep_only.size() + "\n" +
    "sui_full_cuis " + sui_full_cuis.size() + "\n" +
    "sui_keep_all_except " + sui_keep_all_except.size() + "\n" +
    "sui_keep_only " + sui_keep_only.size();
  }
}
