/**************************************************************************
 *
 * Package:    gov.nih.nlm.mms;
 * Interface:  DefaultSubsetSelectionGUI.java
 *
 *************************************************************************/
package gov.nih.nlm.mms;

import gov.nih.nlm.swing.FontSizeManager;
import gov.nih.nlm.swing.SortableJTable;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.ArrayList;
import java.util.List;

import javax.swing.*;
import javax.swing.border.CompoundBorder;
import javax.swing.border.EmptyBorder;
import javax.swing.border.EtchedBorder;
import javax.swing.border.TitledBorder;
import javax.swing.event.TableModelEvent;
import javax.swing.table.AbstractTableModel;

/**
 * Dialog for user to select the {@link DefaultSubset}.
 * 
 * @author Deborah Shapiro
 */

public class DefaultSubsetSelectionGUI extends JDialog implements
    ActionListener {

  //
  // Private Fields
  //
  private DSTableModel dsmodel;

  private SortableJTable default_subsets_table;

  private ApplicationConfiguration app_config;

  private Frame owner;

  public static final java.awt.Cursor wait_cursor = new java.awt.Cursor(
      java.awt.Cursor.WAIT_CURSOR);

  public static final java.awt.Cursor default_cursor = new java.awt.Cursor(
      java.awt.Cursor.DEFAULT_CURSOR);

  public static final MouseListener NULL_MOUSE_LISTENER = new MouseAdapter() {
  };

  private List default_subsets = new ArrayList();

  private DefaultSubset default_subset;

  private boolean is_cancelled = true;

  //
  // Constructor
  //

  /**
   * Instantiates a {@link DefaultSubsetSelectionGUI} from the specified
   * configuration objects.
   * 
   * @param app_config
   *          the {@link ApplicationConfiguration}
   * @param user_config
   *          the {@link UserConfiguration}
   * @param owner {@link Frame}
   */
  public DefaultSubsetSelectionGUI(ApplicationConfiguration app_config,
      UserConfiguration user_config, Frame owner) {
    super(owner);
    this.app_config = app_config;
    this.owner = owner;
    setModal(true);
    setTitle("Select Default Subset");
    initialize();
  };

  //
  // Methods
  //

  /**
   * This method sets up the GUI
   */
  private void initialize() {
    dsmodel = new DSTableModel();

    //
    // Compose menu bar
    //
    JMenuBar menu_bar = new JMenuBar();
    setJMenuBar(menu_bar);
    menu_bar.add(Box.createHorizontalGlue());
    JMenu help_menu = new JMenu("Help");
    menu_bar.add(help_menu);

    //
    // Install UMLS Help Menu Item
    //
    help_menu.add(MMSToolkit.getResourceHelpMenu("Default Subset Help",
        "config/DefaultSubsetHelp.html", owner));

    default_subsets = app_config.getDefaultSubsetsAsList();
    default_subsets_table = new SortableJTable(dsmodel);
    default_subsets_table.getColumnModel().getColumn(0).setPreferredWidth(100);
    default_subsets_table.getColumnModel().getColumn(1).setPreferredWidth(400);

    getContentPane().setLayout(new BorderLayout());
    JPanel table_panel = new JPanel(new GridBagLayout());
    JPanel complete_panel = new JPanel(new BorderLayout());
    GridBagConstraints constraints = new GridBagConstraints();
    constraints.fill = GridBagConstraints.BOTH;
    constraints.insets = new Insets(2, 2, 2, 2);
    constraints.gridx = 0;
    constraints.gridwidth = 1;
    constraints.weightx = 1.0;

    // configure table_panel
    TitledBorder titledBorder = new TitledBorder(new EtchedBorder(
        EtchedBorder.RAISED), "Select Default Subset");
    titledBorder.setTitleJustification(TitledBorder.CENTER);
    table_panel.setBorder(new CompoundBorder(titledBorder, new EmptyBorder(10,
        10, 10, 10)));

    constraints.gridy++;

    String message = "<html><b>"
        + "Select one of the default Metathesaurus subsets below.  The default subset, Level O, includes<br>"
        + "only those sources for which no additional license agreements are necessary. Note: You may later<br>"
        + "use the Source List tab to customize your subset, to include or exclude specific vocabularies.<b></html>";

    JLabel jds_label = new JLabel(message);
    table_panel.add(jds_label, constraints);

    constraints.gridy++;
    default_subsets_table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

    default_subsets_table.setRowSelectionInterval(0, 0);

    JScrollPane jscroll = new JScrollPane(default_subsets_table);
    jscroll.setPreferredSize(new Dimension(500, 100));

    table_panel.add(jscroll, constraints);
    constraints.gridy++;

    //
    // Configure Buttons
    //
    JPanel btn_panel = new JPanel();
    JButton jdone = new JButton("Done");
    jdone.addActionListener(this);
    btn_panel.add(jdone);

    // Create an action object
    Action cancelAction = new AbstractAction("Cancel") {
      {
        // Set action command to "Cancel"
        putValue(Action.ACTION_COMMAND_KEY, getValue(Action.NAME));
      }
      // This method is called when the action is invoked
      public void actionPerformed(ActionEvent e) {
        cancelActionPerformed(e);
      }
    };
    JButton jcancel = new JButton(cancelAction);
    btn_panel.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(
        KeyStroke.getKeyStroke("ESCAPE"), cancelAction.getValue(Action.NAME));
    // Register action
    btn_panel.getActionMap().put(cancelAction.getValue(Action.NAME),
        cancelAction);
    btn_panel.add(jcancel);
    jdone.setPreferredSize(jcancel.getPreferredSize());

    // Add the sub-panels to complete_panel
    complete_panel.add(table_panel, BorderLayout.CENTER);
    complete_panel.add(btn_panel, BorderLayout.SOUTH);
    EtchedBorder etched_border = new EtchedBorder(EtchedBorder.RAISED);
    complete_panel.setBorder(new CompoundBorder(etched_border, new EmptyBorder(
        10, 10, 10, 10)));

    getContentPane().add(complete_panel, BorderLayout.CENTER);
    getRootPane().setDefaultButton(jdone);
    FontSizeManager.adjustFontSize(this);
    this.pack();
    jdone.requestFocus();
    setLocationRelativeTo(null);

    // Configure window listener
    addWindowListener(new WindowAdapter() {
      public void windowClosing(WindowEvent e) {
        InstallConfiguration.getInstallConfiguration().clearProgressMonitor();
        MMSToolkit.cancel();
      }
    });

  };

  /**
   * Returns the {@link DefaultSubset} selected by the user.
   * 
   * @return the {@link DefaultSubset} selected by the user
   */
  public DefaultSubset getSelectedDefaultSubset() {
    return default_subset;
  }

  /**
   * Indicates whether or not the user clicked the "Cancel" button.
   * 
   * @return <code>true</code> if so, <code>false</code> otherwise
   */
  public boolean isCancelled() {
    return is_cancelled;
  }

  /**
   * Used by locally defined Action for cancel.
   *
   * @param e a "Cancel" action event
   */
  private void cancelActionPerformed(ActionEvent e) {
    actionPerformed(e);
  }

  /**
   * Handle user selections.
   *
   * @param e
   *          an action event from a user selection
   */
  public void actionPerformed(ActionEvent e) {

    //
    // Handle "Done" button
    //
    if (e.getActionCommand().equals("Done")) {

      //
      // Validate selection
      //
      int row = default_subsets_table.getSelectedRow();
      if (row == -1) {
        MMSToolkit.notifyUser(DefaultSubsetSelectionGUI.this,
            "You did not choose a default subset.\n" + "Please select a row.");
        return;
      }

      //
      // Identify selection and set it
      //
      String default_subset_name = (String) dsmodel.getValueAt(row, 0);
      for (int i = 0; i < default_subsets.size(); i++) {
        DefaultSubset ds = ((DefaultSubset) default_subsets.get(i));
        if (ds.getName().equals(default_subset_name)) {
          default_subset = ds;
          break;
        }
      }

      //
      // Inidcate no cancel
      //
      is_cancelled = false;

      //
      // Dispose dialog
      //
      dispose();
    }

    //
    // Handle "Cancel" button
    //
    if (e.getActionCommand().equals("Cancel")) {
      is_cancelled = true;
      //
      // Go back to WelcomeGUI, unless MetamorphoSysGUI is open.
      //
      if (WelcomeGUI.getWelcomeGUI().isVisible())
        InstallConfiguration.getInstallConfiguration().clearProgressMonitor();
      dispose();
    }

  }

  /**
   * Table model for the GUI table.
   */
  private class DSTableModel extends AbstractTableModel {

    //
    // Fields
    //
    final String[] columnNames = { "Name", "Description" };

    /**
     * Instantiates an empty DSTableModel.
     */
    public DSTableModel() {
    }

    /**
     * Refresh the table.
     */
    public void refresh() {
      fireTableChanged(new TableModelEvent(DSTableModel.this, 0, getRowCount(),
          TableModelEvent.ALL_COLUMNS, TableModelEvent.INSERT));
    }

    /**
     * Returns the column count.
     * 
     * @return the column count
     */
    public int getColumnCount() {
      return columnNames.length;
    }

    /**
     * Returns the row count.
     * 
     * @return the row count
     */
    public int getRowCount() {
      return default_subsets.size();
    }

    /**
     * Returns the name for the specified column.
     * 
     * @param col
     *          the specified column
     * @return the name for the specified column
     */
    public String getColumnName(int col) {
      return columnNames[col];
    }

    /**
     * Returns the value at the specified row and column.
     * 
     * @param row
     *          the specified row
     * @param col
     *          the specified column
     * @return the value at the specified row and column
     */
    public Object getValueAt(int row, int col) {
      DefaultSubset ds = (DefaultSubset) (default_subsets.get(row));
      if (col == 0)
        return ds.getName();
      else
        return ds.getDescription();
    }

    /**
     * Indicates whether or not the table cell at the specified row and column
     * is editable.
     * 
     * @param row
     *          the specified row
     * @param col
     *          the specified column
     * @return <code>false</code>
     */
    public boolean isCellEditable(int row, int col) {
      return false;
    }

    /**
     * Returns the class representing the type of data in this column.
     * 
     * @param c
     *          the specified column
     * @return the class representing the type of data in this column
     */
    public Class getColumnClass(int c) {
      return String.class;
    }

  } // End DSTableModel

}// End of DefaultSubsetSelectionGUI class

