/**************************************************************************
 *
 * Package:	gov.nih.nlm.mms
 * Object:	Dependency.java
 *
 **************************************************************************/
package gov.nih.nlm.mms;

import gov.nih.nlm.umls.meta.Source;

/**
 * Represents a dependency between two sources.  Used to connect sources
 * together logically to assist in subsetting.
 *
 * @author Deborah Shapiro, Brian Carlsen
 */
public class Dependency implements Comparable {

  //
  // The first source
  //
  private Source source;

  //
  // The dependent source.
  //
  private Source dep_source;

  //
  // Constructors
  //

  /**
   * Instantiates an empty {@link Dependency}.
   */
  public Dependency() {};

  /**
   * Instantiates a {@link Dependency} between
   * the two specified {@link Source}s.
   * @param source a {@link Source}
   * @param dep_source the dependent {@link Source}
   */
  public Dependency(Source source, Source dep_source) {
    setSource(source);
    setDependentSource(dep_source);
  }

  //
  // Methods
  //

  /**
   * Return a hash code based on the source abbreviations.
   * @return a hash code based on the source abbreviations
   */
  public int hashCode() {
    return (source.getVersionedSourceAbbreviation() +
	    dep_source.getVersionedSourceAbbreviation()).hashCode();
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be a
   * {@link Dependency} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if(o == null || !(o instanceof Dependency)) return false;
    Dependency dependency = (Dependency)o;
    return dependency.getSource().equals(this.getSource()) &&
      dependency.getDependentSource().equals(this.getDependentSource());
  }

  /**
   * Compares this {@link Dependency} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {

    if (o instanceof Dependency) {
      int i = source.getVersionedSourceAbbreviation().compareTo((
		(Dependency)o).getSource().getVersionedSourceAbbreviation());
      if(i==0)
	return dep_source.getVersionedSourceAbbreviation().compareTo(
	    ((Dependency)o).getDependentSource().getVersionedSourceAbbreviation());
      else
	return i;
    } else { return 0; }
  }

  /**
   * Returns a {@link String} representation of the dependency.
   * @return a {@link String} representation of the dependency
   */
  public String toString() {
    return(source.getRootSourceAbbreviation() + "|" +
	   source.getSourceFamily() + ":" +
	   dep_source.getRootSourceAbbreviation() + "|" +
	   dep_source.getSourceFamily());
  }

  //
  // Accessor Methods
  //

  /**
   * Returns the {@link Source} object being depended upon.
   * @return the {@link Source} object being depended upon.
   */
  public Source getSource() {
    return source;
  }

  /**
   * Sets the {@link Source} being depended upon.
   * @param source the {@link Source} being depended upon.
   */
  public void setSource(Source source) {
    this.source = source;
  }

  /**
   * Returns the dependent {@link Source}.
   * @return the dependent {@link Source}
   */
  public Source getDependentSource() {
    return dep_source;
  }

  /**
   * Sets the dependent {@link Source}.
   * @param dep_source the dependent {@link Source}
   */
  public void setDependentSource(Source dep_source) {
    this.dep_source = dep_source;
  }

}

