/**************************************************************************
 *
 * Package:    gov.nih.nlm.mms;
 * Interface:  GUIConfigurable.java
 *
 *************************************************************************/
package gov.nih.nlm.mms;

import gov.nih.nlm.swing.DataChangeListener;

import java.awt.Color;

import javax.swing.JPanel;

/**
 * Generically represents an object that can be configured by
 * the {@link MetamorphoSysGUI}.
 *
 * @author Deborah Shapiro
 *
 **/

public interface GUIConfigurable extends Configurable, Comparable {

  /**
   * Returns a {@link JPanel} containing configuration options.
   * @return a {@link JPanel} containing configuration options
   */
  public JPanel getPanel();

  /**
   * Indicates whether or not the configuration information has changed.
   * @return <code>true</code> if data has changed;
   *         <code>false</code> otherwise
   */
  public boolean hasDataChanged();

  /**
   * Sets the value of the data changed flag.
   * @param b the <code>boolean</code> value for the data changed flag
   */
  public void setDataChanged(boolean b);



  /**
   * Sets the data changed flag to <code>false</code>.
   */
  public void resetDataChanged();

  /**
   * Adds a {@link DataChangeListener} that responds to configuration changes.
   * Allows main GUI to record all changes to
   * this gui in the undo/redo mechanism.
   * @param dcl the {@link DataChangeListener} to add
   */
  public void addDataChangeListener(DataChangeListener dcl);

  /**
   * Removes the specified {@link DataChangeListener}.
   * @param dcl the {@link DataChangeListener} to remove
   */
  public void removeDataChangeListener(DataChangeListener dcl);

  /**
   * Returns the {@link Color} that should be used to display
   * the name.
   * @return the {@link Color} that should be used to display the name.
   */
  public Color getColor();

  /**
   * Returns the name.
   * @return the name
   */
  public String getName();

  /**
   * Returns the description to be used in conjunction with the panel
   * when presented to the user.
   * @return the description
   */
  public String getTabDescription();


  /**
   * Presents an advanced options GUI to the user.
   */
  public void showOptions();

  /**
   * Indicates whether or not an options menu item should be supplied
   * for advanced configuration options.
   * @return <code>true</code> if so; <code>false</code> otherwise
   */
  public boolean hasAdvancedOptions();

  /**
   * Returns a {@link JPanel} that will be displayed when the
   * user selects the corresponding menu item from the Help menu
   * and will provide the user with detailed instructions on the tab's use.
   * @return the {@link JPanel} to be used when showing user help
   */
  public JPanel getHelpPanel();

  /**
   * Updates the view to reflect the data model.
   */
  public void setPanelSelections();

  /**
   * Indicates whether or not the validation rules have passed.
   * @return <code>boolean</code><code>true</code> if so;
   * 	     <code>false</code> otherwise
   */
  public boolean validate();

  /**
   * Returns a message indicating why validation failed.  Should only be
   * called if {@link #validate()} returns <code>false</code>.
   * @return the message
   * deprecated
   */
  //public String getValidateMessage();

  /**
   * Returns the messages indicating why validation failed.  Should only be
   * called if {@link #validate()} returns <code>false</code>.
   * @return the failure messages
   */
  public String[] getValidationErrors();

  /**
   * Returns warnings that were indicated during validation.
   * @return the warning messages
   */
  public String[] getValidationWarnings();

  /**
   * Returns the {@link GUIConfigurable} whose tab should next be displayed
   * to allow the user to correct an error or warning.
   * @return GUIConfigurable to be displayed
   */
  public GUIConfigurable getGUIConfigurableForValidationFailure();
}

