/*****************************************************************************
 *
 * Package:    gov.nih.nlm.mms
 * Object:     InstallConfiguration.java
 *
 *****************************************************************************/
package gov.nih.nlm.mms;

import gov.nih.nlm.swing.SwingToolkit;
import gov.nih.nlm.util.SystemToolkit;

import java.awt.Component;
import java.awt.Frame;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.zip.GZIPInputStream;

/**
 * Represents the user's configuration choices up to the point where
 * a configuration file is loaded and {@link MetamorphoSysGUI} is instantiated.
 * It is mostly used in {@link InstallationGUI} and {@link SubsetGUI}.
 *
 * @author Deborah Shapiro
 */
public class InstallConfiguration  {

  //
  // Singleton reference to iteself
  //
  private static InstallConfiguration ic = null;

  //
  // File variables
  //
  private static File destination_meta_dir;
  private static File destination_umls_dir;
  private static File[] metathesaurus_files;
  private static File umls_dir;
  private static File metamorphosys_file;
  private static File metamorphosys_config_dir;
  private static File semantic_network_file;
  private static File lexicon_file;

  //
  // Boolean variables
  //
  private static boolean metathesaurus_checked = true;
  private static boolean metamorphosys_checked = false;
  private static boolean semantic_network_checked = true;
  private static boolean lexicon_checked = true;
  private static boolean lvg_checked = false;
  private static boolean new_config_file = true;
  private static boolean query_to_open = true;

  //
  // Release version (of code/data)
  //
  private static String release_version;

  //
  // Should settings override user MMSYS configuration files
  //
  private static boolean override_config = false;

  //
  // Indicates if user chose Install on the WelcomeGUI.
  // Otherwise the user chose Configure Subset
  //
  private static boolean install_umls = true;

  //
  // Progress monitor
  //
  private static MMSProgressMonitor progress_monitor = null;

  //
  // install.log
  //
  private static BufferedWriter log = null;

  private static final String SEPARATOR = System.getProperty("file.separator");

  /**
   * Instantiates {@link InstallConfiguration}.
   */
  private InstallConfiguration() {
    initialize();
  }

  /**
   * Reset the values.
   */
  protected void initialize() {

    //
    // Set boolean defaults
    //
    metathesaurus_checked = true;
    semantic_network_checked = true;
    lexicon_checked = true;
    lvg_checked = false;
    metamorphosys_checked = false;
    override_config = false;
    install_umls = true;

    destination_meta_dir = null;
    destination_umls_dir = null;
    metathesaurus_files = null;
    umls_dir = null;
    metamorphosys_file = null;
    metamorphosys_config_dir = null;
    semantic_network_file = null;
    lexicon_file = null;

    progress_monitor = null;

    //
    // Set release version
    //
    setReleaseVersion(MMSToolkit.getReleaseVersion());

    //
    // Set File defaults
    //
    File source_dir = null;
    try {
       source_dir = new File("..").getCanonicalFile();
    } catch (IOException ioe) {
      source_dir = new File("..");
    }
    setUmlsDirectory(source_dir);
    setMetamorphoSysConfigurationDirectory(new File("config").getAbsoluteFile());

  };

  /**
   * Returns the singleton reference to the {@link InstallConfiguration}.
   * @return the singleton reference to the {@link InstallConfiguration}
   */
  public static InstallConfiguration getInstallConfiguration() {
    if (ic == null)
      ic = new InstallConfiguration();
    return ic;
  }

  /**
   * Returns whether or not the user is running with a new configuration file.
   * @return <code>true</code> if a new configuration is being made
   *         <code>false</code> if a saved configuration is being used
   */
  public boolean isNewConfiguration() {
    return new_config_file;
  }

  /**
   * Sets whether or not the user is running with a new configuration file.
   * @param new_config_file <code>true</code> if a new configuration is being made
   *         <code>false</code> if a saved configuration is being used
   */
  public void setNewConfiguration(boolean new_config_file) {
    InstallConfiguration.new_config_file = new_config_file;
  }

  /**
   * Returns whether or not the semantic network has been chosen
   * for installation.
   * @return <code>true</code> if semantic network is chosen
   *         <code>false</code> otherwise
   */
  public boolean isSemanticNetworkChecked() {
    return semantic_network_checked;
  }

  /**
   * Sets whether or not the semantic network has been chosen
   * for installation.
   * @param checked according to selection of semantic network
   */
  public void setSemanticNetworkChecked(boolean checked) {
    semantic_network_checked = checked;
  }

  /**
   * Returns the {@link File} that contains the semantic network
   * data.
   * @return {@link File} containing semantic network data
   */
  public File getSemanticNetworkFile() {
    return semantic_network_file;
  }

  /**
   * Sets the {@link File} that contains the semantic network data.
   * @param file {@link File} containing semantic network data
   */
  public void setSemanticNetworkFile(File file) {
    semantic_network_file = file;
  }


  /**
   * Returns whether or not the lexicon has been chosen
   * for installation.
   * @return <code>true</code> if lexicon is chosen
   *         <code>false</code> otherwise
   */
  public boolean isLexiconChecked() {
    return lexicon_checked;
  }

  /**
   * Sets whether or not the lexicon has been chosen
   * for installation.
   * @param checked according to selection of lexicon
   */
  public void setLexiconChecked(boolean checked) {
    lexicon_checked = checked;
  }
  
  /**
   * Returns whether or not LVG has been chosen
   * for installation.
   * @return <code>true</code> if LVG is chosen
   *         <code>false</code> otherwise
   */
  public boolean isLvgChecked() {
    return lvg_checked;
  }

  /**
   * Sets whether or not LVG has been chosen
   * for installation.
   * @param checked according to selection of LVG
   */
  public void setLvgChecked(boolean checked) {
    lvg_checked = checked;
  }
  
  /**
   * Returns the {@link File} that contains the lexicon
   * data.
   * @return {@link File} containing lexicon data
   */
  public File getLexiconFile() {
    return lexicon_file;
  }

  /**
   * Sets the {@link File} that contains the lexicon data.
   * @param file {@link File} containing lexicon data
   */
  public void setLexiconFile(File file) {
    lexicon_file = file;
  }

  /**
   * Indicates whether or not settings from this object should override
   * the configuration files
   * @return <code>true</code> if semantic network is chosen
   *         <code>false</code> otherwise
   */
  public boolean getOverrideConfigurationFiles() {
    return override_config;
  }

  /**
   * Sets the flag indicating whether or not settings from this object should override
   * the configuration files
   * @param override the flag value
   */
  public void setOverrideConfigurationFiles(boolean override) {
    InstallConfiguration.override_config = override;
  }

  /**
   * Returns whether or not the metamorphosys file and nlm files have been chosen
   * for installation.
   * @return <code>true</code> if copy to hard drive is chosen
   *         <code>false</code> otherwise
   */
  public boolean isCopyToHardDriveChecked() {
    return metamorphosys_checked;
  }

  /**
   * Sets whether or not the metamorphosys file and nlm files have been chosen
   * for installation.
   * @param checked according to selection of copy to hard drive
   */
  public void setCopyToHardDriveChecked(boolean checked) {
    metamorphosys_checked = checked;
  }


  /**
   * Returns the {@link File} that contains the MetamorphoSys
   * application.
   * @return {@link File} containing MetamorphoSys
   */
  public File getMetamorphoSysFile() {
    return metamorphosys_file;
  }

  /**
   * Returns the {@link File} for the MetamorphoSys "config" directory
   * to use.
   * @return {@link File} for the MetamorphoSys "config" directory
   */
  public File getMetamorphoSysConfigurationDirectory() {
    return metamorphosys_config_dir;
  }

  /**
   * Returns the {@link File} for the MetamorphoSys "config" directory
   * to use.
   * @param dir {@link File} for the MetamorphoSys "config" directory
   */
  public void setMetamorphoSysConfigurationDirectory(File dir) {
    metamorphosys_config_dir = dir;
  }

  /**
   * Sets the {@link File} that contains the MetamorphoSys application.
   * @param file {@link File} containing MetamorphoSys
   */
  public void setMetamorphoSysFile(File file) {
    metamorphosys_file = file;
  }

  /**
   * Returns whether or not a directory should be created for the
   * Metathesaurus.
   * @return <code>true</code> if metathesaurus is chosen
   *         <code>false</code> otherwise
   */
  public boolean isMetathesaurusChecked() {
    return metathesaurus_checked;
  }

  /**
   * Sets whether or not a directory should be created for the
   * Metathesaurus.
   * @param checked according to selection of metathesaurus
   */
  public void setMetathesaurusChecked(boolean checked) {
    metathesaurus_checked = checked;
  }

  /**
   * Returns the {@link File} directory that contains the Metathesaurus
   * data.
   * @return {@link File} containing Metathesaurus data
   */
  public File getUmlsDirectory() {
    return umls_dir;
  }

  /**
   * Returns the {@link File} that contains the Metathesaurus
   * data.
   * @return {@link File} containing Metathesaurus data
   */
  public File[] getMetathesaurusFiles() {
    return metathesaurus_files;
  }

  /**
   * Sets the {@link File} directory that contains the Metathesaurus data.
   * @param dir {@link File} containing Metathesaurus data
   */
  public void setUmlsDirectory(File dir) {
    umls_dir = dir;

    File[] files = umls_dir.listFiles();
    if (files != null) {
      setLexiconFile(new File(umls_dir,
                            getReleaseVersion().toLowerCase() +
                            "-otherks.nlm"));
      for (int i = 0; i < files.length; i++) {
        String file_name = files[i].getName();
        if (file_name.endsWith("-otherks.nlm") &&
            file_name.startsWith(getReleaseVersion().toLowerCase())) {
            setLexiconFile(files[i]);
            break;
        }
      }
    }

    setSemanticNetworkFile(getLexiconFile());

    if (files != null) {
      setMetathesaurusFiles(new File[] {
          new File(umls_dir, getReleaseVersion().toLowerCase() + "-meta.nlm") } );
      List meta_files = new ArrayList();
      for (int i = 0; i < files.length; i++) {
        String file_name = files[i].getName();
        if (file_name.endsWith("-meta.nlm") &&
            file_name.startsWith(getReleaseVersion().toLowerCase())) {
          meta_files.add(files[i]);
        }
      }
      setMetathesaurusFiles((File[])meta_files.toArray(new File[0]));
    }

    setMetamorphoSysFile(new File(umls_dir, "mmsys.zip"));
  }

  /**
   * Sets the {@link File} that contains the Metathesaurus data.
   * @param files {@link File} containing Metathesaurus data
   */
  public void setMetathesaurusFiles(File[] files) {
    metathesaurus_files = files;
  }

  /**
   * Returns the directory into which the Metathesaurus subset will be made.
   * @return {@link File} directory into which the subset will be made
   */
  public File getMetaDestinationDirectory() {
    return destination_meta_dir;
  }

  /**
   * Sets the directory into which the Metathesaurus subset will be made.
   * @param dir {@link File} directory into which the subsets will be made
   */
  public void setMetaDestinationDirectory(File dir) {
    destination_meta_dir = dir;
  }

  /**
   * Returns the directory into which the UMLS will be installed.
   * @return the {@link File} directory into which the UMLS will be installed
   */
  public File getUmlsDestinationDirectory() {
    return destination_umls_dir;
  }

  /**
   * Sets the directory into which the UMLS will be installed.
   * @param dir the {@link File} directory into which the UMLS will be installed
   */
  public void setUmlsDestinationDirectory(File dir) {
    destination_umls_dir = dir;
    setMetaDestinationDirectory(
      new File(
        new File(
          destination_umls_dir,ic.getReleaseVersion().toUpperCase()),
          "META"));
  }

  /**
   * Sets the release version.
   * @param version the release version
   */
  public void setReleaseVersion(String version) {
    release_version = version;
  }

  /**
   * Returns the release version.
   * @return the release version
   */
  public String getReleaseVersion() {
    return release_version;
  }

  /**
   * Indicates whether the user chose the Install UMLS button on the
   * {@link WelcomeGUI}.
   * @return <code>true</code> if yes
   *         <code>no</code> otherwise the user chose the ConfigureSubset button
   */
  public boolean isInstallUMLS() {
    return install_umls;
  }
  /**
   * Sets whether the user chose the Install UMLS button on the
   * {@link WelcomeGUI}.
   * @param install_umls <code>true</code> if user chose the Install UMLS button
   *         <code>no</code> otherwise the user chose the ConfigureSubset button
   */
  public void setInstallUMLS(boolean install_umls) {
    InstallConfiguration.install_umls = install_umls;
  }

  /**
   * Indicates whether the user should be queried to open last saved configuration
   * when initializing RRF Browser.
   * @return <code>true</code> if yes user should be queried
   *         <code>no</code> otherwise
   */
  public boolean isQueryUserToOpenSavedConfig() {
    return query_to_open;
  }
  /**
   * Sets whether the user should be queried to open the last saved configuration
   * when initializing the RRF Browser.
   * @param query_to_open <code>true</code> if user chose the Install UMLS button
   *         <code>no</code> otherwise the user chose the ConfigureSubset button
   */
  public void setQueryUserToOpenSavedConfig(boolean query_to_open) {
    InstallConfiguration.query_to_open = query_to_open;
  }

  /**
   * Returns a {@link MMSProgressMonitor}.  If one doesn't yet exist, it instantiates
   * one.
   * @return {@link MMSProgressMonitor}
   */
  public MMSProgressMonitor getProgressMonitor() {
    if (progress_monitor == null) {
      if (isInstallUMLS())
        progress_monitor =
          MMSToolkit.usingView() ? new InstallationProgressMonitor() : null;
      else
        progress_monitor =
          MMSToolkit.usingView() ? new ConfigureSubsetProgressMonitor() : null;
    }
    return progress_monitor;
  }

  /**
   * Clears the {@link MMSProgressMonitor}.
   */
  public void clearProgressMonitor() {
    if (progress_monitor != null)
      progress_monitor.dispose();
    progress_monitor = null;
  }

  /**
   * Gets the {@link MetamorphoSysInputStream}.
   * @return input stream
   */
  public MetamorphoSysInputStream getMetamorphoSysInputStream() {
    File meta_nlm_file = null;
    if (getMetathesaurusFiles() != null && getMetathesaurusFiles().length>0)
      meta_nlm_file = getMetathesaurusFiles()[0];
    if (meta_nlm_file != null && meta_nlm_file.exists() && isInstallUMLS()) {
      return new NLMFileMetamorphoSysInputStream();
    }
    File[] files = umls_dir.listFiles();
    if (files != null) {
      for (int i = 0; i < files.length; i++) {
        if (files[i].getAbsolutePath().endsWith("MRFILES.RRF")) {
          return new RichMRMetamorphoSysInputStream();
        }
        if (files[i].getAbsolutePath().endsWith("MRFILES")) {
          return new OriginalMRMetamorphoSysInputStream();
        }
      }
    }
    // return a default so that validation can be run
    return new RichMRMetamorphoSysInputStream();
  }

  /**
   * Installs the Semantic Network and the Specialist Lexicon and Lexical Tools
   * if they are selected on the {@link InstallationGUI}.  Writes the install.log.
   * @param owner parent {@link Component}
   * @return boolean <code>true</code> indicates installation completed 
   *    	           <code>false</code> indicates installation failed
   */
  public static boolean install(Frame owner) {
    if (!ic.isInstallUMLS())
      return true;

    boolean lex_unzipped = false, net_unzipped = false, lvg_unzipped = false;

    //
    // Open install.log file
    //
    File log_file = null;
    try {
      log_file = new File(ic.getUmlsDestinationDirectory() + SEPARATOR +
                          ic.getReleaseVersion().toUpperCase(),
                          "install.log");
      log_file.getParentFile().mkdirs();
      log = new BufferedWriter(new FileWriter(log_file));

    }
    catch (Exception e) {
      e.printStackTrace();
      MMSToolkit.reportError(owner,
          "Unexpected error trying to open the install.log file.\n" +
          "See console for more details.", false);
    }
    try {
    log.write("Installation time:................................." +
                new Date());
    log.newLine();
    log.write("Source directory:.................................." +
                ic.getUmlsDirectory());
    log.newLine();
    log.write("Destination directory:............................." +
                ic.getUmlsDestinationDirectory());
    log.newLine();

    //
    // Prep progress monitor
    //
    final MMSProgressMonitor progress =
        ( (MMSProgressMonitor) ic.getProgressMonitor());

    //
    // unzip LEX if indicated
    //
    if (ic.isLexiconChecked()) {
      log.write("Install Specialist Lexicon and Tools...............");
      Date start = new Date();
      progress.setOperation(InstallationProgressMonitor.INSTALL_LEXICON);
      try {
        lex_unzipped = SwingToolkit.unzip(
            ic.getLexiconFile().getAbsolutePath(),
            destination_umls_dir.getAbsolutePath(),
            ic.getReleaseVersion().toUpperCase() + "/LEX/*",
            owner);
      }
      catch (IOException ioe) {
        log.write("NO (error)");
        log.newLine();
        handleError(owner, ioe, "Error installing Specialist Lexicon and Tools.");
        return false;
      }

      if (lex_unzipped) {
        Date end = new Date();
        log.write("YES (" +
                    MMSToolkit.timeToString(MMSToolkit.
                                            timeDifference(end, start)) + ")");
        log.newLine();
      }
      else {
        log.write("NO (cancelled by user)");
        log.newLine();
        MMSToolkit.notifyUser(owner,
            "Installing the Specialist Lexicon and Tools was cancelled.\n" +
            "This also cancels the install process.");
        log.flush();
        log.close();
        WelcomeGUI.getWelcomeGUI().disableGlassPane();
        InstallConfiguration.getInstallConfiguration().clearProgressMonitor();
        return false;
      }

    }
    else {
      log.write("Install Specialist Lexicon and Tools...............NO");
      log.newLine();
    }

    //
    // unzip LVG if indicated
    //
    if (ic.isLvgChecked()) {
      log.write("Install and Auto-Configure LVG.....................");
      Date start = new Date();
      progress.setOperation(InstallationProgressMonitor.INSTALL_LVG);
      try {
      	File lex_dir = new File(ic.getMetaDestinationDirectory(), "../LEX").getCanonicalFile();
      	File lex_pgms = new File(lex_dir, "/LEX_PGMS");
      	File[] file_list = lex_pgms.listFiles();
      	File tgz_file = null;
      	for (int i=0; i<file_list.length; i++) {
      		if (file_list[i].getAbsolutePath().endsWith(".tgz")) {
      			tgz_file = file_list[i];
      			break;
      		}
      	}
      	
      	// unzip the \LEX\LEX_PGMS\lvg{2005}.tgz file
      	lvg_unzipped = MMSToolkit.readTar(new GZIPInputStream(new FileInputStream(tgz_file)),
      			lex_pgms.getAbsolutePath(), tgz_file, WelcomeGUI.getWelcomeGUI());
      	
      	// get the \LEX\LEX_PGMS\lvg{2005} directory
      	File[] dir_list = lex_pgms.listFiles();
      	File lvg_version_dir = null;
      	for (int i=0; i<dir_list.length; i++) {
      		if (dir_list[i].getAbsolutePath().indexOf("lvg") != -1) {
      		  lvg_version_dir = dir_list[i];
      		  break;
      		}	                                
      	}
    
      	// update the LVG_DIR property in the lvg.properties file
        File lvg_properties = new File(lvg_version_dir,
        		"/data/config/lvg.properties");
        File lvg_properties_out = new File(lvg_version_dir,
    				"/data/config/lvg.properties.out");
        BufferedReader reader = new BufferedReader(new FileReader(lvg_properties));
        FileOutputStream fos = new FileOutputStream(lvg_properties_out);
        BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(fos));
        while (true) {
          String line = reader.readLine();
          if (line == null) {
            break;
          } else if (line.startsWith("LVG_DIR")) {
          	String lex_pgrms_path = lvg_version_dir.getAbsolutePath() + SEPARATOR;
          	lex_pgrms_path = lex_pgrms_path.replaceAll("\\\\", "\\\\\\\\");
          	writer.write("LVG_DIR=" + lex_pgrms_path);         	
          	writer.newLine();
          	continue;
          }
          writer.write(line);
          writer.newLine();
        }
        reader.close();
        writer.close();
        SystemToolkit.copy(lvg_properties_out, lvg_properties);
        lvg_properties_out.delete();
      }
      catch (IOException ioe) {
        log.write("NO (error)");
        log.newLine();
        handleError(owner, ioe, "Error installing LVG.");
        return false;
      }

      if (lvg_unzipped) {
        Date end = new Date();
        log.write("YES (" +
                    MMSToolkit.timeToString(MMSToolkit.
                                            timeDifference(end, start)) + ")");
        log.newLine();
      }
      else {
        log.write("NO (cancelled by user)");
        log.newLine();
        MMSToolkit.notifyUser(owner,
            "Installing LVG was cancelled.\n" +
            "This also cancels the install process.");
        log.flush();
        log.close();
        WelcomeGUI.getWelcomeGUI().disableGlassPane();
        InstallConfiguration.getInstallConfiguration().clearProgressMonitor();
        return false;
      }

    }
    else {
      log.write("Install and Auto-Configure LVG....................NO");
      log.newLine();
    }
   
    
    //
    // unzip NET if indicated
    //
    if (ic.isSemanticNetworkChecked()) {
      log.write("Install Semantic Network...........................");
      //log.newLine();
      Date start = new Date();
      progress.setOperation(InstallationProgressMonitor.INSTALL_NETWORK);
      try {
        net_unzipped = SwingToolkit.unzip(
            ic.getSemanticNetworkFile().
            getAbsolutePath(),
            destination_umls_dir.getAbsolutePath(),
            ic.getReleaseVersion().toUpperCase() + "/NET/*",
            owner);
      }
      catch (IOException ioe) {
        log.write("NO (error)");
        log.newLine();
        handleError(owner, ioe, "Error installing Semantic Network.\n" +
        		ioe.getMessage());
        return false;
      }

      if (net_unzipped) {
        Date end = new Date();
        log.write("YES (" +
                    MMSToolkit.timeToString(MMSToolkit.
                                            timeDifference(end, start)) + ")");
        log.newLine();
      }
      else {
        log.write("NO (cancelled by user)");
        log.newLine();
        MMSToolkit.notifyUser(owner,
            "Installing the Semantic Network was cancelled.\n" +
            "This also cancels the install process.");
        log.flush();
        log.close();
        WelcomeGUI.getWelcomeGUI().disableGlassPane();
        InstallConfiguration.getInstallConfiguration().clearProgressMonitor();
        return false;
      }

    }
    else {
      log.write("Install Semantic Network...........................NO");
      log.newLine();
    }

    //
    // Create a META directory if indicated
    //
    if (ic.isMetathesaurusChecked()) {
      log.write("Install Metathesaurus..............................YES");
      log.newLine();
      File meta_dir = ic.getMetaDestinationDirectory();
      if (!meta_dir.exists())
        meta_dir.mkdirs();
    }
    else {
      log.write("Install Metathesaurus..............................NO");
      log.newLine();
    }

    log.close();

    //
    // Summarize what happened for the user only
    // if we are not continuing to configure subset and
    // don't have an extended progress monitor
    //
    if (!ic.isMetathesaurusChecked()) {
      progress.setOperation(MMSProgressMonitor.DONE);
      SwingToolkit.displayTextFile(owner, "Installation Log", log_file);
    }
    } catch (Exception ex) {
    	MMSToolkit.reportError("Error writing to install.log.");
    }
    return true;
  }


  /**
   * Handle error.
   * @param e the exception
   * @param msg the error message
   */
 private static void handleError(Frame owner, Exception e, String msg) {
   MMSToolkit.printStackTrace(e, log);
   try {
     log.flush();
     log.close();
   } catch (Exception ex) {
			MMSToolkit.reportError("Unable to write to install.log.");		 
   }
    File log_file = new File(
        InstallConfiguration.getInstallConfiguration().getUmlsDestinationDirectory()
            + SEPARATOR
            + InstallConfiguration.getInstallConfiguration().getReleaseVersion().toUpperCase(),
        "install.log");
    SwingToolkit.displayTextFile(owner, "UMLS Install Log", log_file);
   MMSToolkit.reportError(owner, msg, false);
   //disableGlassPane();
   //dispose();
   WelcomeGUI.getWelcomeGUI().disableGlassPane();
   InstallConfiguration.getInstallConfiguration().clearProgressMonitor();
   return;
 }
 


} // end class InstallConfiguration.java

