/**************************************************************************
 *
 * Package:    gov.nih.nlm.mms;
 * Interface:  LicenseAgreementGUI.java
 *
 *************************************************************************/
package gov.nih.nlm.mms;

import gov.nih.nlm.swing.FontSizeManager;
import gov.nih.nlm.swing.HtmlViewer;
import gov.nih.nlm.swing.SwingToolkit;
import gov.nih.nlm.util.BrowserLauncher;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Frame;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.IOException;
import java.net.URL;

import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JEditorPane;
import javax.swing.JPanel;
import javax.swing.border.EmptyBorder;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;

/**
 * Used to present the license agreement (or equivalent URL) to the user and
 * have the user "accept" it before going on.
 * 
 * @author Deborah Shapiro
 */
public class LicenseAgreementGUI extends JDialog {

  //
  // Fields
  //
  private boolean is_cancelled = true;

  //
  // Constructor
  //

  /**
   * Instantiates a {@link LicenseAgreementGUI}. The constructor does not make
   * it visible.
   * 
   * @param mms
   *          the {@link MetamorphoSys} applicatoin
   * @param owner parent {@link Component}
   */
  public LicenseAgreementGUI(MetamorphoSys mms, final Frame owner) {
    super(owner);

    //
    // Set title and modal settings
    //
    setModal(true);
    setTitle("License Agreement Notice");

    //
    // Configure layout
    //
    getContentPane().setLayout(new BorderLayout());

    //
    // Get HTML viewer over license agreement text.
    // Respond to clicks on HyperLinks with the BrowserLauncher
    //
    HtmlViewer html_viewer = new HtmlViewer();
    html_viewer.setPage(this.getClass().getClassLoader().getResource(
        "config/license.html"));
    JEditorPane editor_pane = html_viewer.getEditorPane();
    editor_pane.setContentType("text/html");
    editor_pane.addHyperlinkListener(new HyperlinkListener() {
      public void hyperlinkUpdate(HyperlinkEvent he) {
        if (he.getEventType() == HyperlinkEvent.EventType.ACTIVATED) {
          final String dsc = he.getDescription();
          URL local_url = null;
          try {
            local_url = new URL(dsc);
          } catch (Exception e) {
            MMSToolkit.notifyUser(owner, "Error with URL " + dsc);
          }
          final URL url = local_url;
          Thread t = new Thread(new Runnable() {
            public void run() {
              try {
                BrowserLauncher.openURL(dsc);
              } catch (IOException ioe) {
                SwingToolkit.displayURL(owner, "Notice", url);
              }
            }
          }); // End of new Thread
          t.start();
        }
      }
    });
    final JPanel agreement_panel = new JPanel();
    agreement_panel.setLayout(new BorderLayout());
    agreement_panel.add(html_viewer, BorderLayout.CENTER);

    if (agreement_panel == null)
      return;

    //
    // Add Buttons
    //

    final JPanel base_panel = new JPanel();
    base_panel.setLayout(new BorderLayout());
    base_panel.setBorder(new EmptyBorder(10, 10, 10, 10));
    base_panel.add(agreement_panel, BorderLayout.CENTER);

    JPanel btn_panel = new JPanel();
    JButton done_button = new JButton("Accept");
    getRootPane().setDefaultButton(done_button);
    done_button.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        is_cancelled = false;
        dispose();
      }
    });
    btn_panel.add(done_button);

    JButton cancel_button = new JButton("Do Not Accept");
    cancel_button.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        //
        // Go back to welcome GUI unless we are at the configuration gui for
        // subsetting
        //
        if (WelcomeGUI.getWelcomeGUI().isVisible())
          InstallConfiguration.getInstallConfiguration().clearProgressMonitor();
        dispose();
      }
    });
    done_button.setPreferredSize(cancel_button.getPreferredSize());
    btn_panel.add(cancel_button);
    base_panel.add(btn_panel, BorderLayout.SOUTH);

    //
    // Prep for viewing
    //
    FontSizeManager.addContainer(base_panel);
    if (agreement_panel == null)
      return;

    this.setContentPane(base_panel);
    FontSizeManager.adjustFontSize(this);
    this.pack();
    done_button.requestFocus();

    setLocationRelativeTo(null);

    // Configure window listener
    addWindowListener(new WindowAdapter() {
      public void windowClosing(WindowEvent e) {
        InstallConfiguration.getInstallConfiguration().clearProgressMonitor();
        MMSToolkit.cancel();
      }
    });
  };

  /**
   * Indicates whether or not the license agreement was cancelled or not
   * accepted.
   * 
   * @return <code>true</code> if the GUI was cancelled, <code>false</code>
   *         otherwise
   */
  public boolean isCancelled() {
    return is_cancelled;
  }

} // End of LicenseAgreementGUI class

