/*****************************************************************************
 *
 * Package:    gov.nih.nlm.mms
 * Object:     MMSToolkit.java
 *
 *****************************************************************************/
package gov.nih.nlm.mms;

import gov.nih.nlm.swing.FontSizeManager;
import gov.nih.nlm.swing.GlassPaneListener;
import gov.nih.nlm.swing.SwingToolkit;
import gov.nih.nlm.util.BrowserLauncher;
import gov.nih.nlm.util.SystemToolkit;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Frame;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.io.UnsupportedEncodingException;
import java.lang.reflect.Constructor;
import java.lang.reflect.Modifier;
import java.net.URL;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.jar.JarFile;

import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.ProgressMonitor;
import javax.swing.ScrollPaneConstants;
import javax.swing.border.CompoundBorder;
import javax.swing.border.EmptyBorder;
import javax.swing.border.EtchedBorder;
import javax.swing.border.TitledBorder;
import javax.swing.filechooser.FileFilter;

import com.ice.tar.TarEntry;
import com.ice.tar.TarInputStream;


/**
 * Utility class containing methods for MetamorphoSys applications.
 *
 * @author Brian A. Carlsen
 */
public class MMSToolkit extends SwingToolkit {

  //
  // Constants
  //
  private static MetamorphoSysGUI gui;

  //
  // Private Fields
  //
  private static String release_dat = "../release.dat";
  final private static Properties release_dat_properties = new Properties();
  static {
    try {
      setReleaseDatLocation(release_dat);
    } catch (IOException ioe) { }
  }

  //
  // Standard MMSToolkit methods
  //
  /**
   * Sets the "using view" flag.
   * @param view a <code>boolean</code> flag indicating whether or
   *        not the current application is using a GUI.
   */
  public static void setUsingView(boolean view) {
    // Set it in SwingToolkit also
    SwingToolkit.setProperty(SwingToolkit.VIEW, String.valueOf(view));
  }

  /**
   * Prompt user to choose a file.
   * @param title the title of the dialog
   * @param button_text the text on the button
   * @param selection_mode the selection mode (i.e. Files/Directories/both)
   * @param current_dir the current directory
   * @param selected_file {@link File}
   * @param owner parent {@link Component}
   * @return the {@link File} selected by the user
   */
  public static File chooseFile(String title,
                                String button_text,
                                int selection_mode,
                                File current_dir,
                                File selected_file,
                                Frame owner) {
    if (usingView()) {
      JFileChooser file_chooser =
          prepareFileChooser(current_dir, selected_file, title, selection_mode);
      int result = file_chooser.showDialog(owner, button_text);
      if (result == JFileChooser.APPROVE_OPTION) {
        java.io.File filename = file_chooser.getSelectedFile();
        return filename;
      }
      else
	return null;
    } else
      return null;
  }

  /**
   * Prompts user to choose a file, specifying a file filter.
   *
   * @param title          the title of the dialog
   * @param button_text    the text on the button
   * @param selection_mode the selection mode (i.e. Files/Directories/both)
   * @param current_dir    the current directory
   * @param selected_file  {@link File}
   * @param fileFilter     {@link FileFilter} on accepted file type(s)
   * @param owner          parent {@link Component}
   * @return the {@link File} selected by the user
   */
  public static File chooseFile(String title, String button_text,
      int selection_mode, File current_dir, File selected_file,
      FileFilter fileFilter, Frame owner) {
    if (usingView()) {
      JFileChooser file_chooser =
          prepareFileChooser(current_dir, selected_file, title, selection_mode);
      if (fileFilter != null) {
        file_chooser.setFileFilter(fileFilter);
      }
      int result = file_chooser.showDialog(owner, button_text);
      if (result == JFileChooser.APPROVE_OPTION) {
        File filename = file_chooser.getSelectedFile();
        return filename;
      }
      else
        return null;
    }
    else
      return null;
  }

  /**
   * Prompts user to choose a file to save, specifying a file filter and
   * file extension.
   *
   * @param title          the title of the dialog
   * @param button_text    the text on the button
   * @param selection_mode the selection mode (i.e. Files/Directories/both)
   * @param current_dir    the current directory
   * @param selected_file  {@link File}
   * @param fileFilter     {@link FileFilter} on accepted file type(s)
   * @param fileExtension  desired file extension
   * @param owner          parent {@link Component}
   * @return the {@link File} selected by the user
   */
  public static File chooseFile(String title, String button_text,
      int selection_mode, File current_dir, File selected_file,
      FileFilter fileFilter, String fileExtension, Frame owner) {
    if (usingView()) {
      JFileChooser file_chooser =
          prepareFileChooser(current_dir, selected_file, title, selection_mode);
      file_chooser.setDialogType(JFileChooser.SAVE_DIALOG);
      if (fileFilter != null) {
        file_chooser.setFileFilter(fileFilter);
      }
      int result = file_chooser.showDialog(owner, button_text);
      if (result == JFileChooser.APPROVE_OPTION) {
        File selectedFile = file_chooser.getSelectedFile();
        if (!getExtension(selectedFile).equals(fileExtension)) {
          selectedFile = new File(selectedFile.getParent(),
              selectedFile.getName() + "." + fileExtension);
        }
        if (canSaveFile(selectedFile, file_chooser)) {
          return selectedFile;
        }
        else {
          return null;
        }
      }
      else
        return null;
    }
    else
      return null;
  }

  /**
   * Create and set up a {@link JFileChooser} dialog.
   * @param current_dir    the current directory
   * @param selected_file  {@link File}
   * @param title          the title of the dialog
   * @param selection_mode the selection mode (i.e. Files/Directories/both)
   * @return a configured {@link JFileChooser} dialog.
   */
  private static JFileChooser prepareFileChooser(File current_dir,
      File selected_file, String title, int selection_mode) {
    File l_current_dir = null;
    File l_selected_file = null;
    try {
      if (current_dir != null)
        l_current_dir = current_dir.getCanonicalFile();
    } catch (Exception e) {
      // do nothing
    }
    try {
      if (l_selected_file != null)
        l_selected_file = selected_file.getCanonicalFile();
    } catch (Exception e) {
      // do nothing
    }

    JFileChooser file_chooser = new JFileChooser();
    file_chooser.setBounds(1, 0, 500, 300);
    file_chooser.setDialogTitle(title);
    file_chooser.setFileSelectionMode(selection_mode);
    if (l_selected_file != null)
      file_chooser.setSelectedFile(l_selected_file);
    if (l_current_dir != null && !l_current_dir.equals(l_selected_file))
      file_chooser.setCurrentDirectory(l_current_dir);
    else if (l_current_dir != null)
      file_chooser.setCurrentDirectory(l_current_dir.getParentFile());
    FontSizeManager.adjustFontSize(file_chooser);
    //file_chooser.setFileHidingEnabled(false);
    return file_chooser;
  }

  /**
   * Returns the extension portion of the file's name.
   * @param f {@link File} on which to find extension
   * @return the file's extension, or an empty {@link String} if none.
   * @see FileFilter#accept
   */
  public static String getExtension(File f) {
    if (f != null) {
      String filename = f.getName();
      int i = filename.lastIndexOf('.');
      if (i > 0 && i < filename.length() - 1) {
        return filename.substring(i + 1).toLowerCase();
      }
    }
    return "";
  }

  /**
   * Checks whether the file can be saved.  If the file exists and is writable,
   * permission is requested from the user to determine the answer.
   *
   * @return <code>true</code> if the file can be written; <code>false</code>
   *         otherwise.
   */
  private static boolean canSaveFile(File file, JFileChooser fileChooser) {
    if (file.exists()) {
      // the file exists, can we write over it?
      if (file.canWrite()) {
        // we can write over the existing file, but does the user
        // really want to do that?
        if (JOptionPane.showConfirmDialog(fileChooser.getParent(),
            "File already exists. Do you want to overwrite it?", "File Exists",
            JOptionPane.YES_NO_OPTION) == JOptionPane.YES_OPTION) {
          // the file exists, we can write over it,
          // and the user wants to do that
          return true;
        }
        else {
          return false;
        }
      }
      else {
        JOptionPane.showMessageDialog(fileChooser.getParent(),
            "File is read-only. Cannot write to file " + file.getName() + ".",
            "Check File", JOptionPane.ERROR_MESSAGE);
        return false;
      }
    }
    return true;
  }

  /**
   * Indicates whether or not the given {@link String} has
   * any extended unicode characters in it.
   * @param str given {@link String}
   * @return boolean <code>true</code> extended Unicode characters are present
   *                 <code>false</code> extended Unicode characters are not present
   * @throws UnsupportedEncodingException
   */
  public static boolean containsUnicode(String str)  throws UnsupportedEncodingException {
    // If the length does not equal the byte count it is unicode
    return str.length() != str.getBytes("UTF-8").length;
  }

  /**
   * Returns <code>true</code> if the given archive {@link File}
   * contains the given directory.
   * @param file archive {@link File}
   * @param dir directory name
   * @return <code>true</code> if archive contains the given directory
   *         <code>false</code> otherwise
   * @throws Exception
   */
  public static boolean archiveContainsDir(File file, String dir) throws Exception {
    boolean found = false;
    if (!file.exists())
      throw new Exception("File does not exist: " + file.getAbsolutePath());

    String [] zip_entries = SystemToolkit.getZipEntries(file.getAbsolutePath());

    String ew_match = "/" + dir + "/";
    String io_match = "\t" + dir + "/";
    String io_match2 = " " + dir + "/";
    for (int i = 0; i<zip_entries.length; i++) {
    	String line = zip_entries[i];
      if (line.endsWith(ew_match) || line.indexOf(io_match) != -1 || line.indexOf(io_match2) != -1)
        found = true;
    }

    return found;

   }

  /**
   * Convert a <code>long</code> representing a time to
   * a string representing hours, minutes, and seconds.
   * @param time the time in milliseconds
   * @return a {@link String} representation "HH:MM:SS"
   */
  public static String timeToString( long time ) {
    if (time < 1000) {
      return "00:00:00."+time;
    }
    else {
      int hours = (int)(time / 3600000);
      time = time % 3600000;
      int minutes = (int)(time / 60000);
      time = time % 60000;
      int seconds = (int)(time / 1000);

      String h = "0"+hours;
      String m = "0"+minutes;
      String s = "0"+seconds;
      return h.substring(h.length()-2) + ":" +
	m.substring(m.length()-2) + ":" +
	s.substring(s.length()-2);
    }

  }

  /**
   * Compute the time difference between the specified parameters.
   * @param now the later time
   * @param then the earlier time
   * @return a <code>long</code> indicating the number
   *         of milliseconds between the specified dates
   */
  public static long timeDifference( java.util.Date now,
				    java.util.Date then) {
    return now.getTime() - then.getTime();
  };
  
  /**
   * Scans all <code>.jar</code> files in the <code>CLASSPATH</code>
   * and returns all fully qualified class names that implement
   * the specified interface.
   *
   * @param interface_name the fully qualified interface name
   * @return a {@link String}<code>[]</code> of all fully qualified
   *        class names implementing that interface
   * @throws ClassNotFoundException
   * @throws IOException
   */
  public static String [] getInterfaceImplementations(String interface_name) 
    throws ClassNotFoundException, IOException {
  	List dir_list = new ArrayList();
  	dir_list.add(new File("lib"));
  	dir_list.add(new File("ext"));
  	return getInterfaceImplementations(interface_name, dir_list);
  }
  
  /**
   * Scans all <code>.jar</code> files in the <code>CLASSPATH</code>
   * and returns all fully qualified class names that implement
   * the specified interface.
   *
   * @param interface_name the fully qualified interface name
   * @param dir_list {@link List}
   * @return a {@link String}<code>[]</code> of all fully qualified
   *        class names implementing that interface
   * @throws ClassNotFoundException
   * @throws IOException
   */
  public static String [] getInterfaceImplementations(String interface_name,
  		List dir_list)
  throws ClassNotFoundException, IOException {
  	List interface_implementations = new ArrayList();
  	String class_path = System.getProperty("java.class.path");
  	String path_separator = System.getProperty("path.separator");
  	StringTokenizer path_names =
  		new StringTokenizer(class_path, path_separator);
  	
  	// load interface
  	Class interface_class = Class.forName(interface_name);
  	
  	while (path_names.hasMoreElements()) {
  		String path = path_names.nextToken();
  		//  Get only JAR files
  		if (path.endsWith(".jar"))  {
  			JarFile jf = null;
  			File f = new File(path);
  			boolean match = false;
  			for (int i=0; i<dir_list.size(); i++) {
  				if (((File)dir_list.get(i)).getCanonicalPath().equals(f.getParentFile().getCanonicalPath()))
  					match = true;
  			}
  			if (f.exists() && match) {
  				jf = new JarFile (f);
  			} else {
  				// .jar file in classpath, but does not exist
  				continue;
  			}
  			
  			Enumeration enumeration = jf.entries();
  			while (enumeration.hasMoreElements()) {
  				
  				//  Replace path separators with '/'
  				char file_separator = '/';
  				String entry = (enumeration.nextElement()).toString().replace(file_separator,'.');
  				
  				//  Remove ".class" extension
  				if (entry.endsWith(".class")) {
  					entry = entry.substring(0, entry.length()-".class".length());
  				} else {
  					continue;
  				}
  				Class entry_class;
  				try {
  					entry_class = Class.forName(entry);
  				} catch (Throwable e) {
  					continue;
  				}
  				if ((entry_class.getModifiers() & Modifier.ABSTRACT) != 0) {
  					continue;
  				}
  				if (interface_class.isAssignableFrom(entry_class)) {
  					interface_implementations.add(entry);
  				}
  			}
  		}
  	}
  	
  	return (String[]) interface_implementations.toArray(new String[0]);
  };


  /**
   * Cleans up system resources, then exits.
   * @param return_value the <code>System.exit(int)</code> status code
   */
  public static void Exit (int return_value) {
    if (gui != null) gui.cancel();
    System.err.println("Exiting ....");
    System.exit(return_value);
  }

  /**
   * Closes given {@link JFrame} and returns to the {@link WelcomeGUI}.
   * @param frame {JFrame} of sub-application to exit
   */
  public static void exitToWelcomeGUI(JFrame frame) {
    frame.dispose();
    if (WelcomeGUI.getWelcomeGUI().isVisible())
    	WelcomeGUI.getWelcomeGUI().disableGlassPane();
    else
    	Exit(0);
  }
  
  /**
   * Cleans up system resources, shows the log file, then exits.
   * @param title for the log
   * @param log_file {@link File} for log
   */
  public static void showLogAndExitToWelcomeGUI (String title, File log_file) {
    if (gui != null) gui.cancel();
    SwingToolkit.displayTextFile(gui, title, log_file);
    exitToWelcomeGUI(gui);
  }


  /**
   * Helper function to make panels for the {@link MetamorphoSysGUI}
   * tabbed pane. Each panel displays the configuration GUI for one
   * filter.
   *
   * @param table a {@link JTable} to be displayed
   * @param title title for the panel
   * @return a {@link JPanel} with the table and title on it
   */
  public static JPanel makeCard(JTable table, String title) {
    JScrollPane scroll_pane;

    table.setAutoResizeMode(JTable.AUTO_RESIZE_SUBSEQUENT_COLUMNS);

    TitledBorder titled_border = new TitledBorder(new EtchedBorder(EtchedBorder.RAISED), title);
    titled_border.setTitleJustification(TitledBorder.CENTER);

    JPanel table_panel = new JPanel(new BorderLayout());
    table_panel.setBorder(new CompoundBorder(titled_border, new EmptyBorder(10, 10, 10, 10)));
    //table_panel.add(table.getTableHeader(),BorderLayout.PAGE_START);
    //table_panel.add(table,BorderLayout.CENTER);
    scroll_pane = new JScrollPane(table,
	ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED,
	ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
    table_panel.add(scroll_pane);
    return table_panel;
  }



  /**
   * Instantiates an object of the specified fully qualified type.
   * @param name a fully qualified class name
   * @return {@link Object} that was instantiated
   */
  public static Object instantiateObject(String name) {
    Object o = null;
    try {
      Class cls = Class.forName(name);
      Constructor constr;
      constr = cls.getConstructor(new Class [] {});
      o = constr.newInstance(new Object [] {});
    } catch (Exception e) {
      if (name.startsWith("gov.nih.nlm")) {
        e.printStackTrace();
        MMSToolkit.reportError("Radically unexpected error. \nPlease see console.\n\n " + e.getMessage(), true);
      } else {
	e.printStackTrace();
	MMSToolkit.reportError(
         "Failed to create class "  + name + ".\n" +
	 "You most likely forgot a library in the\n" +
	 "METAMSYS/META/ext directory that was used\n" +
	 "with the previous release.", false);
      }
    }
    return o;
  }


  /**
   * Returns a {@link JMenuItem} for a "Help" menu that opens a browser with
   * the specified URL.  The title is used for the help menu item label.
   * @param title for help menu
   * @param url for help menu display
   * @param owner parent {@link Component}
   * @return {@link JMenuItem} for a "Help" menu item
   */
  public static JMenuItem getResourceHelpMenu(final String title,
                                              final String url,
                                              final Frame owner) {
    JMenuItem menu_item = new JMenuItem(title);
    menu_item.addActionListener(
      new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          // Wrap html documentation in a panel
          JPanel help_panel =
              SwingToolkit.getPanelForHTML(this.getClass().getClassLoader().getResource(url));

          if (help_panel == null) {
            MMSToolkit.notifyUser(owner, "No help available.");
            return;
          }

          help_panel.setBackground(Color.gray);
          final JPanel base_panel = new JPanel(new BorderLayout());
          base_panel.setBorder(new EmptyBorder(10, 10, 10, 10));
          base_panel.add(help_panel, BorderLayout.CENTER);

          final JDialog help_frame = new JDialog(owner);
          help_frame.setModal(true);

          JButton done_button = new JButton("Done");
          done_button.addActionListener(
            new ActionListener() {
              public void actionPerformed(ActionEvent e) {
                help_frame.dispose();
              }
            });
          base_panel.add(done_button, BorderLayout.SOUTH);
          FontSizeManager.addContainer(base_panel);
          help_frame.setContentPane(base_panel);

          help_frame.getRootPane().setDefaultButton(done_button);
          help_frame.setTitle(title);
          help_frame.pack();
          done_button.requestFocus();
          help_frame.setLocationRelativeTo(owner);
          help_frame.setVisible(true);
        }
      }
    );
    return menu_item;
  }


  /**
   * Returns a {@link JMenuItem} for a "Help" menu that opens a browser with
   * the specified URL.  The title is used for the help menu item label.
   * @param title for help menu
   * @param url to be displayed
   * @param gpl {@link GlassPaneListener}
   * @return {@link JMenuItem} for a "Help" menu item
   */
  public static JMenuItem getURLHelpMenu(
      final String title, final String url, final GlassPaneListener gpl) {
    JMenuItem menu_item =
        new JMenuItem(title + " " + url);
    URL local_url = null;
    try {
      local_url = new URL(url);
    }
    catch (Exception ex) {
      MMSToolkit.notifyUser(
          "Error with URL " + url);
    }
    final URL umls_url = local_url;
    menu_item.addActionListener(
        new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        gpl.enableGlassPane();
        Thread t = new Thread(new Runnable() {
          public void run() {
            try {
              BrowserLauncher.openURL(url);
              gpl.disableGlassPane();
            }
            catch (IOException ioe) {
              SwingToolkit.displayURL((Component)gpl,  title,
                                      umls_url);
              gpl.disableGlassPane();
            }
          }
        }); // End of new Thread
        t.start();
      }
    }
    );
    return menu_item;
  }

  /**
   * Returns the application timestamp from <code>MMSYS/config/timestamp.dat</code>.
   * @return the application timestamp from <code>MMSYS/config/timestamp.dat</code>.
   */
  public static String getApplicationBuildDate() {
    // If the timestamp.dat file exists, read it and return
    String timestamp = "";
    try {
      BufferedReader in = new BufferedReader(
        new InputStreamReader(
          (InputStream)MMSToolkit.class.getClassLoader().getResource(
            "config/timestamp.dat").getContent()));
      if ((timestamp = in.readLine()) == null)
        timestamp = "";
      in.close();
      return timestamp;
    } catch(IOException e) {
      return "<error reading timestamp>";
    }
  }

  /**
   * Returns the application version from <code>MMSYS/config/version.dat</code>.
   * @return the application version from <code>MMSYS/config/version.dat</code>.
   */
  public static String getApplicationVersion() {
    // If the version.dat file exists, read it and return
    String version = "";
    try {
      BufferedReader in = new BufferedReader(
        new InputStreamReader(
          (InputStream)MMSToolkit.class.getClassLoader().getResource(
            "config/version.dat").getContent()));
      if ((version = in.readLine()) == null)
        version = "";
      in.close();
      return version;
    } catch(IOException e) {
      return "<error reading version>";
    }
  }

  /**
   * Returns release version of the UMLS from <code>release.dat</code>.
   * @return release version of the UMLS from <code>release.dat</code>
   */
  public static String getReleaseVersion() {
    if (release_dat_properties.containsKey("umls.release.name"))
      return release_dat_properties.getProperty("umls.release.name");
    MMSToolkit.reportError(getGui(),
                           "umls.release.name not found in release.dat file.",
                           false);
    return "";
  }

  /**
   * Returns release description of the UMLS from <code>release.dat</code>.
   * @return release description of the UMLS from <code>release.dat</code>
   */
  public static String getReleaseDescription() {
    if (release_dat_properties.containsKey("umls.release.description"))
      return release_dat_properties.getProperty("umls.release.description");
    MMSToolkit.reportError(getGui(),
                           "umls.release.description not found in release.dat file.",
                           false);
    return "";

  }

  /**
   * Returns release date of the UMLS from <code>release.dat</code>.
   * @return release date of the UMLS from <code>release.dat</code>
   */
  public static String getReleaseDate() {
    if (release_dat_properties.containsKey("umls.release.date"))
      return release_dat_properties.getProperty("umls.release.date");
    MMSToolkit.reportError(getGui(),
                           "umls.release.dat not found in release.dat file.",
                           false);
    return "";
  }

  /**
   * Returns build date of the UMLS from <code>release.dat</code>.
   * @return build date of the UMLS from <code>release.dat</code>
   */
  public static String getBuildDate() {
    if (release_dat_properties.containsKey("nlm.build.date"))
      return release_dat_properties.getProperty("nlm.build.date");
    return "No nlm.build.date";
  }
  
  /**
   * Returns the lvg version from <code>release.dat</code>.
   * @return lvg version from <code>release.dat</code>
   */
  public static String getLvgVersion() {
    if (release_dat_properties.containsKey("umls.lvg.version"))
      return release_dat_properties.getProperty("umls.lvg.version");
    MMSToolkit.reportError(getGui(),
                           "umls.lvg.version not found in release.dat file.",
                           false);
    return "";
  }
  
  /**
   * Helper method for writing <code>release.dat</code>.
   * @param mr_dir directory location for release.dat
   * @throws IOException if the file cannot be written
   */
  public static void writeReleaseDatFile(String mr_dir) throws IOException {
    FileOutputStream out = new FileOutputStream(new File(mr_dir, "release.dat"));
    release_dat_properties.put("mmsys.version", MMSToolkit.getApplicationVersion());
    release_dat_properties.put("mmsys.build.date", MMSToolkit.getApplicationBuildDate());
    release_dat_properties.store(out, " Release Metadata ");
    out.close();
  }

  /**
   * Get a copy of the release.dat Properties with the property list filled in.
   * @return a copy of the release.dat Properties.
   */
  public static Properties getReleaseDatProperties() {
    // Create Properties object with an empty property list with defaults
    //  specified by release_dat_properties
    Properties properties = new Properties(release_dat_properties);
    // Then create a property list with same keys and values as the default
    //  property list
    Enumeration e = properties.propertyNames();
    while (e.hasMoreElements()) {
      String propertyName = (String) e.nextElement();
      properties.put(propertyName, properties.getProperty(propertyName));
    }
    // Update MetamorphoSys version and build date
    properties.put("mmsys.version", MMSToolkit.getApplicationVersion());
    properties.put("mmsys.build.date", MMSToolkit.getApplicationBuildDate());
    return properties;
  }

  /**
   * Sets the location of the <cod>release.dat</code> file.
   * Opens the release.dat file and load the properties.
   * @param rd the location of the release.dat file
   * @throws IOException
   */
  public static void setReleaseDatLocation(String rd) throws IOException {
    release_dat = rd;
    File release_file = new File(release_dat);
    if (release_file.exists()) {
      // Open and load the properties file
      Properties props = new Properties();
      FileInputStream propInputStream = new FileInputStream(release_file);
      props.load(propInputStream);
      Enumeration keys = props.propertyNames();
      // Overwrite any properties in release_dat_properties
      while (keys.hasMoreElements()) {
        String key = (String) keys.nextElement();
        String value = props.getProperty(key);
        release_dat_properties.put(key, value);
      }
    }
  }

  /**
   * Copies the contents of the given {@link InputStream} to a new directory
   * with a given file name.  It treats the stream as a character stream
   * by wrapping it with
   * @param resource_stream {@link InputStream} of resource to be copied
   * @param write_file {@link String} name of subset file
   * @param subset_dir {@link String} subset directory
   * @throws IOException if anything goes wrong
   */
  public static void copyResourceToFile(InputStream resource_stream, String write_file, String subset_dir)
    throws  IOException {
    //
    // Open Reader/Writer
    //
    BufferedReader in = new BufferedReader(
        new InputStreamReader(resource_stream,"UTF-8"));
    BufferedWriter output_file =
        new BufferedWriter(
          new OutputStreamWriter(
            new FileOutputStream(
              new File(subset_dir, write_file)),
                  "UTF-8"));
    //
    // copy contents and close
    //
    String line = null;
    while ((line = in.readLine()) != null) {
      output_file.write(line);
      output_file.newLine();
      output_file.flush();
    }
    in.close();
    output_file.close();
  }

  /**
   * Sets a reference to the {@link MetamorphoSysGUI}.
   * @param gui the {@link MetamorphoSysGUI}
   */
  public static void setGui(MetamorphoSysGUI gui) {
    MMSToolkit.gui = gui;
  }

  /**
   * Returns the {@link MetamorphoSysGUI}.
   * @return the {@link MetamorphoSysGUI}
   */
  public static MetamorphoSysGUI getGui() {
    return gui;
  }

  /**
   * Determines whether or not a cancel operation should call System.Exit.
   */
  public static void cancel() {
    if (getGui() == null && !WelcomeGUI.getWelcomeGUI().isVisible()) {
      Exit(0);
    } else {
      return;
    }
  }

  /**
   * Splits a string on word boundaries so that each element of the string
   * array returned represents a "line" of the original string that is no
   * longer than the specified max line length. This is very useful for
   * breaking up long strings into 80 character chunks for displaying in a "
   * terminal.
   * @param to_split the {@link String} to split up
   * @param line_length the maximum line length, or <code>-1</code> to
   *        accept the default setting.
   * @return a {@link String}<code>[]</code> of the lines
   */
  public static String[] splitString(String to_split, int line_length) {
    // default max
    int max_length = 80;
    if (line_length > 0)
      max_length = line_length;
    String word = null;
    StringTokenizer st = new StringTokenizer(to_split, " ");
    String[] result = new String[1000];
    int first_word = 1;
    int char_count = 0;
    int i = 0;
    while (st.hasMoreTokens()) {
      word = st.nextToken();
      if (char_count + ((first_word > 0) ? 0 : 1) + word.length() < max_length) {
        if (result[i] == null)
          result[i] = "";
        result[i] = result[i] + ((first_word > 0) ? "" : " ") + word;
        char_count += ((first_word > 0) ? 0 : 1) + word.length();
      } else {
        if (first_word == 0)
          i++;
        result[i] = word;
        char_count = word.length();
      }
      if (first_word > 0)
        first_word = 0;
    }
    return result;
  }

  /**
   * Prints the stack trace from the given {@link Exception} to the
   * given {@link BufferedWriter}.
   * @param e {@link Exception}
   * @param b {@link BufferedWriter}
   */
  public static void printStackTrace(Exception e, BufferedWriter b) {
        StringWriter sw = new StringWriter();
        PrintWriter pw = new PrintWriter(sw);
        e.printStackTrace(pw);
        try {
          b.write(sw.toString());
          b.newLine();
        } catch (Exception ex) {
            MMSToolkit.reportError("Unable to print stack trace to BufferedWriter.");
        }
    }
 
  public static boolean readTar(InputStream in, String untarDir, 
  		final File tgz_file, final Frame owner) throws IOException{
    //
    // Declare variables
    //
    ProgressMonitor pm = null;
    boolean using_view = usingView();

    //
    // Set up progress monitor if using view
    //
    if (using_view) {
      String title = "Extracting " + tgz_file;
      String initial_msg = "Starting ... ";
      pm = new ProgressMonitor(owner, title, initial_msg, 0, 100);
      pm.setMillisToDecideToPopup(2);
      pm.setMaximum(100);
      pm.setProgress(1);
    } else {
      System.out.println("Extracting " + tgz_file);
    }
    
    // approximate compression at a factor of 3
    long tgz_file_length = tgz_file.length() * 3;
      
    TarInputStream tin = new TarInputStream(in);   
    TarEntry tarEntry = tin.getNextEntry();
    
    long byte_counter = 0;
    if (new File(untarDir).exists()) {
			while (tarEntry != null) {
        byte_counter += tarEntry.getSize();
 
        int progress = new Long((byte_counter * 100) / tgz_file_length).intValue();
        if (usingView()) {
          pm.setNote("Operation is " + progress + "% complete.");
          pm.setProgress(progress);
          if (pm.isCanceled())
          	return false;
        } else {
        	System.out.println("Operation is " + progress + "% complete.");
        }
        
				File destPath = new File(untarDir + File.separatorChar
						+ tarEntry.getName());
				if (!tarEntry.isDirectory()) {
					FileOutputStream fout = new FileOutputStream(destPath);
					tin.copyEntryContents(fout);
					fout.close();
				} else {
					destPath.mkdir();
				}
				tarEntry = tin.getNextEntry();
				
			  /*boolean ret_val = true;
		    if (using_view) {
		      ret_val = !pm.isCanceled();
		      pm.close();
		    }
		    return ret_val; */
			}
			tin.close();
			return true;
		} else {
			System.out.println("That destination directory doesn't exist! "
					+ untarDir);
			return false;
		}
    
 } 
} // end class MMSToolkit.java

