/**************************************************************************
 *
 * Package:	gov.nih.nlm.mms
 * Object:	MetamorphoSysGUI.java
 *
 **************************************************************************/
package gov.nih.nlm.mms;

import gov.nih.nlm.mms.browser.RRFBrowserGUI;
import gov.nih.nlm.mms.filters.Filter;
import gov.nih.nlm.mms.filters.SourcesToRemoveFilter;
import gov.nih.nlm.swing.DataChangeEvent;
import gov.nih.nlm.swing.DataChangeListener;
import gov.nih.nlm.swing.DecreaseFontAction;
import gov.nih.nlm.swing.FontSizeManager;
import gov.nih.nlm.swing.GlassComponent;
import gov.nih.nlm.swing.GlassPaneListener;
import gov.nih.nlm.swing.IncreaseFontAction;
import gov.nih.nlm.swing.XTabbedPane;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Stack;

import javax.swing.Box;
import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.JTextPane;
import javax.swing.KeyStroke;
import javax.swing.ScrollPaneConstants;
import javax.swing.border.CompoundBorder;
import javax.swing.border.EmptyBorder;
import javax.swing.border.EtchedBorder;
import javax.swing.border.TitledBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

/**
 * GUI application for configuring the {@link MetamorphoSys} subsetting
 * operation. This {@link JFrame} provides the means to completely configure a
 * subsetting run. The screen is built around a series of tabbed panes for each
 * configurable thing involved in the processing of a subset. A user can
 * configure input and ouptut options, as well as configuring the basic three
 * filters (sources, precedence, suppressibility). Users can enable inactive
 * filters or import new ones as well, which adds additional tabs with
 * configuration options.
 * 
 * The application is really a skeletal structure built around the
 * {@link GUIConfigurable} interface and roughly lets the plugged-in components
 * decide what the user is to do.
 * 
 * @author Deborah Shapiro
 */
public class MetamorphoSysGUI extends JFrame implements DataChangeListener,
    ChangeListener, GlassPaneListener {

  //
  // Private Fields
  //

  //
  // References to the configuration objects
  //
  private ApplicationConfiguration default_configuration = null;

  private UserConfiguration user_configuration = null;

  //
  // Reference to the application
  //
  private MetamorphoSys mms = null;

  //
  // Map of gui configurable objects to the components representing them
  //
  private Map filter_to_component_map = new HashMap();

  //
  // Map of tabbed panes to the corresponding GUIConfigurable
  //
  private Map component_to_gc_map = new HashMap();

  //
  // Undoable and redoable actions
  //
  private Stack undoable = new Stack();

  private Stack redoable = new Stack();

  //
  // List of data change listeners
  //
  private List listeners = new ArrayList();

  //
  // Reference to the previously selected tab index.
  //
  private int previous_tab = 0;

  //
  // state variable tracking whether or not to
  // pay atention to state changes (used for validating filters)
  //
  private boolean ignore_state_changed = false;

  //
  // GUI components
  //
  private XTabbedPane tabbed_pane = null;

  private JMenu reset_menu = null, help_menu = null, options_menu = null;

  private JMenuItem undo_menu_item = null, redo_menu_item = null,
      options_menu_input = null, options_menu_output = null;

  private JTextField cui_input_field = null, cui_output_field = null;
  private JPanel input_stream_content_panel = null, output_stream_content_panel = null;

  //
  // For glass pane
  //
  private GlassComponent gc = null;

  //
  // Constructor
  //

  /**
   * Instantiates a {@link MetamorphoSysGUI} and builds the frame.
   * 
   * @param mms
   *          the {@link MetamorphoSys} object, used for configuration should
   *          override settings in the configuration files.
   */
  public MetamorphoSysGUI(MetamorphoSys mms) {
    super();
    MMSToolkit.setGui(this);

    //
    // Handle glass pane
    //
    gc = new GlassComponent(this);
    setGlassPane(gc);
    
    //
    // Get configuration and filter references
    //
    this.mms = mms;
    default_configuration = mms.getApplicationConfiguration();
    user_configuration = mms.getUserConfiguration();

    //
    // If so indicated, the InstallationConfiguration settings
    // should override the configuration file settings
    //
    InstallConfiguration ic = InstallConfiguration.getInstallConfiguration();
    if (ic.getOverrideConfigurationFiles()) {

      //
      // Configure the META/ directory to subset into
      //
      if (ic.getMetaDestinationDirectory() != null) {
        user_configuration.setSubsetDirectory(ic.getMetaDestinationDirectory()
            .getAbsolutePath());
        default_configuration.setSubsetDirectory(ic
            .getMetaDestinationDirectory().getAbsolutePath());
        user_configuration.removeProperty(user_configuration.getOutputStream(),
            "subset_dir");
      }

      //
      // Configure the MetamorphoSys "config" dir
      //
      File config_dir = ic.getMetamorphoSysConfigurationDirectory();
      if (config_dir != null && config_dir.exists())
        default_configuration.setConfigDirectory(config_dir.getAbsolutePath());

      //
      // configure input/output streams
      //
      try {
        // if using a new config file, preserve output handler settings
          if (ic.getMetamorphoSysInputStream() instanceof NLMFileMetamorphoSysInputStream) {
            user_configuration
                .setOutputStream(new RichMRMetamorphoSysOutputStream());
          } else if (ic.getMetamorphoSysInputStream() instanceof RichMRMetamorphoSysInputStream) {
            user_configuration
                .setOutputStream(new RichMRMetamorphoSysOutputStream());
          } else {
            user_configuration
                .setOutputStream(new OriginalMRMetamorphoSysOutputStream());
          }

        if (ic.getMetamorphoSysInputStream() instanceof NLMFileMetamorphoSysInputStream) {
          user_configuration
              .setInputStream(new NLMFileMetamorphoSysInputStream());
        } else if (ic.getMetamorphoSysInputStream() instanceof RichMRMetamorphoSysInputStream) {
          user_configuration
              .setInputStream(new RichMRMetamorphoSysInputStream());
        } else {
          user_configuration
              .setInputStream(new OriginalMRMetamorphoSysInputStream());
        }

      } catch (Exception e) {
        e.printStackTrace();
        MMSToolkit.reportError("Problem configuring output handler.\n"
            + "The most likely problem is a missing \n"
            + "mr(plus)colsfiles.dat file in the config \n" + "directory.",
            true);
      }

      //
      // Configure source path
      //
      File install_file = ic.getUmlsDirectory();
      if (install_file != null && install_file.exists()) {
        user_configuration.clearSourcePaths();
        user_configuration.addSourcePath(install_file.getAbsolutePath());
        default_configuration.clearSourcePaths();
        default_configuration.addSourcePath(install_file.getAbsolutePath());
        user_configuration.removeProperty(user_configuration.getInputStream(),
            "source_path");
      }
    }

    //
    // Add metamorphosys as listner for the source pane
    //
    addSourcesToRemoveListener(mms);

    initialize();

    //
    // Give user opportunity to save
    //
    user_configuration.setPropertiesChanged();

    this.setVisible(true);
  }

  /**
   * Initializes the GUI.
   */
  public void initialize() {

    // Size is now set by the filter tables
    // setSize(800, 600);

    //
    // Build and configure tabbed pane
    // The change listener resonds to users switching tabs
    //
    JPanel base_panel = new JPanel(new BorderLayout());
    tabbed_pane = new XTabbedPane();
    tabbed_pane.addChangeListener(this);
    tabbed_pane.freezePane(0);
    tabbed_pane.freezePane(1);
    base_panel.setBorder(new EmptyBorder(20, 20, 20, 20));

    //
    // Compose menu bar
    //
    JMenuBar menu_bar = new JMenuBar();
    setJMenuBar(menu_bar);
    JMenu file_menu = new JMenu("File");
    menu_bar.add(file_menu);
    JMenu edit_menu = new JMenu("Edit");
    menu_bar.add(edit_menu);
    options_menu = new JMenu("Options");
    menu_bar.add(options_menu);
    reset_menu = new JMenu("Reset");
    menu_bar.add(reset_menu);
    JMenu done_menu = new JMenu("Done");
    menu_bar.add(done_menu);
    menu_bar.add(Box.createHorizontalGlue());
    help_menu = new JMenu("Help");
    menu_bar.add(help_menu);

    //
    // File Menu
    //

    //
    // Enable/Disable Filters
    //
    JMenuItem enable_filter_menu_item = new JMenuItem(new EnableFilterAction(
        this, mms));
    file_menu.add(enable_filter_menu_item);

    //
    // Import Filter
    //
    JMenuItem import_filter_menu_item = new JMenuItem(new ImportFilterAction(
        this, mms));
    file_menu.add(import_filter_menu_item);

    //
    // New Configuration
    //
    JMenuItem new_config_menu_item = new JMenuItem(new NewConfigAction(mms,
        this) {

      public void actionPerformed(ActionEvent ae) {
        InstallConfiguration.getInstallConfiguration()
            .setOverrideConfigurationFiles(true);
        super.actionPerformed(ae);
      }
    });
    new_config_menu_item.setAccelerator(KeyStroke.getKeyStroke('N',
        java.awt.Event.CTRL_MASK, false));
    file_menu.add(new_config_menu_item);

    //
    // Open Configuration
    //
    JMenuItem open_config_menu_item = new JMenuItem(new OpenConfigAction(mms,
        this) {

      public void actionPerformed(ActionEvent ae) {

        // Indicate override files, but set the new inputs and outputs
        InstallConfiguration ic = InstallConfiguration.getInstallConfiguration();
        ic.setOverrideConfigurationFiles(true);
        ic.setMetaDestinationDirectory(new File(user_configuration.getSubsetDirectory()));
        ic.setUmlsDirectory(new File(user_configuration.getSourcePaths()[0]));

        super.actionPerformed(ae);
      }
    });
    open_config_menu_item.setAccelerator(KeyStroke.getKeyStroke('O',
        java.awt.Event.CTRL_MASK, false));
    file_menu.add(open_config_menu_item);

    //
    // Save Configuration
    //
    JMenuItem save_config_menu_item = new JMenuItem(new SaveConfigAction(mms,
        this));
    save_config_menu_item.setAccelerator(KeyStroke.getKeyStroke('S',
        java.awt.Event.CTRL_MASK, false));
    file_menu.add(save_config_menu_item);

    //
    // Exit application
    //
    JMenuItem exit_menu_item = new JMenuItem(new ExitAction(mms));
    exit_menu_item.setAccelerator(KeyStroke.getKeyStroke('Q',
        java.awt.Event.CTRL_MASK, false));
    file_menu.add(exit_menu_item);

    //
    // Edit Menu
    //

    //
    // Undo
    //
    edit_menu.add(undo_menu_item = new JMenuItem(new UndoAction(this)));
    undo_menu_item.setAccelerator(KeyStroke.getKeyStroke('Z',
        java.awt.Event.CTRL_MASK, false));
    undo_menu_item.setEnabled(false);

    //
    // Redo
    //
    edit_menu.add(redo_menu_item = new JMenuItem(new RedoAction(this)));
    redo_menu_item.setAccelerator(KeyStroke.getKeyStroke('Y',
        java.awt.Event.CTRL_MASK, false));
    redo_menu_item.setEnabled(false);

    //
    // Increase font
    //
    JMenuItem if_item = new JMenuItem(new IncreaseFontAction());
    edit_menu.add(if_item);

    //
    // Decrease font
    //
    JMenuItem df_item = new JMenuItem(new DecreaseFontAction());
    edit_menu.add(df_item);

    //
    // Configuration Done Menu
    //

    //
    // Begin MetamorphoSys subset operation
    //
    JMenuItem begin_menu_item = new JMenuItem(new BeginSubsetAction(mms, this));
    done_menu.add(begin_menu_item);

    //
    // Help Menu
    //

    //
    // About the application
    //
    JMenuItem about_menu_item = new JMenuItem(new AboutAction(
        "About MetamorphoSys", this));
    help_menu.add(about_menu_item);

    //
    // Prepare options menu
    // The addFilterTab call below adds options
    // menu items for each of the active filters
    //
    JMenuItem application_options_item = new JMenuItem(
        new AdvancedOptionsAction(user_configuration));
    application_options_item.setEnabled(true);
    options_menu.add(application_options_item);

    //
    // Configure active filters and input/output streams.
    //
    GUIConfigurable gc = null;
    List gui_configurables = new ArrayList();
    gui_configurables = user_configuration.getActiveFilters();
    gui_configurables.add(0, user_configuration.getInputStream());
    gui_configurables.add(1, user_configuration.getOutputStream());

    for (int i = 0; i < gui_configurables.size(); i++) {
      gc = (GUIConfigurable) gui_configurables.get(i);
      initializeGUIConfigurable(gc);
    }

    //
    // Configure frame
    //
    base_panel.add(tabbed_pane, BorderLayout.CENTER);
    this.setTitle("UMLS Metathesaurus Configuration "
        + default_configuration.getReleaseVersion());
    this.setResizable(true);
    this.setContentPane(base_panel);

    //
    // Configure window listener
    //
    addWindowListener(new WindowAdapter() {
      public void windowClosing(WindowEvent e) {

        InstallConfiguration.getInstallConfiguration().clearProgressMonitor();
        MMSToolkit.exitToWelcomeGUI(MetamorphoSysGUI.this);
      }
    }

    );

    //
    // connect to font size manager, pack and make visible
    //
    FontSizeManager.addContainer(this);
    this.pack();
    this.setLocationRelativeTo(null);

  }

  //
  // Methods
  //
  /**
   * Configure the gui aspects of a {@link GUIConfigurable}.
   * 
   * @param gc
   *          {@link GUIConfigurable}
   */
  public void initializeGUIConfigurable(GUIConfigurable gc) {
    mms.initializeConfigurable(gc);

    addTab(gc);

    // Configure settings
    gc.setPanelSelections();
    gc.addDataChangeListener(this);

    if (gc instanceof SourcesToRemoveListener)
      addSourcesToRemoveListener((SourcesToRemoveListener) gc);
  }

  /**
   * Removes and returns top tuple off the undo stack.
   * 
   * @return an {@link Object}<code>[]</code> tuple from the undo stack
   */
  public Object[] popOffUndoStack() {
    Object[] ret_val = new Object[0];
    if (!undoable.isEmpty()) {
      ret_val = (Object[]) undoable.pop();
      if (!undoable.isEmpty()) {
        Object[] next = (Object[]) undoable.peek();
        undo_menu_item.setText(((UndoHandler) next[0]).getTitle());
      } else {
        undo_menu_item.setEnabled(false);
      }
    }
    return ret_val;
  }

  /**
   * Removes and returns top tuple off the redo stack.
   * 
   * @return an {@link Object}<code>[]</code> tuple from the redo stack
   */
  public Object[] popOffRedoStack() {
    Object[] ret_val = new Object[0];
    if (!redoable.isEmpty()) {
      ret_val = (Object[]) redoable.pop();
      if (!redoable.isEmpty()) {
        Object[] next = (Object[]) redoable.peek();
        redo_menu_item.setText(((RedoHandler) next[0]).getTitle());
      } else {
        redo_menu_item.setEnabled(false);
      }
    }
    return ret_val;
  }

  /**
   * Pushes an {@link Object}<code>[]</code> tuple onto the undo stack.
   * 
   * @param tuple
   *          an {@link Object}<code>[]</code> tuple
   */
  public void pushOnUndoStack(Object[] tuple) {
    undoable.push(tuple);
    undo_menu_item.setEnabled(true);
    undo_menu_item.setText(((UndoHandler) tuple[0]).getTitle());
  }

  /**
   * Pushes an {@link Object}<code>[]</code> tuple onto the redo stack.
   * 
   * @param tuple
   *          an {@link Object}<code>[]</code> tuple
   */
  public void pushOnRedoStack(Object[] tuple) {
    redoable.push(tuple);
    redo_menu_item.setEnabled(true);
    redo_menu_item.setText(((RedoHandler) tuple[0]).getTitle());
  }

  /**
   * Programatically selects the tab in the tabbed pane corresponding to the
   * specified {@link GUIConfigurable}.
   * 
   * @param gc
   *          the specified {@link GUIConfigurable}
   */
  public void selectTabForFilter(GUIConfigurable gc) {
    for (int i = 0; i < tabbed_pane.getTabCount(); i++) {
      if (tabbed_pane.getTitleAt(i).equals(getTitle(gc)))
        tabbed_pane.setSelectedIndex(i);
      if (gc instanceof MetamorphoSysInputStream)
        tabbed_pane.setSelectedIndex(tabbed_pane.indexOfTab("Input Options"));
      if (gc instanceof MetamorphoSysOutputStream)
        tabbed_pane.setSelectedIndex(tabbed_pane.indexOfTab("Output Options"));
    }
  }

  /**
   * Returns the GUI {@link Component} for the specified {@link Filter}. The
   * application maintains a private hash map that allows this mapping to take
   * place. If the map does not contain an entry for the filter, a panel is
   * built, added to the map and returned.
   * 
   * @param gui_configurable
   *          the specified {@link GUIConfigurable}
   * @return the GUI {@link Component} for the specified {@link Filter}.
   */
  public Component getComponentForTab(GUIConfigurable gui_configurable) {
    if (filter_to_component_map.containsKey(gui_configurable)) {
      return (Component) filter_to_component_map.get(gui_configurable);
    }

    //
    // Create the panel for the whole tab
    //
    JPanel tab_panel = new JPanel(new BorderLayout());

    //
    // Create the comment panel
    //
    JPanel comment_panel = new JPanel(new BorderLayout());

    //
    // Add the comments to the comment panel
    //
    comment_panel.add(makeComment(gui_configurable.getTabDescription()),
        BorderLayout.NORTH);

    //
    // Add the comments to the tab panel
    //
    tab_panel.add(comment_panel, BorderLayout.NORTH);

    //
    // Create the panel container for the GUIConfigurable panel and add it
    //
    JPanel gui_panel = new JPanel(new BorderLayout());
    gui_panel.add(gui_configurable.getPanel(), BorderLayout.CENTER);

    //
    // For input/ouptut options, add the chooser panel
    //
    if (gui_configurable instanceof MetamorphoSysInputStream) {
      comment_panel.add(getInputStreamChooserPanel(gui_panel),
          BorderLayout.CENTER);
      input_stream_content_panel = gui_panel;
    }
    if (gui_configurable instanceof MetamorphoSysOutputStream) {
      comment_panel.add(getOutputStreamChooserPanel(gui_panel),
          BorderLayout.CENTER);
      output_stream_content_panel = gui_panel;
    }

    //
    // Add GUI Panel (the preferred size of this controlls
    // the preferred size of the whole frame.
    //
    JScrollPane sp = new JScrollPane(gui_panel,
        ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED,
        ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
    sp.setPreferredSize(new Dimension(700, 500));
    tab_panel.add(sp, BorderLayout.CENTER);

    // add filter to map & return
    filter_to_component_map.put(gui_configurable, tab_panel);
    component_to_gc_map.put(tab_panel, gui_configurable);
    FontSizeManager.addContainer(tab_panel);
    return tab_panel;
  }

  /**
   * Adds a {@link GUIConfigurable} to the GUI. Calls methods to create the gui
   * configurable component, and to make reset, options and help menu items.
   * 
   * @param gc
   *          a {@link GUIConfigurable} to add to the GUI.
   */
  public void addTab(GUIConfigurable gc) {
    // Builds the gc panel if not already cached
    tabbed_pane.add(getComponentForTab(gc), getTitle(gc));

    // Options Menu
    if (gc.hasAdvancedOptions() || gc instanceof MetamorphoSysInputStream
        || gc instanceof MetamorphoSysOutputStream)
      options_menu.add(getOptionsMenuItem(gc));

    // Reset Menu (already handles font size manager)
    // don't add reset_item for output stream - it shares one with input stream
    if (!(gc instanceof MetamorphoSysOutputStream)) {
      JMenuItem reset_item = getResetMenuItem(gc);
      if (reset_item != null)
        reset_menu.add(reset_item);
    }

    // Help Menu (already handles fsm)
    if (gc.getHelpPanel() != null) {
      JMenuItem help_item = getHelpMenuItem(gc);
      if (help_item != null)
        help_menu.add(help_item);
    }
  }

  /**
   * Removes a {@link GUIConfigurable} tab from the GUI.
   * 
   * @param gc
   *          a {@link GUIConfigurable} to remove from the GUI.
   */
  public void removeTab(GUIConfigurable gc) {
    if (gc instanceof SourcesToRemoveListener)
      removeSourcesToRemoveListener((SourcesToRemoveListener) gc);

    tabbed_pane.remove(getComponentForTab(gc));

    // Remove options menu item
    for (int i = 0; i < options_menu.getItemCount(); i++) {
      JMenuItem menu_item = options_menu.getItem(i);
      if (menu_item.getText().indexOf(getTitle(gc)) != -1) {
        options_menu.remove(i);
        break;
      }
    }

    // Remove reset menu item
    for (int i = 0; i < reset_menu.getItemCount(); i++) {
      JMenuItem menu_item = reset_menu.getItem(i);
      if (menu_item.getText().indexOf(getTitle(gc)) != -1) {
        reset_menu.remove(i);
        break;
      }
    }

    // Remove help menu item
    for (int i = 0; i < help_menu.getItemCount(); i++) {
      JMenuItem menu_item = help_menu.getItem(i);
      if (menu_item.getText().indexOf(getTitle(gc)) != -1) {
        help_menu.remove(i);
        break;
      }
    }
  }

  /**
   * Confirms that {@link GUIConfigurable}s are in valid states. Notifies user
   * if a {@link GUIConfigurable} is not in a valid state and switches the tab
   * pane focus to the display for that {@link GUIConfigurable} so that it can
   * be corrected.
   * 
   * @return <code>true</code> if all {@link GUIConfigurable}s are valid
   *         <code>false</code> if any of the {@link GUIConfigurable}s are
   *         invalid
   */
  public boolean validateGUIConfigurables() {
    fireSourcesToRemoveChanged();

    List active_filters = user_configuration.getActiveFilters();
    Filter filter = null;
    for (int i = 0; i < active_filters.size(); i++) {
      filter = (Filter) active_filters.get(i);
      if (!handleValidation(filter))
        return false;
    }
    MetamorphoSysInputStream is = user_configuration.getInputStream();
    if (!is.validate()) {
      if (!handleValidation(is))
        return false;
    }
    MetamorphoSysOutputStream os = user_configuration.getOutputStream();
    if (!os.validate()) {
      if (!handleValidation(os))
        return false;
    }

    return true;
  }

  /**
   * Handle validation errors and warnings appropriately for the given
   * {@link GUIConfigurable}.
   * 
   * @param gui_config
   *          {@link GUIConfigurable}
   * @return <code>boolean</code>
   */
  private boolean handleValidation(GUIConfigurable gui_config) {
    if (!gui_config.validate()) {
      // need to ignore ChangeEvents when moving back to previous
      // tab, so that the user notification does not pop-up twice
      ignore_state_changed = true;
      tabbed_pane.setSelectedIndex(previous_tab);
      ignore_state_changed = false;
      MMSToolkit.notifyUser(this, getValidationReport(gui_config
          .getValidationErrors()));
      return false;
    } else if (gui_config.getValidationWarnings() != null
        && gui_config.getValidationWarnings().length > 0) {
      if (MMSToolkit.confirmRequestWithOption(null, getValidationWarningReport(
          gui_config.getValidationWarnings(), gui_config.getName()),
          "validate_" + gui_config.getName())) {

        GUIConfigurable gc = gui_config
            .getGUIConfigurableForValidationFailure();

        // need to ignore ChangeEvents when moving back to previous
        // tab, so that the user notification does not pop-up twice
        ignore_state_changed = true;
        selectTabForFilter(gc);
        ignore_state_changed = false;
        return false;
      }
    }
    return true;
  }

  /**
   * Returns a report of the validation issues to be displayed to the user.
   * 
   * @param issues
   *          warnings or errors
   * @return {@link String} report
   */
  private String getValidationReport(String[] issues) {
    StringBuffer sb = new StringBuffer();
    for (int i = 0; i < issues.length; i++) {
      sb.append(issues[i]).append("\n\n");
    }
    return sb.toString();
  }

  /**
   * Returns a report of the validation warnings with a user query appended to
   * the end.
   * 
   * @param issues
   *          warnings
   * @param title
   *          name of {link GUIConfigurable} causing warnings
   * @return {@link String} report
   */
  private String getValidationWarningReport(String[] issues, String title) {
    return title + " Warning" + ((issues.length > 1) ? "s\n\n" : "\n\n")
        + getValidationReport(issues)
        + "Would you like to resolve the warning issue(s)?\n\n";
  }

  //
  // Accessor methods
  //

  /**
   * Adds the specified listener to the list of listeners to be informed when
   * the sources to remove list is changed.
   * 
   * @param l
   *          the {@link SourcesToRemoveListener}
   */
  public void addSourcesToRemoveListener(SourcesToRemoveListener l) {
    listeners.add(l);
  }

  /**
   * Removes the specfied listener from the list of listeners to be informed
   * when the sources to remove list is changed.
   * 
   * @param l
   *          the {@link SourcesToRemoveListener}
   */
  public void removeSourcesToRemoveListener(SourcesToRemoveListener l) {
    listeners.remove(l);
  }

  /**
   * Returns the {@link UserConfiguration}.
   * 
   * @return the {@link UserConfiguration}
   */
  public UserConfiguration getUserConfiguration() {
    return user_configuration;
  }

  /**
   * Returns the {@link ApplicationConfiguration}.
   * 
   * @return the {@link ApplicationConfiguration}
   */
  public ApplicationConfiguration getApplicationConfiguration() {
    return default_configuration;
  }

  //
  // ChangeListener Implementation
  //

  /**
   * Responds to a user switching tabs in the tabbed pane. When this happens, we
   * validate the filter being switched away from and prevent the action if
   * validation fails.
   * 
   * @param e
   *          a {@link ChangeEvent}
   */
  public void stateChanged(ChangeEvent e) {
    if (!ignore_state_changed) {
      // Always fire this, it is safer that way
      fireSourcesToRemoveChanged();
      XTabbedPane tabbed_pane = ((XTabbedPane) e.getSource());
      GUIConfigurable gc = (GUIConfigurable) component_to_gc_map
          .get(tabbed_pane.getComponentAt(previous_tab));
      if (!handleValidation(gc)) {
        return;
      }

      // update the filter order based on user reorganizing filter tabs
      XTabbedPane xtabbed_pane = (XTabbedPane) e.getSource();
      int last_drag_index = xtabbed_pane.getLastDragIndex();
      int last_tab_index = xtabbed_pane.getLastTabIndex();
      boolean drag_completed = xtabbed_pane.isDragCompleted();

      if (last_drag_index != -1 && last_tab_index != -1 && last_drag_index >= 2
          && last_tab_index >= 2 && drag_completed) {
        user_configuration.reorderActiveFilters(last_drag_index - 2,
            last_tab_index - 2);
      }

    }
    // save reference to previous tab
    previous_tab = ((XTabbedPane) e.getSource()).getSelectedIndex();
  }

  /**
   * Sets the gui to ignore reponding to {@link ChangeEvent}s.
   * 
   * @param b
   *          <code>boolean</code>
   */
  public void setIgnoreEvents(boolean b) {
    ignore_state_changed = b;
  }

  //
  // DataChangeListener Implementation
  //

  /**
   * Responds to a change in the user configuration for one of the active
   * filters. The main function is to push the filter state onto the undo stack
   * and mark the user configuration as changed.
   * 
   * @param e
   *          a {@link DataChangeEvent}
   */
  public void dataChanged(DataChangeEvent e) {
    GUIConfigurable gc = (GUIConfigurable) e.getSource();
    Properties p = (Properties) gc.getProperties().clone();
    Boolean hdc = new Boolean(gc.hasDataChanged());
    Object[] tuple = { new GUIConfigurableUndoHandler(gc, this), p, hdc };
    pushOnUndoStack(tuple);
    redoable.clear();
    redo_menu_item.setEnabled(false);
  }

  //
  // Protected Methods
  //

  /**
   * Fires a {@link SourcesToRemoveEvent} to inform all listening filters to
   * respond to the change in the sources to exclude list.
   */
  protected void fireSourcesToRemoveChanged() {
    SourcesToRemoveFilter sources_to_exclude_filter = null;
    List active_filters = user_configuration.getActiveFilters();
    for (int i = 0; i < active_filters.size(); i++) {
      Filter filter = (Filter) active_filters.get(i);
      if (filter instanceof SourcesToRemoveFilter)
        sources_to_exclude_filter = (SourcesToRemoveFilter) filter;
    }

    SourcesToRemoveEvent e = new SourcesToRemoveEvent(this,
        sources_to_exclude_filter.getSourcesToRemove());
    Iterator iter = listeners.iterator();
    while (iter.hasNext()) {
      SourcesToRemoveListener strl = (SourcesToRemoveListener) iter.next();
      strl.sourcesToRemoveChanged(e);
    }
  }

  //
  // Private Methods
  //

  private JPanel getInputStreamChooserPanel(JPanel content_panel) {
    // Configure panel
    JPanel files_panel = new JPanel(new BorderLayout());

    // Configure border
    TitledBorder titledBorder = new TitledBorder(new EtchedBorder(
        EtchedBorder.RAISED), "Input Format Options");
    titledBorder.setTitleJustification(TitledBorder.CENTER);
    files_panel.setBorder(new CompoundBorder(titledBorder, new EmptyBorder(10,
        10, 10, 10)));

    //
    // Add label
    //
    JLabel cui_input_label = new JLabel("Select Input Format");
    cui_input_label.setForeground(Color.black);
    files_panel.add(cui_input_label, BorderLayout.NORTH);

    //
    // Add text field
    //
    JPanel field_button_panel = new JPanel(new FlowLayout());
    cui_input_field = new JTextField(user_configuration.getInputStream()
        .getName(), 50);
    cui_input_field.setEditable(false);
    cui_input_field.setBackground(Color.white);
    cui_input_field.setForeground(user_configuration.getInputStream()
        .getColor());
    field_button_panel.add(cui_input_field);

    //
    // Add browse button
    //
    JButton cui_input_button = new JButton(new ConceptInputStreamAction(this,
        cui_input_field, content_panel));
    cui_input_button.setText("Browse..");
    field_button_panel.add(cui_input_button);

    files_panel.add(field_button_panel, BorderLayout.CENTER);
    return files_panel;
  }

  /**
   * Returns a panel for selecting output stream.
   * 
   * @param content_panel
   *          {@link JPanel}
   * @return {@link JPanel} selecting output stream panel
   */
  private JPanel getOutputStreamChooserPanel(JPanel content_panel) {

    JPanel files_panel = new JPanel(new BorderLayout());

    //
    // Configure border
    //
    TitledBorder titledBorder = new TitledBorder(new EtchedBorder(
        EtchedBorder.RAISED), "Output Format Options");
    titledBorder.setTitleJustification(TitledBorder.CENTER);
    files_panel.setBorder(new CompoundBorder(titledBorder, new EmptyBorder(10,
        10, 10, 10)));

    //
    // Add label
    //
    JLabel cui_output_label = new JLabel("Select Output Format");
    cui_output_label.setForeground(Color.black);
    files_panel.add(cui_output_label, BorderLayout.NORTH);

    //
    // Add text field
    //
    JPanel field_button_panel = new JPanel(new FlowLayout());
    cui_output_field = new JTextField(user_configuration.getOutputStream()
        .getName(), 50);
    cui_output_field.setEditable(false);
    cui_output_field.setBackground(Color.white);
    cui_output_field.setForeground(user_configuration.getOutputStream()
        .getColor());
    field_button_panel.add(cui_output_field);

    //
    // Add button
    //
    JButton cui_output_button = new JButton(new ConceptOutputStreamAction(this,
        cui_output_field, content_panel));
    cui_output_button.setText("Browse..");
    field_button_panel.add(cui_output_button);

    // Compose the panel and return it
    files_panel.add(field_button_panel, BorderLayout.CENTER);
    return files_panel;
  }

  /**
   * Helper method to make the comment area for a filter.
   * 
   * @param msg
   *          the lines of the message
   * @return a {@link JScrollPane} with the comments on it
   */
  private JScrollPane makeComment(String msg) {
    JTextPane area = new JTextPane();
    area.setContentType("text/html");

    area.setText("<html><font style=\"font-family:sans-serif;font-size:85%\">"
        + msg.replaceAll("\n", "<br>") + "</font></html>");
    area.setEditable(false);

    area.setBackground((new JPanel()).getBackground());
    area.setBorder(new EmptyBorder(5, 5, 5, 5));
    area.setPreferredSize(new JTextArea(6, 50).getPreferredSize());
    JScrollPane scroll = new JScrollPane(area,
        ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED,
        ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
    scroll.setPreferredSize(area.getPreferredSize());
    area.setCaretPosition(1);
    area.setCaretPosition(0);
    return (scroll);
  }

  /**
   * Helper method for building a <code>Help</code> menu item for a particular
   * {@link GUIConfigurable}.
   * 
   * @param gc
   *          a {@link GUIConfigurable} from which to build a help menu item
   * @return a {@link JMenuItem} to add to the <code>Help</code> menu
   */
  private JMenuItem getHelpMenuItem(GUIConfigurable gc) {
    final GUIConfigurable fgc = gc;

    //
    // Create menu item
    //
    JMenuItem help_item = new JMenuItem(getTitle(gc) + " Help");

    //
    // Add action listener (it opens the help frame)
    //
    help_item.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        JPanel help_panel = null;
        if (fgc instanceof MetamorphoSysInputStream)
          help_panel = user_configuration.getInputStream().getHelpPanel();
        else if (fgc instanceof MetamorphoSysOutputStream)
          help_panel = user_configuration.getOutputStream().getHelpPanel();
        else
          help_panel = fgc.getHelpPanel();

        //
        // If no help, say no help
        //
        if (help_panel == null) {
          MMSToolkit.notifyUser(MetamorphoSysGUI.this, "No help available.");
          return;
        }

        help_panel.setBackground(Color.gray);
        final JPanel base_panel = new JPanel(new BorderLayout());
        base_panel.setBorder(new EmptyBorder(10, 10, 10, 10));
        base_panel.add(help_panel, BorderLayout.CENTER);

        final JFrame help_frame = new JFrame(getTitle(fgc));

        JButton done_button = new JButton("Done");
        done_button.addActionListener(new ActionListener() {
          public void actionPerformed(ActionEvent e) {
            help_frame.dispose();
          }
        });
        getRootPane().setDefaultButton(done_button);
        base_panel.add(done_button, BorderLayout.SOUTH);
        FontSizeManager.addContainer(base_panel);
        help_frame.setContentPane(base_panel);
        help_frame.pack();
        done_button.requestFocus();
        help_frame.setLocationRelativeTo(null);
        help_frame.setVisible(true);
      }
    });

    JPanel wrapper = new JPanel();
    wrapper.add(help_item);
    FontSizeManager.addContainer(wrapper);
    return help_item;
  }

  /**
   * Helper method for building a <code>Reset</code> menu item for a
   * particular {@link GUIConfigurable}.
   * 
   * @param gconfig
   *          a {@link GUIConfigurable} from which to build a reset menu item
   * @return a {@link JMenuItem} to add to the <code>Reset</code> menu
   */
  private JMenuItem getResetMenuItem(GUIConfigurable gconfig) {
    final GUIConfigurable gc = gconfig;
    JMenuItem reset_item = new JMenuItem();
    if (gc instanceof MetamorphoSysInputStream ||
    		gc instanceof MetamorphoSysOutputStream) {
      reset_item.setText("Reset Input/Output Options");
      reset_item.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
        	// Reset INPUT stream
          // fake a data change event for the undo stack
          //MetamorphoSysGUI.this.dataChanged(new DataChangeEvent(
            //  user_configuration.getInputStream()));
          MetamorphoSysInputStream new_input_stream =
          	InstallConfiguration.getInstallConfiguration().getMetamorphoSysInputStream();
          new_input_stream.setConfiguration(default_configuration);
          changeInputStream(new_input_stream,
          		cui_input_field, input_stream_content_panel);
          //MetamorphoSysGUI.this.dataChanged(new DataChangeEvent(
          //    user_configuration.getInputStream()));
          user_configuration.getInputStream().resetToDefaults();
          user_configuration.getInputStream().setPanelSelections();
          
          // Reset OUTPUT stream
          //MetamorphoSysGUI.this.dataChanged(new DataChangeEvent(
            //  user_configuration.getOutputStream()));
          MetamorphoSysOutputStream new_output_stream = null;
          try {
						if (new_input_stream instanceof OriginalMRMetamorphoSysInputStream)
							new_output_stream = new OriginalMRMetamorphoSysOutputStream();
						else
							new_output_stream = new RichMRMetamorphoSysOutputStream();
					} catch (Exception ex) {
						ex.printStackTrace();
						MMSToolkit.reportError("Problem reseting output stream.\n");
					}
          new_input_stream.setConfiguration(default_configuration);
          //MetamorphoSysGUI.this.dataChanged(new DataChangeEvent(
          //    user_configuration.getOutputStream()));
          changeOutputStream(new_output_stream,
          		cui_output_field, output_stream_content_panel);
          user_configuration.getOutputStream().resetToDefaults();
          user_configuration.getOutputStream().setPanelSelections();
        }
      });
    } else {
      reset_item.setText("Reset " + getTitle(gc));
      reset_item.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          // fake a data change event for the undo stack
          MetamorphoSysGUI.this.dataChanged(new DataChangeEvent(gc));
          gc.resetToDefaults();
          gc.setPanelSelections();
        }
      });
    }

    JPanel wrapper = new JPanel();
    wrapper.add(reset_item);

    FontSizeManager.adjustFontSize(wrapper);

    return reset_item;
  }

  /**
   * Returns the title for the specified {@link GUIConfigurable}. It is
   * determined based on what type of object the thing is. It will be either a
   * {@link Filter} a {@link MetamorphoSysInputStream}, or a
   * {@link MetamorphoSysOutputStream}.
   * 
   * @param gc
   *          the specified {@link GUIConfigurable}
   * @return the title for the specified {@link GUIConfigurable}
   */
  private String getTitle(GUIConfigurable gc) {
    if (gc instanceof MetamorphoSysInputStream)
      return "Input Options";
    else if (gc instanceof MetamorphoSysOutputStream)
      return "Output Options";
    else if (gc instanceof Filter)
      return gc.getName();

    return "";
  }

  /**
   * Helper method for building a <code>Options</code> menu item for a
   * particular {@link GUIConfigurable}.
   * 
   * @param gc
   *          a {@link GUIConfigurable} from which to build a options menu item
   * @return a {@link JMenuItem} to add to the <code>Options</code> menu
   */
  private JMenuItem getOptionsMenuItem(GUIConfigurable gc) {
    JMenuItem options_item = new JMenuItem();
    final GUIConfigurable fgc = gc;
    options_item.setText("Advanced " + getTitle(gc) + " Options");
    JPanel wrapper = new JPanel();
    wrapper.add(options_item);
    FontSizeManager.adjustFontSize(wrapper);
    if (gc instanceof MetamorphoSysInputStream) {
      options_item.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          user_configuration.getInputStream().showOptions();
        }
      });
      options_menu_input = options_item;
      options_menu_input.setEnabled(gc.hasAdvancedOptions());
    } else if (gc instanceof MetamorphoSysOutputStream) {
      options_item.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          user_configuration.getOutputStream().showOptions();
        }
      });
      options_menu_output = options_item;
      options_menu_output.setEnabled(gc.hasAdvancedOptions());
    } else {
      options_item.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          fgc.showOptions();
        }
      });
    }

    return options_item;
  }

  /**
   * Enables or disables options menu items for input and output handlers.
   */
  public void checkEnableOptionsMenu() {
    options_menu_input.setEnabled(user_configuration.getInputStream()
        .hasAdvancedOptions());
    options_menu_output.setEnabled(user_configuration.getOutputStream()
        .hasAdvancedOptions());
  }

  /**
   * Changes the title of a {@link JPanel} with a compound, titled border.
   * 
   * @param panel
   *          a {@link JPanel} with a compound, titled border
   * @param title
   *          the new title
   */
  public void resetTitleForFilter(JPanel panel, String title) {
    CompoundBorder compound_border = (CompoundBorder) panel.getBorder();
    TitledBorder titled_border = (TitledBorder) compound_border
        .getOutsideBorder();
    titled_border.setTitle(title);

    //
    // Make sure that the border is updated immediately
    //
    panel.repaint();
  }

  /**
   * Changes the title of a tab on the {@link XTabbedPane}.
   * 
   * @param filter
   *          {@link Filter} whose tab title will be changed
   * @param title
   *          the new title for the tab
   */
  public void resetTabTitleForFilter(Filter filter, String title) {
    Component tabbed_pane_component = getComponentForTab(filter);
    tabbed_pane.setTitleAt(tabbed_pane.indexOfComponent(tabbed_pane_component),
        title);
  }

  /**
   * Handles a cancel operation from the progress monitor.
   */
  public void cancel() {
    try {
      // close input stream in case it is using native unzip
      user_configuration.getInputStream().close();
      if (mms.getLog() != null) {
        mms.getLog().write("MetamorphoSys was cancelled.");
        mms.getLog().flush();
        mms.getLog().close();
      }
    } catch (Exception e) {
      // too late to worry about it, don't bother.
    }
  }

  /**
   * Updates the {@link MetamorphoSysInputStream} on the gui.
   * 
   * @param input_stream
   *          {@link MetamorphoSysInputStream}
   * @param misf
   *          {@link JTextField} where name of input stream is displayed
   * @param content_panel
   *          {@link JPanel} associated with the
   *          {@link MetamorphoSysInputStream}
   */
  public void changeInputStream(MetamorphoSysInputStream input_stream,
      JTextField misf, JPanel content_panel) {

    GUIConfigurable old_input_stream = user_configuration.getInputStream();

    if (user_configuration.getInputStream() instanceof SourcesToRemoveListener)
      removeSourcesToRemoveListener((SourcesToRemoveListener) user_configuration
          .getInputStream());

    Component c = (Component) filter_to_component_map.get(old_input_stream);

    //
    // Set user configuration input stream
    //
    user_configuration.setInputStream(input_stream);
    input_stream.setConfiguration(default_configuration);

    component_to_gc_map.put(c, input_stream);

    filter_to_component_map.put(input_stream, c);

    //
    // Set text field name and color
    //
    misf.setText(input_stream.getName());
    misf.setForeground(user_configuration.getInputStream().getColor());

    //
    // Check if we need to enable options
    //
    checkEnableOptionsMenu();

    //
    // update the tab gui for the new input stream
    //
    content_panel.remove(0);
    content_panel.add(user_configuration.getInputStream().getPanel());

    if (user_configuration.getInputStream() instanceof SourcesToRemoveListener)
      addSourcesToRemoveListener((SourcesToRemoveListener) user_configuration
          .getInputStream());
    input_stream.setProperties(user_configuration.getProperties(input_stream));
    input_stream.setPanelSelections();
    validate();
  }

  /**
   * Updates the {@link MetamorphoSysOutputStream} on the gui.
   * 
   * @param output_stream
   *          {@link MetamorphoSysOutputStream}
   * @param mosf
   *          {@link JTextField} where name of output stream is displayed
   * @param content_panel
   *          {@link JPanel} associated with the
   *          {@link MetamorphoSysOutputStream}
   */
  public void changeOutputStream(MetamorphoSysOutputStream output_stream,
      JTextField mosf, JPanel content_panel) {
    GUIConfigurable old_output_stream = user_configuration.getOutputStream();

    if (user_configuration.getOutputStream() instanceof SourcesToRemoveListener)
      removeSourcesToRemoveListener((SourcesToRemoveListener) user_configuration
          .getOutputStream());

    Component c = (Component) filter_to_component_map.get(old_output_stream);

    //
    // Set user configuration output stream
    //
    user_configuration.setOutputStream(output_stream);
    output_stream.setConfiguration(default_configuration);

    component_to_gc_map.put(c, output_stream);

    filter_to_component_map.put(output_stream, c);

    //
    // Set text field name and color
    //
    mosf.setText(output_stream.getName());
    mosf.setForeground(user_configuration.getOutputStream().getColor());

    //
    // Check if we need to enable options
    //
    checkEnableOptionsMenu();

    //
    // update the tab gui for the new output stream
    //
    content_panel.remove(0);
    content_panel.add(user_configuration.getOutputStream().getPanel());

    if (user_configuration.getOutputStream() instanceof SourcesToRemoveListener)
      addSourcesToRemoveListener((SourcesToRemoveListener) user_configuration
          .getOutputStream());
    output_stream
        .setProperties(user_configuration.getProperties(output_stream));
    output_stream.setPanelSelections();

    validate();
  }
  
  //
  // GlassPaneListener Implementation
  //
  /**
   * Enable glass pane.
   */
  public void enableGlassPane() {
    gc.setVisible(true);
  }

  /**
   * Disable glass pane.
   */
  public void disableGlassPane() {
    gc.setVisible(false);
  }
  
}
