/***************************************************************************
 *
 * Package:     gov.nih.nlm.mms
 * Interface:   MetamorphoSysInputStream.java
 *
 **************************************************************************/
package gov.nih.nlm.mms;

import gov.nih.nlm.umls.io.MetathesaurusInputStream;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.IndexEntry;

/**
 * Generically represents an input source for Metathesaurus data.
 *
 * @author Deborah Shapiro
 */
public interface MetamorphoSysInputStream extends MetathesaurusInputStream,
   GUIConfigurable, SubsetListener {

 /**
  * Return an <code>int</code> between 0 and 100 indicating
  * how far along the input stream is in reading {@link Concept}s.
  * This information can be used for a progress monitor.
  * @return <code>int</code> representing percentage complete
  */
 public int getConceptProgress();

 /**
  * Return an <code>int</code> between 0 and 100 indicating
  * how far along the input stream is in reading {@link IndexEntry}
  * elements. This information can be used for a progress monitor.
  * @return <code>int</code> representing percentage complete
  */
 public int getIndexProgress();

  /**
   * Validates the source path.  This involves
   * checking that it both exists, is readable and that it has
   * access to MR data.
   * @param source_path path to be validated
   * @return <code>true</code> if validation passes
   *         <code>false</code> otherwise
   */
  public boolean validateSourcePath(String source_path);

  /**
   * Validates the source paths.  This involves
   * checking that they both exist, are readable and that they have
   * access to MR data.
   * @param source_paths to be validated
   * @return <code>true</code> if validation passes
   *         <code>false</code> otherwise
   */
  public boolean validateSourcePaths(String[] source_paths);

  /**
   * For each source path it attempts to determine the version of the release being
   * used and validate it against the version from the properties file.
   * If the version is determined to be a mismatch or if it cannot
   * be determined, <code>false</code> is returned.
   * @param source_paths to be validated
   * @return <code>true</code> if validation passes
   *         <code>false</code> otherwise
   */
  public boolean validateVersions(String[] source_paths);

  /**
    * Attempts to determine the version of the release being
    * used and validate it against the version from the properties file.
    * If the version is determined to be a mismatch or if it cannot
    * be determined, <code>false</code> is returned.
    * @param path source path to be validated
   * @return <code>true</code> if validation passes
   *         <code>false</code> otherwise
    */
  public boolean validateVersion(String path);

  /**
   * Indicates the type of file accepted by the file chooser.
   * @return {@link String}
   */
  public int getChooserType();

  /**
   * Sets the subsetting mode to efficient.  In the efficient mode, the
   * "Initialize Cuis" phase of the subsetting only references MRCONSO and
   * MRSTY.
   * @param efficient <code>true</code> if subsetting should be run in efficient mode
   *                  <code>false</code> if not
   */
  public void setEfficientMode(boolean efficient);

  /**
   * Returns the mode in which subsetting will be run.  In the efficient mode, the
   * "Initialize Cuis" phase of the subsetting only references MRCONSO and
   * MRSTY.
   * @return <code>true</code> if subsetting will be run in efficient mode
   *         <code>false</code> otherwise
   */
  public boolean isEfficientMode();

  /**
   * Indicates whether this input source has full context data, such as that
   * included in a MRCXT file.
   *
   * @return <code>true</code> if this input source has full context data;
   *         <code>false</code> otherwise.
   */
  boolean hasFullContextData();
}
