/**************************************************************************
 *
 * Package:     gov.nih.nlm.mms
 * Object:      OriginalMRMetamorphoSysOutputStream.java
 *
 **************************************************************************/
package gov.nih.nlm.mms;

import gov.nih.nlm.swing.DataChangeEvent;
import gov.nih.nlm.swing.DataChangeListener;
import gov.nih.nlm.swing.IntegerDocument;
import gov.nih.nlm.swing.SwingToolkit;
import gov.nih.nlm.umls.io.InputStreamFileStatisticsConfiguration;
import gov.nih.nlm.umls.io.OriginalMRMetathesaurusOutputStream;
import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.Attribute;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.Context;
import gov.nih.nlm.umls.meta.ContextMember;
import gov.nih.nlm.umls.meta.Definition;
import gov.nih.nlm.umls.meta.IndexEntry;
import gov.nih.nlm.umls.meta.Language;
import gov.nih.nlm.umls.meta.SourceTermType;
import gov.nih.nlm.util.ColumnStatistics;
import gov.nih.nlm.util.FileStatistics;
import gov.nih.nlm.util.SystemToolkit;
import gov.nih.nlm.util.UTF8OutputStreamWriter;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.FlowLayout;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.border.CompoundBorder;
import javax.swing.border.EmptyBorder;
import javax.swing.border.EtchedBorder;
import javax.swing.border.TitledBorder;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.text.Document;

/**
 * Represents a {@link MetamorphoSysOutputStream} that writes a directory
 * containing Original Release Format Metathesaurus files.
 * 
 * @author Deborah Shapiro
 */
public class OriginalMRMetamorphoSysOutputStream extends
    OriginalMRMetathesaurusOutputStream implements MetamorphoSysOutputStream {

  //
  // Protected Fields
  //

  //
  // The user and application configuration objects
  //
  protected UserConfiguration user_config;

  protected ApplicationConfiguration default_config;

  //
  // tracks whether or not data has changed
  //
  protected boolean data_changed = false;

  //
  // List of data change listeners
  //
  protected List listeners = new ArrayList();

  //
  // Validation message
  //
  protected String validate_msg;

  //
  // state variables
  //
  private boolean ignore_events = false, truncate = false,
      remove_unicode = false, calculate_md5s = false, remove_mth_only = false,
      write_oracle = false, write_mysql = false, no_mrcxt_file = false;

  private int max_field_length = 4000;

  //
  // GUI elements
  //
  private JTextField max_length_text_field = null;

  private JCheckBox truncation_button = null;

  private JCheckBox versioned_button = null;

  private JCheckBox remove_unicode_checkbox = null;

  private JCheckBox add_unicode_bom_checkbox = null;

  private JCheckBox calculate_md5s_checkbox = null;

  private JCheckBox remove_mth_only_checkbox = null;

  private JCheckBox write_oracle_load_checkbox = null;

  private JCheckBox write_mysql_load_checkbox = null;

  private JTextField sd_field = null;

  //
  // Map of MRFiles to their MD5 digests
  //
  private Map mrfiles_to_md5s = null;

  //
  // Bytes for \n character, used in cross-platform MD5 calculations
  //
  private final byte[] md5_lt = "\n".getBytes("UTF-8");
 
  private static HashMap files_to_index_fields = new HashMap();
  
  static {
  	// Index the following fields
  	// Note: Indexes on more than one column must be without a space
  	// such as "CUI,SUI"
  	Object[] mrcon_indexes = { "CUI", "SUI" };
  	files_to_index_fields.put("MRCON", Arrays.asList(mrcon_indexes));
  	Object[] mrso_indexes = { "CUI", "SUI", "SAB", "CODE" };
  	files_to_index_fields.put("MRSO", Arrays.asList(mrso_indexes));
  	Object[] mrsty_indexes = { "CUI" };
  	files_to_index_fields.put("MRSTY", Arrays.asList(mrsty_indexes));
  	Object[] mrdef_indexes = { "CUI", "SAB" };
  	files_to_index_fields.put("MRDEF", Arrays.asList(mrdef_indexes));
  	Object[] mrrank_indexes = { "SAB" };
  	files_to_index_fields.put("MRRANK", Arrays.asList(mrrank_indexes));
  	Object[] mrrel_indexes = { "CUI1", "CUI2", "SAB" };
  	files_to_index_fields.put("MRREL", Arrays.asList(mrrel_indexes));
  	Object[] mrsab_indexes = {"RSAB"};
  	files_to_index_fields.put("MRSAB", Arrays.asList(mrsab_indexes));
  	Object[] mrcoc_indexes = { "CUI1", "CUI2", "SOC" };
  	files_to_index_fields.put("MRCOC", Arrays.asList(mrcoc_indexes));
  	Object[] mrcxt_indexes = { "CUI,SUI", "SAB" };
  	files_to_index_fields.put("MRCXT", Arrays.asList(mrcxt_indexes));
  	Object[] mrsat_indexes = { "CUI", "ATN", "SAB" };
  	files_to_index_fields.put("MRSAT", Arrays.asList(mrsat_indexes));
  	Object[] mratx_indexes = { "CUI", "SAB" };
  	files_to_index_fields.put("MRATX", Arrays.asList(mratx_indexes));
  	Object[] mrxnw_eng_indexes = { "NWD" };
  	files_to_index_fields.put("MRXNW.ENG", Arrays.asList(mrxnw_eng_indexes));
  	Object[] mrxns_eng_indexes = { "NSTR" };
  	files_to_index_fields.put("MRXNS.ENG", Arrays.asList(mrxns_eng_indexes));
  	files_to_index_fields.put("AMBIG.SUI", Arrays
  			.asList(new Object[] { "SUI" }));
  	files_to_index_fields.put("AMBIG.LUI", Arrays
  			.asList(new Object[] { "LUI" }));
  }

  
  //
  // Constructor
  //

  /**
   * Instantiates an empty {@link OriginalMRMetamorphoSysOutputStream}.
   * 
   * @throws IOException
   *           if anything goes wrong
   */
  public OriginalMRMetamorphoSysOutputStream() throws IOException {
    super(new InputStreamFileStatisticsConfiguration(
        (InputStream) RichMRMetamorphoSysOutputStream.class.getClassLoader()
            .getResource("config/mrcolsfiles.dat").getContent()));
  }

  //
  // Methods
  //

  /**
   * Returns the short name for display purposes.
   * 
   * @return the short name for display purposes.
   */
  public String getName() {
    return "Original Release Format";
  }

  /**
   * Sets the {@link UserConfiguration}.
   * 
   * @param user_config
   *          the {@link UserConfiguration}.
   */
  public void setConfiguration(UserConfiguration user_config) {
    this.user_config = user_config;
  }

  /**
   * Provides output stream access to the cui list.
   * 
   * @param cui_list
   *          {@link CuiList}
   */
  public void setCuiList(CuiList cui_list) {
    // do nothing
  }

  /**
   * Sets the {@link ApplicationConfiguration} used for default settings.
   * 
   * @param c
   *          the {@link ApplicationConfiguration}
   */
  public void setConfiguration(ApplicationConfiguration c) {
    default_config = c;
  }

  /**
   * Returns details of the configuration for use in a log.
   * 
   * @return details of the configuration for use in a log.
   */
  public String getConfigurationDetails() {
    StringBuffer sb = new StringBuffer();
    String ls = System.getProperty("line.separator");
    sb.append("Metathesaurus Output: Original Release Format.").append(ls);
    if (truncate)
      sb.append("Truncate long fields to length: " + max_field_length).append(
          ls);
    else
      sb.append("Long fields were not truncated.").append(ls);
    if (versioned_output)
      sb.append(
          "Source Abbreviations were written out with a "
              + "versioned representation.").append(ls);
    else
      sb.append(
          "Source Abbreviations were written out with a "
              + "versionless (root) representation.").append(ls);
    if (remove_unicode)
      sb.append("Removed fields containing UTF-8 characters.").append(ls);
    else
      sb.append("Fields containing UTF-8 characters were not removed.").append(
          ls);
    if (add_unicode_bom)
      sb.append("Prepended UTF-8 BOM characters to output files.").append(ls);
    else
      sb.append("UTF-8 BOM characters were not added to output files.").append(
          ls);
    if (calculate_md5s)
      sb.append("Calculated md5s written to mmsys.md5.").append(ls);
    else
      sb.append("Md5s were not calculated.").append(ls);
    if (remove_mth_only)
      sb.append("Removed concepts containing only MTH atoms.").append(ls);
    else
      sb.append("Concepts containing only MTH atoms were not removed.").append(
          ls);
    if (write_oracle)
      sb.append("Oracle load scripts were written.").append(ls);
    if (write_mysql)
      sb.append("MySQL load scripts were written.").append(ls);
    if (!no_mrcxt_file)
      sb.append("MRCXT was written to the destination directory.").append(
          ls);

    return sb.toString();
  }

  /**
   * Opens all target files in the specified directory. The set of target files
   * is the standard MR files in Original Release Format.
   * 
   * @param mr_dir
   *          the subset directory
   * @throws IOException
   *           if the files cannot be opened
   */
  public void open(String mr_dir) throws IOException {
    // if no mrcxt file is to be written don't make one
    no_mrcxt_file = !user_config.getInputStream().hasFullContextData();
    if (no_mrcxt_file)
      target_files = new String[] { "MRATX", "MRCOC", "MRCOLS", "MRCON",
          "MRCUI", "MRDEF", "MRFILES", "MRRANK", "MRREL", "MRSAB", "MRSAT",
          "MRSO", "MRSTY", "MRXNS.ENG", "MRXNW.ENG" };
    else
      target_files = new String[] { "MRATX", "MRCOC", "MRCOLS", "MRCON",
          "MRCUI", "MRCXT", "MRDEF", "MRFILES", "MRRANK", "MRREL", "MRSAB", "MRSAT",
          "MRSO", "MRSTY", "MRXNS.ENG", "MRXNW.ENG" };
    super.open(mr_dir);
    mrfiles_to_md5s = new HashMap();
  }

  /**
   * Closes target files that were opened by this class's openTargets() method
   * 
   * @throws IOException
   *           if any files cannot be closed, or if the <code>release.dat</code>
   *           file or DB load scripts cannot be written.
   */
  public void close() throws IOException {
    super.close();
    try {
      MMSToolkit.writeReleaseDatFile(mr_dir);

      if (write_oracle) {
        writeOracleLoadScripts();
      }
      if (write_mysql) {
        writeMysqlLoadScripts();
      }
    } catch (IOException ie) {
      throw ie;
    }

    if (write_oracle || write_mysql) {
      MMSToolkit.copyResourceToFile((InputStream) this.getClass()
          .getClassLoader().getResource("config/README_DB.txt").getContent(),
          "README_DB.txt", mr_dir);
    }

    //
    // Open mmsys.md5 file
    //
    if (calculateMD5s()) {
      BufferedWriter log = null;
      try {
        log = new BufferedWriter(new FileWriter(new File(user_config
            .getSubsetDirectory(), "mmsys.md5")));
      } catch (Exception e) {
        e.printStackTrace();
        MMSToolkit.reportError(
            "Exception opening mmsys.md5: " + e.getMessage(), true);
      }

      // Recompute MD5 for sorted files.
      try {
        mrfiles_to_md5s.put("MRCUI", SystemToolkit.md5CrossPlatform(new File(
            mr_dir, "MRCUI")));
        mrfiles_to_md5s.put("MRATX", SystemToolkit.md5CrossPlatform(new File(
            mr_dir, "MRATX")));
        mrfiles_to_md5s.put("AMBIG.LUI", SystemToolkit
            .md5CrossPlatform(new File(mr_dir, "AMBIG.LUI")));
        mrfiles_to_md5s.put("AMBIG.SUI", SystemToolkit
            .md5CrossPlatform(new File(mr_dir, "AMBIG.SUI")));
        mrfiles_to_md5s.put("MRCOLS", SystemToolkit
            .md5CrossPlatform(new File(mr_dir, "MRCOLS")));
        mrfiles_to_md5s.put("MRFILES", SystemToolkit
            .md5CrossPlatform(new File(mr_dir, "MRFILES")));
      } catch (NoSuchAlgorithmException ne) {
        ne.printStackTrace();
      }

      String[] md5_strings = new String[mrfiles_to_md5s.keySet().size()];
      int i = 0;
      Iterator iter = mrfiles_to_md5s.keySet().iterator();
      while (iter.hasNext()) {
        String file_name = (String) iter.next();
        Object value = mrfiles_to_md5s.get(file_name);
        if (value instanceof MessageDigest)
          md5_strings[i++] = ("MD5 (" + file_name + ") = " + SystemToolkit
              .toHexString(((MessageDigest) value).digest()));
        else
          md5_strings[i++] = ("MD5 (" + file_name + ") = " + value.toString());
      }
      Arrays.sort(md5_strings);
      for (int j = 0; j < md5_strings.length; j++) {
        log.write(md5_strings[j]);
        log.newLine();
      }
      log.close();
    }
  }


  /**
   * Returns <code>true</code> if the specified LUI is suppressible.
   * @param lui the LUI value to check
   * @param atoms the {@link Atom}s
   * @return <code>true</code> if the LUI is suppressible, <code>false</code> otherwise
   */
  protected boolean isLuiSuppressible(String lui, Atom[] atoms) {

    //
    // Iterate through atoms
    //
    for (int i=0; i<atoms.length; i++) {
      if (lui.equals(atoms[i].getLui())) {
        //
        // If any atom is not suppressible, the LUI is not suppressible
        //
        if (!isSuppressible(atoms[i]))
          return false;
      }
    }

    //
    // If all atoms are suppressible, LUI is suppressible
    //
    return true;
  }
  
  /**
   * Provides support for atom level suppressibility based on values
   * "Y", "N", "E", and "O".
   * @param atom {@link Atom}
   * @return <code>true</code> if atom is suppressible
   * <code>false</code> otherwise
   */
  private boolean isSuppressible(Atom atom) {
  	if (atom.getSuppress() != null && atom.getTermType() != null) {
  		if (atom.getSuppress().getAbbreviation().equals("Y"))
  			return true;
  		else if (atom.getSuppress().getAbbreviation().equals("N"))
  			return false;
  		else if (atom.getSuppress().getAbbreviation().equals("E"))
  			return true;
  		else if (atom.getSuppress().getAbbreviation().equals("O")) {
  			SourceTermType stty = null;
  			if (atom.getSource() != null && atom.getTermType() != null)
  				stty = default_config.getSourceTermType(
  					atom.getSource().getRsab(), atom.getTermType().getAbbreviation());
        if (stty != null)
  			  return user_config.isSourceTermTypeSuppressed(stty);
        else
        	return false;
  		}
  	}
  	return false;
  }
  
  /**
   * Writes MRDEF entries for a {@link Concept}. Truncates long fields and/or
   * removes unicode characters as user specifies.
   * 
   * @param concept
   *          the {@link Concept} to write
   * @throws IOException
   *           if anything goes wrong
   */
  protected void writeMRDEF(Concept concept) throws IOException {
    Definition[] definitions = concept.getDefinitions();
    if (truncate || remove_unicode) {
      for (int i = 0; i < definitions.length; i++) {
        if (truncate) {
          String def = definitions[i].getDef();
          if (def.length() > max_field_length) {
            def = truncate(def, max_field_length);
          }
          definitions[i].setDef(def);
        }
        if (remove_unicode) {
          String def = definitions[i].getDef();
          if (MMSToolkit.containsUnicode(def)) {
            concept.removeDefinition(definitions[i]);
          }
        }
      }
    }
    super.writeMRDEF(concept);
  }

  /**
   * Writes MRSAT entries for a {@link Concept}. Truncates long fields and/or
   * removes unicode characters as user specifies.
   * 
   * @param concept
   *          the {@link Concept} to write
   * @throws IOException
   *           if anything goes wrong
   */
  protected void writeMRSAT(Concept concept) throws IOException {
    Attribute[] attributes = concept.getAttributes();
    if (truncate || remove_unicode) {
      for (int i = 0; i < attributes.length; i++) {
        if (truncate) {
          String atv = attributes[i].getAtv();
          if (atv.length() > max_field_length) {
            atv = truncate(atv, max_field_length);
          }
          attributes[i].setAtv(atv);
        }
        if (remove_unicode) {
          String atv = attributes[i].getAtv();
          if (MMSToolkit.containsUnicode(atv)) {
            concept.removeAttribute(attributes[i]);
          }
        }
      }
    }
    super.writeMRSAT(concept);
  }

  /**
   * Writes MRCXT entries for a {@link Concept}. Truncates long fields and/or
   * removes unicode characters as user specifies.
   * 
   * @param concept
   *          the {@link Concept} to write
   * @throws IOException
   *           if anything goes wrong
   */
  protected void writeMRCXT(Concept concept) throws IOException {
    if (!no_mrcxt_file) {
      Context[] contexts = concept.getContexts();
      if (remove_unicode) {
        for (int i = 0; i < contexts.length; i++) {
          ContextMember[] context_members = contexts[i].getContextMembers();
          for (int j = 0; j < context_members.length; j++) {
            // removing the whole context
            String str = context_members[j].getRelatedAtom().getStr();
            if (MMSToolkit.containsUnicode(str)) {
              concept.removeContext(contexts[i]);
            }
          }
        }
      }
      super.writeMRCXT(concept);
    }
  }

  /**
   * Writes index entries with the "remove utf8" settings in mind.
   * 
   * @param ie
   *          {@link IndexEntry} to be written
   * @throws IOException
   *           if anything goes wrong
   */
  public void writeIndexEntry(IndexEntry ie) throws IOException {
    if (!remove_unicode || !MMSToolkit.containsUnicode(ie.getValue())) {
      super.writeIndexEntry(ie);
    }
  }

  /**
   * Prints the given line to the file with the given name.
   * 
   * @param line
   *          the line to be printed
   * @param file_name
   *          printing destination
   * @throws IOException
   */
  protected void printLine(String line, String file_name) throws IOException {

    if (calculateMD5s()) {
      try {
        if (!mrfiles_to_md5s.containsKey(file_name))
          mrfiles_to_md5s.put(file_name, MessageDigest.getInstance("MD5"));

        MessageDigest md5 = (MessageDigest) mrfiles_to_md5s.get(file_name);
        md5.update(line.getBytes("UTF-8"));
        md5.update(md5_lt);
      } catch (Exception e) {
        e.printStackTrace();
      }
    }
    super.printLine(line, file_name);
  }

  /**
   * Indicates whether or not the output stream is configured to remove
   * non-ASCII Unicode lines.
   * 
   * @return <code>true</code> if so, <code>false</code> otherwise
   */
  public boolean removeUnicode() {
    return remove_unicode;
  }

  /**
   * Indicates whether or not the output stream is configured to remove concepts
   * with only MTH atoms.
   * 
   * @return <code>boolean</code>
   */
  public boolean removeMTHOnlyConcepts() {
    return remove_mth_only;
  }

  /**
   * Indicates whether or not the output stream is configured to calculate md5
   * values for all of the output files.
   * 
   * @return <code>boolean</code>
   */
  public boolean calculateMD5s() {
    return calculate_md5s;
  }

  /**
   * Truncate the string to the specified length.
   * 
   * @param str
   *          {@link String} to be truncated
   * @param length
   *          specified length for truncation
   * @return truncated {@link String}
   */
  private String truncate(String str, int length) {
    if (length < 15)
      return str;
    str = str.substring(0, length - 14);
    // trim incomplete word at the end of string
    int space_index = str.lastIndexOf(" ");
    if (space_index != -1)
      str = str.substring(0, space_index);
    str = str + "...[truncated]";
    return str;
  }

  /**
   * Returns the GUI panel that will be displayed on one of the tabs on the main
   * GUI window.
   * 
   * @return {@link JPanel} that displays graphical interface for a tab
   */
  public JPanel getPanel() {

    //
    // Create panel for all elements
    //
    JPanel full_panel = new JPanel();

    //
    // Use BoxLayout
    //
    full_panel.setLayout(new BoxLayout(full_panel, BoxLayout.Y_AXIS));

    //
    // Files subpanel - Subset folder: label, text field, browse button
    //
    JPanel files_panel = new JPanel(new BorderLayout());

    //
    // Configure border
    //
    TitledBorder titledBorder = new TitledBorder(new EtchedBorder(
        EtchedBorder.RAISED), getName() + " Configuration");
    titledBorder.setTitleJustification(TitledBorder.CENTER);
    files_panel.setBorder(new CompoundBorder(titledBorder, new EmptyBorder(10,
        10, 10, 10)));

    //
    // Add label
    //
    JLabel sd_label = new JLabel("Subset Folder - Location of Subset Files");
    sd_label.setForeground(Color.black);
    files_panel.add(sd_label, BorderLayout.NORTH);

    //
    // Add field/button
    //
    JPanel field_button_panel = new JPanel(new FlowLayout());
    sd_field = new JTextField(user_config.getSubsetDirectory(), 50);
    field_button_panel.add(sd_field);
    JButton sd_button = new JButton(new SubsetDirAction());
    field_button_panel.add(sd_button);
    files_panel.add(field_button_panel, BorderLayout.SOUTH);

    //
    // Add files panel to overall panel
    //
    full_panel.add(files_panel);

    versioned_button = new JCheckBox(
        "Output versioned source abbreviations rather than versionless source abbreviations.");
    truncation_button = new JCheckBox("Truncate long fields to");
    remove_unicode_checkbox = new JCheckBox(
        "Remove records containing extended UTF-8 characters.");
    add_unicode_bom_checkbox = new JCheckBox(
        "Add UTF-8 BOM characters to output files.");
    calculate_md5s_checkbox = new JCheckBox(
        "Calculate MD5s for output files - writes mmsys.md5 file.");
    remove_mth_only_checkbox = new JCheckBox(
        "Remove concepts containing only MTH atoms.");
    write_oracle_load_checkbox = new JCheckBox("Write Oracle load script.");
    write_mysql_load_checkbox = new JCheckBox("Write MySQL load script.");

    //
    // Ad "load DB" subpanel
    //
    JPanel load_db_panel = new JPanel(new GridLayout(1, 2));

    titledBorder = new TitledBorder(new EtchedBorder(EtchedBorder.RAISED),
        "Write Database Load Scripts");
    titledBorder.setTitleJustification(TitledBorder.CENTER);
    load_db_panel.setBorder(new CompoundBorder(titledBorder, new EmptyBorder(
        10, 10, 10, 10)));

    //
    // Add Oracle checkbox
    //
    write_oracle_load_checkbox.setSelected(write_oracle);
    write_oracle_load_checkbox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (ignore_events)
          return;
        fireDataChanged();
        if (!write_oracle)
          write_oracle = true;
        else
          write_oracle = false;
      }
    });
    load_db_panel.add(write_oracle_load_checkbox);

    //
    // Add MySQL checkbox
    //
    write_mysql_load_checkbox.setSelected(write_mysql);
    write_mysql_load_checkbox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (ignore_events)
          return;
        fireDataChanged();
        if (!write_mysql)
          write_mysql = true;
        else
          write_mysql = false;
      }
    });
    load_db_panel.add(write_mysql_load_checkbox);

    //
    // Add "versioned/root" sab output panel
    //
    JPanel sab_panel = new JPanel(new GridLayout(1, 1));
    titledBorder = new TitledBorder(new EtchedBorder(EtchedBorder.RAISED),
        "Source Abbreviation Format");
    titledBorder.setTitleJustification(TitledBorder.CENTER);
    sab_panel.setBorder(new CompoundBorder(titledBorder, new EmptyBorder(10,
        10, 10, 10)));

    versioned_button.setSelected(user_config.isOutputVersioned());
    versioned_button.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (ignore_events)
          return;
        fireDataChanged();
        if (!versioned_output)
          versioned_output = true;
        else
          versioned_output = false;

      }
    });

    sab_panel.add(versioned_button);

    //
    // Field length truncate subpanel
    //
    JPanel t_panel = new JPanel(new GridLayout(1, 1));
    max_length_text_field = new JTextField(15);
    max_length_text_field.setDocument(new IntegerDocument());
    max_length_text_field.setText(new Integer(max_field_length).toString());
    max_length_text_field.getDocument().addDocumentListener(
        new IntegerDocumentListener());
    max_length_text_field.getDocument().putProperty("field", "def def");

    titledBorder = new TitledBorder(new EtchedBorder(EtchedBorder.RAISED),
        "Maximum Field Length");
    titledBorder.setTitleJustification(TitledBorder.CENTER);
    t_panel.setBorder(new CompoundBorder(titledBorder, new EmptyBorder(10, 10,
        10, 10)));

    truncation_button.setSelected(truncate);
    truncation_button.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (ignore_events)
          return;
        fireDataChanged();
        if (!truncate)
          truncate = true;
        else
          truncate = false;
      }
    });

    //
    // Truncate subpanel
    //
    JPanel length_panel = new JPanel();
    FlowLayout f_layout = new FlowLayout(FlowLayout.LEFT, 0, 0);
    length_panel.setLayout(f_layout);
    length_panel.add(truncation_button);
    length_panel.add(max_length_text_field);
    JLabel chars = new JLabel(" characters.");
    chars.setForeground(Color.black);
    length_panel.add(chars);

    t_panel.add(length_panel);

    //
    // Unicode subpanel
    //
    JPanel unicode_panel = new JPanel(new GridLayout(1, 1));
    titledBorder = new TitledBorder(new EtchedBorder(EtchedBorder.RAISED),
        "Eliminate Extended Unicode Characters");
    titledBorder.setTitleJustification(TitledBorder.CENTER);
    unicode_panel.setBorder(new CompoundBorder(titledBorder, new EmptyBorder(
        10, 10, 10, 10)));

    remove_unicode_checkbox.setSelected(remove_unicode);
    remove_unicode_checkbox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (ignore_events)
          return;
        fireDataChanged();
        if (!remove_unicode)
          remove_unicode = true;
        else
          remove_unicode = false;
      }
    });
    unicode_panel.add(remove_unicode_checkbox);

    //
    // MTH subpanel
    //
    JPanel mth_panel = new JPanel(new GridLayout(1, 1));
    titledBorder = new TitledBorder(new EtchedBorder(EtchedBorder.RAISED),
        "Remove MTH only concepts");
    titledBorder.setTitleJustification(TitledBorder.CENTER);
    mth_panel.setBorder(new CompoundBorder(titledBorder, new EmptyBorder(10,
        10, 10, 10)));

    remove_mth_only_checkbox.setSelected(remove_mth_only);
    remove_mth_only_checkbox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (ignore_events)
          return;
        fireDataChanged();
        if (!remove_mth_only)
          remove_mth_only = true;
        else
          remove_mth_only = false;
      }
    });
    mth_panel.add(remove_mth_only_checkbox);

    //
    // MD5 subpanel
    //
    JPanel md5_panel = new JPanel(new GridLayout(1, 1));
    titledBorder = new TitledBorder(new EtchedBorder(EtchedBorder.RAISED),
        "Calculate MD5 values for output files");
    titledBorder.setTitleJustification(TitledBorder.CENTER);
    md5_panel.setBorder(new CompoundBorder(titledBorder, new EmptyBorder(10,
        10, 10, 10)));

    calculate_md5s_checkbox.setSelected(calculate_md5s);
    calculate_md5s_checkbox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (ignore_events)
          return;
        fireDataChanged();
        if (!calculate_md5s)
          calculate_md5s = true;
        else
          calculate_md5s = false;
      }
    });
    md5_panel.add(calculate_md5s_checkbox);

    //
    // BOM subpanel
    //
    JPanel bom_panel = new JPanel(new GridLayout(1, 1));
    titledBorder = new TitledBorder(new EtchedBorder(EtchedBorder.RAISED),
        "Prepend UTF-8 BOM characters to output files");
    titledBorder.setTitleJustification(TitledBorder.CENTER);
    bom_panel.setBorder(new CompoundBorder(titledBorder, new EmptyBorder(10,
        10, 10, 10)));

    add_unicode_bom_checkbox.setSelected(add_unicode_bom);
    add_unicode_bom_checkbox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (ignore_events)
          return;
        fireDataChanged();
        if (!add_unicode_bom)
          add_unicode_bom = true;
        else
          add_unicode_bom = false;
      }
    });
    bom_panel.add(add_unicode_bom_checkbox);

    //
    // Compose full panel
    //
    full_panel.add(load_db_panel);
    full_panel.add(sab_panel);
    full_panel.add(t_panel);
    full_panel.add(unicode_panel);
    full_panel.add(mth_panel);
    full_panel.add(md5_panel);
    full_panel.add(bom_panel);

    //
    // Wrap and return it.
    //
    JPanel wrapper_panel = new JPanel(new BorderLayout());
    wrapper_panel.add(full_panel, BorderLayout.NORTH);
    return wrapper_panel;

  }

  /**
   * Informs {@link SubsetListener} that the Initialize Cui List phase is
   * starting.
   */
  public void beginInitializeCuiList() {
  }

  /**
   * Informs {@link SubsetListener} that the Initialize Cui List phase is
   * ending.
   */
  public void endInitializeCuiList() {
  }

  /**
   * Informs {@link SubsetListener} that the Subset Content phase is starting.
   */
  public void beginSubsetContent() {
  }

  /**
   * Informs {@link SubsetListener} that the Subset Content phase is ending.
   */
  public void endSubsetContent() {
  }

  /**
   * Enforces integer field validation.
   */
  class IntegerDocumentListener implements DocumentListener {

    /**
     * Handle insert changes to text field.
     * 
     * @param e
     *          the {@link DocumentEvent} change
     */
    public void insertUpdate(DocumentEvent e) {
      if (ignore_events)
        return;
      updateLog(e);
    }

    /**
     * Handle remove changes to text field.
     * 
     * @param e
     *          the {@link DocumentEvent} change
     */
    public void removeUpdate(DocumentEvent e) {
      if (ignore_events)
        return;
      updateLog(e);
    }

    /**
     * Does nothing.
     * 
     * @param e
     *          the {@link DocumentEvent} change
     */
    public void changedUpdate(DocumentEvent e) {
      // Plain text components don't fire these events
    }

    /**
     * Handle update to state variables.
     * 
     * @param e
     *          the {@link DocumentEvent}
     */
    public void updateLog(DocumentEvent e) {
      Document doc = (Document) e.getDocument();
      if (doc.getProperty("field").equals("def def")
          && !max_length_text_field.getText().equals("")) {
        fireDataChanged();
        try {
          max_field_length = Integer.valueOf(max_length_text_field.getText())
              .intValue();
        } catch (Exception ex) {
        }
      }
    }
  }

  /**
   * Indicates whether or not the values on this gui have changed.
   * 
   * @return <code>boolean</code> <code>true</code> if data has changed
   *         <code>false</code> otherwise
   */
  public boolean hasDataChanged() {
    return data_changed;
  }

  /**
   * Sets the "has data changed" flag.
   * 
   * @param b
   *          <code>boolean</code>
   */
  public void setDataChanged(boolean b) {
    data_changed = b;
  }

  /**
   * Returns the current state of the output stream as a {@link Properties}
   * object.
   * 
   * @return {@link Properties} only those directly related to the gui
   */
  public Properties getProperties() {
    Properties p = new Properties();
    p.setProperty("versioned_output", String.valueOf(versioned_output));
    p.setProperty("truncate", String.valueOf(truncate));
    p.setProperty("remove_utf8", String.valueOf(remove_unicode));
    p.setProperty("add_unicode_bom", String.valueOf(add_unicode_bom));
    p.setProperty("calculate_md5s", String.valueOf(calculate_md5s));
    p.setProperty("remove_mth_only", String.valueOf(remove_mth_only));
    p.setProperty("max_field_length", String.valueOf(max_field_length));
    p.setProperty("subset_dir", user_config.getSubsetDirectory());
    p.setProperty("write_oracle", String.valueOf(write_oracle));
    p.setProperty("write_mysql", String.valueOf(write_mysql));
    p.setProperty("no_mrcxt_file", String.valueOf(no_mrcxt_file));

    return p;

  }

  /**
   * Sets the current state of the output stream from a {@link Properties}
   * object.
   * 
   * @param p
   *          {@link Properties} only those directly related to the gui
   */
  public void setProperties(Properties p) {

    if (p.containsKey("versioned_output")) {
      versioned_output = Boolean.valueOf(p.getProperty("versioned_output"))
          .booleanValue();
    }

    if (p.containsKey("truncate")) {
      truncate = Boolean.valueOf(p.getProperty("truncate")).booleanValue();
    }

    if (p.containsKey("remove_utf8")) {
      remove_unicode = Boolean.valueOf(p.getProperty("remove_utf8"))
          .booleanValue();
    }
    if (p.containsKey("add_unicode_bom")) {
      add_unicode_bom = Boolean.valueOf(p.getProperty("add_unicode_bom"))
          .booleanValue();
    }

    if (p.containsKey("calculate_md5s")) {
      calculate_md5s = Boolean.valueOf(p.getProperty("calculate_md5s"))
          .booleanValue();
    }

    if (p.containsKey("remove_mth_only")) {
      remove_mth_only = Boolean.valueOf(p.getProperty("remove_mth_only"))
          .booleanValue();
    }

    if (p.containsKey("max_field_length")) {
      max_field_length = Integer.valueOf(p.getProperty("max_field_length"))
          .intValue();
    }

    if (p.containsKey("subset_dir")) {
      user_config.setSubsetDirectory(p.getProperty("subset_dir"));
    }

    if (p.containsKey("write_oracle")) {
      write_oracle = Boolean.valueOf(p.getProperty("write_oracle"))
          .booleanValue();
    }

    if (p.containsKey("write_mysql")) {
      write_mysql = Boolean.valueOf(p.getProperty("write_mysql"))
          .booleanValue();
    }

    if (p.containsKey("no_mrcxt_file")) {
      no_mrcxt_file = Boolean.valueOf(p.getProperty("no_mrcxt_file"))
          .booleanValue();
    }

  }

  /**
   * Resets the "has data changed" flag to indicate that no data has been
   * changed since the last save.
   */
  public void resetDataChanged() {
    data_changed = false;
  }

  /**
   * Adds a {@link DataChangeListener} that will be notified every time a
   * {@link #fireDataChanged()} call is made. Allows main GUI to record all
   * changes to this gui in the undo/redo mechanism.
   * 
   * @param dcl
   *          {@link DataChangeListener}
   */
  public void addDataChangeListener(DataChangeListener dcl) {
    listeners.add(dcl);
  }

  /**
   * Removes a {@link DataChangeListener} so that it will no longer be notified
   * every time a {@link #fireDataChanged} call is made from this gui.
   * 
   * @param dcl
   *          {@link DataChangeListener}
   */
  public void removeDataChangeListener(DataChangeListener dcl) {
    listeners.remove(dcl);
  }

  /**
   * Indicates the {@link Color} that should be used to display the name.
   * 
   * @return {@link Color}
   */
  public Color getColor() {
    return Color.green.darker();
  }

  /**
   * Returns the description of the screen and includes instructions on use.
   * 
   * @return {@link String} text of filter instructions
   */
  public String getTabDescription() {
    return "Select data output options to customize the subset and create "
        + "additional files if desired.  See Help for more information.\n\n"
        + "Users may proceed to the other Options tabs in any order by "
        + "clicking on the tabs across the top of the screen or by clicking "
        + "\"Done\" on the File bar and selecting \"Begin Subset\".";
  }

  /**
   * Provides a mechanism for resetting the variables on a tab back to their
   * default values. This is not used to actually change the gui.
   * setPanelSelections() is always called after this method and will update the
   * gui based on the variable values.
   */
  public void resetToDefaults() {
    ignore_events = true;
    fireDataChanged();
    ApplicationConfiguration default_config = MMSToolkit.getGui()
        .getApplicationConfiguration();

    max_field_length = Integer.valueOf(
        default_config.getProperty(this, "max_field_length")).intValue();
    truncate = Boolean.valueOf(default_config.getProperty(this, "truncate"))
        .booleanValue();
    versioned_output = Boolean.valueOf(
        default_config.getProperty(this, "versioned_output")).booleanValue();
    remove_unicode = Boolean.valueOf(
        default_config.getProperty(this, "remove_utf8")).booleanValue();
    add_unicode_bom = Boolean.valueOf(
        default_config.getProperty(this, "add_unicode_bom")).booleanValue();
    calculate_md5s = Boolean.valueOf(
        default_config.getProperty(this, "calculate_md5s")).booleanValue();
    remove_mth_only = Boolean.valueOf(
        default_config.getProperty(this, "remove_mth_only")).booleanValue();
    write_oracle = Boolean.valueOf(
        default_config.getProperty(this, "write_oracle")).booleanValue();
    write_mysql = Boolean.valueOf(
        default_config.getProperty(this, "write_mysql")).booleanValue();
    no_mrcxt_file = Boolean.valueOf(
        default_config.getProperty(this, "no_mrcxt_file")).booleanValue();

    ignore_events = false;

  }

  /**
   * Presents an advanced options GUI to the user.
   */
  public void showOptions() {
    // do nothing
  }

  /**
   * Indicates whether or not an options menu item should be supplied for
   * advanced configuration options.
   * 
   * @return <code>true</code> if so; <code>false</code> otherwise
   */
  public boolean hasAdvancedOptions() {
    return false;
  }

  /**
   * Returns a {@link JPanel} that will be displayed when the user selects the
   * corresponding menu item from the Help menu and will provide the user with
   * detailed instructions on the tab's use.
   * 
   * @return {@link JPanel} to be added to the Help menu
   */
  public JPanel getHelpPanel() {
    return SwingToolkit.getPanelForHTML(this.getClass().getClassLoader()
        .getResource("config/OutputOptionsHelp.html"));
  }

  /**
   * Method makes selections on the panel.
   */
  public void setPanelSelections() {
    //
    // ignore events so that when undo calls this method it doesn't
    // add to the undo stack when resetting the JTextFields
    //
    ignore_events = true;
    max_length_text_field.setText(new Integer(max_field_length).toString());

    if (versioned_output)
      versioned_button.setSelected(true);
    else
      versioned_button.setSelected(false);
    if (truncate)
      truncation_button.setSelected(true);
    else
      truncation_button.setSelected(false);
    if (remove_unicode)
      remove_unicode_checkbox.setSelected(true);
    else
      remove_unicode_checkbox.setSelected(false);
    if (add_unicode_bom)
      add_unicode_bom_checkbox.setSelected(true);
    else
      add_unicode_bom_checkbox.setSelected(false);
    if (calculate_md5s)
      calculate_md5s_checkbox.setSelected(true);
    else
      calculate_md5s_checkbox.setSelected(false);
    if (remove_mth_only)
      remove_mth_only_checkbox.setSelected(true);
    else
      remove_mth_only_checkbox.setSelected(false);
    if (write_oracle)
      write_oracle_load_checkbox.setSelected(true);
    else
      write_oracle_load_checkbox.setSelected(false);

    if (write_mysql)
      write_mysql_load_checkbox.setSelected(true);
    else
      write_mysql_load_checkbox.setSelected(false);

    sd_field.setText(user_config.getSubsetDirectory());

    ignore_events = false;

  }

  /**
   * This method indicates whether or not the {@link GUIConfigurable} is valid.
   * 
   * @return <code>boolean</code><code>true</code> if
   *         {@link GUIConfigurable} is valid <code>false</code> if
   *         {@link GUIConfigurable} is invalid
   */
  public boolean validate() {
    File subset_directory = new File(user_config.getSubsetDirectory());
    if (subset_directory.exists() && subset_directory.canWrite()) {
      return true;
    }

    //
    // If does not exist, offer to create it
    //
    if (!subset_directory.exists()) {
      if (MMSToolkit.confirmRequest(MMSToolkit.getGui(), "The subset folder\n "
          + subset_directory + " \n does not exist.\n"
          + "Would you like to create a subset folder\n " + "with this path?")) {
        user_config.setSubsetDirectory(subset_directory.getAbsolutePath());
        if (subset_directory.mkdirs()) {
          return true;
        } else {
          validate_msg = "There was a problem creating the directory,\n"
              + "please try a different directory.";
          return false;
        }
      } else {
        validate_msg = "The subset folder\n " + subset_directory
            + " \n does not exist.\n" + "Please choose another subset folder.";
        return false;
      }
    }

    //
    // if not writeable, report error
    //
    if (!subset_directory.canWrite()) {
      validate_msg = "The subset folder\n " + subset_directory
          + " \nis write protected.\n" + "Please choose another subset folder.";
      return false;
    } else {
      // if (!sd_field.getText().equals(user_config.getSubsetDirectory()))
      // user_config.setPropertiesChanged();
      // user_config.setSubsetDirectory(sd_field.getText());
      return true;
    }
  }

  /**
   * Returns the message that the user will see if this {@link GUIConfigurable}
   * is invalid.
   * 
   * @return {@link String} with error message
   */
  public String getValidateMessage() {
    return "";
  }

  /**
   * Returns the messages indicating why validation failed. Should only be
   * called if {@link #validate()} returns <code>false</code>.
   * 
   * @return the failure messages
   */
  public String[] getValidationErrors() {
    return new String[] { getValidateMessage() };
  }

  /**
   * Returns warnings that were indicated during validation.
   * 
   * @return the warning messages
   */
  public String[] getValidationWarnings() {
    return new String[0];
  }

  /**
   * Returns the {@link GUIConfigurable} whose tab should next be displayed to
   * allow the user to correct an error or warning.
   * 
   * @return GUIConfigurable to be displayed
   */
  public GUIConfigurable getGUIConfigurableForValidationFailure() {
    return null;
  }

  /**
   * Enables filters to inform listeners of data changed event.
   */
  protected void fireDataChanged() {
    for (int i = 0; i < listeners.size(); i++) {
      ((DataChangeListener) listeners.get(i)).dataChanged(new DataChangeEvent(
          this));
    }
    data_changed = true;
  }

  /**
   * Enables filters to indicate that data has changed without notifying the
   * listeners or making this change undoable.
   */
  protected void fireNotUndoableDataChanged() {
    data_changed = true;
  }

  /**
   * Opens the specified file in the directory passed to {@link #open(String)}.
   * 
   * @param name
   *          the filename
   * @return a {@link BufferedWriter} for the file
   * @throws IOException
   *           if the file could not be opened.
   */
  protected BufferedWriter openTargetFile(String name) throws IOException {
    if (!name.endsWith(".x") && !name.endsWith(".ctl")
        && !name.endsWith(".sql") && !name.endsWith(".csh")
        && !name.endsWith(".bat"))
      return super.openTargetFile(name);
    else {
      FileOutputStream fos = new FileOutputStream(new File(mr_dir, name));

      return new BufferedWriter(new UTF8OutputStreamWriter(fos, false),
          1024 * 256);
    }
  }

  /**
   * Calls methods that write out scripts necessary for loading subset into
   * Oracle.
   * 
   * @throws IOException
   */
  private void writeOracleLoadScripts() throws IOException {
    // Open files and initialize statistics
    for (int i = 0; i < target_files.length; i++) {
      writeOracleCtlFile(target_files[i]);
    }

    List languages = Language.getLanguagesAsList();
    Collections.sort(languages);
    for (int i = 0; i < languages.size(); i++) {
      String base_file = "MRXW."
          + ((Language) languages.get(i)).getAbbreviation();
      writeOracleCtlFile(base_file);
    }
    writeOracleCtlFile("AMBIG.SUI");
    writeOracleCtlFile("AMBIG.LUI");

    String script_type = System.getProperty("script_type");
    if (script_type == null) {
      MMSToolkit.reportError(MMSToolkit.getGui(),
          "Expected command line parameter \"-Dscript_type\" missing.\n"
              + "Cannot create load script.", false);
      script_type = "";
    }
    if (!script_type.equals(".bat")) {
      writePopulateOracleCshFile();
    } else {
      writePopulateOracleBatFile();
    }
    writeOracleTablesFile();
    writeOracleIndexesFile();
    MMSToolkit.copyResourceToFile(
        (InputStream) this.getClass().getClassLoader().getResource(
            "config/mrconso_oracle.sql").getContent(), "mrconso_oracle.sql",
        mr_dir);
  }

  /**
   * Writes an SQL*Loader control file to the specified file name.
   * 
   * @param file_name
   *          the file name
   * @throws IOException
   */
  private void writeOracleCtlFile(String file_name) throws IOException {

    //
    // MRXW.SPA becomes MRXW_SPA
    //
    String revised_file_name = file_name.replace('.', '_');

    //
    // Open the file
    //
    BufferedWriter target = openTargetFile(revised_file_name + ".ctl");

    //
    // Write the control file header (do not use direct load)
    //
    target.write("options (direct=true)");
    target.newLine();
    target.write("load data");
    target.newLine();
    target.write("infile '" + revised_file_name + ".dat' \"str X'7c");
    target.write(lineTerminatorToHex());
    target.write("'\"");
    target.newLine();
    target.write("badfile '" + revised_file_name + ".bad'");
    target.newLine();
    target.write("discardfile '" + revised_file_name + ".dsc'");
    target.newLine();
    target.write("truncate");
    target.newLine();
    target.write("into table " + revised_file_name);
    target.newLine();
    target.write("fields terminated by '|'");
    target.newLine();
    target.write("trailing nullcols");
    target.newLine();

    //
    // Write the schema section of the control file
    //
    target.write("(");
    FileStatistics fs = config.getFileStatistics(file_name);
    ColumnStatistics[] col_stats = fs.getAllColumnStatistics();
    for (int i = 0; i < col_stats.length; i++) {
      String type = getOracleCtlType(col_stats[i].getDataType());
      target.write(col_stats[i].getColumnName() + "\t" + type);
      if (i != col_stats.length - 1) {
        target.write(",");
        target.newLine();
      }
    }
    target.write("");
    target.newLine();
    target.write(")");

    target.close();
  }

  /**
   * Gets the SQL*Loader control file data type for an ANSI SQL-92 data type.
   * 
   * @param sql_type
   *          {@link String}
   * @return {@link String}
   */
  private String getOracleCtlType(String sql_type) {
    String size_suffix = "";
    if (sql_type.indexOf('(') != -1) {
      size_suffix = sql_type.substring(sql_type.indexOf('('));
      if (size_suffix.equals("(0)")) {
        size_suffix = "";
      }
    }
    if (sql_type.startsWith("char") || sql_type.startsWith("varchar")) {
      sql_type = "char" + size_suffix;
    } else if (sql_type.equals("integer")) {
      sql_type = "integer external";
    } else if (sql_type.startsWith("numeric")) {
      sql_type = "float external";
    }
    return sql_type;
  }

  /**
   * Writes "tables.sql" script which creates all necessary tables in the db.
   * 
   * @throws IOException
   */
  private void writeOracleTablesFile() throws IOException {
    BufferedWriter tables_sql = openTargetFile("oracle_tables.sql");
    for (int i = 0; i < target_files.length; i++) {
      writeOracleSqlTable(target_files[i], tables_sql);
    }
    List languages = Language.getLanguagesAsList();
    Collections.sort(languages);
    for (int i = 0; i < languages.size(); i++) {
      String base_file = "MRXW."
          + ((Language) languages.get(i)).getAbbreviation();
      writeOracleSqlTable(base_file, tables_sql);
    }

    writeOracleSqlTable("AMBIG.SUI", tables_sql);
    writeOracleSqlTable("AMBIG.LUI", tables_sql);

    tables_sql.close();
  }

  /**
   * Writes a file called "indexes.sql" that is used for creating indexes on the
   * tables loaded into the db.
   * 
   * @throws IOException
   */
  private void writeOracleIndexesFile() throws IOException {

    BufferedWriter target = openTargetFile("oracle_indexes.sql");
    for (int i = 0; i < target_files.length; i++) {
      writeOracleIndexesForTable(target_files[i], target,
          (List) files_to_index_fields.get(target_files[i]));
    }
    List languages = Language.getLanguagesAsList();
    Collections.sort(languages);
    Object[] indexes_indexes = { "WD" };
    for (int i = 0; i < languages.size(); i++) {
      String base_file = "MRXW."
          + ((Language) languages.get(i)).getAbbreviation();
      writeOracleIndexesForTable(base_file, target, Arrays
          .asList(indexes_indexes));
    }

    writeOracleIndexesForTable("AMBIG.SUI", target,
        (List) files_to_index_fields.get("AMBIG.SUI"));
    writeOracleIndexesForTable("AMBIG.LUI", target,
        (List) files_to_index_fields.get("AMBIG.LUI"));
    target.close();
  }

  /**
   * Write specifics for creating a given table to the "tables.sql" file.
   * 
   * @param file_name
   *          {@link String} name of the table/file to be created
   * @param target
   *          {@link BufferedWriter} "tables.sql"
   * @throws IOException
   */
  private void writeOracleSqlTable(String file_name, BufferedWriter target)
      throws IOException {

    // Something like MRXW.ENG may get passed
    // in, convert it to MRXW_ENG
    String revised_file_name = file_name.replace('.', '_');

    //
    // Write the schema script for the table
    //

    target.write("DROP TABLE " + revised_file_name + ";");
    target.newLine();
    target.write("CREATE TABLE " + revised_file_name + " (");
    target.newLine();
    FileStatistics these_stats = config.getFileStatistics(file_name);
    ColumnStatistics[] col_stats = these_stats.getAllColumnStatistics();
    for (int i = 0; i < col_stats.length; i++) {
      String type = col_stats[i].getDataType();

      // If we have something like varchar(6000), where
      // the length is > 4000 characters, make it 4000
      // so that Oracle does not complain.
      if (type.startsWith("varchar(")
          && Integer.valueOf(
              type.substring(type.indexOf('(') + 1, type.indexOf(')')))
              .intValue() > 4000) {
        type = "varchar2(4000)";

        // Oracle uses VARCHAR2, so convert varchar
      }
      if (type.startsWith("varchar")) {
        type = "varchar2" + type.substring(type.indexOf('('));
      }
      target.write("    " + col_stats[i].getColumnName() + "\t" + type);
      if (col_stats[i].getMinLength() != 0) {
        target.write(" NOT NULL");
      }
      if (i != col_stats.length - 1) {
        target.write(",");
        target.newLine();
      }
    }
    target.newLine();
    target.write(")");
    target.newLine();
    target.write("PCTFREE 10 PCTUSED 80");
    target.newLine();
    long initial = these_stats.getByteCount();
    if (initial > (1024 * 1024 * 1024))
      initial = (1024 * 1024 * 1024);
    target.write("STORAGE (INITIAL " + initial + " NEXT 100M PCTINCREASE 0);");
    target.newLine();
    target.newLine();
  }

  /**
   * Writes to the "indexes.sql" file all indexes related to the given
   * table_name.
   * 
   * @param table_name
   *          {@link String}
   * @param target
   *          {@link BufferedWriter} "indexes.sql"
   * @param dex
   *          {@link List} of columns needing indexes on the given table_name
   * @throws IOException
   */
  private void writeOracleIndexesForTable(String table_name,
      BufferedWriter target, List dex) throws IOException {
    // Something like MRXW.ENG may get passed
    // in, convert it to MRXW_ENG
    String revised_table_name = table_name.replace('.', '_');
    if (dex == null) {
      return;
    }
    ArrayList indexes = new ArrayList(dex);

    for (int i = 0; i < indexes.size(); i++) {
      // Write index definition
      target.newLine();
      target.write("DROP INDEX X_" + revised_table_name + "_"
          + ((String) indexes.get(i)).replace(',', '_') + ";");
      target.newLine();
      target.write("CREATE INDEX X_" + revised_table_name + "_"
          + ((String) indexes.get(i)).replace(',', '_') + " ON "
          + revised_table_name + "(" + indexes.get(i) + ")");
      target.newLine();
      target.write("PCTFREE 10 STORAGE (INITIAL 10M NEXT 10M)");
      target.newLine();
      target.write("COMPUTE STATISTICS;");
      target.newLine();
    }
  }

  /**
   * Write a UNIX .csh script for loading all of the files into a database.
   * 
   * @throws IOException
   */
  private void writePopulateOracleCshFile() throws IOException {

    // Open the file
    BufferedWriter csh_file = openTargetFile("populate_oracle_db.csh");

    // Write the header
    csh_file.write("#!/bin/csh -f");
    csh_file.newLine();
    csh_file.write("#");
    csh_file.newLine();
    csh_file.write("# RECOMMENDATIONS (READ THIS FIRST)");
    csh_file.newLine();
    csh_file.write("# When creating an Oracle database for loading UMLS content,");
    csh_file.newLine();
    csh_file.write("# you should perform your configuration with the following");
    csh_file.newLine();
    csh_file.write("# recommendations in mind:");
    csh_file.newLine();
    csh_file.write("#  o Create the database using UTF-8 as the default and NLS character set");
    csh_file.newLine();
    csh_file.write("#  o Do not load content as SYS or SYSTEM, instead create a separate user,");
    csh_file.newLine();
    csh_file.write("#    e.g. UMLS");
    csh_file.newLine();
    csh_file.write("#  o Make sure the user is able to load the volume of data, consider");
    csh_file.newLine();
    csh_file.write("#     o Allocate large enough data files to the user's tablespace OR,");
    csh_file.newLine();
    csh_file.write("#       enable AUTOEXTEND on this user's tablespace and don't specify a MAXSIZE");
    csh_file.newLine();
    csh_file.write("#     o Set the user quota to UNLIMITED on the tablespace");
    csh_file.newLine();
    csh_file.write("#");
    csh_file.newLine();
    csh_file.write("# Database connection parameters");
    csh_file.newLine();
    csh_file.write("# Please edit these variables to reflect your environment");
    csh_file.newLine();
    csh_file.write("#");
    csh_file.newLine();
    csh_file.write("set ORACLE_HOME=<path to ORACLE_HOME>");
    csh_file.newLine();
    csh_file.write("set user=<username>");
    csh_file.newLine();
    csh_file.write("set password=<password>");
    csh_file.newLine();
    csh_file.write("set tns_name=<tns_name>");
    csh_file.newLine();

    List languages = Language.getLanguagesAsList();

    csh_file.write("setenv NLS_LANG AMERICAN_AMERICA.UTF8");
    csh_file.newLine();

    csh_file.write("");
    csh_file.newLine();

    csh_file.write("echo \"----------------------------------------\"");
    csh_file.newLine();
    csh_file.write("echo \"Starting ... `/bin/date`\"");
    csh_file.newLine();
    csh_file.write("echo \"----------------------------------------\"");
    csh_file.newLine();
    csh_file.write("echo \"\"");
    csh_file.newLine();

    //
    // For each core content file, create the corresponding table
    //
    csh_file.write("$ORACLE_HOME/bin/sqlplus $user/$password@$tns_name < "
        + "oracle_tables.sql");
    csh_file.newLine();
    csh_file.newLine();
    //
    // Load the data using SQL*Loader
    //
    for (int i = 0; i < target_files.length; i++) {
      csh_file.write("/bin/mv -f " + target_files[i] + " "
          + target_files[i].replace('.', '_') + ".dat");
      csh_file.newLine();
      csh_file.write("$ORACLE_HOME/bin/sqlldr $user/$password@$tns_name "
          + "control=\"" + target_files[i].replace('.', '_') + ".ctl\"");
      csh_file.newLine();
      csh_file.write("/bin/mv -f " + target_files[i].replace('.', '_')
          + ".dat " + target_files[i]);
      csh_file.newLine();
    }

    Collections.sort(languages);
    for (int i = 0; i < languages.size(); i++) {
      String base_file = "MRXW_"
          + ((Language) languages.get(i)).getAbbreviation();
      csh_file.write("/bin/mv -f " + base_file.replace('_', '.') + " "
          + base_file + ".dat");
      csh_file.newLine();
      csh_file.write("$ORACLE_HOME/bin/sqlldr $user/$password@$tns_name "
          + "control=\"" + base_file + ".ctl\"");
      csh_file.newLine();
      csh_file.write("/bin/mv -f " + base_file + ".dat "
          + base_file.replace('_', '.'));
      csh_file.newLine();

    }
    csh_file.newLine();

    csh_file.write("/bin/mv -f AMBIG.SUI AMBIG_SUI.dat");
    csh_file.newLine();
    csh_file.write("$ORACLE_HOME/bin/sqlldr $user/$password@$tns_name "
        + "control=\"AMBIG_SUI.ctl\"");
    csh_file.newLine();
    csh_file.write("/bin/mv -f AMBIG_SUI.dat AMBIG.SUI");
    csh_file.newLine();

    csh_file.write("/bin/mv -f AMBIG.LUI AMBIG_LUI.dat");
    csh_file.newLine();
    csh_file.write("$ORACLE_HOME/bin/sqlldr $user/$password@$tns_name "
        + "control=\"AMBIG_LUI.ctl\"");
    csh_file.newLine();
    csh_file.write("/bin/mv -f AMBIG_LUI.dat AMBIG.LUI");
    csh_file.newLine();

    //
    // Then create the relevant indexes
    //
    csh_file.write("$ORACLE_HOME/bin/sqlplus $user/$password@$tns_name < "
        + "oracle_indexes.sql");
    csh_file.newLine();
    //
    // Then create and index the MRCONSO table
    //
    csh_file.write("$ORACLE_HOME/bin/sqlplus $user/$password@$tns_name < "
        + "mrconso_oracle.sql");
    csh_file.newLine();

    csh_file.write("echo");
    csh_file.newLine();
    csh_file.write("echo \"The oracle_grants.sql file was not loaded.\"");
    csh_file.newLine();
    csh_file.write("echo \"The oracle_synonyms.sql file was not loaded.\"");
    csh_file.newLine();
    csh_file.write("echo");
    csh_file.newLine();

    csh_file.write("echo \"----------------------------------------\"");
    csh_file.newLine();
    csh_file.write("echo \"Finished ... `/bin/date`\"");
    csh_file.newLine();
    csh_file.write("echo \"----------------------------------------\"");
    csh_file.newLine();
    csh_file.close();
  }

  /**
   * Write a Windows .bat script for loading all of the files into a database.
   * 
   * @throws IOException
   */
  private void writePopulateOracleBatFile() throws IOException {

    // Open the file
    BufferedWriter bat_file = openTargetFile("populate_oracle_db.bat");

    // Write the header
    bat_file.write("::");
    bat_file.newLine();
    bat_file.write(":: RECOMMENDATIONS (READ THIS FIRST)");
    bat_file.newLine();
    bat_file.write(":: When creating an Oracle database for loading UMLS content,");
    bat_file.newLine();
    bat_file.write(":: you should perform your configuration with the following");
    bat_file.newLine();
    bat_file.write(":: recommendations in mind:");
    bat_file.newLine();
    bat_file.write("::  o Create the database using UTF-8 as the default and NLS character set");
    bat_file.newLine();
    bat_file.write("::  o Do not load content as SYS or SYSTEM, instead create a separate user,");
    bat_file.newLine();
    bat_file.write("::    e.g. UMLS");
    bat_file.newLine();
    bat_file.write("::  o Make sure the user is able to load the volume of data, consider");
    bat_file.newLine();
    bat_file.write("::     o Allocate large enough data files to the user's tablespace OR,");
    bat_file.newLine();
    bat_file.write("::       enable AUTOEXTEND on this user's tablespace and don't specify a MAXSIZE");
    bat_file.newLine();
    bat_file.write("::     o Set the user quota to UNLIMITED on the tablespace");
    bat_file.newLine();
    bat_file.write("::");
    bat_file.newLine();
    bat_file.write(":: Database connection parameters");
    bat_file.newLine();
    bat_file.write(":: Please edit these variables to reflect your environment");
    bat_file.newLine();
    bat_file.write("::");
    bat_file.newLine();
    bat_file.write("set ORACLE_HOME=<path to ORACLE_HOME>");
    bat_file.newLine();
    bat_file.write("set user=<username>");
    bat_file.newLine();
    bat_file.write("set password=<password>");
    bat_file.newLine();
    bat_file.write("set tns_name=<tns_name>");
    bat_file.newLine();

    List languages = Language.getLanguagesAsList();

    bat_file.write("echo ----------------------------------------");
    bat_file.newLine();
    bat_file.write("echo Starting ... ");
    bat_file.newLine();
    bat_file.write("echo ----------------------------------------");
    bat_file.newLine();
    bat_file.write("echo. ");
    bat_file.newLine();

    //
    // For each core content, create the corresponding table
    //
    bat_file
        .write("%ORACLE_HOME%\\bin\\sqlplus %user%/%password%@%tns_name% < "
            + "oracle_tables.sql");
    bat_file.newLine();

    //
    // Load the data using SQL*Loader
    //
    for (int i = 0; i < target_files.length; i++) {
      bat_file.newLine();
      bat_file.write("rename " + target_files[i] + " "
          + target_files[i].replace('.', '_') + ".dat");
      bat_file.newLine();
      bat_file.write("%ORACLE_HOME%\\bin\\sqlldr %user%/%password%@%tns_name% "
          + "control=\"" + target_files[i].replace('.', '_') + ".ctl\"");
      bat_file.newLine();
      bat_file.write("rename " + target_files[i].replace('.', '_') + ".dat "
          + target_files[i]);
      bat_file.newLine();

    }
    Collections.sort(languages);
    for (int i = 0; i < languages.size(); i++) {

      bat_file.newLine();
      String base_file = "MRXW_"
          + ((Language) languages.get(i)).getAbbreviation();
      bat_file.write("rename " + base_file.replace('_', '.') + " " + base_file
          + ".dat");
      bat_file.newLine();
      bat_file.write("%ORACLE_HOME%\\bin\\sqlldr %user%/%password%@%tns_name% "
          + "control=\"" + base_file + ".ctl\"");
      bat_file.newLine();
      bat_file.write("rename " + base_file + ".dat "
          + base_file.replace('_', '.'));
      bat_file.newLine();
    }
    bat_file.newLine();
    bat_file.write("rename AMBIG.SUI AMBIG_SUI.dat");
    bat_file.newLine();
    bat_file.write("%ORACLE_HOME%\\bin\\sqlldr %user%/%password%@%tns_name% "
        + "control=\"AMBIG_SUI.ctl\"");
    bat_file.newLine();
    bat_file.write("rename AMBIG_SUI.dat AMBIG.SUI");
    bat_file.newLine();

    bat_file.write("rename AMBIG.LUI AMBIG_LUI.dat");
    bat_file.newLine();
    bat_file.write("%ORACLE_HOME%\\bin\\sqlldr %user%/%password%@%tns_name% "
        + "control=\"AMBIG_LUI.ctl\"");
    bat_file.newLine();
    bat_file.write("rename AMBIG_LUI.dat AMBIG.LUI");
    bat_file.newLine();

    //
    // Then create the relevant indexes
    //
    bat_file
        .write("%ORACLE_HOME%\\bin\\sqlplus %user%/%password%@%tns_name% < "
            + "oracle_indexes.sql");
    bat_file.newLine();

    //
    // Then create and index the MRCONSO table
    //
    bat_file
        .write("%ORACLE_HOME%\\bin\\sqlplus %user%/%password%@%tns_name% < "
            + "mrconso_oracle.sql");
    bat_file.newLine();

    bat_file.write("echo");
    bat_file.newLine();
    bat_file.write("echo The oracle_grants.sql file was not loaded.");
    bat_file.newLine();
    bat_file.write("echo The oracle_synonyms.sql file was not loaded.");
    bat_file.newLine();
    bat_file.write("echo");
    bat_file.newLine();

    bat_file.write("echo ----------------------------------------");
    bat_file.newLine();
    bat_file.write("echo Finished  ");
    bat_file.newLine();
    bat_file.write("echo ----------------------------------------");
    bat_file.close();
  }

  /**
   * Converts the system line separator to a string of hex characters. This is
   * used to identify the record terminator in the SQL*Loader control file.
   * 
   * @return {@link String}
   */
  private String lineTerminatorToHex() {
    String lt = System.getProperty("line.separator");
    StringBuffer sb = new StringBuffer();
    for (int i = 0; i < lt.length(); i++) {
      String hex = Integer.toHexString((int) lt.charAt(i));
      if (hex.length() == 1) {
        sb.append("0");
      }
      sb.append(hex);
    }
    return sb.toString();
  }

  /**
   * Calls methods that write out scripts necessary for loading subset into
   * mysql.
   * 
   * @throws IOException
   */
  private void writeMysqlLoadScripts() throws IOException {
    if (!System.getProperty("script_type").equals(".bat")) {
      writePopulateMysqlCshFile();
    } else {
      writePopulateMysqlBatFile();
    }
    writeMysqlTablesFile();
    writeMysqlIndexesFile();
    // writeMysqlGrantFile();
    writeCreateMysqlDbFile();
    MMSToolkit.copyResourceToFile((InputStream) this.getClass()
        .getClassLoader().getResource("config/mrconso_mysql.sql").getContent(),
        "mrconso_mysql.sql", mr_dir);
  }

  /**
   * Writes a create statement for a table in a MySQL database.
   * 
   * @param file_name
   *          the name of the file/table
   * @param target
   *          the script to write to
   * @throws IOException
   *           if anything goes wrong
   */
  private void writeMysqlTable(String file_name, BufferedWriter target)
      throws IOException {
    String revised_file_name = file_name.replace('.', '_');

    target.write("DROP TABLE IF EXISTS " + revised_file_name + ";");
    target.newLine();
    target.write("CREATE TABLE " + revised_file_name + " (");
    target.newLine();
    FileStatistics these_stats = config.getFileStatistics(file_name);
    ColumnStatistics[] col_stats = these_stats.getAllColumnStatistics();
    int avg_row_length = 0;
    for (int i = 0; i < col_stats.length; i++) {
      String type = col_stats[i].getDataType();

      // If we have something like varchar(6000), where
      // the length is > 255 characters, make it type blob
      // so that mysql does not complain.
      if (type.startsWith("varchar(")
          && Integer.valueOf(
              type.substring(type.indexOf('(') + 1, type.indexOf(')')))
              .intValue() > 255) {
        type = "blob";
      }
      if (type.startsWith("varchar")) {
        type = "varchar" + type.substring(type.indexOf('('));
      }
      target.write("    " + col_stats[i].getColumnName() + "\t" + type);
      if (col_stats[i].getMinLength() != 0) {
        target.write(" NOT NULL");
      }
      if (i != col_stats.length - 1) {
        target.write(",");
        target.newLine();
      }

      // Determine average line length
      avg_row_length += col_stats[i].getAverageLength();

    }
    target.newLine();
    target.write(") CHARACTER SET utf8");

    // print MAX_ROWS clause
    if (these_stats.getByteCount() > 4000000000L) {
      target.write(" MAX_ROWS=" + (these_stats.getLineCount() * 2)
          + " AVG_ROW_LENGTH=" + avg_row_length);
    }

    // Finish the table
    target.write(";");
    target.newLine();

  }

  /**
   * Writes the MySQL populate script.
   * 
   * @throws IOException
   *           if anything goes wrong
   */
  private void writePopulateMysqlCshFile() throws IOException {
    BufferedWriter csh_file = openTargetFile("populate_mysql_db.csh");
    csh_file.write("#!/bin/csh -f");
    csh_file.newLine();
    csh_file.write("#");
    csh_file.newLine();
    csh_file.write("# RECOMMENDATIONS (READ THIS FIRST)");
    csh_file.newLine();
    csh_file.write("# When creating a MySQL database for loading UMLS content,");
    csh_file.newLine();
    csh_file.write("# you should perform your configuration with the following");
    csh_file.newLine();
    csh_file.write("# recommendations in mind:");
    csh_file.newLine();
    csh_file.write("#  o Create the database using UTF-8 as the default character set");
    csh_file.newLine();
    csh_file.write("#  o Create a separate user, e.g. UMLS");
    csh_file.newLine();
    csh_file.write("#  o To make sure the user is able to load the volume of data,");
    csh_file.newLine();
    csh_file.write("#    you may need to set tablespace/data file or table parameters");
    csh_file.newLine();
    csh_file.write("#    depending on your database engine, e.g. MyISAM, InnoDB");
    csh_file.newLine();
    csh_file.write("#");
    csh_file.newLine();
    csh_file.write("# Database connection parameters");
    csh_file.newLine();
    csh_file.write("#");
    csh_file.newLine();
    csh_file.write("set MYSQL_HOME=<path to MYSQL_HOME>");
    csh_file.newLine();
    csh_file.write("set user=<username>");
    csh_file.newLine();
    csh_file.write("set password=<password>");
    csh_file.newLine();
    csh_file.write("set db_name=<db_name>");
    csh_file.newLine();
    csh_file.newLine();

    csh_file.write("echo \"----------------------------------------\"");
    csh_file.newLine();
    csh_file.write("echo \"Starting ... `/bin/date`\"");
    csh_file.newLine();
    csh_file.write("echo \"----------------------------------------\"");
    csh_file.newLine();
    csh_file.write("echo \"\"");
    csh_file.newLine();

    csh_file.write("$MYSQL_HOME/bin/mysql -u $user -p$password  $db_name"
        + " < mysql_tables.sql");
    csh_file.newLine();
    csh_file.write("$MYSQL_HOME/bin/mysql -u $user -p$password  $db_name"
        + " < mysql_indexes.sql");
    csh_file.newLine();
    csh_file.write("$MYSQL_HOME/bin/mysql -u $user -p$password  $db_name"
        + " < mrconso_mysql.sql");
    csh_file.newLine();

    csh_file.write("echo \"----------------------------------------\"");
    csh_file.newLine();
    csh_file.write("echo \"Finished ... `/bin/date`\"");
    csh_file.newLine();
    csh_file.write("echo \"----------------------------------------\"");
    csh_file.close();
  }

  /**
   * Writes the MySQL populate .bat script.
   * 
   * @throws IOException
   *           if anything goes wrong
   */
  private void writePopulateMysqlBatFile() throws IOException {
    BufferedWriter bat_file = openTargetFile("populate_mysql_db.bat");

    bat_file.write("::");
    bat_file.newLine();
    bat_file.write(":: RECOMMENDATIONS (READ THIS FIRST)");
    bat_file.newLine();
    bat_file.write(":: When creating a MySQL database for loading UMLS content,");
    bat_file.newLine();
    bat_file.write(":: you should perform your configuration with the following");
    bat_file.newLine();
    bat_file.write(":: recommendations in mind:");
    bat_file.newLine();
    bat_file.write("::  o Create the database using UTF-8 as the default character set");
    bat_file.newLine();
    bat_file.write("::  o Create a separate user, e.g. UMLS");
    bat_file.newLine();
    bat_file.write("::  o To make sure the user is able to load the volume of data,");
    bat_file.newLine();
    bat_file.write("::    you may need to set tablespace/data file or table parameters");
    bat_file.newLine();
    bat_file.write("::    depending on your database engine, e.g. MyISAM, InnoDB");
    bat_file.newLine();
    bat_file.write("::");
    bat_file.newLine();
    bat_file.write(":: Database connection parameters");
    bat_file.newLine();
    bat_file.write("::");
    bat_file.newLine();
    bat_file.write("set MYSQL_HOME=<path to MYSQL_HOME>");
    bat_file.newLine();
    bat_file.write("set user=<username>");
    bat_file.newLine();
    bat_file.write("set password=<password>");
    bat_file.newLine();
    bat_file.write("set db_name=<db_name>");
    bat_file.newLine();
    bat_file.newLine();

    bat_file.write("echo ----------------------------------------");
    bat_file.newLine();
    bat_file.write("echo Starting ... ");
    bat_file.newLine();
    bat_file.write("echo ----------------------------------------");
    bat_file.newLine();
    bat_file.write("echo. ");
    bat_file.newLine();
    bat_file.flush();

    bat_file
        .write("%MYSQL_HOME%\\bin\\mysql -u %user% -p%password% --local-infile=1 %db_name% "
            + "< mysql_tables.sql");
    bat_file.newLine();
    bat_file
        .write("%MYSQL_HOME%\\bin\\mysql -u %user% -p%password% --local-infile=1 %db_name% "
            + "< mysql_indexes.sql");
    bat_file.newLine();
    bat_file
        .write("%MYSQL_HOME%\\bin\\mysql -u %user% -p%password% --local-infile=1 %db_name% "
            + "< mrconso_mysql.sql");
    bat_file.newLine();

    bat_file.write("echo ----------------------------------------");
    bat_file.newLine();
    bat_file.write("echo Finished  ");
    bat_file.newLine();
    bat_file.write("echo ----------------------------------------");
    bat_file.close();
  }

  /**
   * Writes the MySQL create script (.sql).
   * 
   * @throws IOException
   *           if anything goes wrong.
   */
  private void writeCreateMysqlDbFile() throws IOException {
    BufferedWriter sql_file = openTargetFile("create_mysql_db.sql");
    sql_file
        .write("CREATE DATABASE IF NOT EXISTS %db_name% CHARACTER SET UTF8;");
    sql_file.newLine();
    sql_file.close();
  }

  /**
   * Writes the MySQL create/load .sql script.
   * 
   * @throws IOException
   *           if anything goes wrong
   */
  private void writeMysqlTablesFile() throws IOException {
    BufferedWriter tables_sql = openTargetFile("mysql_tables.sql");
    List languages = Language.getLanguagesAsList();

    String sep = System.getProperty("line.separator");
    if (sep.equals("\n"))
      sep = "\\n";
    if (sep.equals("\r\n"))
      sep = "\\r\\n";
    if (sep.equals("\r"))
      sep = "\\r";

    for (int i = 0; i < target_files.length; i++) {
      writeMysqlTable(target_files[i], tables_sql);
      tables_sql.write("load data local infile \'" + target_files[i]
          + "\' into table " + target_files[i].replace('.', '_')
          + " fields terminated by \'|\' lines terminated by \'" + sep + "\';");
      tables_sql.newLine();
      tables_sql.newLine();
    }

    Collections.sort(languages);
    for (int i = 0; i < languages.size(); i++) {
      String base_file = "MRXW."
          + ((Language) languages.get(i)).getAbbreviation();
      writeMysqlTable(base_file, tables_sql);
      tables_sql.write("load data local infile \'" + base_file
          + "\' into table " + base_file.replace('.', '_')
          + " fields terminated by \'|\' lines terminated by \'" + sep + "\';");
      tables_sql.newLine();
      tables_sql.newLine();
    }

    writeMysqlTable("AMBIG.SUI", tables_sql);
    tables_sql.write("load data local infile \'" + "AMBIG.SUI"
        + "\' into table " + "AMBIG.SUI".replace('.', '_')
        + " fields terminated by \'|\' lines terminated by \'" + sep + "\';");
    tables_sql.newLine();
    tables_sql.newLine();

    writeMysqlTable("AMBIG.LUI", tables_sql);
    tables_sql.write("load data local infile \'" + "AMBIG.LUI"
        + "\' into table " + "AMBIG.LUI".replace('.', '_')
        + " fields terminated by \'|\' lines terminated by \'" + sep + "\';");
    tables_sql.newLine();

    tables_sql.close();
  }

  /**
   * Writes a file called "indexes.sql" that is used for creating indexes on the
   * tables loaded into the db.
   * 
   * @throws IOException
   */
  private void writeMysqlIndexesFile() throws IOException {

    BufferedWriter target = openTargetFile("mysql_indexes.sql");
    for (int i = 0; i < target_files.length; i++) {
      writeMysqlIndexesForTable(target_files[i], target,
          (List) files_to_index_fields.get(target_files[i]));
    }
    List languages = Language.getLanguagesAsList();
    Object[] indexes_indexes = { "WD" };
    Collections.sort(languages);
    for (int i = 0; i < languages.size(); i++) {
      String base_file = "MRXW."
          + ((Language) languages.get(i)).getAbbreviation();
      writeMysqlIndexesForTable(base_file, target, Arrays
          .asList(indexes_indexes));
    }
    writeMysqlIndexesForTable("AMBIG.SUI", target, (List) files_to_index_fields
        .get("AMBIG.SUI"));
    writeMysqlIndexesForTable("AMBIG.LUI", target, (List) files_to_index_fields
        .get("AMBIG.LUI"));
    target.close();
  }

  /**
   * Writes to the "indexes.sql" file all indexes related to the given
   * table_name.
   * 
   * @param table_name
   *          {@link String}
   * @param target
   *          {@link BufferedWriter} "indexes.sql"
   * @param dex
   *          {@link List} of columns needing indexes on the given table_name
   * @throws IOException
   */
  private void writeMysqlIndexesForTable(String table_name,
      BufferedWriter target, List dex) throws IOException {
    // Something like MRXW.ENG may get passed
    // in, convert it to MRXW_ENG
    String revised_table_name = table_name.replace('.', '_');
    if (dex == null) {
      return;
    }
    ArrayList indexes = new ArrayList(dex);

    for (int i = 0; i < indexes.size(); i++) {
      // Write index definition
      target.newLine();

      // if data type is a blob, put index only on first 255 chars
      String length = "";
      FileStatistics fs = (FileStatistics) config.getFileStatistics(table_name);
      ColumnStatistics[] col_stats = fs.getAllColumnStatistics();
      for (int j = 0; j < col_stats.length; j++) {
        if (col_stats[j].getColumnName().equals(indexes.get(i))
            && col_stats[j].getDataType().startsWith("varchar")
            && new Integer(col_stats[j].getDataTypeLength()).intValue() > 255) {
          length = "(255)";
        }
      }

      target.write("CREATE INDEX X_" + revised_table_name + "_"
          + ((String) indexes.get(i)).replace(',', '_') + " ON "
          + revised_table_name + "(" + indexes.get(i) + length + ");");
      target.newLine();
    }
  }

  /**
   * Compares this object with the specified object for order.  Returns a
   * negative integer, zero, or a positive integer as this object is less
   * than, equal to, or greater than the specified object.<p>
   * <p/>
   * Required method for implementing {@link Comparable} interface.
   * Indicates the means for comparison between two
   * {@link MetamorphoSysOutputStream}s.
   * @param o the Object to be compared.
   * @return a negative integer, zero, or a positive integer as this object
   *         is less than, equal to, or greater than the specified object.
   * @throws ClassCastException if the specified object's type is not
   *         {@link GUIConfigurable}.
   */
  public int compareTo(Object o) {
    if (!(o instanceof GUIConfigurable)) {
      throw new ClassCastException();
    }
    return getName().compareTo(((GUIConfigurable) o).getName());
  }

  /**
   * "Browse" button handler.
   */
  private class SubsetDirAction extends AbstractAction {

    /**
     * Instantiates a {@link SubsetDirAction}.
     */
    private SubsetDirAction() {
      super();

      putValue(Action.NAME, "Browse..");
      putValue(Action.SHORT_DESCRIPTION, "Subset Folder");
    }

    /**
     * Prompts the user for a new subset directory and sets this value in the
     * text field.
     * 
     * @param e
     *          an {@link ActionEvent}
     */
    public void actionPerformed(ActionEvent e) {
      if (ignore_events) {
        return;
      }
      File dir = new File(user_config.getSubsetDirectory());
      File subset_dir = MMSToolkit.chooseFile(
          "Select and Open the Subset Folder", "Open..",
          JFileChooser.DIRECTORIES_ONLY, (dir.exists() ? dir : null), null,
          MMSToolkit.getGui());
      if (subset_dir != null) {
        if (!validate()) {
          MMSToolkit.reportError(MMSToolkit.getGui(), getValidateMessage(),
              false);
          return;
        }
        fireDataChanged();
        user_config.setSubsetDirectory(subset_dir.getAbsolutePath());
        sd_field.setText(subset_dir.getAbsolutePath());
      }
    }
  }

}
