/**************************************************************************
 *
 * Package:     gov.nih.nlm.mms
 * Object:      RichMRMetamorphoSysInputStream.java
 *
 **************************************************************************/
package gov.nih.nlm.mms;

import gov.nih.nlm.swing.DataChangeEvent;
import gov.nih.nlm.swing.DataChangeListener;
import gov.nih.nlm.swing.SwingToolkit;
import gov.nih.nlm.umls.io.RichMRMetathesaurusInputStream;
import gov.nih.nlm.umls.meta.Concept;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.border.CompoundBorder;
import javax.swing.border.EmptyBorder;
import javax.swing.border.EtchedBorder;
import javax.swing.border.TitledBorder;

/**
 * Represents a {@link MetamorphoSysInputStream} that reads content from a
 * directory containing Rich Release Format Metathesaurus files.
 * 
 * @author Deborah Shapiro
 */
public class RichMRMetamorphoSysInputStream extends
    RichMRMetathesaurusInputStream implements MetamorphoSysInputStream {

  //
  // Private Fields
  //

  //
  // Data model for the source paths
  //
  protected UserConfiguration user_configuration;

  //
  // Data model for default configuraiton
  //
  protected ApplicationConfiguration default_config;

  //
  // CUI list reference
  //
  protected CuiList cui_list = null;

  //
  // Validation error message
  protected String validate_msg = null;

  //
  // Boolean state variables
  //
  private boolean data_changed = false, ignore_events = false,
      enable_efficient = false, init_cui = false;

  //
  // Data change listeners
  //
  private List listeners = new ArrayList();

  //
  // Text field showing user selected source path
  //
  private JTextField sd_field = null;

  //
  // Constructors
  //

  /**
   * Instantiates an empty {@link RichMRMetamorphoSysInputStream}.
   */
  public RichMRMetamorphoSysInputStream() {
  }

  //
  // Methods
  //

  /**
   * Sets the {@link UserConfiguration}.
   * 
   * @param user_configuration
   *          the {@link UserConfiguration}.
   */
  public void setConfiguration(UserConfiguration user_configuration) {
    this.user_configuration = user_configuration;
  }

  /**
   * Returns the short name for display purposes.
   * 
   * @return the short name for display purposes.
   */
  public String getName() {
    return "Rich Release Format";
  }

  /**
   * Returns a message for the log.
   * 
   * @return a message for the log
   */
  public String getConfigurationDetails() {
    return "MetamorphoSys Input: Rich Release Format";
  }

  /**
   * Sets the {@link CuiList} reference.
   * 
   * @param cui_list
   *          {@link CuiList}
   */
  public void setCuiList(CuiList cui_list) {
    this.cui_list = cui_list;
  }

  /**
   * Returns a {@link Concept} containing data from the MR Files. Different from
   * UMLS objects implementation in that it requires that the {@link Concept} be
   * present on the {@link CuiList} before reading the {@link Concept}'s data
   * from the MR Files.
   * 
   * @return a {@link Concept} containing data from the MR Files.
   * @throws IOException
   *           if files could not be read
   */
  public Concept readConcept() throws IOException {
    String current_concept = "";
    this_concept.clear();

    //
    // If in efficient mode and during initialize_cui stage
    //
    if (enable_efficient && init_cui) {

      //
      // Create Atoms for each group of rows from MRCONSO with the given CUI
      //
      if (getAtoms(this_concept, current_concept) == null)
        return null;

      current_concept = this_concept.getCui();

      //
      // Create SemanticTypes for each row in MRSTY with the given CUI
      //
      getSemanticTypes(this_concept, current_concept);

      //
      // Concept is fully populated (for efficient mode), return it
      //
      return this_concept;

    } else {

      //
      // Create Atoms for each group of rows from MRCONSO with the given CUI
      //

      if (getAtoms(this_concept, current_concept) == null)
        return null;

      current_concept = this_concept.getCui();

      if (init_cui || cui_list.inList(current_concept)) {
        //
        // Create Definitions for each row in MRDEF with the given CUI
        //
        getDefinitions(this_concept, current_concept);

        //
        // Create SemanticTypes for each row in MRSTY with the given CUI
        //
        getSemanticTypes(this_concept, current_concept);

        //
        // Create Relationships for each row in MRREL with the given CUI
        //
        getRelationships(this_concept, current_concept);

        //
        // Create Attributes for each row in MRSAT with the given CUI
        //
        getAttributes(this_concept, current_concept);

        //
        // Create Coocurrence for each row in MRCOC with the given CUI
        //
        getCoocurrences(this_concept, current_concept);

        //
        // Create Contexts and ContextMembers for rows in MRCXT
        //
        getContexts(this_concept, current_concept);

        //
        // Create ComponentHistories for rows in MRHIST with the given CUI
        //
        getComponentHistories(this_concept, current_concept);

        //
        // Create Mappings for rows in MRMAP with the given CUI
        //
        getMappings(this_concept, current_concept);
      }

      //
      // Concept is fully populated, return it
      //
      return this_concept;
    }
  }

  /**
   * Close files opened during {@link #open(String[])}.
   * 
   * @throws IOException
   *           if files can not be closed
   */
  public void close() throws IOException {
    super.close();
  }

  /**
   * Returns GUI panel that will be displayed on one of the tabs on the main GUI
   * window.
   * 
   * @return {@link JPanel} that displays graphical interface for a tab
   */
  public JPanel getPanel() {

    //
    // Configure panel
    //
    JPanel files_panel = new JPanel(new BorderLayout());

    //
    // Configure border
    //
    TitledBorder titledBorder = new TitledBorder(new EtchedBorder(
        EtchedBorder.RAISED), getName() + " Configuration");
    titledBorder.setTitleJustification(TitledBorder.CENTER);
    files_panel.setBorder(new CompoundBorder(titledBorder, new EmptyBorder(10,
        10, 10, 10)));

    //
    // Add panel fields
    //
    // Configure source directory chooser
    //
    JLabel sd_label = new JLabel(
        "Source Folder - Location of Metathesaurus Files");
    sd_label.setForeground(Color.black);
    files_panel.add(sd_label, BorderLayout.NORTH);

    JPanel field_button_panel = new JPanel(new FlowLayout());
    String[] paths = user_configuration.getSourcePaths();
    sd_field = new JTextField(paths[0], 50);
    field_button_panel.add(sd_field);
    JButton sd_button = new JButton(new SourceFileAction());
    field_button_panel.add(sd_button);
    files_panel.add(field_button_panel, BorderLayout.CENTER);

    JPanel tmppanel = new JPanel(new BorderLayout());
    files_panel.add(tmppanel, BorderLayout.SOUTH);

    return files_panel;
  }

  /**
   * Returns a boolean indicating whether or not the values on this gui have
   * changed.
   * 
   * @return <code>boolean</code> <code>true</code> if data has changed
   *         <code>false</code> otherwise
   */
  public boolean hasDataChanged() {
    return data_changed;
  }

  /**
   * Sets the "has data changed" value.
   * 
   * @param b
   *          <code>boolean</code> indicating whether or not data in the GUI
   *          configuration has changed.
   */
  public void setDataChanged(boolean b) {
    data_changed = b;
  }

  /**
   * Returns the current state of the input stream as a {@link Properties}
   * object.
   * 
   * @return {@link Properties} only those directly related to the gui
   */
  public Properties getProperties() {
    Properties p = new Properties();
    p.setProperty("source_path", user_configuration.getSourcePaths()[0]);
    return p;

  }

  /**
   * Sets the current state of the input stream from a {@link Properties}
   * object.
   * 
   * @param p
   *          {@link Properties} pertaining to the gui
   */
  public void setProperties(Properties p) {
    if (p.containsKey("source_path")) {
      user_configuration.setSourcePaths(new String[] { p
          .getProperty("source_path") });
    }
  }

  /**
   * Resets the "has data changed" flag to indicate that no data has been
   * changed since the last save.
   */
  public void resetDataChanged() {
    data_changed = false;
  }

  /**
   * Adds a {@link DataChangeListener} that will be notified every time a
   * {@link #fireDataChanged()} call is made. Allows main GUI to record all
   * changes to this gui in the undo/redo mechanism.
   * 
   * @param dcl
   *          {@link DataChangeListener}
   */
  public void addDataChangeListener(DataChangeListener dcl) {
    listeners.add(dcl);
  }

  /**
   * Removes a {@link DataChangeListener} so that it will no longer be notified
   * every time a {@link #fireDataChanged} call is made from this gui.
   * 
   * @param dcl
   *          {@link DataChangeListener}
   */
  public void removeDataChangeListener(DataChangeListener dcl) {
    listeners.remove(dcl);
  }

  /**
   * Returns the {@link Color} that should be used to display the name.
   * 
   * @return {@link Color}
   */
  public Color getColor() {
    return Color.blue;
  }

  /**
   * Returns a basic description of how to configure the input stream.
   * 
   * @return {@link String} text of input stream instructions
   */
  public String getTabDescription() {
    return "Select the format in which subset data was saved and indicate the "
        + "location of the source folder that contains the existing subset "
        + "configuration.  See Help for more information.\n\n"
        + "Users may proceed to the other Options tabs in any order by clicking "
        + "on the tabs across the top of the screen or by clicking \"Done\" on "
        + "the File bar and selecting \"Begin Subset\".";
  }

  /**
   * Resets the state of the input stream to its default state. This is not used
   * to actually change the gui. setPanelSelections() is always called after
   * this method and will update the gui based on the variable values.
   */
  public void resetToDefaults() {
    ignore_events = true;
    fireDataChanged();
    user_configuration.setSourcePaths(default_config.getSourcePaths());
    ignore_events = false;
  }

  /**
   * Presents an advanced options GUI to the user.
   */
  public void showOptions() {
    // do nothing
  }

  /**
   * Indicates whether or not an options menu item should be supplied for
   * advanced configuration options.
   * 
   * @return <code>true</code> if so; <code>false</code> otherwise
   */
  public boolean hasAdvancedOptions() {
    return false;
  }

  /**
   * Provides a {@link JPanel} that will be displayed when the user selects the
   * corresponding menu item from the Help menu and will provide the user with
   * detailed instructions on the tab's use.
   * 
   * @return {@link JPanel} to be added to the Help menu
   */
  public JPanel getHelpPanel() {
    return SwingToolkit.getPanelForHTML(this.getClass().getClassLoader()
        .getResource("config/InputOptionsHelp.html"));
  }

  /**
   * Applies the input stream state to the GUI panel.
   */
  public void setPanelSelections() {
    ignore_events = true;

    //
    // Configure directory
    //
    sd_field.setText(user_configuration.getSourcePaths()[0]);

    ignore_events = false;

  }

  /**
   * Indicates whether or not the {@link GUIConfigurable} is valid.
   * 
   * @return <code>boolean</code><code>true</code> if
   *         {@link GUIConfigurable} is valid <code>false</code> if
   *         {@link GUIConfigurable} is invalid
   */
  public boolean validate() {
    String[] paths = user_configuration.getSourcePaths();
    return validateSourcePaths(paths);
  }

  /**
   * Returns the message that the user will see if this {@link GUIConfigurable}
   * is invalid.
   * 
   * @return {@link String} with error message
   */
  public String getValidateMessage() {
    return (validate_msg == null) ? "" : validate_msg;
  }

  /**
   * Returns the messages indicating why validation failed. Should only be
   * called if {@link #validate()} returns <code>false</code>.
   * 
   * @return the failure messages
   */
  public String[] getValidationErrors() {
    return new String[] { getValidateMessage() };
  }

  /**
   * Returns warnings that were indicated during validation.
   * 
   * @return the warning messages
   */
  public String[] getValidationWarnings() {
    return new String[0];
  }

  /**
   * Returns the {@link GUIConfigurable} whose tab should next be displayed to
   * allow the user to correct an error or warning.
   * 
   * @return GUIConfigurable to be displayed
   */
  public GUIConfigurable getGUIConfigurableForValidationFailure() {
    return null;
  }

  /**
   * Validates the source paths. This involves checking that they both exist,
   * are readable and that they have access to MR data.
   * 
   * @param source_paths
   *          to be validated
   * @return <code>true</code> if validation passes <code>false</code>
   *         otherwise
   */
  public boolean validateSourcePaths(String[] source_paths) {
    return validateSourcePath(source_paths[0]);
  }

  /**
   * Validates the source path. This involves checking that it both exists, is
   * readable and that it has access to MR data.
   * 
   * @param source_path
   *          path to be validated
   * @return <code>true</code> if validation passes <code>false</code>
   *         otherwise
   */
  public boolean validateSourcePath(String source_path) {
    File source_dir = new File(source_path);
    if (!source_dir.isDirectory()) {
      validate_msg = "Invalid source location.\nPlease select a directory.";
      return false;
    } else if (!source_dir.exists()) {
      validate_msg = "Invalid source location.\nPlease select a directory that exists.";
      return false;
    } else if (!source_dir.canRead()) {
      validate_msg = "Invalid source location:" + "\n\n" + source_path
          + "\n\nThis directory is read-protected.";
      return false;
    } else {
      File[] files = source_dir.listFiles();
      for (int i = 0; i < files.length; i++) {
        if (files[i].getName().endsWith("MRFILES.RRF")) {
          return true;
        }
      }
    }
    // if this is called when the MetamorphoSysGUI is open, msg can mention
    // subsetting and selecting NLM format
    if (MMSToolkit.getGui() != null)
      validate_msg = "The source location you have selected, \n\n" +
    								source_path + "\n\n" +
    							 "does not contain the correct source files.\n" + 
    							 "Please enter a location that contains an RRF subset.\n" +
    							 "If this is your first subset, you must select NLM Data File Format."; 
    // otherwise this code may be called from MRCXT Builder or another app
    // that just needs to validate a rrf source directory
    else
      validate_msg = "The source location you have selected, \n\n" +
		source_path + "\n\n" +
		"does not contain the correct source files.\n" + 
		"Please enter a location that contains an RRF subset."; 
    	
    return false;
  }

  /**
   * For each source path it attempts to determine the version of the release
   * being used and validate it against the version from the properties file. If
   * the version is determined to be a mismatch of if it cannot be determined,
   * <code>false</code> is returned.
   * 
   * @param paths
   *          to be validated
   * @return <code>true</code> if validation passes <code>false</code>
   *         otherwise
   */
  public boolean validateVersions(String[] paths) {
    for (int i = 0; i < paths.length; i++) {
      if (validateVersion(paths[i]))
        continue;
      else
        return false;
    }
    return true;
  }

  /**
   * Attempts to determine the version of the release being used and validate it
   * against the version from the properties file. If the version is determined
   * to be a mismatch of if it cannot be determined, <code>false</code> is
   * returned.
   * 
   * @param source_path
   *          to be validated
   * @return <code>true</code> if validation passes <code>false</code>
   *         otherwise
   */
  public boolean validateVersion(String source_path) {
    InstallConfiguration ic = InstallConfiguration.getInstallConfiguration();
    // Get parent of installation directory, check the name
    File install_dir = new File(source_path);
    if (install_dir.isDirectory() && install_dir.getParent() != null) {
      File parent_dir = new File(install_dir.getParent());

      // Check for release.dat first in the installation directory.
      // If it is not there, check next in the directory above the installation
      // dir.
      // If it is in either of these places, compare the release_version in the
      // release.dat with that from the configuration file.
      File release_file = new File(source_path, "release.dat");
      if (!release_file.exists()) {
        release_file = new File(parent_dir, "release.dat");
      }
      if (release_file.exists()) {
        String mms_version = "";
        if (release_file.exists()) {
          // Open and load the properties file
          Properties props = new Properties();
          try {
            FileInputStream propInputStream = new FileInputStream(release_file);
            props.load(propInputStream);
            if (props.containsKey("umls.release.name"))
              mms_version = props.getProperty("umls.release.name");
          } catch (IOException ioe) {
            // do nothing
          }
        }

        if (mms_version.startsWith(ic.getReleaseVersion())) {
          return true;
        } else {
          validate_msg = "Warning: MetamorphoSys is not configured to run \n"
              + "against this version of the data, subsetting may be \n"
              + "incorrect. To correct this problem, either use the\n"
              + ic.getReleaseVersion() + " release of the Metathesaurus or \n"
              + "use the MetamorphoSys distributed with the " + mms_version
              + "\nrelease of the Metathesaurus. ";
        }
      }

      // If the release.dat wasn't found, check the name of the parent dir,
      // in the default installation, it will match the release version
      if (parent_dir.getName().startsWith(ic.getReleaseVersion())
          && !release_file.exists()) {
        return true;
      } else {
        validate_msg = "Warning: MetamorphoSys is configured to run against\n"
            + "the " + ic.getReleaseVersion()
            + " version of the Metathesaurus, \n"
            + "but the version of the data being used could not be\n"
            + "determined.  Please make sure you are subsetting the\n"
            + "correct version of the Metathesaurus.";
      }
    } else {
      validate_msg = "Invalid source location:\n\n" + install_dir
          + "\n\nPlease select a valid directory. ";
    }
    return false;
  }

  /**
   * Enables filters to inform listeners of data changed event.
   */
  protected void fireDataChanged() {
    for (int i = 0; i < listeners.size(); i++)
      ((DataChangeListener) listeners.get(i)).dataChanged(new DataChangeEvent(
          this));
    data_changed = true;
  }

  /**
   * Enables filters to indicate that data has changed without notifying the
   * listeners or making this change undoable.
   */
  protected void fireNotUndoableDataChanged() {
    data_changed = true;
  }

  /**
   * Indicates the type of file accepted by the {@link JFileChooser}.
   * 
   * @return {@link String}
   */
  public int getChooserType() {
    return JFileChooser.DIRECTORIES_ONLY;
  }

  /**
   * Sets the {@link ApplicationConfiguration} representing the default state.
   * 
   * @param c
   *          {@link ApplicationConfiguration}
   */
  public void setConfiguration(ApplicationConfiguration c) {
    default_config = c;
  }

  /**
   * Informs {@link SubsetListener} that the Initialize Cui List phase is
   * starting.
   */
  public void beginInitializeCuiList() {
    init_cui = true;
  }

  /**
   * Informs {@link SubsetListener} that the Initialize Cui List phase is
   * ending.
   */
  public void endInitializeCuiList() {
    init_cui = false;
  }

  /**
   * Informs {@link SubsetListener} that the Subset Content phase is starting.
   */
  public void beginSubsetContent() {
  }

  /**
   * Informs {@link SubsetListener} that the Subset Content phase is ending.
   */
  public void endSubsetContent() {
  }

  /**
   * Sets the subsetting mode to efficient. In the efficient mode, the
   * "Initialize Cuis" phase of the subsetting only references MRCONSO and
   * MRSTY.
   * 
   * @param efficient
   *          <code>true</code> if subsetting should be run in efficient mode
   *          <code>false</code> if not
   */
  public void setEfficientMode(boolean efficient) {
    enable_efficient = efficient;
  }

  /**
   * Returns the mode in which subsetting will be run. In the efficient mode,
   * the "Initialize Cuis" phase of the subsetting only references MRCONSO and
   * MRSTY.
   * 
   * @return <code>true</code> if subsetting will be run in efficent mode
   *         <code>false</code> otherwise
   */
  public boolean isEfficientMode() {
    return enable_efficient;
  }

  /**
   * Indicates whether this input source has full context data, such as that
   * included in a MRCXT file.
   *
   * @return <code>true</code> if this input source has full context data;
   *         <code>false</code> otherwise.
   */
  public boolean hasFullContextData() {
    String[] source_paths = user_configuration.getSourcePaths();
    File mrcxtFile = new File(source_paths[0], "MRCXT.RRF" + add_suffix);
    if (mrcxtFile != null && mrcxtFile.exists()) {
      return true;
    }
    else {
      return false;
    }
  }

  /**
   * Compares this object with the specified object for order.  Returns a
   * negative integer, zero, or a positive integer as this object is less
   * than, equal to, or greater than the specified object.<p>
   * <p/>
   * Required method for implementing {@link Comparable} interface.
   * Indicates the means for comparison between two
   * {@link MetamorphoSysInputStream}s.
   * @param o the Object to be compared.
   * @return a negative integer, zero, or a positive integer as this object
   *         is less than, equal to, or greater than the specified object.
   * @throws ClassCastException if the specified object's type is not
   *         {@link GUIConfigurable}.
   */
  public int compareTo(Object o) {
    if (!(o instanceof GUIConfigurable)) {
      throw new ClassCastException();
    }
    return getName().compareTo(((GUIConfigurable) o).getName());
  }

  /**
   * Action for the "Browse" button.
   */
  private class SourceFileAction extends AbstractAction {

    /**
     * Instantiates a {@link SourceFileAction}.
     */
    public SourceFileAction() {
      super();
      putValue(Action.NAME, "Browse..");
      putValue(Action.SHORT_DESCRIPTION, "Source Folder");
    }

    /**
     * Prompts the user for a new subset directory and sets this value in the
     * text field.
     * 
     * @param e
     *          an {@link ActionEvent}
     */
    public void actionPerformed(ActionEvent e) {

      //
      // If ignoring events, return
      //
      if (ignore_events)
        return;

      //
      // Get the selected file as a starting point
      //
      File selected_file = new File(user_configuration.getSourcePaths()[0]);

      //
      // Prompt the user to choose a file
      //
      File source_dir = MMSToolkit.chooseFile(
          "Select and Open the Source Folder", "Open..", getChooserType(),
          selected_file, selected_file, MMSToolkit.getGui());

      //
      // If a directory was selected, set the source path
      //
      if (source_dir != null) {
        String[] source_paths = { source_dir.getAbsolutePath() };

        //
        // Validate user selection
        //
        if (!validateSourcePaths(source_paths)) {
          MMSToolkit.reportError(MMSToolkit.getGui(), getValidateMessage(),
              false);
          return;
        }

        //
        // If validation passes, apply change
        //
        fireDataChanged();
        user_configuration.setSourcePaths(source_paths);

        //
        // Update GUI
        //
        sd_field.setText(source_dir.getAbsolutePath());

        //
        // Validate versions (just a warning message)
        //
        if (!validateVersions(source_paths)) {
          MMSToolkit.reportError(MMSToolkit.getGui(), getValidateMessage(),
              false);
        }
      }
    } // end actionPerformed
  } // end inner class
}