/*****************************************************************************
 *
 * Package: gov.nih.nlm.mms
 * Object:  SaveConfigAction.java
 *
 *****************************************************************************/
package gov.nih.nlm.mms;

import java.awt.Component;
import java.awt.Frame;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.JFileChooser;
import javax.swing.KeyStroke;
import javax.swing.filechooser.FileFilter;

/**
 * Saves a config file. Used for the <code>File->Save Configuration</code>
 * menu of {@link MetamorphoSysGUI}.
 * 
 * @author Brian Carlsen, Deborah Shapiro
 */
public class SaveConfigAction extends AbstractAction {

  //
  // Protected Fields
  //
  protected MetamorphoSys mms;

  protected Frame owner;

  protected static File lastConfigFile = new File("config", "mmsys.prop.sav");

  //
  // Constructor
  //

  /**
   * Instantiates a {@link SaveConfigAction}.
   * 
   * @param mms
   *          the {@link MetamorphoSys}
   * @param owner parent {@link Component}
   */
  public SaveConfigAction(MetamorphoSys mms, Frame owner) {
    super();
    this.mms = mms;
    this.owner = owner;

    // configure action
    putValue(Action.NAME, "Save Configuration..");
    putValue(Action.SHORT_DESCRIPTION, "Save Configuration");
    putValue(Action.MNEMONIC_KEY, new Integer((int) 's'));
    putValue(Action.ACCELERATOR_KEY, KeyStroke.getKeyStroke('S',
    		Toolkit.getDefaultToolkit().getMenuShortcutKeyMask()));
  }

  /**
   * Saves the {@link UserConfiguration} to a file.
   * 
   * @param e
   *          an {@link ActionEvent}
   */
  public void actionPerformed(ActionEvent e) {
    saveConfig();
  }

  /**
   * Helper method for saving the configuration.
   */
  protected void saveConfig() {

    //
    // Get configuration references
    //
    UserConfiguration user_configuration = mms.getUserConfiguration();
    ApplicationConfiguration default_configuration = mms
        .getApplicationConfiguration();

    lastConfigFile =
      new File(default_configuration.getConfigDirectory(),"mmsys.prop.sav");
    
    //
    // Validate filters before saving
    //
    if (!MMSToolkit.getGui().validateGUIConfigurables())
      return;

    //
    // Determine "default" save directory
    //
    File to_save = new File(user_configuration
        .getCurrentConfigurationFilename());
    File save_dir;
    if (to_save.getName().startsWith("mmsys.")
        && to_save.getName().endsWith(".prop")) {
      save_dir = new File(user_configuration.getSubsetDirectory())
          .getParentFile();
      to_save = null;
    } else
      save_dir = to_save.getParentFile();

    //
    // Prompt user for file to save to
    //
    FileFilter propertiesFileFilter = new PropertiesFileFilter();
    to_save = MMSToolkit.chooseFile("Save Configuration File", "Save",
        JFileChooser.FILES_ONLY, save_dir, to_save, propertiesFileFilter,
        "properties", owner);

    //
    // File was selected, retrieve it and save configuration
    //
    if (to_save != null) {
      lastConfigFile = to_save;

      try {

        if (to_save.getParentFile() != null
            && !to_save.getParentFile().canWrite())
          throw new LocalException("Directory is not writable: "
              + to_save.getParentFile().getAbsolutePath());

        user_configuration.setCurrentConfigFile(to_save.toString());

        //
        // Save config to file
        //
        user_configuration.saveState(to_save);

        //
        // Write been_saved.txt
        //
        String config_dir = mms.getApplicationConfiguration()
            .getConfigDirectory();
        if (new File(config_dir).canWrite()) {
          try {
            new BufferedWriter(new FileWriter(new File(config_dir,
                "been_saved.txt")));
          } catch (Exception ex) {
            // 
            // Treat this as if canWrite() returned true
            // for a DVD RW drive with a read-only DVD in it
            //
          }
        }

        //
        // File saved, reset data changed
        //
        user_configuration.resetDataChanged();
        
        backupConfig();

      } catch (LocalException le) {
        //
        // Error saving file, give user another chance
        //
        MMSToolkit.reportError(owner,
            "There was an error saving your configuration.\n\n"
                + le.getMessage() + "\n\n" + "Please try again.", false);
        saveConfig();
        return;
      }
    }
  }
  
  /**
   * Backup the config file to mmsys.prop.sav.
   */
  public void backupConfig() {
    try {
      //
      // ALWAYS save to mmsys.prop.sav also (if config directory is not on DVD)
      //
      File config_file = new File(mms.getApplicationConfiguration().getConfigDirectory(),
          "mmsys.prop.sav");
      if (config_file.exists() && config_file.getParentFile().canWrite()
          && config_file.canWrite()) {
        mms.getUserConfiguration().saveState(config_file);
      }
    } catch (LocalException le) {
      // 
      // Treat this as if canWrite() returned true
      // for a DVD RW drive with a read-only DVD in it
      //
    }

  }

  /**
   * {@link FileFilter} class for {@link JFileChooser} to limit files to
   * .properties file extension.
   */
  protected static class PropertiesFileFilter extends FileFilter {
    /**
     * Indicates whether a file has an extension of .properties, or is a
     * directory to allow navigation.
     * @param file file or directory
     * @return <code>true</code> if the file is a .properties file or a
     *         directory.
     */
    public boolean accept(File file) {
      String fileName = file.getName();

      // If it is a directory -- we want to show it so return true.
      if (file.isDirectory())
        return true;

      if (fileName.toUpperCase().endsWith(".PROPERTIES"))
        return true;

      return false;
    }

    /**
     * Returns the .properties file filter description.
     * @return the filter description.
     */
    public String getDescription() {
      return "Java Properties Files (*.properties)";
    }
  }
}
