/**************************************************************************
 *
 * Package:    gov.nih.nlm.mms;
 * Interface:  SubsetGUI.java
 *
 *************************************************************************/
package gov.nih.nlm.mms;

import gov.nih.nlm.swing.FontSizeManager;
import gov.nih.nlm.swing.GlassComponent;
import gov.nih.nlm.swing.GlassPaneListener;
import gov.nih.nlm.util.FieldedStringTokenizer;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.*;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Properties;

import javax.swing.*;
import javax.swing.border.CompoundBorder;
import javax.swing.border.EmptyBorder;
import javax.swing.border.EtchedBorder;

/**
 * This dialog is used to choose the configuration options. The choices are to
 * make a new configuration, to open a specified configuration file, or to use
 * the last saved configuration.
 * 
 * @author Deborah Shapiro
 */
public class SubsetGUI extends JFrame implements GlassPaneListener {

  private GlassComponent gc = null;

  /**
   * Instantiates the GUI. The constructor does not make it visible.
   * 
   * @param mms
   *          {@link MetamorphoSys}
   */
  public SubsetGUI(MetamorphoSys mms) {
    super();
    ApplicationConfiguration default_config = mms.getApplicationConfiguration();
    final InstallConfiguration ic = InstallConfiguration
        .getInstallConfiguration();

    //
    // Indicate which mode we are in and that the users's
    // selections now should override those in the config file chosen.
    //
    ic.setOverrideConfigurationFiles(true);
    ic.setInstallUMLS(false);
    mms.getUserConfiguration().setInputStream(ic.getMetamorphoSysInputStream());

    //
    // Disable welcome gui
    //
    WelcomeGUI.getWelcomeGUI().enableGlassPane();

    //
    // Create glass component and install
    //
    gc = new GlassComponent(this);
    setGlassPane(gc);

    //
    // Set title
    //
    setTitle("Customize Subset");

    //
    // Configure layout
    //
    getContentPane().setLayout(new BorderLayout());
    GridBagLayout gridbag = new GridBagLayout();
    JPanel choice_panel = new JPanel(gridbag);
    GridBagConstraints constraints = new GridBagConstraints();
    constraints.fill = GridBagConstraints.BOTH;
    constraints.insets = new Insets(2, 2, 2, 2);
    constraints.gridy = 0;
    constraints.gridx = 0;
    constraints.gridwidth = 1;
    constraints.gridy++;

    //
    // Compose menu bar
    //
    JMenuBar menu_bar = new JMenuBar();
    setJMenuBar(menu_bar);
    menu_bar.add(Box.createHorizontalGlue());
    JMenu help_menu = new JMenu("Help");
    menu_bar.add(help_menu);
    buildHelpMenu(help_menu);

    //
    // add Source Directory Installation Row
    //
    constraints.gridx = 0;
    choice_panel.add(new JLabel("Source"), constraints);

    //
    // add Source JTextField
    //
    constraints.gridx = 2;
    final JTextField source_field = new JTextField(40);
    source_field.setEditable(false);

    final String source_prop =
        System.getProperty("gov.nih.nlm.mms.SubsetGUI.source");
    if (source_prop != null) {
      setUmlsDirectory(new File(source_prop), source_field, ic);
    }
    else {
      try {
        //
        // If "been_saved.txt" file exists, prepopulate Source field from saved
        //  value
        //
        if (new File(default_config.getConfigDirectory(),
                     "been_saved.txt").exists()) {
          Properties props = new Properties();
            FileInputStream propInputStream = new FileInputStream(new File(
                default_config.getConfigDirectory(), "mmsys.prop.sav"));
            props.load(propInputStream);
          // Set the source path according to the source_paths property
          if (props.containsKey("source_paths")) {
            String source_path_list = props.getProperty("source_paths");
            String[] source_paths =
                FieldedStringTokenizer.split(source_path_list, ";");
            setUmlsDirectory(new File(source_paths[0]), source_field, ic);
          }
        }
        //
        // else set default metathesaurus file
        //
        else {
          source_field.setText(new File("..").getCanonicalPath());
        }
      } catch (IOException ioe) {
        source_field.setText("..");
      }
    }
    source_field.setBackground(Color.white);
    gridbag.setConstraints(source_field, constraints);
    choice_panel.add(source_field);

    //
    // add Source Browse JButton
    //
    constraints.gridx = 3;
    JButton source_browse = new JButton("Browse..");
    source_browse.setPreferredSize(new Dimension(100, 22));
    source_browse.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        try {

          File selected_file = null;
          if (!source_field.getText().equals("")) {
            selected_file = new File(source_field.getText()).getCanonicalFile();
          }
          File source_dir = MMSToolkit.chooseFile(
              "Select and Open the source directory", "Open..", ic
                  .getMetamorphoSysInputStream().getChooserType(),
              selected_file, selected_file, SubsetGUI.this);
          if (source_dir != null) {
            setUmlsDirectory(source_dir, source_field, ic);
          }
        } catch (Exception ex) {
          ex.printStackTrace();
          MMSToolkit.reportError(SubsetGUI.this,
              "Unexpected error while setting source directory.\n"
                  + "See console for details.", false);
        }
      }
    });
    gridbag.setConstraints(source_browse, constraints);
    choice_panel.add(source_browse);

    //
    // add Source JLabel comment
    //
    constraints.gridy++;
    constraints.gridx = 2;
    constraints.insets = new Insets(0, 6, 10, 0);
    JLabel comment = new JLabel("(this directory contains Metathesaurus files)");
    choice_panel.add(comment, constraints);
    constraints.insets = new Insets(2, 2, 2, 2);
    constraints.gridy++;

    //
    // add destination Directory Installation Row
    //
    constraints.gridy++;
    constraints.gridx = 0;
    constraints.gridwidth = 1;

    choice_panel.add(new JLabel("Destination"), constraints);

    //
    // add destination JTextField
    //
    constraints.gridx = 2;
    final JTextField destination_field = new JTextField(40);
    destination_field.setEditable(false);
    destination_field.setBackground(Color.white);
    gridbag.setConstraints(destination_field, constraints);
    choice_panel.add(destination_field);

    final String destination_prop =
        System.getProperty("gov.nih.nlm.mms.SubsetGUI.destination");
    if (destination_prop != null) {
      setMetaDestinationDirectory(new File(destination_prop), destination_field,
                                  ic);
    }
    else {
      try {
        //
        // If "been_saved.txt" file exists, prepopulate Destination field from
        //  saved value
        //
        if (new File(mms.getApplicationConfiguration().getConfigDirectory(),
                     "been_saved.txt").exists()) {
          Properties props = new Properties();
          FileInputStream propInputStream = new FileInputStream(new File(
              default_config.getConfigDirectory(), "mmsys.prop.sav"));
          props.load(propInputStream);
          // Set the destination path according to the subsetname property
          if (props.containsKey("subsetname")) {
            String file_name = props.getProperty("subsetname");
            setMetaDestinationDirectory(new File(file_name), destination_field, ic);
          }
        }
      } catch (IOException ioe) {
      }
    }

    //
    // add destination Browse JButton
    //
    constraints.gridx = 3;
    JButton destination_browse = new JButton("Browse..");
    destination_browse.setPreferredSize(new Dimension(100, 22));
    destination_browse.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        try {
          File selected_file = null;

          if (!destination_field.getText().equals("")) {
            selected_file = new File(destination_field.getText())
                .getCanonicalFile();
          } else {
            selected_file = new File(source_field.getText()).getParentFile();
          }

          File destination_dir = MMSToolkit.chooseFile(
              "Select and Open the destination directory", "Open..",
              JFileChooser.DIRECTORIES_ONLY, selected_file, selected_file,
              SubsetGUI.this);
          if (destination_dir != null) {
            setMetaDestinationDirectory(destination_dir, destination_field, ic);
          }
        } catch (Exception ex) {
          ex.printStackTrace();
          MMSToolkit.reportError(SubsetGUI.this,
              "Unexpected error while setting destination directory.\n"
                  + "See console for details.", false);
        }
      }
    });

    gridbag.setConstraints(destination_browse, constraints);
    choice_panel.add(destination_browse);

    //
    // add destination JLabel comment
    //
    constraints.gridy++;
    constraints.gridx = 2;
    constraints.insets = new Insets(0, 6, 10, 0);
    comment = new JLabel("(this is where the output files will be)");
    choice_panel.add(comment, constraints);
    constraints.insets = new Insets(2, 2, 2, 2);
    constraints.gridy++;

    //
    // Button Panel
    //
    JPanel button_panel = new JPanel(new FlowLayout());

    // New Configuration Button
    Action local_new_config = new NewConfigAction(mms, WelcomeGUI
        .getWelcomeGUI()) {

      public void actionPerformed(ActionEvent e) {
        if (!validateChoices())
          return;
        MMSProgressMonitor progress = ic.getProgressMonitor();
        progress.setOperation(MMSProgressMonitor.CONFIG_SUBSET);
        dispose();
        super.actionPerformed(e);
        if (isCancelled()) {
          disableGlassPane();
          return;
        }
      }
    };
    JButton jnew = new JButton(local_new_config);
    button_panel.add(jnew);

    // Open Configuration Button
    constraints.gridy++;
    Action local_open_config = new OpenConfigAction(mms, this) {
      public void actionPerformed(ActionEvent e) {
        if (!validateChoices())
          return;
        if (isCancelled()) {
          disableGlassPane();
          return;
        }
        MMSProgressMonitor progress = ic.getProgressMonitor();
        progress.setOperation(MMSProgressMonitor.CONFIG_SUBSET);
        dispose();
        super.actionPerformed(e);
      }
    };
    JButton jopen = new JButton(local_open_config);
    button_panel.add(jopen);

    // Last Saved Configuration Button
    constraints.gridy++;
    OpenConfigAction oca = new OpenConfigAction(mms, WelcomeGUI.getWelcomeGUI()) {
      public void actionPerformed(ActionEvent e) {
        if (!validateChoices())
          return;
        MMSProgressMonitor progress = ic.getProgressMonitor();
        progress.setOperation(MMSProgressMonitor.CONFIG_SUBSET);
        dispose();
        super.actionPerformed(e);
      }
    };

    oca.putValue(Action.NAME, "Last Saved Configuration..");
    oca.putValue(Action.SHORT_DESCRIPTION, "Last Saved Configuration");
    // Open and load the properties file in order to get the name of the
    // last saved configuration for the JTextField
    Properties props = new Properties();
    String file_name = "";
    try {
      FileInputStream propInputStream = new FileInputStream(new File(
          default_config.getConfigDirectory(), "mmsys.prop.sav"));
      props.load(propInputStream);
    } catch (IOException ioe) {
      ioe.printStackTrace();
      MMSToolkit.reportError(SubsetGUI.this,
          "Error while loading properties file\n"
              + "Please try opening another properties file", false);
    }
    // Set the path to the current config file
    if (props.containsKey("current_config_file")) {
      file_name = props.getProperty("current_config_file");
    }
    oca.setFileToOpen(new File(file_name));

    JButton jsaved = new JButton(oca);

    //
    // Only add "Last Saved" if "been_saved.txt" file exists
    //
    if (new File(mms.getApplicationConfiguration().getConfigDirectory(),
        "been_saved.txt").exists())
      button_panel.add(jsaved);

    //
    // Cancel button
    //
    // Create an action object
    Action cancelAction = new AbstractAction("Cancel") {
      {
        // Set action command to "Cancel"
        putValue(Action.ACTION_COMMAND_KEY, getValue(Action.NAME));
      }

      // This method is called when the action is invoked
      public void actionPerformed(ActionEvent e) {
        ic.clearProgressMonitor();
        dispose();
        return;
      }
    };
    JButton jcancel = new JButton(cancelAction);
    // Register keystroke
    button_panel.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(
        KeyStroke.getKeyStroke("ESCAPE"), cancelAction.getValue(Action.NAME));
    // Register action
    button_panel.getActionMap().put(cancelAction.getValue(Action.NAME),
        cancelAction);
    button_panel.add(jcancel);

    EtchedBorder etched_border = new EtchedBorder(EtchedBorder.RAISED);
    choice_panel.setBorder(new CompoundBorder(etched_border, new EmptyBorder(
        10, 10, 10, 10)));

    getContentPane().add(choice_panel, BorderLayout.CENTER);
    getContentPane().add(button_panel, BorderLayout.SOUTH);
    FontSizeManager.adjustFontSize(this);
    this.pack();
    setLocationRelativeTo(null);

    //
    // Configure window listener
    //
    addWindowListener(new WindowAdapter() {
      public void windowClosing(WindowEvent e) {
        ic.clearProgressMonitor();
        dispose();
        return;
      }
    });

  }

  /**
   * Updates the Source text field and sets the directory that contains the
   * Metathesaurus data.
   *
   * @param source_dir  source directory {@link File} of the Metathesaurus data
   * @param source_field  Source text field
   * @param ic  installation configuration
   */
  private static void setUmlsDirectory(File source_dir, JTextField source_field,
      InstallConfiguration ic) {
    source_field.setText(source_dir.getAbsolutePath());
    ic.setUmlsDirectory(source_dir);
  }

  /**
   * Updates the Destination text field and sets the directory into which the
   * Metathesaurus subset will be made.
   *
   * @param destination_dir  destination directory {@link File} for the subset
   * @param destination_field  Destination text field
   * @param ic  installation configuration
   */
  private static void setMetaDestinationDirectory(File destination_dir,
      JTextField destination_field, InstallConfiguration ic) {
    destination_field.setText(destination_dir.getAbsolutePath());
    ic.setMetaDestinationDirectory(destination_dir);
  }

  /**
   * Builds the help menu items.
   * 
   * @param help_menu
   *          the help {@link JMenu}
   */
  private void buildHelpMenu(JMenu help_menu) {

    //
    // Configure Subset Help Menu Item
    //
    help_menu.add(MMSToolkit.getResourceHelpMenu("Configure Subset Help",
        "config/ConfigureSubsetHelp.html", this));

    //
    // UMLS Homepage menu item
    //
    help_menu.add(MMSToolkit.getURLHelpMenu("UMLS Homepage",
        "http://www.nlm.nih.gov/research/umls/", this));
    //
    // Docs Home Page Menu Item
    //
    help_menu.add(MMSToolkit.getURLHelpMenu("UMLS Documentation",
        "http://www.nlm.nih.gov/research/umls/umlsdoc.html", this));

    //
    // NLM Home Page Menu Item
    //
    help_menu.add(MMSToolkit.getURLHelpMenu("NLM Homepage",
        "http://www.nlm.nih.gov", this));
  }

  /**
   * Enable glass pane.
   */
  public void enableGlassPane() {
    gc.setVisible(true);
  }

  /**
   * Disable glass pane.
   */
  public void disableGlassPane() {
    gc.setVisible(false);
  }

  /**
   * Close this window and re-enable the {@link WelcomeGUI}.
   */
  public void dispose() {
    WelcomeGUI.getWelcomeGUI().disableGlassPane();
    super.dispose();
  }

  /**
   * Validates the user selections for directories.
   * 
   * @return <code>true</code> if validation passes <code>false</code>
   *         otherwise
   */
  private boolean validateChoices() {
    final InstallConfiguration ic = InstallConfiguration
        .getInstallConfiguration();

    //
    // Validate source dir
    //
    enableGlassPane();

    MetamorphoSysInputStream is = ic.getMetamorphoSysInputStream();
    String[] source_paths = new String[1];
    boolean test = false;
    try {
      source_paths[0] = ic.getUmlsDirectory().getCanonicalPath();
      test = !is.validateSourcePaths(source_paths);
    } catch (IOException ioe) {
      test = true;
    }
    if (test) {
      MMSToolkit.reportError(SubsetGUI.this, "Invalid source path:" + "\n\n"
          + source_paths[0] + "\n\nMust be a directory containing"
          + "\na Rich Release Format subset,"
          + "\nor an Original Release Format subset."
          + "\nPlease choose another source path.", false);
      disableGlassPane();
      return false;
    }

    //
    // Validate destination dir
    //

    if (ic.getMetaDestinationDirectory() == null) {
      MMSToolkit.notifyUser(WelcomeGUI.getWelcomeGUI(),
          "Please choose a destination directory.");
      disableGlassPane();
      return false;
    } else if (!ic.getMetaDestinationDirectory().exists()) {

      //
      // Create destination dir if not exists
      //
      boolean confirm = MMSToolkit.confirmRequest(WelcomeGUI.getWelcomeGUI(),
          "The destination directory does not exist.\n"
              + "Would you like to create it?");
      if (confirm) {
        ic.getMetaDestinationDirectory().mkdirs();
      } else {
        MMSToolkit.notifyUser(WelcomeGUI.getWelcomeGUI(),
            "Please select a destination directory that exists.");
        disableGlassPane();
        return false;
      }
    } else if (!ic.getMetaDestinationDirectory().canWrite()) {
      MMSToolkit.notifyUser(WelcomeGUI.getWelcomeGUI(),
          "The destination directory is write protected.");
      disableGlassPane();
      return false;
    }

    return true;
  }

} // End of SubsetGUI class
