/**************************************************************************
 *
 * Package:	gov.nih.nlm.mms
 * Object:	UserConfiguration.java
 *
 **************************************************************************/
package gov.nih.nlm.mms;

import gov.nih.nlm.mms.filters.ConceptContentSelector;
import gov.nih.nlm.mms.filters.ConceptManipulator;
import gov.nih.nlm.mms.filters.ConceptSelector;
import gov.nih.nlm.mms.filters.Filter;
import gov.nih.nlm.mms.filters.PrecedenceFilter;
import gov.nih.nlm.mms.filters.SourcesToRemoveFilter;
import gov.nih.nlm.mms.filters.SuppressibleTermgroupsFilter;
import gov.nih.nlm.umls.meta.SourceTermType;
import gov.nih.nlm.util.FieldedStringTokenizer;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

/**
 * Represents a set of user configurations, including all of the relevant filter
 * configuration data.
 * 
 * @author Deborah Shapiro, Brian Carlsen
 */
public class UserConfiguration {

  //
  // Lists of active and valid filters
  //
  private List active_filters = new ArrayList();

  private List valid_filters = new ArrayList();

  //
  // GUIConfigurable-specific properties
  //
  private Map gc_properties = new HashMap();

  //
  // Some boolean settings
  //
  private boolean automatic_selection = false, properties_changed = false;

  //
  // Current config file name and subset directory name
  //
  private String current_config_file, subset_dir;

  //
  // List of input paths
  //
  private ArrayList source_paths = new ArrayList();

  //
  // User defined Precedence List as indicated by user
  // in Precedence Filter.
  //
  private List precedence = new ArrayList();
  
  //
  // User defined list of suppressed Source/Term Types 
  // as indicated in SuppressibleTermgroupsFilter
  //
  private List user_suppressed_termgroups = new ArrayList();
  
  //
  // Various data for messages to user
  //
  private String release_version, release_date, release_description;

  //
  // Input and output stream references
  //
  private MetamorphoSysInputStream mmsys_input_stream;

  private MetamorphoSysOutputStream mmsys_output_stream;

  //
  // Information about user selected "default subset" configuration
  //
  protected List default_subsets = new ArrayList();

  protected DefaultSubset selected_default_subset;

  //
  // Information tracking warnings when loading the user config properties
  //
  private boolean load_warnings = false;

  private StringBuffer load_warnings_report = new StringBuffer();

  //
  // Constructor
  //

  /**
   * Instantiates a {@link UserConfiguration} from the specified properties
   * {@link File}.
   * 
   * @param config_file
   *          a valid Java properties {@link File}
   * @throws LocalException
   *           if there is a problem loading the properties file.
   */
  public UserConfiguration(File config_file) throws LocalException {
    String path = null;
    try {
      path = config_file.getCanonicalPath();
    } catch (IOException e) {
      throw new LocalException(
          "The following properties file could not be found\n\n"
              + config_file
              + "\n\n"
              + "Try opening a different properties file.\n"
              + "Verify existance of configuration directory if error persists.");
    }

    // Load should be overridden by subclasses for this to work
    load(path);

  }

  /**
   * Instantiates a {@link UserConfiguration} from the specified
   * {@link Properties}.
   * 
   * @param p
   *          {@link Properties}
   * @throws LocalException
   *           if there is a problem loading the properties.
   */
  public UserConfiguration(Properties p) throws LocalException {
    // Load should be overridden by subclasses for this to work
    loadProperties(p);
  }

  //
  // Accessor Methods
  //
  /**
   * Returns a {@link List} of all {@link SourceTermType} objects in the 
   * precedence order.
   * 
   * @return a {@link List} of all {@link SourceTermType} objects.
   */
  public List getPrecedence() {
    return new ArrayList(precedence);
  }
  
  /**
   * Sets the {@link List} of all {@link SourceTermType} objects in the
   * precedence order.
   * @param precedence {@link List}
   */
  public void setPrecedence(List precedence) {
  	this.precedence = precedence;
  }
  
  /**
   * Returns a {@link List} of user suppressed {@link SourceTermType} objects.
   * 
   * @return a {@link List} of all user suppressed {@link SourceTermType} objects.
   */
  public List getUserSuppressedSourceTermTypes() {
    return new ArrayList(user_suppressed_termgroups);
  } 
  
  /**
   * Sets the {@link List} of user suppressed {@link SourceTermType} objects.
   * @param sup_tgps {@link List}
   */
  public void setUserSuppressedSourceTermTypes(List sup_tgps) {
  	this.user_suppressed_termgroups = sup_tgps;
  } 
  
  /**
   * Indicates if given {@link SourceTermType} is suppressed by the user.
   * @param sab_tty {@link SourceTermType} in question
   * @return <code>true</code> if {@link SourceTermType} is suppressed
   * <code>false</code> otherwise
   */
  public boolean isSourceTermTypeSuppressed(SourceTermType sab_tty) {
  	return user_suppressed_termgroups.contains(sab_tty);
  }
  
  /**
   * Returns paths to the the installation sources.
   * 
   * @return paths to the the installation sources.
   */
  public String[] getSourcePaths() {
    return (String[]) source_paths.toArray(new String[0]);
  }

  /**
   * Sets the paths to the installation sources.
   * 
   * @param ip
   *          the paths to the installation sources
   */
  public void setSourcePaths(String[] ip) {
    source_paths.clear();
    for (int i = 0; i < ip.length; i++)
      addSourcePath(ip[i]);
    setPropertiesChanged();
  }

  /**
   * Adds a path to the list of installation sources.
   * 
   * @param ip
   *          source path to be added
   */
  public void addSourcePath(String ip) {
    String file = ip;
    try {
      file = new File(file).getCanonicalPath();
    } catch (Exception ex) {
      // do nothing
    }
    source_paths.add(file);
  }

  /**
   * Removes a path from the list of installation sources.
   * 
   * @param ip
   *          source path to be removed
   */
  public void removeSourcePath(String ip) {
    source_paths.remove(ip);
  }

  /**
   * Clears the list of installation sources.
   */
  public void clearSourcePaths() {
    source_paths.clear();
  }

  /**
   * Returns path to the the subset folder.
   * 
   * @return path to the the subset folder.
   */
  public String getSubsetDirectory() {
    return subset_dir;
  }

  /**
   * Sets the path to the subset folder.
   * 
   * @param sd
   *          the path to the subset folder
   */
  public void setSubsetDirectory(String sd) {
    subset_dir = new File(sd).getAbsolutePath();
    setPropertiesChanged();
  }

  /**
   * Returns the {@link ConceptInputStream} selected by the user.
   * 
   * @return the {@link ConceptInputStream} selected by the user
   */
  public MetamorphoSysInputStream getInputStream() {
    return mmsys_input_stream;
  }

  /**
   * Sets the class name of the {@link ConceptInputStream} to be used during
   * subsetting and instantiates it.
   * 
   * @param input_stream
   *          input stream
   */
  public void setInputStream(MetamorphoSysInputStream input_stream) {
    mmsys_input_stream = input_stream;
    mmsys_input_stream.setConfiguration(this);
  }

  /**
   * Returns the {@link ConceptOutputStream} selected by the user.
   * 
   * @return the {@link ConceptOutputStream} selected by the user
   */
  public MetamorphoSysOutputStream getOutputStream() {
    return mmsys_output_stream;
  }

  /**
   * Sets the {@link ConceptOutputStream} to be used during subsetting and
   * instantiates it.
   * 
   * @param output_stream
   *          output stream
   */
  public void setOutputStream(MetamorphoSysOutputStream output_stream) {
    mmsys_output_stream = output_stream;
    mmsys_output_stream.setConfiguration(this);
  }

  /**
   * Returns the path to the current configuration file.
   * 
   * @return the path to the current configuration file.
   */
  public String getCurrentConfigurationFilename() {
    return current_config_file;
  }

  /**
   * Sets the path to the current configuration file.
   * 
   * @param file
   *          the {@link String} path to the current configuration file
   */
  public void setCurrentConfigFile(String file) {
    current_config_file = file;
    try {
      current_config_file = new File(current_config_file).getAbsolutePath();
    } catch (Exception ex) {
      // do nothing
    }
  }

  /**
   * Returns a flag indicating whether or not "auto selection mode" is on.
   * 
   * @return <code>true</code> if automatic selection mode is enabled,
   *         <code>false</code> otherwise.
   */
  public boolean getAutomaticSelection() {
    return automatic_selection;
  }

  /**
   * Sets the "automatic selection mode" flag.
   * 
   * @param b
   *          the "automatic selection mode" flag
   */
  public void setAutomaticSelection(boolean b) {
    setPropertiesChanged();
    automatic_selection = b;
  }

  /**
   * Returns the {@link List} of active {@link ConceptSelector}s. These are
   * {@link ConceptSelector}s that have been loaded or imported <i>and</i>
   * enabled.
   * 
   * @return the {@link List} of active {@link ConceptSelector}s
   */
  public List getConceptSelectors() {
    List concept_selectors = new ArrayList();
    for (int i = 0; i < active_filters.size(); i++) {
      if (active_filters.get(i) instanceof ConceptSelector)
        concept_selectors.add(active_filters.get(i));
    }
    // if (getOutputStream() instanceof ConceptSelector)
    // concept_selectors.add(getOutputStream());
    return concept_selectors;
  }

  /**
   * Returns the {@link List} of active {@link ConceptContentSelector}s. These
   * are {@link ConceptContentSelector}s that have been loaded or imported
   * <i>and</i> enabled.
   * 
   * @return the {@link List} of active {@link ConceptContentSelector}s
   */
  public List getConceptContentSelectors() {
    List concept_content_selectors = new ArrayList();
    for (int i = 0; i < active_filters.size(); i++) {
      if (active_filters.get(i) instanceof ConceptContentSelector)
        concept_content_selectors.add(active_filters.get(i));
    }
    // if (getOutputStream() instanceof ConceptContentSelector)
    // concept_content_selectors.add(getOutputStream());
    return concept_content_selectors;
  }

  /**
   * Returns the {@link List} of active {@link ConceptManipulator}s. These are
   * {@link ConceptManipulator}s that have been loaded or imported <i>and</i>
   * enabled.
   * 
   * @return the {@link List} of active {@link ConceptManipulator}s
   */
  public List getConceptManipulators() {
    List concept_manipulators = new ArrayList();
    for (int i = 0; i < active_filters.size(); i++) {
      if (active_filters.get(i) instanceof ConceptManipulator)
        concept_manipulators.add(active_filters.get(i));
    }
    // if (getOutputStream() instanceof ConceptManipulator)
    // concept_manipulators.add(getOutputStream());
    return concept_manipulators;
  }
  
  /**
   * Returns the {@link List} of old style filters.  These are not
   * instances of {@link ConceptSelector}, {@link ConceptManipulator},
   * or {@link ConceptContentSelector}.
   * @return the {@link List} of active old style filters
   */
  public List getOldStyleFilters() {
    List old_style_filters = new ArrayList();
    for (int i=0; i<active_filters.size(); i++) {
      if (!(active_filters.get(i) instanceof ConceptSelector) &&
      		!(active_filters.get(i) instanceof ConceptContentSelector) &&
      		!(active_filters.get(i) instanceof ConceptManipulator))
        old_style_filters.add(active_filters.get(i));
    }
    return old_style_filters;
  }
  
  /**
   * Returns the {@link List} of active {@link Filter}s.
   * These are filters that have been loaded or imported <i>and</i> enabled.
   * @return the {@link List} of active {@link Filter}s
   */
  public List getActiveFilters() {
    List filters = new ArrayList(active_filters);
    // filters.add(getOutputStream());
    return filters;
  }

  /**
   * Moves the active filter at initial_index to insert it before final_index.
   * 
   * @param initial_index
   *          position of filter initially
   * @param final_index
   *          position of filter after reorder
   */
  public void reorderActiveFilters(int initial_index, int final_index) {
    Filter saved_filter = (Filter) active_filters.get(initial_index);
    active_filters.remove(initial_index);
    active_filters.add(final_index, saved_filter);
  }

  /**
   * Returns the {@link List} of active {@link Filter}s without the
   * {@link MetamorphoSysOutputStream}. These are filters that have been loaded
   * or imported <i>and</i> enabled.
   * 
   * @return the {@link List} of active {@link Filter}s
   * 
   * public List getActiveFiltersOnly() { List filters = new
   * ArrayList(active_filters); return filters; }
   */

  /**
   * Returns the {@link List} of valid {@link Filter}s. These are filters that
   * have been loaded but not enabled.
   * 
   * @return the {@link List} of valid {@link Filter}s
   */
  public List getValidFilters() {
    List filters = new ArrayList(valid_filters);
    // filters.add(getOutputStream());
    return filters;
  }

  /**
   * Returns the {@link List} of valid {@link Filter}s without the
   * {@link MetamorphoSysOutputStream}. These are filters that have been loaded
   * but not enabled.
   * 
   * @return the {@link List} of valid {@link Filter}s
   * 
   * public List getValidFiltersOnly() { List filters = new
   * ArrayList(valid_filters); return filters; }
   */

  /**
   * Returns <code>true</code> if output stream should write versioned format.
   * <code>false</code> if output stream should write versionless format.
   * 
   * @return <code>true</code> if output stream should write versioned format
   *         <code>false</code> if output stream should write versionless
   *         format
   */
  public boolean isOutputVersioned() {

    return false;
  }

  /**
   * Return the {@link Properties} corresponding to the specified
   * {@link Configurable}.
   * 
   * @param f
   *          the {@link Configurable}
   * @return the {@link Properties} corresponding to the specified
   *         {@link Configurable}
   */
  public Properties getProperties(Configurable f) {
    Properties l_prop = new Properties();
    Iterator iter = gc_properties.keySet().iterator();
    while (iter.hasNext()) {
      String prop = (String) iter.next();
      if (prop.startsWith(f.getClass().getName() + ".")) {
        l_prop.setProperty(prop.substring(prop.lastIndexOf(".") + 1),
            (String) gc_properties.get(prop));
      }
    }
    return l_prop;
  }

  /**
   * Set the {@link Properties} corresponding to the specified
   * {@link Configurable}.
   * 
   * @param f
   *          {@link Configurable}
   */
  public void setProperties(Configurable f) {
    clearProperties(f);
    Properties f_props = f.getProperties();
    Enumeration e = f_props.propertyNames();
    while (e.hasMoreElements()) {
      String key = (String) e.nextElement();
      String prop = f_props.getProperty(key);
      addProperty(f.getClass().getName() + "." + key, prop);
    }
    setPropertiesChanged();
  }

  /**
   * Returns the {@link GUIConfigurable}-specific property value.
   * 
   * @param f
   *          a {@link GUIConfigurable}
   * @param key
   *          a property name specific to the specified {@link GUIConfigurable}
   * @return the {@link GUIConfigurable}-specific property value
   */
  public String getProperty(GUIConfigurable f, String key) {
    Properties this_filter = getProperties(f);
    if (this_filter.getProperty(key) == null)
      return null;
    else
      return this_filter.getProperty(key);
  }

  /**
   * Removes the {@link GUIConfigurable}-specific property value.
   * 
   * @param f
   *          a {@link GUIConfigurable}
   * @param key
   *          a property name specific to the specified {@link GUIConfigurable}
   */
  public void removeProperty(GUIConfigurable f, String key) {
    gc_properties.remove(f.getClass().getName() + "." + key);
  }

  /**
   * Clears the properties associated with a particular {@link Configurable}.
   * 
   * @param f
   *          the {@link Configurable}
   */
  public void clearProperties(Configurable f) {
    Set keyset = new HashSet(gc_properties.keySet());
    Iterator iter = keyset.iterator();
    while (iter.hasNext()) {
      String prop = (String) iter.next();
      if (prop.startsWith(f.getClass().getName() + "."))
        gc_properties.remove(prop);
    }
  }

  /**
   * Enables the specified valid filter. This involves adding it to the active
   * filter list.
   * 
   * @param f
   *          the {@link Filter} to enable
   */
  public void enableFilter(Filter f) {
    if (valid_filters.contains(f) && !active_filters.contains(f))
      active_filters.add(f);
  }

  /**
   * Imports and enables the specified filter. This involves adding it to the
   * valid filter list.
   * 
   * @param f
   *          the {@link Filter} to import
   */
  public void importFilter(Filter f) {
    valid_filters.add(f);
    enableFilter(f);
  }

  /**
   * Clears the active filter list.
   */
  public void clearFilters() {
    active_filters.clear();
  }

  /**
   * Returns a flag indicating whether or not the user has changed the
   * underlying configuration since it was last loaded or saved.
   * 
   * @return <code>true</code> if changed have been made, <code>false</code>
   *         otherwise
   */
  public boolean hasDataChanged() {
    for (int i = 0; i < active_filters.size(); i++) {
      Filter f = (Filter) active_filters.get(i);
      if (f.hasDataChanged()) {
        return true;
      }
    }
    if (properties_changed) {
      return true;
    }
    return false;
  }

  /**
   * Resets the "has data changed" flag to false.
   */
  public void resetDataChanged() {
    for (int i = 0; i < active_filters.size(); i++) {
      Filter f = (Filter) active_filters.get(i);
      f.resetDataChanged();
    }
    properties_changed = false;
  }

  /**
   * Sets the flag indicating that the non-filter properties have changed.
   */
  public void setPropertiesChanged() {
    properties_changed = true;
  }

  /**
   * Returns a {@link String} indicating the release version corresponding to
   * the configuration file being used.
   * 
   * @return a {@link String} indicating the release version corresponding to
   *         the configuration file being used.
   */
  public String getReleaseVersion() {
    return release_version;
  }

  /**
   * Returns a {@link String} indicating the release date corresponding to the
   * configuration file being used.
   * 
   * @return a {@link String} indicating the release date corresponding to the
   *         configuration file being used.
   */
  public String getReleaseDate() {
    return release_date;
  }

  /**
   * Returns a {@link String} indicating the release description corresponding
   * to the configuration file being used.
   * 
   * @return a {@link String} indicating the release description corresponding
   *         to the configuration file being used.
   */
  public String getReleaseDescription() {
    return release_description;
  }

  /**
   * Returns a report indicating all warnings that occurred while loading the
   * user configuration file.
   * 
   * @return {@link String} report of warnings
   */
  public String getLoadWarningsReport() {
    return load_warnings_report.toString();
  }

  /**
   * Indicates if there were warnings while loading the user configuration file.
   * 
   * @return <code>boolean</code> true if warnings occurred false otherwise
   */
  public boolean didLoadCauseWarnings() {
    return load_warnings;
  }

  //
  // Protected Methods
  //

  /**
   * Loads {@link Properties} from the specified file. This method should be
   * overridden by subclasses. It opens the named properties file, configures
   * the class and returns the {@link Properties} object intact.
   * 
   * @param file
   *          a {@link String} filename
   * @return {@link Properties}
   * @throws LocalException
   *           if there is a problem loading the properties
   */
  protected Properties load(String file) throws LocalException {
    // Open and load the properties file
    Properties props = new Properties();
    try {
      FileInputStream propInputStream = new FileInputStream(file);
      props.load(propInputStream);
    } catch (IOException ioe) {
      ioe.printStackTrace();
      throw new LocalException("Error while loading properties file\n"
          + "Please try opening another properties file");
    }
    return loadProperties(props);
  }

  /**
   * Configures the class based on the {@link Properties}.
   * 
   * @param props
   *          user-level {@link Properties}
   * @return {@link Properties}
   * @throws LocalException
   *           if there is a problem loading the properties
   */
  protected Properties loadProperties(Properties props) throws LocalException {

    // Load the source paths
    if (props.containsKey("source_paths")) {
      String source_path_list = props.getProperty("source_paths");
      setSourcePaths(FieldedStringTokenizer.split(source_path_list, ";"));
    } else {
      setSourcePaths(new String[] {});
    }

    // Load the subset directory name
    if (props.containsKey("subsetname")) {
      setSubsetDirectory(props.getProperty("subsetname"));
      subset_dir = new File(getSubsetDirectory()).getAbsolutePath();
    } else {
      setSubsetDirectory("");
    }

    // Load and instantiate the ConceptInputStream handler
    if (props.containsKey("mmsys_input_stream")) {
      setInputStream((MetamorphoSysInputStream) MMSToolkit
          .instantiateObject(props.getProperty("mmsys_input_stream")));
    } else {
      load_warnings = true;
      setInputStream(new RichMRMetamorphoSysInputStream());
      load_warnings_report.append("Properties file does not indicate which\n"
          + "input handler should be used.\n"
          + "Rich Release Format will be used unless you \n"
          + "indicate otherwise on the Input Options tab.\n\n");
    }

    // Load and instantiate the ConceptOutputStream handler
    if (props.containsKey("mmsys_output_stream")) {
      setOutputStream((MetamorphoSysOutputStream) MMSToolkit
          .instantiateObject(props.getProperty("mmsys_output_stream")));
    } else {
      load_warnings = true;
      try {
        setOutputStream(new RichMRMetamorphoSysOutputStream());
      } catch (Exception e) {
        // do nothing
      }
      load_warnings_report.append("Properties file does not indicate which\n"
          + "output handler should be used.\n"
          + "Rich Release Format will be used unless you \n"
          + "indicate otherwise on the Output Options tab.\n\n");
    }

    // Set the path to the current config file
    if (props.containsKey("current_config_file")) {
      setCurrentConfigFile(props.getProperty("current_config_file"));
    }

    // Set the auto-selection option
    if (props.containsKey("automatic_selection")) {
      setAutomaticSelection(Boolean.valueOf(
          props.getProperty("automatic_selection")).booleanValue());
    } else {
      load_warnings = true;
      setAutomaticSelection(false);
      load_warnings_report
          .append("Properties file does not indicate if related\n"
              + "items should be automatically selected.\n"
              + "Related items will not be automatically selected\n"
              + "unless you indicate otherwise on the Advanced Options menu item.\n\n");
    }

    // Load and instantiate valid filters
    if (props.containsKey("valid_filters")) {
      String valid_filters_prop = props.getProperty("valid_filters");
      String[] lines = FieldedStringTokenizer.split(valid_filters_prop, ";");
      boolean invalid_filter = false;
      for (int i = 0; i < lines.length; i++) {
        try {
          valid_filters.add(Class.forName(lines[i]).newInstance());
        } catch (Exception e) {
          load_warnings = true;
          invalid_filter = true;
          load_warnings_report.append(lines[i] + " is no longer a \n"
              + "valid filter and will not be available to the user.\n");
        }
      }
      if (invalid_filter)
        load_warnings_report.append("\n");
    }
    if (valid_filters.size() == 0) {
      load_warnings = true;
      valid_filters.add(new SourcesToRemoveFilter());
      valid_filters.add(new PrecedenceFilter());
      valid_filters.add(new SuppressibleTermgroupsFilter());
      load_warnings_report
          .append("Properties file does not have a valid filters property.\n"
              + "The default valid filters list will be used instead.\n\n");
    }

    // Put valid filters on the active filters list
    if (props.containsKey("active_filters")) {
      String active_filters_prop = (String) props.get("active_filters");
      String[] tokens = FieldedStringTokenizer.split(active_filters_prop, ";");
      int k = 0;
      boolean invalid_filter = false;
      for (int i = 0; i < tokens.length; i++) {
        boolean active_filter_found = false;
        for (int j = 0; j < valid_filters.size(); j++) {
          if (tokens[i].equals(valid_filters.get(j).getClass().getName())) {
            active_filters.add(k++, valid_filters.get(j));
            active_filter_found = true;
          }
        }
        if (!active_filter_found) {
          load_warnings = true;
          invalid_filter = true;
          load_warnings_report
              .append(tokens[i]
                  + " is no longer a \n"
                  + "valid active filter and will not be used while making the subset.\n");
        }
      }
      if (invalid_filter)
        load_warnings_report.append("\n");
    }
    if (active_filters.size() < 3) {
      load_warnings = true;
      active_filters.add(new SourcesToRemoveFilter());
      active_filters.add(new PrecedenceFilter());
      active_filters.add(new SuppressibleTermgroupsFilter());
      load_warnings_report
          .append("Properties file does not have an active filters property.\n"
              + "The default active filters list will be used instead.\n\n");
    }

    // Add all properties with a '.' in the key
    // to gc_properties Set.
    // These are the filter-specific properties
    Enumeration e = props.propertyNames();
    while (e.hasMoreElements()) {
      String key = (String) e.nextElement();
      if (key.indexOf(".") != -1) {
        String prop = props.getProperty(key);
        gc_properties.put(key, prop);
      }
    }

    // The release version of the current config file
    if (props.containsKey("release_version")) {
      release_version = (String) props.get("release_version");
    } else {
      load_warnings = true;
      release_version = "unknown";
      load_warnings_report
          .append("Properties file does not indicate the release version.\n"
              + "The release version will be set to \"unknown\".\n\n");
    }

    // The release date of the current config file
    if (props.containsKey("release_date")) {
      release_date = (String) props.get("release_date");
    } else {
      load_warnings = true;
      load_warnings_report
          .append("Properties file does not indicate the release date.\n"
              + "The release date will be blank.\n\n");
    }

    // The release description of the current config file
    if (props.containsKey("release_description")) {
      release_description = (String) props.get("release_description");
    } else {
      load_warnings = true;
      load_warnings_report
          .append("Properties file does not indicate the release description.\n"
              + "The release description will be blank.\n\n");
    }

    // Valid default subsets
    if (props.containsKey("default_subsets")) {
      String default_subsets_prop = props.getProperty("default_subsets");
      String[] default_subset_ids = FieldedStringTokenizer.split(
          default_subsets_prop, ";");
      for (int i = 0; i < default_subset_ids.length; i++) {
        DefaultSubset ds = new DefaultSubset();
        ds.setId(default_subset_ids[i]);
        ds.setConfigurationFilename("mmsys." + default_subset_ids[i] + ".prop");
        String key = "default_subset_name_" + default_subset_ids[i];
        if (props.containsKey(key)) {
          ds.setName(props.getProperty(key));
        }
        key = "default_subset_description_" + default_subset_ids[i];
        if (props.containsKey(key)) {
          ds.setDescription(props.getProperty(key));
        }
        default_subsets.add(ds);
      }
    } else {
      load_warnings = true;
      String[] default_subset_ids = { "a", "b", "c" };
      for (int i = 0; i < default_subset_ids.length; i++) {
        DefaultSubset ds = new DefaultSubset();
        ds.setId(default_subset_ids[i]);
        ds.setConfigurationFilename("mmsys." + default_subset_ids[i] + ".prop");
        default_subsets.add(ds);
      }
      load_warnings_report
          .append("Properties file does not indicate the default subsets.\n"
              + "The default set of subsets will be used.\n\n");
    }

    if (props.containsKey("default_subset")) {
      String default_subset_id = props.getProperty("default_subset");
      DefaultSubset[] dss = getDefaultSubsets();
      for (int i = 0; i < dss.length; i++) {
        if (dss[i].getId().equals(default_subset_id)) {
          setSelectedDefaultSubset(dss[i]);
        }
      }
    } else {
      load_warnings = true;
      setSelectedDefaultSubset((DefaultSubset) default_subsets.get(0));
      load_warnings_report
          .append("Properties file does not indicate the selected default subset.\n"
              + "The Level 0 default subset will be used instead.\n\n");
    }

    return props;
  }

  /**
   * Saves the configuration properties to the specified {@link File}.
   * 
   * @param file_path
   *          the {@link File}
   * @throws LocalException
   *           if there is an error while saving the configuration
   */
  public void saveState(File file_path) throws LocalException {

    if (!file_path.canWrite() && file_path.exists()) {
      throw new LocalException("The file " + file_path.getClass().getName()
          + " is write protected.  Configuration not saved.");
    }

    // Save user configuration to a properties object
    Properties properties = new Properties();

    if (current_config_file != null)
      properties.setProperty("current_config_file", current_config_file);

    StringBuffer source_paths_line = new StringBuffer();
    String[] source_paths = getSourcePaths();
    for (int i = 0; i < source_paths.length; i++) {
      source_paths_line.append(source_paths[i]);
      if (i < source_paths.length - 1)
        source_paths_line.append(";");
    }
    properties.setProperty("source_paths", source_paths_line.toString());

    properties.setProperty("subsetname", subset_dir);

    // go through all active filters and write out all filter
    // specific properties
    // update gc_properties HashSet

    for (int i = 0; i < getValidFilters().size(); i++) {
      // should skip filter if filter hasn't changed
      Filter f = (Filter) getValidFilters().get(i);
      Properties props = f.getProperties();
      clearProperties(f);
      Enumeration e = props.propertyNames();
      while (e.hasMoreElements()) {
        String key = (String) e.nextElement();
        String prop = props.getProperty(key);
        addProperty(f.getClass().getName() + "." + key, prop);
        properties.setProperty(f.getClass().getName() + "." + key, prop);
      }
    }

    GUIConfigurable gc = getInputStream();
    Properties props = gc.getProperties();
    clearProperties(gc);
    Enumeration e = props.propertyNames();
    while (e.hasMoreElements()) {
      String key = (String) e.nextElement();
      String prop = props.getProperty(key);
      addProperty(gc.getClass().getName() + "." + key, prop);
      properties.setProperty(gc.getClass().getName() + "." + key, prop);
    }

    gc = getOutputStream();
    props = gc.getProperties();
    clearProperties(gc);
    e = props.propertyNames();
    while (e.hasMoreElements()) {
      String key = (String) e.nextElement();
      String prop = props.getProperty(key);
      addProperty(gc.getClass().getName() + "." + key, prop);
      properties.setProperty(gc.getClass().getName() + "." + key, prop);
    }

    StringBuffer active_filters_line = new StringBuffer();
    List af = getActiveFilters();
    for (int i = 0; i < af.size(); i++) {
      Filter f = (Filter) af.get(i);
      active_filters_line.append(f.getClass().getName());
      if (i < af.size() - 1)
        active_filters_line.append(";");

    }
    properties.setProperty("active_filters", active_filters_line.toString());

    StringBuffer valid_filters_line = new StringBuffer();
    List vf = getValidFilters();
    for (int i = 0; i < vf.size(); i++) {
      Filter f = (Filter) vf.get(i);
      valid_filters_line.append(f.getClass().getName());
      if (i < vf.size() - 1)
        valid_filters_line.append(";");
    }
    properties.setProperty("valid_filters", valid_filters_line.toString());

    String id = getSelectedDefaultSubset().getId();
    properties
        .setProperty("default_subset", getSelectedDefaultSubset().getId());
    properties.setProperty("default_subsets", id);
    properties.setProperty(("default_subset_name_" + id),
        getSelectedDefaultSubset().getName());
    properties.setProperty(("default_subset_description_" + id),
        getSelectedDefaultSubset().getDescription());

    // write automatic selection property
    properties.setProperty("automatic_selection", new Boolean(
        automatic_selection).toString());
    properties.setProperty("mmsys_input_stream", getInputStream().getClass()
        .getName());
    properties.setProperty("mmsys_output_stream", getOutputStream().getClass()
        .getName());

    // write release information
    properties.setProperty("release_date", release_date);
    properties.setProperty("release_description", release_description);
    properties.setProperty("release_version", release_version);

    // Save properties file
    try {
      FileOutputStream out;
      out = new FileOutputStream(file_path);
      properties.store(out, " Configuration Properties File ");
      out.close();

    } catch (IOException e1) {
      throw new LocalException(e1.getMessage());
    }
  }

  /**
   * Sets the selected {@link DefaultSubset}.
   * 
   * @param ds
   *          the selected {@link DefaultSubset}
   */
  public void setSelectedDefaultSubset(DefaultSubset ds) {
    selected_default_subset = ds;
  }

  /**
   * Returns the selected {@link DefaultSubset}.
   * 
   * @return the selected {@link DefaultSubset}
   */
  public DefaultSubset getSelectedDefaultSubset() {
    return selected_default_subset;
  }

  /**
   * Returns all available {@link DefaultSubset}s.
   * 
   * @return all available {@link DefaultSubset}s
   */
  public DefaultSubset[] getDefaultSubsets() {
    return (DefaultSubset[]) default_subsets.toArray(new DefaultSubset[0]);
  }

  /**
   * Returns all available {@link DefaultSubset}s as a {@link List}.
   * 
   * @return all available {@link DefaultSubset}s as a {@link List}
   */
  public List getDefaultSubsetsAsList() {
    return default_subsets;
  }

  //
  // Private methods
  //

  /**
   * Adds a property. This is used to set filter properties.
   * 
   * @param key
   *          the {@link String} key
   * @param value
   *          the {@link String} value
   */
  public void addProperty(String key, String value) {
    gc_properties.put(key, value);
    setPropertiesChanged();
  }

}
