/*****************************************************************************
 * Package: gov.nih.nlm.mms.browser
 * Object:  ConceptListManager.java
 *
 ****************************************************************************/
package gov.nih.nlm.mms.browser;

import gov.nih.nlm.mms.UserConfiguration;
import gov.nih.nlm.umls.io.RichMRFilesSingleConceptInputStream;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.impl.ConceptImpl;
import gov.nih.nlm.util.ObjectXMLSerializer;

import java.io.IOException;

/**
 * Manages the concept history.
 *
 * @author Deborah Shapiro
 */
public class ConceptListManager  {

  //
  // User config
  //
  protected UserConfiguration user_configuration;

  //
  // Reports  manager
  //
  protected ReportsManager reports_manager;

  //
  // Finder manager
  //
  protected FinderManager finder_manager;

  //
  // concept history
  //
  protected ConceptList concept_list;

  //
  // Sets an object into XML
  //
  protected ObjectXMLSerializer xml_ser = new ObjectXMLSerializer();

  //
  // Constructor
  //

  /**
   * Instantiates an {@link ConceptListManager}.
   */
  public ConceptListManager() { }

  /**
   * Sets the {@link ConceptList}.
   * @param concept_list the {@link ConceptList}
   */
  public void setConceptList(ConceptList concept_list) {
    this.concept_list = concept_list;
    if (concept_list.getCurrentConcept() != null) {
      Concept full_concept = readConcept(concept_list.getCurrentConcept());
      if (full_concept != null) {
        reports_manager.setConcept(full_concept);
      }
    }
  }

  /**
   * Returns the {@link ConceptList}.
   * @return the {@link ConceptList}
   */
  public ConceptList getConceptList() {
    return concept_list;
  }

  /**
   * Adds a {@link Concept} for display in the current view.
   * Fully populates the {@link Concept} before displaying it.
   * Manages the forward and backward aparatus.
   * This must be called from a separate thread.
   * @param concept the {@link Concept} to view
   */
  public void addConcept(Concept concept)  {
    if (concept.equals(concept_list.getCurrentConcept()))
      return;

    Concept full_concept = readConcept(concept);

    if (full_concept != null) {

      Concept c = new ConceptImpl();
      c.setCui(full_concept.getCui());
      concept_list.getBackwardStack().push(concept_list.getCurrentConcept());
      concept_list.setCurrentConcept(c);
      concept_list.getForwardStack().clear();

      // set the new current concept in all of the concept views
      reports_manager.setConcept(full_concept);
    }

  }

  /**
   * Refresh the current concept view if set.
   */
  public void refresh() {
    if (concept_list.getCurrentConcept() != null) {
      Concept full_concept = readConcept(concept_list.getCurrentConcept());
      reports_manager.setConcept(full_concept);
    } 
  }

  /**
   * Reads and fully populates a {@link Concept} that otherwise has just a
   * CUI value.
   * @param concept the {@link Concept} to read (with CUI field populated)
   * @return the fully populated concept {@link Concept}
   */
  public Concept readConcept(Concept concept){
    Concept full_concept = null;
    try {
      RichMRFilesSingleConceptInputStream rrf =
          new RichMRFilesSingleConceptInputStream(concept.getCui(),
                                                  user_configuration.getSourcePaths()[0]);
      full_concept = rrf.getConcept();
    } catch (IOException ioe) {
      RuntimeException re = new RuntimeException("Failed to get concept.");
      re.initCause(ioe);
      throw re;
    }
    return full_concept;

  }

  /**
   * Moves the concept list forwards by one.
   */
  public void forward() {
    Concept c = new ConceptImpl();
    c.setCui(concept_list.getCurrentConcept().getCui());
    concept_list.getBackwardStack().push(c);
    concept_list.setCurrentConcept((Concept)(concept_list.getForwardStack().pop()));
    // set the new current concept in all of the concept views
    reports_manager.setConcept(readConcept(concept_list.getCurrentConcept()));
  }

  /**
   * Moves the concept list backwards by one.
   */
  public void backward() {
    Concept c = new ConceptImpl();
    c.setCui(concept_list.getCurrentConcept().getCui());
    concept_list.getForwardStack().push(c);
    concept_list.setCurrentConcept((Concept)(concept_list.getBackwardStack().pop()));
    // set the new current concept in all of the concept views
    reports_manager.setConcept(readConcept(concept_list.getCurrentConcept()));
  }


  /**
   * Method provides the filter with values from the user configuration
   * file.
   * @param c {@link UserConfiguration}
   */
  public void setConfiguration(UserConfiguration c) {
    user_configuration = c;
  }

  /**
   * Sets the {@link ReportsManager}.
   * @param reports_manager {@link ReportsManager}
   */
  public void setReportsManager(ReportsManager reports_manager) {
    this.reports_manager = reports_manager;
  }

  /**
   * Returns the {@link Concept} currently being viewed.
   * @return {@link Concept}
   */
  public Concept getCurrentConcept() {
    if (concept_list == null) return null;
    return concept_list.getCurrentConcept();
  }

  /**
   * Indicates if the backwards stack is empty.
   * @return <code>boolean</code>
   */
  public boolean isBackHistoryEmpty() {
    if (concept_list == null) return true;
    return concept_list.getBackwardStack().size() < 2;
  }

  /**
   * Indicates if the forwards stack is empty.
   * @return <code>boolean</code>
   */
  public boolean isForwardHistoryEmpty() {
    if (concept_list == null) return true;
    return concept_list.getForwardStack().isEmpty();
  }

  /**
   * Resets all navigation components.
   */
  public void clearConceptList() {
    concept_list.getBackwardStack().clear();
    concept_list.getForwardStack().clear();
    concept_list.setCurrentConcept(null);
  }

}
