/**************************************************************************
 *
 * Package:    gov.nih.nlm.mms.browser;
 * Class:      ExpandableConceptView.java
 *
 *************************************************************************/
package gov.nih.nlm.mms.browser;

import gov.nih.nlm.mms.MMSToolkit;
import gov.nih.nlm.mms.filters.SourcesToRemoveFilter;
import gov.nih.nlm.swing.DataChangeListener;
import gov.nih.nlm.swing.SuperJList;
import gov.nih.nlm.umls.io.RichMRFilesIndexLookup;
import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.AtomAttribute;
import gov.nih.nlm.umls.meta.AtomCoocurrence;
import gov.nih.nlm.umls.meta.AtomElement;
import gov.nih.nlm.umls.meta.AtomRelationship;
import gov.nih.nlm.umls.meta.Attribute;
import gov.nih.nlm.umls.meta.ChainRestrictor;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.Context;
import gov.nih.nlm.umls.meta.Coocurrence;
import gov.nih.nlm.umls.meta.Definition;
import gov.nih.nlm.umls.meta.Language;
import gov.nih.nlm.umls.meta.Mapping;
import gov.nih.nlm.umls.meta.Relationship;
import gov.nih.nlm.umls.meta.RelationshipAttribute;
import gov.nih.nlm.umls.meta.Restrictor;
import gov.nih.nlm.umls.meta.SemanticType;
import gov.nih.nlm.umls.meta.Source;
import gov.nih.nlm.umls.meta.SourceElement;
import gov.nih.nlm.umls.meta.SourceTermType;
import gov.nih.nlm.umls.meta.impl.ConceptImpl;
import gov.nih.nlm.util.FieldedStringTokenizer;
import gov.nih.nlm.util.FinderHandler;
import gov.nih.nlm.util.HTMLDocumentRenderer;
import gov.nih.nlm.util.OnePagePrintable;
import gov.nih.nlm.util.SystemToolkit;

import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.print.Printable;
import java.awt.print.PrinterException;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import javax.swing.*;
import javax.swing.border.CompoundBorder;
import javax.swing.border.EmptyBorder;
import javax.swing.border.EtchedBorder;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;

/**
 * {@link ConceptView} that displays the concept report with a formatted
 * appearance that is customizable by the user.
 * 
 * @author Deborah Shapiro
 * 
 */

public class ExpandableConceptView extends AbstractConceptView implements
    ActionListener {

  //
  // GUI
  //
  private JEditorPane report_pane;

  //
  // Current cui
  //
  private Concept concept;

  //
  // Source(s) to display (if restricted)
  //
  private Set display_sources = new HashSet();

  //
  // Language(s) to display (if restricted)
  //
  private Set display_languages = new HashSet();

  //
  // Restrictor
  //
  private ChainRestrictor restrictor;

  //
  // Atom Comparator
  //
  private Comparator atom_comparator = null;

  //
  // Prepare string buffer
  //
  private StringBuffer report = new StringBuffer(10000);

  //
  // Mechanism for looking up children and siblings
  //
  private RichMRFilesIndexLookup index_lookup = null;

  //
  // Indicates which nodes have already been expanded
  //
  private Map expanded_hm = null;

  //
  // plus and minus icons for table
  //
  private URL plus_url = null, minus_url = null;

  //
  // data structures to hold concept components
  //
  private Atom[] sorted_atoms;

  private SemanticType[] stys;

  private Definition[] defs;

  private Context[] contexts;

  private Coocurrence[] cocs;

  private Relationship[] rels;

  private Attribute[] atts;

  private Mapping[] maps;

  //
  // Maps rel names to lists of rels
  //
  private Map rel_name_to_rels = new HashMap();

  //
  // precedence ranking of source termgroups
  //
  private List precedence;

  //
  // string to cui map used for hyperlinks
  //
  private Map str_to_cui = new HashMap();

  //
  // array of atom comparator choices
  //
  private Comparator[] comparators;

  private Rectangle rect;

  //
  // Prepares separators for atom reports
  // and link parameters for HTML reports
  //
  private String bold_begin = null;

  private String bold_end = null;

  final private String line_end = "<br>" + System.getProperty("line.separator");

  /**
   * Options Menu
   */

  private int page_size = 100;

  /**
   * Instantiates a {@link ExpandableConceptView}.
   */
  public ExpandableConceptView() {
    restrictor = new ChainRestrictor();
    restrictor.setComparator(new SourceElement.BySourceComparator());
    expanded_hm = new HashMap();
    try {
      plus_url = this.getClass().getClassLoader().getResource(
          "config/plus2.gif");
      minus_url = this.getClass().getClassLoader().getResource(
          "config/minus2.gif");
    } catch (Exception ex) {
      throw new RuntimeException("config/*.gif file could not be loaded.");
    }

    // Make the atom comparator array
    comparators = new Comparator[3];

    // dummy comparator place saver for lexical class rank order
    comparators[0] = new Comparator() {
      public int compare(Object o1, Object o2) {
        return 0;
      };

      public String toString() {
        return "Sorted by Lexical Class Rank";
      }

      public boolean equals(Object o) {
        return this.toString().equals(((Comparator) o).toString());
      }
    };
    // alphabetical sort
    comparators[1] = new Comparator() {
      public int compare(Object o1, Object o2) {
        if (o1 == null || o2 == null || !(o1 instanceof Atom)
            || !(o2 instanceof Atom))
          return 0;
        final Atom c1 = (Atom) o1;
        final Atom c2 = (Atom) o2;

        return c1.getStr().compareTo(c2.getStr());
      };

      public String toString() {
        return "Sorted by String";
      }

      public boolean equals(Object o) {
        return this.toString().equals(((Comparator) o).toString());
      }
    };
    // source, alphabetical sort
    comparators[2] = new Comparator() {
      public int compare(Object o1, Object o2) {
        if (o1 == null || o2 == null || !(o1 instanceof Atom)
            || !(o2 instanceof Atom))
          return 0;
        final Atom c1 = (Atom) o1;
        final Atom c2 = (Atom) o2;

        if (c1.getSource().compareTo(c2.getSource()) == 0)
          return c1.getStr().compareTo(c2.getStr());
        else
          return c1.getSource().compareTo(c2.getSource());
      };

      public String toString() {
        return "Sorted by Source, String";
      }

      public boolean equals(Object o) {
        return this.toString().equals(((Comparator) o).toString());
      }
    };

    // set default to "Sorted by Source, String"
    atom_comparator = comparators[2];
  }

  /**
   * Returns a {@link JPanel} containing the concept report.
   * 
   * @return a {@link JPanel} containing the concept report
   */
  public JPanel getPanel() {
    JPanel report_panel = new JPanel();
    report_panel.setLayout(new BoxLayout(report_panel, BoxLayout.Y_AXIS));
    report_pane = new JEditorPane();
    report_pane.addHyperlinkListener(new HyperlinkListener() {
      public void hyperlinkUpdate(HyperlinkEvent he) {
        rect = report_pane.getVisibleRect();
        if (he.getEventType() == HyperlinkEvent.EventType.ACTIVATED) {
          final String ui = he.getURL() == null ?
              he.getDescription() : he.getURL().toString();
          if (ui.startsWith("CUI:")) {
            RRFBrowserGUI.getRRFBrowserGUI().enableGlassPane();
            Thread t = new Thread(new Runnable() {
              public void run() {
                Concept local_concept = null;
                local_concept = new ConceptImpl();
                local_concept.setCui(ui.substring(4));
                clm.addConcept(local_concept);
                RRFBrowserGUI.getRRFBrowserGUI().disableGlassPane();
              }
            });
            t.start();
          } else if (ui.startsWith("Back")) {
            // change the range of records to be displayed
            String key = ui.substring("Back".length());
            int record = ((Integer) expanded_hm.get(key)).intValue();
            if (record > page_size)
              record = record - page_size;
            else
              record = 0;
            expanded_hm.put(key, new Integer(record));

            Thread thread = new Thread(new Runnable() {
              public void run() {
                redrawConcept(concept);
                RRFBrowserGUI.getRRFBrowserGUI().disableGlassPane();
              }
            });
            RRFBrowserGUI.getRRFBrowserGUI().enableGlassPane();
            thread.start();
          } else if (ui.startsWith("Forward")) {
            // change the range of records to be displayed
            String key = ui.substring("Forward".length());
            int record = ((Integer) expanded_hm.get(key)).intValue();
            record = record + page_size;
            expanded_hm.put(key, new Integer(record));

            Thread thread = new Thread(new Runnable() {
              public void run() {
                redrawConcept(concept);
                RRFBrowserGUI.getRRFBrowserGUI().disableGlassPane();
              }
            });
            RRFBrowserGUI.getRRFBrowserGUI().enableGlassPane();
            thread.start();

          } else if (expanded_hm.containsKey(ui)) {
            expanded_hm.remove(ui);
            Thread thread = new Thread(new Runnable() {
              public void run() {
                redrawConcept(concept);
                RRFBrowserGUI.getRRFBrowserGUI().disableGlassPane();
              }
            });
            RRFBrowserGUI.getRRFBrowserGUI().enableGlassPane();
            thread.start();

          } else if (!expanded_hm.containsKey(ui)) {
            Thread thread = new Thread(new Runnable() {
              public void run() {
                expanded_hm.put(ui, new Integer(0));
                redrawConcept(concept);
                RRFBrowserGUI.getRRFBrowserGUI().disableGlassPane();
              }
            });
            RRFBrowserGUI.getRRFBrowserGUI().enableGlassPane();
            thread.start();
          }
        }
      }
    });

    report_pane.setEditable(false);
    report_pane.setContentType(getContentType());
    JScrollPane sp = new JScrollPane(report_pane,
        ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED,
        ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
    report_panel.add(sp, BorderLayout.CENTER);

    //
    // Create the popup menu.
    //
    JPopupMenu popup = new JPopupMenu();
    JMenuItem option_menu_item = new JMenuItem("Options...");
    option_menu_item.addActionListener(this);
    popup.add(option_menu_item);

    popup.addSeparator();

    JMenuItem copy_item = new JMenuItem("Copy");
    copy_item.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        Thread thread = new Thread(new Runnable() {
          public void run() {
            try {
              report_pane.copy();
            } catch (Exception e) {
              e.printStackTrace();
              MMSToolkit.reportError(RRFBrowserGUI.getRRFBrowserGUI(),
                  "Unexpected error expanding report.\n"
                      + "See the console for more information.", false);
            }
            RRFBrowserGUI.getRRFBrowserGUI().disableGlassPane();
          }
        });
        RRFBrowserGUI.getRRFBrowserGUI().enableGlassPane();
        thread.start();
      }
    });
    popup.add(copy_item);

    popup.addSeparator();

    JMenuItem expand_tree_item = new JMenuItem("Expand Complete Tree");
    expand_tree_item.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        Thread thread = new Thread(new Runnable() {
          public void run() {
            try {
              expandAllNodes();
            } catch (Exception e) {
              e.printStackTrace();
              MMSToolkit.reportError(RRFBrowserGUI.getRRFBrowserGUI(),
                  "Unexpected error expanding report.\n"
                      + "See the console for more information.", false);
            }
            RRFBrowserGUI.getRRFBrowserGUI().disableGlassPane();
          }
        });
        RRFBrowserGUI.getRRFBrowserGUI().enableGlassPane();
        thread.start();
      }
    });
    popup.add(expand_tree_item);

    JMenuItem collapse_tree_item = new JMenuItem("Collapse Complete Tree");
    collapse_tree_item.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        Thread thread = new Thread(new Runnable() {
          public void run() {
            try {
              collapseAllNodes();
            } catch (Exception e) {
              e.printStackTrace();
              MMSToolkit.reportError(RRFBrowserGUI.getRRFBrowserGUI(),
                  "Unexpected error collapsing report.\n"
                      + "See the console for more information.", false);
            }
            RRFBrowserGUI.getRRFBrowserGUI().disableGlassPane();
          }
        });
        RRFBrowserGUI.getRRFBrowserGUI().enableGlassPane();
        thread.start();
      }
    });

    popup.add(collapse_tree_item);

    // Add listener to components that can bring up popup menus.
    PopupListener popup_listener = new PopupListener(popup);
    report_pane.addMouseListener(popup_listener);

    return report_panel;
  }

  public void actionPerformed(ActionEvent e) {
    JMenuItem source = (JMenuItem) (e.getSource());
    String s = source.getText();
    if (s.equals("Options...")) {
      showOptions();
    }
    if (concept != null) {
      RRFBrowserGUI.getRRFBrowserGUI().enableGlassPane();

      Thread t = new Thread(new Runnable() {
        public void run() {

          clm.refresh();
          RRFBrowserGUI.getRRFBrowserGUI().disableGlassPane();
        }
      });
      t.start();

    }
  }

  /**
   * Returns the content type used for the {@link JEditorPane} display.
   * 
   * @return content type
   */
  public String getContentType() {
    return "text/html";
  }

  /**
   * Indicates the {@link FinderHandler} that will execute the find operation
   * for this {@link ConceptView}
   * 
   * @return {@link FinderHandler}
   */
  public FinderHandler getFinderHandler() {
    JEditorPaneFinderHandler handler = new JEditorPaneFinderHandler(report_pane);
    handler.setDialog(RRFBrowserGUI.getFindDialog());
    return handler;
  }

  /**
   * Sets the {@link Concept} to be displayed.
   * 
   * @param concept
   *          {@link Concept}
   */
  public void setConcept(Concept concept) {
    this.concept = concept;
    expanded_hm.clear();
    rel_name_to_rels.clear();

    // add the default expanded sections
    expanded_hm.put("Definitions", null);
    expanded_hm.put("Atoms", null);
    expanded_hm.put("Semantic Types", null);

    final StringBuffer sb = new StringBuffer(10000);
    sb.append("<html>");
    sb.append(getReport(concept));
    sb.append("</html>");

    try {
      SwingUtilities.invokeAndWait(new Runnable() {
        public void run() {
          report_pane.setText(sb.toString());
          report_pane.requestFocus();
          report_pane.setCaretPosition(0);
        }
      });
    } catch (Exception e) {
      e.printStackTrace();
      MMSToolkit.reportError(RRFBrowserGUI.getRRFBrowserGUI(),
          "Unexpected error refreshing report.\n"
              + "See the console for more information.", false);
    }

  }

  /**
   * Sets the {@link Concept} to be displayed.
   * 
   * @param concept
   *          {@link Concept}
   */
  private void redrawConcept(Concept concept) {
    this.concept = concept;
    final StringBuffer sb = new StringBuffer(10000);
    sb.append("<html>");
    sb.append(getReport(concept));
    sb.append("</html>");
    report_pane.setText(sb.toString());
    try {
      SwingUtilities.invokeAndWait(new Runnable() {
        public void run() {
          report_pane.scrollRectToVisible(rect);
        }
      });
    } catch (Exception e) {
      e.printStackTrace();
      MMSToolkit.reportError(RRFBrowserGUI.getRRFBrowserGUI(),
          "Unexpected error refreshing report.\n"
              + "See the console for more information.", false);
    }
  }

  private void expandAllNodes() {
    expanded_hm.put("ALL", null);
    rect = report_pane.getVisibleRect();
    redrawConcept(concept);
    try {
      SwingUtilities.invokeAndWait(new Runnable() {
        public void run() {
          report_pane.requestFocus();
          report_pane.setCaretPosition(0);
        }
      });
    } catch (Exception e) {
      e.printStackTrace();
      MMSToolkit.reportError(RRFBrowserGUI.getRRFBrowserGUI(),
          "Unexpected error refreshing report.\n"
              + "See the console for more information.", false);
    }
  }

  private void collapseAllNodes() {
    expanded_hm.clear();
    rect = report_pane.getVisibleRect();
    redrawConcept(concept);

    try {
      SwingUtilities.invokeAndWait(new Runnable() {
        public void run() {
          report_pane.requestFocus();
          report_pane.setCaretPosition(0);
        }
      });
    } catch (Exception e) {
      e.printStackTrace();
      MMSToolkit.reportError(RRFBrowserGUI.getRRFBrowserGUI(),
          "Unexpected error refreshing report.\n"
              + "See the console for more information.", false);
    }
  }

  /**
   * Sets the {@link Restrictor} that will be used to restrict the parts of the
   * {@link Concept} that will be displayed in this {@link ConceptView}.
   * 
   * @param restrictor
   *          {@link Restrictor}
   */
  public void setRestrictor(Restrictor restrictor) {
    this.restrictor = (ChainRestrictor) restrictor;
  }

  /**
   * Sets the {@link Comparator} that will be used to sort the {@link Atom}s.
   * 
   * @param atom_comparator
   *          {@link Comparator}
   */
  public void setAtomComparator(Comparator atom_comparator) {
    this.atom_comparator = atom_comparator;
  }

  /**
   * Sets the page size. This is the number of records that will be displayed at
   * one time for nodes with a lot of data.
   * 
   * @param page_size
   *          int
   */
  public void setPageSize(int page_size) {
    this.page_size = page_size;
  }

  /**
   * Resets the configuration back to default values. When used with a GUI this
   * method should change the data model behind the GUI and allow
   * <code>setPanelSelections</code> to actually update the GUI.
   */
  public void resetToDefaults() {
    report_pane.setText("");
    expanded_hm.clear();
  }

  /**
   * Method returns all properties specific to the {@link ConceptView}.
   * 
   * @return {@link Properties} only those directly related to the concept view.
   */
  public Properties getProperties() {
    Properties new_filter_props = new Properties();

    new_filter_props
        .setProperty("page_size", new Integer(page_size).toString());

    List sources = new ArrayList(display_sources);
    StringBuffer line = new StringBuffer();
    for (int i = 0; i < sources.size(); i++) {
      line.append(((Source) sources.get(i)).getRootSourceAbbreviation());
      line.append("|");
      line.append(((Source) sources.get(i)).getSourceFamily());
      if (i <= sources.size() - 2)
        line.append(";");
    }
    new_filter_props.setProperty("display_sources", line.toString());

    List lats = new ArrayList(display_languages);
    line = new StringBuffer();
    for (int i = 0; i < lats.size(); i++) {
      line.append(((Language) lats.get(i)).getAbbreviation());
      if (i <= lats.size() - 2)
        line.append(";");
    }
    new_filter_props.setProperty("display_languages", line.toString());

    new_filter_props.setProperty("atom_comparator", atom_comparator.toString());

    return new_filter_props;
  }

  /**
   * Sets the view properties.
   * @param p {@link Properties}
   */
  public void setProperties(Properties p) {
    Properties filter_props = p;
    if (filter_props.containsKey("page_size")) {
      setPageSize(new Integer((String) filter_props.get("page_size"))
          .intValue());
    }
    if (filter_props.containsKey("display_sources")) {
      display_sources.clear();
      String remove_prop = (String) filter_props.get("display_sources");
      String[] tokens = FieldedStringTokenizer.split(remove_prop, ";");
      for (int i = 0; i < tokens.length; i++) {
        Source local_source = Source.getSource(tokens[i].substring(0, tokens[i]
            .indexOf("|")));
        if (local_source != null)
          display_sources.add(local_source);
      }
    }
    if (filter_props.containsKey("display_languages")) {
      display_languages.clear();
      String remove_prop = (String) filter_props.get("display_languages");
      String[] tokens = FieldedStringTokenizer.split(remove_prop, ";");
      for (int i = 0; i < tokens.length; i++) {
        Language local_lat = Language.getLanguage(tokens[i]);
        if (local_lat != null)
          display_languages.add(local_lat);
      }
    }
    if (filter_props.containsKey("atom_comparator")) {
      String comparator_prop = (String) filter_props.get("atom_comparator");
      for (int i = 0; i < comparators.length; i++) {
        if (comparator_prop.equals(comparators[i].toString())) {
          atom_comparator = comparators[i];
          break;
        }
      }
    }

    // make and set the restrictor
    ChainRestrictor restrictor = new ChainRestrictor();
    if (!display_sources.isEmpty()) {
      restrictor.add(new SourceElement.BySourceRestrictor(
          (Source[]) display_sources.toArray(new Source[0])));

      restrictor.add(new AtomElement.BySourceRestrictor(
          (Source[]) display_sources.toArray(new Source[0])));
    }
    if (!display_languages.isEmpty()) {
      restrictor.add(new AtomElement.ByLanguageRestrictor(
          (Language[]) display_languages.toArray(new Language[0])));

      restrictor.add(new Atom.ByLanguageRestrictor(
          (Language[]) display_languages.toArray(new Language[0])));
    }
    restrictor.setComparator(new SourceElement.BySourceComparator());
    setRestrictor(restrictor);

  }

  /**
   * Adds a {@link DataChangeListener} that responds to configuration changes.
   * Allows main GUI to record all changes to this gui in the undo/redo
   * mechanism.
   * 
   * @param dcl
   *          the {@link DataChangeListener} to add
   */
  public void addDataChangeListener(DataChangeListener dcl) {

  }

  /**
   * Removes the specified {@link DataChangeListener}.
   * 
   * @param dcl
   *          the {@link DataChangeListener} to remove
   */
  public void removeDataChangeListener(DataChangeListener dcl) {

  }

  /**
   * Returns the name.
   * 
   * @return the name
   */
  public String getName() {
    return "Report View";
  }

  /**
   * Presents an advanced options GUI to the user.
   */
  public void showOptions() {
    final ExpandableConceptViewOptionsGUI pretty_concept_view_gui = new ExpandableConceptViewOptionsGUI(
        this);
    pretty_concept_view_gui.setVisible(true);
  }

  /**
   * Indicates whether or not an options menu item should be supplied for
   * advanced configuration options.
   * 
   * @return <code>true</code> if so; <code>false</code> otherwise
   */
  public boolean hasAdvancedOptions() {
    return true;
  }

  /**
   * Prints the concept report.
   * 
   * @throws PrinterException
   */
  public void print() throws PrinterException {
    HTMLDocumentRenderer hdr = new HTMLDocumentRenderer();
    hdr.setTitle("Concept: " + concept.getCui());
    JEditorPane editor = new JEditorPane();
    editor.setContentType("text/html");
    String s = SystemToolkit.removeLinks(report_pane.getText());
    editor.setText(s.replaceFirst("<div>",
        "<div style=\"font-family:monospace; font-size:" + font_point_size
            + "pt\">"));
    editor.setDoubleBuffered(false);
    hdr.print(editor);
  }

  /**
   * Returns the {@link Printable} that handles printing.
   * 
   * @return {@link Printable} that handles printing
   */
  public Printable getPrintable() {
    JEditorPane editor = new JEditorPane();
    editor.setContentType("text/html");
    String s = SystemToolkit.removeLinks(report_pane.getText());
    editor.setText(s.replaceFirst("<div>", "<div style=\"font-family:"
        + font_family + "; font-size:" + font_point_size + "pt\">"));
    editor.setDoubleBuffered(false);
    return new OnePagePrintable(editor, "Concept: " + concept.getCui());
  }

  private StringBuffer addConcept(StringBuffer report) {
    StringBuffer concept_row = new StringBuffer();
    StringBuffer concept_header = new StringBuffer();
    concept_header.append(bold_begin).append("&nbsp;Concept: ");
    concept_header.append("[CUI: " + concept.getCui() + "] ").append(bold_end);

    if (sorted_atoms.length == 0)
      concept_row.append("");
    else {
      concept_row.append(getPreferredAtom(sorted_atoms, precedence).replaceAll(
          "<", "&lt;").replaceAll(">", "&gt;"));
    }
    if (expanded_hm.containsKey("Concept") || expanded_hm.containsKey("ALL")) {
      if (expanded_hm.containsKey("ALL"))
        expanded_hm.put("Concept", null);
      report.append(
          "<a href=\"" + "Concept" + "\"><img src = \"" + minus_url
              + "\" border=0 alt=\"minus\"></a>").append(concept_header)
          .append(concept_row.toString()).append(line_end);
      report = addConceptAttributes(report);
    } else
      report.append(
          "<a href=\"" + "Concept" + "\"><img src = \"" + plus_url
              + "\" border=0></a>").append(concept_header).append(
          concept_row.toString()).append(line_end);
    return report;
  }

  private StringBuffer addConceptAttributes(StringBuffer report) {
    final StringBuffer st_buffer = new StringBuffer(200);
    st_buffer.append("<font color=\"#777777\">");
    for (int i = 0; i < atts.length; i++) {
      if (!(atts[i] instanceof AtomAttribute)
          && !(atts[i] instanceof RelationshipAttribute)
          && atts[i].getAtn().getAbbreviation().equals("ST")) {
        st_buffer.append("&nbsp;&nbsp;");
        st_buffer.append("ST Status ");
        st_buffer.append(atts[i].getAttributeValue()).append(" ");
        st_buffer.append("[");
        st_buffer.append(atts[i].getAttributeIdentifier()).append("/");
        st_buffer.append(atts[i].getSource().getRsab());
        st_buffer.append("]");
        st_buffer.append(line_end);
        continue;
      }

      else if (!(atts[i] instanceof AtomAttribute)
          && !(atts[i] instanceof RelationshipAttribute)
          && atts[i].getAtn().getAbbreviation().equals("MR")) {
        st_buffer.append("&nbsp;&nbsp;");
        st_buffer.append("MR Major Revision Date ");
        st_buffer.append(atts[i].getAttributeValue()).append(" ");
        st_buffer.append("[");
        st_buffer.append(atts[i].getAttributeIdentifier()).append("/");
        st_buffer.append(atts[i].getSource().getRsab());
        st_buffer.append("]");
        st_buffer.append(line_end);
        continue;
      }

      else if (!(atts[i] instanceof AtomAttribute)
          && !(atts[i] instanceof RelationshipAttribute)
          && atts[i].getAtn().getAbbreviation().equals("DA")) {
        st_buffer.append("&nbsp;&nbsp;");
        st_buffer.append("DA Date Added ");
        st_buffer.append(atts[i].getAttributeValue()).append(" ");
        st_buffer.append("[");
        st_buffer.append(atts[i].getAttributeIdentifier()).append("/");
        st_buffer.append(atts[i].getSource().getRsab());
        st_buffer.append("]");
        st_buffer.append(line_end);
        continue;
      }
    }
    st_buffer.append("</font>");

    return report.append(st_buffer.toString());
  }

  private StringBuffer addSemanticTypes(StringBuffer report) {
    // don't need to restrict
    stys = concept.getSemanticTypes();

    final StringBuffer sty_header = new StringBuffer(100);
    sty_header.append(bold_begin).append("&nbsp;Semantic Type");
    sty_header.append(stys.length == 1 ? "" : "s").append(bold_end);

    final StringBuffer sty_rows = new StringBuffer(1000);
    for (int i = 0; i < stys.length; i++) {
      sty_rows.append("&nbsp;&nbsp;");
      sty_rows.append(stys[i].getSty().getSty()).append(line_end);
    }
    if (expanded_hm.containsKey("Semantic Types")
        || expanded_hm.containsKey("ALL")) {
      if (expanded_hm.containsKey("ALL"))
        expanded_hm.put("Semantic Types", null);
      report.append(
          "<a href=\"" + "Semantic Types" + "\"><img src = \"" + minus_url
              + "\" border=0></a>").append(sty_header).append(line_end).append(
          sty_rows.toString());
    } else
      report.append(
          "<a href=\"" + "Semantic Types" + "\"><img src = \"" + plus_url
              + "\" border=0></a>").append(sty_header).append(line_end);

    return report;
  }

  private StringBuffer addDefinitions(StringBuffer report) {
    defs = concept.getDefinitions(restrictor);
    final StringBuffer def_buffer = new StringBuffer(1000);
    final StringBuffer def_header = new StringBuffer(100);

    def_header.append(bold_begin).append("&nbsp;Definition");
    def_header.append(defs.length == 1 ? "" : "s").append(bold_end);

    for (int i = 0; i < defs.length; i++) {
      final Definition def = defs[i];

      final StringBuffer def_line = new StringBuffer(def.getDef().length() + 40);
      if (def.getAtom() == null || def.getAtom().getAui().equals(""))
        def_line.append("Concept DEF");
      else {
        def_line.append(def.getAtom().getSource().getRsab()).append("/");
        def_line.append(def.getAtom().getTermType());
      }
      def_line.append("|");
      def_line.append(def.getDef());

      final String[] lines = MMSToolkit.splitString(def_line.toString(), 65);
      for (int j = 0; j < lines.length; j++) {
        if (lines[j] == null)
          break;

        def_buffer.append("&nbsp;&nbsp;").append(lines[j]);
        if (i < defs.length - 1
            || (i == defs.length - 1 && lines[j + 1] != null))
          def_buffer.append(line_end);
      }
      def_buffer.append(line_end);
    }

    if (!def_buffer.toString().equals("")) {
      if (expanded_hm.containsKey("Definitions")
          || expanded_hm.containsKey("ALL")) {
        if (expanded_hm.containsKey("ALL"))
          expanded_hm.put("Definitions", null);
        report.append("<a href=\"" + "Definitions" + "\"><img src = \""
            + minus_url + "\" border=0></a>");
        report.append(def_header).append(line_end).append(def_buffer);
      } else {
        report.append("<a href=\"" + "Definitions" + "\"><img src = \""
            + plus_url + "\" border=0></a>");
        report.append(def_header).append(line_end);
      }
    }
    return report;
  }

private StringBuffer addAtoms(StringBuffer report) {
    final StringBuffer atom_buffer = new StringBuffer(1000);
    final StringBuffer atom_header = new StringBuffer(150);

    String prev_lui = null;
    String prev_sui = null;

    //
    // Write atoms
    //

    // sort again depending on atom comparator
    try {
      if (atom_comparator != null
          && atom_comparator.toString().equals("Sorted by Lexical Class Rank"))
        Arrays.sort(sorted_atoms, new Atom.AtomRankComparator(sorted_atoms,
            (SourceTermType[]) precedence.toArray(new SourceTermType[0]), true,
            true, true));
      else
        Arrays.sort(sorted_atoms, atom_comparator);
    } catch (Exception e) {
      MMSToolkit.reportError(RRFBrowserGUI.getRRFBrowserGUI(), e.getMessage(),
          false);
      return report;
    }

    for (int i = 0; i < sorted_atoms.length; i++) {

      final Atom atom = sorted_atoms[i];

      final StringBuffer indent = new StringBuffer();

      //
      // Determine indentation level for Lexical Class Rank Sorting
      //
      if (i > 0
          && atom_comparator.toString().equals("Sorted by Lexical Class Rank")) {
        if (prev_lui != null
            && prev_lui.toString().equals(atom.getLui().toString())) {
          indent.append("&nbsp;&nbsp;");
          if (prev_sui != null
              && prev_sui.toString().equals(atom.getSui().toString()))
            indent.append("&nbsp;&nbsp;");
        }
      }
      indent.append("&nbsp;&nbsp;");
      atom_buffer.append(indent);

      boolean add_attributes = false;
      if (atom.getAttributes().length > 0) {

        if (expanded_hm.containsKey("Atom " + atom.getAui()) ||
            expanded_hm.containsKey("ALL")) {
          expanded_hm.put("Atom " + atom.getAui(), null);
          atom_buffer.append("<a href=\"" + "Atom " + atom.getAui() +
              "\"><img src = \"" + minus_url + "\" border=0></a>");
          add_attributes = true;
        } else {
          atom_buffer.append("<a href=\"" + "Atom " + atom.getAui()
              + "\"><img src = " + plus_url + " border=0></a>");
        }
          
        atom_buffer.append("&nbsp;");
      } else if (!atom_comparator.toString().equals(
          "Sorted by Lexical Class Rank")) {
        atom_buffer.append("&nbsp;&nbsp;");
      } 

      atom_buffer.append(atom.getStr().replaceAll("<", "&lt;").replaceAll(">",
      "&gt;"));
      atom_buffer.append(" [");
      atom_buffer.append(atom.getAui());
      atom_buffer.append("/");
      atom_buffer.append(atom.getSource().getRsab());
      atom_buffer.append("/");
      atom_buffer.append(atom.getTermType());
      atom_buffer.append("/");
      atom_buffer.append(atom.getCode());
      atom_buffer.append("] ");
      if (atom.getSuppress() != null
          && !atom.getSuppress().getAbbreviation().equals("N"))
        atom_buffer.append(atom.getSuppress().getAbbreviation());
      atom_buffer.append(line_end);

      if (add_attributes)
        atom_buffer.append(addAtomAttributes(atom));

      prev_lui = atom.getLui();
      prev_sui = atom.getSui();

    }

    if (!atom_buffer.toString().equals("")) {
      atom_header.append(bold_begin).append("&nbsp;Atoms (").append(
          sorted_atoms.length);
      atom_header.append("): ").append(bold_end).append("(");
      atom_header.append(atom_comparator.toString()).append(")");
      if (expanded_hm.containsKey("Atoms") || expanded_hm.containsKey("ALL")) {
        if (expanded_hm.containsKey("ALL")) expanded_hm.put("Atoms", null);
        report.append("<a href=\"" + "Atoms" + "\"><img src = \"" + minus_url + "\" border=0></a>");
        report.append(atom_header.toString()).append(line_end).append(atom_buffer);
      } else {
        report.append("<a href=\"" + "Atoms" + "\"><img src = \"" + plus_url +
                      "\" border=0></a>");
        report.append(atom_header.toString()).append(line_end);
      }
    }
    return report;
  }  private StringBuffer addAtomAttributes(Atom atom) {
    final StringBuffer atom_buffer = new StringBuffer(200);
    atom_buffer.append("<font color=\"#777777\">");
    Attribute[] atom_atts = atom.getAttributes();
    for (int j = 0; j < atom_atts.length; j++) {
      final AtomAttribute att = (AtomAttribute) atom_atts[j];
      atom_buffer.append("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
      if (atom_comparator.toString().equals("Sorted by Lexical Class Rank"))
        atom_buffer.append("&nbsp;&nbsp;");
      atom_buffer.append(att.getAttributeName().getAbbreviation()).append(
          "&nbsp;");
      atom_buffer.append(att.getAttributeValue()).append("&nbsp;");
      atom_buffer.append("[");
      atom_buffer.append(att.getAttributeIdentifier()).append("/");
      atom_buffer.append(att.getSource().getRsab());
      atom_buffer.append("]");
      atom_buffer.append(line_end);
    }
    atom_buffer.append("</font>");

    return atom_buffer;

  }

  private StringBuffer addLexicalRels(StringBuffer report) {
    String atom1 = null;
    String atom2 = null;
    rels = concept.getRelationships(restrictor);
    final StringBuffer local_buffer = new StringBuffer(1000);
    final StringBuffer lexical_rels_header = new StringBuffer(150);

    //
    // Write within concept rels
    //
    int within_concept_rels_ct = 0;
    for (int i = 0; i < rels.length; i++) {

      final Relationship rel = rels[i];

      if (!rel.getConcept().equals(rel.getRelatedConcept()))
        continue;
      within_concept_rels_ct++;

      if (rel instanceof AtomRelationship) {
        final AtomRelationship atom_rel = (AtomRelationship) rel;

        atom1 = atom_rel.getAtom().getStr();
        atom2 = atom_rel.getRelatedAtom().getStr();
        
        // Skip case that spans concept
        if (atom1 == null || atom2 == null) continue;
        String aui1 = atom_rel.getAtom().getAui();
        String aui2 = atom_rel.getRelatedAtom().getAui();

        if ((atom1.length() < atom2.length())
            || ((atom1.length() == atom2.length()) && (aui1.compareTo(aui2) < 0))) {
          local_buffer.append("&nbsp;&nbsp;").append(atom1).append(
              "&nbsp;&nbsp;");
          local_buffer.append("[");
          local_buffer.append(atom_rel.getRel().getAbbreviation());
          local_buffer.append("/").append(
              atom_rel.getSource().getRootSourceAbbreviation());
          local_buffer.append("]&nbsp;&nbsp;");
          local_buffer.append(atom2);
          local_buffer.append(line_end);
        }
      }
    }

    lexical_rels_header.append(bold_begin).append(
        "&nbsp;Within-Concept Relationships (").append(
        within_concept_rels_ct / 2).append(") ").append(bold_end);

    if (!local_buffer.toString().equals("")) {
      if (expanded_hm.containsKey("Lexical Rels")
          || expanded_hm.containsKey("ALL")) {
        if (expanded_hm.containsKey("ALL"))
          expanded_hm.put("Lexical Rels", null);
        report.append("<a href=\"" + "Lexical Rels" + "\"><img src = \""
            + minus_url + "\" border=0></a>");
        report.append(lexical_rels_header.toString()).append(line_end).append(
            local_buffer);
      } else {
        report.append("<a href=\"" + "Lexical Rels" + "\"><img src = \""
            + plus_url + "\" border=0></a>");
        report.append(lexical_rels_header.toString()).append(line_end);
      }
    }
    return report;
  }

  private StringBuffer addCoocurrences(StringBuffer report) {
    cocs = concept.getCoocurrences(restrictor);
    final StringBuffer coc_header = new StringBuffer(150);

    if (cocs.length != 0) {
      coc_header.append(bold_begin).append(
          "&nbsp;Coocurrences (" + cocs.length + ") ").append(bold_end);
      if (expanded_hm.containsKey("Coocurrences")
          || expanded_hm.containsKey("ALL")) {
        if (expanded_hm.containsKey("ALL"))
          expanded_hm.put("Coocurrences", new Integer(0));
        report.append("<a href=\"" + "Coocurrences" + "\"><img src = \""
            + minus_url + "\" border=0></a>");
        report.append(coc_header.toString()).append(line_end);
        report.append(getCoocurrences());
      } else {
        report.append("<a href=\"" + "Coocurrences" + "\"><img src = \""
            + plus_url + "\" border=0></a>");
        report.append(coc_header.toString()).append(line_end);
      }
    }
    return report;
  }

  private void addPageIcons(StringBuffer sb, String key, int low_range,
      int high_range, boolean at_high_end) {

    sb.append("<font color=\"green\">");
    sb.append("&nbsp;&nbsp;");
    if (expanded_hm.containsKey("Full" + key))
      sb.append("[all:").append("<a href=\"" + "Full" + key + "\">page</a>")
          .append("]");
    else {
      sb.append("[").append("<a href=\"" + "Full" + key + "\">all</a>").append(
          ":page]");

      if (low_range == 0)
        sb.append("&nbsp;").append("[prev:");
      else
        sb.append("&nbsp;").append(
            "[" + "<a href=\"" + "Back" + key + "\">prev</a>" + ":");
      if (at_high_end)
        sb.append("next]");
      else
        sb.append("<a href=\"" + "Forward" + key + "\">next</a>" + "]");

      sb.append("&nbsp;").append("(records ").append(low_range + 1).append(
          " to ").append(high_range).append(")");
    }
    sb.append(line_end);
    sb.append("</font>");
  }

  private int getLowestDisplayedRecord(String key) {
    if (expanded_hm.containsKey(key)) {
      Object value = expanded_hm.get(key);
      if (value == null)
        return 0;

      int int_value = ((Integer) value).intValue();
      if (int_value < 0)
        return 0;
      else
        return int_value;
    } else {
      return 0;
    }
  }

  private StringBuffer getCoocurrences() {
    StringBuffer coc_buffer = new StringBuffer(1500);
    //
    // Write cocs not within concept
    //
    int low_range = 0;
    int high_range = cocs.length;
    if (cocs.length > page_size) {
      if (!expanded_hm.containsKey("FullCoocurrences")) {
        low_range = getLowestDisplayedRecord("Coocurrences");
        high_range = low_range + page_size;
        if (high_range > cocs.length)
          high_range = cocs.length;
      }
      addPageIcons(coc_buffer, "Coocurrences", low_range, high_range,
          high_range >= cocs.length);
    }
    for (int i = low_range; i < high_range; i++) {

      final Coocurrence coc = cocs[i];

      if (coc.getConcept().equals(coc.getRelatedConcept()))
        continue;
      coc_buffer.append(" ");

      if (coc instanceof AtomCoocurrence) {
        final AtomCoocurrence atom_coc = (AtomCoocurrence) coc;
        Atom related_atom = atom_coc.getRelatedAtom();
        if (related_atom.getAui().equals(""))
          continue;
        try {
          related_atom = index_lookup.findAtomByAui(related_atom.getAui());
        } catch (Exception ioe) {
          ioe.printStackTrace();
          MMSToolkit
              .notifyUser(
                  RRFBrowserGUI.getRRFBrowserGUI(),
                  "Unexpected error looking concept up by this CUI:"
                      + related_atom.getAui()
                      + "\n"
                      + "The most likely explanation is that your subset is missing\n"
                      + "an indexes/AUI_MRCONSO.x index file. \n"
                      + "See the console for more information.");
          continue;
        }

        coc_buffer.append("&nbsp;&nbsp;").append("[");
        String cot_cof = coc.getCot().getAbbreviation() + ":" + coc.getCof();
        coc_buffer.append(cot_cof);
        coc_buffer.append("]");
        for (int k = (9 - cot_cof.length()); k > 0; k--) {
          coc_buffer.append("&nbsp;");
        }
        coc_buffer.append(related_atom.getStr()).append("&nbsp;");
        coc_buffer.append("[");
        coc_buffer.append(related_atom.getSource().getRsab()).append("/");
        coc_buffer.append(related_atom.getTermType().getAbbreviation()).append(
            "|");
        coc_buffer.append(atom_coc.getCoa().replaceAll("<>", "&lt;&gt;"))
            .append("|");
        coc_buffer.append(atom_coc.getSource().getRsab());
        coc_buffer.append("] ");
        coc_buffer.append("<a href=\"CUI:" + related_atom.getConcept().getCui() + "\">"
            + related_atom.getConcept().getCui() + "</a>");
        coc_buffer.append(line_end);
      }
    }
    if (cocs.length > page_size) {
      addPageIcons(coc_buffer, "Coocurrences", low_range, high_range,
          high_range >= cocs.length);
    }
    return coc_buffer;
  }

  private StringBuffer addMappings(StringBuffer report) {
    maps = concept.getMappings();
    final StringBuffer map_header = new StringBuffer(150);

    if (maps.length != 0) {
      map_header.append(bold_begin).append(
          "&nbsp;Mappings (" + maps.length + ") ").append(bold_end);
      if (expanded_hm.containsKey("Mappings") || expanded_hm.containsKey("ALL")) {
        if (expanded_hm.containsKey("ALL"))
          expanded_hm.put("Mappings", null);
        report.append("<a href=\"" + "Mappings" + "\"><img src = \""
            + minus_url + "\" border=0></a>");
        report.append(map_header.toString()).append(line_end);
        report.append(getMappings());
      } else {
        report.append("<a href=\"" + "Mappings" + "\"><img src = \"" + plus_url
            + "\" border=0></a>");
        report.append(map_header.toString()).append(line_end);
      }
    }
    return report;

  }

  private StringBuffer getMappings() {
    StringBuffer map_buffer = new StringBuffer(1500);

    //
    // Write cocs not within concept
    //
    int low_range = 0;
    int high_range = maps.length;
    if (maps.length > page_size) {
      if (!expanded_hm.containsKey("FullMappings")) {
        low_range = getLowestDisplayedRecord("Mappings");
        high_range = low_range + page_size;
        if (high_range > maps.length)
          high_range = maps.length;
      }
      addPageIcons(map_buffer, "Mappings", low_range, high_range,
          high_range >= maps.length);
    }
    for (int i = low_range; i < high_range; i++) {

      final Mapping map = maps[i];

      map_buffer.append(" ");

      map_buffer.append("&nbsp;&nbsp;");
      map_buffer.append(map.getSource().getRootSourceAbbreviation())
          .append(" ");
      map_buffer.append((map.getFrom() != null) ? map.getFrom().getExpr() : "")
          .append(" ");
      map_buffer.append(
          (map.getFrom() != null && map.getFrom().getType() != null) ? map
              .getFrom().getType().getAbbreviation() : "").append(" ");
      map_buffer.append(
          (map.getRel() != null) ? map.getRel().getAbbreviation() : "").append(
          " ");
      map_buffer.append(
          (map.getRela() != null) ? map.getRela().getAbbreviation() : "")
          .append(" ");
      map_buffer.append((map.getTo() != null) ? map.getTo().getExpr() : "")
          .append(" ");
      map_buffer.append(
          (map.getTo() != null && map.getTo().getType() != null) ? map.getTo()
              .getType().getAbbreviation() : "").append(" ");

      map_buffer.append(line_end);

    }
    if (maps.length > page_size) {
      addPageIcons(map_buffer, "Mappings", low_range, high_range,
          high_range >= maps.length);
    }

    return map_buffer;
  }

  private StringBuffer addContexts(StringBuffer report) {
    contexts = concept.getContexts(restrictor);
    final StringBuffer cxt_buffer = new StringBuffer(1000);
    final StringBuffer cxt_header = new StringBuffer(150);
    Arrays.sort(contexts, new ContextsComparator());

    for (int i = 0; i < contexts.length; i++) {

      final Context context = contexts[i];

      boolean context_components = false;
      cxt_buffer.append("&nbsp;&nbsp;");
      if (expanded_hm.containsKey("Context," + i)
          || expanded_hm.containsKey("ALL")) {
        if (expanded_hm.containsKey("ALL"))
          expanded_hm.put("Context," + i, null);
        cxt_buffer.append("<a href=\"" + "Context," + i + "\"><img src = \""
            + minus_url + "\" border=0></a>");
        context_components = true;
      } else {
        cxt_buffer.append("<a href=\"" + "Context," + i + "\"><img src = \""
            + plus_url + "\" border=0></a>");
      }

      cxt_buffer.append("&nbsp;").append(context.getSource().getRsab()).append(
          "/");
      cxt_buffer.append(context.getAtom().getTermType()).append("/");
      cxt_buffer.append(context.getAtom().getCode()).append("&nbsp;");
      cxt_buffer.append(context.getContextNumber()).append("&nbsp;");

      // get the CCP atom
      Atom ccp_atom = context.getAtom();
      String ccp_string = ccp_atom.getStr();

      cxt_buffer.append(ccp_string);
      cxt_buffer.append(line_end);

      if (context_components)
        cxt_buffer.append(addContextComponents(i));
    }

    if (!cxt_buffer.toString().equals("")) {
      cxt_header.append(bold_begin).append("&nbsp;Contexts (").append(
          contexts.length).append(")").append(bold_end);
      if (expanded_hm.containsKey("Context") || expanded_hm.containsKey("ALL")) {
        if (expanded_hm.containsKey("ALL"))
          expanded_hm.put("Context", null);
        report.append("<a href=\"" + "Context" + "\"><img src = \"" + minus_url
            + "\" border=0></a>");
        report.append(cxt_header.toString()).append(line_end)
            .append(cxt_buffer);
      } else {
        report.append("<a href=\"" + "Context" + "\"><img src = \"" + plus_url
            + "\" border=0></a>");
        report.append(cxt_header.toString()).append(line_end);
      }
    }
    return report;
  }

  private StringBuffer addContextComponents(int context_index) {
    final StringBuffer local_buffer = new StringBuffer(1000);
    // add ANC node if needed
    if (!contexts[context_index].getPtr().equals("")) {
      StringBuffer anc_header = new StringBuffer();
      anc_header.append(bold_begin).append("&nbsp;ANC ").append(bold_end);

      local_buffer.append("&nbsp;&nbsp;&nbsp;&nbsp;");
      if (expanded_hm.containsKey("Context," + context_index + ",ANC")
          || expanded_hm.containsKey("ALL")) {
        if (expanded_hm.containsKey("ALL"))
          expanded_hm.put("Context," + context_index + ",ANC", null);
        local_buffer.append("<a href=\"" + "Context," + context_index + ",ANC"
            + "\"><img src = \"" + minus_url + "\" border=0></a>");
        local_buffer.append(anc_header.toString()).append(line_end);
        local_buffer.append(addContextAncestors(context_index));
      } else {
        local_buffer.append("<a href=\"" + "Context," + context_index + ",ANC"
            + "\"><img src = \"" + plus_url + "\" border=0></a>");
        local_buffer.append(anc_header.toString()).append(line_end);
      }
    }

    // add SIB node if needed
    int siblings_ct = 0;
    try {
      siblings_ct = index_lookup.countSiblingContexts(contexts[context_index]);
    } catch (IOException ioe) {
      ioe.printStackTrace();
      MMSToolkit.notifyUser(RRFBrowserGUI.getRRFBrowserGUI(),
          "Unexpected error looking concept up the sibling count." + "\n"
              + "The most likely explanation is that your subset is missing\n"
              + "an indexes/PTR_AUI_RELA_MRHIER.x index file. \n"
              + "See the console for more information.");

    }
    // don't want to count self string
    if (siblings_ct > 1) {
      StringBuffer sib_header = new StringBuffer();
      sib_header.append(bold_begin).append("&nbsp;SIB ").append(bold_end);

      local_buffer.append("&nbsp;&nbsp;&nbsp;&nbsp;");
      if (expanded_hm.containsKey("Context," + context_index + ",SIB")
          || expanded_hm.containsKey("ALL")) {
        if (expanded_hm.containsKey("ALL"))
          expanded_hm.put("Context," + context_index + ",SIB", null);
        local_buffer.append("<a href=\"" + "Context," + context_index + ",SIB"
            + "\"><img src = \"" + minus_url + "\" border=0></a>");
        local_buffer.append(sib_header.toString()).append(line_end);
        local_buffer.append(addContextSiblings(context_index));
      } else {
        local_buffer.append("<a href=\"" + "Context," + context_index + ",SIB"
            + "\"><img src = \"" + plus_url + "\" border=0></a>");
        local_buffer.append(sib_header.toString()).append(line_end);
      }

    }

    // add CHD node if needed
    int chd_ct = 0;
    try {
      chd_ct = index_lookup.countChildContexts(contexts[context_index]);
    } catch (IOException ioe) {
      ioe.printStackTrace();
      MMSToolkit.notifyUser(RRFBrowserGUI.getRRFBrowserGUI(),
          "Unexpected error looking concept up the child count." + "\n"
              + "The most likely explanation is that your subset is missing\n"
              + "an indexes/PTR_AUI_RELA_MRHIER.x index file. \n"
              + "See the console for more information.");

    }
    if (chd_ct > 0) {
      StringBuffer chd_header = new StringBuffer();
      chd_header.append(bold_begin).append("&nbsp;CHD ").append(bold_end);
      local_buffer.append("&nbsp;&nbsp;&nbsp;&nbsp;");
      if (expanded_hm.containsKey("Context," + context_index + ",CHD")
          || expanded_hm.containsKey("ALL")) {
        if (expanded_hm.containsKey("ALL"))
          expanded_hm.put("Context," + context_index + ",CHD", null);

        local_buffer.append("<a href=\"" + "Context," + context_index + ",CHD"
            + "\"><img src = \"" + minus_url + "\" border=0></a>");
        local_buffer.append(chd_header.toString()).append(line_end);
        local_buffer.append(addContextChildren(context_index));
      } else {
        local_buffer.append("<a href=\"" + "Context," + context_index + ",CHD"
            + "\"><img src = \"" + plus_url + "\" border=0></a>");
        local_buffer.append(chd_header.toString()).append(line_end);
      }
    }
    return local_buffer;
  }

  private StringBuffer addContextAncestors(int context_index) {
    StringBuffer local_buffer = new StringBuffer(1000);

    String path_to_root = contexts[context_index].getPtr();
    final String[] tokens = FieldedStringTokenizer.split(path_to_root, ".");
    final StringBuffer indent = new StringBuffer(
        "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
    for (int j = 0; j < tokens.length; j++) {
      Atom atom = null;
      try {
        if (!tokens[j].equals(""))
          atom = index_lookup.findAtomByAui(tokens[j]);
      } catch (Exception e) {
        e.printStackTrace();
        MMSToolkit
            .notifyUser(
                RRFBrowserGUI.getRRFBrowserGUI(),
                "Unexpected error looking concept up by this AUI:"
                    + tokens[j]
                    + "\n"
                    + "The most likely explanation is that your subset is missing\n"
                    + "an indexes/AUI_MRCONSO.x index file. \n"
                    + "See the console for more information.");
        continue;
      }
      local_buffer.append(indent);
      local_buffer.append("<a href=\"CUI:" + atom.getConcept().getCui() + "\">"
          + atom.getStr() + "</a>");
      local_buffer.append(line_end);
      indent.append("&nbsp;");
    }

    return local_buffer;
  }

  private StringBuffer addContextChildren(int context_index) {
    StringBuffer local_buffer = new StringBuffer(1000);
    // get the children
    Set children_set = new HashSet();
    Context[] children = null;
    try {
      children = index_lookup.findChildContexts(contexts[context_index]);
    } catch (IOException ioe) {
      ioe.printStackTrace();
      MMSToolkit.notifyUser(RRFBrowserGUI.getRRFBrowserGUI(),
          "Unexpected error looking concept up the child contexts." + "\n"
              + "The most likely explanation is that your subset is missing\n"
              + "an indexes/PTR_AUI_RELA_MRHIER.x index file. \n"
              + "See the console for more information.");

    }

    Atom child_atom;
    for (int j = 0; j < children.length; j++) {
      try {
        child_atom = index_lookup.findAtomByAui(children[j].getAtom().getAui());
      } catch (Exception e) {
        e.printStackTrace();
        MMSToolkit
            .notifyUser(
                RRFBrowserGUI.getRRFBrowserGUI(),
                "Unexpected error looking concept up by this AUI:"
                    + contexts[context_index].getAtom().getAui()
                    + "\n"
                    + "The most likely explanation is that your subset is missing\n"
                    + "an indexes/AUI_MRCONSO.x index file. \n"
                    + "See the console for more information.");
        continue;
      }
      children_set.add(child_atom.getStr());
      str_to_cui.put(child_atom.getStr(), child_atom.getConcept().getCui());
    }

    String[] children_array = (String[]) children_set.toArray(new String[0]);
    Arrays.sort(children_array);

    for (int i = 0; i < children_array.length; i++) {
      local_buffer.append("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
      local_buffer.append(
          "<a href=\"CUI:" + str_to_cui.get(children_array[i]) + "\">"
              + children_array[i] + "</a>").append(line_end);
    }

    return local_buffer;
  }

  private StringBuffer addContextSiblings(int context_index) {
    StringBuffer local_buffer = new StringBuffer(1000);
    Context[] siblings = null;
    Set siblings_set = new HashSet();
    try {
      siblings = index_lookup.findSiblingContexts(contexts[context_index]);
    } catch (IOException ioe) {
      ioe.printStackTrace();
      MMSToolkit.notifyUser(RRFBrowserGUI.getRRFBrowserGUI(),
          "Unexpected error looking concept up the sibling contexts." + "\n"
              + "The most likely explanation is that your subset is missing\n"
              + "an indexes/PTR_AUI_RELA_MRHIER.x index file. \n"
              + "See the console for more information.");
    }

    for (int j = 0; j < siblings.length; j++) {
      Atom sibling_atom;
      try {
        sibling_atom = index_lookup.findAtomByAui(siblings[j].getAtom()
            .getAui());
      } catch (Exception e) {
        e.printStackTrace();
        MMSToolkit
            .notifyUser(
                RRFBrowserGUI.getRRFBrowserGUI(),
                "Unexpected error looking concept up by this AUI:"
                    + siblings[j].getAtom().getAui()
                    + "\n"
                    + "The most likely explanation is that your subset is missing\n"
                    + "an indexes/AUI_MRCONSO.x index file. \n"
                    + "See the console for more information.");
        continue;
      }
      String sibling_string = sibling_atom.getStr();
      if (!sibling_string.equals(contexts[context_index].getAtom().getStr())) {
        siblings_set.add(sibling_string);
        str_to_cui.put(sibling_string, sibling_atom.getConcept().getCui());
      }

    }

    String[] sibling_array = (String[]) siblings_set.toArray(new String[0]);
    Arrays.sort(sibling_array);

    for (int i = 0; i < sibling_array.length; i++) {
      local_buffer.append("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
      local_buffer.append(
          "<a href=\"CUI:" + str_to_cui.get(sibling_array[i]) + "\">"
              + sibling_array[i] + "</a>").append(line_end);
    }
    return local_buffer;
  }

  private StringBuffer addRelationships(StringBuffer report) {
    final StringBuffer rel_buffer = new StringBuffer(1000);
    final StringBuffer rel_header = new StringBuffer(150);
    int within_concept_ct = 0;
    
    //
    // Write rels not within concept
    //
    for (int i = 0; i < rels.length; i++) {

      final Relationship rel = rels[i];

      if (rel.getConcept().equals(rel.getRelatedConcept())) {
      	within_concept_ct += 1;
        continue;
      }

      // add rel to rel_name hashmap to categorize by rel name
      if (rel_name_to_rels.containsKey(rel.getRel().getAbbreviation())) {
        HashSet rels = (HashSet) rel_name_to_rels.get(rel.getRel()
            .getAbbreviation());
        rels.add(rel);
      } else {
        HashSet rels = new HashSet();
        rels.add(rel);
        rel_name_to_rels.put(rel.getRel().getAbbreviation(), rels);
      }
    }

    // sort rel name groups
    List rel_names = new ArrayList(rel_name_to_rels.keySet());
    Collections.sort(rel_names);

    // add a node for each rel name group
    for (int i = 0; i < rel_names.size(); i++) {
      int count = ((HashSet) rel_name_to_rels.get(rel_names.get(i))).size();
      if (expanded_hm.containsKey(rel_names.get(i))
          || expanded_hm.containsKey("ALL")) {
        if (expanded_hm.containsKey("ALL"))
          expanded_hm.put(rel_names.get(i), null);
        rel_buffer.append("&nbsp;&nbsp;").append(
            "<a href=\"" + rel_names.get(i) + "\"><img src = \"" + minus_url
                + "\" border=0></a>");
        rel_buffer.append("&nbsp;").append(
            rel_names.get(i).toString() + " (" + count + ")").append(line_end);

        // list relationships in this group
        addRelationshipName(rel_buffer, rel_names.get(i).toString());
      } else {
        rel_buffer.append("&nbsp;&nbsp;").append(
            "<a href=\"" + rel_names.get(i) + "\"><img src = \"" + plus_url
                + "\" border=0></a>");
        rel_buffer.append("&nbsp;").append(
            rel_names.get(i).toString() + " (" + count + ")");
        rel_buffer.append(line_end);
      }
    }
    // add Relationship header and subgroup nodes if needed
    if (!rel_buffer.toString().equals("")) {
      rel_header.append(bold_begin).append("&nbsp;Relationships (").append(
          rels.length - within_concept_ct).append("):").append(bold_end);
      if (expanded_hm.containsKey("Relationships")
          || expanded_hm.containsKey("ALL")) {
        if (expanded_hm.containsKey("ALL"))
          expanded_hm.put("Relationships", null);
        report.append("<a href=\"" + "Relationships" + "\"><img src = \""
            + minus_url + "\" border=0></a>");
        report.append(rel_header.toString()).append(line_end)
            .append(rel_buffer);
      } else {
        report.append("<a href=\"" + "Relationships" + "\"><img src = \""
            + plus_url + "\" border=0></a>");
        report.append(rel_header.toString()).append(line_end);
      }
    }

    return report;

  }

  private StringBuffer addRelationshipName(StringBuffer buffer, String rel_name) {
    // get list of rels associated with clicked on rel name
    Iterator iter = rel_name_to_rels.keySet().iterator();
    ArrayList rels_list = new ArrayList();
    while (iter.hasNext()) {
      String this_rel_name = (String) iter.next();
      if (rel_name.indexOf(this_rel_name) != -1) {
        HashSet rels_hs = (HashSet) rel_name_to_rels.get(this_rel_name);
        rels_list = new ArrayList(rels_hs);
        Collections.sort(rels_list, new RelationshipsComparator());
        break;
      }
    }
    // collect relationship data for all rels in rel name group
    int low_range = 0;
    int high_range = rels_list.size();
    if (rels_list.size() > page_size) {
      if (!expanded_hm.containsKey("Full" + rel_name)) {
        low_range = getLowestDisplayedRecord(rel_name);
        high_range = low_range + page_size;
        if (high_range > rels_list.size())
          high_range = rels_list.size();
      }
      addPageIcons(buffer, rel_name, low_range, high_range,
          high_range >= cocs.length);
    }
    for (int i = low_range; i < high_range; i++) {

      final Relationship rel = (Relationship) rels_list.get(i);
      final RelationshipAttribute[] rel_atts = rel.getAttributes(restrictor);
      StringBuffer rel_buffer = new StringBuffer(1000);

      if (rel.getConcept().equals(rel.getRelatedConcept()))
        continue;

      if (rel instanceof AtomRelationship) {
        final AtomRelationship atom_rel = (AtomRelationship) rel;
        Atom related_atom = atom_rel.getRelatedAtom();
        try {
          related_atom = index_lookup.findAtomByAui(related_atom.getAui());
        } catch (Exception ioe) {
          ioe.printStackTrace();
          MMSToolkit
              .notifyUser(
                  RRFBrowserGUI.getRRFBrowserGUI(),
                  "Unexpected error looking concept up by this AUI:"
                      + related_atom.getAui()
                      + "\n"
                      + "The most likely explanation is that your subset is missing\n"
                      + "an indexes/AUI_MRCONSO.x index file. \n"
                      + "See the console for more information.");
          continue;
        }
        // need this keep b/c don't have related atom info earlier
        if (!restrictor.keep(related_atom))
          continue;

        if (rel_atts.length > 0) {
          if (expanded_hm.containsKey("Relationships," + related_atom.getStr())
              || expanded_hm.containsKey("ALL")) {
            if (expanded_hm.containsKey("ALL"))
              expanded_hm.put("Relationships " + related_atom.getStr(), null);
            rel_buffer.append("&nbsp;&nbsp;&nbsp;&nbsp;").append(
                "<a href=\"" + "Relationships," + related_atom.getStr()
                    + "\"><img src = \"" + minus_url + "\" border=0></a>")
                .append("&nbsp;");
          } else {
            rel_buffer.append("&nbsp;&nbsp;&nbsp;&nbsp;").append(
                "<a href=\"" + "Relationships," + related_atom.getStr()
                    + "\"><img src = \"" + plus_url + "\" border=0></a>")
                .append("&nbsp;");
          }
        } else {
          rel_buffer.append("&nbsp;&nbsp;&nbsp;&nbsp;");
        }
        rel_buffer.append(related_atom.getStr()).append(" ");
        rel_buffer.append("[");
        rel_buffer.append(related_atom.getSource().getRsab()).append("/");
        rel_buffer.append(related_atom.getTermType().getAbbreviation()).append(
            "|");
        rel_buffer.append(atom_rel.getRela().getAbbreviation()).append("|");
        rel_buffer.append(atom_rel.getSource().getRsab());
        rel_buffer.append("] ");
        rel_buffer.append("<a href=\"CUI:" + related_atom.getConcept().getCui() + "\">"
            + related_atom.getConcept().getCui() + "</a>");
        rel_buffer.append(line_end);
        if (rel_atts.length > 0) {
          if (expanded_hm.containsKey("Relationships," + related_atom.getStr())
              || expanded_hm.containsKey("ALL")) {
            rel_buffer.append(addRelationshipAttributes(rel_atts));
          }
        }
      } else {

        rel_buffer.append("&nbsp;&nbsp;&nbsp;&nbsp;");
        try {
          Concept concept2 = index_lookup.findConceptByCui(rel
              .getRelatedConcept().getCui());
          Atom[] rel_atoms = concept2.getAtoms(restrictor);
          if (rel_atoms.length > 0)
            rel_buffer.append(getPreferredAtom(rel_atoms, precedence));
        } catch (Exception ioe) {
          ioe.printStackTrace();
          MMSToolkit
              .notifyUser(
                  RRFBrowserGUI.getRRFBrowserGUI(),
                  "Unexpected error looking concept up by this CUI:"
                      + rels[i].getRelatedConcept().getCui()
                      + "\n"
                      + "The most likely explanation is that your subset is missing\n"
                      + "an indexes/AUI_MRCONSO.x index file. \n"
                      + "See the console for more information.");
        }

        rel_buffer.append("[");
        rel_buffer.append(rel.getRela().getAbbreviation()).append("|");
        rel_buffer.append(rel.getSource().getRsab());
        rel_buffer.append("] ");
        rel_buffer.append("<a href=\"CUI:" + rel.getRelatedConcept().getCui() + "\">"
            + rel.getRelatedConcept().getCui() + "</a>");
        rel_buffer.append(line_end);
      }

      buffer.append(rel_buffer);
    }
    if (rels_list.size() > page_size) {
      addPageIcons(buffer, rel_name, low_range, high_range,
          high_range >= cocs.length);
    }

    return buffer;
  }

  private StringBuffer addRelationshipAttributes(
      RelationshipAttribute[] rel_atts) {
    StringBuffer att_buffer = new StringBuffer(1000);
    for (int j = 0; j < rel_atts.length; j++) {
      final RelationshipAttribute att = rel_atts[j];

      att_buffer.append("<font color=\"#777777\">");
      att_buffer.append("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
      att_buffer.append(att.getAttributeName().getAbbreviation()).append(" ");
      att_buffer.append(att.getAttributeValue()).append(" ");
      att_buffer.append("[");
      att_buffer.append(att.getAttributeIdentifier()).append("/");
      att_buffer.append(att.getSource().getRsab());
      att_buffer.append("]");
      att_buffer.append(line_end);
      att_buffer.append("</font>");
    }
    return att_buffer;
  }

  /**
   * This method takes a concept object and produces a concept report from it.
   * This concept report is then returned as a string.
   * 
   * @param concept
   *          An object {@link Concept}.
   * @return An object {@link String} representation of report.
   */
  private String getReport(Concept concept) {
    // do this to set the atom_comparator correctly
    if (atom_comparator == null)
      new ExpandableConceptViewOptionsGUI(this);

    report = new StringBuffer(10000);

    if (getContentType().equals("text/html")) {
      report.append(
          "<div style=\"font-family: " + font_family + "; font-size:"
              + font_point_size + "pt\">").append(line_end);
      bold_begin = "<b>";
      bold_end = "</b>";
    }

    //
    // Concept information
    // CN Section
    //
    //
    // Get atoms from concept
    //
    // Get list of source ttys in precedence order
    precedence = new ArrayList();
    try {
      BufferedReader in = new BufferedReader(new InputStreamReader(
          new FileInputStream(new File(user_configuration.getSourcePaths()[0],
              "MRRANK.RRF")), "UTF-8"));
      String line;
      while ((line = in.readLine()) != null) {
        String[] tokens = FieldedStringTokenizer.split(line, "|");
        SourceTermType source_tty = default_configuration.getSourceTermType(
            tokens[1], tokens[2]);
        precedence.add(source_tty);
      }
      in.close();
    } catch (Exception e) {
      Exception e2 = new Exception("Problem getting ranking from MRRANK.RRF");
      e2.initCause(e);
    }

    sorted_atoms = concept.getAtoms(restrictor);
    atts = concept.getAttributes();

    //
    // Concept Section
    //
    addConcept(report);

    //
    // STY Section
    //
    addSemanticTypes(report);

    //
    // DEF Section
    //
    addDefinitions(report);

    //
    // ATOMS Section
    //
    addAtoms(report);

    //
    // WITHIN CONCEPT RELS section
    //
    addLexicalRels(report);

    //
    // CONTEXT Section
    //
    addContexts(report);

    //
    // RELATIONSHIPS section
    //
    addRelationships(report);

    //
    // COOCURRENCES section
    //
    addCoocurrences(report);

    //
    // MAPPINGS section
    //
    addMappings(report);

    if (expanded_hm.containsKey("ALL"))
      expanded_hm.remove("ALL");

    if (getContentType().equals("text/html")) {
      report.append("</div>").append(line_end);
    }

    // Return the report
    return report.toString();
  }

  /**
   * Returns the preferred atom's string from amongst the given {@link Atom}[].
   * Uses the Atom.AtomRankComparator and the given precidence list to do
   * ranking.
   * 
   * @param atoms
   *          {@link Atom}[]
   * @param precedence
   *          {@link List}
   * @return preferred atom's STR string
   */
  public String getPreferredAtom(Atom[] atoms, List precedence) {
    try {
      Arrays.sort(atoms, new Atom.AtomRankComparator(atoms,
          (SourceTermType[]) precedence.toArray(new SourceTermType[0]), true,
          true, true));
    } catch (Exception e) {
      MMSToolkit.reportError(RRFBrowserGUI.getRRFBrowserGUI(), e.getMessage(),
          false);
      return report.toString();
    }
    return atoms[0].getStr();
  }

  /**
   * Handles opening of a source directory.
   * 
   * @param source_dir
   *          the source directory
   */
  public void subsetDirectoryOpened(File source_dir) {
    try {
      index_lookup = new RichMRFilesIndexLookup(user_configuration
          .getSourcePaths()[0]);
    } catch (IOException ioe) {
      ioe.printStackTrace();
      MMSToolkit.reportError(RRFBrowserGUI.getRRFBrowserGUI(),
          "Unexpected error opening the subset directory.", false);
    }
  }

  public class PopupListener extends MouseAdapter {
    JPopupMenu popup;

    PopupListener(JPopupMenu popupMenu) {
      popup = popupMenu;
    }

    public void mousePressed(MouseEvent e) {
      maybeShowPopup(e);
    }

    public void mouseReleased(MouseEvent e) {
      maybeShowPopup(e);
    }

    private void maybeShowPopup(MouseEvent e) {
      if (e.isPopupTrigger()) {
        popup.show(e.getComponent(), e.getX(), e.getY());
      }
    }
  }

  /**
   * Sorts the contexts.
   */
  private class ContextsComparator implements Comparator {

    //
    // Constructors
    //

    /**
     * Instantiates a comparator.
     */
    public ContextsComparator() {
    }

    //
    // Implementation of Comparator interface
    //

    /**
     * Compares the specified field of each {@link String} object.
     * 
     * @param o1
     *          the first {@link Object} to be compared.
     * @param o2
     *          the second {@link Object} to be compared.
     * @return An <code>int</code> which might be negative integer, zero, or a
     *         positive integer as the first argument is less than, equal to, or
     *         greater than the second.
     */
    public int compare(Object o1, Object o2) {
      if (o1 == null || o2 == null || !(o1 instanceof Context)
          || !(o2 instanceof Context))
        return 0;
      Context s1 = (Context) o1;
      Context s2 = (Context) o2;

      String source1 = s1.getSource().getRsab();
      String source2 = s2.getSource().getRsab();

      return source1.compareTo(source2);
    }
  }

  /**
   * Sorts the relationships.
   */
  private class RelationshipsComparator implements Comparator {

    //
    // Constructors
    //

    /**
     * Instantiates a comparator.
     */
    public RelationshipsComparator() {
    }

    //
    // Implementation of Comparator interface
    //

    /**
     * Compares the specified field of each {@link String} object.
     * 
     * @param o1
     *          the first {@link Object} to be compared.
     * @param o2
     *          the second {@link Object} to be compared.
     * @return An <code>int</code> which might be negative integer, zero, or a
     *         positive integer as the first argument is less than, equal to, or
     *         greater than the second.
     */
    public int compare(Object o1, Object o2) {
      if (o1 == null || o2 == null || !(o1 instanceof Relationship)
          || !(o2 instanceof Relationship))
        return 0;
      Relationship s1 = (Relationship) o1;
      Relationship s2 = (Relationship) o2;

      int cmp = s1.getRelatedConcept().getCui().compareTo(
          s2.getRelatedConcept().getCui());
      if (cmp == 0) {
        cmp = s1.getSource().getRsab().compareTo(s2.getSource().getRsab());
      }
      if (cmp == 0) {
        cmp = s1.compareTo(s2);
      }

      return cmp;
    }
  }

  /**
   * Provides the Report View Options GUI.
   */
  private class ExpandableConceptViewOptionsGUI extends JDialog {

    //
    // Private Fields
    //
    private SuperJList source_list = new SuperJList();

    private SuperJList language_list = new SuperJList();

    private JComboBox atom_comparator_display;

    private JComboBox page_size_display;

    /**
     * Default constructor.
     * 
     * @param f
     *          {@link ExpandableConceptView}
     */
    public ExpandableConceptViewOptionsGUI(ExpandableConceptView f) {
      super(RRFBrowserGUI.getRRFBrowserGUI());
      setModal(true);
      setTitle("Report View Options");
      initialize();
    };

    //
    // Methods
    //

    /**
     * This method sets up the GUI
     */
    private void initialize() {

      final DataChangeListener dcl = new DataChangeListener();

      getContentPane().setLayout(new BorderLayout());

      // configure source_lang_panel
      JPanel source_lang_panel = new JPanel();
      BoxLayout source_lang_layout = new BoxLayout(source_lang_panel,
          BoxLayout.Y_AXIS);
      source_lang_panel.setLayout(source_lang_layout);

      JLabel list_label = new JLabel("Select source(s) for display.");
      source_lang_panel.add(list_label);
      SourcesToRemoveFilter strf = new SourcesToRemoveFilter();
      strf.setConfiguration(default_configuration);
      strf.setProperties(user_configuration.getProperties(strf));
      List sources = (strf.removeSelectedSources() ? strf
          .getSelectedSourcesInverse() : strf.getSelectedSources());
      Source[] sorted_sources = (Source[]) sources.toArray(new Source[0]);
      Arrays.sort(sorted_sources);
      source_list = new SuperJList(sorted_sources);
      source_list
          .setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
      if (!display_sources.isEmpty()) {
        Source[] source_array = (Source[]) display_sources
            .toArray(new Source[0]);
        source_list.setSelectedValues(source_array, false);
      }

      JScrollPane sp = new JScrollPane(source_list,
          ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED,
          ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
      source_lang_panel.add(sp);

      source_lang_panel.add(new JLabel("\n"));
      JLabel lang_label = new JLabel("Select language(s) for display.");
      source_lang_panel.add(lang_label);
      Language[] sorted_languages = Language.getLanguages();
      Arrays.sort(sorted_languages);
      language_list = new SuperJList(sorted_languages);
      language_list
          .setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
      if (!display_languages.isEmpty()) {
        Language[] language_array = (Language[]) display_languages
            .toArray(new Language[0]);
        language_list.setSelectedValues(language_array, false);
      }

      JScrollPane lsp = new JScrollPane(language_list,
          ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED,
          ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
      source_lang_panel.add(lsp);

      // Add the atom comparator display
      atom_comparator_display = new JComboBox();
      atom_comparator_display.setEditable(false);
      for (int i = 0; i < comparators.length; i++) {
        atom_comparator_display.addItem(comparators[i]);
      }
      if (atom_comparator == null) {
        atom_comparator_display.setSelectedItem(comparators[2]);
        setAtomComparator(comparators[2]);
      } else
        atom_comparator_display.setSelectedItem(atom_comparator);
      source_lang_panel.add(new JLabel("\n"));
      JLabel comp_label = new JLabel("Select atom sorting algorithm.");
      source_lang_panel.add(comp_label);
      source_lang_panel.add(atom_comparator_display);

      // Add the page size display
      page_size_display = new JComboBox();
      page_size_display.setEditable(false);
      String[] page_size_choices = { "50", "100", "200", "500", "1000" };
      for (int i = 0; i < page_size_choices.length; i++) {
        page_size_display.addItem(page_size_choices[i]);
      }
      if (page_size == 0) {
        page_size_display.setSelectedItem(page_size_choices[1]);
        setPageSize(new Integer(page_size_choices[1]).intValue());
      } else
        page_size_display.setSelectedItem(new String(new Integer(page_size)
            .toString()));

      source_lang_panel.add(new JLabel("\n"));
      JLabel page_label = new JLabel("Select page size for display.");
      source_lang_panel.add(page_label);
      source_lang_panel.add(page_size_display);

      // configure done_panel
      FlowLayout done_layout = new FlowLayout();
      done_layout.setVgap(10);
      JPanel done_panel = new JPanel(done_layout);
      JPanel complete_panel = new JPanel(new BorderLayout(10, 10));

      JButton jdone = new JButton("Done");
      jdone.addActionListener(dcl);
      done_panel.add(jdone);

      // Create an action object
      Action cancelAction = new AbstractAction("Cancel") {
        {
          // Set action command to "Cancel"
          putValue(Action.ACTION_COMMAND_KEY, getValue(Action.NAME));
        }

        // This method is called when the action is invoked
        public void actionPerformed(ActionEvent e) {
          dcl.actionPerformed(e);
        }
      };
      JButton jcancel = new JButton(cancelAction);
      // Register keystroke
      done_panel.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(
          KeyStroke.getKeyStroke("ESCAPE"), cancelAction.getValue(Action.NAME));
      // Register action
      done_panel.getActionMap().put(cancelAction.getValue(Action.NAME),
          cancelAction);
      done_panel.add(jcancel);

      // Add the sub-panels to complete_panel
      complete_panel.add(source_lang_panel, BorderLayout.CENTER);
      complete_panel.add(done_panel, BorderLayout.SOUTH);
      EtchedBorder etched_border = new EtchedBorder(EtchedBorder.RAISED);
      complete_panel.setBorder(new CompoundBorder(etched_border,
          new EmptyBorder(10, 10, 10, 10)));

      getContentPane().add(complete_panel, BorderLayout.CENTER);
      getRootPane().setDefaultButton(jdone);
      this.pack();
      this.setLocationRelativeTo(null);
    };

    //
    // Inner classes
    //
    public class DataChangeListener implements java.awt.event.ActionListener {

      // constructor
      public DataChangeListener() {
      };

      // implementation of ActionListener interface
      public void actionPerformed(ActionEvent e) {

        if (e.getActionCommand().equals("Done")) {

          display_sources.clear();
          Object[] selected = source_list.getSelectedValues();
          for (int i = 0; i < selected.length; i++) {
            display_sources.add((Source) selected[i]);
          }
          display_languages.clear();
          selected = language_list.getSelectedValues();
          for (int i = 0; i < selected.length; i++) {
            display_languages.add((Language) selected[i]);
          }

          // make and set the restrictor
          ChainRestrictor restrictor = new ChainRestrictor();
          if (!display_sources.isEmpty()) {
            restrictor.add(new SourceElement.BySourceRestrictor(
                (Source[]) display_sources.toArray(new Source[0])));

            restrictor.add(new AtomElement.BySourceRestrictor(
                (Source[]) display_sources.toArray(new Source[0])));
          }
          if (!display_languages.isEmpty()) {
            restrictor.add(new AtomElement.ByLanguageRestrictor(
                (Language[]) display_languages.toArray(new Language[0])));

            restrictor.add(new Atom.ByLanguageRestrictor(
                (Language[]) display_languages.toArray(new Language[0])));
          }
          restrictor.setComparator(new SourceElement.BySourceComparator());
          ExpandableConceptView.this.setRestrictor(restrictor);
          ExpandableConceptView.this
              .setAtomComparator((Comparator) atom_comparator_display
                  .getSelectedItem());
          ExpandableConceptView.this.setPageSize(new Integer(
              (String) page_size_display.getSelectedItem()).intValue());

          if (concept != null) {
            RRFBrowserGUI.getRRFBrowserGUI().enableGlassPane();

            Thread t = new Thread(new Runnable() {
              public void run() {

                clm.refresh();
                RRFBrowserGUI.getRRFBrowserGUI().disableGlassPane();
              }
            });
            t.start();

          }
          ExpandableConceptViewOptionsGUI.this.dispose();
        } else if (e.getActionCommand().equals("Cancel"))
          dispose();
      }
    }
  } // End of ExpandableConceptViewOptionsGUI class

}
