/*****************************************************************************
 * Package: gov.nih.nlm.mms.browser
 * Object:  FinderManager.java
 *
 ****************************************************************************/
package gov.nih.nlm.mms.browser;

import gov.nih.nlm.mms.GUIConfigurable;
import gov.nih.nlm.mms.filters.Filter;

import java.awt.BorderLayout;
import java.awt.Component;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.swing.JPanel;
import javax.swing.JTabbedPane;
import javax.swing.SwingUtilities;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

/**
 * Manages the {@link Finder}s.
 *
 * @author Deborah Shapiro
 */
public class FinderManager implements ChangeListener {

  //
  // Finder set
  //
  private final Set finders = new HashSet(10);

  //
  // Current finder
  //
  private Finder current_finder;

  //
  // Concept list manager
  //
  private ConceptListManager clm;

  //
  // Tabbed pane to be displayed on the FinderManager - holds Finders
  //
  private final JTabbedPane tabbed_pane = new JTabbedPane();

  //
  // Map to get a display component from a Finder
  //
  private final Map gc_to_component_map = new HashMap();

  //
  // Map to get a Finder from the display component
  //
  private final Map component_to_gc_map = new HashMap();

  //
  // Constructor
  //

  /**
   * Instantiates an {@link FinderManager}.
   */
  public FinderManager() { }

  /**
   * Adds a {@link Finder} to the {@link FinderManager}.
   * Provides another search technique to the user.
   * @param finder {@link Finder}
   */
  public void addFinder(Finder finder) {
    finders.add(finder);
    finder.setConceptListManager(clm);
  }

  /**
   * Returns a {@link JPanel} with a {@link JTabbedPane}
   * displaying all of the {@link FinderManager}'s {@link Finder}s.
   * @return {@link JPanel}
   */
  public JPanel getPanel() {
    JPanel finder_manager_panel = new JPanel(new BorderLayout());
    tabbed_pane.addChangeListener(this);
    Finder[] finder_array = (Finder[])finders.toArray(new Finder[0]);
    Arrays.sort(finder_array);
    int initial_index = 0;
    for (int i=0; i<finder_array.length; i++) {
      Finder this_finder = (Finder)finder_array[i];
      tabbed_pane.add(getComponentForTab(this_finder), this_finder.getName());
      if (this_finder.getClass().getName().endsWith("WordSearchFinder"))
        initial_index = i;
    }
    finder_manager_panel.add(tabbed_pane, BorderLayout.CENTER);
    tabbed_pane.setSelectedIndex(initial_index);
    return finder_manager_panel;
  }

  /**
   * Returns the GUI {@link Component} for the specified {@link Filter}.
   * The application maintains a private hash map that allows
   * this mapping to take place.  If the map does not contain
   * an entry for the filter, a panel is built, added to the map
   * and returned.
   * @param gui_configurable the specified {@link GUIConfigurable}
   * @return the GUI {@link Component} for the specified {@link Filter}.
   */
  public Component getComponentForTab(GUIConfigurable gui_configurable) {
    if (gc_to_component_map.containsKey(gui_configurable)) {
      return (Component)gc_to_component_map.get(gui_configurable);
    }

    //
    // Create the panel for the whole tab
    //
    JPanel tab_panel = new JPanel(new BorderLayout());

    //
    // Create the panel container for the GUIConfigurable panel and add it
    //
    JPanel gui_panel = new JPanel(new BorderLayout());
    gui_panel.add(gui_configurable.getPanel(), BorderLayout.CENTER);

    tab_panel.add(gui_panel, BorderLayout.CENTER);

    // add gc and component to maps & return
    gc_to_component_map.put(gui_configurable, tab_panel);
    component_to_gc_map.put(tab_panel, gui_configurable);
    return tab_panel;
  }


  /**
   * Returns the {@link Finder} that is currently being
   * displayed on the {@link JTabbedPane}.
   * @return {@link Finder}
   */
  public Finder getCurrentFinder() {
    return current_finder;
  }

  /**
   * Sets the {@link ConceptListManager} for the {@link RRFBrowserGUI}.
   * @param clm {@link ConceptListManager}
   */
  public void setConceptListManager(ConceptListManager clm) {
    this.clm = clm;
  }

  /**
   * Resets the graphical components of the {@link Finder}s.
   */
  public void clear() {
    Iterator iter = finders.iterator();
    while (iter.hasNext()) {
      ((Finder)iter.next()).resetToDefaults();
    }
  }

  /**
   * Returns the {@link Set} of {@link Finder}s.
   * @return {@link Set}
   */
  public Set getFinders() {
    return finders;
  }

  /**
   * Responds to a user switching tabs in the tabbed pane.
   * Resets the current view.
   * @param e a {@link ChangeEvent}
   */
  public void stateChanged(ChangeEvent e) {
    Component c = ((JTabbedPane)e.getSource()).getSelectedComponent();
    current_finder = ((Finder)component_to_gc_map.get(c));
  }
  
  /**
   * Notifies all the {@link Finder}s of the font family
   * to be used for text display.
   * @param font_family {@link String}
   */
  public void setFontFamily(final String font_family) {
  	Finder[] l_finders = (Finder[])finders.toArray(new Finder[0]);
    for (int i=0; i<l_finders.length; i++) {
      final Finder finder = l_finders[i];
      finder.setFontFamily(font_family);
      finder.refresh(); 
    }
  }
  
  /**
   * Notifies all the {@link Finder}s of the font point size
   * to be used for text display.
   * @param font_point_size int
   */
  public void setFontPointSize(final int font_point_size) {
  	Finder[] l_finders = (Finder[])finders.toArray(new Finder[0]);
  	for (int i=0; i<l_finders.length; i++) {
      final Finder finder = l_finders[i];
      finder.setFontPointSize(font_point_size);
      finder.refresh();
  	}
  }
}
