/*****************************************************************************
 * Package: gov.nih.nlm.mms.browser
 * Object:  JEditorPaneFinderHandler.java
 *
 ****************************************************************************/
package gov.nih.nlm.mms.browser;

import gov.nih.nlm.swing.FindDialog;
import gov.nih.nlm.util.FinderHandler;

import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.JDialog;
import javax.swing.JEditorPane;
import javax.swing.text.Document;

/**
 * Handles a find operation on a {@link JEditorPane}.
 *
 */
public class JEditorPaneFinderHandler implements FinderHandler{

  /** JEditorPane containing the document to search in */
  private JEditorPane editor;

  /** search text from the document */
  private String searchText;

  /** search phrase to find */
  private String phrase;

  /** last start position, the search phrase was found at in the document */
  private int lastPos;

  /** two fields to correct position differences during replace operations */
  private int offset;

  /** indicates if a find is already in progress */
  private boolean findInProgress = false;

  /** gui that displays find options */
  private FindDialog dialog = null;

  /**
   * Constructor.
   * @param editor_pane {@link JEditorPane}
   */
  public JEditorPaneFinderHandler(JEditorPane editor_pane) {
    setEditor(editor_pane);
  }

  /**
   * Sets the {@link JDialog} for the handler.
   * @param find_dialog {@link JDialog}
   */
  public void setDialog(FindDialog find_dialog) {
    dialog = find_dialog;
  }

  /**
   * Set the JEditorPane holding the document to be searched
   * @param editor  the JEditorPane holding the document to be searched
   */
  public void setEditor(JEditorPane editor) {
    this.editor = editor;
    editor.getCaret().setSelectionVisible(true);

  }


  /* --------- Find implementation start ------ */

  /**
   * Indicates the find operation has been cancelled on the gui.
   */
  public void cancel() {
    findInProgress = false;
  }
  
  /**
   * Allows the handler to request focus.
   *
   */  
  public void requestFocus() {
  	editor.requestFocus();
  }

  /**
   * Initialize a find operation by reading all relevant settings and
   * locking the dialog window.
   * @throws Exception
   */
  public void init() throws Exception {
    if (!findInProgress) {
      searchText = editor.getDocument().getText(0, editor.getDocument().getLength());
      phrase = dialog.getSearchField();
      offset = 0;
      if (!dialog.isMatchCase()) {
        phrase = phrase.toLowerCase();
        searchText = searchText.toLowerCase();
      }
      if (dialog.isStartOnTop()) {
        if (dialog.isSearchUp()) {
          lastPos = editor.getDocument().getLength();
        }
        else {
          lastPos = 0;
        }
      }
      else {
        lastPos = editor.getSelectionStart();
      }
    }
  }

  /**
   * Initiate a find or find next operation, whatever applies. If no (more)
   * hits are found, a message is displayed and the dialog is unlocked for a
   * new search operation.
   * @throws Exception
   */
  public void find() throws Exception {
    findInProgress = true;

    if (!dialog.isSearchUp()) {

        boolean result = doFind();
        if (result) {
          return;
        }
        else {
          lastPos = 0;
          searchText = editor.getDocument().getText(0, editor.getDocument().getLength());
          if (!dialog.isMatchCase()) {
            phrase = phrase.toLowerCase();
            searchText = searchText.toLowerCase();
          }
        }

    }
    else {

        boolean result = doFind();
        if (result) {
          return;
        }
        else {
          lastPos = editor.getDocument().getLength();
          searchText = editor.getDocument().getText(0, editor.getDocument().getLength());
          if (!dialog.isMatchCase()) {
            phrase = phrase.toLowerCase();
            searchText = searchText.toLowerCase();
          }
        }
      }


    dialog.terminateOperation();
    findInProgress = false;
  }

  /**
   * Look for the next occurrence of the search phrase either as whole word
   * or as part of a word, depending on the current dialog setting.
   *
   * <p>If the phrase is found (again), its position is 'remembered' for a
   * possible findNext and its postion is highlighted in the underlying
   * <code>JEditorPane</code>.
   *
   * @return  true, if the phrase was found (again), false if not
   *
   * @see javax.swing.JEditorPane
   */
  public boolean doFind() {
    boolean found = false;
    int start = findNext();

    if (start > 0) {
      lastPos = start;
      if (dialog.isSearchDown()) {
        start += offset;
      }

      found = true;

      // highlight selection
      editor.scrollRectToVisible(new Rectangle(
          new Point(0, (int)editor.getVisibleRect().getLocation().getY())));

      editor.setSelectionStart(start);
      editor.setSelectionEnd(start + phrase.length());
      editor.repaint();
    }

    return found;
  }

  /**
   * Find the next occurrence of a searched phrase from the last position
   * either up or down.
   *
   * @return the start position of the next occurrence or
   *                          0 if no more hits were found
   */
  private int findNext() {
    int start;
    if (dialog.isSearchUp()) {
      if (lastPos < editor.getDocument().getLength()) {
        start = searchText.lastIndexOf(phrase, lastPos - 1);
      }
      else {
        start = searchText.lastIndexOf(phrase, lastPos);
      }
    }
    else {
      if (lastPos > 0) {
        start = searchText.indexOf(phrase, lastPos + phrase.length());
      }
      else {
        start = searchText.indexOf(phrase, lastPos);
      }
    }
    return start;
  }

  /* ----------- Find implementation end ------- */

}
