/**************************************************************************
 *
 * Package:	gov.nih.nlm.mms.browser
 * Object:	RRFBrowserGUI.java
 *
 **************************************************************************/
package gov.nih.nlm.mms.browser;

import gov.nih.nlm.mms.AboutAction;
import gov.nih.nlm.mms.ApplicationConfiguration;
import gov.nih.nlm.mms.GUIConfigurable;
import gov.nih.nlm.mms.InstallConfiguration;
import gov.nih.nlm.mms.LocalException;
import gov.nih.nlm.mms.MMSToolkit;
import gov.nih.nlm.mms.RichMRMetamorphoSysInputStream;
import gov.nih.nlm.mms.UserConfiguration;
import gov.nih.nlm.swing.FindDialog;
import gov.nih.nlm.swing.GlassComponent;
import gov.nih.nlm.swing.GlassPaneListener;
import gov.nih.nlm.umls.io.FileStatisticsConfiguration;
import gov.nih.nlm.umls.io.IndexBuilder;
import gov.nih.nlm.umls.io.InputStreamFileStatisticsConfiguration;
import gov.nih.nlm.umls.io.SuiCountIndexBuilder;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.MetaDescriptionInitializer;
import gov.nih.nlm.util.FileStatistics;
import gov.nih.nlm.util.FinderHandler;
import gov.nih.nlm.util.PrintPreviewAction;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.Iterator;
import java.util.Properties;
import java.util.Set;

import javax.swing.Action;
import javax.swing.Box;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JSplitPane;
import javax.swing.JToolBar;
import javax.swing.KeyStroke;
import javax.swing.ProgressMonitor;
import javax.swing.SwingUtilities;
import javax.swing.border.EmptyBorder;

/**
 * GUI application for viewing the data in a MetamorphoSys subset. User may use
 * the {@link Finder} panels to search for {@link Concept}s given various
 * search criteria. Search results are displayed in the {@link ReportsManager}.
 * 
 * @author Deborah Shapiro
 */
public class RRFBrowserGUI extends JFrame implements GlassPaneListener {

  //
  // Private Fields
  //

  //
  // References to the configuration objects
  //
  private ApplicationConfiguration default_configuration = null;

  private UserConfiguration user_configuration = null;

  //
  // Concept list manager
  //
  private ConceptListManager clm = null;

  //
  // Currently open subset dir
  //
  private File subset_dir = null;

  //
  // Finder manager
  //
  private FinderManager finder_manager;

  //
  // Reports manager
  //
  private ReportsManager reports_manager;

  //
  // GUI components
  //
  private JSplitPane split_pane;

  private JButton open_subset_button;

  private JButton print_button;

  private JButton print_preview_button;

  private PrintAction print_action;

  private JButton backward_button;

  private JButton forward_button;

  private JComboBox subset_dir_display;
  
  private JComboBox font_family_display;
  
  private JComboBox font_size_display;

  private JMenu help_menu = null;

  private JMenuItem print_menu_item = null;

  private JMenuItem print_preview_menu_item = null;

  private JMenuItem find_menu_item = null;

  private JMenuItem increase_font_menu_item = null;

  private JMenuItem decrease_font_menu_item = null;

  private JMenu change_font_menu = null;

  private JMenuItem[] font_menu_items = null;

  //
  // For glass pane
  //
  private GlassComponent gc = null;

  //
  // Static singleton instance
  //
  private static RRFBrowserGUI rrf_browser_gui = null;

  //
  // For building index files automatically
  //
  private FileStatisticsConfiguration fs_config;

  private static FindDialog find_dialog;

  //
  //
  private boolean ignore_events = false;

  private static Component parent_gui;
  
  private static int[] sizes = {8, 10, 11, 12, 14, 16, 18, 20, 24, 27};
  
  

  //
  // Constructor
  //

  /**
   * Instantiates a {@link RRFBrowserGUI} and builds the frame.
   */
  private RRFBrowserGUI() {
    super();

    try {
      fs_config = new InputStreamFileStatisticsConfiguration(
          (InputStream) RRFBrowserGUI.class.getClassLoader().getResource(
              "config/mrpluscolsfiles.dat").getContent());
    } catch (Exception e) {
      e.printStackTrace();
      MMSToolkit.reportError("Unexpected error loading mrpluscolsfiles.dat. \n"
          + "See the console for more details.", true);
    }

    find_dialog = new FindDialog(this);

    //
    // Handle glass pane
    //
    gc = new GlassComponent(this);
    setGlassPane(gc);

    //
    // Get configuration references
    //

    try {
      default_configuration = new ApplicationConfiguration(new File("config",
          "mmsys.a.prop"));
      user_configuration = new UserConfiguration(new File("config",
          "mmsys.a.prop"));
      user_configuration.setInputStream(new RichMRMetamorphoSysInputStream());
    } catch (LocalException le) {
      le.printStackTrace();
      MMSToolkit.reportError(this, "Failed to load mmsys.a.prop.", false);
      find_dialog.dispose();
      MMSToolkit.exitToWelcomeGUI(RRFBrowserGUI.getRRFBrowserGUI());
    }

    initialize();

  }
  
  /**
   * Returns the singleton reference to the {@link RRFBrowserGUI}.
   * 
   * @return the singleton reference to the {@link RRFBrowserGUI}
   */
  public static RRFBrowserGUI getRRFBrowserGUI() {
    if (rrf_browser_gui == null)
      rrf_browser_gui = new RRFBrowserGUI();
    return rrf_browser_gui;
  }

  public static RRFBrowserGUI newRRFBrowserGUI() {
  	if (rrf_browser_gui != null)
  		rrf_browser_gui.dispose();
    rrf_browser_gui = new RRFBrowserGUI();
    return rrf_browser_gui;
  }
  
  /**
   * Returns the singleton reference to the {@link RRFBrowserGUI}. Used when
   * browser is called from another application.
   * 
   * @param parent
   *          parent gui from which the RRF Browser is requested
   * @return the singleton reference to the {@link RRFBrowserGUI}
   */
  public static RRFBrowserGUI getRRFBrowserGUI(Component parent) {
    parent_gui = parent;
    return getRRFBrowserGUI();
  }

  public static FindDialog getFindDialog() {
    return find_dialog;
  }

  /**
   * Initializes the GUI.
   */
protected void initialize() {

    final JPanel base_panel = new JPanel(new BorderLayout());
    
    //
    // Build and configure tabbed pane
    // The change listener resonds to users switching tabs
    //
    split_pane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);
    split_pane.setBorder(new EmptyBorder(20, 20, 20, 20));

    //
    // Menu Bar with top-level menus
    //
    final JMenuBar menu_bar = new JMenuBar();
    setJMenuBar(menu_bar);
    final JMenu file_menu = new JMenu("File");
    menu_bar.add(file_menu);
    final JMenu edit_menu = new JMenu("Edit");
    menu_bar.add(edit_menu);
    final JMenu options_menu = new JMenu("Options");
    menu_bar.add(options_menu);
    menu_bar.add(Box.createHorizontalGlue());
    help_menu = new JMenu("Help");
    menu_bar.add(help_menu);

    //
    // File Menu
    //

    //
    // Open Subset for Browsing
    //
    JMenuItem open_subset_menu_item = new JMenuItem(new OpenSubsetAction(
        user_configuration, clm));
    if (open_subset_menu_item.getIcon() != null) {
      open_subset_menu_item.setIcon(null); // a text only menu item
    }
    open_subset_menu_item.setAccelerator(KeyStroke.getKeyStroke('O',
        java.awt.Event.CTRL_MASK, false));
    file_menu.add(open_subset_menu_item);

    //
    // Print Concept Report
    //
    print_menu_item = new JMenuItem(new PrintAction());
    if (print_menu_item.getIcon() != null) {
      print_menu_item.setIcon(null); // a text only menu item
    }
    print_menu_item.setAccelerator(KeyStroke.getKeyStroke('P',
        java.awt.Event.CTRL_MASK, false));
    file_menu.add(print_menu_item);

    //
    // Print Preview
    //
    print_preview_menu_item = new JMenuItem(new ConceptViewPrintPreviewAction(
        "config/PrintPreview24.gif", print_action));
    if (print_preview_menu_item.getIcon() != null) {
      print_preview_menu_item.setIcon(null); // a text only menu item
    }
    file_menu.add(print_preview_menu_item);

    //
    // Exit application
    //
    JMenuItem exit_menu_item = new JMenuItem(new ExitRRFAction());
    exit_menu_item.setAccelerator(KeyStroke.getKeyStroke('Q',
        java.awt.Event.CTRL_MASK, false));
    file_menu.add(exit_menu_item);

    //
    // Edit Menu
    //

    //
    // Find
    //
    find_menu_item = new JMenuItem(new FindAction());
    if (find_menu_item.getIcon() != null) {
      find_menu_item.setIcon(null); // a text only menu item
    }
    find_menu_item.setAccelerator(KeyStroke.getKeyStroke('F',
        java.awt.Event.CTRL_MASK, false));
    edit_menu.add(find_menu_item);

    //
    // Increase Font Size
    //
    increase_font_menu_item = new JMenuItem("Increase Font Size");
    if (increase_font_menu_item.getIcon() != null) {
      increase_font_menu_item.setIcon(null); // a text only menu item
    }
    increase_font_menu_item.setAccelerator(KeyStroke.getKeyStroke(']',
        java.awt.Event.CTRL_MASK,false));
    increase_font_menu_item.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
    	if (!ignore_events) {
    	ignore_events = true;
        enableGlassPane();
        Thread t = new Thread (new Runnable() {
          public void run() {
            int font_point_size = BrowserState.getBrowserState().getFontPointSize();
            for (int i=0; i<sizes.length; i++) {
            	if (font_point_size == sizes[i] && (i != sizes.length -1)) {
            		font_point_size = sizes[i+1];
            		break;
            	}
            }
            setFontSize(font_point_size);
            disableGlassPane();
            ignore_events = false;
          }
        });
        t.start();
      }
      }
    });
    edit_menu.add(increase_font_menu_item);

    //
    // Decrease Font Size
    //
    decrease_font_menu_item = new JMenuItem("Decrease Font Size");
    if (decrease_font_menu_item.getIcon() != null) {
      decrease_font_menu_item.setIcon(null); // a text only menu item
    }
    decrease_font_menu_item.setAccelerator(KeyStroke.getKeyStroke('[',
        java.awt.Event.CTRL_MASK, false));
    decrease_font_menu_item.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
    	if (!ignore_events) {
    	ignore_events = true;
        enableGlassPane();
        Thread t = new Thread(new Runnable() {
          public void run() {
            int font_point_size = BrowserState.getBrowserState()
                .getFontPointSize();
            for (int i=0; i<sizes.length; i++) {
            	if (font_point_size == sizes[i] && (i != 0)) {
            		font_point_size = sizes[i-1];
            		break;
            	}
            }
            setFontSize(font_point_size);
            disableGlassPane();
            ignore_events = false;
          }
        });
        t.start();
      }
      }
    });
    edit_menu.add(decrease_font_menu_item);

    //
    // Change Font
    //
    change_font_menu = new JMenu("Change Font");
    if (change_font_menu.getIcon() != null) {
      change_font_menu.setIcon(null); // a text only menu item
    }
    edit_menu.add(change_font_menu);

    //
    // Font sub-menus
    //
    String[] fonts = {"Monospaced", "Serif", "SansSerif"};
    font_menu_items = new JMenuItem[fonts.length];
    for (int i = 0; i < fonts.length; i++) {
      font_menu_items[i] = new JMenuItem(fonts[i]);
      final JMenuItem font_menu_item = font_menu_items[i];
      final String final_font = fonts[i];
      font_menu_item.setEnabled(!fonts[i].equals(BrowserState.getBrowserState()
          .getFontFamily()));
      font_menu_item.setIcon(null);
      font_menu_item.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          ignore_events = true;
          enableGlassPane();
          Thread t = new Thread(new Runnable() {
            public void run() {
              setFontStyle(final_font);
              disableGlassPane();
              ignore_events = false;
            }
          });
          t.start();
        }
      });
      change_font_menu.add(font_menu_item);
    }

    
    //
    // Help Menu
    //

    //
    // About the application
    //
    JMenuItem about_menu_item = new JMenuItem(new AboutAction(
        "About RRF Browser", this));
    help_menu.add(about_menu_item);

    //
    // RRF Browser Documentation menu item
    //
    help_menu.add(MMSToolkit.getURLHelpMenu("RRF Browser Documentation  ",
        "http://www.nlm.nih.gov/research/umls/rrf_help.html", this));

    //
    // Add the FinderManager and ReportsManager panels
    //
    finder_manager = new FinderManager();
    reports_manager = new ReportsManager();

    //
    // Set the ConceptListManager
    //
    clm = new GUIConceptListManager();
    clm.setConfiguration(user_configuration);
    clm.setReportsManager(reports_manager);
    finder_manager.setConceptListManager(clm);
    reports_manager.setConceptListManager(clm);

    //
    // Find and instantiate all the Finders
    //
    String[] finders = null;
    try {
      finders = MMSToolkit
          .getInterfaceImplementations("gov.nih.nlm.mms.browser.Finder");
    } catch (ClassNotFoundException cnfe) {
      //
      // this exception will never happen b/c the list to choose from consists
      // of classes that are known to exist
      //
      cnfe.printStackTrace();
    } catch (IOException ie) {
      ie.printStackTrace();
      MMSToolkit
          .reportError(this,
              "Unexpected error while looking for installed finder formats.\n"
                  + "This is a serious error that cannot be recovered from.",
              false);
      find_dialog.dispose();
      MMSToolkit.exitToWelcomeGUI(RRFBrowserGUI.getRRFBrowserGUI());
    }
    
    //
    // Add finders to findermanager, get finder options menus
    //
    for (int i = 0; i < finders.length; i++) {
      Finder this_finder = (Finder) MMSToolkit.instantiateObject(finders[i]);
      this_finder.setConfiguration(default_configuration);
      this_finder.setConfiguration(user_configuration);
      finder_manager.addFinder(this_finder);

      if (this_finder.hasAdvancedOptions())
        options_menu.add(getOptionsMenuItem(this_finder));
    }

    //
    // Find and instantiate all the ConceptViews
    //
    String[] views = null;
    try {
      views = MMSToolkit
          .getInterfaceImplementations("gov.nih.nlm.mms.browser.ConceptView");
    } catch (ClassNotFoundException cnfe) {
      //
      // this exception will never happen b/c the list to choose from consists
      // of classes that are known to exist
      //
      cnfe.printStackTrace();
    } catch (IOException ie) {
      ie.printStackTrace();
      MMSToolkit
          .reportError(this,
              "Unexpected error while looking for installed concept view formats.\n"
                  + "This is a serious error that cannot be recovered from.",
              false);
      find_dialog.dispose();
      MMSToolkit.exitToWelcomeGUI(RRFBrowserGUI.getRRFBrowserGUI());
    }
    
    //
    // Add concept views to manager, pick up options menus
    //
    for (int i = 0; i < views.length; i++) {
      ConceptView this_cv = (ConceptView) MMSToolkit
          .instantiateObject(views[i]);
      this_cv.setConfiguration(default_configuration);
      this_cv.setConfiguration(user_configuration);
      reports_manager.addConceptView(this_cv);

      if (this_cv.hasAdvancedOptions())
        options_menu.add(getOptionsMenuItem(this_cv));
    }

    //
    // Tool Bar
    //

    //
    // Add the buttons
    //
    // Add the Open Subset button
    JToolBar tool_bar = new JToolBar();
    open_subset_button = new JButton(new OpenSubsetAction(user_configuration, clm));
    if (open_subset_button.getIcon() != null) {
      open_subset_button.setText(""); // an icon-only button
    }
    tool_bar.add(open_subset_button);

    // Add the Print button
    print_action = new PrintAction();
    print_button = new JButton(print_action);
    if (print_button.getIcon() != null) {
      print_button.setText(""); // an icon-only button
    }
    tool_bar.add(print_button);

    // Add the Print preview button
    print_preview_button = new JButton(new ConceptViewPrintPreviewAction(
        "config/PrintPreview24.gif", print_action));
    if (print_preview_button.getIcon() != null) {
      print_preview_button.setText(""); // an icon-only button
    }
    tool_bar.add(print_preview_button);

    // Add the Backward button
    backward_button = new JButton(new BackwardAction(clm));
    if (backward_button.getIcon() != null) {
      backward_button.setText(""); // an icon-only button
    }
    tool_bar.add(backward_button);

    // Add the Forward button
    forward_button = new JButton(new ForwardAction(clm));
    if (forward_button.getIcon() != null) {
      forward_button.setText(""); // an icon-only button
    }
    tool_bar.add(forward_button);
    tool_bar.addSeparator(new Dimension(20, 10));
    
    // Add the font style display
    font_family_display = new JComboBox();
    font_family_display.setEditable(false);
    font_family_display.setEnabled(false);
    font_family_display.setMaximumSize(new Dimension(120, 25));
    font_family_display.setToolTipText("Font Style");
    for (int i=0; i<fonts.length; i++) {
    	font_family_display.addItem(fonts[i]);
    }
    font_family_display.setSelectedItem(BrowserState.getBrowserState().getFontFamily());
    font_family_display.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        final Object family = font_family_display.getSelectedItem();
        if (family != null && !ignore_events) {
          ignore_events = true;
          enableGlassPane();
          Thread t = new Thread(new Runnable() {
            public void run() { 
              setFontStyle((String)family);
              disableGlassPane();
              ignore_events = false;
            }
          });
          t.start();
        }
      }
    });
    tool_bar.add(font_family_display);
    tool_bar.addSeparator(new Dimension(10, 10));
    
    // Add the font size display
    font_size_display = new JComboBox();
    font_size_display.setEditable(false);
    font_size_display.setEnabled(false);
    font_size_display.setMaximumSize(new Dimension(40, 25));
    font_size_display.setToolTipText("Font Point Size");
    for (int i=0; i<sizes.length; i++) {
    	font_size_display.addItem(new Integer(sizes[i]));
    }
    font_size_display.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        final Object size = font_size_display.getSelectedItem();
        if (size != null && !ignore_events) {
          ignore_events = true;
          enableGlassPane();
          Thread t = new Thread(new Runnable() {
            public void run() {
              setFontSize(((Integer)size).intValue());
              disableGlassPane();
              ignore_events = false;
            }
          });
          t.start();
        }
      }
    });
    tool_bar.add(font_size_display);
    
    // Add the subset dir display
    subset_dir_display = new JComboBox();
    subset_dir_display.setEditable(false);
    subset_dir_display.setEnabled(false);
    subset_dir_display.setMaximumSize(new Dimension(500, 25));
    subset_dir_display.setToolTipText("Subset Directory");
    subset_dir_display.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        final Object dir = subset_dir_display.getSelectedItem();
        if (dir != null && !ignore_events) {
          enableGlassPane();
          Thread t = new Thread(new Runnable() {
            public void run() {
              subsetDirectoryOpened(new File(dir.toString()));
              disableGlassPane();
            }
          });
          t.start();
        }      	
      }
    });

    tool_bar.addSeparator(new Dimension(100, 10));
    tool_bar.add(subset_dir_display);
    tool_bar.addSeparator(new Dimension(20, 10));

    //
    // Set the base content pane
    //
    JPanel left_component = finder_manager.getPanel();
    left_component.setPreferredSize(new Dimension(300, 400));
    split_pane.setLeftComponent(left_component);
    JPanel right_component = reports_manager.getPanel();
    right_component.setPreferredSize(new Dimension(500, 400));
    split_pane.setRightComponent(right_component);
    base_panel.add(tool_bar, BorderLayout.NORTH);
    base_panel.add(split_pane, BorderLayout.CENTER);

    setEnableButtons();

    //
    // Configure frame
    //
    this.setTitle("Rich Release Format Browser "
        + default_configuration.getReleaseVersion());
    this.setResizable(true);
    this.setContentPane(base_panel);

    //
    // Configure window listener
    //
    addWindowListener(new WindowAdapter() {
      public void windowClosing(WindowEvent e) {
      	find_dialog.dispose();
        MMSToolkit.exitToWelcomeGUI(RRFBrowserGUI.getRRFBrowserGUI());
      }
    });

    //
    // pack and make visible
    //
    this.pack();
    this.setLocationRelativeTo(null);

    find_dialog.centerDialog(this);

  }

/**
 * Check the previous state, if it exists, offer to load it
 * for the user.
 *
 */
  protected void checkPreviousState() {
    
    //
    // Check if the browser state was previously saved, offer to reload it
    //
    BrowserState in_object = null;
    InstallConfiguration ic = InstallConfiguration.getInstallConfiguration();

    //
    // Use dir_saved.txt file
    //
    if (new File("config", "dir_saved.txt").exists()) {

      boolean open_dir_saved = true;
      if (ic.isQueryUserToOpenSavedConfig() && subset_dir == null)
        open_dir_saved = MMSToolkit.confirmRequest(parent_gui,
            "Would you like to load the most \n" + 
            "recently viewed concepts?");

      if (open_dir_saved) {
        try {
          // read in BrowserState object from dir_saved.txt
          File file = new File("config", "dir_saved.txt");
          in_object = BrowserState.getBrowserState(file);

        } catch (Exception e) {
          MMSToolkit.reportError(this,
              "Unexpected error reading dir_saved.txt file.\n"
                  + "Subset Directory must be chosen using the \n"
                  + "File->Open menu ", false);
        }
      }
    }
    
    
    //
    // Do nothing here, start clean
    //
    else if (ic.isQueryUserToOpenSavedConfig()) {
      // user doesn't want to use dir_saved.txt and it will be deleted
      new File("config", "dir_saved.txt").delete();
    } 
    
   
    //
    // If user wants to use previous config, load it here and call
    // setDirectoryOpened
    //
    if (in_object != null) {
      try {
        // get the current directory
        final File subset_file = new File(in_object.getCurrentDirectory());
        if (subset_file.exists()) {
          enableGlassPane();
          Thread t = new Thread(new Runnable() {
            public void run() {
              subsetDirectoryOpened(subset_file);
              disableGlassPane();
            }
          });
          t.start();
        } else
          new File("config", "dir_saved.txt").delete();

      } catch (Exception e) {
        e.printStackTrace();
        MMSToolkit.reportError(this,
            "Unexpected error opening subset directory", false);
      }
    }
        
  }

  //
  // Methods
  //
  /**
   * Returns the currently selected subset directory.
   * 
   * @return {@link File}
   */
  public File getSubsetDirectory() {
    return subset_dir;
  }

  /**
   * Logic for control of enabling and disabling gui buttons.
   */
  public void setEnableButtons() {
    backward_button.setEnabled(!clm.isBackHistoryEmpty());
    forward_button.setEnabled(!clm.isForwardHistoryEmpty());
    setSplitPaneEnabled(subset_dir != null);
    setPrintComponentsEnabled(clm.getCurrentConcept() != null);
    enableEditMenuItems(subset_dir != null);
  }

  /**
   * Handles opening of a source directory.
   * 
   * @param source_dir
   *          the source directory
   */
  public void subsetDirectoryOpened(File source_dir) {
    //
    // Set instance variable tracking subset dir
    //
    subset_dir = source_dir;

    //
    // Open the config file (or open the default if it can't be found)
    //
    try {
      File file = new File(subset_dir, "config.prop");
      if (file.exists())
        user_configuration = new UserConfiguration(file);
      else {
        user_configuration = new UserConfiguration(new File("config",
            "mmsys.a.prop"));
      }

      //
      // We know it's RRF, so fake the input stream setting
      //
      user_configuration.setInputStream(new RichMRMetamorphoSysInputStream());

      //
      // Initialize from MRDOC and MRSAB
      //
      MetaDescriptionInitializer.initialize(
          new File(subset_dir, "MRDOC.RRF"),
          new File(subset_dir, "MRSAB.RRF"), 
          null, null);

    } catch (Exception le) {
      le.printStackTrace();
      MMSToolkit.reportError(this,
          "Unexpected error loading subset configuration.\n "
              + "The most likely problem is that you opened a \n "
              + "directory containing a subset from a previous version \n"
              + "of the Metathesaurus. \n"
              + "This may cause unexpected behavior. \n"
              + "See the console for more details.", false);
    }
    
    //
    // Set source path to subset dir
    //
    user_configuration.setSourcePaths(new String[] { 
        subset_dir.getAbsolutePath() });

    //
    // Set frame title, assumes we are operating outside the event queue thread
    //
    try {
      SwingUtilities.invokeAndWait(new Runnable() {
        public void run() {
          setTitle("Rich Release Format Browser "
              + default_configuration.getReleaseVersion());

        }
      });
    } catch (Exception e) {
      e.printStackTrace();
      MMSToolkit.reportError(this, 
          "Unexpected error setting title." +
          "\n See console for more details.", false);
    }

    //
    // Check indexes, build if necessary
    //
    buildIndexes();

    //
    // Configure concept list manager
    //
    clm.setConfiguration(user_configuration);

    //
    // Configure finder manager
    //
    String font_family = BrowserState.getBrowserState().getFontFamily();
    int font_point_size = BrowserState.getBrowserState().getFontPointSize();

    finder_manager.clear();
    Set finders = finder_manager.getFinders();
    Iterator iter = finders.iterator();
    while (iter.hasNext()) {
      Finder f = (Finder) iter.next();
      f.setConfiguration(default_configuration);
      f.setConfiguration(user_configuration);
      f.subsetDirectoryOpened(subset_dir);
      Properties this_finder_props = BrowserState.getBrowserState()
          .getProperties(f);
      if (this_finder_props != null)
        f.setProperties(this_finder_props);
    }

    //
    // Configure reports manager
    //
    reports_manager.clear();
    Set views = reports_manager.getConceptViews();
    iter = views.iterator();
    while (iter.hasNext()) {
      ConceptView cv = (ConceptView) iter.next();
      cv.setConfiguration(default_configuration);
      cv.setConfiguration(user_configuration);
      cv.subsetDirectoryOpened(subset_dir);
      Properties this_report_props = BrowserState.getBrowserState()
          .getProperties(cv);
      if (this_report_props != null)
        cv.setProperties(this_report_props);
    }

    //
    // Reinitialize the finder mechanism
    //
    ConceptView current_view = reports_manager.getCurrentView();
    FinderHandler finder_handler = current_view.getFinderHandler();
    try {
      finder_handler.init();
    } catch (Exception ex) {
      ex.printStackTrace();
      MMSToolkit.reportError(this,
          "Unexpected error initializing the find dialog.\n"
              + "Please see console for details.", false);
    }
    finder_handler.setDialog(RRFBrowserGUI.getFindDialog());
    RRFBrowserGUI.getFindDialog().setHandler(finder_handler);

    //
    // Load the concept list
    //
    ConceptList cl = BrowserState.getBrowserState().getConceptListForDirectory(
        subset_dir.getAbsolutePath());
    if (cl == null) {
      cl = new ConceptList();
      BrowserState.getBrowserState().addDirectory(subset_dir.getAbsolutePath(), cl);
      try {
        BrowserState.save(new File("config", "dir_saved.txt"));
      } catch (Exception e) {
        //
        // Treat this exception as though canWrite() returned true
        // for a DVD RW drive.
        //
      }
    }
    clm.setConceptList(cl);
    
    // moved here after clm.setConceptList(cl)
    // avoids NPE that occurs if new subset doesn't have
    // last current cui
    ignore_events = true;
    setFontStyle(font_family);
    setFontSize(font_point_size);
    ignore_events = false;
    
    //
    // Load directory history, font family, and font point size
    //
    ignore_events = true;
    String[] dirs = BrowserState.getBrowserState().getDirectoryHistory();
    subset_dir_display.removeAllItems();
    for (int i = 0; i < dirs.length; i++) {
      subset_dir_display.addItem(dirs[i]);
    }
    subset_dir_display.setSelectedItem(subset_dir.getAbsolutePath());
    subset_dir_display.setEnabled(true);
    
    font_family_display.setEnabled(true);
    font_size_display.setEnabled(true);
    ignore_events = false;


    //
    // Enable buttons
    //
    try {
      SwingUtilities.invokeAndWait(new Runnable() {
        public void run() {
          setEnableButtons();
        }
      });
    } catch (Exception e) {
      e.printStackTrace();
      MMSToolkit.reportError(this, "Unexpected error enabling buttons."
          + "\n See console for more details.", false);
    }

    //
    // Refresh concept views
    //
    clm.refresh();

  }

  //
  // Private Methods
  //
  /**
   * Used to update font style on browser.
   * @param font_style desired font style
   */
  private void setFontStyle(String font_style) {
	  final String final_font = font_style;
	  for (int j = 0; j < font_menu_items.length; j++) {
		  if (font_menu_items[j].getText().equals(final_font))
			  font_menu_items[j].setEnabled(false);
		  else
			  font_menu_items[j].setEnabled(true);
	  }
	  BrowserState.getBrowserState().setFontFamily(final_font);
	  getFinderManager().setFontFamily(final_font);
	  getReportsManager().setFontFamily(final_font);
	  
	  try {
		  SwingUtilities.invokeAndWait(new Runnable() {
			  public void run() {
				  font_family_display.setSelectedItem(final_font);
			  }
		  });
	  } catch (Exception e) {
		  e.printStackTrace();
		  MMSToolkit.reportError(RRFBrowserGUI.this,
				  "Unexpected error updating font style."
				  + "\n See console for more details.", false);
	  }
  }  
  
  /**
   * Used to update font size on browser.
   * @param font_size desired font size
   */
  private void setFontSize(int font_size) {
	  final int font_point_size = font_size;
	  if (font_point_size > 8)
		  decrease_font_menu_item.setEnabled(true);
	  else
		  decrease_font_menu_item.setEnabled(false);
	  if (font_point_size >= 27)
		  increase_font_menu_item.setEnabled(false);
	  else
		  increase_font_menu_item.setEnabled(true);
	  BrowserState.getBrowserState().setFontPointSize(font_point_size);
	  getFinderManager().setFontPointSize(font_point_size);
	  getReportsManager().setFontPointSize(font_point_size);
	  
      try {
          SwingUtilities.invokeAndWait(new Runnable() {
            public void run() {
          	  font_size_display.setSelectedItem(new Integer(font_point_size));
            }
          });
        } catch (Exception e) {
          e.printStackTrace();
          MMSToolkit.reportError(RRFBrowserGUI.this,
              "Unexpected error updating font size."
                  + "\n See console for more details.", false);
        }
  }
  
  /**
	 * Sets the {@link JSplitPane} and all of its {@link Component}s to be
	 * enabled or disabled.
	 * 
	 * @param b
	 *            <code>boolean</code>
	 */
  private void setSplitPaneEnabled(boolean b) {
    split_pane.setEnabled(b);
    setComponentEnabled(split_pane, b);
  }

  /**
   * Sets the print toolbar button and menu item to be enabled or disabled.
   * 
   * @param b
   *          <code>boolean</code> true or false
   */
  private void setPrintComponentsEnabled(boolean b) {
    print_menu_item.setEnabled(b);
    print_action.setEnabled(b);
    print_preview_menu_item.setEnabled(b);
    print_preview_button.setEnabled(b);
  }

  /**
   * Sets a given {@link Container} and all of its descendents to be enabled or
   * disabled.
   * 
   * @param c
   *          {@link Container}
   * @param b
   *          <code>boolean</code>
   */
  private void setComponentEnabled(Container c, boolean b) {
    Component[] components = c.getComponents();
    for (int i = 0; i < components.length; i++) {
      if (components[i] instanceof JButton)
        return;
      components[i].setEnabled(b);
      if (components[i] instanceof Container)
        setComponentEnabled((Container) components[i], b);
    }
  }

  /**
   * Enables/Disables the Edit menu items.
   * 
   * @param b
   *          <code>true</code> or <code>false</code>
   */
  private void enableEditMenuItems(boolean b) {
    find_menu_item.setEnabled(b);
    increase_font_menu_item.setEnabled(b);
    decrease_font_menu_item.setEnabled(b);
    change_font_menu.setEnabled(b);
  }

  /**
   * Helper method for building a <code>Options</code> menu item for a
   * particular {@link GUIConfigurable}.
   * 
   * @param gc
   *          a {@link GUIConfigurable} from which to build a options menu item
   * @return a {@link JMenuItem} to add to the <code>Options</code> menu
   */
  private JMenuItem getOptionsMenuItem(GUIConfigurable gc) {
    JMenuItem options_item = new JMenuItem();
    final GUIConfigurable fgc = gc;
    options_item.setText(gc.getName() + " Options");
    JPanel wrapper = new JPanel();
    wrapper.add(options_item);

    options_item.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        fgc.showOptions();
      }
    });

    return options_item;
  }

  /**
   * Returns a reference to the {@link FinderManager}.
   * 
   * @return a reference to the {@link FinderManager}
   */
  public FinderManager getFinderManager() {
    return finder_manager;
  }

  /**
   * Returns a reference to the {@link ReportsManager}.
   * 
   * @return a reference to the {@link ReportsManager}
   */
  public ReportsManager getReportsManager() {
    return reports_manager;
  }

  /**
   * Returns a reference to the {@link ConceptListManager}.
   * 
   * @return a reference to the {@link ConceptListManager}
   */
  public ConceptListManager getConceptListManager() {
    return clm;
  }

  public FileStatisticsConfiguration getFileStatisticsConfiguration() {
    return fs_config;
  }

  /**
   * Saves {@link BrowserState} to dir_saved.txt file so that history and
   * settings can be reset on future runs of the application.
   * 
   * @throws java.lang.Exception
   */
  public void writeCuiToFile() throws Exception {
    if (new File("config").canWrite() && subset_dir != null) {
      BrowserState browser_state = BrowserState.getBrowserState();
      Iterator iter = finder_manager.getFinders().iterator();
      while (iter.hasNext()) {
        Finder finder = (Finder) iter.next();
        browser_state.addProperties(finder, finder.getProperties());
      }
      iter = reports_manager.getConceptViews().iterator();
      while (iter.hasNext()) {
        ConceptView report = (ConceptView) iter.next();
        browser_state.addProperties(report, report.getProperties());
      }

      try {
        BrowserState.save(new File("config", "dir_saved.txt"));
      } catch (IOException ioe) {
        //
        // Treat this exception as though canWrite() returned true
        // for a DVD RW drive.
        //
      }
      // BufferedWriter pw = new BufferedWriter(new FileWriter(new
      // File("config",
      // "dir_saved.txt")));
      // pw.write(xml_ser.toXML(BrowserState.getBrowserState()));
      // pw.close();
    }
  }

  private void buildIndexes() {

    //
  	// Check if indexes exist
  	//
    File aui_mrconso_index_file = new File(new File(subset_dir, "indexes"), "AUI_MRCONSO.x");
    File ptr_aui_rela_index_file = new File(new File(subset_dir, "indexes"),
    "PTR_AUI_RELA_MRHIER.x");
    File sui_ct_mrxw_index_file = new File(new File(subset_dir, "indexes"), "SUI_CT_MRXW.x");

    int missing_ct = 0;
    if(!sui_ct_mrxw_index_file.exists())
    	missing_ct++;
    if(!ptr_aui_rela_index_file.exists()) 
    	missing_ct++;
    if(!aui_mrconso_index_file.exists())
    	missing_ct++;
    if (missing_ct > 0) {
      MMSToolkit.notifyUser(
        this,
        "The browser has detected " +
        (missing_ct > 1 ? "missing index files\n" : "a missing index file\n") +
        "in your subset.\n\n" +
        (!aui_mrconso_index_file.exists()?"AUI_MRCONSO.x\n" : "") +
        (!ptr_aui_rela_index_file.exists() ? "PTR_AUI_RELA.x\n" : "") +
        (!sui_ct_mrxw_index_file.exists() ? "SUI_CT_MRXW.x\n" : "") +
        "will now be built.", "Missing Index Files");
    }
    
    //
    // Build AUI_MRCONSO.x
    //
    if (!aui_mrconso_index_file.exists()) {
      final IndexBuilder index_builder = new IndexBuilder(user_configuration
          .getSourcePaths()[0]);
      final FileStatistics mrconso_stats = fs_config
          .getFileStatistics("MRCONSO");
      final ProgressMonitor pm = new ProgressMonitor(this,
          "Building AUI_MRCONSO.x index", "Starting...", 0, 100);
      try {
        index_builder.buildIndex("MRCONSO", new String[] { "AUI" },
            mrconso_stats, pm);
      } catch (Exception e) {
        if (e.getMessage().indexOf("cancelled") != -1)
          MMSToolkit.reportError("Index build was cancelled.\n"
              + "Browser requires indexes to run properly.\n", true);
        e.printStackTrace();
        MMSToolkit.reportError("Problem building index.\n"
            + "Browser requires indexes to run properly.\n", true);
      }
      pm.setProgress(100);
    }

    //
    // Build PTR_AUI_RELA_MRHIER.x
    //

    if (!ptr_aui_rela_index_file.exists()) {
      final IndexBuilder index_builder = new IndexBuilder(user_configuration
          .getSourcePaths()[0]);
      final FileStatistics mrhier_stats = fs_config.getFileStatistics("MRHIER");
      final ProgressMonitor pm = new ProgressMonitor(this,
          "Building PTR_AUI_RELA_MRHIER.x index", "Starting...", 0, 100);

      pm.setMillisToDecideToPopup(0);
      try {
        index_builder.buildIndex("MRHIER",
            new String[] { "PTR", "AUI", "RELA" }, mrhier_stats, pm);
      } catch (Exception e) {
        if (e.getMessage().indexOf("cancelled") != -1)
          MMSToolkit.reportError("Index build was cancelled.\n"
              + "Browser requires indexes to run properly.\n", true);
        e.printStackTrace();
        MMSToolkit.reportError("Problem building index.\n"
            + "Browser requires indexes to run properly.\n", true);

      }
      pm.setProgress(100);
    }

    //
    // Build SUI_CT_MRXW.x
    //
    if (!sui_ct_mrxw_index_file.exists()) {
      final SuiCountIndexBuilder index_builder = new SuiCountIndexBuilder(
          user_configuration.getSourcePaths()[0]);
      final ProgressMonitor pm = new ProgressMonitor(this,
          "Building SUI_CT_MRXW.x index", "Starting...", 0, 100);

      pm.setMillisToDecideToPopup(0);
      try {
        index_builder.buildSuiCountWordIndex(pm);
      } catch (Exception e) {
        if (e.getMessage().indexOf("cancelled") != -1)
          MMSToolkit.reportError("Index build was cancelled.\n"
              + "Browser requires indexes to run properly.\n", true);
        e.printStackTrace();
        MMSToolkit.reportError("Problem building index.\n"
            + "Browser requires indexes to run properly.\n", true);

      }
      pm.setProgress(100);
    }

  }

  //
  // GlassPaneListener Implementation
  //

  /**
   * Enable glass pane.
   */
  public void enableGlassPane() {
    gc.setVisible(true);
  }

  /**
   * Disable glass pane.
   */
  public void disableGlassPane() {
    gc.setVisible(false);
  }

  /**
   * {@link ConceptListManager} that is GUI aware.
   */
  private class GUIConceptListManager extends ConceptListManager {

    //
    // Constructor
    //

    /**
     * Instantiates an {@link ConceptListManager}.
     */
    public GUIConceptListManager() {
    }

    /**
     * Indicates a {@link Concept} for display in the current view. Fully
     * populates the {@link Concept} before displaying it. Manages the forward
     * and backward aparatus. This method expects to be running in a separate
     * thread.
     * 
     * @param concept
     *          {@link Concept}
     */
    public void addConcept(final Concept concept) {

      try {
        GUIConceptListManager.super.addConcept(concept);

        writeCuiToFile();

        SwingUtilities.invokeAndWait(new Runnable() {
          public void run() {
            setEnableButtons();
            setTitle("Rich Release Format Browser "
                + default_configuration.getReleaseVersion() + " "
                + concept.getCui());

            find_dialog.reset();
          }
        });
      } catch (Exception e) {
        e.printStackTrace();
        MMSToolkit.reportError(RRFBrowserGUI.this,
            "Unexpected error generating concept view."
                + "\n See console for more details.", false);
      }

    }

    /**
     * Refreshes the current {@link Concept} view. This method expects to be
     * running in a separate thread.
     */
    public void refresh() {
      try {
        GUIConceptListManager.super.refresh();
        find_dialog.reset();
      } catch (Exception e) {
        e.printStackTrace();
        MMSToolkit.reportError(RRFBrowserGUI.this,
            "Unexpected error refreshing concept view."
                + "\n See console for more details.", false);
      }
    }

    /**
     * Navigate backward. This method expects to be running in a separate
     * thread.
     */
    public void backward() {
      try {
        GUIConceptListManager.super.backward();

        SwingUtilities.invokeAndWait(new Runnable() {
          public void run() {
            setTitle("Rich Release Format Browser "
                + default_configuration.getReleaseVersion() + " "
                + clm.getCurrentConcept().getCui());
            setEnableButtons();
            find_dialog.reset();
          }
        });

        writeCuiToFile();
      } catch (Exception ex) {
        ex.printStackTrace();
        MMSToolkit.notifyUser(RRFBrowserGUI.this,
            "There was an error navigating backward.\n"
                + "See the console for more details.");
      }
    }

    /**
     * Navigate forward. This method expects to be running in a separate thread.
     */
    public void forward() {
      try {
        GUIConceptListManager.super.forward();

        SwingUtilities.invokeAndWait(new Runnable() {
          public void run() {
            setTitle("Rich Release Format Browser "
                + default_configuration.getReleaseVersion() + " "
                + clm.getCurrentConcept().getCui());
            setEnableButtons();
            find_dialog.reset();
          }
        });

        writeCuiToFile();
      } catch (Exception ex) {
        ex.printStackTrace();
        MMSToolkit.notifyUser(RRFBrowserGUI.this,
            "There was an error navigating forward.\n"
                + "See the console for more details.");
      }
    }

    /**
     * Clear the concept list.
     */
    public void clear() {
      super.clearConceptList();
      finder_manager.clear();
      this.reports_manager.clear();
      find_dialog.reset();
    }
  }

  private class ConceptViewPrintPreviewAction extends PrintPreviewAction {

    /**
     * Instantiates a {@link ConceptViewPrintPreviewAction}.
     * 
     * @param icon_path
     *          the path to the icon
     * @param print_action
     *          the print action
     */
    public ConceptViewPrintPreviewAction(String icon_path, Action print_action) {
      super(icon_path, print_action);
    }

    /**
     * Action performed.
     * 
     * @param e
     *          {@link ActionEvent}
     */
    public void actionPerformed(ActionEvent e) {
      enableGlassPane();
      setPrintable(getReportsManager().getCurrentView().getPrintable());
      setTitle("Preview Concept");
      super.actionPerformed(e);
    }

    public void windowClosed(WindowEvent we) {
      disableGlassPane();
    }
  }

  /**
   * Application entry point.
   * 
   * @param args
   *          command line arguments
   */
  public static void main(String[] args) {

    // Instantiate Application
    try {
      MMSToolkit.setUsingView(true);
      MMSToolkit.setReleaseDatLocation("../release.dat");

      rrf_browser_gui = RRFBrowserGUI.newRRFBrowserGUI();
      rrf_browser_gui.setVisible(true);
      rrf_browser_gui.checkPreviousState();
      rrf_browser_gui.toFront();

      //
      // remind user to open a RRF subset
      //
      if (!new File("config", "dir_saved.txt").exists()) {
        MMSToolkit.notifyUser(rrf_browser_gui,
            "Use the File->Open Subset menu to open a\n"
                + "directory with valid RRF files.");
      }

    } catch (Exception e) {
      e.printStackTrace();
      MMSToolkit.reportError(rrf_browser_gui,
          "Unexpected exception.\nSee user console for details.", false);
    }
  }

}
