/*****************************************************************************
 * Package: gov.nih.nlm.mms.browser
 * Object:  ReportsManager.java
 *
 ****************************************************************************/
package gov.nih.nlm.mms.browser;

import gov.nih.nlm.mms.GUIConfigurable;
import gov.nih.nlm.mms.MMSToolkit;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.ContentView;
import gov.nih.nlm.util.FinderHandler;

import java.awt.BorderLayout;
import java.awt.Component;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Filter;

import javax.swing.JPanel;
import javax.swing.JTabbedPane;
import javax.swing.SwingUtilities;
import javax.swing.border.EmptyBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;


/**
 * Manages the various {@link ConceptView}s.
 *
 * @author Deborah Shapiro
 */
public class ReportsManager implements ChangeListener {

  //
  // Concept view set
  //
  private final List concept_views = new ArrayList();

  //
  // current view
  //
  private ConceptView current_view;

  //
  // Concept list manager
  //
  private ConceptListManager clm;

  //
  // Map to get a display component from a ConceptView
  //
  private final Map gc_to_component_map = new HashMap();

  //
  // Map to get a ConceptView from the display component
  //
  private final Map component_to_gc_map = new HashMap();

  //
  // Constructor
  //

  /**
   * Instantiates an {@link ReportsManager}.
   */
  public ReportsManager() { }

  /**
   * Returns a {@link JPanel} with a {@link JTabbedPane}
   * displaying all of the {@link ReportsManager}'s {@link ConceptView}s.
   * @return {@link JPanel}
   */
  public JPanel getPanel() {
    JPanel reports_manager_panel = new JPanel(new BorderLayout(5,5));
    JTabbedPane tabbed_pane = new JTabbedPane();
    tabbed_pane.addChangeListener(this);
    tabbed_pane.setBorder(new EmptyBorder(0,5,0,0));

    ConceptView[] concept_view_array =
        (ConceptView[])concept_views.toArray(new ConceptView[0]);
    Arrays.sort(concept_view_array);

    int initial_index = 0;
    for (int i=0; i<concept_view_array.length; i++) {
      ConceptView this_cv = (ConceptView)concept_view_array[i];
      tabbed_pane.add(getComponentForTab(this_cv), this_cv.getName());
      this_cv.setConceptListManager(clm);
      if (this_cv.getClass().getName().endsWith("ExpandableConceptView"))
        initial_index = i;
    }
    tabbed_pane.setSelectedIndex(initial_index);

    reports_manager_panel.add(tabbed_pane, BorderLayout.CENTER);
    return reports_manager_panel;

  }

  /**
   * Returns the GUI {@link Component} for the specified {@link Filter}.
   * The application maintains a private hash map that allows
   * this mapping to take place.  If the map does not contain
   * an entry for the filter, a panel is built, added to the map
   * and returned.
   * @param gui_configurable the specified {@link GUIConfigurable}
   * @return the GUI {@link Component} for the specified {@link Filter}.
   */
  public Component getComponentForTab(GUIConfigurable gui_configurable) {
    if (gc_to_component_map.containsKey(gui_configurable)) {
      return (Component)gc_to_component_map.get(gui_configurable);
    }

    //
    // Create the panel for the whole tab
    //
    JPanel tab_panel = new JPanel(new BorderLayout());

    tab_panel.add(gui_configurable.getPanel(), BorderLayout.CENTER);

    // add gc and component to maps & return
    gc_to_component_map.put(gui_configurable, tab_panel);
    component_to_gc_map.put(tab_panel, gui_configurable);
    return tab_panel;
  }


  /**
   * Adds a {@link ConceptView} to the {@link ReportsManager}.
   * Provides another concept report format and style for the user.
   * @param concept_view {@link ConceptView}
   */
  public void addConceptView(ConceptView concept_view) {
    concept_views.add(concept_view);
    current_view = concept_view;
  }

  /**
   * Sets the {@link ConceptListManager}.
   * @param clm {@link ConceptListManager}
   */
  public void setConceptListManager(ConceptListManager clm) {
    this.clm = clm;
  }

  /**
   * Sets a {@link Concept} for display in the {@link ReportsManager}'s
   * chosen {@link ContentView} panel.
   * @param concept {@link Concept} to be displayed
   */
  public void setConcept(Concept concept) {
    final Set concept_views = getConceptViews();
    final Iterator iter = concept_views.iterator();
    while (iter.hasNext()) {
      ( (ConceptView) iter.next()).setConcept(concept);
    }
  }

  /**
   * Returns the {@link ConceptView} that is currently being
   * displayed on the {@link JTabbedPane}.
   * @return {@link ConceptView}
   */
  public ConceptView getCurrentView() {
    return current_view;
  }

  /**
   * Resets the graphical components of the {@link ConceptView}s.
   */
  public void clear() {
    Iterator iter = concept_views.iterator();
    while (iter.hasNext()) {
      ((ConceptView)iter.next()).resetToDefaults();
    }
  }

  /**
   * Returns a {@link Set} of the {@link ConceptView}s.
   * @return concept views
   */
  public Set getConceptViews() {
    return new HashSet(concept_views);
  }

  /**
   * Responds to a user switching tabs in the tabbed pane.
   * Resets the current view.
   * @param e a {@link ChangeEvent}
   */
  public void stateChanged(ChangeEvent e) {
    Component c = ((JTabbedPane)e.getSource()).getSelectedComponent();
    current_view = ((ConceptView)component_to_gc_map.get(c));
    FinderHandler finder_handler = current_view.getFinderHandler();
    try {
      finder_handler.init();
    } catch (Exception ex) {
      ex.printStackTrace();
      MMSToolkit.reportError(RRFBrowserGUI.getRRFBrowserGUI(),
                             "Unexpected error initializing the find dialog.\n" +
                             "Please see console for details.", false);
    }
    finder_handler.setDialog(RRFBrowserGUI.getFindDialog());
    RRFBrowserGUI.getFindDialog().setHandler(finder_handler);
  }

  /**
   * Notifies all the {@link ConceptView}s of the font family
   * to be used for text display.  Prompts an update of the display.
   * @param font_family {@link String}
   */
  public void setFontFamily(final String font_family) {
    for (int i = 0; i < concept_views.size(); i++) {
      final ConceptView l_concept_view = ((ConceptView) concept_views.get(i));
      l_concept_view.setFontFamily(font_family);
      if (clm.getCurrentConcept() != null) {
        final Concept current_concept = clm
            .readConcept(clm.getCurrentConcept());
        l_concept_view.setConcept(current_concept);
      }
    }
  }
  
  /**
   * Notifies all the {@link ConceptView}s of the font point size to be used
   * for text display. Prompts an update of the display.
   * 
   * @param font_point_size
   *          int
   */
  public void setFontPointSize(final int font_point_size) {
    for (int i = 0; i < concept_views.size(); i++) {
      final ConceptView l_concept_view = ((ConceptView) concept_views.get(i));
      l_concept_view.setFontPointSize(font_point_size);
      if (clm.getCurrentConcept() != null) {
        final Concept current_concept = clm
            .readConcept(clm.getCurrentConcept());
        l_concept_view.setConcept(current_concept);
      }
    }

  }
}
