/**************************************************************************
 *
 * Package:    gov.nih.nlm.mms.browser;
 * Class:      WordSearchFinder.java
 *
 *************************************************************************/
package gov.nih.nlm.mms.browser;

import gov.nih.nlm.mms.MMSToolkit;
import gov.nih.nlm.swing.DataChangeListener;
import gov.nih.nlm.swing.SuperJList;
import gov.nih.nlm.swing.SwingToolkit;
import gov.nih.nlm.umls.io.RichMRFilesIndexLookup;
import gov.nih.nlm.umls.io.SuiCountIndexBuilder;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.Language;
import gov.nih.nlm.util.FieldedStringTokenizer;

import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Comparator;
import java.util.Properties;
import java.util.Vector;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.ProgressMonitor;
import javax.swing.ScrollPaneConstants;
import javax.swing.SwingUtilities;
import javax.swing.border.EmptyBorder;

/**
 * Finds {@link Concept}s via word searches.
 * 
 * @author Deborah Shapiro
 */
public class WordSearchFinder extends AbstractFinder {

  public static String regular_mode = "Regular Search Mode";

  public static String normalized_mode = "Normalized Search Mode";

  //
  // GUI elements
  //
  private JTextField search_field;

  private JButton go_button;

  private JList cui_choice_list;

  private JScrollPane cui_list_sp;

  private JButton next_button;

  private JButton prev_button;

  //
  // Language restriction
  //
  private Language language;

  //
  // Sort comparator
  //
  private Comparator named_comparator;

  //
  // Search mode
  //
  private String search_mode = regular_mode;

  //
  // Lvg configuration file directory
  //
  private String lvg_config_file;

  //
  // results label
  //
  private JLabel list_label;

  //
  // Search label
  //
  private JLabel search_label;

  //
  // Full list of word search concept results
  //
  private Concept[] concept_candidates;

  //
  // index of highest ranked concept currently displayed
  //
  private int highest_displayed_concept = 0;

  //
  // index of lowest ranked concept currently displayed
  //
  private int first_displayed = 0;

  //
  // ranking threshold - all concepts ranked below threshold will
  // not be displayed
  //
  private double threshold = 0.001;

  /**
   * Instantiates a {@link WordSearchFinder}.
   */
  public WordSearchFinder() {
  }

  /**
   * Returns a {@link JPanel} containing configuration options.
   * 
   * @return a {@link JPanel} containing configuration options
   */
  public JPanel getPanel() {

    final JPanel full_panel = new JPanel(new BorderLayout(5, 5));
    full_panel.setBorder(new EmptyBorder(5, 5, 5, 5));

    // init word search field and button
    final JPanel button_panel = new JPanel(new BorderLayout(5, 5));
    search_label = new JLabel("Enter search terms: ("
        + getLanguage().getDescription() + ")");
    button_panel.add(search_label, BorderLayout.NORTH);
    search_field = new JTextField(8);
    search_field.addKeyListener(new KeyAdapter() {
      public void keyTyped(KeyEvent e) {
        char c = e.getKeyChar();
        if (c == KeyEvent.VK_ENTER) {
          WordSearchAction wsa = new WordSearchAction();
          wsa.actionPerformed(new ActionEvent(this, 0, ""));
        }
      }
    });
    button_panel.add(search_field, BorderLayout.CENTER);
    go_button = new JButton(new WordSearchAction());
    button_panel.add(go_button, BorderLayout.SOUTH);

    // initialize cui choice list
    final JPanel cui_choice_panel = new JPanel(new BorderLayout(5, 5));
    final JPanel label_panel = new JPanel(new BorderLayout(5, 5));
    final JPanel prev_next_panel = new JPanel(new FlowLayout());
    list_label = new JLabel("Select a concept.");
    label_panel.add(list_label, BorderLayout.WEST);
    next_button = new JButton(new ForwardAction());
    if (next_button.getIcon() != null) {
      next_button.setText(""); // an icon-only button
    }
    next_button.setMargin(new Insets(2, 2, 2, 2));

    prev_button = new JButton(new BackwardAction());
    if (prev_button.getIcon() != null) {
      prev_button.setText(""); // an icon-only button
    }
    prev_button.setMargin(new Insets(2, 2, 2, 2));

    prev_next_panel.add(prev_button);
    prev_next_panel.add(next_button);
    label_panel.add(prev_next_panel, BorderLayout.EAST);
    cui_choice_panel.add(label_panel, BorderLayout.NORTH);
    cui_choice_list = new SuperJList(new String[] {});
    cui_choice_list.addMouseListener(new MouseAdapter() {

      public void mouseClicked(MouseEvent e) {
        if (e.getClickCount() == 2) {
          Thread t = new Thread(new Runnable() {
            public void run() {
            	if (concept_candidates == null || 
            			concept_candidates.length == 0)
            		return;
              clm.addConcept( (Concept) cui_choice_list.getSelectedValue());
              RRFBrowserGUI.getRRFBrowserGUI().disableGlassPane();
            }
          });
          t.start();

        }
      }
    });

    cui_list_sp = new JScrollPane(cui_choice_list,
        ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED,
        ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
    cui_choice_panel.add(cui_list_sp, BorderLayout.CENTER);

    full_panel.add(button_panel, BorderLayout.NORTH);
    full_panel.add(cui_choice_panel, BorderLayout.CENTER);
    
    
    //
    //Create the popup menu.
    //
    JPopupMenu popup = new JPopupMenu();
    JMenuItem option_menu_item = new JMenuItem("Options...");
    option_menu_item.addActionListener( 
    	new ActionListener() {
    	  public void actionPerformed(ActionEvent e) {
    	    JMenuItem source = (JMenuItem)(e.getSource());
    	    String s = source.getText();
    	    if (s.equals("Options...")) {
    	      showOptions();
    	    }
    	  }
    	}
    );
    popup.add(option_menu_item);
    
    /*popup.addSeparator();
    
    JMenuItem copy_item = new JMenuItem("Copy");
    copy_item.addActionListener(
      new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          Thread thread = new Thread(
            new Runnable() {
              public void run() {
                try {
                	//panel.copy();
                } catch (Exception e) {
                  e.printStackTrace();
                  MMSToolkit.reportError(RRFBrowserGUI.getRRFBrowserGUI(),
                  "Unexpected error expanding report.\n" +
                  "See the console for more information.", false);
                }
                RRFBrowserGUI.getRRFBrowserGUI().disableGlassPane();
              }
            }
          );
          RRFBrowserGUI.getRRFBrowserGUI().enableGlassPane();
          thread.start();
        }
      }
    );
    popup.add(copy_item);*/

    //Add listener to components that can bring up popup menus.
    PopupListener popup_listener = new PopupListener(popup);
    full_panel.addMouseListener(popup_listener);

    return full_panel;
  }

  /**
   * Method returns all properties specific to the {@link Finder}.
   * 
   * @return {@link Properties} only those directly related to the finder.
   */
  public Properties getProperties() {
    Properties new_filter_props = new Properties();
    new_filter_props.setProperty("threshold", new Double(threshold).toString());
    new_filter_props.setProperty("language", getLanguage().getAbbreviation());
    new_filter_props.setProperty("search_mode", getSearchMode());
    if (getLvgConfigFile() != null)
      new_filter_props.setProperty("lvg_config_file", getLvgConfigFile());
    return new_filter_props;
  }

  /**
   * Sets the finder properties.
   * @param p
   *          {@link Properties}
   */
  public void setProperties(Properties p) {
    Properties filter_props = p;
    if (filter_props.containsKey("threshold")) {
      setThreshold(new Double((String) filter_props.get("threshold"))
          .doubleValue());
    }
    if (filter_props.containsKey("language")) {
      setLanguage(Language.getLanguage((String) filter_props.get("language")));
    }
    if (filter_props.containsKey("search_mode")) {
      setSearchMode((String) filter_props.get("search_mode"));
    }
    if (filter_props.containsKey("lvg_config_file")) {
      setLvgConfigFile((String) filter_props.get("lvg_config_file"));
    }

  }

  /**
   * Returns the rank threshold.
   * 
   * @return the rank threshold.
   */
  public double getThreshold() {
    return threshold;
  }

  /**
   * Sets the rank threshold.
   * 
   * @param threshold
   *          the rank threshold
   */
  public void setThreshold(double threshold) {
    this.threshold = threshold;
  }

  /**
   * Handles opening of a source directory.
   * 
   * @param source_dir
   *          the source directory
   */
  public void subsetDirectoryOpened(File source_dir) {
		if (next_button != null)
			next_button.setEnabled(false);
		if (prev_button != null)
			prev_button.setEnabled(false);

		File lex_dir = null;
		try {
			lex_dir = new File(source_dir, "../LEX").getCanonicalFile();
		} catch (IOException ioe) {
			// do nothing
			return;
		}
		File lex_pgms = new File(lex_dir, "/LEX_PGMS");

		// get the \LEX\LEX_PGMS\lvg{2005} directory
		File[] dir_list = lex_pgms.listFiles();
		if (dir_list == null)
			return;
		File lvg_version_dir = null;
		for (int i = 0; i < dir_list.length; i++) {
			if (dir_list[i].getAbsolutePath().indexOf("lvg") != -1) {
				lvg_version_dir = dir_list[i];
				break;
			}
		}
		if (lvg_version_dir == null)
			return;

		setLvgConfigFile(new File(lvg_version_dir, "/data/config/lvg.properties")
				.getAbsolutePath());

	}

  /**
	 * Returns the {@link Language} of the index that will be searched. The
	 * default is "ENG".
	 * @return {@link Language}
	 */
  public Language getLanguage() {
    if (language == null)
      return Language.getLanguage("ENG");
    return language;
  }

  /**
   * Returns the {@link Comparator} that will be used for sorting the word
   * search results.
   * 
   * @return {@link Comparator}
   */
  public Comparator getComparator() {
    return named_comparator;
  }

  public String getSearchMode() {
    return search_mode;
  }

  public void setSearchMode(String search_mode) {
    this.search_mode = search_mode;
  }

  public String getLvgConfigFile() {
    return lvg_config_file;
  }

  public void setLvgConfigFile(String lvg_config_file) {
    this.lvg_config_file = lvg_config_file;
  }

  /**
   * Sets the {@link Comparator} thtat will be used for sorting the word search
   * results.
   * 
   * @param comparator
   *          {@link Comparator}
   */
  public void setComparator(Comparator comparator) {
    named_comparator = comparator;
  }

  /**
   * Sets the {@link Language} of the index that will be searched.
   * 
   * @param lat
   *          {@link Language}
   */
  public void setLanguage(Language lat) {
    language = lat;
    search_label.setText("Enter search terms: (" + lat.getDescription() + ")");
  }

  /**
   * Resets the configuration back to default values. When used with a GUI this
   * method should change the data model behind the GUI and allow
   * <code>setPanelSelections</code> to actually update the GUI.
   */
  public void resetToDefaults() {
    search_field.setText("");
    cui_choice_list.setListData(new Object[0]);
    list_label.setText("Select a concept.");
  }

  /**
   * Highlights the text displayed in the word text field.
   */
  public void highlightText() {
    search_field.requestFocus();
    search_field.setSelectionStart(0);
    search_field.setSelectionEnd(search_field.getText().length());
    cui_choice_list.setListData(new Object[0]);
    list_label.setText("Select a concept.");

  }

  /**
   * Adds a {@link DataChangeListener} that responds to configuration changes.
   * Allows main GUI to record all changes to this gui in the undo/redo
   * mechanism.
   * 
   * @param dcl
   *          the {@link DataChangeListener} to add
   */
  public void addDataChangeListener(DataChangeListener dcl) {
  }

  /**
   * Removes the specified {@link DataChangeListener}.
   * 
   * @param dcl
   *          the {@link DataChangeListener} to remove
   */
  public void removeDataChangeListener(DataChangeListener dcl) {
  }

  /**
   * Returns the name.
   * 
   * @return the name
   */
  public String getName() {
    return "Word Search";
  }

  /**
   * Presents an advanced options GUI to the user.
   */
  public void showOptions() {
    final WordSearchFinderOptionsGUI word_search_finder_gui = new WordSearchFinderOptionsGUI(
        this);
    word_search_finder_gui.setVisible(true);
  }

  /**
   * Indicates whether or not an options menu item should be supplied for
   * advanced configuration options.
   * 
   * @return <code>true</code> if so; <code>false</code> otherwise
   */
  public boolean hasAdvancedOptions() {
    return true;
  }
  
  /**
   * Redraws the text portions of the {@link Finder}.
   */
  public void refresh() {
  	try {
  		SwingUtilities.invokeAndWait(new Runnable() {
  			public void run() {
  				  				
  				search_field.setFont(new Font(font_family, Font.PLAIN, font_point_size));
  				cui_choice_list.setFont(new Font(font_family, Font.PLAIN, font_point_size));
  				
  				int row_ct = 100;
  				Concept[] this_page_list = new Concept[row_ct];
  				int index = 0;
  				highest_displayed_concept = 0;
  				if (concept_candidates == null)
  					return;
  				for (int i = highest_displayed_concept; i < highest_displayed_concept
  				+ row_ct && i < concept_candidates.length; i++)
  					this_page_list[index++] = concept_candidates[i];
  				
  				list_label.setText("Select a concept. (" + (highest_displayed_concept + 1)
  						+ " to " + (highest_displayed_concept + index) + " of "
  						+ concept_candidates.length + " results)");
  				first_displayed = highest_displayed_concept + 1;
  				highest_displayed_concept += index;
  				
  				cui_choice_list.setListData(this_page_list);
  				
  				prev_button.setEnabled(false);
  				next_button.setEnabled(highest_displayed_concept < concept_candidates.length);
  			}
  		});
  	} catch (Exception e) {
  		e.printStackTrace();
  		MMSToolkit.reportError(RRFBrowserGUI.getRRFBrowserGUI(),
  				"Unexpected error refreshing report.\n"
  				+ "See the console for more information.", false);
  	}
  }  
  
  
  
  //
  // Inner classes
  //
  public class PopupListener extends MouseAdapter {
    JPopupMenu popup;

    PopupListener(JPopupMenu popupMenu) {
      popup = popupMenu;
    }

    public void mousePressed(MouseEvent e) {
      maybeShowPopup(e);
    }

    public void mouseReleased(MouseEvent e) {
      maybeShowPopup(e);
    }

    private void maybeShowPopup(MouseEvent e) {
      if (e.isPopupTrigger()) {
        popup.show(e.getComponent(),
                   e.getX(), e.getY());
      }
    }
  }
  
  /**
   * Action for the "Go" button.
   */
  private class WordSearchAction extends AbstractAction {

    //
    // Constructor
    //

    /**
     * Instantiates a {@link WordSearchAction}.
     */
    public WordSearchAction() {
      super();
      putValue(Action.NAME, "Search..");
      putValue(Action.SHORT_DESCRIPTION, "Word Search");
    }

    /**
     * @param e
     *          an {@link ActionEvent}
     */
    public void actionPerformed(ActionEvent e) {
      RRFBrowserGUI.getRRFBrowserGUI().enableGlassPane();
      Thread t = new Thread(new Runnable() {
        public void run() {
          try {

            //
            // Validate search field
            //
            if (search_field.getText().equals("")) {
              MMSToolkit.reportError(RRFBrowserGUI.getRRFBrowserGUI(),
                  "A search term must be entered.", false);
              RRFBrowserGUI.getRRFBrowserGUI().disableGlassPane();
              return;
            }

            //
            // Ensure SUI_CT.x has been built if norm word search is chosen
            //
            if (getSearchMode().equals(normalized_mode)) {
              File index_file = new File(new File(user_configuration
                  .getSourcePaths()[0], "indexes"), "SUI_CT.x");
              if (!index_file.exists()) {
                MMSToolkit.notifyUser(RRFBrowserGUI.getRRFBrowserGUI(),
                    "The browser has detected a missing index file\n"
                        + "in your subset.  SUI_CT.x will now be built.");
                final SuiCountIndexBuilder index_builder = new SuiCountIndexBuilder(
                    user_configuration.getSourcePaths()[0]);

                final ProgressMonitor pm = new ProgressMonitor(null,
                    "Building SUI_CT.x index", "Starting...", 0, 100);

                pm.setMillisToDecideToPopup(0);
                try {
                  index_builder.buildSuiCountIndex(pm);
                } catch (Exception e) {
                  if (e.getMessage().indexOf("cancelled") != -1)
                    MMSToolkit
                        .reportError(
                            "Index build was cancelled.\n"
                                + "Normalized word search requires SUI_CT.x to run properly.\n",
                            false);
                  e.printStackTrace();
                  MMSToolkit
                      .reportError(
                          "Problem building index.\n"
                              + "Normalized word search requires SUI_CT.x to run properly.\n",
                          false);

                }
                pm.setProgress(100);
              }

            }

            //
            // Configure lookup object
            //
            final RichMRFilesIndexLookup index_lookup = new RichMRFilesIndexLookup(
                user_configuration.getSourcePaths()[0]);
            index_lookup.setLanguage(getLanguage().getAbbreviation());
            index_lookup.setThreshold(getThreshold());

            //
            // Compute search string
            //
            final String delim = " \t-({[)}]_!@#%&\\:;\"',.?/~+=|<>$`^";
            final String[] words = FieldedStringTokenizer.split(search_field
                .getText(), delim);

            //
            // Normalized Search Mode
            //
            if (getSearchMode().equals(normalized_mode)) {

              // Open and load the LVG Configuration file
              Properties props = new Properties();
              try {
                FileInputStream propInputStream = new FileInputStream(
                    getLvgConfigFile());
                props.load(propInputStream);
              } catch (IOException ioe) {
                ioe.printStackTrace();
                throw new Exception(
                    "Error while loading lvg configuration file.\n"
                        + "Please try opening another properties file.");
              }

              // get the LVG_DIR and use it to find the lvg*dist.jar file
							File lvg_dir = new File(props.getProperty("LVG_DIR"));
							File lib_dir = new File(lvg_dir, "lib");
							File[] file_list = lib_dir.listFiles();
							File jar_file = null;
							for (int i = 0; i < file_list.length; i++) {
								if (file_list[i].getAbsolutePath().endsWith("dist.jar")) {
									jar_file = file_list[i];
									break;
								}
							}
							
							if (jar_file.getAbsolutePath().indexOf(MMSToolkit.getLvgVersion()) == -1) {
								String user_version = jar_file.getAbsolutePath().substring(
										jar_file.getAbsolutePath().lastIndexOf("lvg"));
								user_version = user_version.substring(3, user_version.indexOf("dist"));
								MMSToolkit.notifyUser("It is highly recommended to use the current\n" +
											                "Lvg version " + MMSToolkit.getLvgVersion() + 
											                ", rather than version " + user_version + ".\n" +
											                "Otherwise, results may be unexpected.\n");
							}
							URL url = jar_file.toURL();

              URLClassLoader ucl = URLClassLoader
                  .newInstance(new URL[] { url });

              // use url of jar file to find NormApi method
              Class cl = ucl.loadClass("gov.nih.nlm.nls.lvg.Api.NormApi");
              Constructor constr = cl
                  .getConstructor(new Class[] { String.class });
              Object ac = constr
                  .newInstance(new Object[] { getLvgConfigFile() });

              // invoke NormApi.Mutate() class on search terms
              Vector results = (Vector) cl.getMethod("Mutate",
                  new Class[] { String.class }).invoke(ac,
                  new Object[] { search_field.getText() });

              if (results.size() != 0) {
                final String[] result_words = FieldedStringTokenizer.split(
                    (String) results.get(0), delim);

                concept_candidates = index_lookup
                    .findConceptsByNormWords(result_words);
              } else {
                MMSToolkit.reportError(RRFBrowserGUI.getRRFBrowserGUI(),
                    "Normalization resulted in no search terms.\n"
                        + "Please enter other search terms.", false);
                highlightText();
                RRFBrowserGUI.getRRFBrowserGUI().disableGlassPane();
                return;
              }

            }
            //
            // Regular Search Mode
            //
            else if (getSearchMode().equals(regular_mode)) {
              concept_candidates = index_lookup.findConceptsByWords(words);
            }
            index_lookup.close();

            if (concept_candidates.length == 0) {
              highlightText();
              MMSToolkit.reportError(RRFBrowserGUI.getRRFBrowserGUI(),
                  "No concepts matched the " + getLanguage().getDescription()
                      + " search string.", false);
              RRFBrowserGUI.getRRFBrowserGUI().disableGlassPane();
              return;
            }

            final int row_ct = 100;
            String text = "Select a concept. (";
            if (concept_candidates.length > row_ct) {
              text += "1 to " + row_ct + " of ";
              text += (concept_candidates.length);
              first_displayed = 1;
              highest_displayed_concept = row_ct;
            } else {
              text += concept_candidates.length;
              first_displayed = 1;
              highest_displayed_concept = concept_candidates.length;
            }

            final String list_label_text;
            if (concept_candidates.length > 1 || concept_candidates.length == 0)
              list_label_text = text + " results)";
            else if (concept_candidates.length == 1) {
              list_label_text = text + " result)";
              Thread t = new Thread(new Runnable() {
                public void run() {
                  clm.addConcept(concept_candidates[0]);
                  RRFBrowserGUI.getRRFBrowserGUI().disableGlassPane();
                }
              });
              t.start();
            } else
              list_label_text = "";

            SwingUtilities.invokeAndWait(new Runnable() {
              public void run() {

                list_label.setText(list_label_text);
                if (concept_candidates.length > row_ct) {
                  Concept[] first_page_candidates = new Concept[row_ct];
                  for (int i = 0; i < row_ct; i++) {
                    first_page_candidates[i] = concept_candidates[i];
                  }
                  cui_choice_list.setListData(first_page_candidates);
                  next_button.setEnabled(true);
                } else {
                  cui_choice_list.setListData(concept_candidates);
                  next_button.setEnabled(false);
                }

                prev_button.setEnabled(false);
              }
            });

          } catch (Exception e) {
            e.printStackTrace();
            MMSToolkit.reportError(RRFBrowserGUI.getRRFBrowserGUI(),
                "Unexpected error handling word search."
                    + "\n See console for more details.", false);
          }
          RRFBrowserGUI.getRRFBrowserGUI().disableGlassPane();
        }
      });
      t.start();

    } // end actionPerformed

  } // end inner class

  public class ForwardAction extends AbstractAction {

    /**
     * Instantiates a {@link ForwardAction}.
     */
    public ForwardAction() {
      super();

      putValue(Action.NAME, "Forward");
      putValue(Action.SHORT_DESCRIPTION, "Next page");
      ImageIcon clm_icon = null;
      BufferedImage clm_image = null;
      try {
        URL gif_url = this.getClass().getClassLoader().getResource(
            "config/arrows_blue_005.gif");
        clm_image = SwingToolkit.readScaledImage(gif_url, 17, false);
        clm_icon = new ImageIcon(clm_image);
      } catch (Exception ex) {
        throw new RuntimeException(
            "Scaled version of config/arrows_blue_005.gif could not be loaded.");
      }
      putValue(Action.SMALL_ICON, clm_icon);
      setEnabled(false);
    }

    /**
     * Action performed.
     * 
     * @param e
     *          {@link ActionEvent}
     */
    public void actionPerformed(ActionEvent e) {
      if (highest_displayed_concept >= concept_candidates.length)
        return;
      int row_ct = 100;
      Concept[] this_page_list = new Concept[row_ct];
      int index = 0;
      for (int i = highest_displayed_concept; i < highest_displayed_concept
          + row_ct
          && i < concept_candidates.length; i++)
        this_page_list[index++] = concept_candidates[i];
      list_label.setText("Select a concept. ("
          + (highest_displayed_concept + 1) + " to "
          + (highest_displayed_concept + index) + " of "
          + concept_candidates.length + " results)");
      first_displayed = highest_displayed_concept + 1;
      highest_displayed_concept += index;

      cui_choice_list.setListData(this_page_list);

      prev_button.setEnabled(true);
      next_button
          .setEnabled(highest_displayed_concept < concept_candidates.length);
    }
  }

  public class BackwardAction extends AbstractAction {

    /**
     * Instantiates a {@link BackwardAction}.
     */
    public BackwardAction() {
      super();

      putValue(Action.NAME, "Previous");
      putValue(Action.SHORT_DESCRIPTION, "Previous page");
      ImageIcon clm_icon = null;
      BufferedImage clm_image = null;
      try {
        URL gif_url = this.getClass().getClassLoader().getResource(
            "config/arrows_blue_004.gif");
        clm_image = SwingToolkit.readScaledImage(gif_url, 17, false);
        clm_icon = new ImageIcon(clm_image);
      } catch (Exception ex) {
        throw new RuntimeException(
            "Scaled version of config/arrows_blue_004.gif could not be loaded.");
      }
      putValue(Action.SMALL_ICON, clm_icon);

      setEnabled(false);
    }

    /**
     * Action performed.
     * 
     * @param e
     *          {@link ActionEvent}
     */
    public void actionPerformed(ActionEvent e) {
      int row_ct = 100;
      if (first_displayed - (row_ct + 1) < 0)
        return;
      Concept[] this_page_list = new Concept[row_ct];
      int index = 0;
      for (int i = first_displayed - (row_ct + 1); i < first_displayed - 1; i++)
        this_page_list[index++] = concept_candidates[i];
      first_displayed -= row_ct;
      highest_displayed_concept = first_displayed + (row_ct - 1);
      list_label.setText("Select a concept. (" + first_displayed + " to "
          + highest_displayed_concept + " of " + concept_candidates.length
          + " results)");

      cui_choice_list.setListData(this_page_list);

      next_button.setEnabled(true);
      prev_button.setEnabled(first_displayed != 1);
    }
  }

}
