/**************************************************************************
 *
 * Package:	gov.nih.nlm.mms.cxt
 * Object:	MRCXTBuilderOutputStream.java
 *
 **************************************************************************/
package gov.nih.nlm.mms.cxt;

import gov.nih.nlm.umls.io.PushBackReader;
import gov.nih.nlm.umls.io.RichMRFilesConceptInputStream;
import gov.nih.nlm.umls.io.RichMRFilesIndexLookup;
import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.Context;
import gov.nih.nlm.util.FieldedStringTokenizer;
import gov.nih.nlm.util.SystemToolkit;
import gov.nih.nlm.util.UTF8OutputStreamWriter;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.List;

import javax.swing.ProgressMonitor;
import javax.swing.SwingUtilities;

/**
 * Builds <tt>MRCXT.RRF</tt> from <tt>MRHIER.RRF</tt> and <tt>MRCONSO.RRF</tt> with the
 * help of a few indexes.
 *
 * @author Deborah Shapiro
 */
public class MRCXTBuilderOutputStream extends RichMRFilesConceptInputStream {

  //
  // Used for lookups
  //
  private RichMRFilesIndexLookup index_lookup = null;

  //
  // Output file
  //
  private BufferedWriter mrcxt_target = null;

  //
  // write versioned SABs?
  //
  private boolean versioned_output = false;

  //
  // MRFILES data
  //
  private PushBackReader mrfiles_source = null;

  //
  // Total MRHIER lines
  //
  private int mrhier_lines = 0;

  //
  // Lines seen so far
  //
  private int mrhier_ct = 0;

  //
  // cancelled?
  //
  private boolean is_cancelled = false;

  //
  // Progress monitor during build
  //
  private ProgressMonitor progress = null;

  //
  // Constructor
  //

  /**
   * Instantiates an empty {@link MRCXTBuilderOutputStream}.
   */
  public MRCXTBuilderOutputStream() { }

  /**
   * Writes <tt>MRCXT.RRF</tt>
   * @param gui the {@link MRCXTBuilderGUI}
   * @return int count of mrhier lines
   * @throws java.lang.Exception if anything goes wrong
   */
  public int  writeMRCXT(MRCXTBuilderGUI gui) throws Exception {

    mrhier_ct = 0;
    progress =new ProgressMonitor(gui,
                                  "Building MRCXT.RRF",
                                  "Starting...", 0, 100);

    // encourage progress monitor to display immediately
    progress.setMillisToDecideToPopup(0);
    progress.setMillisToPopup(1);

    //
    // Wait cursor
    //
    gui.enableGlassPane();

    //
    // Open MRCXT file, confirm with user if already exists
    //
    final File source_dir = gui.getSourceDir();
    open(new String[] {source_dir.getAbsolutePath()});

    final File mrcxt_file = new File(source_dir, "MRCXT.RRF");

    mrcxt_target = new BufferedWriter(new UTF8OutputStreamWriter(
      new FileOutputStream(mrcxt_file), gui.getAddUnicodeBOM()),1024*256);

    //
    // Get options
    //
    versioned_output = gui.getVersionedSabs();
    final List included_sources = gui.getIncludedSources();
    int max_contexts = gui.getMaxContexts();

    //
    // Get lookup handler
    //
    index_lookup = new
        RichMRFilesIndexLookup(gui.getSourceDir().getAbsolutePath());
    index_lookup.setMaxResults(-1);

    //
    // Read MRFILES to get number of rows in MRHIER
    //
    final Object[][] mrfiles =
        SystemToolkit.readFieldedReader(mrfiles_source, "|");
    for (int i = 0; i < mrfiles.length; i++) {
      if (mrfiles[i][0].equals("MRHIER.RRF"))
        mrhier_lines = new Integer( (String) mrfiles[i][4]).intValue();
    }

    //
    // progress monitor variables
    //
    int progress_ct = 999;
    int prev_progress = 0;
    progress.setProgress(1);

    int cancel_progress_ct = 0;

    //
    // iterate through mrhier
    //
    String line = null;
    final int threshold = (int)(mrhier_lines/101);
    while ( (line = mrhier_source.readLine()) != null) {
      progress_ct++;
      mrhier_ct++;

      final Context this_cxt = newInstanceFromMRHIER(line);

      //
      // Restrict to max_contexts (calculation based on cxn).
      //
      if (max_contexts != -1 && this_cxt.getCxn() > max_contexts)
          continue;

      // If source is included, write context
      //
      if (included_sources.contains(this_cxt.getSource())) {
        Atom atom = index_lookup.findAtomByAui(this_cxt.getAtom().getAui());
        this_cxt.setAtom(atom);

        //
        // write ancestors
        //
        writeAncestors(this_cxt,gui);

        //
        // write self
        //
        writeSelf(this_cxt,gui);

        //
        // write children
        //
        if (gui.getBuildChildren()) {
          writeChildren(this_cxt,gui);
        }

        //
        // write siblings
        //
        if (gui.getBuildSiblings() &&
            this_cxt.getSource().getCxty().indexOf("NOSIB") == -1 &&
            !this_cxt.isTopLevel()) {
          writeSiblings(this_cxt,gui);
        }
      }

      if (cancel_progress_ct++ >= 1000) {
        //
        // Check if cancelled
        //
        if (progress.isCanceled()) {
          is_cancelled = true;
          mrcxt_target.close();
          mrhier_source.close();
          gui.disableGlassPane();
          progress.close();
          return 0;
        }
        cancel_progress_ct = 0;
      }

      //
      // update progress monitor
      //
      if (progress_ct >= threshold) {
        progress_ct = 0;
        final int cur_progress = getProgress();


        //
        // Update if progress has changed
        //
        if (cur_progress > prev_progress) {
          prev_progress = cur_progress;
          SwingUtilities.invokeLater(
            new Runnable() {
              public void run() {
                progress.setProgress(cur_progress);
                progress.setNote(cur_progress + "% completed");
              }
            }
          );
        }
      }
    }

    //
    // Close files
    //
    mrcxt_target.close();
    mrhier_source.close();



    return mrhier_lines;
  }

  /**
   * Sorts the MRCXT.RRF file.
   * @param gui {@link MRCXTBuilderGUI}
   * @throws java.lang.Exception
   */
  public void sortMRCXT(MRCXTBuilderGUI gui) throws Exception {
    final File source_dir = gui.getSourceDir();

    progress.setNote("Sorting MRCXT.RRF");
    ProgressMonitor p2 = new ProgressMonitor(null, null, null, 0, 100) {
      public void setProgress(int pct) {
        progress.setProgress( (int) (pct * .2) + 80);
        progress.setNote( ( (int) (pct * .2) + 80) + "% completed");
      }
    };
    SystemToolkit.sort(new File(source_dir, "MRCXT.RRF").getAbsolutePath(), true,
                       p2);

   //
   // Finish progress monitor
   //
    progress.setProgress(100);
    progress.setNote("Finished!");
    progress.close();

    gui.disableGlassPane();
    gui.dispose();

  }

  /**
   * Indicates whether or not the license agreement was cancelled or
   * not accepted.
   * @return <code>true</code> if the GUI was cancelled, <code>false</code>
   * otherwise
   */
  public boolean isCancelled() {
    return is_cancelled;
  }

  /**
   * Returns the percentage of the MRHIER.RRF file that
   * has been converted and written to MRCXT.RRF.
   * @return <code>int</code> percentage
   */
  private int getProgress() {
    return (int)((mrhier_ct*80) / mrhier_lines);
  }

  /**
   * Writes ancestors.
   * @param this_cxt the current {@link Context}
   * @param gui the {@link MRCXTBuilderGUI}
   * @throws Exception if anything goes wrong
   */
  private void writeAncestors(Context this_cxt, MRCXTBuilderGUI gui) throws Exception {

    final String ptr = this_cxt.getPtr();
    final String[] tokens = FieldedStringTokenizer.split(ptr, ".");

    final StringBuffer hcd_ptr = new StringBuffer(100);
    for (int i = 0; i < tokens.length; i++) {

      final Atom rel_atom = index_lookup.findAtomByAui(tokens[i]);
      final StringBuffer sb = new StringBuffer(110);
      final Atom atom = this_cxt.getAtom();
      sb.append(atom.getConcept().getCui()).append("|");
      sb.append(atom.getSui()).append("|");
      sb.append( (this_cxt.getAtom() != null) ? this_cxt.getAtom().getAui() : "").
          append("|");
      if (versioned_output)
        sb.append(this_cxt.getSource().getVersionedSourceAbbreviation());
      else
        sb.append(this_cxt.getSource().getRootSourceAbbreviation());
      sb.append("|");
      sb.append(atom.getCode()).append("|");
      sb.append(this_cxt.getCxn()).append("|");
      sb.append("ANC|");
      sb.append(i + 1);
      sb.append("|");
      sb.append(rel_atom.getStr()).append("|");
      sb.append(rel_atom.getConcept().getCui()).append("|");
      sb.append( (rel_atom != null) ? rel_atom.getAui() : "").append("|");

      //
      // get the hcd for this ancestor
      //
      if (!hcd_ptr.toString().equals(""))
        hcd_ptr.append(".");
      hcd_ptr.append(tokens[i]);
      final Context anc_cxt = index_lookup.findAncestorContext(this_cxt, hcd_ptr.toString());

      if (anc_cxt != null) {
        if (i == 0)  // hcd is "" for the ANC 1
          sb.append("|");
        else
          sb.append(anc_cxt.getHcd()).append("|");
        sb.append(anc_cxt.getRela() != null ?
                anc_cxt.getRela().getAbbreviation() : "").append("|");
      } else {
        sb.append("||");
      }
      sb.append("|");
      sb.append( (this_cxt.getCvf() != null) ?
                this_cxt.getCvf().toString() : "").append("|");


      mrcxt_target.write(sb.toString());
      mrcxt_target.newLine();
   }
  }

  /**
   * Write self rows.
   * @param this_cxt current {@link Context}
   * @param gui the {@link MRCXTBuilderGUI}
   * @throws Exception if anything goes wrong
  */
  private void writeSelf(Context this_cxt, MRCXTBuilderGUI gui) throws Exception {

      final StringBuffer sb = new StringBuffer(110);
      final Atom atom = this_cxt.getAtom();

      sb.append(atom.getConcept().getCui()).append("|");
      sb.append(atom.getSui()).append("|");
      sb.append( (this_cxt.getAtom() != null) ? this_cxt.getAtom().getAui() : "").
          append("|");
      if (versioned_output)
        sb.append(this_cxt.getSource().getVersionedSourceAbbreviation());
      else
        sb.append(this_cxt.getSource().getRootSourceAbbreviation());
      sb.append("|");
      sb.append(atom.getCode()).append("|");
      sb.append(this_cxt.getCxn()).append("|");
      sb.append("CCP||"); // CXL
      sb.append(atom.getStr()).append("|");
      sb.append(atom.getConcept().getCui()).append("|");
      sb.append( (atom != null) ? atom.getAui() : "").append("|");
      sb.append(this_cxt.getHcd()).append("|");
      sb.append(this_cxt.getRela() != null ?
                this_cxt.getRela().getAbbreviation() : "").append("||");
      sb.append( (this_cxt.getCvf() != null) ?
                this_cxt.getCvf().toString() : "").append("|");
      mrcxt_target.write(sb.toString());
      mrcxt_target.newLine();
  }

  /**
   * Writes children.
   * @param this_cxt current {@link Context}
   * @param gui the {@link MRCXTBuilderGUI}
   * @throws Exception if anything goes wrong
   */
  private void writeChildren(Context this_cxt, MRCXTBuilderGUI gui) throws Exception{

    final Context[] child_cxts = index_lookup.findChildContexts(this_cxt);
    for (int i = 0; i < child_cxts.length; i++) {

      final Atom rel_atom = index_lookup.findAtomByAui(child_cxts[i].getAtom().getAui());
      final StringBuffer sb = new StringBuffer(110);
      sb.append(this_cxt.getAtom().getConcept().getCui()).append("|");
      sb.append(this_cxt.getAtom().getSui()).append("|");
      sb.append(this_cxt.getAtom().getAui()).
          append("|");
      if (versioned_output)
        sb.append(this_cxt.getSource().getVersionedSourceAbbreviation());
      else
        sb.append(this_cxt.getSource().getRootSourceAbbreviation());
      sb.append("|");
      sb.append(this_cxt.getAtom().getCode()).append("|");
      sb.append(this_cxt.getCxn()).append("|");
      sb.append("CHD|");
      sb.append("|");  //CXL
      sb.append(rel_atom.getStr()).append("|");
      sb.append(rel_atom.getConcept().getCui()).append("|");
      sb.append( (rel_atom != null) ? rel_atom.getAui() : "").append("|");
      sb.append(child_cxts[i].getHcd()).append("|");
      sb.append(child_cxts[i].getRela() != null ?
                child_cxts[i].getRela().getAbbreviation() : "").append("|");
      if (gui.getComputeXC() && index_lookup.countChildContexts(child_cxts[i]) > 0)
        sb.append("+");
      sb.append("|");
      sb.append( (this_cxt.getCvf() != null) ?
                this_cxt.getCvf().toString() : "").append("|");
      mrcxt_target.write(sb.toString());
      mrcxt_target.newLine();
    }
  }

  /**
   * Write Siblings
   * @param this_cxt current {@link Context}
   * @param gui the {@link MRCXTBuilderGUI}
   * @throws Exception if anything goes wrong
   */
  private void writeSiblings(Context this_cxt, MRCXTBuilderGUI gui) throws Exception {
    final Context[] sib_cxts = index_lookup.findSiblingContexts(this_cxt);

    for (int i = 0; i < sib_cxts.length; i++) {
      final Atom rel_atom = index_lookup.findAtomByAui(sib_cxts[i].getAtom().getAui());

      //
      // don't include the self row as a sibling
      //
      if (this_cxt.getAtom().getAtomIdentifier().equals(rel_atom.getAtomIdentifier()))
        continue;

      final StringBuffer sb = new StringBuffer(110);
      sb.append(this_cxt.getAtom().getConcept().getCui()).append("|");
      sb.append(this_cxt.getAtom().getSui()).append("|");
      sb.append( (this_cxt.getAtom() != null) ? this_cxt.getAtom().getAui():"").
          append("|");
      if (versioned_output)
        sb.append(this_cxt.getSource().getVersionedSourceAbbreviation());
      else
        sb.append(this_cxt.getSource().getRootSourceAbbreviation());
      sb.append("|");
      sb.append(this_cxt.getAtom().getCode()).append("|");
      sb.append(this_cxt.getCxn()).append("|");
      sb.append("SIB|");
      sb.append("|");  //CXL
      sb.append(rel_atom.getStr()).append("|");
      sb.append(rel_atom.getConcept().getCui()).append("|");
      sb.append( (rel_atom != null) ? rel_atom.getAui() : "").append("|");
      sb.append(sib_cxts[i].getHcd()).append("|");
      sb.append(sib_cxts[i].getRela() != null ?
                sib_cxts[i].getRela().getAbbreviation() : "").append("|");
      if (gui.getComputeXC() && index_lookup.countChildContexts(sib_cxts[i]) > 0)
        sb.append("+");
      sb.append("|");
      sb.append( (this_cxt.getCvf() != null) ?
                this_cxt.getCvf().toString() : "").append("|");
      mrcxt_target.write(sb.toString());
      mrcxt_target.newLine();
    }
  }

  /**
   * Opens the Rich Release Format MR files.
   * @throws IOException if any files cannot be opened
   */
  protected void openMRFiles() throws IOException {
    mrhier_source = openSourceFile("MRHIER.RRF");
    mrfiles_source = openSourceFile("MRFILES.RRF");
  }


}
