/**************************************************************************
 *
 * Package:	gov.nih.nlm.mms.guis
 * Object:	RemoveByTermgroupListSelectionListener.java
 *
 **************************************************************************/

package gov.nih.nlm.mms.cxt;
import gov.nih.nlm.mms.MMSToolkit;
import gov.nih.nlm.swing.IgnorableListener;
import gov.nih.nlm.swing.ListDialog;
import gov.nih.nlm.swing.SortableJTable;
import gov.nih.nlm.swing.TableSortEvent;
import gov.nih.nlm.swing.TableSortListener;
import gov.nih.nlm.umls.meta.Source;
import gov.nih.nlm.umls.meta.SourceTermType;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;

import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

/**
 * List selection listener for the remove term types table.
 * Initializes and updates selections on the the removable term types table.
 *
 * @author Deborah Shapiro, Brian Carlsen
 */

public class MRCXTBuilderSourceListSelectionListener
  implements ListSelectionListener, IgnorableListener, TableSortListener{

  //
  // Private Fields
  //

  //
  // Table
  //
  private SortableJTable table;

  //
  // gui
  //
  private MRCXTBuilderGUI gui;

  //
  // flag indicating whether or not events are being ignored
  //
  private boolean ignore_events = false;

  private boolean[] selection_before;

  //
  // Constructors
  //
  /**
   * Instantiates a {@link MRCXTBuilderSourceListSelectionListener}
   * with the specified {@link MRCXTBuilderGUI} and {@link SortableJTable}.
   * @param gui the parent gui {@link MRCXTBuilderGUI}
   * @param table the {@link SortableJTable} of {@link Source}s
   */
  public MRCXTBuilderSourceListSelectionListener(
    MRCXTBuilderGUI gui, SortableJTable table) {
    this.table = table;
    this.gui = gui;
  }

  //
  // Methods
  //


  public void setSelections() {
    setIgnoreEvents(true);

    table.clearSelection();
    selection_before = new boolean[table.getModel().getRowCount()];

    final List sources = gui.getSourceList();
    for (int i = 0; i < sources.size(); i++) {
      final Source source = (Source)sources.get(i);
      //
      // check if row should be visible in table
      //
      int row = gui.getRowForSource(source);
      if (row != -1) {
        addSource(row);
        selection_before[table.mapIndex(row)] = true;
      }
    }

    setIgnoreEvents(false);
  }



  /**
   * Listens for changes in the list and validates them.
   * @param e the {@link ListSelectionEvent}
   */
  public void valueChanged(ListSelectionEvent e) {
	  // prevents proceding through the code twice
	  // once for the event related to deselecting a row(s)
	  // once for the event related to selecting a row(s)
	  if (e.getValueIsAdjusting())
		  return;
	  
	  // Only process events we care about
	  if (ignoreEvents()) {
		  return;
	  }

    //gui.fireDataChanged();

    //logic to avoid calling up multiple confirm_tgs
    //dialog boxes when user doesn't Cntrl + click and only one
    //source row is selected
    int single_row = -1;
    if (table.getSelectedRowCount() == 1 &&
    	       table.getSelectedRow() >= e.getFirstIndex() &&
    	       table.getSelectedRow() < e.getLastIndex()){
      single_row = table.getSelectedRow();
    }

    HashSet sources_seen = new HashSet();

    for (int i = e.getFirstIndex(); i <= e.getLastIndex(); i++) {

      Source this_source = (Source)gui.getSourceList().get(i);

      //Set current selection for this row and update gui
      boolean selected = table.isRowSelected(i);
      if (selected)
        gui.addSelectedSource(this_source);
      else
        gui.removeSelectedSource(this_source);

        //Determine if selection changed
      boolean selection_changed =
          (selection_before[table.mapIndex(i)] != selected);

      //Set selection_before for the next event
      selection_before[table.mapIndex(i)] = selected;

      HashSet set = new HashSet();
      Object[] related_sources = null;
      // Do not continue if selection for this row didn't change.
      // single_row variable prevents dialog box from
      // being opened repeatedly when a single row is clicked
      // thereby keeping all the source term types previously selected

      if (selection_changed && (single_row == -1 || single_row == i)) {

    	  set.addAll(Source.
   		       getSourcesForFamilyAsSet(this_source.getSourceFamily()));

        if (set.contains(this_source))
          set.remove(this_source);

          // related_tgs should be given all sources in set b/c if
          // automatic selection is chosen, user will not be given a choice
          // of which related sources to select
        related_sources = set.toArray();

        if (!sources_seen.contains(this_source)) {

          // get confirm_tgs set
          Source[] confirm_sources =
              (Source[]) set.toArray(new Source[0]);
          Arrays.sort(confirm_sources);

          // if you are removing source term types and all related rows are already
          // selected, don't raise the message
          // likewise if you are keeping source term types and all related rows are
          // already unselected, don't raise the message
          boolean raise_message = false;
          if (table.isRowSelected(i)) {
            for (int j = 0; j < confirm_sources.length; j++) {
              int k = gui.getRowForSource(confirm_sources[j]);
              if (!table.isRowSelected(k))
                raise_message = true;
            }
          }
          else {
            for (int j = 0; j < confirm_sources.length; j++) {
              int k = gui.getRowForSource(confirm_sources[j]);
              if (table.isRowSelected(k))
                raise_message = true;
            }
          }

          Object[] end_source_names = null;
          if (!set.isEmpty() && raise_message) {
            String message = "";
            if ( (!table.isRowSelected(i))) {
              message = "<html><b>You have selected not to build contexts from " +
                  this_source.getVersionedSourceAbbreviation() +
                  ".  When contexts are not built from </b><br> <b>" + 
                  this_source.getVersionedSourceAbbreviation() +          
                  ", typically they are not built from the following sources either. " +
                  " </b><br> <b>Choose which sources will not have contexts built.</b></ul>";
            
            }
            else {
              message = "<html><b>You have chosen to build contexts from " +
                  this_source.getVersionedSourceAbbreviation() + 
                  ".  When building contexts from </b><br> <b>" + 
                  this_source.getVersionedSourceAbbreviation() + 
                  " typically contexts should be built from the following sources as well. " +
                  " </b><br> <b>Choose which sources should also have contexts.</b></ul>";
            }

            //
            // Look up source term type names
            //
            final String[] start_source_names = new String[confirm_sources.length];
            for (int k1 = 0; k1 < start_source_names.length; k1++)
              start_source_names[k1] = confirm_sources[k1].getVsab();
            Arrays.sort(start_source_names);

            end_source_names =
                ListDialog.showListMultipleMode(MMSToolkit.getGui(), message,
                                                "Confirm Selection",
                                                start_source_names, null);
          }

          // Cancel was pressed in dialog or none were chosen
          if (related_sources == null)
            continue;


          //
          // Take selected sty names and look back up the sty objects
          //
          if (end_source_names != null) {
            related_sources = new Source[end_source_names.length];
            int ct = 0;
            for (int k1 = 0; k1 < confirm_sources.length; k1++) {
              String name = confirm_sources[k1].getVsab();
              for (int k2 = 0; k2 < end_source_names.length; k2++)
                if (name.equals((String) end_source_names[k2]))
                  related_sources[ct++] = confirm_sources[k1];
            }
          }
        }

        sources_seen.add(this_source);

        //remove or maintain from the table based on dialog box results
        for (int j = 0; j < related_sources.length; j++) {

          int k =
              gui.getRowForSource((Source)related_sources[j]);

          // Unselect if selected
          if (k > -1 && table.isRowSelected(k) && !selected) {
            // prevent dialog box from being raised for each
            // of the other sources with the same source term type
            selection_before[table.mapIndex(k)] = false;
            removeSource(k);
            // Select if unselected
          }
          else if (k > -1 && !table.isRowSelected(k) && selected) {
            selection_before[table.mapIndex(k)] = true;
            addSource(k);
          }
        }
      } // end (if selection changed or only a single row is selected)
    } // end from (first index, last index, etc.)
  }

  /**
   * Returns <code>boolean</code> to indicate whether or not
   * {@link ListSelectionEvent}s are to be ignored.
   * @return <code>boolean</code>
   */
  public boolean ignoreEvents() {
    return ignore_events;
  }

  /**
   * Turns off responses to the {@link ListSelectionEvent}s.
   * @param b <code>boolean</code> <code>true</code> if events are to be ignored
   *				 <code>false</code> if events are to be processed
   */
  public void setIgnoreEvents(boolean b) {
    ignore_events = b;
  }

  /**
   * Selects a row and adds the {@link Source} on that row to the list of
   *  {@link Source}s.
   * @param row number to be selected and removed
   */
  public void addSource(int row) {
    table.addRowSelectionInterval(row, row);
    gui.addSelectedSource(gui.getSourceListAsArray()[row]);
  }

  /**
   * Deselects row and removes the {@link Source} on that row from the list of
   *  {@link Source}s.
   * @param row number to be deselected and not removed
   */
  public void removeSource(int row) {
    table.removeRowSelectionInterval(row, row);
    gui.removeSelectedSource(gui.getSourceListAsArray()[row]);
  }

  //
  // TableSortListener Implementation
  //

  /**
   * Handle start of table sorting.
   * @param e {@link TableSortEvent}
   */
  public void tableSortStarted(TableSortEvent e) {
    setIgnoreEvents(true);
  }

  /**
   * Handle finish of table sorting.
   * @param e {@link TableSortEvent}
   */
  public void tableSortFinished(TableSortEvent e) {
    setIgnoreEvents(false);
  }

}

