/**************************************************************************
 *
 * Package:	gov.nih.nlm.mms.filters
 * Object:	AttType.java
 *
 **************************************************************************/

package gov.nih.nlm.mms.filters;

import gov.nih.nlm.umls.meta.Source;

import java.util.StringTokenizer;

/**
 * Class contains all fields relevant to an attribute type
 *
 * @author Deborah Shapiro, Brian Carlsen
 */

public class AttType implements Comparable {
  //
  // Private Fields
  //
  private Source source;
  private String att_type;  //attribute type

  //
  // Constructors
  //
  /**
   * This is the default no-argument constructor
   */
  public AttType() {};

  /**
   * This constructor takes a String in the form
   * sab|type .
   * @param line {@link String} containing all fields needed
   *   for a {@link AttType}
   */
  public AttType(String line) {
    String[] members = new String[2];
    StringTokenizer tokens = new StringTokenizer(line, "|");
    int i=0;
    while(tokens.hasMoreTokens()) {
      members[i++] = tokens.nextToken();
    }
    //source.setSourceAbbreviation(members[0]);
    setSource(Source.getSource(members[0]));
    setType(members[1]);
  }

  /**
   * This constructor takes initial values for each of the fields in
   * the {@link AttType}.
   * @param sab {@link String} An initial value for the source abbreviation
   * @param type {@link String}
   */
  public AttType(String sab, String type) {
    setSource(Source.getSource(sab));
    setType(type);
  }

  /**
   * This constructor takes initial values for each of the fields in
   * the {@link AttType}.
   * @param source {@link Source}
   * @param type {@link String}
   */
  public AttType(Source source, String type) {
    setSource(source);
    setType(type);
  }

  //
  // Methods
  //
  /**
   * Get hashcode based on source abbreviation and attribute type.
   * @return int hash code for this {@link AttType}
   */
  public int hashCode() {
    return (source.getRootSourceAbbreviation() + att_type).hashCode();
  }

  /**
   * This method returns true if the parameter has a
   * source abbreviation and a attribute type that have the same
   * values as those in this {@link AttType}.
   * @param o {@link Object} to be compared with this {@link AttType}
   * @return boolean <code>true</code> if parameter o equals {@link AttType}
   *   		     <code>false</code> if parameter o is not equal to {@link AttType}
   */
  public boolean equals(Object o) {
    AttType att_type = (AttType)o;
    return
        (att_type.getSource().getRootSourceAbbreviation().equals(
         this.getSource().getRootSourceAbbreviation())) &&
        (att_type.getType().equals(this.getType()));
  }

  /**
   * This method compares this {@link AttType} with a parameter
   * {@link Object} based on the source abbreviation.
   * @param o {@link Object} to be compared with this {@link AttType}
   * @return int
   *		 0 if the parameter is not a {@link AttType}
   */
  public int compareTo(Object o) {
    if(o instanceof AttType) {
      AttType a = (AttType)o;
      if (getSource() == a.getSource()) return getType().compareTo(a.getType());
      return getSource().getRootSourceAbbreviation().compareTo(
        ((AttType)o).getSource().getRootSourceAbbreviation());
    } else
      return 0;
  }

  /**
   * Get a string representation of the {@link AttType} that includes
   * the two fields that make the {@link AttType} unique.
   * @return {@link String} representation of the {@link AttType}
   */
  public String toString() {
     return
      (this.getSource().getRootSourceAbbreviation() + "|" + this.getType());
  }

  //
  // Accessor Methods
  //
  /**
   * This method returns the name of the source.
   * @return {@link Source}
   */
  public Source getSource() {
    return source;
  }

  /**
   * This method sets the {@link Source} value of the attribute type.
   * @param source {@link Source}
   */
  public void setSource(Source source) {
    this.source = source;
  }

  /**
   * This method returns the attribute type.
   * @return {@link String}
   */
  public String getType() {
    return att_type;
  }

  /**
   * This method sets the value of the attribute type.
   * @param type value for the attribute type
   */
  public void setType(String type) {
    this.att_type = type;
  }
}

