/**************************************************************************
 *
 * Package:    gov.nih.nlm.mms.filters;
 * Interface:  AttributeTypesToRemoveFilter.java
 *
 *************************************************************************/
package gov.nih.nlm.mms.filters;

import gov.nih.nlm.mms.ApplicationConfiguration;
import gov.nih.nlm.mms.MMSToolkit;
import gov.nih.nlm.mms.SourcesToRemoveEvent;
import gov.nih.nlm.mms.SourcesToRemoveListener;
import gov.nih.nlm.mms.UserConfiguration;
import gov.nih.nlm.swing.NoEditTableModel;
import gov.nih.nlm.swing.SortableJTable;
import gov.nih.nlm.swing.SwingToolkit;
import gov.nih.nlm.umls.meta.Attribute;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.Source;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.TreeSet;

import javax.swing.ButtonGroup;
import javax.swing.JPanel;
import javax.swing.JRadioButton;

/**
 *
 * This Filter removes all CUIs with a given attribute type.
 *
 * @author Deborah Shapiro
 *
 **/

public class AttributeTypesToRemoveFilter extends AbstractFilter
  implements SourcesToRemoveListener, ConceptContentSelector {

  // Reference to default configuration object
  protected ApplicationConfiguration default_config;

  // Reference to user configuration object
  protected UserConfiguration user_config;

  // List of attribute_types to remove
  protected List attribute_types_to_remove = new ArrayList();

  // List of all attribute types
  protected static ArrayList attribute_types = new ArrayList();

  // Reference to the table (used by resetToDefaults)
  protected SortableJTable attribute_type_table = null;

  // Reference to the listener
  protected AttributeTypesToRemoveListSelectionListener listener = null;

  // Sources currently chosen to be removed
  protected List sources_to_remove = new ArrayList();

  private JPanel panel = null;
  private JPanel full_panel = null;
  private JRadioButton remove_attributes = null;
  private JRadioButton keep_attributes = null;
  private boolean remove_selected_attributes = true;

  //
  // Constructor
  //

  /**
   * Every filter is required to have a no-arg constructor.
   */
  public AttributeTypesToRemoveFilter() {  }

  //
  // Methods
  //

  /**
   * Method provides the GUI panel that will display the filter's
   * screen as one of the tabs on the main GUI window.  The code can
   * assume that this method is called after
   * {@link #setConfiguration(ApplicationConfiguration)} and
   * {@link #setProperties(Properties)}.  There is no data
   * change as a result of calling this method.
   * @return {@link JPanel} that displays graphical interface
   * for the filter
   */
  public JPanel getPanel(){
    if (full_panel != null)
      return full_panel;

    full_panel = new JPanel();


    JPanel toggle_panel = new JPanel(new BorderLayout());
    remove_attributes =
        new JRadioButton("Select attributes to EXCLUDE from subset");
    keep_attributes =
        new JRadioButton("Select attributes to INCLUDE in subset");
    ButtonGroup group = new ButtonGroup();
    group.add(remove_attributes);
    group.add(keep_attributes);

    remove_attributes.setSelected(removeSelectedAttributes());
    remove_attributes.addActionListener(
      new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          setRemoveSelectedAttributes(true);
          MMSToolkit.getGui().resetTitleForFilter(panel, "Attributes to Exclude");
        }
      }
    );

    toggle_panel.add(remove_attributes, BorderLayout.CENTER);

    keep_attributes.setSelected(!removeSelectedAttributes());
    keep_attributes.addActionListener(
      new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          setRemoveSelectedAttributes(false);
          MMSToolkit.getGui().resetTitleForFilter(panel, "Attributes to Include");
        }
      }
    );

    toggle_panel.add(keep_attributes, BorderLayout.SOUTH);

    attribute_type_table = new SortableJTable(getModel());

    attribute_type_table.getTableHeader().getColumnModel().getColumn(0).setPreferredWidth(300);
    attribute_type_table.getTableHeader().getColumnModel().getColumn(1).setPreferredWidth(150);
    attribute_type_table.getTableHeader().getColumnModel().getColumn(2).setPreferredWidth(75);

    listener = new AttributeTypesToRemoveListSelectionListener(this,
      attribute_type_table);
    attribute_type_table.getSelectionModel().addListSelectionListener(listener);

    if (removeSelectedAttributes())
      panel = MMSToolkit.makeCard(attribute_type_table, " Attributes to Exclude ");
    else
      panel = MMSToolkit.makeCard(attribute_type_table, " Attributes to Include ");
    full_panel.setLayout(new BorderLayout());
    full_panel.add(toggle_panel, BorderLayout.NORTH);
    full_panel.add(panel, BorderLayout.CENTER);
    return full_panel;

  }

  /**
   * This method is responsible for selecting the rows
   * of the table based on the properties for this filter.
   */
  public void setPanelSelections() {
    listener.setSelections(getAttributeTypesToRemove());
    if(removeSelectedAttributes()) {
      MMSToolkit.getGui().resetTitleForFilter(panel, "Attributes to Exclude");
      remove_attributes.setSelected(true);
    } else {
      MMSToolkit.getGui().resetTitleForFilter(panel, "Attributes to Include");
      keep_attributes.setSelected(true);
    }
  }

  /**
   * Method provides the description of the filter's screen and
   * includes instructions on use.
   * @return {@link String} text of instructions for use of panel
   */
  public String getTabDescription() {
    return "Please select one or more attribute types to remove from the UMLS " +
        "Metathesaurus.  This filter removes only attribute data and not entire concepts " +
        "from the output subset.";
  }

  protected NoEditTableModel getModel() {
    ArrayList model = new ArrayList(attribute_types.size());
    for(int i=0; i<attribute_types.size(); i++) {
      AttType rt = (AttType)attribute_types.get(i);
      if(!sources_to_remove.contains(rt.getSource()))
	model.add(rt);
    }
    String[][] att_data = new String[model.size()][3];
    for(int i=0; i< model.size(); i++) {
      att_data[i][0] = ((AttType)model.get(i)).getSource().getOfficialName();
      att_data[i][1] = ((AttType)model.get(i)).getSource().getVersionedSourceAbbreviation();
      att_data[i][2] = ((AttType)model.get(i)).getType();
    }
    Object [] headers = {"Source", "Source Abbreviation", "Attribute Name"};
    return new NoEditTableModel(att_data, headers);
  }

  /**
   * Provides functionality for a "Reset To Defaults" Menu item.
   * The implementation removes all selections and
   * fires a "data changed" event.
   */
  public void resetToDefaults() {
    setAttributeTypesToRemove(new ArrayList());
  }

  /**
   * Method provides a <code>JPanel</code> that will be added to the main GUI
   * help menu to allow the user to consult a help window
   * that is particular to this filter.
   * @return <code>JPanel</code> to be added to the Help menu
   */
  public JPanel getHelpPanel() {
    // Wrap html documentation in a panel
    return SwingToolkit.getPanelForHTML(
      this.getClass().getClassLoader().getResource(
      "config/AttributeTypesToRemoveHelp.html"));
  }

  /**
   * Method provides the filter with values from the default properties
   * file.  This is the first method called after the class is constructed.
   * @param c {@link ApplicationConfiguration} default values
   */
  public void setConfiguration(ApplicationConfiguration c) {
    default_config = c;
  }

  /**
   * Method returns all properties specific to the filter.
   * @return {@link Properties} only those directly related to the filter
   *        They will start with the filter's class name which will be
   *        concatenated to the name of the property.
   *        Filter.property_name
   */
  public Properties getProperties() {
    Properties new_filter_props = new Properties();

    new_filter_props.setProperty("remove_selected_attributes",
      new Boolean(removeSelectedAttributes()).toString());

    // Rebuild attribute_types_to_remove property
    List ltr = getAttributeTypesToRemove();
    StringBuffer line = new StringBuffer();
    for(int i=0; i< ltr.size(); i++) {
      line.append(((AttType)ltr.get(i)).getSource().getRootSourceAbbreviation());
      line.append("|");
      line.append(((AttType)ltr.get(i)).getType());
      if(i < ltr.size() - 1)
	line.append(";");
    }
    new_filter_props.setProperty("selected_attribute_types", line.toString());
    return new_filter_props;
  }

  /**
   * Method provides the filter with all properties specific to the
   * filter.
   * @param props {@link Properties}
   */
  public void setProperties(Properties props) {
    attribute_types.clear();
    try {
       BufferedReader in = new BufferedReader(
         new InputStreamReader(
           (InputStream)this.getClass().getClassLoader().getResource(
             "config/att_types.dat").getContent()));

       String line;
       while ( (line = in.readLine()) != null) {
         AttType at = new AttType(line);
         at.setSource(Source.getSource(at.getSource().getVersionedSourceAbbreviation()));
         attribute_types.add(at);
       }

       in.close();

     } catch (IOException e) {
       MMSToolkit.reportError( MMSToolkit.getGui(), e +
         "Attribute Types To Exclude filter requires an\n" +
         "att_types.dat file to exist in the config directory.\n" +
         "This missing file will cause the configuration table\n" +
         "for this filter to be empty.", false);
     }

    // does not fire data changed
    // because it is in response to Init, Load, or Undo
    if(props.containsKey("remove_selected_attributes")) {
      setRemoveSelectedAttributes(new Boolean((String)props.get(
        "remove_selected_attributes")).booleanValue());
    }

    // Get the selected_attribute_types property
    // keep backwards compatible to read attribute_types_to_remove
    if(props.containsKey("attribute_types_to_remove") ||
       props.containsKey("selected_attribute_types")) {

      // Clear existing settings
      attribute_types_to_remove.clear();

      // Set attribute_types to remove from property
      String selected_prop = "";
      if (props.containsKey("selected_attribute_types"))
        selected_prop = (String)props.get("selected_attribute_types");
      else
        selected_prop = (String)props.get("attribute_types_to_remove");
      StringTokenizer tokens = new StringTokenizer(selected_prop, ";");
      while(tokens.hasMoreTokens()) {
	String line = tokens.nextToken();
	Source source = Source.getSource(line.substring(0, line.indexOf('|')));
        if (source != null) {
          String type = line.substring(line.indexOf('|') + 1);
          AttType rel = new AttType(source, type);
          addAttributeTypeToRemove(rel);
        }
      }
    }

    // we do not need to set tahble model of attribute_types_table because it is
    // handled by getPanel() and attribute_types property never changes
  }

  /**
   * Method which determines if a given CUI will be removed from the
   * release.
   * @param concept {@link Concept}
   */
  public void applySelections(Concept concept) {
    Attribute[] attributes = concept.getAttributes();
    for(int i=0; i<attributes.length; i++) {
      Attribute attribute = attributes[i];
      if((removeSelectedAttributes() && attribute_types_to_remove.contains(
          new AttType(attribute.getSource().getVersionedSourceAbbreviation(),
                      attribute.getAttributeName().getAbbreviation())))
        ||
        (!removeSelectedAttributes() && !attribute_types_to_remove.contains(
          new AttType(attribute.getSource().getVersionedSourceAbbreviation(),
                      attribute.getAttributeName().getAbbreviation())))) {
        concept.removeAttribute(attributes[i]);
      }
    }
  }

  /**
   * Returns the name.
   * @return {@link String}
   */
  public String getName() {
    return "Attribute Type List";
  }

  /**
   * Method returns a list of the attribute types to be excluded for
   * display in the configuration file.
   * @return {@link String}
   */
  public String getConfigurationDetails() {
    StringBuffer sb = new StringBuffer();
    String ls = System.getProperty("line.separator");
    if (removeSelectedAttributes())
      sb.append("Terms with the following attribute types were excluded from the subset.").append(ls);

    else
      sb.append("Terms with the following attribute types were included in the subset.").append(ls);    Iterator iter = attribute_types_to_remove.iterator();
    boolean empty_list = true;
    while (iter.hasNext()) {
      AttType rt = (AttType)iter.next();
      // don't print out AttributeTypes if their sources are to be excluded from subset
      if(sources_to_remove.contains(rt.getSource()))
        continue;
      String sn = rt.getSource().getOfficialName();
      String sab = rt.getSource().getVersionedSourceAbbreviation();
      sb.append("  " +
	       ((sn.length() > 62) ? sn.substring(0,59) + "..." : sn) +
	       "                                                                    ".
                 substring(0,63-((sn.length()>62)?62:sn.length())) +
	       sab +
	       "                                                                                  ".
                 substring(0,82-(62+sab.length())) +
	       rt.getType()).append(ls);
      empty_list = false;
    }
    if(empty_list)
      sb.append("  <none>").append(ls);
    return sb.toString();
  }

  //
  // Filter Specific Methods
  //
  /**
   * This method indicates whether the selections represent attributes to
   * be excluded or attributes to be kept.
   * @return <code>true</code> selected attributes will be excluded
   * <code>false</code> selected attributes will be kept
   */
  public boolean removeSelectedAttributes() {
    return remove_selected_attributes;
  }

  /**
   * This method sets whether or not the selections represent attributes to
   * be excluded or attributes to be kept.
   * @param b <code>boolean</code>
   */
  public void setRemoveSelectedAttributes(boolean b) {
    remove_selected_attributes = b;
  }

  /**
   * This method returns the {@link List} containing all attribute types.
   * @return {@link List} of all attribute types
   */
  public List getAttributeTypes() {
    return attribute_types;
  }

  /**
   * This method returns the {@link List} containing all
   * {@link AttType}s that share the same source.
   * @param s {@link Source}
   * @return {@link List} of {@link AttType}s sharing given {@link Source}
   */
   public List getAttributeTypes(Source s) {
     ArrayList list = new ArrayList();
     for(int i=0; i<attribute_types.size(); i++) {
       if(((AttType)attribute_types.get(i)).getSource().equals(s))
       list.add((AttType)attribute_types.get(i));
     }
     return list;
   }

  /**
   * This method returns the {@link Set} containing all
   * {@link AttType}s that share the same {@link Source}.
   * @param s {@link Source}
   * @return {@link Set} of {@link AttType}s sharing given {@link Source}
   */
  public Set getAttributeTypesAsSet(Source s) {
    return new TreeSet(getAttributeTypes(s));
  }

  /**
   * This method clears the data structure that holds the list of attribute
   * types to be excluded.
   */
  public void clearAttributeTypesToRemove() {
    attribute_types_to_remove.clear();
  }

  /**
   * This method adds a {@link AttType} to the data structure that holds
   * the list of attribute types to be excluded.
   * @param attribute_type {@link AttType} to be excluded from subset
   */
  public void addAttributeTypeToRemove(AttType attribute_type) {
    if(!attribute_types_to_remove.contains(attribute_type))
      attribute_types_to_remove.add(attribute_type);
  }

  /**
   * This method removes a {@link AttType} from the data structure that
   * holds the list of attribute_types to be excluded.
   * @param attribute_type {@link AttType} to be included in subset
   * */
  public void removeAttributeTypeToRemove(AttType attribute_type) {
    attribute_types_to_remove.remove(attribute_type);
  }

  /**
   * This method returns the {@link List} containing all
   * AttributeTypes to be excluded.
   * @return {@link List}
   */
  public List getAttributeTypesToRemove() {
    return attribute_types_to_remove;
  }

  /**
   * This method replaces the {@link List} of {@link AttType}s to be removed
   * with the one provided.
   * @param atr {@link List} of {@link AttType}s
   */
  public void setAttributeTypesToRemove(List atr) {
    attribute_types_to_remove.clear();
    for(int i=0; i<atr.size(); i++) {
      AttType r = (AttType)atr.get(i);
      addAttributeTypeToRemove(r);
    }
  }

  /**
   * This method is used by the {@link AttributeTypesToRemoveListSelectionListener}
   * and indicates whether the automatic selection option on the MetamorphoSys Options
   * dialog box has been chosen.  Automatic selection allows the application
   * to select attribute types sharing the same source
   * without prompting the user.
   * @return automatic_selection <code>true</code>
   *				 <code>false</code>
   */
  public boolean getAutomaticSelection() {
    return user_config.getAutomaticSelection();
  }

  /**
   * Method provides the filter with values from the user configuration
   * file.
   * @param c {@link UserConfiguration}
   */
  public void setConfiguration(UserConfiguration c) {
    user_config = c;
  }

  //
  // SourcesToRemoveListener Implementation
  //

  /**
   * This method compensates for the {@link AttType}s with sources to be
   * excluded that are therefore not displayed in the attribute type table.
   * @param at {@link AttType}
   * @return int row where given {@link AttType} is displayed
   */
  public int getRowForAttributeType(AttType at) {
    if(sources_to_remove.contains(at.getSource()))
      return -1;
    int row =0;
    for(int i=0; i<attribute_types.size(); i++) {
      AttType att = (AttType)attribute_types.get(i);
      if(!sources_to_remove.contains(att.getSource())) {
	row++;
        if(at.equals(att))
 	  return attribute_type_table.reverseMapIndex(row-1);
      }
    }
    return -1;
  }

  /**
   * This method compensates for the {@link AttType}s with sources to be
   * excluded that are therefore not displayed in the attribute type table.
   * @param a int row number where {@link AttType} is requested
   * @return {@link AttType} at given row
   */
  public AttType getAttributeTypeAtRow(int a) {
    int row = attribute_type_table.mapIndex(a);
    int subset_row =-1;
    AttType att = null;
    for(int i=0; i<attribute_types.size(); i++) {
      att = (AttType)attribute_types.get(i);
      if(!sources_to_remove.contains(att.getSource())) {
	subset_row++;
      }
      if(subset_row == row)
	return att;
    }
    return att;
  }

  /**
   * This method is called when there is a change to the selections
   * on the sources to exclude tab.  It resets the attribute type
   * table model to only include those attribute types that are not
   * from excluded sources.
   * @param e {@link SourcesToRemoveEvent}
   */
  public void sourcesToRemoveChanged(SourcesToRemoveEvent e) {
    sources_to_remove = e.getSourcesToRemove();
    attribute_type_table.setModel(getModel());
    attribute_type_table.getColumnModel().getColumn(0).setPreferredWidth(300);
    attribute_type_table.getColumnModel().getColumn(1).setPreferredWidth(150);
    attribute_type_table.getColumnModel().getColumn(2).setPreferredWidth(75);
    listener.setSelections(getAttributeTypesToRemove());
  }

  /**
   * Method indicates the {@link Color} that should be used to display
   * the name.
   * @return {@link Color}
   */
  public Color getColor() {
    return null;
  }

  /**
   * Indicates if {@link Filter} can use the efficient mode during
   * the "Initialize Cuis" phase of the subsetting process.
   * The efficient mode reads only MRCONSO and MRSTY during the
   * "Initialize Cuis" phase.
   * @return <code>true</code> if efficient mode is acceptable
   *         <code>false</code> if efficient mode is not acceptable
   */
  public boolean isEfficientEnabled() {
    return true;
  }

}

