/**************************************************************************
 *
 * Package:    gov.nih.nlm.mms.filters;
 * Interface:  ContentViewFilter.java
 *
 *************************************************************************/
package gov.nih.nlm.mms.filters;

import gov.nih.nlm.mms.ApplicationConfiguration;
import gov.nih.nlm.mms.GUIConfigurable;
import gov.nih.nlm.mms.SourcesToRemoveEvent;
import gov.nih.nlm.mms.SourcesToRemoveListener;
import gov.nih.nlm.swing.SwingToolkit;
import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.Attribute;
import gov.nih.nlm.umls.meta.ComponentHistory;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.ContentView;
import gov.nih.nlm.umls.meta.Context;
import gov.nih.nlm.umls.meta.Coocurrence;
import gov.nih.nlm.umls.meta.Definition;
import gov.nih.nlm.umls.meta.Mapping;
import gov.nih.nlm.umls.meta.Relationship;

import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import javax.swing.BoxLayout;
import javax.swing.JCheckBox;
import javax.swing.JPanel;
import javax.swing.border.CompoundBorder;
import javax.swing.border.EmptyBorder;
import javax.swing.border.EtchedBorder;
import javax.swing.border.TitledBorder;


/**
 * Keeps only content in selected content views.  Content view definitions
 * are entirely data driven from the "default" property file.
 **/
public class ContentViewFilter extends AbstractFilter
    implements SourcesToRemoveListener, ConceptContentSelector {

    //
    // List of sources to remove (for validation)
    //
    private List sources_to_remove = null;

    //
    // Indicates whether events should be ignored
    //
    private boolean ignore_events = false;

    //
    // maps ContentViewOption id to the ContentViewOption
    //
    private final Map id_to_content_views = new HashMap();

    //
    // set of selected ContentViewOption ids
    //
    private final Set selected_views = new HashSet();

    //
    // represents the bit-wise OR of all selected views
    //
    private final ContentView combined_view = new ContentView();

    //
    // Validation message
    //
    private String validate_msg;
    private List warning_msgs = new ArrayList();

    //
    // Constructors
    //

    /**
     * Instantiates an empty {@link ContentViewFilter}.
     */
    public ContentViewFilter() { }

    /**
     * Returns a configuration panel.  Allows user to choose which
     * content views should be applied.
     * @return the configuration {@link JPanel}
     */
    public JPanel getPanel(){

      JPanel panel = new JPanel();
      panel.setLayout(new BoxLayout(panel, BoxLayout.Y_AXIS));

      TitledBorder titledBorder = new TitledBorder(
              new EtchedBorder(EtchedBorder.RAISED), "Content View Filter");
      titledBorder.setTitleJustification(TitledBorder.CENTER);
      panel.setBorder(new CompoundBorder(titledBorder,
                                         new EmptyBorder(10, 10, 10, 10)));

      // add all ContentViewOptions to panel
      Iterator iter = id_to_content_views.entrySet().iterator();
      while(iter.hasNext()) {
        Map.Entry entry = (Map.Entry)iter.next();
        ContentViewOption option = (ContentViewOption)entry.getValue();
        option.initCheckBox();
        panel.add(option.getCheckBox());
      }

      return panel;
    }

    /**
     * Method provides a {@link JPanel} that will be called from
     * the help menu to allow the user to reference a help window
     * that is particular to this filter.
     * @return {@link JPanel} to be added to the Help menu
     */
    public JPanel getHelpPanel() {
      return SwingToolkit.getPanelForHTML(
        this.getClass().getClassLoader().getResource(
        "config/ContentViewHelp.html"));
    }

    /**
     * Return the description to appear above the main configuration
     * panel in the {#link MetamorphoSysGUI}.
     * @return tab description {@link String}
     */
    public String getTabDescription() {
	final StringBuffer b = new StringBuffer(200);
        b.append("Restricts the subset to content from one or more of the following content views.");
        return b.toString();
    }

    /**
     * There are no defaults because this filter has nothing to configure.
     */
    public void resetToDefaults() {
      setProperties(default_configuration.getProperties(this));
    }

    /**
     * Gives the filter access to the default configuration object.
     * @param c {@link ApplicationConfiguration}
     */
    public void setConfiguration(ApplicationConfiguration c) {
      default_configuration = c;

      //
      // Get ContentViewOptions from default configuration
      //
      final Properties default_props = default_configuration.getProperties(this);
      final Enumeration defaults = default_props.propertyNames();
      while(defaults.hasMoreElements()) {
        final String prop_name = (String)defaults.nextElement();
        if (prop_name.equals("selected_views")) {
          continue;
        }
        final String id = prop_name.substring(prop_name.lastIndexOf('_')+1);
        ContentViewOption this_option = null;
        if (id_to_content_views.containsKey(id))
          this_option = (ContentViewOption)id_to_content_views.get(id);
        else
          this_option = new ContentViewOption();
        this_option.setProperty(prop_name, default_props.getProperty(prop_name));
        id_to_content_views.put(id, this_option);
      }
    }

    /**
     * Returns a {@link Properties} object with the configuration.
     * @return a {@link Properties} object with the configuration.
     */
    public Properties getProperties() {

      final Properties p = new Properties();

      //
      // return selected view ids
      //
      if (selected_views.isEmpty())
        return p;

      final StringBuffer line = new StringBuffer();
      final Iterator sv_iter = selected_views.iterator();
      while (sv_iter.hasNext())
        line.append(sv_iter.next()).append(sv_iter.hasNext() ? "," : "");
      p.setProperty("selected_views", line.toString());
      return p;
    }

    /**
     * Not an active method, because this filter doesn't have any
     * filter specific properties.
     * @param props {@link Properties}
     */
    public void setProperties(Properties props) {
      //
      // Get selected views from user configuration
      //
      selected_views.clear();
      final String line = props.getProperty("selected_views");
      final String[] sv = line.split("[,]");
      for (int i=0; i<sv.length; i++)
        addSelectedView(sv[i]);
    }

    /**
     * Apply the filter to the specified {@link Concept}. Atoms whose
     * Content View Fields don't provide a bit-wise AND with the selected
     * views, that is non-zero, are removed.
     * @param concept {@link Concept}
     */
    public void applySelections(Concept concept) {

      //
      // DO NOT explicitly remove STYs based on CVF
      // because most will not direclty participate
      //

      final Atom[] atoms = concept.getAtoms();
      for (int i = atoms.length - 1; i >= 0; i--) {
        final ContentView cv = atoms[i].getContentView();
        if (cv == null || !cv.keep(combined_view))
          concept.removeAtom(atoms[i]);
      }

      if (concept.isEmpty())
        return;

      //
      // Remove definitions based on source
      //
      final Definition[] definitions = concept.getDefinitions();
      for (int i = definitions.length - 1; i >= 0; i--) {
        final Definition def = definitions[i];
        if (def.getContentView() == null ||
            !def.getContentView().keep(combined_view))
          concept.removeDefinition(def);
      }

      //
      // Remove attributes based on source
      //
      final Attribute[] attributes = concept.getAttributes();
      for (int i = attributes.length - 1; i >= 0; i--) {
        final Attribute attribute = attributes[i];
        if (attribute.getContentView() == null ||
            !attribute.getContentView().keep(combined_view))
          concept.removeAttribute(attribute);
      }

      //
      // Remove coocurrences based on source
      //
      final Coocurrence[] coocurrences = concept.getCoocurrences();
      for (int i = coocurrences.length - 1; i >= 0; i--) {
        final Coocurrence coc = coocurrences[i];
        if (coc.getContentView() == null ||
            !coc.getContentView().keep(combined_view))
          concept.removeCoocurrence(coc);
      }

      //
      // Remove relationships based on source
      //
      final Relationship[] relationships = concept.getRelationships();
      for (int i = relationships.length - 1; i >= 0; i--) {
        final Relationship rel = relationships[i];
        if (rel.getContentView() == null ||
            !rel.getContentView().keep(combined_view))
          concept.removeRelationship(rel);
      }

      //
      // Remove contexts based on source
      //
      final Context[] contexts = concept.getContexts();
      for (int i = contexts.length - 1; i >= 0; i--) {
        final Context context = contexts[i];
        if (context.getContentView() == null ||
            !context.getContentView().keep(combined_view))
          concept.removeContext(context);
      }

      //
      // Remove mappings based on source
      //
      final Mapping[] mappings = concept.getMappings();
      for (int i = mappings.length - 1; i >= 0; i--) {
        final Mapping mapping = mappings[i];
        if (mapping.getContentView() == null ||
            !mapping.getContentView().keep(combined_view))
          concept.removeMapping(mapping);
      }

      //
      // Remove history based on source
      //
      final ComponentHistory[] chs = concept.getComponentHistories();
      for (int i = chs.length - 1; i >= 0; i--) {
        final ComponentHistory ch = chs[i];
        if (ch.getContentView() == null ||
            !ch.getContentView().keep(combined_view))
          concept.removeComponentHistory(ch);
      }
    }

    /**
     * Inactive method because there are no items to configure on this filter.
     */
    public void setPanelSelections() {
      //
      // ignore events so that when undo calls this method it doesn't
      // add to the undo stack when resetting the JCheckBoxes
      //
      ignore_events = true;
      final Iterator iter = id_to_content_views.values().iterator();
      while (iter.hasNext()) {
        final ContentViewOption option = (ContentViewOption)iter.next();
        option.getCheckBox().setSelected(selected_views.contains(option.getId()));
      }
      ignore_events = false;
    }

    /**
     * Returns the name of the {@link Filter}.
     * @return name of the filter
     */
    public String getName() {
	return "ContentView Filter";
    }

    /**
     * Returns the statement that should appear in the subset log file
     * regarding this {@link Filter}.
     * @return filter statement for log file
     */
    public String getConfigurationDetails() {
      final StringBuffer sb = new StringBuffer(500);
      sb.append("The subset was restricted to the following content views:");
      sb.append(System.getProperty("line.separator"));
      final Iterator iter = id_to_content_views.values().iterator();
      boolean found = false;
      while (iter.hasNext()) {
        final ContentViewOption option = (ContentViewOption)iter.next();
        if (selected_views.contains(option.getId())) {
          found = true;
          sb.append("   ").append(option.getName())
              .append(": ").append(option.getDescription())
              .append(System.getProperty("line.separator"));
        }
      }
      if (found)
        return sb.toString();
      else
        return "The subset was not restricted based on content views (none selected)" +
            System.getProperty("line.separator");
    }

  /**
   * Method indicates the {@link Color} that should be used to display
   * the name.
   * @return {@link Color}
   */
  public Color getColor() {
    return null;
  }

  /**
   * Indicates if {@link Filter} can use the efficient mode during
   * the "Initialize Cuis" phase of the subsetting process.
   * The efficient mode reads only MRCONSO and MRSTY during the
   * "Initialize Cuis" phase.
   * @return <code>true</code> if efficient mode is acceptable
   *         <code>false</code> if efficient mode is not acceptable
   */
  public boolean isEfficientEnabled() {
    return true;
  }

  /**
   * This method is the default method for filters that don't
   * check for validity and it returns true indicating a valid
   * filter.
   * @return <code>true</code> filter is valid
   */
  public boolean validate() {
    warning_msgs.clear();
    if (combined_view.getValue() == 0) {
      validate_msg =
          "ERROR: You have not selected any content views.\n" +
          "You must either select one of the views or\n" +
          "disable this filter.";
      return false;
    }
    if (!sources_to_remove.isEmpty()) {
      warning_msgs.add(
          "WARNING: There are sources selected for removal.\n" +
          "It is highly recommended that you set the 'Source List'\n" +
          " filter to keep all sources (or to remove no sources).");
    };
    if (user_configuration.getOutputStream().removeUnicode())
      warning_msgs.add("WARNING: You are encouranged not to remove records\n" +
                            "containing extended UTF-8 characters when \n" +
                            "running the ContentView filter.\n" +
                            "Please go to the Output Handler tab and uncheck\n" +
                            "the Remove records with extended UTF-8 characters box.");

    return true;
  }

  /**
   * Returns the validation message (if validation falied).
   * @return the validation message (if validation falied)
   */
  public String getValidateMessage() { return validate_msg; }

  /**
   * Returns the messages indicating why validation failed.  Should only be
   * called if {@link #validate()} returns <code>false</code>.
   * @return the failure messages
   */
  public String[] getValidationErrors() {
    return new String[] {getValidateMessage()};
  }

  /**
   * Returns warnings that were indicated during validation.
   * @return the warning messages
   */
  public String[] getValidationWarnings() {
    if (warning_msgs.size() == 0) return null;
    return (String[])warning_msgs.toArray(new String[0]);
  }

  /**
   * Returns the {@link GUIConfigurable} whose tab should next be displayed
   * to allow the user to correct an error or warning.
   * @return GUIConfigurable to be displayed
   */
  public GUIConfigurable getGUIConfigurableForValidationFailure() {
    if (sources_to_remove.size() != 0)
      return new SourcesToRemoveFilter();
    if (user_configuration.getOutputStream().removeUnicode())
      return user_configuration.getOutputStream();
    return null;

  }

  /**
   * Listens for changes in the sources to remove list.
   * @param stre the event
   */
  public void sourcesToRemoveChanged(SourcesToRemoveEvent stre) {
    sources_to_remove = stre.getSourcesToRemove();
  }

  /**
   * Adds a selected view to the group of ones that will be applied.
   * @param id for the related ContentViewOption
   */
  private void addSelectedView(String id) {
    selected_views.add(id);
    makeCombinedView();
  }

  /**
   * Removes a selected view from the group of ones that will be applied.
   * @param id for the related ContentViewOption
   */
  private void removeSelectedView(String id) {
    selected_views.remove(id);
    makeCombinedView();
  }

  /**
   * Calculates the bit-wise OR for all currently selected views.
   */
  private void makeCombinedView() {
    long combined_value = 0L;
    final Iterator iter = id_to_content_views.values().iterator();
    while (iter.hasNext()) {
      final ContentViewOption cvo = (ContentViewOption)iter.next();
      if (selected_views.contains(cvo.getId()))
        combined_value = combined_value | new Long(cvo.getCode()).longValue();
    }
    combined_view.setValue(combined_value);
  }

  //
  // Inner Classes
  //

  /**
   * Represents a single defined content view.
   */
  private class ContentViewOption {

    //
    // CV name
    //
    private String name;

    //
    // CV description
    //
    private String desc;

    //
    // CV bit
    //
    private String code;

    //
    // CV id
    //
    private String id;

    //
    // JCheckBox that appears on the gui panel
    //
    private JCheckBox option_checkbox;

    /**
     * Instantiates a {@link ContentViewOption}
     */
    public ContentViewOption() {}

    /**
     * Initialize the {@link JCheckBox} for this ContentViewOption.
     */
    private void initCheckBox() {
      option_checkbox = new JCheckBox(getName() + ": " + getDescription(), true);
      option_checkbox.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          if (!ignore_events) fireDataChanged();
          if (option_checkbox.isSelected())
            addSelectedView(getId());
          else
            removeSelectedView(getId());
        }
      });
    }

    /**
     * Returns the {@link JCheckBox} for this ContentViewOption.
     * @return {@link JCheckBox}
     */
    private JCheckBox getCheckBox() {
      return option_checkbox;
    }

    /**
     * Initializes the ContentViewOption with data from the properties.
     * @param key name of the property
     * @param value value of the property
     */
    private void setProperty(String key, String value) {
      if (key.startsWith("name")) {
        setName(value);
        setId(key.substring(key.lastIndexOf('_')+1));
      } else if (key.startsWith("code"))
        setCode(value);
      else if (key.startsWith("description"))
        setDescription(value);
    }

    /**
     * Sets the code of the ContentViewOption.
     * @param code the code
     */
    private void setCode(String code) {
      this.code = code;
    }

    /**
     * Returns the code.
     * @return code
     */
    private String getCode() {
      return code;
    }

    /**
     * Sets the description of the ContentViewOption.
     * @param desc description
     */
    private void setDescription(String desc) {
      this.desc = desc;
    }

    /**
     * Returns the description of the ContentViewOption.
     * @return description
     */
    private String getDescription() {
      return desc;
    }

    /**
     * Sets the name of the ContentViewOption.
     * @param name the name of the ContentViewOption
     */
    private void setName(String name) {
      this.name = name;
    }

    /**
     * Returns the name of the ContentViewOption.
     * @return name
     */
    private String getName() {
      return name;
    }

    /**
     * Sets the id of the {@link ContentViewOption}.
     * @param id the id of the {@link ContentViewOption}
     */
    private void setId(String id) {
      this.id = id;
    }

    /**
     * Returns the id of the ContentViewOption.
     * @return id
     */
    private String getId() {
      return id;
    }
  }
}

