/**************************************************************************
 *
 * Package:    gov.nih.nlm.mms.filters;
 * Interface:  LanguagesToRemoveFilter.java
 *
 *************************************************************************/
package gov.nih.nlm.mms.filters;

import gov.nih.nlm.mms.ApplicationConfiguration;
import gov.nih.nlm.mms.MMSToolkit;
import gov.nih.nlm.swing.IgnorableListener;
import gov.nih.nlm.swing.NoEditTableModel;
import gov.nih.nlm.swing.SortableJTable;
import gov.nih.nlm.swing.SwingToolkit;
import gov.nih.nlm.swing.TableSortEvent;
import gov.nih.nlm.swing.TableSortListener;
import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.Language;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.StringTokenizer;

import javax.swing.JPanel;
import javax.swing.event.ListSelectionEvent;

/**
 *
 * This Filter removes all CUI components with a given language
 *
 * @author Deborah Shapiro
 *
 **/

public class LanguagesToRemoveFilter extends AbstractFilter
  implements TableSortListener, IgnorableListener, ConceptContentSelector {

  // Flag to ignore valueChanged events
  private boolean ignore_events = false;

  // List of languages to remove
  private List languages_to_remove = new ArrayList();

  // Reference to the table (used by resetToDefaults)
  private SortableJTable language_table = null;

  //
  // Constructor
  //

  /**
   * Every filter is required to have a no-arg constructor.
   */
  public LanguagesToRemoveFilter() {
  }

  //
  // Methods
  //

  /**
   * Method provides the GUI panel that will display the filter's
   * screen as one of the tabs on the main GUI window.  The code can
   * assume that this method is called after
   * {@link #setConfiguration(ApplicationConfiguration)} and
   * {@link #setProperties(Properties)}.  There is no data
   * change as a result of calling this method.
   * @return {@link JPanel} that displays graphical interface
   * for the filter
   */
  public JPanel getPanel(){

    //
    // Create GUI table
    //
    List lat_list = Language.getLanguagesAsList();
    Collections.sort(lat_list);
    String [][] lat_data = new String[lat_list.size()][2];
    for (int i = 0; i < lat_data.length; i++) {
      lat_data[i][0] = ((Language)lat_list.get(i)).getDescription();
      lat_data[i][1] = ((Language)lat_list.get(i)).getAbbreviation();
    }

    language_table =
      new SortableJTable(new NoEditTableModel(
       lat_data, new Object[] {"Language", "Language Abbreviation"}));

    language_table.getSelectionModel().addListSelectionListener(this);
    language_table.addTableSortListener(this);

    // Package table into a Panel for use in GUI
    JPanel panel = MMSToolkit.makeCard(language_table,
				       " Languages to Exclude ");

    return panel;
  }

  /**
   * This method is responsible for selecting the rows
   * of the table based on the properties for this filter.
   */
  public void setPanelSelections() {
    setIgnoreEvents(true);
    language_table.clearSelection();
    for (int i = 0; i < language_table.getRowCount(); i++)
      if (isLanguageToRemove((String)language_table.getValueAt(i,1)))
	language_table.addRowSelectionInterval(i,i);
    setIgnoreEvents(false);
  }

  /**
   * Method provides the description of the filter's screen and
   * includes instructions on use.
   * @return {@link String} text of instructions for use of panel
   */
  public String getTabDescription() {
    return
      "Please select one or more languages to "+
      "remove from the UMLS Metathesaurus.  " +
      "All terms with one of the selected languages will " +
      "be removed, as well as all attributes and relationships " +
      "connected to those terms.";
  }

  /**
   * Provides functionality for a "Reset To Defaults" Menu item.
   * The implementation sets all rows for selection except for "ENG".
   */
  public void resetToDefaults() {
    // Everything but ENG should be selected.
    clearLanguagesToRemove();
    for (int i = 0; i < language_table.getRowCount(); i++) {
      String lat = (String)language_table.getValueAt(i,1);
      if (!lat.equals("ENG")) {
	addLanguageToRemove(Language.getLanguage(lat));
      }
    }
  }

  /**
   * Method provides a {@link JPanel} that will be added to the main GUI
   * help menu to allow the user to consult a help window
   * that is particular to this filter.
   * @return {@link JPanel} to be called from the Options menu
   */
  public JPanel getHelpPanel() {
    // Wrap html documentation in a panel
    return SwingToolkit.getPanelForHTML(
      this.getClass().getClassLoader().getResource(
      "config/LanguagesToRemoveHelp.html"));
  }

  /**
   * Method provides the filter with values from the default properties
   * file.  This is the first method called after the class is constructed.
   * @param c {@link ApplicationConfiguration} default values
   */
  public void setConfiguration(ApplicationConfiguration c) {
  }

  /**
   * Method returns all properties specific to the filter.
   * They will start with the filter's class name which will be
   * concatenated to the name of the property.
   * Filter.property_name
   * @return {@link Properties} only those directly related to the filter
   */
  public Properties getProperties() {
    // Rebuild languages_to_remove property
    Properties new_filter_props = new Properties();
    List ltr = getLanguagesToRemove();
    StringBuffer line = new StringBuffer();
    for(int i=0; i< ltr.size(); i++) {
      line.append(((Language)ltr.get(i)).getAbbreviation());
      if(i < ltr.size() - 1)
	line.append(";");
    }
    new_filter_props.setProperty("selected_languages", line.toString());
    return new_filter_props;
  }

  /**
   * Method provides the filter with all properties specific to the
   * filter.
   * @param props {@link Properties} all properties in the config file
   */
  public void setProperties(Properties props) {

    // Get the selected_languages property
    // make backwards compatible with languages_to_remove property name
    if(props.containsKey("languages_to_remove") ||
       props.containsKey("selected_languages")) {

      // Clear existing settings
      languages_to_remove.clear();

      // Set languages to remove from property
      // which has the form "DUT;ENG;GER;..."
      String selected_prop = "";
      if (props.containsKey("selected_languages"))
        selected_prop = (String)props.get("selected_languages");
      else
        selected_prop = (String)props.get("languages_to_remove");
      StringTokenizer tokens = new StringTokenizer(selected_prop, ";");
      while(tokens.hasMoreTokens()) {
        Language local_language = Language.getLanguage(tokens.nextToken());
        if (local_language != null)
          addLanguageToRemove(local_language);
      }

    }

    // does not fire data changed
    // because it is in response to Init, Load, or Undo

  }

  /**
   * Method determines if a given CUI component will be
   * removed from the release.
   * @param concept {@link Concept}
   */
  public void applySelections(Concept concept) {
    Atom [] atoms = concept.getAtoms();
    // Determine which atoms should remain in the Concept
    for(int i=0; i<atoms.length; i++) {
      String lat = atoms[i].getLanguage().getAbbreviation();
      if(isLanguageToRemove(lat)) {
        concept.removeAtom(atoms[i]);
      }
    }
  }

  /**
   * Returns the name.
   * @return {@link String}
   */
  public String getName() {
    return "Languages To Exclude";
  }

  /**
   * Method returns a list of the languages that were excluded for display in
   * the configuration file.
   * @return {@link String}
   */
  public String getConfigurationDetails() {
    StringBuffer sb = new StringBuffer();
    String ls = System.getProperty("line.separator");
    sb.append(
      "Terms with the following languages were excluded from subset.").append(ls);
    Iterator iter = languages_to_remove.iterator();
    boolean empty_list = true;
    while (iter.hasNext()) {
      Language lang = (Language)iter.next();
      sb.append("  " +
	       lang.getDescription() +
	       "                                                                                   ".substring(0,68-(lang.getDescription().length())) +
	       lang.getAbbreviation()).append(ls);
      empty_list = false;
    }
    if(empty_list)
      sb.append("  <none>").append(ls);
    return sb.toString();
  }

  /**
   * This method clears the data structure that holds the list of languages
   * to be excluded.
   */
  public void clearLanguagesToRemove() {
    languages_to_remove.clear();
  }

  /**
   * This method adds a {@link Language} to the data structure that holds
   * the list of languages to be excluded.
   * @param language {@link Language}
   */
  public void addLanguageToRemove(Language language) {
    if(!languages_to_remove.contains(language))
      languages_to_remove.add(language);
  }

  /**
   * This method removes a {@link Language} from the data structure that
   * holds the list of languages to be excluded.
   * @param language {@link Language}
   */
  public void removeLanguageToRemove(Language language) {
    languages_to_remove.remove(language);
  }

  /**
   * This method returns true if terms with the specified language
   * should be excluded.
   * @param language {@link String}
   * @return <code>boolean</code>
   */
  public boolean isLanguageToRemove(String language) {
    for(int i=0; i<languages_to_remove.size(); i++) {
      if(((Language)languages_to_remove.get(i)).getAbbreviation().equals(language))
	return true;
    }
    return false;
  }

  /**
   * This method returns the {@link List} containing all
   * languages to be excluded.
   * @return {@link List}
   */
  public List getLanguagesToRemove() {
    return languages_to_remove;
  }

  //
  // IgnorableListener Implementation
  //

  /**
   * Method returns <code>boolean</code> to indicate whether or not
   * {@link ListSelectionEvent}s are to be ignored.
   * @return <code>boolean</code> <code>true</code> to be ignored
   * 			 	<code>false</code> otherwise
   */
  public boolean ignoreEvents() { return ignore_events; }

  /**
   * Method that turns off responses to the {@link ListSelectionEvent}s
   * @param b <code>boolean</code>
   */
  public void setIgnoreEvents(boolean b) { ignore_events = b; }

  //
  // TableSortListener Implementation
  //

  /**
   * Handle start of table sorting.
   * @param e {@link TableSortEvent}
   */
  public void tableSortStarted(TableSortEvent e) {
    setIgnoreEvents(true);
  }

  /**
   * Handle finish of table sorting.
   * @param e {@link TableSortEvent}
   */
  public void tableSortFinished(TableSortEvent e) {
    setIgnoreEvents(false);
  }

  //
  // ListSelectionListener Implentation
  //

  /**
   * Sets the list of languages to be excluded when selection changes.
   * @param e {@link ListSelectionEvent}
   */
  public void valueChanged (ListSelectionEvent e) {

    if (e.getValueIsAdjusting() || ignoreEvents()) return;

    // Review entire table and re-build list of languages to remove
    List new_selections = new ArrayList();
    for (int i = 0; i <= language_table.getRowCount(); i++) {
      if (language_table.getSelectionModel().isSelectedIndex(i))
	new_selections.add(Language.getLanguage(
          (String)language_table.getValueAt(i,1)));
    }

    boolean changed = false;

    List a = new ArrayList(new_selections);
    List b = new ArrayList(languages_to_remove);
    a.removeAll(languages_to_remove);
    b.removeAll(new_selections);
    changed = !(a.size() == 0 && b.size() == 0);
    if (changed) {
      fireDataChanged();
      languages_to_remove = new_selections;
    }

  }

  /**
   * Method indicates the {@link Color} that should be used to display
   * the name.
   * @return {@link Color}
   */
  public Color getColor() {
    return null;
  }

  /**
   * Indicates if {@link Filter} can use the efficient mode during
   * the "Initialize Cuis" phase of the subsetting process.
   * The efficient mode reads only MRCONSO and MRSTY during the
   * "Initialize Cuis" phase.
   * @return <code>true</code> if efficient mode is acceptable
   *         <code>false</code> if efficient mode is not acceptable
   */
  public boolean isEfficientEnabled() {
    return true;
  }

}

