/**************************************************************************
 *
 * Package:    gov.nih.nlm.mms.filters;
 * Object:  PrecedenceFilter.java
 *
 *************************************************************************/
package gov.nih.nlm.mms.filters;

import gov.nih.nlm.mms.ApplicationConfiguration;
import gov.nih.nlm.mms.MMSToolkit;
import gov.nih.nlm.mms.SourcesToRemoveEvent;
import gov.nih.nlm.mms.SourcesToRemoveListener;
import gov.nih.nlm.swing.CutandPasteDNDTable;
import gov.nih.nlm.swing.CutandPasteModeEvent;
import gov.nih.nlm.swing.CutandPasteModeListener;
import gov.nih.nlm.swing.SwingToolkit;
import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.SourceTermType;
import gov.nih.nlm.umls.meta.StringType;
import gov.nih.nlm.umls.meta.TermStatus;
import gov.nih.nlm.util.FieldedStringTokenizer;

import java.awt.Color;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.StringTokenizer;

import javax.swing.JPanel;

/**
 * This Filter allows the user to reorder the source/termgroup rows in the
 * table to indicate the ranking of termgroups desired.
 *
 * @author Deborah Shapiro
 *
 **/

public class PrecedenceFilter extends AbstractFilter
		implements CutandPasteModeListener, FocusListener,
                SourcesToRemoveListener, ConceptManipulator {

  //
  // Precedence lists
  //
  //private List precedence = new ArrayList();
  private List default_precedence = new ArrayList();

  //
  // GUI stuff
  //
  private CutandPasteDNDTable precedence_table;
  private JPanel panel;
  private PrecedenceTableModel prec_model;

  //
  // Wordind delimiter list
  //
  private final String delim = " \t-({[)}]_!@#%&*\\:;\"',.?/~+=|<>$`^";

  //
  // List of sources to remove
  //
  private List sources_to_remove = new ArrayList();

  //
  // TS and STT comaparators
  //
  //private AtomRankComparator ts_comparator = new AtomRankComparator(false);
  //private AtomRankComparator stt_comparator = new AtomRankComparator(true);


  // Maps SourceTermType objects to strings representing the
  // ranking (for getRank).  Lazily initialized in applyFilter.
  private Map termgroup_rank_map = null;

  //
  // Constructor
  //

  /**
   * Instantiates a {@link PrecedenceFilter}.
   */
  public PrecedenceFilter() { }

  //
  // Methods
  //

  /**
   * Provides the GUI panel that will display the filter's
   * screen as one of the tabs on the main GUI window.
   * @return {@link JPanel} that displays graphical interface for the filter
   */
  public JPanel getPanel(){
    prec_model = new PrecedenceTableModel(default_configuration, this);
    precedence_table = new CutandPasteDNDTable(prec_model);
    precedence_table.getTableHeader().getColumnModel().getColumn(0).setPreferredWidth(300);
    precedence_table.getTableHeader().getColumnModel().getColumn(1).setPreferredWidth(150);
    precedence_table.getTableHeader().getColumnModel().getColumn(2).setPreferredWidth(75);
    precedence_table.addCutandPasteModeListener(this);
    precedence_table.addFocusListener(this);
    panel = MMSToolkit.makeCard(precedence_table, " Precedence ");
    return panel;
  }

  /**
   * Applies state to panel selections.
   */
  public void setPanelSelections() {
    prec_model = (PrecedenceTableModel)precedence_table.getModel();
    prec_model.setData((ArrayList)getPrecedenceSubset());
    prec_model.fireTableDataChanged();
  }

  /**
   * Returns the description.
   * @return the description
   */
  public String getTabDescription() {
    return "This screen sets the relative ranking of source/term type combinations.  " +
        "The ranking of sources and term types will affect the output subset.  " +
        "In particular, the name of a concept will be determined by the highest " +
        "ranking term type in that concept. To move rows, either cut and paste " +
        "rows using keyboard shortcuts, or drag and drop. See Help for more information.\n\n" +
        "Users may proceed to the other Options tabs in any order by clicking " +
        "on the tabs across the top of the screen or by clicking \"Done\" on the " +
        "File bar and selecting \"Begin Subset\".";
  }

  /**
   * Resets filter state to default.
   */
  public void resetToDefaults() {
    user_configuration.setPrecedence(
    		new ArrayList(default_precedence));
    if(precedence_table != null) {
      precedence_table.setCutMode(false);
      MMSToolkit.getGui().resetTitleForFilter(panel, "Precedence");
    }
  }

  /**
   * Returns a {@link JPanel} that will be added to the main GUI
   * help menu to allow the user to consult a help window
   * that is particular to this filter.
   * @return {@link JPanel} to be added to the Help menu
   */
  public JPanel getHelpPanel() {
    return SwingToolkit.getPanelForHTML(
      this.getClass().getClassLoader().getResource(
      "config/PrecedenceHelp.html"));
  }

  /**
   * Sets the {@link ApplicationConfiguration}.
   * @param c the {@link ApplicationConfiguration}
   */
  public void setConfiguration(ApplicationConfiguration c) {
    default_configuration = c;
    default_precedence.clear();
    if (default_configuration.getProperty(this, "precedence") != null) {
      final String prec_prop = (String)default_configuration.getProperty(this, "precedence");
      final StringTokenizer tokens = new StringTokenizer(prec_prop, ";");
      while(tokens.hasMoreTokens()) {
        final String line = tokens.nextToken();
        default_precedence.add(default_configuration.getSourceTermType(
            line.substring(0, line.indexOf('|')),
            line.substring(line.indexOf('|')+1)));
      }
    } else {
      MMSToolkit.reportError("There is no default precedence list", true);
    }
  }

  /**
   * Returns all {@link Properties} specific to the filter. Used for undo/redo
   * and saving state.
   * @return {@link Properties} only those directly related to the filter.
   */
  public Properties getProperties() {
    final Properties new_filter_props = new Properties();
    final List str = getPrecedence();
    final StringBuffer sb = new StringBuffer();
    for(int i = 0; i < str.size(); i++) {
      final SourceTermType stt = (SourceTermType)str.get(i);
      sb.append((stt.getSource().getRsab()));
      sb.append("|");
      sb.append((stt.getTermType().getAbbreviation()));
      sb.append(";");
    }
    new_filter_props.setProperty("precedence", sb.toString());
    return new_filter_props;
  }

  /**
   * Sets all {@link Properties} specific to the filter.
   * @param p the {@link Properties}
   */
  public void setProperties(Properties p) {
    clearPrecedence();
    // first get list from user_configuration
    if (p.containsKey("precedence")) {
      final String prec_prop = (String)p.get("precedence");
      final StringTokenizer tokens = new StringTokenizer(prec_prop, ";");
      List l_prec_list = new ArrayList();
      while(tokens.hasMoreTokens()) {
        final String line = tokens.nextToken();
        final SourceTermType tg = (default_configuration.getSourceTermType(
            line.substring(0, line.indexOf('|')),
            line.substring(line.indexOf('|')+1)));
        if (tg != null)
          l_prec_list.add(tg);
      }
      // supplement list from user_configuration (which may be old and
      // incomplete) with list from default_configuration
      for (int i=0; i<default_precedence.size(); i++) {
    	  SourceTermType tgrp = (SourceTermType)default_precedence.get(i);
    	  if (!l_prec_list.contains(tgrp))
    		  l_prec_list.add(tgrp);
      }
      user_configuration.setPrecedence(l_prec_list);
    }

    if (precedence_table != null) {
      precedence_table.setCutMode(false);
      MMSToolkit.getGui().resetTitleForFilter(panel, "Precedence");
    }
  }

  /**
   * Manipulates data in the {@link Concept}, but does not remove any data
   * from the concept.
   * @param concept {@link Concept}
   */
  public void applyChanges(Concept concept) {

    //
    // If in init_cui phase, don't bother ranking
    //
    if (init_cui) return;

    // Fields used to store info about PF strings
    String [] pf_words = null;
    String prev_lat = "";
    final HashSet top_luis = new HashSet();

    //
    // Comparison Strings
    //
    String sorted_lc_pf_string = null;
    String sorted_pf_string = null;
    String lc_pf_string = null;
    String [] sorted_pf_words = null;
    String [] lc_pf_words = null;
    String [] sorted_lc_pf_words = null;

    // Initialize the termgroup_rank_map
    if (termgroup_rank_map == null) {
      termgroup_rank_map = new HashMap(400);
      List l_prec = user_configuration.getPrecedence();
      for (int i = 0; i < l_prec.size(); i++) {
        final SourceTermType tmgrp = (SourceTermType)l_prec.get(i);
        final int rnk = 9999 - i;
        final String r = "0000" + rnk;
        final String pad_i = r.substring(r.length() - 4);
        termgroup_rank_map.put(tmgrp,pad_i);
      }
    }

    final Atom[] atoms = concept.getAtoms();
    // Sort the atoms, group by language
    try {
      Arrays.sort(atoms, new Atom.AtomRankComparator(atoms,
          (SourceTermType[]) getPrecedence().toArray(new SourceTermType[0]),
          true, false, false));
    } catch (Exception e) {
      e.printStackTrace();
    }

    for (int i=0; i<atoms.length; i++) {
      final Atom atom = atoms[i];
      // If the language changes, we have the next P
      if (!atom.getLat().getAbbreviation().equals(prev_lat))
  	    top_luis.add(atom.getLui());

      // If this atom has a "top lui" lui, assign ts=P else ts=S
      if (top_luis.contains(atom.getLui())) {
        atom.setTs(TermStatus.getTermStatus("P"));
      } else {
        atom.setTs(TermStatus.getTermStatus("S"));
      }

      prev_lat = atom.getLat().getAbbreviation();
    }

    // Assign stt field
    try {
      Arrays.sort(atoms, new Atom.AtomRankComparator(atoms,
          (SourceTermType[]) getPrecedence().toArray(new SourceTermType[0]),
          true, true, true));
    } catch (Exception e) {
      e.printStackTrace();
    }
    String prev_lui="L0000000", pf_str;
    final HashMap sui_hm = new HashMap();
    final HashSet suis_seen = new HashSet();

    for (int j = 0; j < atoms.length; j++) {
      final Atom this_atom = (Atom)atoms[j];
      if (suis_seen.contains(this_atom.getSui())) {
        this_atom.setIspref("N");
      } else {
        this_atom.setIspref("Y");
        suis_seen.add(this_atom.getSui());
      }
      if (!prev_lui.equals(this_atom.getLui())) {
        this_atom.setStt(StringType.getStringType("PF"));
        prev_lui = this_atom.getLui();
        pf_str = this_atom.getStr();
        sui_hm.clear();
        sui_hm.put(this_atom.getSui(), "PF");

        // Tokenize string into words, save the words
        pf_words = FieldedStringTokenizer.split(pf_str, delim);
        sorted_pf_words = new String[pf_words.length];
        lc_pf_words = new String[pf_words.length];
        sorted_lc_pf_words = new String[pf_words.length];

        // Iterate through the words of the PF string
        for (int i = 0; i < pf_words.length; i++) {
          final String pf_word = pf_words[i];
          // Save the words in a sorted fashion
          // (these are sorted later)
          sorted_pf_words[i] = pf_word;

          // Save the lowercased words
          lc_pf_words[i] = pf_word.toLowerCase();

          // Save the lowercased words in a sorted fashion
          // (these are sorted later)
          sorted_lc_pf_words[i] = lc_pf_words[i];
        }

        // sort the words themselves
        Arrays.sort(sorted_pf_words, 0, pf_words.length);

        // sort the lowercase words
        Arrays.sort(sorted_lc_pf_words, 0, pf_words.length);


        sorted_lc_pf_string = null;
        sorted_pf_string = null;
        lc_pf_string = null;

      } else {

        // if this cui|lui|sui has already been assigned an stt
        // simply use the previously assigned one and continue
        if(sui_hm.containsKey(this_atom.getSui())) {
	      this_atom.setStt(StringType.getStringType(
            (String)sui_hm.get(this_atom.getSui())));
          continue;
        }

        String stt = "V";

        // Tokenize string into words
        final String[] cmp_words = FieldedStringTokenizer.split(this_atom.getStr(), delim);
        final String [] sorted_cmp_words = new String[cmp_words.length];
        final String [] lc_cmp_words = new String[cmp_words.length];
        final String [] sorted_lc_cmp_words = new String[cmp_words.length];

        // Iterate through words
        for (int i=0; i < cmp_words.length; i++) {

          final String cmp_word = cmp_words[i];

          // Save the words in a sorted fashion
          // (this is sorted later)
          sorted_cmp_words[i] = cmp_word;

          // Save the lowercased words
          lc_cmp_words[i] = cmp_word.toLowerCase();

          // Save the lowercased words
          // (this is sorted later)
          sorted_lc_cmp_words[i] = lc_cmp_words[i];
        }

        // Sort the words
        Arrays.sort(sorted_cmp_words, 0, cmp_words.length);

        // Sort the lowercased words
        Arrays.sort(sorted_lc_cmp_words, 0, cmp_words.length);

        // Generate strings from words by appending them
        // together separated by spaces
        final String sorted_lc_cmp_string = join(sorted_lc_cmp_words);
        final String sorted_cmp_string = join(sorted_cmp_words);
        final String lc_cmp_string = join(lc_cmp_words);

        if (sorted_lc_pf_string == null && sorted_lc_pf_words != null)
          sorted_lc_pf_string = join(sorted_lc_pf_words);
        else if (sorted_lc_pf_words == null)
          sorted_lc_pf_string = "";
        if (sorted_pf_string == null && sorted_pf_words != null)
          sorted_pf_string = join(sorted_pf_words);
        else if (sorted_pf_words == null)
          sorted_pf_string = "";
        if (lc_pf_string == null && lc_pf_words != null)
          lc_pf_string = join(lc_pf_words);
        else if (lc_pf_words == null)
          lc_pf_string = "";

        // If the sorted lowercased PF string equals
        // the sorted lowercased variant string, but
        // the sorted non-lowercased PF string does not equal
        // the sorted non-lowercased variant string,
        // then there is a case variant
        if(sorted_lc_pf_string.equals(sorted_lc_cmp_string) &&
           !sorted_pf_string.equals(sorted_cmp_string)) {
          stt = stt + "C";
        }

        // If the sorted lowercased PF string equals
        // the sorted lowercased variant string, but
        // the non-sorted lowercased PF string does not equal
        // the non-sorted lowercased variant string,
        // then there is a word-order variant
        if(sorted_lc_pf_string.equals(sorted_lc_cmp_string) &&
          !lc_pf_string.equals(lc_cmp_string)) {
          stt = stt + "W";
        }
        if(stt.equals("V"))
          stt = "VO";
        this_atom.setStt(StringType.getStringType(stt));
        sui_hm.put(this_atom.getSui(), stt);
      }
    }
  }

  /**
   * Concatenates all {@link String}s in input array into
   * a single {@link String} separated by spaces.
   * @param strings {@link String}[]  of {@link String}s
   *        to be concatenated
   * @return {@link String}
   */
  private String join(final String [] strings) {
    final StringBuffer join_buffer = new StringBuffer(100);
    for(int i = 0; i < strings.length; i++) {
      final String str = strings[i];
      join_buffer.append(str);
      if(i<strings.length -1 && !str.equals(""))
        join_buffer.append(" ");
    }
    return join_buffer.toString();
  }

  /**
   * Returns the name.
   * @return {@link String}
   */
  public String getName() {
    return "Precedence";
  }

  /**
   * Method returns a list of the precedence order for display in
   * the configuration file.
   * @return {@link String}
   */
  public String getConfigurationDetails() {
    final StringBuffer sb = new StringBuffer();
    final String ls = System.getProperty("line.separator");
    sb.append("Precedence Ranking of Source/Term Type").append(ls);
    final Iterator iter = getPrecedenceSubset().iterator();
    boolean empty_list = true;
    while(iter.hasNext()) {
      final SourceTermType tg = (SourceTermType)iter.next();
      // don't print out SourceTermTypes if their sources are to be excluded from subset
      if(sources_to_remove.contains(tg.getSource()))
        continue;
      String sn = tg.getSource().getOfficialName();
      sb.append("  " +
                ((sn.length() > 67) ? sn.substring(0,64) + "..." : sn) +
                "                                                                  ".substring(0,68-((sn.length()>67)?67:sn.length())) +
                tg.getSource().getVersionedSourceAbbreviation() +
                "|" +
                tg.getTermType().getAbbreviation()).append(ls);
      empty_list = false;
    }
    if(empty_list)
      sb.append("  <none>").append(ls);
    return sb.toString();
  }

  /**
   * This method clears the data structure that holds the precedence list.
   */
  public void clearPrecedence() {
    user_configuration.setPrecedence(new ArrayList());
  }

  /**
   * This method returns the {@link ArrayList} ranking
   * the {@link SourceTermType}s according to precedence.
   * @return precedence {@link ArrayList}
   */
  public List getPrecedence() {
    return user_configuration.getPrecedence();
  }


  /**
   * This method makes updates to the precedence list based on changes made
   * by the user with the drag & drop interface.
   * @param drag_row_index <code>int[]</code>- indexes of all rows to be
   * 					dragged to new location
   * @param drop_row_index <code>int</code> - index of location where dragged
   * 					rows should be dropped
   */
  public void moveSourceTermTypes(int[] drag_row_index, int drop_row_index) {
		if (drag_row_index[0] == drop_row_index
				|| drag_row_index[0] == drop_row_index - 1)
			return;
		fireDataChanged();
		List l_prec_list = user_configuration.getPrecedence();
		for (int i = 0; i < drag_row_index.length; i++) {
			l_prec_list.add(drop_row_index, l_prec_list.get(drag_row_index[i]));
			if (drop_row_index < drag_row_index[i]) {
				drop_row_index++;
				l_prec_list.remove(drag_row_index[i] + 1);
			} else if (drop_row_index > drag_row_index[i]) {
				for (int j = i + 1; j < drag_row_index.length; j++) {
					drag_row_index[j]--;
				}
				l_prec_list.remove(drag_row_index[i]);
			}
		}
		user_configuration.setPrecedence(l_prec_list);
		setPanelSelections();
	}


  /**
	 * This method is called when a cut has appropriately been done while in cut
	 * mode.
	 * @param e {@link CutandPasteModeEvent}
	 */
  public void cutDoneinCutMode(CutandPasteModeEvent e) {
    MMSToolkit.getGui().resetTitleForFilter(panel, "Precedence - paste mode");
  }

  /**
   * This method is called when a cut has been attempted illegally
   * while in paste mode.
   * @param e {@link CutandPasteModeEvent}
   */
  public void cutDoneinPasteMode(CutandPasteModeEvent e) {
    MMSToolkit.notifyUser(MMSToolkit.getGui(),
        "You already have rows on your clipboard.  You must now paste.");
  }

  /**
   * This method is called when a paste has appropriately been done
   * while in paste mode.
   * @param e {@link CutandPasteModeEvent}
   */
  public void pasteDoneinPasteMode(CutandPasteModeEvent e) {
    MMSToolkit.getGui().resetTitleForFilter(panel, "Precedence");
  }

  /**
   * This method is called when a paste has been attempted illegally
   * while in cut mode.
   * @param e {@link CutandPasteModeEvent}
   */
  public void pasteDoneinCutMode(CutandPasteModeEvent e) {
    MMSToolkit.notifyUser(MMSToolkit.getGui(),
        "You have nothing on your clipboard.  You must cut rows.");
  }

  /**
   * This method indicates that this filter is only valid when
   * any rows that have been cut have also been pasted.
   * @return <code>true</code> if filter is valid
   * 	     <code>false</code> if cut has been done, but paste has
   * 				not been done.
   */
  public boolean validate() {
    if(precedence_table == null)
      return true;
    if(precedence_table.getCutMode())
      return false;
    else
      return true;
  }

  /**
   * This method returns the message that the user will see if this
   * filter is invalid.
   * @return {@link String} with error message
   */
  public String getValidateMessage() {
    return "You have not pasted the row(s) on your clipboard.\nSelect a paste location and press <Cntrl-V> to paste the row(s).";
  }

  /**
   * Provides means to keep focus on the {@link PrecedenceFilter}
   * so as not to leave the filter when a cut has not yet been
   * pasted.
   * @param e {@link FocusEvent}
   */
  public void focusGained(FocusEvent e) {
  }

  /**
   * Provides means to keep focus on the {@link PrecedenceFilter}
   * so as not to leave the filter when a cut has not yet been
   * pasted.
   * @param e {@link FocusEvent}
   */
  public void focusLost(FocusEvent e) {
    focusGained(e);
  }

  //
  // SourcesToRemoveListener Implementation
  //
  /**
   * This method is called when there is a change to the selections
   * on the sources to exclude tab.  It resets the attribute type
   * table model to only include those attribute types that are not
   * from excluded sources.
   * @param e {@link SourcesToRemoveEvent}
   */
  public void sourcesToRemoveChanged(SourcesToRemoveEvent e) {
    sources_to_remove = e.getSourcesToRemove();
    prec_model = (PrecedenceTableModel)precedence_table.getModel();
    prec_model.setData((ArrayList)getPrecedenceSubset());
    prec_model.fireTableDataChanged();
  }

  /**
   * Gets the subset of rows from the precedence data that will not
   * be excluded based on the exclusions chosen on the Sources to
   * Exclude tab.
   * @return {@link List} of {@link SourceTermType}s
   */
  public List getPrecedenceSubset() {
    ArrayList prec_subset = new ArrayList();
    List l_prec_list = user_configuration.getPrecedence();
    
    for(int i=0; i<l_prec_list.size(); i++) {
      if(!sources_to_remove.contains(
        ((SourceTermType)l_prec_list.get(i)).getSource())) {
        prec_subset.add(l_prec_list.get(i));
      }
    }
    return prec_subset;
  }

  /**
   * Method indicates the {@link Color} that should be used to display
   * the name.
   * @return {@link Color}
   */
  public Color getColor() {
    return null;
  }

  /**
   * Indicates if {@link Filter} can use the efficient mode during
   * the "Initialize Cuis" phase of the subsetting process.
   * The efficient mode reads only MRCONSO and MRSTY during the
   * "Initialize Cuis" phase.
   * @return <code>true</code> if efficient mode is acceptable
   *         <code>false</code> if efficient mode is not acceptable
   */
  public boolean isEfficientEnabled() {
    return true;
  }

}

