/**************************************************************************
 *
 * Package:	gov.nih.nlm.mms.filters
 * Object:	PrecedenceTableModel.java
 *
 **************************************************************************/

package gov.nih.nlm.mms.filters;

import gov.nih.nlm.mms.ApplicationConfiguration;
import gov.nih.nlm.swing.CutandPasteDNDTableModel;
import gov.nih.nlm.umls.meta.SourceTermType;

import java.util.ArrayList;

import javax.swing.table.AbstractTableModel;

/**
 * Manages and initializes the data for the Precedence table
 * 
 * @author Deborah Shapiro, Brian Carlsen
 */

public class PrecedenceTableModel extends AbstractTableModel implements
    CutandPasteDNDTableModel {

  //
  // Private Fields
  //
  final private String[] column_names = { "Full Source Name",
      "Source Abbreviation", "Term Type" };

  final private Class[] column_classes = { String.class, String.class,
      String.class };

  private PrecedenceFilter f;

  private ArrayList data = new ArrayList();

  private ArrayList cut_data = new ArrayList();

  //
  // Constructor
  //
  /**
   * This constructor takes the initial values for the precedence table from the
   * default configuration file.
   * 
   * @param config
   *          {@link ApplicationConfiguration} object
   * @param filter
   *          {@link PrecedenceFilter}
   */
  public PrecedenceTableModel(ApplicationConfiguration config,
      PrecedenceFilter filter) {
    super();
    this.f = filter;
  }

  //
  // Methods
  //
  /*
   * Method for intaking data into the model. @param {@link ArrayList}
   */
  public void setData(ArrayList d) {
    data = d;
  }

  /**
   * Returns the number of columns in the precedence table.
   * 
   * @return the number of columns
   */
  public int getColumnCount() {
    return column_names.length;
  }

  /**
   * Returns the number of rows in the precedence table.
   * 
   * @return the number of rows
   */
  public int getRowCount() {
    return data.size();
  }

  /**
   * Returns the name of a column in the precedence table.
   * 
   * @param column_number
   *          The <code>int</code> number of the column in question
   * @return the {@link String} name of the column in the table in position
   *         column_number
   */
  public String getColumnName(int column_number) {
    return column_names[column_number];
  }

  /**
   * Returns the value at the cell of the precedence table.
   * 
   * @param row
   *          The <code>int</code> row number of the cell
   * @param column
   *          The <code>int</code> column number of the cell
   * @return the {@link Object} value at the cell at row and column
   */
  public Object getValueAt(int row, int column) {
    SourceTermType tg = (SourceTermType) data.get(row);
    if (column == 0)
      return tg.getSource().getOfficialName();
    else if (column == 1)
      return tg.getSource().getVersionedSourceAbbreviation();
    else if (column == 2)
      return tg.getTermType().getAbbreviation();
    else
      return null;
  }

  /**
   * Returns the {@link Class} of the {@link Object}s in column with index
   * column_number.
   * 
   * @param column_number
   *          <code>int</code> indicates the column in question
   * @return the {@link Class} of the column
   */
  public Class getColumnClass(int column_number) {
    return column_classes[column_number];
  }

  /**
   * Moves rows in table from initial/drag location to final/drop location.
   * 
   * @param drag_row_index
   *          are the indexes of the rows that are to be dragged
   * @param drop_row_index
   *          is the index where the dragged row is to be dropped
   */
  public void moveRow(int[] drag_row_index, int drop_row_index) {
    int prec_drop_row_index = getRowForSourceTermType((SourceTermType) data
        .get(drop_row_index));
    int[] prec_drag_row_index = new int[drag_row_index.length];
    for (int i = 0; i < drag_row_index.length; i++) {
      prec_drag_row_index[i] = getRowForSourceTermType((SourceTermType) data
          .get(drag_row_index[i]));
    }
    f.moveSourceTermTypes(prec_drag_row_index, prec_drop_row_index);
    this.fireTableDataChanged();
  }

  /**
   * Removes rows in table that are cut. Saves the cut data for when the paste
   * event occurs.
   * 
   * @param rows_to_cut
   *          are the indexes of the rows that are to be cut
   */
  public void cutRow(int[] rows_to_cut) {
    f.fireDataChanged();
    cut_data.clear();
    for (int i = 0; i < rows_to_cut.length; i++) {
      cut_data.add(data.get(rows_to_cut[i]));
    }
    for (int i = 0; i < rows_to_cut.length; i++) {
      data.remove(rows_to_cut[i] - i);
      f.getPrecedence().remove(
          getRowForSourceTermType((SourceTermType) cut_data.get(i)));
    }
    fireTableDataChanged();
  }

  /**
   * Pastes rows in table that were cut.
   * 
   * @param paste_location
   *          is the index where the cut row will be pasted
   */
  public void pasteRow(int paste_location) {
    for (int i = 0; i < cut_data.size(); i++) {
      int prec_paste_location = getRowForSourceTermType((SourceTermType) data
          .get(paste_location));
      f.getPrecedence().add(prec_paste_location, cut_data.get(i));
      data.add(paste_location, (SourceTermType) cut_data.get(i));
      paste_location++;
    }
    fireTableDataChanged();
  }

  /**
   * Gets row from the {@link PrecedenceFilter}'s precedence array for the
   * given {@link SourceTermType}.
   * 
   * @param tg
   *          {@link SourceTermType}
   * @return <code>int</code> row for the given {@link SourceTermType}
   */
  public int getRowForSourceTermType(SourceTermType tg) {
    for (int i = 0; i < f.getPrecedence().size(); i++) {
      SourceTermType term = (SourceTermType) f.getPrecedence().get(i);
      if (term.equals(tg))
        return i;
    }
    return -1;
  }
}
