/**************************************************************************
 *
 * Package:    gov.nih.nlm.mms.filters;
 * Interface:  RelationshipTypesToRemoveFilter.java
 *
 *************************************************************************/
package gov.nih.nlm.mms.filters;

import gov.nih.nlm.mms.ApplicationConfiguration;
import gov.nih.nlm.mms.MMSToolkit;
import gov.nih.nlm.mms.SourcesToRemoveEvent;
import gov.nih.nlm.mms.SourcesToRemoveListener;
import gov.nih.nlm.mms.UserConfiguration;
import gov.nih.nlm.swing.NoEditTableModel;
import gov.nih.nlm.swing.SortableJTable;
import gov.nih.nlm.swing.SwingToolkit;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.Relationship;
import gov.nih.nlm.umls.meta.Source;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.BufferedReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.TreeSet;

import javax.swing.ButtonGroup;
import javax.swing.JPanel;
import javax.swing.JRadioButton;

/**
 *
 * This Filter removes all CUIs with a given relationship type.
 *
 * @author Deborah Shapiro
 *
 **/

public class RelationshipTypesToRemoveFilter extends AbstractFilter
  implements SourcesToRemoveListener, ConceptContentSelector {

  // Reference to user configuration object
  private UserConfiguration user_config;

  // List of relationship_types to remove
  private List relationship_types_to_remove = new ArrayList();

  // List of all relationship types
  private static ArrayList relationship_types = new ArrayList();

  // Reference to the table (used by resetToDefaults)
  private SortableJTable relationship_type_table = null;

  // Reference to the listener
  private RelationshipTypesToRemoveListSelectionListener listener = null;

  // Sources currently chosen to be removed
  private List sources_to_remove = new ArrayList();

  private JPanel panel = null;
  private JPanel full_panel = null;
  private JRadioButton remove_rels = null;
  private JRadioButton keep_rels = null;
  private boolean remove_selected_rels = true;

  //
  // Constructor
  //

  /**
   * Every filter is required to have a no-arg constructor.
   */
  public RelationshipTypesToRemoveFilter() {  }

  //
  // Methods
  //

  /**
   * Method provides the GUI panel that will display the filter's
   * screen as one of the tabs on the main GUI window.  The code can
   * assume that this method is called after
   * {@link #setConfiguration(ApplicationConfiguration)} and
   * {@link #setProperties(Properties)}.  There is no data
   * change as a result of calling this method.
   * @return {@link JPanel} that displays graphical interface
   * for the filter
   */
  public JPanel getPanel(){
    if (full_panel != null)
      return full_panel;

    full_panel = new JPanel();


    JPanel toggle_panel = new JPanel(new BorderLayout());
    remove_rels =
        new JRadioButton("Select relationships to EXCLUDE from subset");
    keep_rels =
        new JRadioButton("Select relationships to INCLUDE in subset");
    ButtonGroup group = new ButtonGroup();
    group.add(remove_rels);
    group.add(keep_rels);

    remove_rels.setSelected(removeSelectedRelationships());
    remove_rels.addActionListener(
      new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          setRemoveSelectedRelationships(true);
          MMSToolkit.getGui().resetTitleForFilter(panel, "Relationships to Exclude");
        }
      }
    );

    toggle_panel.add(remove_rels, BorderLayout.CENTER);

    keep_rels.setSelected(!removeSelectedRelationships());
    keep_rels.addActionListener(
      new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          setRemoveSelectedRelationships(false);
          MMSToolkit.getGui().resetTitleForFilter(panel, "Relationships to Include");
        }
      }
    );

    toggle_panel.add(keep_rels, BorderLayout.SOUTH);

    relationship_type_table = new SortableJTable(getModel());
    relationship_type_table.getTableHeader().getColumnModel().getColumn(0).setPreferredWidth(300);
    relationship_type_table.getTableHeader().getColumnModel().getColumn(1).setPreferredWidth(150);
    relationship_type_table.getTableHeader().getColumnModel().getColumn(2).setPreferredWidth(75);

    listener = new RelationshipTypesToRemoveListSelectionListener(this,
      relationship_type_table);
    relationship_type_table.getSelectionModel().addListSelectionListener(listener);


    if (removeSelectedRelationships())
      panel = MMSToolkit.makeCard(relationship_type_table, " Relationships to Exclude ");
    else
      panel = MMSToolkit.makeCard(relationship_type_table, " Relationships to Include ");
    full_panel.setLayout(new BorderLayout());
    full_panel.add(toggle_panel, BorderLayout.NORTH);
    full_panel.add(panel, BorderLayout.CENTER);
    return full_panel;
  }

  /**
   * This method is responsible for selecting the rows
   * of the table based on the properties for this filter.
   */
  public void setPanelSelections() {
    listener.setSelections(getRelationshipTypesToRemove());

    if(removeSelectedRelationships()) {
      MMSToolkit.getGui().resetTitleForFilter(panel, "Relationships to Exclude");
      remove_rels.setSelected(true);
    } else {
      MMSToolkit.getGui().resetTitleForFilter(panel, "Relationships to Include");
      keep_rels.setSelected(true);
    }
  }

  /**
   * Method provides the description of the filter's screen and
   * includes instructions on use.
   * @return {@link String} text of instructions for use of panel
   */
  public String getTabDescription() {
    return "Please select one or more relationship types to remove from or include in the " +
    "UMLS Metathesaurus.  This filter removes only relationship data and not " +
    "entire concepts from the output subset.";
  }

  /**
   * Returns the table model.
   * @return {@link NoEditTableModel}
   */
  private NoEditTableModel getModel() {
    ArrayList model = new ArrayList(relationship_types.size());
    for(int i=0; i<relationship_types.size(); i++) {
      RelType rt = (RelType)relationship_types.get(i);
      if(!sources_to_remove.contains(rt.getSource()))
	model.add(rt);
    }
    String[][] rel_data = new String[model.size()][3];
    for(int i=0; i< model.size(); i++) {
      rel_data[i][0] = ((RelType)model.get(i)).getSource().getOfficialName();
      rel_data[i][1] = ((RelType)model.get(i)).getSource().getVersionedSourceAbbreviation();
      rel_data[i][2] = ((RelType)model.get(i)).getType();
    }
    Object [] headers = {"Source", "Source Abbreviation", "Relationship Type"};
    return new NoEditTableModel(rel_data, headers);
  }

  /**
   * Method allows the user to reset the filter's settings
   * to defaults.
   */
  public void resetToDefaults() {
    setRelationshipTypesToRemove(new ArrayList());
    setRemoveSelectedRelationships(true);
  }

  /**
   * Method provides a {@link JPanel} that will be added to the main GUI
   * help menu to allow the user to consult a help window
   * that is particular to this filter.
   * @return {@link JPanel} to be added to the Options menu
   */
  public JPanel getHelpPanel() {
    // Wrap html documentation in a panel
    return SwingToolkit.getPanelForHTML(
      this.getClass().getClassLoader().getResource(
      "config/RelationshipTypesToRemoveHelp.html"));
  }

  /**
   * Method provides the filter with values from the default properties
   * file.  This is the first method called after the class is constructed.
   * @param c {@link ApplicationConfiguration} default values
   */
  public void setConfiguration(ApplicationConfiguration c) {
  }

  /**
   * Method returns all properties specific to the filter.
   * @return {@link Properties} only those directly restyed to the filter.
   *        They will start with the filter's class name which will be
   *        concatenated to the name of the property.
   *        Filter.property_name
   */
  public Properties getProperties() {
    // Rebuild relationship_types_to_remove property
    Properties new_filter_props = new Properties();
    List ltr = getRelationshipTypesToRemove();
    StringBuffer line = new StringBuffer();
    for(int i=0; i< ltr.size(); i++) {
      line.append((
        (RelType)ltr.get(i)).getSource().getRootSourceAbbreviation());
      line.append("|");
      line.append(((RelType)ltr.get(i)).getType());
      if(i < ltr.size() - 1)
	line.append(";");
    }
    new_filter_props.setProperty(
      "selected_relationship_types", line.toString());

    new_filter_props.setProperty("remove_selected_rels",
      new Boolean(removeSelectedRelationships()).toString());

    return new_filter_props;
  }

  /**
   * Method provides the filter with all properties specific to the
   * filter.
   * @param props {@link Properties}
   */
  public void setProperties(Properties props) {
    try {
      relationship_types.clear();
      BufferedReader in = new BufferedReader(
        new InputStreamReader(
          (InputStream)this.getClass().getClassLoader().getResource(
            "config/rel_types.dat").getContent()));
      String line;
      while ( (line = in.readLine()) != null) {
        RelType rt = new RelType(line);
        relationship_types.add(rt);
      }
      in.close();
    } catch (Exception e) {
      MMSToolkit.reportError(MMSToolkit.getGui(),
        "Relationship To Exclude filter requires a\n" +
        "rel_types.dat file to exist in the config directory.\n" +
        "This missing file will cause the configuration table\n" +
        "for this filter to be empty.", false);
    }

    // does not fire data changed
    // because it is in response to Init, Load, or Undo

    // Get the selected_relationship_types property
    // make backwards compatible to recognize relationship_types_to_remove
    if(props.containsKey("relationship_types_to_remove") ||
       props.containsKey("selected_relationship_types")) {

      // Clear existing settings
      relationship_types_to_remove.clear();

      // Set relationship_types to remove from property
      String selected_prop = "";
      if (props.containsKey("selected_relationship_types"))
        selected_prop = (String)props.get("selected_relationship_types");
      else
        selected_prop = (String)props.get("relationship_types_to_remove");
      StringTokenizer tokens = new StringTokenizer(selected_prop, ";");
      while(tokens.hasMoreTokens()) {
	String line = tokens.nextToken();
	Source source = Source.getSource(
          line.substring(0, line.indexOf('|')));
        if (source != null) {
          String type = line.substring(line.indexOf('|') + 1);
          RelType rel = new RelType(source, type);
          addRelationshipTypeToRemove(rel);
        }
      }
    }
    if(props.containsKey("remove_selected_rels")) {
      setRemoveSelectedRelationships(new Boolean((String)props.get(
        "remove_selected_rels")).booleanValue());
    }
    // we do not need to set tahble model of relationship_types_table because it
    // is handled by getPanel() and relationship_types property never changes
  }

  /**
   * Method which determines if a given CUI will be removed from the
   * release.
   * @param concept {@link Concept}
   */
  public void applySelections(Concept concept) {
    Relationship[] relationships = concept.getRelationships();
    for(int i=0; i<relationships.length; i++) {
      Relationship rel = relationships[i];
      if((removeSelectedRelationships() &&
        isRelationshipTypeToRemove(getRelationshipTypeForRelationship(rel)))
         ||
        (!removeSelectedRelationships() &&
        !isRelationshipTypeToRemove(getRelationshipTypeForRelationship(rel)))) {
          concept.removeRelationship(relationships[i]);
      }
    }
  }

  /**
   * Returns the name.
   * @return {@link String}
   */
  public String getName() {
    return "Relationship Type List";
  }

  /**
   * Method returns a list of the relationship types to be excluded for
   * display in the configuration file.
   * @return {@link String}
   */
  public String getConfigurationDetails() {
    StringBuffer sb = new StringBuffer();
    String ls = System.getProperty("line.separator");
    if (removeSelectedRelationships())
      sb.append("The following relationship types were excluded from the subset.").append(ls);
    else
      sb.append("The following relationship types were included in the subset.").append(ls);
    Iterator iter = relationship_types_to_remove.iterator();
    boolean empty_list = true;
    while (iter.hasNext()) {
      RelType rt = (RelType)iter.next();
      // don't print out RelationshipTypes if their sources are to be excluded from subset
      if(sources_to_remove.contains(rt.getSource()))
        continue;
      String sn = rt.getSource().getOfficialName();
      String sab = rt.getSource().getVersionedSourceAbbreviation();
      sb.append("  " +
                ( (sn.length() > 62) ? sn.substring(0, 59) + "..." : sn) +
          "                                                                    ".
          substring(0, 63 - ( (sn.length() > 62) ? 62 : sn.length())) +
          sab +
          "                                                                                  ".
          substring(0, 82 - (62 + sab.length())) +
          rt.getType()).append(ls);
      empty_list = false;
    }
    if (empty_list)
      sb.append("  <none>").append(ls);
    return sb.toString();
  }

  //
  // Filter Specific Methods
  //

  /**
   * This method returns the {@link List} containing all {@link RelType}s.
   * @return {@link List} of {@link RelType}s
   */
  public List getRelationshipTypes() {
    return relationship_types;
  }

  /**
   * This method returns the {@link List} containing all
   * {@link RelType}s that share the same {@link Source}.
   * @param s {@link Source}
   * @return {@link List} of {@link RelType}s
   */
   public List getRelationshipTypes(Source s) {
     ArrayList list = new ArrayList();
     for(int i=0; i<relationship_types.size(); i++) {
       if(((RelType)relationship_types.get(i)).getSource().equals(s))
       list.add((RelType)relationship_types.get(i));
     }
     return list;
   }

  /**
   * This method returns the {@link Set} containing all
   * {@link RelType}s that share the same {@link Source}.
   * @param s {@link Source}
   * @return  {@link Set} of {@link RelType}s
   */
  public Set getRelTypesAsSet(Source s) {
    return new TreeSet(getRelationshipTypes(s));
  }

  /**
   * This method clears the data structure that holds the list of relationship
   * types to be excluded.
   */
  public void clearRelationshipTypesToRemove() {
    relationship_types_to_remove.clear();
  }

  /**
   * This method adds a {@link RelType} to the data structure that holds
   * the list of relationship types to be excluded.
   * @param relationship_type {@link RelType}
   */
  public void addRelationshipTypeToRemove(RelType relationship_type) {
    if(!relationship_types_to_remove.contains(relationship_type))
      relationship_types_to_remove.add(relationship_type);
  }

  /**
   * This method removes a relationship_type from the data structure that
   * holds the list of relationship_types to be excluded.
   * @param relationship_type {@link RelType}
   */
  public void removeRelationshipTypeToRemove(RelType relationship_type) {
    relationship_types_to_remove.remove(relationship_type);
  }

  /**
   * This method returns true if the given {@link RelType}
   * is on the list of those to be exluded from the subset.
   * @param relationship_type {@link RelType}
   * @return <code>boolean</code>
   */
  public boolean isRelationshipTypeToRemove(RelType relationship_type) {
    if(relationship_type == null)
      return false;
    return relationship_types_to_remove.contains(relationship_type);
  }

  /**
   * This method returns the {@link List} containing all
   * {@link RelType}s to be excluded.
   * @return {@link List}
   */
  public List getRelationshipTypesToRemove() {
    return relationship_types_to_remove;
  }

  /**
   * This method replaces the list of {@link RelType}s to exclude with
   * the one provided.
   * @param rtr {@link List} of {@link RelType}s
   */
  public void setRelationshipTypesToRemove(List rtr) {
    relationship_types_to_remove.clear();
    for(int i=0; i<rtr.size(); i++) {
      RelType r = (RelType)rtr.get(i);
      addRelationshipTypeToRemove(r);
    }
  }

  /**
   * Returns the {@link RelType} for the specified {@link Relationship}.
   * @param rel the specified {@link Relationship}
   * @return the {@link RelType} for the specified {@link Relationship}
   */
  private RelType getRelationshipTypeForRelationship(Relationship rel) {
    for(int i=0; i<relationship_types.size(); i++) {
      RelType this_rel_type =
        (RelType)relationship_types.get(i);
      if((this_rel_type.getType().indexOf(rel.getRel().getAbbreviation()) != -1) &&
	 (this_rel_type.getSource().equals(rel.getSource())))
	return this_rel_type;

    }
    return null;
  }

  /**
   * This method indicates whether the selections represent relationship types to
   * be excluded or relationship types to be kept.
   * @return <code>true</code> selected relationship types will be excluded
   * <code>false</code> selected relationship types will be kept
   */
  public boolean removeSelectedRelationships() {
    return remove_selected_rels;
  }

  /**
   * This method sets whether or not the selections represent relationship types to
   * be excluded or relationship types to be kept.
   * @param b <code>boolean</code>
   */
  public void setRemoveSelectedRelationships(boolean b) {
    remove_selected_rels = b;
  }
  /**
   * This method indicates whether the automatic_selection option has been chosen.
   * Automatic selection allows the application to select relationship types
   * sharing the same source without prompting the user.
   * @return automatic_selection <code>boolean</code><code>true</code>
   *				 <code>false</code>
   */
  public boolean getAutomaticSelection() {
    return user_config.getAutomaticSelection();
  }

  /**
   * Method provides the filter with values from the user configuration
   * file.
   * @param c {@link UserConfiguration}
   */
  public void setConfiguration(UserConfiguration c) {
    user_config = c;
  }

  //
  // SourcesToRemoveListener Implementation
  //
  /**
   * This method compensates for the {@link RelType}s with sources to be
   * excluded that are therefore not displayed in the relationship type table.
   * @param rt {@link RelType}
   * @return int row where given {@link RelType} is displayed
   */
  public int getRowForRelationshipType(RelType rt) {
    if(sources_to_remove.contains(rt.getSource()))
      return -1;
    int row =0;
    for(int i=0; i<relationship_types.size(); i++) {
      RelType rel = (RelType)relationship_types.get(i);
      if(!sources_to_remove.contains(rel.getSource())) {
	row++;
        if(rt.equals(rel))
 	  return relationship_type_table.reverseMapIndex(row-1);
      }
    }
    return -1;
  }

  /**
   * This method compensates for the {@link RelType}s with sources to be
   * excluded that are therefore not displayed in the attribute type table.
   * @param r <code>int</code> row number where {@link RelType} is requested
   * @return {@link RelType} at given row
   */
  public RelType getRelationshipTypeAtRow(int r) {
    int row = relationship_type_table.mapIndex(r);
    int subset_row =-1;
    RelType rel = null;
    for(int i=0; i<relationship_types.size(); i++) {
      rel = (RelType)relationship_types.get(i);
      if(!sources_to_remove.contains(rel.getSource())) {
	subset_row++;
      }
      if(subset_row == row)
	return rel;
    }
    return rel;
  }

  /**
   * This method is called when there is a change to the selections
   * on the sources to exclude tab.  It resets the relationship type
   * table model to only include those attribute types that are not
   * from excluded sources.
   * @param e {@link SourcesToRemoveEvent}
   */
  public void sourcesToRemoveChanged(SourcesToRemoveEvent e) {
    sources_to_remove = e.getSourcesToRemove();
    relationship_type_table.setModel(getModel());
    relationship_type_table.getTableHeader().getColumnModel().getColumn(0).setPreferredWidth(300);
    relationship_type_table.getTableHeader().getColumnModel().getColumn(1).setPreferredWidth(150);
    relationship_type_table.getTableHeader().getColumnModel().getColumn(2).setPreferredWidth(75);
    listener.setSelections(getRelationshipTypesToRemove());
  }

  /**
   * Method indicates the {@link Color} that should be used to display
   * the name.
   * @return {@link Color}
   */
  public Color getColor() {
    return null;
  }

  /**
   * Indicates if {@link Filter} can use the efficient mode during
   * the "Initialize Cuis" phase of the subsetting process.
   * The efficient mode reads only MRCONSO and MRSTY during the
   * "Initialize Cuis" phase.
   * @return <code>true</code> if efficient mode is acceptable
   *         <code>false</code> if efficient mode is not acceptable
   */
  public boolean isEfficientEnabled() {
    return true;
  }

}

