/***************************************************************************
 *
 * Package:	gov.nih.nlm.mms.filters
 * Object:	RelationshipTypesToRemoveListSelectionListener.java
 *
 **************************************************************************/

package gov.nih.nlm.mms.filters;

import gov.nih.nlm.swing.IgnorableListener;
import gov.nih.nlm.swing.SortableJTable;

import java.util.List;

import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

/**
 * List selection listener for the Relationship Types To Exclude table.
 * Initializes and updates selections on the the relationships table.
 * Makes selections based on user's options settings.
 *
 * @author Deborah Shapiro, Brian Carlsen
 */

public class RelationshipTypesToRemoveListSelectionListener
  implements ListSelectionListener, IgnorableListener {
  //
  // Private Fields
  //
  private SortableJTable table;
  private RelationshipTypesToRemoveFilter filter;
  private boolean ignore_events = false;

  //
  // Constructors
  //
  /**
   * Initializes selections on the sources that will be excluded.
   * @param filter {@link RelationshipTypesToRemoveFilter}
   * @param table {@link SortableJTable}
   */
  public RelationshipTypesToRemoveListSelectionListener(
    RelationshipTypesToRemoveFilter filter, SortableJTable table) {
    this.table = table;
    this.filter = filter;
  }

  /**
   * Sets selections on the {@link SortableJTable}
   * based on the relationship_types_to_remove data structure in the
   * {@link RelationshipTypesToRemoveFilter}.
   * Records initial set of selected relationship types in order to determine
   * at later points whether selections have changed.
   * @param rtr {@link List} of relationship types to be excluded from subset
   */
  public void setSelections(List rtr) {
    setIgnoreEvents(true);
    table.clearSelection();

    for(int i=0; i< rtr.size(); i++) {
      RelType s = (RelType)rtr.get(i);

      int row = filter.getRowForRelationshipType(s);
      if(row != -1) {
        table.addRowSelectionInterval(row,row);
      }
    }
    setIgnoreEvents(false);
  }

  //
  // Methods
  //
  /**
   * Method that recognizes a {@link ListSelectionEvent} and responds
   * by going through logic to determine if other relationship type rows should
   * be selected as well.
   * @param e {@link ListSelectionEvent}
   */
  public void valueChanged(ListSelectionEvent e) {

    if(e.getValueIsAdjusting()) return;

    // Only process events we care about
    if(ignoreEvents()) {
      return;
    }

    filter.fireDataChanged();

    for(int i = e.getFirstIndex(); i<=e.getLastIndex(); i++) {

      RelType this_rel= filter.getRelationshipTypeAtRow(i);

      //Set current selection for this row and update filter
      boolean selected = table.isRowSelected(i);
      if(selected)
      	filter.addRelationshipTypeToRemove(this_rel);
      else
      	filter.removeRelationshipTypeToRemove(this_rel);

    } // end from (first index, last index, etc.)
  }

  /**
   * Method returns <code>boolean</code> to indicate whether or not
   * {@link ListSelectionEvent}s are to be ignored.
   * @return <code>boolean</code>
   */
  public boolean ignoreEvents() {
    return ignore_events;
  }

  /**
   * Method that turns off responses to the {@link ListSelectionEvent}s.
   * @param b <code>boolean</code> <code>true</code> if events are to be ignored
   *				 <code>false</code> if events are to be processed
   */
  public void setIgnoreEvents(boolean b) {
    ignore_events = b;
  }
}

