/**************************************************************************
 *
 * Package:    gov.nih.nlm.mms.filters;
 * Interface:  RemoveByTermgroupFilter.java
 *
 *************************************************************************/
package gov.nih.nlm.mms.filters;

import gov.nih.nlm.mms.ApplicationConfiguration;
import gov.nih.nlm.mms.MMSToolkit;
import gov.nih.nlm.mms.SourcesToRemoveEvent;
import gov.nih.nlm.mms.SourcesToRemoveListener;
import gov.nih.nlm.swing.NoEditTableModel;
import gov.nih.nlm.swing.SortableJTable;
import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.Source;
import gov.nih.nlm.umls.meta.SourceTermType;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.StringTokenizer;

import javax.swing.ButtonGroup;
import javax.swing.JPanel;
import javax.swing.JRadioButton;

/**
 * Computes which content is removable based on source/term types and removes
 * it.
 * 
 * @author Deborah Shapiro
 */

public class RemoveByTermgroupFilter extends AbstractFilter implements
    SourcesToRemoveListener, ConceptContentSelector {

  //
  // List of user selected removable termgroups
  //
  private List removable_termgroups = new ArrayList();

  //
  // List of all termgroups
  //
  private List complete_termgroups = new ArrayList();

  //
  // List of default removable termgroups
  //
  private List default_removable_termgroups = new ArrayList();

  //
  // Subset of default removable termgroups
  // that are still selected (used to determine whether or
  // not user must confirm deselecting a certain termgroup)
  //
  private List confirm_default_removable_termgroups = new ArrayList();

  //
  // List of sources to remove
  //
  private List sources_to_remove = new ArrayList();

  //
  // GUI Elements
  //
  private SortableJTable termgroup_table;

  private RemoveByTermgroupListSelectionListener listener = null;

  private JPanel panel = null;

  private JPanel full_panel = null;

  private JRadioButton remove_tgs = null;

  private JRadioButton keep_tgs = null;

  private boolean remove_selected_tgs = true;

  //
  // Constructor
  //

  /**
   * Instantiates an empty {@link RemoveByTermgroupFilter}.
   */
  public RemoveByTermgroupFilter() {
  }

  //
  // Methods
  //

  /**
   * Returns the GUI panel that will display the filter's screen as one of the
   * tabs on the main GUI window.
   * 
   * @return {@link JPanel} that displays graphical interface for the filter
   */
  public JPanel getPanel() {
    if (full_panel != null)
      return full_panel;

    full_panel = new JPanel();

    JPanel toggle_panel = new JPanel(new BorderLayout());
    remove_tgs = new JRadioButton(
        "Select source term types to EXCLUDE from subset");
    keep_tgs = new JRadioButton("Select source term types to INCLUDE in subset");
    ButtonGroup group = new ButtonGroup();
    group.add(remove_tgs);
    group.add(keep_tgs);

    remove_tgs.setSelected(removeSelectedTermgroups());
    remove_tgs.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        setRemoveSelectedTermgroups(true);
        MMSToolkit.getGui().resetTitleForFilter(panel,
            "Source Term Types to Exclude");
      }
    });

    toggle_panel.add(remove_tgs, BorderLayout.CENTER);

    keep_tgs.setSelected(!removeSelectedTermgroups());
    keep_tgs.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        setRemoveSelectedTermgroups(false);
        MMSToolkit.getGui().resetTitleForFilter(panel,
            "Source Term Types to Include");
      }
    });

    toggle_panel.add(keep_tgs, BorderLayout.SOUTH);

    termgroup_table = new SortableJTable(getModel());
    termgroup_table.getTableHeader().getColumnModel().getColumn(0)
        .setPreferredWidth(300);
    termgroup_table.getTableHeader().getColumnModel().getColumn(1)
        .setPreferredWidth(150);
    termgroup_table.getTableHeader().getColumnModel().getColumn(2)
        .setPreferredWidth(75);
    listener = new RemoveByTermgroupListSelectionListener(this, termgroup_table);
    termgroup_table.getSelectionModel().addListSelectionListener(listener);

    if (removeSelectedTermgroups())
      panel = MMSToolkit.makeCard(termgroup_table,
          " Source Term Types to Exclude ");
    else
      panel = MMSToolkit.makeCard(termgroup_table,
          " Source Term Types to Include ");
    full_panel.setLayout(new BorderLayout());
    full_panel.add(toggle_panel, BorderLayout.NORTH);
    full_panel.add(panel, BorderLayout.CENTER);
    return full_panel;

  }

  /**
   * Returns the table model.
   * 
   * @return {@link NoEditTableModel}
   */
  private NoEditTableModel getModel() {
    complete_termgroups = default_configuration.getSourceTermTypes();
    final List model = new ArrayList(complete_termgroups.size());
    for (int i = 0; i < complete_termgroups.size(); i++) {
      final SourceTermType tg = (SourceTermType) complete_termgroups.get(i);
      if (!sources_to_remove.contains(tg.getSource()))
        model.add(tg);
    }
    final String[][] tg_data = new String[model.size()][3];
    for (int i = 0; i < model.size(); i++) {
      final SourceTermType stt = (SourceTermType) model.get(i);
      tg_data[i][0] = stt.getSource().getOfficialName();
      tg_data[i][1] = stt.getSource().getVersionedSourceAbbreviation();
      tg_data[i][2] = stt.getTermType().getAbbreviation();
    }
    final Object[] headers = { "Source", "Source Abbreviation", "Term Type" };
    return new NoEditTableModel(tg_data, headers);
  }

  /**
   * Selects rows on the source/term type table for rows of termgroups that
   * should be removed.
   */
  public void setPanelSelections() {

    listener.setSelections();

    if (removeSelectedTermgroups()) {
      MMSToolkit.getGui().resetTitleForFilter(panel,
          "Source Term Types to Exclude");
      remove_tgs.setSelected(true);
    } else {
      MMSToolkit.getGui().resetTitleForFilter(panel,
          "Source Term Types to Include");
      keep_tgs.setSelected(true);
    }
  }

  /**
   * Returns the description.
   * 
   * @return the description
   */
  public String getTabDescription() {
    return "This screen contains the Source Term Type Filter, which specifies "
        + "source term type combinations to be excluded or included in the subset.  "
        + "Users can customize "
        + "the subset by selecting and deselecting source term type combinations.  ";
  }

  /**
   * Resets state to defaults.
   */
  public void resetToDefaults() {
    setRemovableTermgroups(new ArrayList(default_removable_termgroups));
    confirm_default_removable_termgroups = new ArrayList(
        default_removable_termgroups);
    setRemoveSelectedTermgroups(true);
  }

  /**
   * Shows an advanced options GUI to the user.
   */
  public void showOptions() {
  }

  /**
   * Indicates whether or not an options menu item should be supplied for
   * advanced configuration options.
   * 
   * @return <code>true</code> if so; <code>false</code> otherwise
   */
  public boolean hasAdvancedOptions() {
    return false;
  }

  /**
   * Returns a help {@link JPanel}.
   * 
   * @return a help {@link JPanel}
   */
  public JPanel getHelpPanel() {
    return null;
  }

  /**
   * Sets the {@link ApplicationConfiguration}. Used for "default" selections.
   * 
   * @param c
   *          the {@link ApplicationConfiguration}
   */
  public void setConfiguration(ApplicationConfiguration c) {
    default_configuration = c;

    //
    // What are the default removed termgroups.
    //
    default_removable_termgroups = new ArrayList();
    confirm_default_removable_termgroups = new ArrayList();
    if (default_configuration.getProperty(this, "removed_termgroups") != null) {
      final String remove_prop = default_configuration.getProperty(this,
          "removed_termgroups");
      final StringTokenizer tokens = new StringTokenizer(remove_prop, ";");
      while (tokens.hasMoreTokens()) {
        final String line = tokens.nextToken();
        final SourceTermType ts = (default_configuration.getSourceTermType(line
            .substring(0, line.indexOf('|')), line
            .substring(line.indexOf('|') + 1)));
        default_removable_termgroups.add(ts);
        confirm_default_removable_termgroups.add(ts);
      }
    }
  }

  /**
   * Returns filter state.
   * 
   * @return filter state as {@link Properties}
   */
  public Properties getProperties() {
    final Properties new_filter_props = new Properties();
    final List stgs = getRemovableTermgroups();
    final StringBuffer line = new StringBuffer();
    if (stgs != null) {
      for (int i = 0; i < stgs.size(); i++) {
        final SourceTermType stg = (SourceTermType) stgs.get(i);
        line.append(stg.getSource().getRsab().toString());
        line.append("|");
        line.append(stg.getTermType().getAbbreviation());
        if (i <= stgs.size() - 2)
          line.append(";");
      }
    }
    new_filter_props.setProperty("removed_termgroups", line.toString());

    new_filter_props.setProperty("remove_selected_termgroups", new Boolean(
        removeSelectedTermgroups()).toString());

    return new_filter_props;
  }

  /**
   * Sets filter state.
   * 
   * @param p
   *          {@link Properties} representing filter state
   */
  public void setProperties(Properties p) {

    //
    // What are the removed termgroups
    //
    if (p.containsKey("removed_termgroups")) {
      removable_termgroups = new ArrayList();
      final String remove_prop = p.getProperty("removed_termgroups");
      final StringTokenizer tokens = new StringTokenizer(remove_prop, ";");
      while (tokens.hasMoreTokens()) {
        final String line = tokens.nextToken();
        final SourceTermType tg = (default_configuration.getSourceTermType(line
            .substring(0, line.indexOf('|')), line
            .substring(line.indexOf('|') + 1)));
        if (tg != null) {
          // tg.setSuppress(true);
          removable_termgroups.add(tg);
        }
      }
    }
    if (p.containsKey("remove_selected_termgroups")) {
      setRemoveSelectedTermgroups(new Boolean((String) p
          .get("remove_selected_termgroups")).booleanValue());
    }

  }

  /**
   * Sets the suppressibility of atoms and removes suppressible content if
   * advanced options indicate to do so.
   * 
   * @param concept
   *          the current {@link Concept}
   */
  public void applySelections(Concept concept) {
    final Atom[] atoms = concept.getAtoms();
    for (int i = atoms.length - 1; i >= 0; i--) {
      final Atom atom = (Atom) atoms[i];
      final SourceTermType tg = (default_configuration.getSourceTermType(atom
          .getSource().getRsab(), atom.getTermType().getAbbreviation()));
      if ((removeSelectedTermgroups() && getRemovableTermgroups().contains(tg))
          || (!removeSelectedTermgroups() && !getRemovableTermgroups()
              .contains(tg))) {
        concept.removeAtom(atom);
      }
    }
  }

  /**
   * Returns the filter name.
   * 
   * @return the filter name
   */
  public String getName() {
    return "Source Term Type List";
  }

  /**
   * This method indicates whether the automatic_selection option has been
   * chosen. Automatic selection allows the application to select relationship
   * types sharing the same source without prompting the user.
   * 
   * @return automatic_selection <code>boolean</code><code>true</code>
   *				 <code>false</code>
   */
  public boolean getAutomaticSelection() {
    return user_configuration.getAutomaticSelection();
  }

  /**
   * Returns configuration details for use in <code>mmsys.log</code>.
   * 
   * @return configuration details for use in <code>mmsys.log</code>
   */
  public String getConfigurationDetails() {
    final StringBuffer sb = new StringBuffer();
    final String ls = System.getProperty("line.separator");
    if (removeSelectedTermgroups())
      sb.append("Source Term Type whose terms were removed").append(ls);
    else
      sb.append("Source Term Type whose terms were included").append(ls);
    final Iterator iter = removable_termgroups.iterator();
    boolean empty_list = true;
    while (iter.hasNext()) {
      final SourceTermType tg = (SourceTermType) iter.next();
      // don't print out SourceTermTypes if their sources are to be excluded
      // from subset
      if (sources_to_remove.contains(tg.getSource()))
        continue;
      final String sn = tg.getSource().getOfficialName();
      sb
          .append(
              "  "
                  + ((sn.length() > 67) ? sn.substring(0, 64) + "..." : sn)
                  + "                                                                  "
                      .substring(0,
                          68 - ((sn.length() > 67) ? 67 : sn.length()))
                  + tg.getSource().getVersionedSourceAbbreviation() + "|"
                  + tg.getTermType().getAbbreviation()).append(ls);
      empty_list = false;
    }

    if (empty_list)
      sb.append("  <none>").append(ls);
    sb.append(ls);

    return sb.toString();
  }

  /**
   * Clears the list of removable {@link SourceTermType}s.
   */
  protected void clearRemovableTermgroups() {
    removable_termgroups.clear();
  }

  /**
   * Adds a {@link SourceTermType} to the list of removable termgroups.
   * 
   * @param tg
   *          the {@link SourceTermType}
   */
  protected void addRemovableTermgroup(SourceTermType tg) {
    if (!removable_termgroups.contains(tg)) {
      // tg.setIsSuppressible(true);
      removable_termgroups.add(tg);
    }
  }

  /**
   * Removes a {@link SourceTermType} from the list of removable termgroups.
   * 
   * @param tg
   *          {@link SourceTermType}
   */
  protected void removeRemovableTermgroup(SourceTermType tg) {
    removable_termgroups.remove(tg);
    // tg.setIsSuppressible(false);
  }

  /**
   * Returns an {@link List} containing all {@link SourceTermType}s to be
   * removed from the default termgroup list.
   * 
   * @return {@link List} of removed termgroups
   */
  public List getDefaultRemovableTermgroups() {
    return default_removable_termgroups;
  }

  /**
   * Returns a list of {@link SourceTermType}s that have been confirmed by the
   * user via a dialog box to be removable. Prevents user from being queried
   * twice about a termgroup's removability.
   * 
   * @return {@link List} of removed termgroups
   */
  public List getConfirmDefaultRemovableTermgroups() {
    return confirm_default_removable_termgroups;
  }

  /**
   * Clears the list of {@link SourceTermType}s that have been confirmed by the
   * user to be removable.
   */
  public void clearConfirmDefaultRemovableTermgroups() {
    confirm_default_removable_termgroups.clear();
  }

  /**
   * This method indicates whether the selections represent source term types to
   * be excluded or source term types to be kept.
   * 
   * @return <code>true</code> selected source term types will be excluded
   *         <code>false</code> selected source term types will be kept
   */
  public boolean removeSelectedTermgroups() {
    return remove_selected_tgs;
  }

  /**
   * This method sets whether or not the selections represent source term types
   * to be excluded or source term types to be kept.
   * 
   * @param b
   *          <code>boolean</code>
   */
  public void setRemoveSelectedTermgroups(boolean b) {
    remove_selected_tgs = b;
  }

  /**
   * This method returns the {@link List} containing all {@link SourceTermType}s.
   * 
   * @return {@link List} of {@link SourceTermType}s
   */
  public List getSourceTermTypes() {
    return complete_termgroups;
  }

  /**
   * This method returns the {@link List} containing all {@link SourceTermType}s
   * that share the same {@link Source}.
   * 
   * @param s
   *          {@link Source}
   * @return {@link List} of {@link RelType}s
   */
  public List getSourceTermTypes(Source s) {
    ArrayList list = new ArrayList();
    for (int i = 0; i < complete_termgroups.size(); i++) {
      if (((SourceTermType) complete_termgroups.get(i)).getSource().equals(s))
        list.add((SourceTermType) complete_termgroups.get(i));
    }
    return list;
  }

  /**
   * Sets the removable termgroups list with the one provided.
   * 
   * @param stgs
   *          the {@link List} of {@link SourceTermType}s
   */
  public void setRemovableTermgroups(List stgs) {
    removable_termgroups = stgs;
  }

  /**
   * Returns the {@link List} containing all removable {@link SourceTermType}s.
   * 
   * @return {@link List} of removable termgroups
   */
  public List getRemovableTermgroups() {
    if (removable_termgroups == null)
      return new ArrayList();
    return removable_termgroups;
  }

  /**
   * Returns the table that includes {@link SourceTermType}s with the removable
   * termgroups highlighted.
   * 
   * @return {@link SortableJTable}
   */
  public SortableJTable getRemovableTermgroupsTable() {
    return termgroup_table;
  }

  //
  // SourcesToRemoveListener Implementation
  //

  /**
   * Resets the removability table when the user tabs between filters. Removes
   * entries visible in the table based on which sources are to be excluded.
   * 
   * @param e
   *          {@link SourcesToRemoveEvent}
   */
  public void sourcesToRemoveChanged(SourcesToRemoveEvent e) {
    sources_to_remove = e.getSourcesToRemove();
    termgroup_table.setModel(getModel());
    termgroup_table.getTableHeader().getColumnModel().getColumn(0)
        .setPreferredWidth(300);
    termgroup_table.getTableHeader().getColumnModel().getColumn(1)
        .setPreferredWidth(150);
    termgroup_table.getTableHeader().getColumnModel().getColumn(2)
        .setPreferredWidth(75);
    setPanelSelections();
  }

  /**
   * Maps a {@link SourceTermType} to a row number in the table.
   * 
   * This is necessary because only {@link SourceTermType}s from sources that
   * are not to be excluded will be visible in the table.
   * 
   * @param tg
   *          the {@link SourceTermType}
   * @return <code>int</code> row number of termgroup in table -1 if termgroup
   *         is to be excluded
   */
  public int getRowForTermgroup(SourceTermType tg) {
    if (sources_to_remove.contains(tg.getSource()))
      return -1;
    int row = 0;
    for (int i = 0; i < complete_termgroups.size(); i++) {
      final SourceTermType termgroup = (SourceTermType) complete_termgroups
          .get(i);
      if (!sources_to_remove.contains(termgroup.getSource())) {
        row++;
        if (tg.equals(termgroup))
          return termgroup_table.reverseMapIndex(row - 1);
      }
    }
    return -1;
  }

  /**
   * Maps a row number to the {@link SourceTermType}. Adjusts for the
   * {@link SourceTermType}s that are not currently visible in the table
   * because their sources are to be excluded.
   * 
   * @param tg
   *          <code>int</code> row in the table
   * @return {@link SourceTermType} located at that row in the table
   */
  public SourceTermType getTermgroupAtRow(int tg) {
    int row = termgroup_table.mapIndex(tg);
    int subset_row = -1;
    SourceTermType termgroup = null;
    for (int i = 0; i < complete_termgroups.size(); i++) {
      termgroup = (SourceTermType) complete_termgroups.get(i);
      if (!sources_to_remove.contains(termgroup.getSource())) {
        subset_row++;
      }
      if (subset_row == row)
        return termgroup;
    }
    return termgroup;
  }

  /**
   * Returns the {@link Color} that should be used to display the name.
   * 
   * @return the {@link Color}
   */
  public Color getColor() {
    return null;
  }

  /**
   * Indicates if {@link Filter} can use the efficient mode during the
   * "Initialize Cuis" phase of the subsetting process. The efficient mode reads
   * only MRCONSO and MRSTY during the "Initialize Cuis" phase.
   * 
   * @return <code>true</code>
   */
  public boolean isEfficientEnabled() {
    return true;
  }

}
