/**************************************************************************
 *
 * Package:	gov.nih.nlm.mms.filters
 * Object:	RemoveByTermgroupListSelectionListener.java
 *
 **************************************************************************/

package gov.nih.nlm.mms.filters;
import gov.nih.nlm.mms.MMSToolkit;
import gov.nih.nlm.swing.IgnorableListener;
import gov.nih.nlm.swing.ListDialog;
import gov.nih.nlm.swing.SortableJTable;
import gov.nih.nlm.swing.TableSortEvent;
import gov.nih.nlm.swing.TableSortListener;
import gov.nih.nlm.umls.meta.SourceTermType;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;

import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

/**
 * List selection listener for the remove term types table.
 * Initializes and updates selections on the the removable term types table.
 *
 * @author Deborah Shapiro, Brian Carlsen
 */

public class RemoveByTermgroupListSelectionListener
  implements ListSelectionListener, IgnorableListener, TableSortListener{

  //
  // Private Fields
  //

  //
  // Table
  //
  private SortableJTable table;

  //
  // Filter
  //
  private RemoveByTermgroupFilter filter;

  //
  // flag indicating whether or not events are being ignored
  //
  private boolean ignore_events = false;

  private boolean[] selection_before;

  //
  // Constructors
  //
  /**
   * Instantiates a {@link RemoveByTermgroupListSelectionListener}
   * with the specified {@link RemoveByTermgroupFilter} and {@link SortableJTable}.
   * @param filter the parent filter {@link RemoveByTermgroupFilter}
   * @param table the {@link SortableJTable} of {@link SourceTermType}s
   */
  public RemoveByTermgroupListSelectionListener(
    RemoveByTermgroupFilter filter, SortableJTable table) {
    this.table = table;
    this.filter = filter;
  }

  //
  // Methods
  //


  public void setSelections() {
    setIgnoreEvents(true);

    table.clearSelection();
    selection_before = new boolean[table.getModel().getRowCount()];

    final List stgs = new ArrayList(filter.getRemovableTermgroups());
    for (int i = 0; i < stgs.size(); i++) {
      final SourceTermType tg = (SourceTermType) stgs.get(i);
      //
      // check if row should be visible in table
      //
      int row = filter.getRowForTermgroup(tg);
      if (row != -1) {
        addRemovableTermgroup(row);
        selection_before[table.mapIndex(row)] = true;
      }
    }

    setIgnoreEvents(false);
  }



  /**
   * Listens for changes in the list and validates them.
   * @param e the {@link ListSelectionEvent}
   */
  public void valueChanged(ListSelectionEvent e) {
    if (e.getValueIsAdjusting())
      return;

    // Only process events we care about
    if (ignoreEvents()) {
      return;
    }

    filter.fireDataChanged();

    //logic to avoid calling up multiple confirm_tgs
    //dialog boxes when user doesn't Cntrl + click and only one
    //source row is selected
    int single_row = -1;
    if (table.getSelectedRowCount() == 1) {
      single_row = table.getSelectedRow();
    }

    HashSet sources_seen = new HashSet();

    for (int i = e.getFirstIndex(); i <= e.getLastIndex(); i++) {

      SourceTermType this_tg = filter.getTermgroupAtRow(i);

      //Set current selection for this row and update filter
      boolean selected = table.isRowSelected(i);
      if (selected)
        filter.addRemovableTermgroup(this_tg);
      else
        filter.removeRemovableTermgroup(this_tg);

        //Determine if selection changed
      boolean selection_changed =
          (selection_before[table.mapIndex(i)] != selected);

      //Set selection_before for the next event
      selection_before[table.mapIndex(i)] = selected;

      HashSet set = new HashSet();
      Object[] related_tgs = null;
      // Do not continue if selection for this row didn't change.
      // single_row variable prevents dialog box from
      // being opened repeatedly when a single row is clicked
      // thereby keeping all the source term types previously selected

      if (selection_changed && (single_row == -1 || single_row == i)) {

        set.addAll(filter.getSourceTermTypes(this_tg.getSource()));

        if (set.contains(this_tg))
          set.remove(this_tg);

          // related_tgs should be given all sources in set b/c if
          // automatic selection is chosen, user will not be given a choice
          // of which related sources to select
        related_tgs = set.toArray();

        if (!filter.getAutomaticSelection() &&
            !sources_seen.contains(this_tg.getSource())) {

          // get confirm_tgs set
          SourceTermType[] confirm_tgs =
              (SourceTermType[]) set.toArray(new SourceTermType[0]);
          Arrays.sort(confirm_tgs);

          // if you are removing source term types and all related rows are already
          // selected, don't raise the message
          // likewise if you are keeping source term types and all related rows are
          // already unselected, don't raise the message
          boolean raise_message = false;
          if (table.isRowSelected(i)) {
            for (int j = 0; j < confirm_tgs.length; j++) {
              int k = filter.getRowForTermgroup(confirm_tgs[j]);
              if (!table.isRowSelected(k))
                raise_message = true;
            }
          }
          else {
            for (int j = 0; j < confirm_tgs.length; j++) {
              int k = filter.getRowForTermgroup(confirm_tgs[j]);
              if (table.isRowSelected(k))
                raise_message = true;
            }
          }

          Object[] end_tg_names = null;
          if (!set.isEmpty() && raise_message) {
            String message = "";
            if ( (table.isRowSelected(i) && filter.removeSelectedTermgroups()) ||
                (!table.isRowSelected(i) && !filter.removeSelectedTermgroups())) {
              message = "<html><b>You have selected " +
                  this_tg.getSource().getVersionedSourceAbbreviation() + "|" +
                  this_tg.getTermType() +
                  " for removal, the following source term types</b><br> <b> typically should be removed with " +
                  this_tg.getSource().getVersionedSourceAbbreviation() + "|" +
                  this_tg.getTermType() +
                  ".  Choose which ones should also be removed.</b></ul>";
            }
            else {
              message = "<html><b>You have chosen to keep source term type " +
                  this_tg.getSource().getVersionedSourceAbbreviation() + "|" +
                  this_tg.getTermType() +
                  ".  The following source term types </b><br> <b> typically should be kept with " +
                  this_tg.getSource().getVersionedSourceAbbreviation() + "|" +
                  this_tg.getTermType() +
                  ".  Choose which ones should also be kept.</b></ul>";
            }

            //
            // Look up source term type names
            //
            final String[] start_tg_names = new String[confirm_tgs.length];
            for (int k1 = 0; k1 < start_tg_names.length; k1++)
              start_tg_names[k1] = confirm_tgs[k1].getSource().getVsab() + "|" +
                  confirm_tgs[k1].getTermType().getAbbreviation();
            Arrays.sort(start_tg_names);

            end_tg_names =
                ListDialog.showListMultipleMode(MMSToolkit.getGui(), message,
                                                "Confirm Selection",
                                                start_tg_names, null);
          }

          // Cancel was pressed in dialog or none were chosen
          if (related_tgs == null)
            continue;


          //
          // Take selected sty names and look back up the sty objects
          //
          if (end_tg_names != null) {
            related_tgs = new SourceTermType[end_tg_names.length];
            int ct = 0;
            for (int k1 = 0; k1 < confirm_tgs.length; k1++) {
              String name = confirm_tgs[k1].getSource().getVsab() + "|" +
                  confirm_tgs[k1].getTermType().getAbbreviation();
              for (int k2 = 0; k2 < end_tg_names.length; k2++)
                if (name.equals((String) end_tg_names[k2]))
                  related_tgs[ct++] = confirm_tgs[k1];
            }
          }
        }

        sources_seen.add(this_tg.getSource());

        //remove or maintain from the table based on dialog box results
        for (int j = 0; j < related_tgs.length; j++) {

          int k =
              filter.getRowForTermgroup( (SourceTermType) related_tgs[j]);

          // Unselect if selected
          if (k > -1 && table.isRowSelected(k) && !selected) {
            // prevent dialog box from being raised for each
            // of the other sources with the same source term type
            selection_before[table.mapIndex(k)] = false;
            removeRemovableTermgroup(k);
            // Select if unselected
          }
          else if (k > -1 && !table.isRowSelected(k) && selected) {
            selection_before[table.mapIndex(k)] = true;
            addRemovableTermgroup(k);
          }
        }
      } // end (if selection changed or only a single row is selected)
    } // end from (first index, last index, etc.)






/*
    //
    // prevents proceding through the code twice
    // once for the event related to deselecting a row(s)
    // once for the event related to selecting a row(s)
    //
    if(e.getValueIsAdjusting()) return;

    //
    // Return if we are ignoring events
    //
    if(ignoreEvents()) return;

    //
    // Handle the deselection of default removable termgroups
    //
    boolean first_loop = true;
    boolean user_choice = false;
    boolean data_change_fired = false;
    for(int i = 0; i<table.getRowCount(); i++) {

      //
      // Is this a default removable termgroup?
      //
      if(!table.isRowSelected(i) &&
         filter.getConfirmDefaultRemovableTermgroups().
         contains(filter.getTermgroupAtRow(i))) {

        //
        // don't want to bring up dialog for each of the
        // unselected default termgroup lines
        //
        user_choice = MMSToolkit.confirmRequest(
          "Your selection will cause the default selections to be \n" +
          "unselected.  Are you sure that you want to do this?");

        //
        // if user doesn't want to proceed to unselect defaults
        // update the GUI so the previously selected lines are selected again
        //
        if (!user_choice) {
          final List stgs = filter.getRemovableTermgroups();
          for (int j = 0; j < stgs.size(); j++) {
            final int row = filter.getRowForTermgroup((SourceTermType)stgs.get(j));
            final ListSelectionModel lsm =
                  (ListSelectionModel)table.getSelectionModel();
            setIgnoreEvents(true);
            lsm.addSelectionInterval(row, row);
            setIgnoreEvents(false);
          }
          return;
        }
        break;
      }
    }

    //
    // Push event onto undo stack
    //
    filter.fireDataChanged();

    //
    // Clear everything in filter and
    // add back in any rows that are selected
    //
    filter.clearRemovableTermgroups();

    final ListSelectionModel lsm = (ListSelectionModel)table.getSelectionModel();
    if (lsm.getMinSelectionIndex() == lsm.getMaxSelectionIndex())
      filter.clearConfirmDefaultRemovableTermgroups();
    for (int i = lsm.getMinSelectionIndex(); i <= lsm.getMaxSelectionIndex(); i++) {

      //
      // if nothing is selected, done
      //
      if (i == -1) break;

      //
      // If row is selected, add it as a removable termgroup
      //
      if (lsm.isSelectedIndex(i)) {
        filter.addRemovableTermgroup(filter.getTermgroupAtRow(i));
      }

      //
      // If row is unselected AND it is a default removable
      // remove it from the ConfirmDefaultSuppresibleTermgroups list
      //
      if (!lsm.isSelectedIndex(i) &&
          filter.getConfirmDefaultRemovableTermgroups().
          contains(filter.getTermgroupAtRow(i))) {
        final SourceTermType tg = filter.getTermgroupAtRow(i);
        filter.getConfirmDefaultRemovableTermgroups().
            remove(filter.getTermgroupAtRow(i));
      }
    }*/
  }

  /**
   * Returns <code>boolean</code> to indicate whether or not
   * {@link ListSelectionEvent}s are to be ignored.
   * @return <code>boolean</code>
   */
  public boolean ignoreEvents() {
    return ignore_events;
  }

  /**
   * Turns off responses to the {@link ListSelectionEvent}s.
   * @param b <code>boolean</code> <code>true</code> if events are to be ignored
   *				 <code>false</code> if events are to be processed
   */
  public void setIgnoreEvents(boolean b) {
    ignore_events = b;
  }

  /**
   * Selects a row and adds the {@link SourceTermType} on that row to the list of
   * removable {@link SourceTermType}s.
   * @param row number to be selected and removed
   */
  public void addRemovableTermgroup(int row) {
    table.addRowSelectionInterval(row, row);
    filter.addRemovableTermgroup(filter.getTermgroupAtRow(row));
  }

  /**
   * Deselects row and removes the {@link SourceTermType} on that row from the list of
   * removable {@link SourceTermType}s.
   * @param row number to be deselected and not removed
   */
  public void removeRemovableTermgroup(int row) {
    table.removeRowSelectionInterval(row, row);
    filter.removeRemovableTermgroup(filter.getTermgroupAtRow(row));
  }

  //
  // TableSortListener Implementation
  //

  /**
   * Handle start of table sorting.
   * @param e {@link TableSortEvent}
   */
  public void tableSortStarted(TableSortEvent e) {
    setIgnoreEvents(true);
  }

  /**
   * Handle finish of table sorting.
   * @param e {@link TableSortEvent}
   */
  public void tableSortFinished(TableSortEvent e) {
    setIgnoreEvents(false);
  }

}

