/**************************************************************************
 *
 * Package:    gov.nih.nlm.mms.filters;
 * Interface:  RxNormFilter.java
 *
 *************************************************************************/
package gov.nih.nlm.mms.filters;

import gov.nih.nlm.mms.ApplicationConfiguration;
import gov.nih.nlm.swing.SwingToolkit;
import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.SemanticType;
import gov.nih.nlm.umls.meta.Source;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Properties;

import javax.swing.JCheckBox;
import javax.swing.JPanel;
import javax.swing.border.CompoundBorder;
import javax.swing.border.EmptyBorder;
import javax.swing.border.EtchedBorder;
import javax.swing.border.TitledBorder;


/**
 *
 * Metathesaurus subset containing concepts connected in the RxNorm "graph"
 * along with their MTH-related concepts. From a spec by SJN, WTH, BAC & SS.
 *
 **/

public class RxNormFilter extends AbstractFilter implements ConceptSelector {

    //
    // Fields
    //
    private ApplicationConfiguration default_config;
    private JCheckBox apply_filter_button;
    private boolean apply_filter = false;

    //
    // Constructors
    //
    /**
     * Constructs a filter.
     */
    public RxNormFilter() {
    }

    /**
     * Returns an empty configuration panel.  This filter has nothing
     * to configure.
     * @return configuration {@link JPanel}
     */
    public JPanel getPanel(){
      JPanel panel = new JPanel(new BorderLayout());

      apply_filter_button =
        new JCheckBox("Apply RxNorm Filter.", true);

      TitledBorder titledBorder = new TitledBorder(
              new EtchedBorder(EtchedBorder.RAISED), "RxNorm Filter");
      titledBorder.setTitleJustification(TitledBorder.CENTER);
      panel.setBorder(new CompoundBorder(titledBorder,
                                         new EmptyBorder(10, 10, 10, 10)));


      apply_filter_button.setSelected(apply_filter);
      apply_filter_button.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          fireDataChanged();
          apply_filter = apply_filter_button.isSelected();
        }
      });

      panel.add(apply_filter_button, BorderLayout.NORTH);

      return panel;
    }

    /**
     * Method provides a {@link JPanel} that will be called from
     * the help menu to allow the user to reference a help window
     * that is particular to this filter.
     * @return {@link JPanel} to be added to the Help menu
     */
    public JPanel getHelpPanel() {
      return SwingToolkit.getPanelForHTML(
        this.getClass().getClassLoader().getResource(
        "config/RxNormHelp.html"));
    }

    /**
     * Return the description to appear above the main configuration
     * panel in the {#link MetamorphoSysGUI}.
     * @return tab description {@link String}
     */
    public String getTabDescription() {
	StringBuffer b = new StringBuffer();
        b.append("Restricts the subset to concepts that meet one of the following criteria: " +
                 "\na) RxNorm atoms, e.g., SCD " +
                 "\nc) Semantic Type = Drug Delivery Device. " +
                 "\nThe RxNorm default subset includes only RxNorm concepts from sources at " +
                 "restriction level 0.  Use the Source List tab to include or exclude RxNorm " +
                 "concepts from other sources.");
        return b.toString();
    }

    /**
     * There are no defaults because this filter has nothing to configure.
     */
    public void resetToDefaults() {
      apply_filter =
        Boolean.valueOf(default_config.getProperty(this, "apply_rxnorm_filter")).booleanValue();
    }


    /**
     * Gives the filter access to the default configuration object.
     * @param c {@link ApplicationConfiguration}
     */
    public void setConfiguration(ApplicationConfiguration c) {
      default_config = c;
    }

    /**
     * Returns a {@link Properties} object with the configuration.
     * @return a {@link Properties} object with the configuration.
     */
    public Properties getProperties() {
      Properties p = new Properties();
      p.setProperty("apply_rxnorm_filter", String.valueOf(apply_filter));
      return p;
    }

    /**
     * Not an active method, because this filter doesn't have any
     * filter specific properties.
     * @param props {@link Properties}
     */
    public void setProperties(Properties props) {
      if(props.containsKey("apply_rxnorm_filter"))
        apply_filter = Boolean.valueOf(props.getProperty("apply_rxnorm_filter")).booleanValue();
    }

    /**
     * Indicates if {@link Concept} should be selected to be kept for the
     * subset.
     * @param concept {@link Concept} to be evaluated for removal
     * @return <code>boolean</code> <code>true</code> if concept should be kept
     *          <code>false</code> otherwise
     */
    public boolean keep(Concept concept) {
      if (apply_filter) {
        boolean keep = false;
        Atom[] atoms = concept.getAtoms();

        for (int i = 0; i < atoms.length; i++) {
          Source source = atoms[i].getSource();
          if (source.getRootSourceAbbreviation().equals("RXNORM")) {
            keep = true;
            break;
          }
        }

        return keep || dddSemanticType(concept);
      }
      return true;
    }

    /**
     * Returns <code>true</code> if concept contains a {@link SemanticType}
     * of type "Drug Delivery Device".
     * @param concept {@link Concept} to be evaluated
     * @return <code>true</code> if {@link Concept} contains a {@link SemanticType}
     *   of type "Drug Delivery Device"
     *   <code>false</code> otherwise
     */
    private boolean dddSemanticType(Concept concept) {
      SemanticType[] stys = concept.getSemanticTypes();
      for (int i=0; i<stys.length; i++) {
	if (stys[i].getSty().equals("Drug Delivery Device"))
	  return true;
      }
      return false;
    }

    /**
     * Inactive method because there are no items to configure on this filter.
     */
    public void setPanelSelections() {
      // ignore events so that when undo calls this method it doesn't
      // add to the undo stack when resetting the JTextFields

      if(apply_filter)
        apply_filter_button.setSelected(true);
      else
        apply_filter_button.setSelected(false);

    }

    /**
     * Returns the name of the {@link Filter}.
     * @return name of the filter
     */
    public String getName() {
	return "RxNorm Filter";
    }

    /**
     * Returns the statement that should appear in the subset log file
     * regarding this {@link Filter}.
     * @return filter statement for log file
     */
    public String getConfigurationDetails() {
      if (apply_filter)
        return "Only concepts in the RxNorm graph were included." + System.getProperty("line.separator");
      else
        return "RxNorm filter was not applied." + System.getProperty("line.separator");
    }

  /**
   * Method indicates the {@link Color} that should be used to display
   * the name.
   * @return {@link Color}
   */
  public Color getColor() {
    return null;
  }

  /**
   * Indicates if {@link Filter} can use the efficient mode during
   * the "Initialize Cuis" phase of the subsetting process.
   * The efficient mode reads only MRCONSO and MRSTY during the
   * "Initialize Cuis" phase.
   * @return <code>true</code> if efficient mode is acceptable
   *         <code>false</code> if efficient mode is not acceptable
   */
  public boolean isEfficientEnabled() {
    return true;
  }
}

