/**************************************************************************
 *
 * Package:	gov.nih.nlm.mms.filters
 * Object:	SemanticTypesToRemoveListSelectionListener.java
 *
 **************************************************************************/

package gov.nih.nlm.mms.filters;

import gov.nih.nlm.mms.MMSToolkit;
import gov.nih.nlm.swing.IgnorableListener;
import gov.nih.nlm.swing.ListDialog;
import gov.nih.nlm.umls.meta.HierarchicalSemanticType;

import java.util.Arrays;
import java.util.HashSet;
import java.util.List;

import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

/**
 * List selection listener for the HierarchicalSemanticTypes To Exclude table.
 * Initializes and updates selections on the the sources table.
 * Makes selections based on user's options settings.
 *
 * @author Deborah Shapiro, Brian Carlsen
 */

public class SemanticTypesToRemoveListSelectionListener
  implements ListSelectionListener, IgnorableListener {
  //
  // Private Fields
  //
  private SemanticTypesToRemoveJTable table;
  private SemanticTypesToRemoveFilter filter;
  private boolean[] selection_before;
  private boolean ignore_events = false;

  //
  // Constructors
  //
  /**
   * Instantiates a {@link SemanticTypesToRemoveListSelectionListener} from the
   * specified filter and table.
   * @param filter the {@link SemanticTypesToRemoveFilter}
   * @param table the {@link SemanticTypesToRemoveJTable}
   */
  public SemanticTypesToRemoveListSelectionListener(
    SemanticTypesToRemoveFilter filter, SemanticTypesToRemoveJTable table) {
    this.table = table;
    this.filter = filter;
  }

  //
  // Methods
  //
  /**
   * Sets selections on the {@link SemanticTypesToRemoveJTable}
   * based on the semantic types data structure in the
   * {@link SemanticTypesToRemoveFilter}.
   * Records initial set of selected semantic types in order to determine
   * at later points whether selections have changed.
   * @param str {@link List} of {@link HierarchicalSemanticType}s to be excluded from subset
   */
  public void setSelections(List str) {
    setIgnoreEvents(true);

    //
    // Clear selections
    //
    table.clearSelection();

    //
    // Get selections before
    //
    selection_before = new boolean[table.getModel().getRowCount()];

    //
    // Iterate through list that should now be selected and and appropriate
    // row selections
    //
    for(int i=0; i< str.size(); i++) {
      final HierarchicalSemanticType s = (HierarchicalSemanticType)str.get(i);
      int row = table.getRowForSemanticType(s);
      table.addRowSelectionInterval(row,row);
      selection_before[table.mapIndex(table.getRowForSemanticType(s))] = true;
    }

    setIgnoreEvents(false);
  }

  /**
   * Handles {@link ListSelectionEvent}s.
   * @param e the {@link ListSelectionEvent}
   */
  public void valueChanged(ListSelectionEvent e) {

    if(e.getValueIsAdjusting()) return;

    //
    // Only process events we care about
    //
    if(ignoreEvents()) return;

    filter.fireDataChanged();

    //
    // logic to avoid calling up multiple confirm_stys
    // dialog boxes when user doesn't Cntrl + click and only one
    // source row is selected
    //
    int single_row = -1;
    if(table.getSelectedRowCount() == 1) {
      single_row = table.getSelectedRow();
    }

    //
    // Make selections for event
    //
    for(int i = e.getFirstIndex(); i<=e.getLastIndex(); i++) {

      final HierarchicalSemanticType this_sty= table.getSemanticTypeAtRow(i);

      //
      // Set current selection for this row and update filter
      //
      boolean selected = table.isRowSelected(i);
      if(selected)
	filter.addSemanticTypeToRemove(this_sty);
      else
	filter.removeSemanticTypeToRemove(this_sty);

      //
      // Determine if selection changed
      //
      boolean selection_changed = (selection_before[table.mapIndex(i)] != selected);

      //
      // Set selection_before for the next event
      //
      selection_before[table.mapIndex(i)] = selected;

      final HashSet set = new HashSet();
      Object[] related_stys = null;

      //
      // Do not continue if selection for this row didn't change.
      // single_row variable prevents dialog box from
      // being opened repeatedly when a single row is clicked
      // thereby keeping all the sources previously selected
      //
      if (selection_changed && (single_row == -1 || single_row == i)) {

        set.addAll(filter.default_configuration.
	    getSemanticTypeChildrenAsSet(this_sty.getStn()));

	if(set.contains(this_sty)) set.remove(this_sty);

        //
	// related_stys should be given all sources in set b/c if
	// automatic selection is chosen, user will not be given a choice
	// of which related sources to select
        //
	related_stys = set.toArray();

	if(!filter.getAutomaticSelection()) {

          //
	  // get confirm_stys set
          //
	  final HierarchicalSemanticType [] confirm_stys =
	    (HierarchicalSemanticType[])set.toArray(new HierarchicalSemanticType[0]);
	  Arrays.sort(confirm_stys);

          //
	  // if you are removing stys and all related rows are already selected,
	  // don't raise the message
	  // likewise if you are keeping stys and all related rows are already
	  // unselected, don't raise the message
          //
	  boolean raise_message = false;
	  if (table.isRowSelected(i)) {
	    for (int j = 0; j < confirm_stys.length; j++) {
	      int k = table.getRowForSemanticType(confirm_stys[j]);
	      if (!table.isRowSelected(k))
	        raise_message = true;
	    }
	  } else {
	    for(int j = 0; j < confirm_stys.length; j++) {
	      int k = table.getRowForSemanticType(confirm_stys[j]);
	      if(table.isRowSelected(k))
	        raise_message = true;
	    }
	  }

          Object[] end_sty_names = null;
	  if (!set.isEmpty() && raise_message) {
	    String message;
            if((table.isRowSelected(i) && filter.removeSelectedSemanticTypes()) ||
              (!table.isRowSelected(i) && !filter.removeSelectedSemanticTypes())) {
	      message = "<html><b>You have selected " + this_sty.getSty() +
                " for removal, the following semantic types</b><br> <b> typically should be removed with " +
                this_sty.getSty() +
                ".  Choose which ones should also be removed.</b></ul>";
	    } else {
	      message = "<html><b>You have chosen to keep semantic type " +
              this_sty.getSty() +
              ".  The following semantic types </b><br> <b> typically should be kept with " +
              this_sty.getSty() +
              ".  Choose which ones should also be kept.</b></ul>";
	    }

            //
            // Look up semantic type names
            //
            final String[] start_sty_names = new String[confirm_stys.length];
            for (int k1 = 0; k1 < start_sty_names.length; k1++)
              start_sty_names[k1] = confirm_stys[k1].getSty();
            Arrays.sort(start_sty_names);

            //
            // Get user selections
            //
            end_sty_names = ListDialog.showListMultipleMode(
                MMSToolkit.getGui(), (Object)message,
                "Confirm Selection", start_sty_names, null);
          }

          //
          // Cancel was pressed in dialog or none were chosen
          //
          if (related_stys == null) continue;

          //
          // Take selected sty names and look back up the sty objects
          //
          if (end_sty_names != null) {
            related_stys = new HierarchicalSemanticType[end_sty_names.length];
            int ct = 0;
            for (int k1 = 0; k1 < confirm_stys.length; k1++) {
              String name = confirm_stys[k1].getSty();
              for (int k2 = 0; k2 < end_sty_names.length; k2++)
                if (name.equals( (String) end_sty_names[k2]))
                  related_stys[ct++] = confirm_stys[k1];
            }
          }
        }

        //
	// remove or maintain from the table based on dialog box results
        //
	for (int j = 0; j < related_stys.length; j++) {

	  final int k = table.getRowForSemanticType((HierarchicalSemanticType)related_stys[j]);

          //
	  // If the table row matches, keep the row
          //
	  if(table.getSemanticTypeAtRow(k).equals(related_stys[j])) {

            //
	    // Unselect if selected
            //
	    if (table.isRowSelected(k) && !selected) {

              //
	      // prevent dialog box from being raised for each
	      // of the related sources
              //
	      selection_before[table.mapIndex(k)] = false;
	      removeSemanticTypeToRemove(k);

	    }

            //
            // Select if selected
            //
            else if(!table.isRowSelected(k) && selected) {
	      selection_before[table.mapIndex(k)] = true;
	      addSemanticTypeToRemove(k);
	    }
	  }
	}
      } // end (if selection changed or only a single row is selected)
    } // end from (first index, last index, etc.)
  }

  /**
   * Indicates whether or not events are being ignored.
   * @return <code>true</code> if so; <code>false</code> otherwise
   */
  public boolean ignoreEvents() {
    return ignore_events;
  }

  /**
   * Sets the flag indicating whether or not to ignore events.
   * @param b <code>boolean</code> <code>true</code> if events are to be ignored;
   *				 <code>false</code> if events are to be processed
   */
  public void setIgnoreEvents(boolean b) {
    ignore_events = b;
  }

  /**
   * Adds the {@link HierarchicalSemanticType} on that row to the list of
   * semantic types to be excluded.
   * @param row <code>int</code>
   */
  private void addSemanticTypeToRemove(int row) {
    setIgnoreEvents(true);
    table.addRowSelectionInterval(row, row);
    filter.addSemanticTypeToRemove(table.getSemanticTypeAtRow(row));
    setIgnoreEvents(false);
  }

  /**
   * Removes the {@link HierarchicalSemanticType} on that row from the
   * list of semantic types to be excluded.
   * @param row <code>int</code>
   */
  private void removeSemanticTypeToRemove(int row) {
    setIgnoreEvents(true);
    table.removeRowSelectionInterval(row, row);
    filter.removeSemanticTypeToRemove(table.getSemanticTypeAtRow(row));
    setIgnoreEvents(false);
  }
}

