/**************************************************************************
 *
 * Package:	gov.nih.nlm.mms.filters
 * Object:	SourcesToRemoveJTable.java
 *
 **************************************************************************/

package gov.nih.nlm.mms.filters;

import gov.nih.nlm.mms.ApplicationConfiguration;
import gov.nih.nlm.swing.SortableJTable;
import gov.nih.nlm.umls.meta.Source;

import java.util.List;

import javax.swing.ListSelectionModel;
import javax.swing.event.ListSelectionEvent;

/**
 * Provides a table for the Sources To Exclude panel that displays all the
 * sources and highlights the sources that will be excluded.
 * 
 * @author Deborah Shapiro, Brian Carlsen
 */

public class SourcesToRemoveJTable extends SortableJTable {

  //
  // Private Fields
  //
  private SourcesToRemoveTableModel source_model;

  private SourcesToRemoveListSelectionListener source_lsl = null;

  /**
   * This constructor takes a default {@link ApplicationConfiguration} and a
   * {@link SourcesToRemoveFilter}.
   * 
   * @param default_config
   *          {@link ApplicationConfiguration}
   * @param filter
   *          {@link SourcesToRemoveFilter}
   */
  public SourcesToRemoveJTable(ApplicationConfiguration default_config,
      SourcesToRemoveFilter filter) {
    super();
    source_model = new SourcesToRemoveTableModel(default_config);
    super.setModel(source_model);
    getColumnModel().getColumn(0).setPreferredWidth(10);
    getColumnModel().getColumn(1).setPreferredWidth(310);
    getColumnModel().getColumn(2).setPreferredWidth(110);
    getColumnModel().getColumn(3).setPreferredWidth(80);
    getColumnModel().getColumn(4).setPreferredWidth(40);
    getColumnModel().getColumn(5).setPreferredWidth(40);
    getColumnModel().getColumn(6).setPreferredWidth(60);
    setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
    ListSelectionModel rowSM = this.getSelectionModel();
    source_lsl = new SourcesToRemoveListSelectionListener(filter, this);
    rowSM.addListSelectionListener(source_lsl);
  }

  //
  // Methods
  //
  /**
   * This method returns the {@link Source} at the given row number.
   * 
   * @param row
   *          <code>int</code>
   * @return {@link Source}
   */
  public Source getSourceAtRow(int row) {
    return source_model.getSourceAtRow(mapIndex(row));
  }

  /**
   * This method returns the row number in the Sources To Exclude table where a
   * given {@link Source} is located.
   * 
   * @param source
   *          {@link Source}
   * @return <code>int</code> row
   */
  public int getRowForSource(Source source) {
    return (reverseMapIndex(source_model.getRowForSource(source)));
  }

  /**
   * This method turns on/off the list selection listener for this class.
   * Ensures that
   * {@link SourcesToRemoveListSelectionListener#valueChanged(ListSelectionEvent)}
   * code doesn't interfere with tableChanged() code by turning off the list
   * selection listener while tableChanged() is executed.
   * 
   * @param b
   *          <code>boolean</code> <code>true</code> turns listener off
   *          <code>false</code> turns listener on
   */
  public void setIgnoreValueChangedEvents(boolean b) {
    if (source_lsl != null) {
      source_lsl.setIgnoreEvents(b);
    }
  }

  /**
   * Sets the selections to the given list of {@link Source}s.
   * 
   * @param sources
   *          {@link List}
   */
  public void setSelections(List sources) {
    source_lsl.setSelections(sources);
  }

}
