/**************************************************************************
 *
 * Package:	gov.nih.nlm.mms.filters
 * Object:	SourcesToRemoveTableModel.java
 *
 **************************************************************************/

package gov.nih.nlm.mms.filters;

import gov.nih.nlm.mms.ApplicationConfiguration;
import gov.nih.nlm.umls.meta.Language;
import gov.nih.nlm.umls.meta.Source;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import javax.swing.table.AbstractTableModel;

/**
 * Stores and initializes the data for the Sources To Exclude table.
 * 
 * @author Deborah Shapiro, Brian Carlsen
 */

public class SourcesToRemoveTableModel extends AbstractTableModel {

  //
  // Private Fields
  //
  final private String[] column_names = { "*", "Full Source Name",
      "Source Abbreviation", "Source Family", "Language", "Level", "Concepts" };

  final private Class[] column_classes = { String.class, String.class,
      String.class, String.class, String.class, Integer.class, Integer.class };

  private ApplicationConfiguration default_config;

  private List sources = new ArrayList();

  //
  // Constructor
  //
  /**
   * This constructor gets the initial values for the sources table.
   * 
   * @param default_config
   *          {@link ApplicationConfiguration}
   */
  public SourcesToRemoveTableModel(ApplicationConfiguration default_config) {
    Source[] source_array = Source.getSources();
    Arrays.sort(source_array);
    for (int i = 0; i < source_array.length; i++)
      sources.add(source_array[i]);
    sources.remove(Source.getSource("SRC"));
    this.default_config = default_config;
  }

  //
  // Methods
  //
  /**
   * Returns the number of columns in the sources to exclude table.
   * 
   * @return <code>int</code> number of columns
   */
  public int getColumnCount() {
    return column_names.length;
  }

  /**
   * Returns the number of rows in the sources to exclude table.
   * 
   * @return <code>int</code> number of rows
   */
  public int getRowCount() {
    return sources.size();
  }

  /**
   * Returns the name of a column in the sources to exclude table
   * 
   * @param column_number
   *          <code>int</code> number of the column in question
   * @return {@link String} name of the column in position column_number
   */
  public String getColumnName(int column_number) {
    return column_names[column_number];
  }

  /**
   * Returns the value at the cell of the sources to exclude table.
   * 
   * @param row
   *          <code>int</code> row number of the cell
   * @param column
   *          <code>int</code> column number of the cell
   * @return the value at the cell at row and column
   */
  public Object getValueAt(int row, int column) {
    Source s = (Source) sources.get(row);
    if (column == 0)
      return (default_config.getReleaseVersion().equals(s.getImeta()) ? "*"
          : "");
    else if (column == 1)
      return s.getOfficialName();
    else if (column == 2)
      return s.getVersionedSourceAbbreviation();
    else if (column == 3)
      return s.getSourceFamily();
    else if (column == 4) {
      Language lat = s.getLanguage();
      if (lat != null)
        return s.getLanguage().getAbbreviation();
      else
        return "";
    } else if (column == 5)
      return new Integer(s.getRestrictionLevel());
    else if (column == 6)
      return new Integer(s.getCfr());
    else
      return null;
  }

  /**
   * Returns the {@link Class} of the objects in the given column.
   * 
   * @param column_number
   *          <code>int</code> indicates the column in question
   * @return the {@link Class} of the column
   */
  public Class getColumnClass(int column_number) {
    return column_classes[column_number];
  }

  /**
   * Returns the {@link Source} at a given row in the Sources To Remove Table.
   * 
   * @param row
   *          <code>int</code> row number
   * @return {@link Source}
   */
  public Source getSourceAtRow(int row) {
    return (Source) sources.get(row);
  }

  /**
   * Returns the row number for a given {@link Source} in the Sources To Remove
   * Table.
   * 
   * @param source
   *          {@link Source}
   * @return <code>int</code> row number
   */
  public int getRowForSource(Source source) {
    for (int i = 0; i < sources.size(); i++) {
      if (source.equals(sources.get(i)))
        return i;
    }
    return -1;
  }
}
