/**************************************************************************
 *
 * Package:    gov.nih.nlm.mms.filters;
 * Interface:  SuppressibleTermgroupsFilter.java
 *
 *************************************************************************/
package gov.nih.nlm.mms.filters;

import gov.nih.nlm.mms.ApplicationConfiguration;
import gov.nih.nlm.mms.MMSToolkit;
import gov.nih.nlm.mms.OriginalMRMetamorphoSysOutputStream;
import gov.nih.nlm.mms.SourcesToRemoveEvent;
import gov.nih.nlm.mms.SourcesToRemoveListener;
import gov.nih.nlm.swing.NoEditTableModel;
import gov.nih.nlm.swing.SortableJTable;
import gov.nih.nlm.swing.SwingToolkit;
import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.Attribute;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.Definition;
import gov.nih.nlm.umls.meta.Relationship;
import gov.nih.nlm.umls.meta.SourceTermType;
import gov.nih.nlm.umls.meta.Suppressible;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.StringTokenizer;

import javax.swing.JPanel;

/**
 * Computes which content is suppressible and (optionally) removes it.
 *
 * @author Deborah Shapiro
 **/

public class SuppressibleTermgroupsFilter extends AbstractFilter
  implements SourcesToRemoveListener, ConceptContentSelector, ConceptManipulator {

  //
  // Indicates whether or not to remove suppressible data
  //
  private boolean remove_suppressible_data = false;
  private boolean remove_source_tty_suppressible_data = false;
  private boolean remove_editor_suppressible_data = false;
  private boolean remove_obsolete_data = false;

  //
  // List of all termgroups
  //
  private List complete_termgroups = null;

  //
  // List of default suppressible termgroups
  //
  private List default_suppressible_termgroups = null;

  //
  // Subset of default suppressible termgroups
  // that are still selected (used to determine whether or
  // not user must confirm deselecting a certain termgroup)
  //
  private List confirm_default_suppressible_termgroups = null;

  //
  // List of sources to remove
  //
  private List sources_to_remove = new ArrayList();

  //
  // GUI Elements
  //
  private SortableJTable termgroup_table;
  private SuppressibleTermgroupsListSelectionListener listener = null;

  //
  // Constructor
  //

  /**
   * Instantiates an empty {@link SuppressibleTermgroupsFilter}.
   */
  public SuppressibleTermgroupsFilter() { }

  //
  // Methods
  //

  /**
   * Returns the GUI panel that will display the filter's
   * screen as one of the tabs on the main GUI window.
   * @return {@link JPanel} that displays graphical interface for the filter
   */
  public JPanel getPanel(){
    termgroup_table = new SortableJTable(getModel());
    termgroup_table.getTableHeader().getColumnModel().getColumn(0).setPreferredWidth(300);
    termgroup_table.getTableHeader().getColumnModel().getColumn(1).setPreferredWidth(150);
    termgroup_table.getTableHeader().getColumnModel().getColumn(2).setPreferredWidth(75);
    listener = new SuppressibleTermgroupsListSelectionListener(this, termgroup_table);
    termgroup_table.getSelectionModel().addListSelectionListener(listener);
    final JPanel panel = MMSToolkit.makeCard(termgroup_table,
      " Select One or More Suppressible Term Types ");
    return panel;
  }

  /**
   * Returns the table model.
   * @return {@link NoEditTableModel}
   */
  private NoEditTableModel getModel() {
    complete_termgroups = default_configuration.getSourceTermTypes();
    final List model = new ArrayList(complete_termgroups.size());
    for (int i = 0; i < complete_termgroups.size(); i++) {
      final SourceTermType tg = (SourceTermType)complete_termgroups.get(i);
      if (!sources_to_remove.contains(tg.getSource()))
    	model.add(tg);
    }
    final String[][] tg_data = new String[model.size()][3];
    for (int i = 0; i < model.size(); i++) {
      final SourceTermType stt = (SourceTermType)model.get(i);
      tg_data[i][0] = stt.getSource().getOfficialName();
      tg_data[i][1] = stt.getSource().getVersionedSourceAbbreviation();
      tg_data[i][2] = stt.getTermType().getAbbreviation();
    }
    final Object [] headers = {"Source", "Source Abbreviation", "Term Type"};
    return new NoEditTableModel(tg_data, headers);
  }

  /**
   * Selects rows on the suppressiblity table for rows of termgroups that should be suppressed.
   */
  public void setPanelSelections() {
    listener.setIgnoreEvents(true);

    termgroup_table.clearSelection();
    final List stgs = new ArrayList(getSuppressibleTermgroups());
    for (int i = 0; i < stgs.size(); i++) {
      final SourceTermType tg = (SourceTermType)stgs.get(i);
      //
      // check if row should be visible in table
      //
      int row = getRowForTermgroup(tg);
      if (row != -1)
        listener.addSuppressibleTermgroup(row);
    }

    listener.setIgnoreEvents(false);
  }

  /**
   * Returns the description.
   * @return the description
   */
  public String getTabDescription() {
    return "This screen contains the Suppressibility Filter, which specifies " +
        "source/term type combinations to be suppressed. Users can customize " +
        "the subset by selecting and deselecting source/term type combinations.  " +
        "See Help for more information.\n\n" +
        "Users may proceed to the other Options tabs in any order by clicking " +
        "on the tabs across the top of the screen or by clicking \"Done\" on " +
        "the File bar and selecting \"Begin Subset\".";
  }

  /**
   * Resets state to defaults.
   */
  public void resetToDefaults() {
    setSuppressibleTermgroups(new ArrayList(default_suppressible_termgroups));
    confirm_default_suppressible_termgroups = new ArrayList(default_suppressible_termgroups);
    setRemoveSuppressibleData(false);
    setRemoveSourceTtySuppressibleData(false);
    setRemoveEditorSuppressibleData(false);
    setRemoveObsoleteData(false);
  }

  /**
   * Shows an advanced options GUI to the user.
   */
  public void showOptions() {
    final SuppressibleTermgroupsOptionsGUI sup_tgps_options_gui =
      new SuppressibleTermgroupsOptionsGUI(this);
    sup_tgps_options_gui.setVisible(true);
  }

  /**
   * Indicates whether or not an options menu item should be supplied
   * for advanced configuration options.
   * @return <code>true</code> if so; <code>false</code> otherwise
   */
  public boolean hasAdvancedOptions() { return true; }

  /**
   * Returns a help {@link JPanel}.
   * @return a help {@link JPanel}
   */
  public JPanel getHelpPanel() {
    return SwingToolkit.getPanelForHTML(
      this.getClass().getClassLoader().getResource(
      "config/SuppressibleTermgroupsHelp.html"));
  }

  /**
   * Sets the {@link ApplicationConfiguration}. Used for "default" selections.
   * @param c the {@link ApplicationConfiguration}
   */
  public void setConfiguration(ApplicationConfiguration c) {
    default_configuration = c;

    //
    // What are the default suppressed termgroups.
    //
    if(default_configuration.getProperty(this, "suppressed_termgroups") != null) {
      default_suppressible_termgroups = new ArrayList();
      confirm_default_suppressible_termgroups = new ArrayList();
      final String suppress_prop =
          default_configuration.getProperty(this, "suppressed_termgroups");
      final StringTokenizer tokens = new StringTokenizer(suppress_prop, ";");
      while (tokens.hasMoreTokens()) {
        final String line = tokens.nextToken();
        final SourceTermType ts = (default_configuration.getSourceTermType(
            line.substring(0, line.indexOf('|')),
            line.substring(line.indexOf('|')+1)));
        default_suppressible_termgroups.add(ts);
        confirm_default_suppressible_termgroups.add(ts);
      }
    } else {
      MMSToolkit.reportError("There are no default suppressed term types", true);
    }
  }

  /**
   * Returns filter state.
   * @return filter state as {@link Properties}
   */
  public Properties getProperties() {
    final Properties new_filter_props = new Properties();

    new_filter_props.setProperty("remove_source_tty_suppressible_data",
    														 new Boolean(removeSourceTtySuppressibleData()).toString());
    new_filter_props.setProperty("remove_editor_suppressible_data", 
    														 new Boolean(removeEditorSuppressibleData()).toString());
    new_filter_props.setProperty("remove_obsolete_data",
    														 new Boolean(removeObsoleteData()).toString());
    final List stgs = getSuppressibleTermgroups();
    final StringBuffer line = new StringBuffer();
    for(int i=0; i< stgs.size(); i++) {
      final SourceTermType stg = (SourceTermType)stgs.get(i);
      line.append(stg.getSource().getRsab().toString());
      line.append("|");
      line.append(stg.getTermType().getAbbreviation());
      if(i <= stgs.size() - 2)
	line.append(";");
    }
    new_filter_props.setProperty("suppressed_termgroups", line.toString());
    return new_filter_props;
  }

  /**
   * Sets filter state.
   * @param p {@link Properties} representing filter state
   */
  public void setProperties(Properties p) {

    //
    // Should suppressible data be removed?
    //

    // if old style config file, set all three remove types if remove_suppressible_data
    // is set
    if (p.containsKey("remove_suppressible_data") &&
    		(new Boolean(p.getProperty("remove_suppressible_data")).booleanValue()) &&
    		!p.containsKey("remove_source_tty_suppressible_data") &&
    		!p.containsKey("remove_editor_suppressible_data") &&
    		!p.containsKey("remove_obsolete_data")) {
    	setRemoveSourceTtySuppressibleData(true);
    	setRemoveEditorSuppressibleData(true);
    	setRemoveObsoleteData(true);
    }
    if (p.containsKey("remove_source_tty_suppressible_data")) {
      setRemoveSourceTtySuppressibleData(
          new Boolean(
            p.getProperty("remove_source_tty_suppressible_data")).booleanValue());
    }
    if (p.containsKey("remove_editor_suppressible_data")) {
      setRemoveEditorSuppressibleData(
          new Boolean(
            p.getProperty("remove_editor_suppressible_data")).booleanValue());
    } 
    if (p.containsKey("remove_obsolete_data")) {
      setRemoveObsoleteData(
          new Boolean(
            p.getProperty("remove_obsolete_data")).booleanValue());
    }    
    //
    // What are the suppressed termgroups
    //
    if (p.containsKey("suppressed_termgroups")) {
    	List supp_termgroups = new ArrayList();
      final String suppress_prop = p.getProperty("suppressed_termgroups");
      final StringTokenizer tokens = new StringTokenizer(suppress_prop, ";");
      while(tokens.hasMoreTokens()) {
        final String line = tokens.nextToken();
        final SourceTermType tg = (default_configuration.getSourceTermType(
            line.substring(0, line.indexOf('|')),
            line.substring(line.indexOf('|')+1)));
        if (tg != null) {
          tg.setSuppress(Suppressible.getSuppressible("Y"));
          supp_termgroups.add(tg);
        }
      }
      setSuppressibleTermgroups(supp_termgroups);
    }
  }

  /**
   * Manipulates data in the {@link Concept}, but does not remove any data
   * from the concept.
   * @param concept {@link Concept}
   */
  public void applyChanges(Concept concept) {
    final Atom[] atoms = concept.getAtoms();
    for(int i = atoms.length-1; i >= 0; i--) {
      final Atom atom = (Atom)atoms[i];
      atom.setSuppress(getSuppressible(atom));
    }
  }

  /**
   * Sets the suppressibility of atoms and removes suppressible content
   * if advanced options indicate to do so.
   * @param concept the current {@link Concept}
   */
  public void applySelections(Concept concept) {

    final Atom[] atoms = concept.getAtoms();
    for(int i = atoms.length-1; i >= 0; i--) {
      final Atom atom = (Atom)atoms[i];

      if (removeSourceTtySuppressibleData() && getSuppressible(atom).getAbbreviation().equals("Y")) {
      	concept.removeAtom(atom);
      } else if (removeEditorSuppressibleData() && getSuppressible(atom).getAbbreviation().equals("E")) {
      	concept.removeAtom(atom);
      } else if (removeObsoleteData() && getSuppressible(atom).getAbbreviation().equals("O")) {
      	concept.removeAtom(atom);
      }
    }

    if (removeSourceTtySuppressibleData()) {
      final Definition[] defs = concept.getDefinitions();
      for (int i = defs.length - 1; i >= 0; i--) {
        final Definition def = (Definition) defs[i];
        if (def.isSuppressible()) {
          concept.removeDefinition(def);
        }
      }
      final Relationship[] rels = concept.getRelationships();
      for (int i = rels.length - 1; i >= 0; i--) {
        final Relationship rel = (Relationship) rels[i];
        if (rel.isSuppressible()) {
          concept.removeRelationship(rel);
        }
      }
      final Attribute[] atts = concept.getAttributes();
      for (int i = atts.length - 1; i >= 0; i--) {
        final Attribute att = (Attribute) atts[i];
        if (att.isSuppressible()) {
          concept.removeAttribute(att);
        }
      }
    }
  }

  /**
   * Computes and returns the suppressibility of the specified atom.
   * @param a the {@link Atom}
   * @return the suppressibility of the specified atom
   */
  private Suppressible getSuppressible(final Atom a) {
    if (a.getSuppress() != null && a.getSuppress().getAbbreviation().equals("O"))
      return a.getSuppress();

    final String ssab = a.getSource().getRootSourceAbbreviation();    
    final SourceTermType tg =
       default_configuration.getSourceTermType(ssab, a.getTty().getAbbreviation());
    if (tg != null && getSuppressibleTermgroups().contains(tg))
      return Suppressible.getSuppressible("Y");
    
    if (default_configuration.isSuppressedTerm(a))
      return Suppressible.getSuppressible("E");
    
    return Suppressible.getSuppressible("N");
  }

  /**
   * Returns the filter name.
   * @return the filter name
   */
  public String getName() {
    return "Suppressibility";
  }

  /**
   * Returns configuration details for use in <code>mmsys.log</code>.
   * @return configuration details for use in <code>mmsys.log</code>
   */
  public String getConfigurationDetails() {
    final StringBuffer sb = new StringBuffer();
    final String ls = System.getProperty("line.separator");
    sb.append(
      "Source/Term Type whose terms were marked suppressible").append(ls);
    // attention
    final Iterator iter = getSuppressibleTermgroups().iterator();
    boolean empty_list = true;
    while(iter.hasNext()) {
      final SourceTermType tg = (SourceTermType)iter.next();
      // don't print out SourceTermTypes if their sources are to be excluded from subset
      if(sources_to_remove.contains(tg.getSource()))
        continue;
      final String sn = tg.getSource().getOfficialName();
      sb.append("  " +
               ((sn.length() > 67) ? sn.substring(0,64) + "..." : sn) +
               "                                                                  ".substring(0,68-((sn.length()>67)?67:sn.length())) +
               tg.getSource().getVersionedSourceAbbreviation() +
	       "|" +
	       tg.getTermType().getAbbreviation()).append(ls);
      empty_list = false;
    }

    if(empty_list)
      sb.append("  <none>").append(ls);
    sb.append(ls);

    if(user_configuration.getOutputStream()
    		instanceof OriginalMRMetamorphoSysOutputStream) {
    	if (removeSuppressibleData())     
        sb.append("Suppressible terms were excluded from the subset.\n");
      else
        sb.append("Suppressible terms were included in the subset.\n");
    } else {
    	if (removeSourceTtySuppressibleData())
    		sb.append("Source Term Type suppressible terms were excluded from the subset.\n");
    	else
    		sb.append("Source Term Type suppressible terms were included in the subset.\n");    		

    	if (removeObsoleteData())
    		sb.append("Obsolete terms were excluded from the subset.\n");
    	else
    		sb.append("Obsolete terms were included in the subset.\n");  
    	if (removeEditorSuppressibleData())
    		sb.append("Editor assigned suppressible terms were excluded from the subset.\n");
    	else
    		sb.append("Editor assigned suppressible terms were included in the subset.\n");  
    }
    return sb.toString();
  }

  /**
   * Clears the list of suppressible {@link SourceTermType}s.
   */
  protected void clearSuppressibleTermgroups() {
    user_configuration.setUserSuppressedSourceTermTypes(new ArrayList());
  }

  /**
   * Adds a {@link SourceTermType} to the list of suppressible termgroups.
   * @param tg the {@link SourceTermType}
   */
  protected void addSuppressibleTermgroup(SourceTermType tg) {
    if(!getSuppressibleTermgroups().contains(tg)) {
      tg.setSuppress(Suppressible.getSuppressible("Y"));
      List l_suppressed = user_configuration.getUserSuppressedSourceTermTypes();
      l_suppressed.add(tg);
      user_configuration.setUserSuppressedSourceTermTypes(l_suppressed);
    }
  }

  /**
   * Removes a {@link SourceTermType} from the list of suppressible termgroups.
   * @param tg {@link SourceTermType}
   */
  protected void removeSuppressibleTermgroup(SourceTermType tg) {
  	List l_suppressed = user_configuration.getUserSuppressedSourceTermTypes();
    l_suppressed.remove(tg);
    tg.setSuppress(Suppressible.getSuppressible("N"));
    user_configuration.setUserSuppressedSourceTermTypes(l_suppressed);
  }

  /**
   * Returns an {@link List} containing all
   * {@link SourceTermType}s to be suppressed from the default termgroup list.
   * @return {@link List} of suppressed termgroups
   */
  public List getDefaultSuppressibleTermgroups() {
    return default_suppressible_termgroups;
  }

  /**
   * Returns a list of {@link SourceTermType}s that have been
   * confirmed by the user via a dialog box to be suppressible.
   * Prevents user from being queried twice about a termgroup's
   * suppressiblity.
   * @return {@link List} of suppressed termgroups
   */
  public List getConfirmDefaultSuppressibleTermgroups() {
    return confirm_default_suppressible_termgroups;
  }

  /**
   * Clears the list of {@link SourceTermType}s that have been confirmed
   * by the user to be suppressible.
   */
  public void clearConfirmDefaultSuppressibleTermgroups() {
    confirm_default_suppressible_termgroups.clear();
  }

  /**
   * Sets the suppressible termgroups list with the one provided.
   * @param stgs the {@link List} of {@link SourceTermType}s
   */
  public void setSuppressibleTermgroups(List stgs) {
    SourceTermType[] ttys = (SourceTermType[])default_configuration.getSourceTermTypes().toArray(new SourceTermType[0]);
    for (int i=0; i<ttys.length; i++) {
    	if (stgs.contains(ttys[i]))
    		addSuppressibleTermgroup(ttys[i]);
    	else
    		removeSuppressibleTermgroup(ttys[i]);
    }
  }

  /**
   * Returns the {@link List} containing all
   * suppressible {@link SourceTermType}s.
   * @return {@link List} of suppressible termgroups
   */
  public List getSuppressibleTermgroups() {
    return user_configuration.getUserSuppressedSourceTermTypes();
  }

  /**
   * Indicates whether the Remove Suppressible Data
   * option has been chosen on the Advanced Suppressibility Options panel.
   * @return <code>true</code> if so, <code>false</code> otherwise
   */
  public  boolean removeSuppressibleData() {
    return remove_suppressible_data;
  }

  /**
   * Sets the flag indicating whether or not suppressible data will be removed
   * from the subset.
   * @param b the flag indicating whether or not suppressible data will be removed
   * from the subset
   */
  public void setRemoveSuppressibleData(boolean b) {
    fireNotUndoableDataChanged();
    remove_suppressible_data = b;
  }
  /**
   * Indicates whether the Remove Source Term Type Suppressible Data
   * option has been chosen on the Advanced Suppressibility Options panel.
   * @return <code>true</code> if so, <code>false</code> otherwise
   */
  public  boolean removeSourceTtySuppressibleData() {
    return remove_source_tty_suppressible_data;
  }

  /**
   * Sets the flag indicating whether or not source term type suppressible data will be removed
   * from the subset.
   * @param b the flag indicating whether or not source term type suppressible data will be removed
   * from the subset
   */
  public void setRemoveSourceTtySuppressibleData(boolean b) {
    fireNotUndoableDataChanged();
    remove_source_tty_suppressible_data = b;
  }
  /**
   * Indicates whether the Remove Editor Assigned Suppressible Data
   * option has been chosen on the Advanced Suppressibility Options panel.
   * @return <code>true</code> if so, <code>false</code> otherwise
   */
  public  boolean removeEditorSuppressibleData() {
    return remove_editor_suppressible_data;
  }

  /**
   * Sets the flag indicating whether or not editor assigned suppressible data will be removed
   * from the subset.
   * @param b the flag indicating whether or not editor assigned suppressible data will be removed
   * from the subset
   */
  public void setRemoveEditorSuppressibleData(boolean b) {
    fireNotUndoableDataChanged();
    remove_editor_suppressible_data = b;
  }
  
  /**
   * Returns the table that includes {@link SourceTermType}s
   * with the suppressible termgroups highlighted.
   * @return {@link SortableJTable}
   */
  public SortableJTable getSuppressibleTermgroupsTable() {
    return termgroup_table;
  }
  /**
   * Indicates whether the Remove Obsolete Data
   * option has been chosen on the Advanced Suppressibility Options panel.
   * @return <code>true</code> if so, <code>false</code> otherwise
   */
  public  boolean removeObsoleteData() {
    return remove_obsolete_data;
  }

  /**
   * Sets the flag indicating whether or not obsolete data will be removed
   * from the subset.
   * @param b the flag indicating whether or not obsolete data will be removed
   * from the subset
   */
  public void setRemoveObsoleteData(boolean b) {
    fireNotUndoableDataChanged();
    remove_obsolete_data = b;
  }
  //
  // SourcesToRemoveListener Implementation
  //

  /**
   * Resets the suppressiblity table when the user tabs between filters.
   * Removes entries visible in the table based on which sources
   * are to be excluded.
   * @param e {@link SourcesToRemoveEvent}
   */
  public void sourcesToRemoveChanged(SourcesToRemoveEvent e) {
    sources_to_remove = e.getSourcesToRemove();
    termgroup_table.setModel(getModel());
    termgroup_table.getTableHeader().getColumnModel().getColumn(0).setPreferredWidth(300);
    termgroup_table.getTableHeader().getColumnModel().getColumn(1).setPreferredWidth(150);
    termgroup_table.getTableHeader().getColumnModel().getColumn(2).setPreferredWidth(75);
    setPanelSelections();
  }


  /**
   * Maps a {@link SourceTermType} to a row number in the table.
   *
   * This is necessary because only {@link SourceTermType}s
   * from sources that are not to be excluded will be visible
   * in the table.
   * @param tg the {@link SourceTermType}
   * @return <code>int</code> row number of termgroup in table
   * 			-1 if termgroup is to be excluded
   */
  public int getRowForTermgroup(SourceTermType tg) {
    if(sources_to_remove.contains(tg.getSource()))
      return -1;
    int row =0;
    for(int i=0; i<complete_termgroups.size(); i++) {
      final SourceTermType termgroup = (SourceTermType)complete_termgroups.get(i);
      if(!sources_to_remove.contains(termgroup.getSource())) {
        row++;
        if(tg.equals(termgroup))
          return termgroup_table.reverseMapIndex(row-1);
      }
    }
    return -1;
  }

  /**
   * Maps a row number to the {@link SourceTermType}.
   * Adjusts for the {@link SourceTermType}s
   * that are not currently visible in the suppressibility table
   * because their sources are to be excluded.
   * @param tg <code>int</code> row in the suppressibility table
   * @return {@link SourceTermType} located at that row in the table
   */
  public SourceTermType getTermgroupAtRow(int tg) {
    int row = termgroup_table.mapIndex(tg);
    int subset_row =-1;
    SourceTermType termgroup = null;
    for(int i=0; i<complete_termgroups.size(); i++) {
      termgroup = (SourceTermType)complete_termgroups.get(i);
      if(!sources_to_remove.contains(termgroup.getSource())) {
	subset_row++;
      }
      if(subset_row == row)
	return termgroup;
    }
    return termgroup;
  }

  /**
   * Returns the {@link Color} that should be used to display
   * the name.
   * @return the {@link Color}
   */
  public Color getColor() {
    return null;
  }

  /**
   * Indicates if {@link Filter} can use the efficient mode during
   * the "Initialize Cuis" phase of the subsetting process.
   * The efficient mode reads only MRCONSO and MRSTY during the
   * "Initialize Cuis" phase.
   * @return <code>true</code>
   */
  public boolean isEfficientEnabled() {
    return true;
  }

}

