/**************************************************************************
 *
 * Package:    gov.nih.nlm.swing;
 * Class:      FindDialog.java
 *
 *************************************************************************/

package gov.nih.nlm.swing;

import gov.nih.nlm.util.FinderHandler;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.BorderFactory;
import javax.swing.ButtonGroup;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JTextField;
import javax.swing.KeyStroke;
import javax.swing.border.TitledBorder;

/**
 * Dialog to manage find on a <code>Document</code> shown in a
 * <code>JEditorPane</code>.
 * 
 * 
 * @author Ulrich Hilger
 * @author CalCom
 * @author <a href="http://www.calcom.de">http://www.calcom.de</a>
 * @author <a href="mailto:info@calcom.de">info@calcom.de</a>
 * @author modifed by Deborah Shapiro {dshapiro@apelon.com}
 * 
 * @version 1.3, April 13, 2002
 * 
 * @see javax.swing.text.Document
 * @see javax.swing.JEditorPane
 */

public class FindDialog extends JDialog {

  /** indicates the current operation */
  private int operation;

  private FinderHandler handler = null;

  /* Constants for method toggleState */
  public static final boolean STATE_LOCKED = false;

  public static final boolean STATE_UNLOCKED = true;

  /* Constants for operation */
  public static final int OP_NONE = 0;

  public static final int OP_FIND = 1;

  /* ---- GUI elements start --------- */

  private TitledBorder titledBorder1;

  private JButton jbtnFindNext = new JButton();

  private JCheckBox jcbStartOnTop = new JCheckBox();

  private JRadioButton jrbDown = new JRadioButton();

  private JPanel jpnlBtn = new JPanel();

  private JPanel jpnlOptions = new JPanel();
  
  private JPanel jpnlFind = new JPanel();

  private JPanel jpnlMain = new JPanel();

  private JRadioButton jrbUp = new JRadioButton();

  private JTextField jtfPhrase = new JTextField();

  private JCheckBox jcbMatchCase = new JCheckBox();

  private JLabel jLabel3 = new JLabel();

  private JLabel jLabel4 = new JLabel();

  private GridBagLayout gridBagLayout4 = new GridBagLayout();

  private GridBagLayout gridBagLayout5 = new GridBagLayout();

  private JButton jbtnClose = new JButton();

  private GridBagLayout gridBagLayout6 = new GridBagLayout();

  private GridLayout gridLayout2 = new GridLayout();
  
  private boolean ignore_events = false;
  
  //private boolean ignore_events = false;

  // Create action objects
  Action cancel_action = new AbstractAction("Cancel") {
    {
      // Set action command to "Cancel"
      putValue(Action.ACTION_COMMAND_KEY, getValue(Action.NAME));
    }

    // This method is called when the action is invoked
    public void actionPerformed(ActionEvent e) {
      jbtnCancel_actionPerformed(e);
    }
  };
  
  Action next_action = new AbstractAction("Next") {
    {
      // Set action command to "Next"
      putValue(Action.ACTION_COMMAND_KEY, getValue(Action.NAME));
    }

    // This method is called when the action is invoked
    public void actionPerformed(ActionEvent e) {
      jbtnFindNext_actionPerformed(e);
    }
  };
  
  private JButton jbtnCancel = new JButton(cancel_action);

  private JCheckBox jcbUnused = new JCheckBox();

  /* ---- GUI elements end --------- */

  /*****************************************************************************
   * ---- Constructor(s) start -----------
   * 
   * /** Construct a <code>FindDialog</code>.
   * 
   * <p>Shows the dialog window modal, packed and centered over the owning
   * <code>Frame</code> after construction.</p>
   * 
   * <p>Using this constructor implies the dialog shall be used in mode
   * MODE_DOCUMENT</p>
   * 
   * @param owner the <code>Frame</code> that owns this dialog @param editors
   * <code>JEditorPane</code>s displaying the <code>Document</code> to
   * seach in
   * 
   * @see javax.swing.JEditorPane
   * @see javax.swing.text.Document
   * @see java.awt.Frame
   */
  public FindDialog(Frame owner) {
    try {
      jbInit();
      initDialogContents();
      centerDialog(owner);
      pack();
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  /* --------- Constructor(s) end ------------- */

  /* --------- Event handling start ------------- */

  /**
   * Sets the {@link FinderHandler} that will execute the find operation.
   * 
   * @param handler
   *          {@link FinderHandler}
   */
  public void setHandler(FinderHandler handler) {
    this.handler = handler;
  }
  
  /**
   * Returns the {@link FinderHandler} that will execute the find operation.
   * 
   * @returns handler
   *          {@link FinderHandler}
   */
  public FinderHandler getHandler() {
    return handler;
  }
  
  /**
   * Resets the {@link FindDialog} to provide a clean slate for
   * a new search when the close button is pressed or the
   * concept is changed.
   */
  public void reset() {

    toggleState(STATE_UNLOCKED);
    initDialogContents();
    handler.cancel();
    try {
      handler.init();
    } catch (Exception ex) {
      ex.printStackTrace();
      SwingToolkit.reportError(this,
          "Unexpected error initializing find operation.\n "
              + "See console for more details.", false);
    }
  }
  
  /**
   * perform a find, when button 'find next' is pressed
   * 
   * @param e
   *          {@link ActionEvent}
   */
  private void jbtnFindNext_actionPerformed(ActionEvent e) {
    operation = OP_FIND;
    toggleState(STATE_LOCKED);
    try {
      handler.init();
    } catch (Exception ex) {
      ex.printStackTrace();
      SwingToolkit.reportError(this,
          "Unexpected error initializing find operation.\n "
              + "See console for more details.", false);
    }
    try {
      handler.find();
    } catch (Exception ex) {
      ex.printStackTrace();
      SwingToolkit.reportError(this,
          "Unexpected error executing find operation.\n "
              + "See console for more details.", false);
    }
  }

  /**
   * Cancels the current find operation and switches the dialog back to normal.
   * 
   * @param e
   *          {@link ActionEvent}
   */
  private void jbtnCancel_actionPerformed(ActionEvent e) {
    handler.cancel();
    try {
      handler.init();
    } catch (Exception ex) {
      ex.printStackTrace();
      SwingToolkit.reportError(this,
          "Unexpected error initializing find operation.\n "
              + "See console for more details.", false);
    }
    toggleState(STATE_UNLOCKED);
    jtfPhrase.requestFocus();
    jtfPhrase.setSelectionStart(0);
    jtfPhrase.setSelectionEnd(jtfPhrase.getText().length());
  }

  /**
   * When Close is pressed, store the pressed button in the result of this
   * dialog and dispose the dialog.
   * 
   * @param e
   *          {@link ActionEvent}
   */
  private void jbtnClose_actionPerformed(ActionEvent e) {
    setVisible(false);
    reset();
  }

  /* --------- Event handling end ------------- */

  /* ----------- Helper methods start ------- */

  /**
   * Set dialog components to their inital state
   */
  private void initDialogContents() {
    jbtnCancel.setEnabled(false);
    jrbUp.setSelected(false);
    jrbDown.setSelected(true);
    jcbMatchCase.setSelected(false);
    jcbStartOnTop.setSelected(true);
    jtfPhrase.setText("");
    jtfPhrase.requestFocus();
  }

  /**
   * Center this dialog window relative to its owning <code>Frame</code>.
   * 
   * @param owner
   *          <code>Frame</code> owning this dialog
   * 
   * @see java.awt.Frame
   */
  public void centerDialog(Frame owner) {
    Dimension dlgSize = getPreferredSize();
    Dimension frmSize = owner.getSize();
    Point loc = owner.getLocation();
    setLocation((frmSize.width - dlgSize.width) / 2 + loc.x,
        (frmSize.height - dlgSize.height) / 2 + loc.y);
  }

  /**
   * Toggle the state of the dialog window.
   * 
   * <p>
   * The state of the dialog is either unlocked (no find in progress) or locked
   * (find in progress).
   * </p>
   * 
   * @param unlocked
   *          one of FindDialog.STATE_LOCKED and FindDialog.STATE_UNLOCKED
   */
  private void toggleState(boolean unlocked) {
    jbtnCancel.setEnabled(!unlocked);
    jtfPhrase.setEnabled(unlocked);
    jLabel3.setEnabled(unlocked);
    jLabel4.setEnabled(unlocked);
    jcbMatchCase.setEnabled(unlocked);
    jcbStartOnTop.setEnabled(unlocked);
    jrbUp.setEnabled(unlocked);
    jrbDown.setEnabled(unlocked);
  }

  /**
   * Terminate the current operation
   */
  public void terminateOperation() {
    switch (operation) {
    case OP_FIND:
      SwingToolkit.notifyUser(this, "No (more) occurrences found.");
      toggleState(STATE_UNLOCKED);
      break;
    }
    operation = OP_NONE;
  }

  /* ----------- Helper methods end ------- */

  /**
   * GUI builder init
   * 
   * @throws Exception
   */
  private void jbInit() throws Exception {
    titledBorder1 = new TitledBorder(BorderFactory.createEtchedBorder(
        Color.white, new Color(142, 142, 142)), "Options");
    ButtonGroup bgSearchDirection = new ButtonGroup();
    jbtnFindNext.addActionListener(new java.awt.event.ActionListener() {
      public void actionPerformed(ActionEvent e) {
        jbtnFindNext_actionPerformed(e);
      }
    });
    
    jbtnFindNext.setText("Find Next");
    jbtnFindNext.setPreferredSize(new Dimension(100, 27));
    jbtnFindNext.setMinimumSize(new Dimension(100, 27));
    jbtnFindNext.setMaximumSize(new Dimension(100, 27));
    // Register keystroke
    jbtnFindNext.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(
        KeyStroke.getKeyStroke('G' , Toolkit.getDefaultToolkit().getMenuShortcutKeyMask() ), 
        next_action.getValue(Action.NAME));
    // Register action
    jbtnFindNext.getActionMap().put(next_action.getValue(Action.NAME),
        next_action);
    //getRootPane().setDefaultButton(jbtnFindNext);
    jcbStartOnTop.setText("Search from start");
    jcbStartOnTop.setToolTipText("");
    jrbDown.setText("Search down");
    jpnlBtn.setLayout(gridBagLayout4);
    jpnlOptions.setBorder(titledBorder1);
    jpnlOptions.setLayout(gridLayout2);
    jpnlFind.setLayout(gridBagLayout5);
    jpnlMain.setLayout(gridBagLayout6);
    jrbUp.setText("Search up");
    jtfPhrase.setMinimumSize(new Dimension(4, 12));
    jtfPhrase.setPreferredSize(new Dimension(63, 12));
    jtfPhrase.setText("jtfPhrase");
    jtfPhrase.addKeyListener(new KeyAdapter() {
      public void keyTyped(KeyEvent e) {
        char c = e.getKeyChar();
        if (c == KeyEvent.VK_ENTER) {
          jbtnFindNext_actionPerformed(new ActionEvent(this, 0, ""));
        }
      }
    });
    jcbMatchCase.setText("Match case");
    jLabel4.setText("Text to find:");
    jbtnClose.setMaximumSize(new Dimension(100, 27));
    jbtnClose.setMinimumSize(new Dimension(100, 27));
    jbtnClose.setPreferredSize(new Dimension(100, 27));
    jbtnClose.setText("Close");
    jbtnClose.addActionListener(new java.awt.event.ActionListener() {
      public void actionPerformed(ActionEvent e) {
        jbtnClose_actionPerformed(e);
      }
    });
    gridLayout2.setRows(4);
    gridLayout2.setColumns(2);
    this.setModal(false);
    this.setTitle("Find");
    jbtnCancel.setMaximumSize(new Dimension(100, 27));
    jbtnCancel.setMinimumSize(new Dimension(100, 27));
    jbtnCancel.setPreferredSize(new Dimension(100, 27));
    jcbUnused.setText("jcbUnused");
    jcbUnused.setVisible(false);
    this.getContentPane().add(jpnlMain, BorderLayout.NORTH);
    // Register keystroke
    jpnlBtn.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(
        KeyStroke.getKeyStroke("ESCAPE"), cancel_action.getValue(Action.NAME));
    // Register action
    jpnlBtn.getActionMap().put(cancel_action.getValue(Action.NAME),
        cancel_action);
    jpnlBtn.add(jbtnFindNext, new GridBagConstraints(0, 0, 1, 1, 0.0, 0.0,
        GridBagConstraints.NORTH, GridBagConstraints.NONE, new Insets(4, 4, 0,
            4), 0, 0));
    jpnlBtn.add(jbtnClose, new GridBagConstraints(0, 3, 1, 1, 0.0, 0.0,
        GridBagConstraints.NORTH, GridBagConstraints.NONE, new Insets(0, 4, 4,
            4), 0, 0));
    jpnlBtn.add(jbtnCancel, new GridBagConstraints(0, 2, 1, 1, 0.0, 0.0,
        GridBagConstraints.NORTH, GridBagConstraints.NONE, new Insets(4, 4, 4,
            4), 0, 0));
    jpnlMain.add(jpnlFind, new GridBagConstraints(0, 0, 1, 1, 1.0, 1.0,
        GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH, new Insets(4, 4,
            4, 4), 0, 0));
    jpnlMain.add(jpnlBtn, new GridBagConstraints(1, 0, 1, 2, 1.0, 1.0,
        GridBagConstraints.NORTHEAST, GridBagConstraints.NONE, new Insets(4, 4,
            4, 4), 0, 0));
    jpnlFind.add(jtfPhrase, new GridBagConstraints(1, 0, 1, 1, 1.0, 0.0,
        GridBagConstraints.WEST, GridBagConstraints.HORIZONTAL, new Insets(4,
            4, 4, 4), 0, 12));
    jpnlFind.add(jLabel4, new GridBagConstraints(0, 0, 1, 1, 0.0, 0.0,
        GridBagConstraints.EAST, GridBagConstraints.NONE,
        new Insets(4, 4, 4, 4), 0, 0));
    jpnlFind.add(jLabel3, new GridBagConstraints(0, 1, 1, 1, 0.0, 0.0,
        GridBagConstraints.EAST, GridBagConstraints.NONE,
        new Insets(4, 4, 4, 4), 0, 0));
    jpnlMain.add(jpnlOptions, new GridBagConstraints(0, 1, 1, 1, 1.0, 1.0,
        GridBagConstraints.SOUTHWEST, GridBagConstraints.NONE, new Insets(4, 4,
            4, 4), 0, 0));
    // removed b/c buggy with indentation
    // jpnlOptions.add(jcbWholeWords, null);
    jpnlOptions.add(jrbUp, null);
    jpnlOptions.add(jcbMatchCase, null);
    jpnlOptions.add(jrbDown, null);
    jpnlOptions.add(jcbStartOnTop, null);
    jpnlOptions.add(jcbUnused, null);
    bgSearchDirection.add(jrbUp);
    bgSearchDirection.add(jrbDown);
    
    addWindowFocusListener(new WindowAdapter() {
      public void windowGainedFocus(WindowEvent e) {
      	if (ignore_events) {
      		ignore_events = false;
      		return;
      	}
      	ignore_events = true;
      	getHandler().requestFocus();
      	requestFocus();
      	jtfPhrase.requestFocus();
      }    	
    });
  }
  
  public boolean isSearchUp() {
    return jrbUp.isSelected();
  }

  public boolean isSearchDown() {
    return jrbDown.isSelected();
  }

  public boolean isMatchCase() {
    return jcbMatchCase.isSelected();
  }

  public boolean isStartOnTop() {
    return jcbStartOnTop.isSelected();
  }

  public String getSearchField() {
    return jtfPhrase.getText();
  }
}