/***************************************************************************
 * Package:     gov.nih.nlm.umls.io
 * Interface:   MetathesaurusInputStream.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.io;

import gov.nih.nlm.umls.meta.IndexEntry;
import gov.nih.nlm.umls.meta.ReleaseMetadata;

import java.io.IOException;
import java.util.List;

/**
 * Generically represents a way of reading the entire Metathesaurus
 * from some input source.  This source can be MR Files, update MR Files,
 * a database, or pretty much anything else that can be defined by one or
 * more "locations" or "paths".
 *
 * @author Deborah Shapiro
 */
public interface MetathesaurusInputStream extends ConceptInputStream {

  /**
   * Read an {@link IndexEntry} from the input stream and return it.
   * This method should provide the mapping between the input
   * files and {@link IndexEntry} objects.  Futhermore, it should be
   * implemented like a stream, in the sense that only the content
   * for the next {@link IndexEntry} should be read into memory.
   * @return the next {@link IndexEntry}
   * @throws IOException if there is a problem reading the index entry
   */
  public IndexEntry readIndexEntry() throws IOException;

  /**
   * Read the {@link ReleaseMetadata} from the input stream and return it.
   * @return the {@link ReleaseMetadata} for this input stream
   * @throws IOException if there is a problem reading the release metadata
   */
  public ReleaseMetadata readReleaseMetadata() throws IOException;

  /**
   * Returns the percentage of the {@link IndexEntry} objects that have been processed.
   * @return the percentage of the {@link IndexEntry} objects that have been processed
   */
  public int getIndexProgress();

  /**
   * Returns the complete list of suis from the AMBIGSUI list.
   * They will be used in calculating entries for the output AMBIGSUI list.
   * @return list of suis
   * @throws IOException
   */
  public String[] getAmbiguousStringIdentifiers() throws IOException;

  /**
   * Returns the complete list of suis from the AMBIGSUI list.
   * They will be used in calculating entries for the output AMBIGSUI list.
   * @return list of suis
   * @throws IOException
   */
  public List getAmbiguousStringIdentifiersAsList() throws IOException;

  /**
   * Returns the complete list of terms from the AMBIGLUI list.
   * They will be used in calculating entries for the output AMBIGLUI list.
   * @return list of terms
   * @throws IOException
   */
  public String[] getAmbiguousTermIdentifiers() throws IOException;

  /**
   * Returns the complete list of terms from the AMBIGLUI list.
   * They will be used in calculating entries for the output AMBIGLUI list.
   * @return list of terms
   * @throws IOException
   */
  public List getAmbiguousTermIdentifiersAsList() throws IOException;

}
