/**************************************************************************
 *
 * Package:     gov.nih.nlm.umls.io
 * Object:      OriginalMRFilesConceptOutputStream.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.io;

import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.AtomAttribute;
import gov.nih.nlm.umls.meta.Attribute;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.Context;
import gov.nih.nlm.umls.meta.ContextMember;
import gov.nih.nlm.umls.meta.Coocurrence;
import gov.nih.nlm.umls.meta.Definition;
import gov.nih.nlm.umls.meta.Language;
import gov.nih.nlm.umls.meta.Mapping;
import gov.nih.nlm.umls.meta.Relationship;
import gov.nih.nlm.umls.meta.SemanticType;
import gov.nih.nlm.util.SystemToolkit;
import gov.nih.nlm.util.UTF8OutputStreamWriter;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

/**
 * This {@link ConceptOutputStream} generates a subset of standard MR files.
 *
 * @author Deborah Shapiro
 */
public class OriginalMRFilesConceptOutputStream implements ConceptOutputStream {

  //
  // Map of file names -> BufferedWriter
  //
  protected Map mr_targets = new HashMap();

  //
  // The target list (sans index files)
  //
  protected static String[] target_files =
    new String[] { "MRATX", "MRCOC", "MRCOLS", "MRCON", "MRCUI", "MRCXT", "MRDEF",
		    "MRFILES", "MRRANK", "MRREL", "MRSAB", "MRSAT",
                    "MRSO", "MRSTY", "MRXNS.ENG", "MRXNW.ENG" };

  //
  // Directory where to write files
  //
  protected String mr_dir;

  //
  // Tracks whether to use VSAB values or RSAB values for SABs.
  //
  protected boolean versioned_output = false;

  //
  // Indicates whether or not to prepend Byte Order Mark (BOM) to indicate
  // UTF-8 encoding.
  //
  protected boolean add_unicode_bom = false;
  
  //
  // List of warnings
  //
  protected List warnings = new ArrayList();


  //
  // Constructor
  //

  /**
   * Instantiates an empty {@link OriginalMRFilesConceptOutputStream}.
   */
  public OriginalMRFilesConceptOutputStream() {}

  //
  // Methods
  //

  /**
   * Opens all target files in the specified directory.  The set of
   * target files is the standard MR files.
   * @param mr_dir the subset directory
   * @throws IOException if the files cannot be opened
   */
  public void open(String mr_dir) throws IOException {

    //
    // Set the output directory
    //
    this.mr_dir = mr_dir;

    //
    // Open files
    //
    for (int i = 0; i < target_files.length; i++) {
      mr_targets.put(target_files[i], openTargetFile(target_files[i]));
    }

    final Language[] languages = Language.getLanguages();
    for (int i = 0; i < languages.length; i++) {
      String file = "MRXW." + languages[i].getAbbreviation();
      mr_targets.put(file, openTargetFile(file));
    }
  }

  /**
   * Closes files opened in {@link #open(String)}.
   * @throws IOException if the stream cannot be closed
   */
  public void close() throws IOException{
    //
    // Iterate through target files, close each one
    //
    final Iterator iter = mr_targets.keySet().iterator();
    while (iter.hasNext())
      ((Writer)mr_targets.get(iter.next())).close();

    //
    // sort unique MRATX
    //
    SystemToolkit.sort(new File(mr_dir,"MRATX").getAbsolutePath(), true);

  }


  /**
   * Writes a {@link Concept} to the appropriate MR files.
   * @param concept the {@link Concept} to write
   * @throws IOException if the files could not be written
   */
  public void writeConcept(Concept concept) throws IOException {

    //
    // Write MRSO
    //
    writeMRSO(concept);

    //
    // Write MRCON
    //
    writeMRCON(concept);

    //
    // Write MRDEF
    //
    writeMRDEF(concept);

    //
    // Write MRSTY
    //
    writeMRSTY(concept);

    //
    // Write MRSAT
    //
    writeMRSAT(concept);

    //
    // Write MRREL
    //
    writeMRREL(concept);

    //
    // Write MRATX
    //
    writeMRATX(concept);

    //
    // Write MRCOC
    //
    writeMRCOC(concept);

    //
    // Write MRCXT
    //
    writeMRCXT(concept);

  }


  /**
   * Writes <code>MRSO</code> entries for a {@link Concept}
   * @param concept the {@link Concept} to write
   * @throws IOException if anything goes wrong
   */
  protected void writeMRSO(Concept concept) throws IOException {

    //
    // Iterate through atoms in concept
    //
    final Atom[] atoms = concept.getAtoms();
    final List mrso_lines = new ArrayList();
    for(int i = 0; i < atoms.length; i++) {

      final Atom atom = atoms[i];

      //
      // Prepare MRSO line
      //
      final StringBuffer sb = new StringBuffer(150);
      sb.append(atom.getConcept().getCui()).append("|");
      sb.append(atom.getLui()).append("|");
      sb.append(atom.getSui()).append("|");
      if (versioned_output)
        sb.append(atom.getSource().getVersionedSourceAbbreviation());
      else
        sb.append(atom.getSource().getRootSourceAbbreviation());
      sb.append("|");
      sb.append(atom.getTty().getAbbreviation()).append("|");
      sb.append(atom.getCode()).append("|");
      sb.append(atom.getSource().getSrl()).append("|");

      //
      // Add line to list
      //
      mrso_lines.add(sb.toString());
    }

    //
    // Sort lines for this concept and write them out
    //
    Collections.sort(mrso_lines);
    String prev_line = null;
    for (int i = 0; i < atoms.length; i++) {
      String this_line = (String)mrso_lines.get(i);
      if (!this_line.equals(prev_line))
        printLine(this_line, "MRSO");
      prev_line = this_line;
    }

  }

  /**
   * Writes <code>MRATX</code> entries for a {@link Concept}
   * @param concept the {@link Concept} to write
   * @throws IOException if anything goes wrong
   */
  private void writeMRATX(Concept concept)
    throws IOException {

    //
    // Iterate through mappings
    //
    final Mapping[] mappings = concept.getMappings();
    for(int i = 0; i < mappings.length; i++) {

      final Mapping mapping = mappings[i];

      //
      // Prepare lines for ATX mappings
      //
      if (mapping.getType().equals("ATX")) {

        //
        // Prepare MRATX line
        //
        final StringBuffer sb = new StringBuffer(200);
        sb.append(mapping.getFrom().getExpr()).append("|");
        if (versioned_output)
          sb.append(mapping.getTo().getSource().getVersionedSourceAbbreviation());
        else
          sb.append(mapping.getTo().getSource().getRootSourceAbbreviation());
        sb.append("|");
        sb.append(mapping.getRel().getAbbreviation()).append("|");
        sb.append(mapping.getTo().getExpr()).append("|");

        //
        // Print line (sort later)
        //
        printLine(sb.toString(), "MRATX");
      }
    }

  }

  /**
   * Writes <code>MRCOC</code> entries for a {@link Concept}
   * @param concept the {@link Concept} to write
   * @throws IOException if anything goes wrong
   */
  private void writeMRCOC(Concept concept)
    throws IOException {

    //
    // Iterate through coocurrences
    //
    final Coocurrence[] coocurrences = concept.getCoocurrences();
    final List mrcoc_lines = new ArrayList();
    for(int i = 0; i < coocurrences.length; i++) {

      final Coocurrence coc = coocurrences[i];

      //
      // Prepare MRCOC line
      //
      final StringBuffer sb = new StringBuffer(200);
      sb.append(coc.getConcept().getCui()).append("|");
      sb.append(coc.getRelatedConcept().getCui()).append("|");
      if (versioned_output)
        sb.append(coc.getSource().getVersionedSourceAbbreviation());
      else
        sb.append(coc.getSource().getRootSourceAbbreviation());
      sb.append("|");
      sb.append(coc.getCot() != null ?
        coc.getCot().getAbbreviation() : "").append("|");
      sb.append(coc.getCof() !=0 ? String.valueOf(coc.getCof()) : "").append("|");
      sb.append(coc.getCoa()).append("|");

      //
      // Add line to list
      //
      mrcoc_lines.add(sb.toString());
    }

    //
    // Sort (unique) lines and write to MRCOC
    //
    Collections.sort(mrcoc_lines);
    String prev_line = null;
    for(int i = 0; i < coocurrences.length; i++) {
      String this_line = (String)mrcoc_lines.get(i);

      if (!this_line.equals(prev_line))
        printLine(this_line, "MRCOC");
      prev_line = this_line;
    }
  }

  /**
   * Writes <code>MRCON</code> entries for a {@link Concept}
   * @param concept the {@link Concept} to write
   * @throws IOException if anything goes wrong
   */
  protected void writeMRCON(Concept concept)
    throws IOException {

    //
    // Obtain atoms and assign LRL field
    //
    final Atom[] atoms = concept.getAtoms();
    final List mrcon_lines = new ArrayList();
    final Map lrl_hm = assignLrl(atoms);

    //
    // Iterate through atoms
    //
    for(int i = 0; i < atoms.length; i++) {

      final Atom atom = atoms[i];

      //
      // Prepare MRCON line
      //
      final StringBuffer sb = new StringBuffer(200);
      sb.append(atom.getConcept().getCui()).append("|");
      sb.append(atom.getLat().getAbbreviation()).append("|");
      if (isLuiSuppressible(atom.getLui(), atoms))
        sb.append(atom.getTs().getAbbreviation().toLowerCase());
      else
        sb.append(atom.getTs().getAbbreviation());
      sb.append("|");
      sb.append(atom.getLui()).append("|");
      sb.append(atom.getStt().getAbbreviation()).append("|");
      sb.append(atom.getSui()).append("|");
      sb.append(atom.getStr()).append("|");
      sb.append(lrl_hm.get(atom.getConcept().getCui() + ":" + atom.getSui())).append("|");

      //
      // Add line to list
      //
      mrcon_lines.add(sb.toString());
    }

    //
    // Sort (unique) lines and write to MRCON
    //
    Collections.sort(mrcon_lines);
    String this_line, prev_line= "";
    for (int i = 0; i < mrcon_lines.size(); i++) {
      this_line = (String)mrcon_lines.get(i);
      if (!this_line.equals(prev_line))
        printLine(this_line, "MRCON");
      prev_line = this_line;
    }
  }

  /**
   * Returns <code>true</code> if the specified LUI is suppressible.
   * @param lui the LUI value to check
   * @param atoms the {@link Atom}s
   * @return <code>true</code> if the LUI is suppressible, <code>false</code> otherwise
   */
  protected boolean isLuiSuppressible(String lui, Atom[] atoms) {

    //
    // Iterate through atoms
    //
    for (int i=0; i<atoms.length; i++) {
      if (lui.equals(atoms[i].getLui())) {
        //
        // If any atom is not suppressible, the LUI is not suppressible
        //
        if (!atoms[i].isSuppressible())
          return false;
      }
    }

    //
    // If all atoms are suppressible, LUI is suppressible
    //
    return true;
  }
  
  /**
   * Assigns the LRL value.
   * @param atoms the {@link Atom}s to be assigned
   * @return {@link Map} of cui:sui tuple mapped to lrl
   */
  protected Map assignLrl(Atom[] atoms) {

    //
    // Iterate through atoms
    //
    final Map lrl_hm = new HashMap();
    for(int i=0; i<atoms.length; i++) {

      //
      // Prepare CUI,SUI tuple as key
      //
      final String cui_sui = atoms[i].getConcept().getCui() + ":" + atoms[i].getSui();

      //
      // If current SRL is lower than LRL for this CUI,SUI, update it to new value
      //
      if(lrl_hm.containsKey(cui_sui)) {
        if(atoms[i].getSource().getSrl().compareTo(lrl_hm.get(cui_sui)) < 0)
          lrl_hm.put(cui_sui, atoms[i].getSource().getSrl());
      }

      //
      // If no LRL has been assigned for this CUI,SUI, use the current SRL
      else
        lrl_hm.put(cui_sui, atoms[i].getSource().getSrl());

    }
    return lrl_hm;
  }

  /**
   * Writes <code>MRCXT</code> entries for a {@link Concept}
   * @param concept the {@link Concept} to write
   * @throws IOException if anything goes wrong
   */
  protected void writeMRCXT(Concept concept)
    throws IOException {

    //
    // Iterate through contexts
    //
    final Context[] contexts = concept.getContexts();
    final List mrcxt_lines = new ArrayList();
    for(int i = 0; i < contexts.length; i++) {

      final Context context = contexts[i];

      //
      // Iterate through context members for each context
      //
      final ContextMember[] context_rels = context.getContextMembers();
      for(int j=0; j<context_rels.length; j++) {

        final ContextMember context_rel = context_rels[j];

        //
        // Prepare MRCXT line
        //
	final StringBuffer sb = new StringBuffer(200);
	sb.append(context.getConcept().getCui()).append("|");
	sb.append(context.getAtom() != null ? context.getAtom().getSui() : "").append("|");
        if (versioned_output)
          sb.append(context.getSource().getVersionedSourceAbbreviation());
        else
          sb.append(context.getSource().getRootSourceAbbreviation());
        sb.append("|");
	sb.append(context.getAtom() != null ? context.getAtom().getCode() : "").append("|");
	sb.append(context.getCxn()).append("|");
	sb.append(context_rel.getCxl()).append("|");
        if (context_rel.getCxl().equals("ANC"))
	  sb.append(context_rel.getRank());
        sb.append("|");
	sb.append(context_rel.getRelatedAtom().getStr()).append("|");
	sb.append(context_rel.getRelatedConcept().getCui()).append("|");
	sb.append(context_rel.getHcd()).append("|");
	sb.append(context_rel.getRela() != null ?
          context_rel.getRela().getAbbreviation() : "").append("|");
	sb.append(context_rel.getXc()).append("|");

        //
        // Add MRCXT line to list
        //
        mrcxt_lines.add(sb.toString());
      }
    }

    //
    // Sort (unique) lines and write to MRCXT)
    //
    Collections.sort(mrcxt_lines);
    String prev_line = "";
    for(int i = 0; i < mrcxt_lines.size(); i++) {
      String this_line = (String)mrcxt_lines.get(i);
      if (!this_line.equals(prev_line))
        printLine(this_line, "MRCXT");
      prev_line = this_line;
    }
  }

  /**
   * Writes <code>MRDEF</code> entries for a {@link Concept}
   * @param concept the {@link Concept} to write
   * @throws IOException if anything goes wrong
   */
  protected void writeMRDEF(Concept concept)
    throws IOException {
    //
    // Iterate through definitions
    //
    Definition[] definitions = concept.getDefinitions();
    List mrdef_lines = new ArrayList();
    for(int i = 0; i < definitions.length; i++) {

      final Definition def = definitions[i];

      //
      // Prepare MRDEF line
      //
      StringBuffer sb = new StringBuffer(200);
      sb.append(def.getConcept().getCui()).append("|");
      if (versioned_output)
        sb.append(def.getSource().getVersionedSourceAbbreviation());
      else
        sb.append(def.getSource().getRootSourceAbbreviation());
      sb.append("|");
      sb.append(def.getDef()).append("|");

      //
      // Add line to list
      //
      mrdef_lines.add(sb.toString());
    }

    //
    // Sort (unique) lines and write to MRDEF
    //
    Collections.sort(mrdef_lines);
    String prev_line = "";
    for (int i = 0; i < definitions.length; i++) {
      String this_line = (String)mrdef_lines.get(i);
      if (!this_line.equals(prev_line))
        printLine(this_line, "MRDEF");
      prev_line = this_line;
    }
  }



  /**
   * Writes <code>MRSTY</code> entries for a {@link Concept}
   * @param concept the {@link Concept} to write
   * @throws IOException if anything goes wrong
   */
  private void writeMRSTY(Concept concept)
    throws IOException {

    //
    // Iterate through STYs
    //
    SemanticType[] semantic_types = concept.getSemanticTypes();
    List mrsty_lines = new ArrayList();
    for (int i = 0; i < semantic_types.length; i++) {

      final SemanticType sty = semantic_types[i];

      //
      // Prepare MRSTY line
      //
      StringBuffer sb = new StringBuffer(200);
      sb.append(sty.getConcept().getCui()).append("|");
      sb.append(sty.getSty().getUi()).append("|");
      sb.append(sty.getSty() == null ? "" : sty.getSty().getSty()).append("|");

      //
      // Add line to list
      //
      mrsty_lines.add(sb.toString());
    }

    //
    // Sort (unique) lines and write to MRSTY
    //
    Collections.sort(mrsty_lines);
    String prev_line = "";
    for(int i = 0; i < semantic_types.length; i++) {
      String this_line = (String)mrsty_lines.get(i);
      if (!prev_line.equals(this_line))
        printLine(this_line, "MRSTY");
      prev_line = this_line;
    }
  }

  /**
   * Writes <code>MRREL</code> entries for a {@link Concept}
   * @param concept the {@link Concept} to write
   * @throws IOException if anything goes wrong
   */
  private void writeMRREL(Concept concept)
    throws IOException {

    //
    // Iterate through relationships
    //
    Relationship[] relationships = concept.getRelationships();
    List mrrel_lines = new ArrayList();
    for(int i = 0; i < relationships.length; i++) {

      final Relationship rel = relationships[i];

      //
      // Prepare MRREL line
      //
      StringBuffer sb = new StringBuffer(200);
      sb.append(rel.getConcept().getCui()).append("|");
      sb.append(rel.getRel() != null ? rel.getRel().getAbbreviation() : "").append("|");
      sb.append(rel.getRelatedConcept().getCui()).append("|");
      if (rel.getRela() != null)
        sb.append(rel.getRela().getAbbreviation());
      sb.append("|");
      if (versioned_output)
        sb.append(rel.getSource().getVersionedSourceAbbreviation());
      else
        sb.append(rel.getSource().getRootSourceAbbreviation());
      sb.append("|");
      if (versioned_output)
        sb.append(rel.getSl().getVersionedSourceAbbreviation());
      else
        sb.append(rel.getSl().getRootSourceAbbreviation());
      sb.append("||");

      //
      // Add line to list
      //
      mrrel_lines.add(sb.toString());
    }

    //
    // Sort (unique) lines and write to MRREL
    //
    Collections.sort(mrrel_lines);
    String prev_line = "";
    for(int i = 0; i < relationships.length; i++) {
      String this_line = (String)mrrel_lines.get(i);
      if (!prev_line.equals(this_line))
        printLine(this_line, "MRREL");
      prev_line = this_line;
    }
  }

  /**
   * Writes <code>MRSAT</code> entries for a {@link Concept}
   * @param concept the {@link Concept} to write
   * @throws IOException if anything goes wrong
   */
  protected void writeMRSAT(Concept concept)
    throws IOException {

    //
    // Iterate through attributes
    //
    final Attribute[] attributes = concept.getAttributes();
    final List mrsat_lines = new ArrayList();
    for(int i = 0; i < attributes.length; i++) {

      final Attribute attribute = attributes[i];

      //
      // Prepare MRSAT line
      //
      final StringBuffer sb = new StringBuffer(200);
      sb.append(attribute.getConcept().getCui()).append("|");
      if (attribute instanceof AtomAttribute && ((AtomAttribute)attribute).getAtom() != null) {
        AtomAttribute aa = (AtomAttribute)attribute;
        sb.append(aa.getAtom().getLui()).append("|");
        sb.append(aa.getAtom().getSui()).append("|");
        sb.append(aa.getAtom().getCode()).append("|");
      } else {
        sb.append("|||");
      }
      sb.append(attribute.getAtn().getAbbreviation()).append("|");
      if (versioned_output)
        sb.append(attribute.getSource().getVersionedSourceAbbreviation());
      else
        sb.append(attribute.getSource().getRootSourceAbbreviation());
      sb.append("|");
      sb.append(attribute.getAtv()).append("|");

      //
      // Add line to list
      //
      mrsat_lines.add(sb.toString());
    }

    //
    // Sort (unique) lines and write to MRSAT
    //
    Collections.sort(mrsat_lines);
    String prev_line = "";
    for (int i = 0; i < attributes.length; i++) {
      String this_line = (String)mrsat_lines.get(i);
      if (!prev_line.equals(this_line))
        printLine(this_line, "MRSAT");
      prev_line = this_line;
    }
  }


  /**
   * Prints the given line to the file with the given name.
   * @param line the line to be printed
   * @param file_name filename to print to
   * @throws IOException
   */
  protected void printLine(String line, String file_name) throws IOException {
    BufferedWriter target = (BufferedWriter)mr_targets.get(file_name);
    target.write(line);
    target.newLine();
    target.flush();
  }

  /**
   * Opens the specified file in the directory passed to {@link #open(String)}.
   * @param name the filename
   * @return a {@link BufferedWriter} for the file
   * @throws IOException if the file could not be opened.
   */
  protected BufferedWriter openTargetFile(String name) throws IOException {
    FileOutputStream fos = new FileOutputStream(new File(mr_dir, name));

    return new BufferedWriter(new UTF8OutputStreamWriter(
      fos, add_unicode_bom));
  }
  
  /**
   * Returns a {@link String}[] of warnings accumulated during
   * processing of the output stream.
   * @return {@link String}[] of warnings
   */
  public String[] getWarnings() {
  	return (String[])warnings.toArray(new String[0]);
  }
}
