/**************************************************************************
 *
 * Package:     gov.nih.nlm.umls.io
 * Object:      RichMRFilesConceptInputStream.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.io;

import gov.nih.nlm.umls.meta.AdditionalRelationshipLabel;
import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.AtomAttribute;
import gov.nih.nlm.umls.meta.AtomCoocurrence;
import gov.nih.nlm.umls.meta.AtomRelationship;
import gov.nih.nlm.umls.meta.Attribute;
import gov.nih.nlm.umls.meta.AttributeName;
import gov.nih.nlm.umls.meta.ComponentHistory;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.ContentView;
import gov.nih.nlm.umls.meta.Context;
import gov.nih.nlm.umls.meta.ContextMember;
import gov.nih.nlm.umls.meta.Coocurrence;
import gov.nih.nlm.umls.meta.CoocurrenceType;
import gov.nih.nlm.umls.meta.Definition;
import gov.nih.nlm.umls.meta.HierarchicalSemanticType;
import gov.nih.nlm.umls.meta.Language;
import gov.nih.nlm.umls.meta.MapObject;
import gov.nih.nlm.umls.meta.Mapping;
import gov.nih.nlm.umls.meta.MetaDescription;
import gov.nih.nlm.umls.meta.Relationship;
import gov.nih.nlm.umls.meta.RelationshipAttribute;
import gov.nih.nlm.umls.meta.RelationshipLabel;
import gov.nih.nlm.umls.meta.SemanticType;
import gov.nih.nlm.umls.meta.Source;
import gov.nih.nlm.umls.meta.SourceIdentifierType;
import gov.nih.nlm.umls.meta.SourceMetaDescription;
import gov.nih.nlm.umls.meta.StringType;
import gov.nih.nlm.umls.meta.Suppressible;
import gov.nih.nlm.umls.meta.TermStatus;
import gov.nih.nlm.umls.meta.TermType;
import gov.nih.nlm.umls.meta.TermTypeGroup;
import gov.nih.nlm.umls.meta.impl.AtomAttributeImpl;
import gov.nih.nlm.umls.meta.impl.AtomCoocurrenceImpl;
import gov.nih.nlm.umls.meta.impl.AtomImpl;
import gov.nih.nlm.umls.meta.impl.AtomRelationshipImpl;
import gov.nih.nlm.umls.meta.impl.AttributeImpl;
import gov.nih.nlm.umls.meta.impl.ComponentHistoryImpl;
import gov.nih.nlm.umls.meta.impl.ConceptImpl;
import gov.nih.nlm.umls.meta.impl.ContextImpl;
import gov.nih.nlm.umls.meta.impl.ContextMemberImpl;
import gov.nih.nlm.umls.meta.impl.CoocurrenceImpl;
import gov.nih.nlm.umls.meta.impl.DefinitionImpl;
import gov.nih.nlm.umls.meta.impl.MapObjectImpl;
import gov.nih.nlm.umls.meta.impl.MappingImpl;
import gov.nih.nlm.umls.meta.impl.MetaDescriptionImpl;
import gov.nih.nlm.umls.meta.impl.PlaceholderAtom;
import gov.nih.nlm.umls.meta.impl.PlaceholderConcept;
import gov.nih.nlm.umls.meta.impl.RelationshipAttributeImpl;
import gov.nih.nlm.umls.meta.impl.RelationshipImpl;
import gov.nih.nlm.umls.meta.impl.SemanticTypeImpl;
import gov.nih.nlm.umls.meta.impl.SourceMetaDescriptionImpl;
import gov.nih.nlm.util.FieldedStringTokenizer;
import gov.nih.nlm.util.SystemToolkit;
import gov.nih.nlm.util.UTF8InputStreamReader;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.SequenceInputStream;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Vector;
import java.util.zip.GZIPInputStream;

/**
 * This {@link ConceptInputStream} builds {@link Concept} objects from a
 * full set of "Rich Release Format" MR files.
 *
 * @author Deborah Shapiro
 */
public class RichMRFilesConceptInputStream implements ConceptInputStream {

  //
  // Private Fields
  //

  //
  // Readers for the various files
  //
  protected PushBackReader
    mrconso_source,
    mrcxt_source,
    mrsty_source,
    mrdef_source,
    mrrel_source,
    mrcoc_source,
    mrsat_source,
    mrhist_source,
    mrmap_source,
    mrhier_source,
    mrdoc_source;

  //protected ArrayList mrcxt_name_variations = new ArrayList();

  //
  // Map of atoms in current concept
  //
  protected Map atom_hm = new HashMap();;

  //
  // Map of relationships in current concept
  //
  protected Map rel_hm = null;

  //
  // List of file names
  //
  protected String [] file_names;

  //
  // Installation sources containing rich MR files.
  //
  protected Object [] install_paths;

  //
  // Current concept
  //
  protected Concept this_concept = new ConceptImpl();

  //
  // Total number of MRCONSO lines
  //
  protected int mrconso_lines = 0;

  //
  // Number of MRCONSO lines read so far
  //
  protected int mrconso_counter = 0;

  //
  // FROMRSAB if current concept is a map set
  //
  protected String from_rsab;

  //
  // FROMVSAB if current concept is a map set
  //
  protected String from_vsab;

  //
  // TORSAB if current concept is a map set
  //
  protected String to_rsab;

  //
  // TOVSAB if current concept is a map set
  //
  protected String to_vsab;

  //
  // Additional suffix to be added after .RRF such as .RRF.del
  //
  protected String add_suffix = "";

  //
  // FieldedStringTokenizer.split()
  //
  protected final String[] tokens = new String[50];
  
  //
  // List of warnings
  //
  protected List warnings = new ArrayList();

  //
  // Constructors
  //

  /**
   * Instantiates an empty {@link RichMRFilesConceptInputStream};
   */
  public RichMRFilesConceptInputStream() { }

  //
  // Methods
  //

  /**
   * Opens all of the MR Files in preparation for subsetting.
   * @param paths the sources containing the files
   * @throws IOException if any files can not be opened
   */
  public void open(String[] paths) throws IOException {

    //
    // Prepare install path info
    //
    final List install_list = new ArrayList();
    //
    // Iterate through paths
    //
    for (int i=0; i<paths.length; i++) {

      //
      // Determine if directory or ZipFile
      //
      final File file_path = new File(paths[i]);
      boolean zip_file = false;
      final File[] file_list = file_path.listFiles();
      for (int j=0; j<file_list.length; j++) {
        if (file_list[j].getAbsolutePath().endsWith("meta.nlm")) {
          install_list.add(file_list[j].getCanonicalPath());
          zip_file = true;
        }
      }
      if (!zip_file)
        install_list.add(file_path);
    }

    install_paths = install_list.toArray();

    //
    // Open the MR files
    //
    openMRFiles();

  }

  /**
   * Opens the Rich Release Format MR files.
   * @throws IOException if any files cannot be opened
   */
  protected void openMRFiles() throws IOException {
    //
    // Open content files
    //
    mrconso_source = openSourceFile("MRCONSO.RRF" + add_suffix);
    mrsty_source = openSourceFile("MRSTY.RRF" + add_suffix);
    mrdef_source = openSourceFile("MRDEF.RRF" + add_suffix);
    mrrel_source = openSourceFile("MRREL.RRF" + add_suffix);
    mrcoc_source = openSourceFile("MRCOC.RRF" + add_suffix);
    mrsat_source = openSourceFile("MRSAT.RRF" + add_suffix);
    mrhist_source = openSourceFile("MRHIST.RRF" + add_suffix);
    mrmap_source = openSourceFile("MRMAP.RRF" + add_suffix);
    mrhier_source = openSourceFile("MRHIER.RRF" + add_suffix);
    mrcxt_source = openSourceFile("MRCXT.RRF" + add_suffix);

    //
    // Cache line counts for MRCONSO and index files
    //
    getLineCounts();

    //
    // Open MRDOC, initialize "meta descriptions"
    //
    mrdoc_source = openSourceFile("MRDOC.RRF" + add_suffix);
    String line;
    while ((line = mrdoc_source.readLine()) != null &&
	   !line.equals("")) {
      MetaDescription[] md_array = new MetaDescription[1];
      md_array[0] = new MetaDescriptionImpl(line);
      if (md_array[0].getKey().equals("ATN")&&
                 md_array[0].getType().equals("expanded_form")) {
        AttributeName.initialize(md_array);
      } else if (md_array[0].getKey().equals("MAPATN")&&
                   md_array[0].getType().equals("expanded_form")) {
        AttributeName.initialize(md_array);
      } else if (md_array[0].getKey().equals("COT")&&
                 md_array[0].getType().equals("expanded_form")) {
        CoocurrenceType.initialize(md_array);
      } else if (md_array[0].getKey().equals("LAT")&&
                 md_array[0].getType().equals("expanded_form")) {
        Language.initialize(md_array);
      } else if ((md_array[0].getKey().equals("STYPE1") ||
                  md_array[0].getKey().equals("STYPE2") ||
                  md_array[0].getKey().equals("STYPE") ||
                  md_array[0].getKey().equals("FROMTYPE") ||
                  md_array[0].getKey().equals("TOTYPE")) &&
                  md_array[0].getType().equals("expanded_form")) {
        SourceIdentifierType.initialize(md_array);
      } else if (md_array[0].getKey().equals("REL")&&
                 md_array[0].getType().equals("expanded_form")) {
        RelationshipLabel.initialize(md_array);
      } else if (md_array[0].getKey().equals("RELA")&&
                 md_array[0].getType().equals("expanded_form")) {
        AdditionalRelationshipLabel.initialize(md_array);
      } else if (md_array[0].getKey().equals("SUPPRESS") &&
                 md_array[0].getType().equals("expanded_form")) {
        Suppressible.initialize(md_array);
      } else if (md_array[0].getKey().equals("STT")&&
                 md_array[0].getType().equals("expanded_form")) {
        StringType.initialize(md_array);
      } else if (md_array[0].getKey().equals("TS")&&
                 md_array[0].getType().equals("expanded_form")) {
        TermStatus.initialize(md_array);
      } else if (md_array[0].getKey().equals("TTY") &&
                 md_array[0].getType().equals("expanded_form")) {
        TermType.initialize(md_array);
      } else if (md_array[0].getKey().equals("TTY") &&
                 md_array[0].getType().equals("tty_class")) {
        TermTypeGroup.initialize(md_array);
      }
    }
    mrdoc_source.close();
    mrdoc_source = openSourceFile("MRDOC.RRF" + add_suffix);

  }

  /**
   * Close files opened during {@link #open(String[])}.
   * @throws IOException if files can not be closed
   */
  public void close() throws IOException {

    //
    // Close content files
    //
    mrconso_source.close();
    mrsty_source.close();
    mrdef_source.close();
    mrrel_source.close();
    mrcxt_source.close();
    mrcoc_source.close();
    mrsat_source.close();
    mrhist_source.close();
    mrmap_source.close();
    mrhier_source.close();

    //
    // Close MRDOC
    //
    mrdoc_source.close();
  }


  /**
   * Returns a {@link Concept} containing data from the MR files.
   * @return a {@link Concept} containing data from the MR files
   * @throws IOException if files could not be read
   */
  public Concept readConcept() throws IOException {
    String current_cui = "";
    this_concept = new ConceptImpl();

    //
    // Create Atoms for each group of rows from MRCONSO with the given CUI
    //
    if (getAtoms(this_concept, current_cui) == null)
      return null;

    current_cui = this_concept.getCui();
    //
    // Create Definitions for each row in MRDEF with the given CUI
    //
    getDefinitions(this_concept, current_cui);

    //
    // Create SemanticTypes for each row in MRSTY with the given CUI
    //
    getSemanticTypes(this_concept, current_cui);

    //
    // Create Relationships for each row in MRREL with the given CUI
    //
    getRelationships(this_concept, current_cui);

    //
    // Create Attributes for each row in MRSAT with the given CUI
    //
    getAttributes(this_concept, current_cui);



    //
    // Create Coocurrence for each row in MRCOC with the given CUI
    //
    getCoocurrences(this_concept, current_cui);

    //
    // Create Contexts and ContextMembers for rows in MRCXT
    //
    getContexts(this_concept, current_cui);

    //
    // Create ComponentHistories for rows in MRHIST with the given CUI
    //
    getComponentHistories(this_concept, current_cui);

    //
    // Create Mappings for rows in MRMAP with the given CUI
    //
    getMappings(this_concept, current_cui);

    //
    // Concept is fully populated, return it
    //
    return this_concept;
  }

  /**
   * Returns the percentage of <code>MRCONSO</code> lines that have been processed.
   * @return the percentage of <code>MRCONSO</code> lines that have been processed
   */
  public int getConceptProgress() {
    return (int)((mrconso_counter/(double)mrconso_lines)*100);
  }

  //
  // Protected Methods
  //

  /**
   * Opens the source file from the most current of the given
   * installation paths in which this source is available.
   * @param name the filename
   * @return a {@link PushBackReader} on the file
   * @throws IOException if the file could not be opened.
   */
  protected PushBackReader openSourceFile(String name) throws IOException {
    final Map name_variations = new HashMap();

    //
    // Iterate through install paths in reverse order
    //
    for (int i = install_paths.length - 1; i >= 0; i--) {
      //
      // Handle ZipFile case
      //
      if (!(install_paths[i] instanceof File)) {

        //
        // Open file.  Handle split files
        //
        final String zip_file = (String)install_paths[i];
        final String[] entries = SystemToolkit.getZipEntries(zip_file);
        for (int j = 0; j < entries.length; j++) {
          final String entry_name = entries[j];

          //
          // File not split, wrap in GZIP stream and return
          //
          if (entry_name.endsWith(name + ".gz")) {
            return new PushBackReader(
              new BufferedReader(
                new InputStreamReader(
                  new GZIPInputStream(
                    SystemToolkit.getZipInputStream(zip_file, entry_name),
                    1024 * 256),
                "UTF-8")));
          }
          // else file split
          else if (entry_name.indexOf(name) != -1) {
            name_variations.put(entry_name.substring(entry_name.indexOf(
                name)), zip_file);
          }
        }
      }
      //
      // Handle directory case
      //
      else {
        File install_dir = (File) install_paths[i];

        //
        // Return a BufferedReader for non-split case
        //
        File[] install_files = install_dir.listFiles();
        for (int j = 0; j < install_files.length; j++) {
          String this_file = install_files[j].toString();
          if (this_file.endsWith(name)) {
            return new PushBackReader(new BufferedReader(new UTF8InputStreamReader(
                new BufferedInputStream(new FileInputStream(new File(install_dir, name)))), 256 * 1024));
          }
          if (this_file.indexOf("CHANGE") != -1) {
            if (new File(install_dir, name).exists())
              return new PushBackReader(new BufferedReader(new UTF8InputStreamReader(
                  new BufferedInputStream(new FileInputStream(new File(install_dir, name)))), 256 * 1024));
          }
        }
      }
    }
    //
    // Handle split file case (create a SequenceInputStream)
    //
    if (name_variations.size() > 1) {
      final String[] name_array =
          (String[]) name_variations.keySet().toArray(new String[0]);
      Arrays.sort(name_array);
      final Vector input_streams = new Vector();
      for (int j = 0; j < name_variations.size(); j++) {
        final String zip_file = (String)name_variations.get(name_array[j]);
        String[] entries = SystemToolkit.getZipEntries(zip_file);
        for (int k = 0; k < entries.length; k++) {
          final String entry_name = entries[k];
          if (entry_name.indexOf(name_array[j]) != -1) {
            input_streams.add(
              new GZIPInputStream(
                SystemToolkit.getZipInputStream(zip_file,entry_name), 1024 * 256));
          }
        }
      }
      SequenceInputStream sequence_input_stream =
          new SequenceInputStream(input_streams.elements());
      return new PushBackReader(new BufferedReader(new InputStreamReader(
          sequence_input_stream, "UTF-8")));
    }

    //
    // If we get here, the file was not found.
    //
    //throw new IOException(name + "file not found.");
    return new PushBackReader(new BufferedReader(new StringReader("")));
  }


  /**
   * Reads through <code>MRFILES.RRF</code> to determine the size
   * of <code>MRCONSO.RRF</code>.
   * @return {@link Map} of file names to lines in file
   * @throws IOException if the file cannot be opened or read
   */
  protected Map getLineCounts() throws IOException {

    final Map lines_in_file = new HashMap();

    //
    // Iniitalize mrconso count
    //
    mrconso_counter = 0;

    //
    // Read MRFILES and cache entries
    //
    final Object[][] mrfiles = SystemToolkit.readFieldedReader(openSourceFile("MRFILES.RRF" + add_suffix), "|");
    for(int i=0; i<mrfiles.length; i++) {
      lines_in_file.put(mrfiles[i][0], mrfiles[i][4]);
    }

    //
    // Save up MRCONSO.RRF line count
    //
    if (lines_in_file.containsKey("MRCONSO.RRF" + add_suffix))
      mrconso_lines = Integer.parseInt((lines_in_file.get("MRCONSO.RRF" + add_suffix)).toString());
    else
      throw new IOException("MRCONSO.RRF" + add_suffix + " not present in MRFILES.RRF" + add_suffix);

    return lines_in_file;
  }

  /**
   * Instantiates an {@link Atom} from a <code>MRCONSO.RRF</code> line.
   * @param line a line from <code>MRCONSO.RRF</code>
   * @return the {@link Atom} matching the line
   * @throws IOException if there is a problem with the file
   */
  protected Atom newInstanceFromMRCONSO(String line) throws IOException{

    //
    // Instantiate a default atom
    //
    final Atom a = new AtomImpl();

    //
    // Parse line and count fields 19
    //
    FieldedStringTokenizer.split(line, "|", 19, tokens);

    //
    // Populate atom
    //
    a.setLanguage(Language.getLanguage(tokens[1]));
    final String tst = tokens[2];
    final TermStatus ts = TermStatus.getTermStatus(tst);
    if (ts == null && !tst.equals("X"))
      throw new IOException(ts + " TermStatus not in MRDOC");
    a.setTermStatus(ts);
    a.setLui(tokens[3]);
    final String sttt = tokens[4];
    final StringType stt = StringType.getStringType(sttt);
    if (stt == null &&
        !sttt.equals("X"))
      throw new IOException(sttt + " StringType not in MRDOC");
    a.setStringType(stt);
    a.setSui(tokens[5]);
    a.setIspref(tokens[6]);
    a.setAui(tokens[7]);
    a.setSaui(tokens[8]);
    a.setScui(tokens[9]);
    a.setSdui(tokens[10]);
    final Source sab = Source.getSource(tokens[11]);
    if (sab == null)
      throw new IOException(tokens[11] + " Source not in MRSAB");
    a.setSource(sab);
    final TermType tty = TermType.getTermType(tokens[12]);
    if (tty == null)
      throw new IOException(tokens[12] + " TermType not in MRDOC");
    a.setTermType(tty);
    a.setCode(tokens[13]);
    a.setStr(tokens[14]);
    a.setSuppress(Suppressible.getSuppressible(tokens[16]));
    a.setCvf(tokens[17].equals("") ? null : new ContentView(tokens[17]));
    return a;
  }

  /**
   * Instantiates {@link Atom}s from the <code>MRCONSO.RRF</code> file matching the
   * specified CUI and adds them to the current {@link Concept}.
   * @param this_concept the current {@link Concept}
   * @param current_cui the current CUI
   * @return {@link Concept} that will contain the {@link Atom}s
   * @throws IOException if anything goes wrong
   */
  protected Concept getAtoms(Concept this_concept, String current_cui)
    throws IOException {
    //
    // Iterate through atoms
    //
    String line = "";
    int atom_counter = 0;
    atom_hm = new HashMap();
    while((line = mrconso_source.readLine()) != null) {
      //
      // Matching line found
      //
      if (line.startsWith(current_cui) || current_cui.length() == 0) {

        //
        // Connect atom and concept
        //
        final Atom atom = newInstanceFromMRCONSO(line);
        atom_hm.put(atom.getAui(), atom);
        atom.setConcept(this_concept);
        this_concept.addAtom(atom);
        if (current_cui.equals(""))
          current_cui = new String(line.substring(0, line.indexOf("|")));
        mrconso_counter++;
        atom_counter++;
      }

      //
      // went too far - break and reset
      //
      else {
        break;
      }
    }
    mrconso_source.push(line);

    //
    // if any atoms were read return the concept
    //
    if(atom_counter != 0) {
      this_concept.setCui(current_cui);
      return this_concept;
    }

    //
    // No atoms read, we are finished
    //
    return null;
  }

  /**
   * Instantiates an {@link Attribute} from a <code>MRSAT.RRF</code> line.
   * @param line a line from <code>MRSAT.RRF</code>
   * @return the {@link Attribute} matching the line
   * @throws IOException if there is a problem with the file
   */
  private  Attribute newInstanceFromMRSAT(String line)
    throws IOException {

    //
    // Parse line and count tokens
    //
    FieldedStringTokenizer.split(line, "|", 14, tokens);

    //
    // For mapping sets, set the mapping SAB values.
    //
    if (tokens[8].equals("FROMRSAB"))
      from_rsab = tokens[10];
    if (tokens[8].equals("TORSAB"))
      to_rsab = tokens[10];
    if (tokens[8].equals("FROMVSAB"))
      from_vsab = tokens[10];
    if (tokens[8].equals("TOVSAB"))
      to_vsab = tokens[10];

    Attribute a = null;

    //
    // Handle the AtomAttribute case
    //
    if (tokens[4].equals("CUI")) {
        a = new AttributeImpl();

    } else if (tokens[4].indexOf("RUI") == -1) {
      AtomAttribute aa = new AtomAttributeImpl();
      Atom atom = getAtom(tokens[3]);
      aa.setAtom(atom);
      atom.addAttribute(aa);
      a = aa;
    }

    //
    // Handle the RelationshipAttribute case
    //
    else {
      RelationshipAttribute ra = new RelationshipAttributeImpl();
      Relationship rel = getRelationship(tokens[3]);
      ra.setRelationship(rel);
      rel.addAttribute(ra);
      a = ra;
    }

    //
    // Common fields
    //
    final SourceIdentifierType stype = SourceIdentifierType.getSourceIdentifierType(tokens[4]);
    if (stype == null)
      throw new IOException(tokens[4] + " SourceIdentifierType not in MRDOC");
    a.setSourceIdentifierType(stype);
    a.setAtui(tokens[6]);
    a.setSatui(tokens[7]);
    final AttributeName atn = AttributeName.getAttributeName(tokens[8]);
    if (atn == null)
      throw new IOException(tokens[8] + " AttributeName not in MRDOC");
    a.setAtn(atn);
    final Source sab = Source.getSource(tokens[9]);
    if (sab == null)
      throw new IOException(tokens[9] + " Source not in MRSAB");
    a.setSource(sab);
    a.setAtv(tokens[10]);
    final Suppressible sup = Suppressible.getSuppressible(tokens[11]);
    if (sup == null)
      throw new IOException(tokens[11] + " Suppressible not in MRDOC");
    a.setSuppress(sup);
    a.setCvf(tokens[12].equals("") ? null : new ContentView(tokens[12]));
    return a;

  }

  /**
   * Gets {@link Attribute}s from the <code>MRSAT.RRF</code> files with the
   * given cui and adds them to the current {@link Concept}.
   * @param this_concept the current {@link Concept}
   * @param current_cui the current CUI
   * @throws IOException if problem reading the file
   */
  protected void getAttributes(Concept this_concept, String current_cui)
    throws IOException {

    //
    // Iterate through MRSAT
    //
    String line = "";
    while((line = mrsat_source.readLine()) != null) {

      //
      // Skip lines that do not match
      //
      if (line.compareTo(current_cui) < 0) { // line_cui < current_cui
        continue;
      }

      //
      // Matching line found, connect attribute and concept
      //
      else if (line.startsWith(current_cui)) {
        final Attribute sat = newInstanceFromMRSAT(line);
        sat.setConcept(this_concept);
        this_concept.addAttribute(sat);
      }

      //
      // went too far - break and reset
      //
      else {
        break;
      }
    }

    mrsat_source.push(line);
  }

  /**
   * Gets {@link Context}s from the <code>MRCXT.RRF</code> and
   * <code>MRHIER.RRF</code> files with the
   * given cui and adds them to this_concept.
   * @param this_concept {@link Concept} that will contain the {@link Context}s
   * @param current_cui {@link String} representation of the cui
   * @throws IOException if problem reading from <code>MRCXT.RRF</code> or
   * <code>MRHIER.RRF</code>
   */
  protected void getContexts(Concept this_concept, String current_cui)
    throws IOException {

    //
    // Iterate through MRHIER.RRF
    //
    String line = "";
    final Map current_cui_contexts = new HashMap();
    final StringBuffer sb = new StringBuffer(4+10+20);

    // Create Contexts for each row in MRHIER with the given CUI
    while((line = mrhier_source.readLine()) != null) {

      //
      // Skip lines that do not match
      //
      if (line.compareTo(current_cui) < 0) { // line_cui < current_cui
        continue;
      }

      //
      //  Match found, connect context and concept
      //
      else if (line.startsWith(current_cui)) {
        final Context c = newInstanceFromMRHIER(line);
        c.setConcept(this_concept);
        this_concept.addContext(c);
        sb.setLength(0);
        sb.append(c.getCxn()).append(c.getAtom().getAui()).append(c.getSource().getRsab());
        current_cui_contexts.put(sb.toString(), c);
      }

      //
      // went too far - break and reset
      //
      else {
        break;
      }
    }

    mrhier_source.push(line);

    //
    // Iterate through MRCXT.RRF
    //
    while((line = mrcxt_source.readLine()) != null) {
      //
      // Skip lines that do not match
      //
      if (line.compareTo(current_cui) < 0)
        continue;

      //
      // Matching line found, connect context with context members
      //
      else if (line.startsWith(current_cui)) {

        FieldedStringTokenizer.split(line, "|", 16, tokens);
        final String mrcxt_aui = tokens[2];
        final String mrcxt_sab = tokens[3];
        int mrcxt_cxn = Integer.parseInt(tokens[5]);
        sb.setLength(0);
        sb.append(mrcxt_cxn).append(mrcxt_aui).append(mrcxt_sab);
        final Context c = (Context)current_cui_contexts.get(sb.toString());
        if (c != null) {
          final ContextMember cxn_rel = newInstanceFromMRCXTRel(line);
          c.addContextMember(cxn_rel);
        }

      }

      //
      // went too far -  break and reset
      //
      else {
        break;
      }
    }

    mrcxt_source.push(line);
  }

  /**
   * Instantiates an {@link Context} from a <code>MRHIER.RRF</codE> line.
   * @param line a line from <code>MRHIER.RRF</code>
   * @return the {@link Context} matching the line
   * @throws IOException if there is a problem with the file
   */
  protected Context newInstanceFromMRHIER(String line)
    throws IOException {

    //
    // Instantiate a default Context
    //
    final Context cxt = new ContextImpl();

    //
    // Parse line and count fields
    //
    FieldedStringTokenizer.split(line, "|", 10, tokens);

    //
    // Populate context
    //
    final Atom atom = getAtom(tokens[1]);
    cxt.setAtom(atom);
    atom.addContext(cxt);
    cxt.setCxn(Integer.parseInt(tokens[2]));
    cxt.setPaui(tokens[3]);
    final Source sab = Source.getSource(tokens[4]);
    if (sab == null)
      throw new IOException(tokens[4] + " Source not in MRSAB");
    cxt.setSource(sab);
    final AdditionalRelationshipLabel rela = AdditionalRelationshipLabel.
        getAdditionalRelationshipLabel(tokens[5]);
    if (rela == null)
      throw new IOException(tokens[5] +
                            " AdditionalRelationshipLabel not in MRDOC");
    cxt.setRela(rela);
    cxt.setPtr(tokens[6]);
    cxt.setHcd(tokens[7]);
    cxt.setCvf(tokens[8].equals("") ? null : new ContentView(tokens[8]));
    return cxt;
  }

  /**
   * Instantiates an {@link ContextMember} from a <code>MRCXT.RRF</code> line.
   * @param line a line from <code>MRCXT.RRF</code>
   * @return the {@link ContextMember} matching the line
   * @throws IOException if there is a problem with the file
   */
  private  ContextMember newInstanceFromMRCXTRel(String line)
    throws IOException {

    //
    // Instantiate an atom context member
    //
    final ContextMember cm = new ContextMemberImpl();

    //
    // Parse line and count fields
    //
    FieldedStringTokenizer.split(line, "|", 16, tokens);

    //
    // Populate context member
    //
    // ignore fields 0-5
    cm.setCxl(tokens[6]);
    cm.setRank(tokens[7].length() == 0 ? 0 : Integer.parseInt(tokens[7]));
    final Concept related_concept = new PlaceholderConcept();
    related_concept.setCui(tokens[9]);
    cm.setRelatedConcept(related_concept);
    final Atom related_atom = new PlaceholderAtom();
    related_atom.setAui(tokens[10]);
    cm.setRelatedAtom(related_atom);
    cm.getRelatedAtom().setStr(tokens[8]);
    cm.setHcd(tokens[11]);
    final AdditionalRelationshipLabel rela = AdditionalRelationshipLabel.getAdditionalRelationshipLabel(tokens[12]);
    if (rela == null)
      throw new IOException(tokens[12] + " AdditionalRelationshipLabel not in MRDOC");
    cm.setRela(rela);
    cm.setXc(tokens[13]);

    return cm;
  }

  /**
   * Instantiates an {@link Coocurrence} from a <code>MRCOC.RRF</code> line.
   * @param line a line from <code>MRCOC.RRF</code>
   * @return the {@link Coocurrence} matching the line
   * @throws IOException if there is a problem with the file
   */
  private  Coocurrence newInstanceFromMRCOC(String line)
    throws IOException {

    //
    // Parse line and count fields
    //
    FieldedStringTokenizer.split(line, "|", 10, tokens);

    //
    // Instantiate either default coocurrence or AtomCoocurrence
    //
    Coocurrence coc = null;

    if (tokens[1].length() == 0)
      coc = new CoocurrenceImpl();
    else {
      final AtomCoocurrence ac = new AtomCoocurrenceImpl();
      final Atom atom = getAtom(tokens[1]);
      ac.setAtom(atom);
      atom.addCoocurrence(coc);
      final Atom related_atom = getAtom(tokens[3]);
      ac.setRelatedAtom(related_atom);
      coc = ac;
    }

    //
    // Populate shared coocurrence fields
    //
    final Concept related_concept = new PlaceholderConcept();
    related_concept.setCui(tokens[2]);
    coc.setRelatedConcept(related_concept);
    final Source sab = Source.getSource(tokens[4]);
    if (sab == null)
      throw new IOException(tokens[4] + " Source not in MRSAB");
    coc.setSource(sab);
    CoocurrenceType cot = CoocurrenceType.getCoocurrenceType(tokens[5]);
    if (cot == null)
      throw new IOException(tokens[5] + " CoocurrenceType not in MRSAB");
    coc.setCot(cot);
    coc.setCof(tokens[6].equals("") ? 0 : Integer.parseInt(tokens[6]));
    coc.setCoa(tokens[7]);
    coc.setCvf(tokens[8].equals("") ? null : new ContentView(tokens[8]));
    return coc;
  }

 /**
   * Gets {@link Coocurrence}s from the <code>MRCOC.RRF</code> file with the
   * given cui and adds them to the current {@link Concept}.
   * @param this_concept the current {@link Concept}
   * @param current_cui the current CUI
   * @throws IOException if problem reading from <code>MRCOC.RRF</code>
   */
  protected void getCoocurrences(Concept this_concept, String current_cui)
    throws IOException {

    //
    // Iterate through MRCOC.RRF
    //
    String line = "";
    while ((line = mrcoc_source.readLine()) != null) {
      //
      // Skip lines that do not match
      //
      if (line.compareTo(current_cui) < 0)
        continue;

      //
      // Matching line found, connect coocurrence and concept
      //
      else if (line.startsWith(current_cui)) {
        final Coocurrence coc = newInstanceFromMRCOC(line);
        coc.setConcept(this_concept);
        this_concept.addCoocurrence(coc);
      }

      //
      // went too far - break and reset
      //
      else {
        break;
      }
    }

    mrcoc_source.push(line);
  }

  /**
   * Instantiates an {@link Definition} from a <code>MRDEF.RRF</code> line.
   * @param line a line from <code>MRDEF.RRF</code>
   * @return the {@link Definition} matching the line
   * @throws IOException if there is a problem with the file
   */
  private  Definition newInstanceFromMRDEF(String line)
    throws IOException {

    //
    // Instantiate a default definition
    //
    final Definition d = new DefinitionImpl();

    //
    // Parse line and count tokens
    //
    FieldedStringTokenizer.split(line, "|", 9, tokens);

    //
    // Populate definition
    //
    final Atom atom = getAtom(tokens[1]);
    d.setAtom(atom);
    atom.addDefinition(d);
    d.setAtui(tokens[2]);
    d.setSatui(tokens[3]);
    final Source sab = Source.getSource(tokens[4]);
    if (sab == null)
      throw new IOException(tokens[4] + " Source not in MRSAB");
    d.setSource(sab);
    d.setDef(tokens[5]);
    final Suppressible sup = Suppressible.getSuppressible(tokens[6]);
    if (sup == null)
      throw new IOException(tokens[6] + " Suppressible not in MRDOC");
    d.setSuppress(sup);
    d.setCvf(tokens[7].equals("") ? null : new ContentView(tokens[7]));

    return d;
  }

 /**
   * Gets {@link Definition}s from the <code>MRDEF.RRF</code> file with the
   * given cui and adds them to the current {@link Concept}.
   * @param this_concept the current {@link Concept}
   * @param current_cui the current CUI
   * @throws IOException if problem reading from <code>MRDEF.RRF</code>
   */
  protected void getDefinitions(Concept this_concept, String current_cui)
      throws IOException {

    //
    // Iterate through MRDEF.RRF
    //
    String line = "";
    while ((line = mrdef_source.readLine()) != null ) {
      //
      // Skip lines that do not match
      //
      if (line.compareTo(current_cui) < 0)
        continue;

      //
      // Match found, connect definition and concept
      //
      else if (line.startsWith(current_cui)) {
        final Definition d = newInstanceFromMRDEF(line);
        d.setConcept(this_concept);
        this_concept.addDefinition(d);
      }

      //
      // went too far - break and reset
      //
      else {
        break;
      }
    }

    mrdef_source.push(line);
  }


  /**
   * Instantiates an {@link SemanticType} from a <code>MRSTY.RRF</code> line.
   * @param line a line from <code>MRSTY.RRF</code>
   * @return the {@link SemanticType} matching the line
   * @throws IOException if there is a problem with the file
   */
  private  SemanticType newInstanceFromMRSTY(String line)
    throws IOException {

    //
    // Instantiate default semantic type
    //
    final SemanticType d = new SemanticTypeImpl();

    //
    // Parse line and count tokens
    //
    FieldedStringTokenizer.split(line, "|", 7, tokens);

    //
    // Populate semantic type
    //
    HierarchicalSemanticType hst = HierarchicalSemanticType.getHierarchicalSemanticType(tokens[3]);
    if (hst == null)
      throw new IOException(tokens[3] + " HierarchicalSemanticType not in MRSTY");
    d.setSty(hst);
    d.setAtui(tokens[4]);

    return d;
  }

  /**
   * Gets {@link SemanticType}s from the <code>MRSTY.RRF</code> file with the
   * given cui and adds them to the current {@link Concept}.
   * @param this_concept the current {@link Concept}
   * @param current_cui the current CUI
   * @throws IOException if problem reading from <code>MRSTY.RRF</code>
   */
  protected void getSemanticTypes(Concept this_concept, String current_cui)
     throws IOException {

    //
    // Iterate through MRSTY.RRF
    //
    String line = "";
    while ((line = mrsty_source.readLine()) != null) {

      //
      // Skip lines that do not match
      //
      if (line.compareTo(current_cui) < 0)
        continue;

      //
      // Matching line found, connect semantic type and concept
      //
      else if (line.startsWith(current_cui)) {
        final SemanticType sty = newInstanceFromMRSTY(line);
        sty.setConcept(this_concept);
        this_concept.addSemanticType(sty);
      }

      //
      // went too far - break and reset
      //
      else {
        break;
      }
    }

    mrsty_source.push(line);
  }

  /**
   * Instantiates an {@link Relationship} from a <code>MRREL.RRF</code> line.
   * @param line a line from <code>MRREL.RRF</code>
   * @return the {@link Relationship} matching the line
   * @throws IOException if there is a problem with the file
   */
  private  Relationship newInstanceFromMRREL(String line)
    throws IOException{

    //
    // Parse line and count tokens
    //
    FieldedStringTokenizer.split(line, "|", 17, tokens);

    //
    // Instantiate either a default relationship or an AtomRelationship
    //
    Relationship r = null;
    if (tokens[1].equals(""))
      r = new RelationshipImpl();
    else {
      final AtomRelationship ar = new AtomRelationshipImpl();

      //
      // Populate AtomRelationship fields
      //
      final Atom atom = getAtom(tokens[1]);
      ar.setAtom(atom);
      atom.addRelationship(ar);
      final Atom related_atom = getAtom(tokens[5]);
      related_atom.setAui(tokens[5]);
      ar.setRelatedAtom(related_atom);
      r = ar;

    }

    //
    // Populate shared fields
    //
    final SourceIdentifierType stype1 = SourceIdentifierType.getSourceIdentifierType(tokens[2]);
    if (stype1 == null)
      throw new IOException(tokens[2] + " SourceIdentifierType not in MRDOC");
    r.setSourceIdentifierType(stype1);

    final SourceIdentifierType stype2 = SourceIdentifierType.getSourceIdentifierType(tokens[6]);
    if (stype2 == null)
      throw new IOException(tokens[6] + " SourceIdentifierType not in MRDOC");
    r.setRelatedSourceIdentifierType(stype2);

    final RelationshipLabel rel = RelationshipLabel.getRelationshipLabel(tokens[3]);
    if (rel == null)
      throw new IOException(tokens[3] + " RelationshipLabel not in MRDOC");
    r.setRel(rel);
    final Concept related_concept = new PlaceholderConcept();
    related_concept.setCui(tokens[4]);
    r.setRelatedConcept(related_concept);
    final AdditionalRelationshipLabel rela = AdditionalRelationshipLabel.getAdditionalRelationshipLabel(tokens[7]);
    if (rela == null)
      throw new IOException(tokens[7] + " AdditionalRelationshipLabel not in MRDOC");
    r.setRela(rela);
    r.setRui(tokens[8]);
    r.setSrui(tokens[9]);
    final Source sab = Source.getSource(tokens[10]);
    if (sab == null)
      throw new IOException(tokens[10] + " Source not in MRSAB");
    r.setSource(sab);
    final Source sl = Source.getSource(tokens[11]);
    if (sl == null)
      throw new IOException(tokens[11] + " Source not in MRSAB");
    r.setSl(sl);
    r.setRg(tokens[12]);
    r.setDir(tokens[13]);
    final Suppressible sup = Suppressible.getSuppressible(tokens[14]);
    if (sup == null)
      throw new IOException(tokens[14] + " Suppressible not in MRDOC");
    r.setSuppress(sup);
    r.setCvf(tokens[15].equals("") ? null : new ContentView(tokens[15]));

    return r;
  }

 /**
   * Gets {@link Relationship}s from the <code>MRREL.RRF</code> file with the
   * given cui and adds them to the current {@link Concept}.
   * @param this_concept the current {@link Concept}
   * @param current_cui the current CUI
   * @throws IOException if problem reading from <code>MRREL.RRF</code>
   */
  protected void getRelationships(Concept this_concept, String current_cui)
    throws IOException {

    //
    // Iterate through MRREL
    //
    rel_hm = new HashMap();
    String line = "";
    while((line = mrrel_source.readLine()) != null) {

      //
      // Skip lines that do not match
      //
      if (line.compareTo(current_cui) < 0)
        continue;

      //
      // Matching line found, connect relationship and concept
      //
      else if (line.startsWith(current_cui)) {
        final Relationship rel = newInstanceFromMRREL(line);
        //
        // Cache relationship by RUI (for connecting attributes lateR)
        //
        rel_hm.put(rel.getRui(), rel);
        rel.setConcept(this_concept);
        this_concept.addRelationship(rel);
      }

      //
      // went too far - break and reset
      //
      else {
        break;
      }
    }

    mrrel_source.push(line);
  }

  /**
   * Instantiates an {@link ComponentHistory} from a <code>MRHIST.RRF</code> line.
   * @param line a line from <code>MRHIST.RRF</code>
   * @return the {@link ComponentHistory} matching the line
   * @throws IOException if there is a problem with the file
   */
  private  ComponentHistory newInstanceFromMRHIST(String line)
    throws IOException {

    //
    // Instantiate default ComponentHistory
    //
    final ComponentHistory ch = new ComponentHistoryImpl();

    //
    // Parse line and count tokens
    //
    FieldedStringTokenizer.split(line, "|", 10, tokens);

    //
    // Populate ComponenHistory
    //
    ch.setUi(tokens[1]);
    final Source sab = Source.getSource(tokens[2]);
    if (sab == null)
      throw new IOException(tokens[2] + " Source not in MRSAB");
    ch.setSource(sab);
    ch.setSver(tokens[3]);
    ch.setChangeType(tokens[4]);
    ch.setChangeKey(tokens[5]);
    ch.setChangeValue(tokens[6]);
    ch.setReason(tokens[7]);
    ch.setCvf(tokens[8].equals("") ? null : new ContentView(tokens[8]));

    return ch;
  }

 /**
   * Gets {@link ComponentHistory}s from the <code>MRHIST.RRF</code> file with the
   * given cui and adds them to the current {@link Concept}.
   * @param this_concept the current {@link Concept}
   * @param current_cui the current CUI
   * @throws IOException if problem reading from <code>MRHIST.RRF</code>
   */
  protected void getComponentHistories(Concept this_concept, String current_cui)
    throws IOException {

    //
    // Iterate through MRHIST.RRF
    //
    String line = "";
    while((line = mrhist_source.readLine()) != null) {
      //
      // Skip lines that do not match
      //
      if (line.compareTo(current_cui) < 0)
        continue;

      //
      // Matching line found, connect component history and concept
      //
      else if (line.startsWith(current_cui)) {
        final ComponentHistory ch = newInstanceFromMRHIST(line);
        ch.setConcept(this_concept);
        this_concept.addComponentHistory(ch);
      }

      //
      // went too far - break and reset
      //
      else {
        break;
      }
    }

    mrhist_source.push(line);
  }

  /**
   * Instantiates an {@link Mapping} from a <code>MRMAP.RRF</code> line.
   * @param line a line from <code>MRMAP.RRF</code>
   * @return the {@link Mapping} matching the line
   * @throws IOException if there is a problem with the file
   */
  private  Mapping newInstanceFromMRMAP(String line)
    throws IOException {

    //
    // Instantiate default mapping (and map objects)
    //
    final Mapping map = new MappingImpl();
    final MapObject from = new MapObjectImpl();
    final MapObject to = new MapObjectImpl();

    //
    // Parse line and count tokens
    //
    FieldedStringTokenizer.split(line, "|", 27, tokens);

    //
    // Populate mapping
    //
    final Source sab = Source.getSource(tokens[1]);
    if (sab == null)
      throw new IOException(tokens[1] + " Source not in MRSAB");
    map.setSource(sab);
    map.setSubsetId(tokens[2]);
    map.setRank(tokens[3]);
    map.setId(tokens[4]);
    map.setSid(tokens[5]);
    map.setFrom(from);
    from.setIdentifier(tokens[6]);
    from.setSourceIdentifier(tokens[7]);
    from.setExpression(tokens[8]);
    final SourceIdentifierType fromtype = SourceIdentifierType.getSourceIdentifierType(tokens[9]);
    if (fromtype == null)
      throw new IOException(tokens[9] + " SourceIdentifierType not in MRDOC");
    from.setType(fromtype);
    from.setRule(tokens[10]);
    from.setRestriction(tokens[11]);
    final Source fromsab = Source.getSource(from_vsab);
    if (from_vsab == null) {
      String source_line = "||" + from_vsab + "|" + from_rsab + "||||||||||||||||||||||";
      SourceMetaDescription[] smd_array = new SourceMetaDescription[1];
      smd_array[0] = new SourceMetaDescriptionImpl(source_line);
      Source.initialize(smd_array);
      from.setSource(Source.getSource(from_vsab));
    } else
      from.setSource(fromsab);
    final RelationshipLabel rel = RelationshipLabel.getRelationshipLabel(tokens[12]);
    if (rel == null)
      throw new IOException(tokens[12] + " RelationshipLabel not in MRDOC");
    map.setRelationshipName(rel);
    final AdditionalRelationshipLabel rela = AdditionalRelationshipLabel.getAdditionalRelationshipLabel(tokens[13]);
    if (rela == null)
      throw new IOException(tokens[13] +
                            " AdditionalRelationshipLabel not in MRDOC");
    map.setAdditionalRelationshipLabel(rela);
    map.setTo(to);
    to.setIdentifier(tokens[14]);
    to.setSourceIdentifier(tokens[15]);
    to.setExpression(tokens[16]);
    final SourceIdentifierType totype = SourceIdentifierType.getSourceIdentifierType(tokens[17]);
    if (totype == null)
      throw new IOException(tokens[15] + " SourceIdentifierType not in MRDOC");
    to.setType(totype);
    to.setRule(tokens[18]);
    to.setRestriction(tokens[19]);
    final Source tosab = Source.getSource(to_vsab);
    if (Source.getSource(to_vsab) == null) {
      String source_line = "||" + to_vsab + "|" + to_rsab + "||||||||||||||||||||||";
      SourceMetaDescription[] smd_array = new SourceMetaDescription[1];
      smd_array[0] = new SourceMetaDescriptionImpl(source_line);
      Source.initialize(smd_array);
      to.setSource(Source.getSource(to_vsab));
    } else
      to.setSource(tosab);
    map.setRule(tokens[20]);
    map.setRes(tokens[21]);
    map.setType(tokens[22]);
    if (tokens[23].length() != 0) {
      final AttributeName atn = AttributeName.getAttributeName(tokens[23]);
      if (atn == null)
        throw new IOException(tokens[23] + " AttributeName not in MRDOC");
      map.setAttributeName(atn);
    }
    map.setAttributeValue(tokens[24]);
    map.setCvf(tokens[25].equals("") ? null : new ContentView(tokens[25]));

    return map;
  }

  /**
   * Gets {@link Mapping}s from the <code>MRMAP.RRF</code> file with the
   * given cui and adds them to the current {@link Concept}.
   * @param this_concept the current {@link Concept}
   * @param current_cui the current CUI
   * @throws IOException if problem reading from <code>MRMAP.RRF</code>
   */
  protected void getMappings(Concept this_concept, String current_cui) throws
      IOException {

    //
    // Iterate through MRMAP.RRF
    //
    String line = "";
    while((line = mrmap_source.readLine()) != null)  {

      //
      // Skip if line does not match
      //
      if (line.compareTo(current_cui) < 0)
        continue;

      //
      // Matching line found, connect mapping and concept
      //
      else if (line.startsWith(current_cui)) {
        final Mapping map = newInstanceFromMRMAP(line);
        map.setConcept(this_concept);
        this_concept.addMapping(map);
      }

      //
      // went too far - break and reset
      //
      else {
        break;
      }
    }

    mrmap_source.push(line);
  }


  //
  // Protected Utility Method
  //

  /**
   * Returns {@link Atom} with the given RUI.
   * @param aui identifier
   * @return {@link Atom}
   */
  private Atom getAtom(String aui) {
    Atom atom = (Atom)atom_hm.get(aui);
    if (atom == null) {
      atom = new PlaceholderAtom();
      atom.setAui(aui);
    }
    return atom;
  }

  /**
   * Returns {@link Relationship} with the given rui.
   * @param rui identifier
   * @return {@link Relationship}
   */
  private Relationship getRelationship(String rui) {
    Relationship relationship = (Relationship)rel_hm.get(rui);
    if (relationship == null) {
      relationship = new RelationshipImpl();
      relationship.setRui(rui);
    }
    return relationship;
  }

  /**
   * Sets an additional suffix.
   * @param suffix to be added to the .RRF suffix
   */
  public void setAdditionalSuffix(String suffix) {
    add_suffix = suffix;
  }
  
  /**
   * Returns a {@link String}[] of warnings accumulated during
   * processing of the input stream.
   * @return {@link String}[] of warnings
   */
  public String[] getWarnings() {
  	return (String[])warnings.toArray(new String[0]);
  }
}
