/**************************************************************************
 *
 * Package:     gov.nih.nlm.umls.io
 * Object:      RichMRFilesConceptOutputStream.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.io;

import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.AtomAttribute;
import gov.nih.nlm.umls.meta.AtomCoocurrence;
import gov.nih.nlm.umls.meta.AtomRelationship;
import gov.nih.nlm.umls.meta.Attribute;
import gov.nih.nlm.umls.meta.ComponentHistory;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.Context;
import gov.nih.nlm.umls.meta.ContextMember;
import gov.nih.nlm.umls.meta.Coocurrence;
import gov.nih.nlm.umls.meta.Definition;
import gov.nih.nlm.umls.meta.Language;
import gov.nih.nlm.umls.meta.Mapping;
import gov.nih.nlm.umls.meta.Relationship;
import gov.nih.nlm.umls.meta.RelationshipAttribute;
import gov.nih.nlm.umls.meta.SemanticType;
import gov.nih.nlm.util.SystemToolkit;
import gov.nih.nlm.util.UTF8OutputStreamWriter;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

/**
 * This {@link ConceptOutputStream} generates a subset of
 * "Rich Release Format" MR files.
 *
 * @author Deborah Shapiro
 */
public class RichMRFilesConceptOutputStream implements ConceptOutputStream {

  //
  // Protected Fields
  //

  //
  // Map of file names -> BufferedWriter
  //
  protected Map mr_targets = new HashMap();

  //
  // The target list (sans index files)
  //


  protected static String[] target_files =
    new String[] { "MRCOC", "MRCOLS", "MRCONSO", "MRCUI", "MRCXT", "MRDEF",
		     "MRDOC", "MRFILES", "MRHIER", "MRHIST", "MRMAP",
		     "MRRANK", "MRREL", "MRSAB",  "MRSAT", "MRSMAP",
                     "MRSTY", "MRXNS_ENG", "MRXNW_ENG", "MRAUI" };

  protected static String file_ext = ".RRF";

  protected Map target_to_output_stream = new HashMap();

  //
  // Directory where to write files
  //
  protected String mr_dir;

  //
  // Indicates whether to use RSAB or VSAB when writing SAB fields
  //
  protected boolean versioned_output = false;

  //
  // Indicates whether or not to prepend Byte Order Mark (BOM) to indicate
  // UTF-8 encoding.
  //
  protected boolean add_unicode_bom = false;
  
  //
  // List of warnings
  //
  protected List warnings = new ArrayList();


  //
  // Constructor
  //

  /**
   * Instantiates an empty {@link RichMRFilesConceptOutputStream}.
   */
  public RichMRFilesConceptOutputStream() {}

  //
  // Methods
  //

  /**
   * Opens all Rich Release Format MR files in the specified directory.
   * @param mr_dir the subset directory
   * @throws IOException if the files cannot be opened
   */
  public void open(String mr_dir) throws IOException {

    //
    // Save directory name
    //
    this.mr_dir = mr_dir;

    //
    // Open files and initialize statistics
    //
    for (int i = 0; i < target_files.length; i++) {
      mr_targets.put(target_files[i],
		     openTargetFile(target_files[i] + file_ext));
    }

    //
    // Open index files and initialize stats
    //
    final Language[] languages = Language.getLanguages();
    for (int i = 0; i < languages.length; i++) {
      final String base_file = "MRXW_" + languages[i].getAbbreviation();
      final String file = base_file + file_ext;
      mr_targets.put(base_file, openTargetFile(file));
    }

  }

  public void flush(String file_name) throws IOException {
    if (mr_targets.keySet().contains(file_name))
      ((BufferedWriter)mr_targets.get(file_name)).flush();
  }

  /**
   * Closes source files that were opened by @link #open(String)}.
   * @throws IOException if the files could not be cloesed
   */
  public void close() throws IOException {

    //
    // Close all files
    //
    final Iterator iter = mr_targets.keySet().iterator();
    while (iter.hasNext()) {
      ( (BufferedWriter) mr_targets.get(iter.next())).close();
    }

    //
    // Sort MRMAP
    //
    SystemToolkit.sort(new File(mr_dir, "MRMAP" + file_ext).getAbsolutePath(), true);

    //
    // Sort MRSMAP
    //
    SystemToolkit.sort(new File(mr_dir, "MRSMAP" + file_ext).getAbsolutePath(), true);

  }

  /**
   * Prints the given line to the file with the given name.
   * @param line the line to be printed
   * @param file_name printing destination
   * @throws IOException
   */
  protected void printLine(String line, String file_name) throws IOException {
    BufferedWriter this_writer = ((BufferedWriter)mr_targets.get(file_name));
    this_writer.write(line);
    this_writer.newLine();
    //this_writer.flush();
  }

  /**
   * Writes a {@link Concept} to the appropriate MR files.
   * @param concept the {@link Concept} to write
   * @throws IOException if the files could not be written
   */
  public void writeConcept(Concept concept) throws IOException {

    //
    // Write MRCONSO
    //
    writeMRCONSO(concept);

    //
    // Write MRDEF
    //
    writeMRDEF(concept);

    //
    // Write MRSTY
    //
    writeMRSTY(concept);

    //
    // Write MRSAT
    //
    writeMRSAT(concept);

    //
    // Write MRREL
    //
    writeMRREL(concept);

    //
    // Write MRCOC
    //
    writeMRCOC(concept);

    //
    // Write MRCXT
    //
    writeMRCXT(concept);

    //
    // Write MRHIER
    //
    writeMRHIER(concept);

    //
    // Write MRMAP
    //
    writeMRMAP(concept);

    //
    // Write MRSMAP
    //
    writeMRSMAP(concept);

    //
    // Write MRHIST
    //
    writeMRHIST(concept);

  }

  /**
   * Writes <code>MRCOC.RRF</code> entries for a {@link Concept}
   * @param concept the {@link Concept} to write
   * @throws IOException if anything goes wrong
   */
  protected void writeMRCOC(Concept concept)
    throws IOException {

    //
    // Iterate through coocurrences
    //
    final Coocurrence[] coocurrences = concept.getCoocurrences();
    final List mrcoc_lines = new ArrayList(coocurrences.length);
    for(int i = 0; i < coocurrences.length; i++) {

      final Coocurrence coc = coocurrences[i];

      //
      // Determine if Coocurrence or AtomCoocurrence
      //
      final AtomCoocurrence atom_coc =
        (coc instanceof AtomCoocurrence) ? (AtomCoocurrence)coc : null;

      //
      // Prepare MRCOC.RRF line
      //
      final StringBuffer sb = new StringBuffer(100);
      sb.append(coc.getConcept().getCui()).append("|");
      sb.append((atom_coc != null) ? atom_coc.getAtom().getAui() : "").append("|");
      sb.append(coc.getRelatedConcept().getCui()).append("|");
      sb.append((atom_coc != null) ? atom_coc.getRelatedAtom().getAui() : "").append("|");
      sb.append(versioned_output ?
                  coc.getSource().getVersionedSourceAbbreviation() :
                  coc.getSource().getRootSourceAbbreviation()).append("|");
      sb.append(coc.getCot() != null ? coc.getCot().getAbbreviation() : "").append("|");
      sb.append(coc.getCof() !=0 ? String.valueOf(coc.getCof()) : "").append("|");
      sb.append(coc.getCoa()).append("|");
      sb.append((coc.getCvf() != null) ? coc.getCvf().toString() : "").append("|");

      //
      // Add line to list
      //
      mrcoc_lines.add(sb.toString());
    }

    //
    // Sort lines and write to MRCOC.RRF
    //
    Collections.sort(mrcoc_lines);
    for(int i = 0; i < coocurrences.length; i++) {
      printLine(mrcoc_lines.get(i).toString(), "MRCOC");
    }
  }

  /**
   * Writes <code>MRCONSO.RRF</code> entries for a {@link Concept}
   * @param concept the {@link Concept} to write
   * @throws IOException if anything goes wrong
   */
  protected void writeMRCONSO(Concept concept)
    throws IOException {

    //
    // Iterate through atoms
    //
    final Atom[] atoms = concept.getAtoms();
    final List mrconso_lines = new ArrayList(atoms.length);
    for(int i = 0; i < atoms.length; i++) {

      //
      // Prepare MRCONSO.RRF line
      //
      final Atom atom = atoms[i];
      final StringBuffer sb = new StringBuffer(150);
      sb.append(atom.getConcept().getCui()).append("|");
      sb.append(atom.getLat().getAbbreviation()).append("|");
      sb.append(atom.getTs().getAbbreviation()).append("|");
      sb.append(atom.getLui()).append("|");
      sb.append(atom.getStt().getAbbreviation()).append("|");
      sb.append(atom.getSui()).append("|");
      sb.append((atom.hasPreferredAtomIdentifier()) ? "Y" : "N").append("|");
      sb.append((atom.getAui() != null) ? atom.getAui() : "").append("|");
      sb.append((atom.getSaui() != null) ? atom.getSaui() : "").append("|");
      sb.append((atom.getScui() != null) ? atom.getScui() : "").append("|");
      sb.append((atom.getSdui() != null) ? atom.getSdui() : "").append("|");
      if (versioned_output)
        sb.append(atom.getSource().getVersionedSourceAbbreviation());
      else
        sb.append(atom.getSource().getRootSourceAbbreviation());
      sb.append("|");
      sb.append(atom.getTty().getAbbreviation()).append("|");
      sb.append(atom.getCode()).append("|");
      sb.append(atom.getStr()).append("|");
      sb.append(atom.getSource().getSrl()).append("|");
      sb.append((atom.getSuppress() != null) ?
                atom.getSuppress().getAbbreviation() : "").append("|");
      sb.append((atom.getCvf() != null) ?
                atom.getCvf().toString() : "").append("|");

      //
      // Add line to list
      //
      mrconso_lines.add(sb.toString());
    }

    //
    // Sort (unique) lines and write to MRCONSO.RRF
    //
    Collections.sort(mrconso_lines);
    String prev_line= "";
    for (int i = 0; i < mrconso_lines.size(); i++) {
      final String this_line = mrconso_lines.get(i).toString();
      if (!this_line.equals(prev_line))
        printLine(this_line, "MRCONSO");
      prev_line = this_line;
    }
  }

  /**
   * Writes <code>MRCXT.RRF</code> entries for a {@link Concept}
   * @param concept the {@link Concept} to write
   * @throws IOException if anything goes wrong
   */
  protected void writeMRCXT(Concept concept)
    throws IOException {

    //
    // Iterate through contexts
    //
    final Context[] contexts = concept.getContexts();
    final List mrcxt_lines = new ArrayList(contexts.length*4);
    for(int i = 0; i < contexts.length; i++) {

      //
      // Iterate through context members
      //
      final Context cxt = contexts[i];
      final ContextMember[] context_rels = cxt.getContextMembers();
      for(int j=0; j<context_rels.length; j++) {

        final ContextMember cxt_rel = context_rels[j];

        //
        // Prepare MRCXT.RRF line
        //
        final StringBuffer sb = new StringBuffer(110);
	sb.append(cxt.getConcept().getCui()).append("|");
	sb.append(cxt.getAtom().getSui()).append("|");
        sb.append((cxt.getAtom() != null) ? cxt.getAtom().getAui() : "").append("|");
        if (versioned_output)
          sb.append(cxt.getSource().getVersionedSourceAbbreviation());
        else
          sb.append(cxt.getSource().getRootSourceAbbreviation());
        sb.append("|");
	sb.append(cxt.getAtom().getCode()).append("|");
	sb.append(cxt.getCxn()).append("|");
	sb.append(cxt_rel.getCxl()).append("|");
        if (cxt_rel.getCxl().equals("ANC"))
	  sb.append(cxt_rel.getRank());
        sb.append("|");
	sb.append(cxt_rel.getRelatedAtom().getStr()).append("|");
	sb.append(cxt_rel.getRelatedConcept().getCui()).append("|");
        sb.append((cxt_rel.getRelatedAtom() != null) ? cxt_rel.getRelatedAtom().getAui() : "").append("|");
	sb.append(cxt_rel.getHcd()).append("|");
	sb.append(cxt_rel.getRela() != null ?
          cxt_rel.getRela().getAbbreviation() : "").append("|");
	sb.append(cxt_rel.getXc()).append("|");
        sb.append((cxt.getCvf() != null) ?
                  cxt.getCvf().toString() : "").append("|");

        //
        // Add line to list
        //
        mrcxt_lines.add(sb.toString());
      }
    }

    //
    // Sort lines and write to MRCXT.RRF
    //
    Collections.sort(mrcxt_lines);
    for(int i = 0; i < mrcxt_lines.size(); i++) {
      printLine(mrcxt_lines.get(i).toString(), "MRCXT");
    }
  }

  /**
   * Writes <code>MRHIER</code> entries for a {@link Concept}
   * @param concept the {@link Concept} to write
   * @throws IOException if anything goes wrong
   */
  protected void writeMRHIER(Concept concept)
    throws IOException {

    //
    // Iterate through contexts
    //
    final Context[] contexts = concept.getContexts();
    final List mrhier_lines = new ArrayList(contexts.length);
    for(int i = 0; i < contexts.length; i++) {

      final Context cxt = contexts[i];

      //
      // Prepare MRHIER.RRF line
      //
      final StringBuffer sb = new StringBuffer(150);
      sb.append(cxt.getConcept().getCui()).append("|");
      sb.append((cxt.getAtom() != null) ? cxt.getAtom().getAui() : "").append("|");
      sb.append(cxt.getCxn()).append("|");
      sb.append((cxt.getPaui() != null) ? cxt.getPaui() : "").append("|");
      if (versioned_output)
        sb.append(cxt.getSource().getVersionedSourceAbbreviation());
      else
        sb.append(cxt.getSource().getRootSourceAbbreviation());
      sb.append("|");
      sb.append(cxt.getRela() != null ?
          cxt.getRela().getAbbreviation() : "").append("|");
      sb.append((cxt.getPtr() != null) ? cxt.getPtr() : "").append("|");
      sb.append((cxt.getHcd() != null) ? cxt.getHcd() : "").append("|");
      sb.append((cxt.getCvf() != null) ?
                cxt.getCvf().toString() : "").append("|");

      //
      // Add line to list
      //
      mrhier_lines.add(sb.toString());
    }

    //
    // Sort lines and write MRHIER.RRF
    //
    Collections.sort(mrhier_lines);
    for (int i = 0; i < contexts.length; i++) {
      printLine(mrhier_lines.get(i).toString(), "MRHIER");
    }
  }

  /**
   * Writes <code>MRDEF.RRF</code> entries for a {@link Concept}
   * @param concept the {@link Concept} to write
   * @throws IOException if anything goes wrong
   */
  protected void writeMRDEF(Concept concept)
    throws IOException {

    //
    // Iterate through definitions
    //
    final Definition[] definitions = concept.getDefinitions();
    final List mrdef_lines = new ArrayList(definitions.length);
    for(int i = 0; i < definitions.length; i++) {

      final Definition def = definitions[i];

      //
      // Prepare MRDEF.RRF line
      //
      final StringBuffer sb = new StringBuffer(1000);
      sb.append(def.getConcept().getCui()).append("|");
      sb.append((def.getAtom() != null) ? def.getAtom().getAui() : "").append("|");
      sb.append((def.getAtui() != null) ? def.getAtui() : "").append("|");
      sb.append((def.getSatui() != null) ? def.getSatui() : "").append("|");
      if (versioned_output)
        sb.append(def.getSource().getVersionedSourceAbbreviation());
      else
        sb.append(def.getSource().getRootSourceAbbreviation());
      sb.append("|");
      sb.append(def.getDef()).append("|");
      sb.append((def.getSuppress() != null) ?
                def.getSuppress().getAbbreviation() : "").append("|");
      sb.append((def.getCvf() != null) ?
                def.getCvf().toString() : "").append("|");

      //
      // Add line to list
      //
      mrdef_lines.add(sb.toString());
    }

    //
    // Sort lines and write to MRDEF.RRF
    //
    Collections.sort(mrdef_lines);
    for (int i = 0; i < definitions.length; i++) {
      printLine(mrdef_lines.get(i).toString(), "MRDEF");
    }
  }

  /**
   * Writes <code>MRSTY.RRF</code> entries for a {@link Concept}
   * @param concept the {@link Concept} to write
   * @throws IOException if anything goes wrong
   */
  protected void writeMRSTY(Concept concept)
    throws IOException {

    //
    // Iterate through semantic types
    //
    final SemanticType[] semantic_types = concept.getSemanticTypes();
    final List mrsty_lines = new ArrayList(semantic_types.length);
    for (int i = 0; i < semantic_types.length; i++) {

      final SemanticType sty = semantic_types[i];

      //
      // Prepare MRSTY.RRF line
      //
      final StringBuffer sb = new StringBuffer(50);
      sb.append(sty.getConcept().getCui()).append("|");
      sb.append(sty.getSty() == null ?
        "" : sty.getSty().getUi()).append("|");
      sb.append(sty.getSty() == null ?
        "" : sty.getSty().getStn()).append("|");
      sb.append(sty.getSty() == null ?
        "" : sty.getSty().getSty()).append("|");
      sb.append(sty.getAtui() == null ?
        "" : sty.getAtui()).append("|");
      sb.append((sty.getCvf() != null) ?
                sty.getCvf().toString() : "").append("|");

      //
      // Add line to list
      //
      mrsty_lines.add(sb.toString());
    }

    //
    // Sort lines and write to MRSTY.RRF
    //
    Collections.sort(mrsty_lines);
    for(int i = 0; i < semantic_types.length; i++) {
      printLine(mrsty_lines.get(i).toString(), "MRSTY");
    }
  }

  /**
   * Writes <code>MRREL.RRF</code> entries for a {@link Concept}
   * @param concept the {@link Concept} to write
   * @throws IOException if anything goes wrong
   */
  protected void writeMRREL(Concept concept)
    throws IOException {

    //
    // Iterate through relationships
    //
    final Relationship[] relationships = concept.getRelationships();
    final List mrrel_lines = new ArrayList(relationships.length);
    for(int i = 0; i < relationships.length; i++) {

      final Relationship rel = relationships[i];

      //
      // Determine if AtomRelationship or Relationship
      //
      final StringBuffer sb = new StringBuffer(100);

      sb.append(relationships[i].getConcept().getCui()).append("|");

      final AtomRelationship ar = (rel instanceof AtomRelationship) ? (AtomRelationship)rel : null;

      if (ar != null) {
        sb.append((ar.getAtom() != null) ? ar.getAtom().getAui() : "").append("|");
        sb.append((ar.getAtom() != null) ?
                   ar.getSourceIdentifierType().getAbbreviation() : "");
      } else {
        sb.append("|CUI");
      }

      //
      // Continue preparing MRREL.RRF line
      //
      sb.append("|");
      sb.append(rel.getRel() != null ?
        rel.getRel().getAbbreviation() : "").append("|");
      sb.append(rel.getRelatedConcept().getCui()).append("|");
      if (ar != null) {
        sb.append((ar.getRelatedAtom() != null) ? ar.getRelatedAtom().getAui() : "").append("|");
        sb.append((ar.getRelatedAtom() != null) ?
                   ar.getRelatedSourceIdentifierType().getAbbreviation() : "");
      } else {
        sb.append("|CUI");
      }
      sb.append("|");
      if (rel.getRela() != null)
        sb.append(rel.getRela().getAbbreviation());
      sb.append("|");
      sb.append(rel.getRui() == null ? "" : rel.getRui()).append("|");
      sb.append(rel.getSrui() == null ? "" : rel.getSrui()).append("|");
      if (versioned_output)
        sb.append(rel.getSource().getVersionedSourceAbbreviation());
      else
        sb.append(rel.getSource().getRootSourceAbbreviation());
      sb.append("|");
      if (versioned_output)
        sb.append(rel.getSl().getVersionedSourceAbbreviation());
      else
        sb.append(rel.getSl().getRootSourceAbbreviation());
      sb.append("|");
      sb.append(rel.getRg() == null ? "" : rel.getRg()).append("|");
      sb.append(rel.getDir() == null ? "" : rel.getDir()).append("|");
      sb.append(rel.getSuppress() == null ?
                "" : rel.getSuppress().getAbbreviation()).append("|");
      sb.append((rel.getCvf() != null) ?
                rel.getCvf().toString() : "").append("|");

      //
      // Add line to list
      //
      mrrel_lines.add(sb.toString());
    }

    //
    // Sort lines and write to MRREL.RRF
    //
    Collections.sort(mrrel_lines);
    for(int i = 0; i < relationships.length; i++) {
      printLine(mrrel_lines.get(i).toString(), "MRREL");
    }
  }

  /**
   * Writes <code>MRSAT.RRF</code> entries for a {@link Concept}
   * @param concept the {@link Concept} to write
   * @throws IOException if anything goes wrong
   */
  protected void writeMRSAT(Concept concept)
    throws IOException {

    //
    // Iterate through attributes
    //
    final Attribute[] attributes = concept.getAttributes();

    final List mrsat_lines = new ArrayList(attributes.length);
    for(int i = 0; i < attributes.length; i++) {

      final Attribute att = attributes[i];

      //
      // Prepare MRSAT.RRF line
      //
      final StringBuffer sb = new StringBuffer(100);
      sb.append(att.getConcept().getCui()).append("|");

      //
      // Determine if AtomAttribute, RelationshipAttribute, or Attribute
      //
      if (att instanceof AtomAttribute) {
        final AtomAttribute aa = (AtomAttribute)att;
        sb.append(aa.getAtom().getLui()).append("|");
        sb.append(aa.getAtom().getSui()).append("|");
        sb.append(aa.getAtom().getAui()).append("|");
        sb.append(aa.getSourceIdentifierType().getAbbreviation()).append("|");
        sb.append(aa.getAtom().getCode()).append("|");
      } else if (att instanceof RelationshipAttribute) {
        sb.append("||");
        final RelationshipAttribute ra = (RelationshipAttribute)att;
        sb.append(ra.getRelationship() != null ? ra.getRelationship().getRui() : "").append("|");
        sb.append(ra.getSourceIdentifierType().getAbbreviation()).append("||");
      } else {
        sb.append("|||CUI||");
      }
      sb.append((att.getAtui() != null) ? att.getAtui() : "").append("|");
      sb.append((att.getSatui() != null) ? att.getSatui() : "").append("|");
      sb.append((att.getAtn() != null) ? att.getAtn().getAbbreviation() : "").append("|");
      if (versioned_output)
        sb.append(att.getSource().getVersionedSourceAbbreviation());
      else
        sb.append(att.getSource().getRootSourceAbbreviation());
      sb.append("|");
      sb.append(att.getAtv()).append("|");
      sb.append((att.getSuppress() != null) ?
                att.getSuppress().getAbbreviation() : "").append("|");
      sb.append((att.getCvf() != null) ?
                att.getCvf().toString() : "").append("|");

      //
      // Add line to list
      //
      mrsat_lines.add(sb.toString());
    }

    //
    // Sort lines and write to MRSAT.RRF
    //
    Collections.sort(mrsat_lines);
    for (int i = 0; i < attributes.length; i++) {
      printLine(mrsat_lines.get(i).toString(), "MRSAT");
    }
  }

  /**
   * Writes <code>MRMAP.RRF</code> entries for a {@link Concept}
   * @param concept the {@link Concept} to write
   * @throws IOException if anything goes wrong
   */
  protected void writeMRMAP(Concept concept)
    throws IOException {

    //
    // Iterate through mappings
    //
    final Mapping[] mappings = concept.getMappings();
    for (int i=0; i<mappings.length; i++) {

      //
      // Prepare MRMAP.RRF line
      //
      final StringBuffer sb = new StringBuffer(100);
      final Mapping mapping = mappings[i];

      sb.append(mapping.getConcept().getCui()).append("|");
      if (versioned_output)
        sb.append(mapping.getSource().getVersionedSourceAbbreviation());
      else
        sb.append(mapping.getSource().getRootSourceAbbreviation());
      sb.append("|");
      sb.append((mapping.getSubsetId() != null) ? mapping.getSubsetId() : "").append("|");
      sb.append((mapping.getRank() != null) ? mapping.getRank() : "").append("|");
      sb.append((mapping.getId() != null) ? mapping.getId() : "").append("|");
      sb.append((mapping.getSid() != null) ? mapping.getSid() : "").append("|");
      sb.append((mapping.getFrom() != null) ? mapping.getFrom().getId() : "").append("|");
      sb.append((mapping.getFrom() != null) ? mapping.getFrom().getSid() : "").append("|");
      sb.append((mapping.getFrom() != null) ? mapping.getFrom().getExpr() : "").append("|");
      sb.append((mapping.getFrom() != null && mapping.getFrom().getType() != null) ?
        mapping.getFrom().getType().getAbbreviation() : "").append("|");
      sb.append((mapping.getFrom() != null) ? mapping.getFrom().getRule() : "").append("|");
      sb.append((mapping.getFrom() != null) ? mapping.getFrom().getRes() : "").append("|");
      sb.append((mapping.getRel() != null ) ? mapping.getRel().getAbbreviation() : "").append("|");
      sb.append((mapping.getRela() != null) ? mapping.getRela().getAbbreviation() : "").append("|");
      sb.append((mapping.getTo() != null) ? mapping.getTo().getId() : "").append("|");
      sb.append((mapping.getTo() != null) ? mapping.getTo().getSid() : "").append("|");
      sb.append((mapping.getTo() != null) ? mapping.getTo().getExpr() : "").append("|");
      sb.append((mapping.getTo() != null && mapping.getTo().getType() != null) ?
        mapping.getTo().getType().getAbbreviation() : "").append("|");
      sb.append((mapping.getTo() != null) ? mapping.getTo().getRule() : "").append("|");
      sb.append((mapping.getTo() != null) ? mapping.getTo().getRes() : "").append("|");
      sb.append((mapping.getRule() != null) ? mapping.getRule() : "").append("|");
      sb.append((mapping.getRes() != null) ? mapping.getRes() : "").append("|");
      sb.append((mapping.getType() != null) ? mapping.getType() : "").append("|");
      sb.append((mapping.getAttributeName() != null) ?
        mapping.getAttributeName().getAbbreviation() : "").append("|");
      sb.append((mapping.getAttributeValue() != null) ?
        mapping.getAttributeValue() : "").append("|");
      sb.append((mapping.getCvf() != null) ?
        mapping.getCvf().toString() : "").append("|");

      //
      // Print MRMAP.RRF line, sort later
      //
      printLine(sb.toString(), "MRMAP");
    }

  }

  /**
   * Writes <code>MRSMAP.RRF</code> entries for a {@link Concept}
   * @param concept the {@link Concept} to write
   * @throws IOException if anything goes wrong
   */
  protected void writeMRSMAP(Concept concept)
    throws IOException {

    //
    // Iterate through mappings
    //
    final Mapping[] mappings = concept.getMappings();
    for (int i=0; i<mappings.length; i++) {

      final Mapping mapping = mappings[i];

      //
      // Continue if not a simple mapping
      //
      if (mapping == null )
        continue;
      if (!((mapping.getSubsetId().equals("") || mapping.getSubsetId().equals("0")) &&
            (mapping.getRank().equals("") || mapping.getRank().equals("0"))))
        continue;

      //
      // Prepare MRSMAP.RRF line
      //
      final StringBuffer sb = new StringBuffer(80);
      sb.append(mapping.getConcept().getCui()).append("|");
      if (versioned_output)
        sb.append(mapping.getSource().getVersionedSourceAbbreviation());
      else
        sb.append(mapping.getSource().getRootSourceAbbreviation());
      sb.append("|");
      sb.append((mapping.getId() != null) ? mapping.getId() : "").append("|");
      sb.append((mapping.getSid() != null) ? mapping.getSid() : "").append("|");
      sb.append((mapping.getFrom() != null) ? mapping.getFrom().getExpr() : "").append("|");
      sb.append((mapping.getFrom() != null && mapping.getFrom().getType() != null) ?
        mapping.getFrom().getType().getAbbreviation() : "").append("|");
      sb.append((mapping.getRel() != null ) ? mapping.getRel().getAbbreviation() : "").append("|");
      sb.append((mapping.getRela() != null) ? mapping.getRela().getAbbreviation() : "").append("|");
      sb.append((mapping.getTo() != null) ? mapping.getTo().getExpr() : "").append("|");
      sb.append((mapping.getTo() != null && mapping.getTo().getType() != null) ?
        mapping.getTo().getType().getAbbreviation() : "").append("|");
      sb.append((mapping.getCvf() != null) ?
        mapping.getCvf().toString() : "").append("|");

      //
      // Print MRSMAP.RRF line (sort later)
      //
      printLine(sb.toString(), "MRSMAP");
    }

  }

  /**
   * Writes <code>MRHIST.RRF</code> entries for a {@link Concept}
   * @param concept the {@link Concept} to write
   * @throws IOException if anything goes wrong
   */
  protected void writeMRHIST(Concept concept)
    throws IOException {

    //
    // Iterate through ComponentHistory
    //
    final ComponentHistory[] histories = concept.getComponentHistories();
    final List mrhist_lines = new ArrayList(histories.length);
    for (int i=0; i<histories.length; i++) {

      //
      // Prepare MRHIST.RRF line
      //
      final StringBuffer sb = new StringBuffer(100);
      final ComponentHistory history = histories[i];
      sb.append(history.getConcept().getCui()).append("|");
      sb.append((history.getUi() != null) ? history.getUi() : "");sb.append("|");
      if (versioned_output)
        sb.append(history.getSource().getVersionedSourceAbbreviation());
      else
        sb.append(history.getSource().getRootSourceAbbreviation());
      sb.append("|");
      sb.append((history.getSver() != null) ? history.getSver() : "").append("|");
      sb.append((history.getChangeType() != null) ? history.getChangeType() : "").append("|");
      sb.append((history.getChangeKey() != null) ? history.getChangeKey() : "").append("|");
      sb.append((history.getChangeValue() != null) ? history.getChangeValue() : "").append("|");
      sb.append((history.getReason() != null) ? history.getReason() : "").append("|");
      sb.append((history.getCvf() != null) ? history.getCvf().toString() : "").append("|");
      //
      // Add line to list
      //
      mrhist_lines.add(sb.toString());
    }

    //
    // Sort lines and write MRHIST.RRF
    //
    Collections.sort(mrhist_lines);
    for (int i = 0; i < mrhist_lines.size(); i++) {
      printLine(mrhist_lines.get(i).toString(), "MRHIST");
    }
  }

  //
  // Protected Methods
  //

  /**
   * Opens the specified file in the directory passed to
   * {@link #open(String)}.
   * @param name the filename
   * @return a {@link BufferedWriter} for the file
   * @throws IOException if the file could not be opened.
   */
  protected BufferedWriter openTargetFile(String name) throws IOException {
    FileOutputStream fos = new FileOutputStream(new File(mr_dir, name));

    target_to_output_stream.put(name, fos);

    return new BufferedWriter(new UTF8OutputStreamWriter(
      fos, add_unicode_bom),128*1024);
  }

  /**
   * DO NOT USE.
   * @param name the name
   * @return the file pos
   * @throws IOException if anything goes wrong
   */
  protected long getFilePosition(String name) throws IOException {
    return ((FileOutputStream)target_to_output_stream.get(name)).getChannel().position();
  }
  
  /**
   * Returns a {@link String}[] of warnings accumulated during
   * processing of the output stream.
   * @return {@link String}[] of warnings
   */
  public String[] getWarnings() {
  	return (String[])warnings.toArray(new String[0]);
  }
}
