/**************************************************************************
 *
 * Package:     gov.nih.nlm.umls.io
 * Object:      RichMRFilesSingleConceptInputStream.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.io;

import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.impl.ConceptImpl;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.StringReader;

/**
 * Builds a single {@link Concept} from a
 * full set of "Rich Release Format" MR files.
 *
 * @author Deborah Shapiro
 */
public class RichMRFilesSingleConceptInputStream  {

  //
  // Private Fields
  //
  private RichMRFilesConceptInputStream rrf = null;
  private String cui = null;
  private String dir = null;
  private Concept concept = null;

  //
  // Constructors
  //

  /**
   * Instantiates an empty {@link RichMRFilesSingleConceptInputStream};
   * @param cui cui
   * @param dir location for files to be read
   * @throws IOException
   */
  public RichMRFilesSingleConceptInputStream(String cui, String dir)
      throws IOException{
    rrf = new RichMRFilesConceptInputStream();

    this.cui = cui;
    this.dir = dir;


  }

  //
  // Methods
  //
  /**
   * Returns a single {@link Concept} from a set of
   * Rich Release files.
   * @return {@link Concept}
   * @throws IOException
   */
  public Concept getConcept() throws IOException {
    rrf.mrconso_source = openSourceFile("MRCONSO.RRF");
    rrf.mrsty_source = openSourceFile("MRSTY.RRF");
    rrf.mrdef_source = openSourceFile("MRDEF.RRF");
    rrf.mrrel_source = openSourceFile("MRREL.RRF");
    rrf.mrcoc_source = openSourceFile("MRCOC.RRF");
    rrf.mrsat_source = openSourceFile("MRSAT.RRF");
    rrf.mrhist_source = openSourceFile("MRHIST.RRF");
    rrf.mrmap_source = openSourceFile("MRMAP.RRF");
    rrf.mrhier_source = openSourceFile("MRHIER.RRF");
    rrf.mrcxt_source = openSourceFile("MRCXT.RRF");
    rrf.mrdoc_source = new PushBackReader(
        new BufferedReader(new StringReader("")));

    concept = rrf.readConcept();
    rrf.close();
    return concept;
  }

  /**
   * Returns a single {@link Concept} from a set of
   * Rich Release files.  Only the atoms of
   * the {@link Concept} are populated.
   * @return {@link Concept}
   * @throws IOException
   */
  public Concept getConceptWithAtomsOnly() throws IOException {
    concept = new ConceptImpl();
    rrf.mrconso_source = openSourceFile("MRCONSO.RRF" );

    rrf.getAtoms(concept, cui);
    rrf.mrconso_source.close();
    return concept;
  }


  /**
   * Opens the source file of the given name. Returns a
   * {@link SearchStringPushBackReader} with the file pointer set
   * to the row that starts with the value of the class variable cui.
   * @param name the filename
   * @return a {@link SearchStringPushBackReader} on the file
   * @throws IOException if the file could not be opened.
   */
  private PushBackReader openSourceFile(String name) throws IOException {
    File f = new File(dir, name);
    if (f.exists()) {
      return new SearchStringPushBackReader(new File(dir, name), cui);
    }
    //
    // If we get here, the file was not found.
    //
    return new PushBackReader(new BufferedReader(new StringReader("")));
  }

}
