/***************************************************************************
 * Package:     gov.nih.nlm.umls.io
 * Interface:   SearchStringPushBackReader.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.io;

import gov.nih.nlm.util.SystemToolkit;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.io.StringReader;

/**
 *
 * Extends {@link PushBackReader} to provide the ability to
 * find a single line of input, given a search string or pattern,
 * and to push it back onto the reader to read it again
 * on the next <code>readLine()</code> call.
 *
 * @author Brian Carlsen
 */
public class SearchStringPushBackReader extends PushBackReader {

  //
  // Character set to use
  //
  protected String char_set;

  //
  // The file to wrap around
  //
  protected RandomAccessFile raf;

  //
  // The "look" pattern
  //
  protected String search_pattern;

  /**
   * Instantiates a new {@link SearchStringPushBackReader} from
   * the specified file and search pattern.
   * @param file the {@link File} to open and search in
   * @param search_str the search pattern
   * @throws IOException
   * @throws FileNotFoundException
   */
  public SearchStringPushBackReader(File file, String search_str) throws
      IOException, FileNotFoundException {
    this(new RandomAccessFile(file, "r"),search_str);
  }

  /**
   * Instantiates a new {@link SearchStringPushBackReader} from
   * the specified file and search pattern.
   * @param raf the {@link RandomAccessFile} to search in
   * @param search_str the search pattern
   * @throws IOException
   * @throws FileNotFoundException
   */
  public SearchStringPushBackReader(RandomAccessFile raf, String search_str) throws
      IOException, FileNotFoundException {
    super(new BufferedReader(new StringReader("")));
    this.char_set = "UTF-8";
    this.raf = raf;
    search_pattern = search_str;
    if (search_pattern.endsWith(".*|")) 
       search_pattern = search_pattern.substring(0, search_pattern.lastIndexOf(".*|"));
    // Treat | as a literal pipe character
    if (search_pattern.indexOf('|') != -1)
      search_pattern = search_pattern.replaceAll("\\|", "\\\\|");
    // Treat . as a literal . character (if .*, see above)
    if (search_pattern.indexOf('.') != -1)
      search_pattern = search_pattern.replaceAll("\\.", "\\\\.");
    search_pattern = search_pattern  + ".*";

    String l_search_str = search_str.endsWith(".*|") ?
        search_str.substring(0, search_str.lastIndexOf(".*|")) : search_str;

    SystemToolkit.seekstr(raf, l_search_str, char_set);
    line = null;
  }

  /**
   * Returns the next line from the reader.
   * @return the next line from the reader
   * @throws IOException
   */
  public String readLine() throws IOException {
    String result = null;
    if (line != null) {
      result = line;
    } else {
      result = SystemToolkit.readLine(raf, char_set);
    }
    if (!result.matches(search_pattern)) {
      push(result);
      return null;
    }
    line = null;
    return result;
  }

  /**
   * Sets the character set to the given parameter.
   * @param char_set character set parameter such as "UTF-8"
   */
  public void setCharSet(String char_set) {
    this.char_set = char_set;
  }

  /**
   * Close the {@link SearchStringPushBackReader}.
   * @throws IOException
   */
  public void close() throws IOException {
    super.close();
    raf.close();
  }
}
