/**************************************************************************
 *
 * Package:     gov.nih.nlm.umls.io
 * Object:      SuiCountIndexBuilder.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.io;

import gov.nih.nlm.util.FieldedStringTokenizer;
import gov.nih.nlm.util.SystemToolkit;
import gov.nih.nlm.util.UTF8InputStreamReader;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;

import javax.swing.ProgressMonitor;

public class SuiCountIndexBuilder extends IndexBuilder {

  //
  // Delimiter list
  //
  private final String delim = " \t-({[)}]_!@#%&*\\:;\"',.?/~+=|<>$`^";

  /**
   * Instantiates a {@link SuiCountIndexBuilder}.
   * @param dir the directory containing the file to index
   */
  public SuiCountIndexBuilder(String dir) {
    super(dir);
  }

  /**
   * Builds an index file called SUI_CT_MRXW.x
   * @param pm {@link ProgressMonitor} for reporting indexing progress
   * @throws java.lang.Exception
   */
  public void buildSuiCountWordIndex(final ProgressMonitor pm) throws Exception {
    final File idir = new File(dir, "indexes");
    if (!idir.exists())
      idir.mkdir();

    // build index file name
    String file_name = "SUI_CT_MRXW.x";
    final File index = new File(idir, file_name);

    // open input streams
    final File in_file = new File(dir, "MRCONSO.RRF");
    long in_file_length = in_file.length();
    FileInputStream fis = new FileInputStream(in_file);
    UTF8InputStreamReader uisr = new UTF8InputStreamReader(
        new BufferedInputStream(fis));
    final BufferedReader in = new BufferedReader(uisr, 512 * 1024);
    final BufferedWriter out = new BufferedWriter(new OutputStreamWriter(
        new FileOutputStream(index), "UTF-8"), 512 * 1024);
    long pos = uisr.hasByteOrderMark() ? 3 : 0;


    // for each line in the file get the file pointer
    int counter = 0;
    int prev_progress = 0;
    String line = null;
    while ( (line = in.readLine()) != null) {

      final String[] tokens = FieldedStringTokenizer.split(line, "|");
      final String[] words =
          SystemToolkit.getIndexWords(FieldedStringTokenizer.split(tokens[14], delim));

      for (int i = 0; i < words.length; i++) {
        final String word = words[i];
        // LAT
        out.write(tokens[1]);
        out.write("|");

        // WORD
        out.write(word.toLowerCase());
        out.write("|");

        // CUI
        out.write(tokens[0]);
        out.write("|");

        // LUI
        out.write(tokens[3]);
        out.write("|");

        // SUI
        out.write(tokens[5]);
        out.write("|");

        // SUI CT
        out.write(String.valueOf(words.length));
        out.newLine();
        //out.flush();

      }
      pos = fis.getChannel().position();

      // progress monitor - the first portion is 30%
      // the sorting is 70%
      counter++;
      if (counter == 10000 && pm != null) {
        counter = 0;
        final int i = (int) ( (pos * 29) / in_file_length) + 1;
        if (i > prev_progress) {
          pm.setProgress(i);
          pm.setNote("Operation is " + i + "% complete.");
          prev_progress = i;
        }
        if (pm.isCanceled()) {
          in.close();
          out.close();
          index.delete();
          throw new Exception("Index build cancelled.");
        }
      }
    }
    in.close();
    out.flush();
    out.close();

    //
    // Set up sort progress monitor
    //
    ProgressMonitor pm2 = null;
    if (pm != null)
      pm2 = new ProgressMonitor(null, null, null, 0, 100) {

        private int prev_progress = 0;
        public void setProgress(int progress) {
          final int i = ( (int) ( ( (progress * 70) / 100) + 30));
          if (i > prev_progress) {
            pm.setProgress(i);
            pm.setNote("Operation is " + i + "% complete.");
            prev_progress = i;
          }
        }
      };

    //
    // sort index
    //
    SystemToolkit.sort(index.getAbsolutePath(), true, pm2);

  }

  /**
   * Builds an index file called SUI_CT.x
   * @param pm {@link ProgressMonitor} for reporting indexing progress
   * @throws java.lang.Exception
   */
  public void buildSuiCountIndex(final ProgressMonitor pm) throws Exception {
    final File idir = new File(dir, "indexes");
    if (!idir.exists())
      idir.mkdir();

    // build index file name
    String file_name = "SUI_CT.x";
    final File index = new File(idir, file_name);

    // open input streams
    final File in_file = new File(dir, "MRCONSO.RRF");
    long in_file_length = in_file.length();
    FileInputStream fis = new FileInputStream(in_file);
 
    UTF8InputStreamReader uisr = new UTF8InputStreamReader(
        new BufferedInputStream(fis));
    final BufferedReader in = new BufferedReader(uisr, 512 * 1024);
    final BufferedWriter out = new BufferedWriter(new OutputStreamWriter(
        new FileOutputStream(index), "UTF-8"), 512 * 1024);
    long pos = uisr.hasByteOrderMark() ? 3 : 0;

    // for each line in the file get the file pointer
    int counter = 0;
    int prev_progress = 0;
    String line = null;
    while ( (line = in.readLine()) != null) {

      final String[] tokens = FieldedStringTokenizer.split(line, "|");
      if (tokens[1].equals("ENG")) {
        final String[] words =
            SystemToolkit.getIndexWords(FieldedStringTokenizer.split(tokens[14],
            delim));

        for (int i = 0; i < words.length; i++) {
          // SUI
          out.write(tokens[5]);
          out.write("|");

          // SUI CT
          out.write(String.valueOf(words.length));
          out.newLine();
          //out.flush();
        }
        pos = fis.getChannel().position();
      }

      // progress monitor - the first portion is 30%
      // the sorting is 70%
      counter++;
      if (counter == 10000 && pm != null) {
        counter = 0;
        final int i = (int) ( (pos * 29) / in_file_length) + 1;
        if (i > prev_progress) {
          pm.setProgress(i);
          pm.setNote("Operation is " + i + "% complete.");
          prev_progress = i;
        }
        if (pm.isCanceled()) {
          in.close();
          out.close();
          index.delete();
          throw new Exception("Index build cancelled.");
        }
      }
    }
    in.close();
    out.flush();
    out.close();

    //
    // Set up sort progress monitor
    //
    ProgressMonitor pm2 = null;
    if (pm != null)
      pm2 = new ProgressMonitor(null, null, null, 0, 100) {

        private int prev_progress = 0;
        public void setProgress(int progress) {
          final int i = ( (int) ( ( (progress * 70) / 100) + 30));
          if (i > prev_progress) {
            pm.setProgress(i);
            pm.setNote("Operation is " + i + "% complete.");
            prev_progress = i;
          }
        }
      };

    //
    // sort index
    //
    SystemToolkit.sort(index.getAbsolutePath(), true, pm2);

  }




}
