/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Class:       AdditionalRelationshipLabel.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Represents an additional relationship label value. This class ensures that each unique
 * additional relationship label is represented by a singleton object.  In order for
 * this to work correctly, the class must be properly initialized.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public final class AdditionalRelationshipLabel extends AbstractExpandedForm {

  // Map by abbreviation to the actual classes
  private static final Map relas = new HashMap();

  // Abbreviation and description for this individual instance.
  private String abbrev;
  private String dsc;

  /**
   * Instantiates an empty {@link AdditionalRelationshipLabel}.
   */
  public AdditionalRelationshipLabel() {};

  /**
   * Instantiates a {@link AdditionalRelationshipLabel} from the
   * specified {@link MetaDescription}.
   * @param meta_dsc a {@link MetaDescription} representing a
   * additional relationship label
   */
  private AdditionalRelationshipLabel(MetaDescription meta_dsc) {
    abbrev = meta_dsc.getValue();
    dsc = meta_dsc.getExplain();
  }

  /**
   * Initializes the {@link Map} used to track all of the
   * {@link AdditionalRelationshipLabel}s that exist at one time.  You may
   * pass all {@link MetaDescription}s as this method only
   * creates {@link AdditionalRelationshipLabel}s for those with a type of "RELA".
   * @param meta_dsc an array of {@link MetaDescription}s
   */
  public static synchronized void initialize(MetaDescription[] meta_dsc) {
    for (int i=0; i<meta_dsc.length; i++) {
      if (meta_dsc[i].getKey().equals("RELA")&&
          meta_dsc[i].getType().equals("expanded_form")) {
	if (relas.containsKey(meta_dsc[i].getValue())) {
	   AdditionalRelationshipLabel rel_att = (AdditionalRelationshipLabel)relas.get(meta_dsc[i].getValue());
           rel_att.dsc = meta_dsc[i].getExplain();
        } else {
          relas.put(meta_dsc[i].getValue(), new AdditionalRelationshipLabel(meta_dsc[i]));
        }
      }
    }
  }

  /**
   * Initializes the {@link Map} used to track all of the
   * {@link AdditionalRelationshipLabel}s that exist at one time.  This
   * is implemented as a call to {@link #initialize(MetaDescription[])}.
   * @param meta_dsc an array of {@link MetaDescription}s
   * @throws IllegalStateException if an attempt is made to call this
   * method more than once
   */
  public static synchronized void initialize(List meta_dsc) {
    initialize((MetaDescription[])meta_dsc.toArray(new MetaDescription[0]));
  }

  /**
   * Returns the {@link AdditionalRelationshipLabel} for the specified abbreviation.
   * This method will return <code>null</code> if the <code>initialize</code>
   * method has not been called or if the specified abbreviation
   * was not one of those used during initialization.
   * @param rela the additional relationship label abbreviation to look up
   * @return the {@link AdditionalRelationshipLabel} for the specified abbreviation
   */
  public static AdditionalRelationshipLabel getAdditionalRelationshipLabel(String rela) {
    return (AdditionalRelationshipLabel)relas.get(rela);
  }

  /**
   * Return all {@link AdditionalRelationshipLabel}s.
   * @return an array of {@link AdditionalRelationshipLabel}s
   */
  public static AdditionalRelationshipLabel[] getAdditionalRelationshipLabels() {
    return (AdditionalRelationshipLabel[])relas.values().toArray(new AdditionalRelationshipLabel[0]);
  }

  /**
   * Return all {@link AdditionalRelationshipLabel}s as a {@link List}.
   * @return all {@link AdditionalRelationshipLabel}s as a {@link List}
   */
  public static List getAdditionalRelationshipLabelsAsList() {
    return new ArrayList(relas.values());
  }

  /**
   * Returns the abbreviation.
   * @return the abbreviation
   */
  public String getAbbreviation() {
    return abbrev;
  }

  /**
   * Returns the description
   * @return the description
   */
  public String getDescription() {
    return dsc;
  }
}