/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Interface:   AttributeContext.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;

import java.util.List;

/**
 * Represents information attributes of a {@link Source} and code
 * within a {@link StringAttribute}.
 * <p>
 * Following is a UML Class diagram showing the neighborhood around {@link AttributeContext}
 * <p><center><img src="AttributeContext.gif"></center>
 * <p>
 * @see StringAttribute
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public interface AttributeContext extends Comparable, SourceElement {

  /**
   * Returns the {@link StringAttribute} that holds this {@link AttributeContext}.
   * @return the {@link StringAttribute} that holds this {@link AttributeContext}
   */
  public StringAttribute getStringAttribute();

  /**
   * Sets the {@link StringAttribute} that holds this {@link AttributeContext}.
   * @param sa the {@link StringAttribute} that holds this {@link AttributeContext}
   */
  public void setStringAttribute(StringAttribute sa);

  /**
   * Returns the {@link Atom} for this {@link AttributeContext}.
   * @return {@link Atom}
   */
  public Atom getAtom();

  /**
   * Sets the {@link Atom} for this {@link AttributeContext}.
   * @param string_source {@link Atom}
   */
  public void setAtom(Atom string_source);

  /**
   * Returns the {@link SourceIdentifierType} for this {@link AttributeContext}.
   * @return {@link SourceIdentifierType}
   */
  public SourceIdentifierType getSourceIdentifierType();

  /**
   * Sets the {@link SourceIdentifierType} for this {@link AttributeContext}.
   * @param source_id_type {@link SourceIdentifierType}
   */
  public void setSourceIdentifierType(SourceIdentifierType source_id_type);

  /**
   * Returns all {@link AttributeValue}s.
   * @return a array of {@link AttributeValue}s
   */
  public AttributeValue [] getAttributeValues();

  /**
   * Returns all {@link AttributeValue}s as a {@link List}.
   * @return a {@link List} of {@link AttributeValue}s
   */
  public List getAttributeValuesAsList();

  /**
   * Adds the specified {@link AttributeValue}.
   * @param av the {@link AttributeValue} to be added
   */
  public void addAttributeValue(AttributeValue av);

  /**
   * Removes the specified {@link AttributeValue}.
   * @param av the {@link AttributeValue} to be removed
   */
  public void removeAttributeValue(AttributeValue av);

  /**
   * Removes all {@link AttributeValue}s.
   */
  public void clearAttributeValues();

  /**
   * Replaces the {@link AttributeValue}s with the ones in the specified array.
   * @param avs a {@link AttributeValue}<code>[]</code>
   */
  public void setAttributeValues(AttributeValue [] avs);

  /**
   * Replaces the {@link AttributeValue}s with the ones in the specified {@link List}.
   * @param avs a {@link List} of {@link AttributeValue}s
   */
  public void setAttributeValues(List avs);

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from key {@link AttributeContext} fields
   */
  public String getKeyFields();
}