/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Class:       ChainRestrictor.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;

import java.util.Comparator;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

/**
 * Default implementation of the {@link gov.nih.nlm.umls.meta.CoocurrenceContext}
 * interface.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class ChainRestrictor implements Restrictor {

  //
  // Fields
  //
  private Set restrictor_set = new HashSet();
  private Comparator comparator = null;

  /**
   * Instantiates an empty {@link ChainRestrictor}.
   */
  public ChainRestrictor() {};

  /**
   * Indicates if the given {@link Object} should be
   * kept based on the {@link Restrictor}s in the chain.
   * @param o {@link Object}
   * @return <code>boolean</code> returns false if the given object is not to
   *                              be kept by all the {@link Restrictor}s in the chain
   *                              true otherwise
   */
  public boolean keep(Object o) {
    Iterator iter = restrictor_set.iterator();
    while(iter.hasNext()) {
      Restrictor res = (Restrictor)iter.next();
      if (!res.keep(o))
        return false;
    }
    return true;
  }

  /**
   * Adds the given {@link Restrictor} to
   * the chain so that it will be applied.
   * @param restrictor {@link Restrictor}
   */
  public void add(Restrictor restrictor) {
    add(restrictor, false);
  }

  /**
   * Adds the given {@link Restrictor} to the
   * chain and uses it for the sort order.
   * @param restrictor {@link Restrictor}
   * @param is_comparator <code>boolean</code>
   */
  public void add(Restrictor restrictor, boolean is_comparator) {
    restrictor_set.add(restrictor);
    comparator = restrictor;
  }

  /**
   * Sets the {@link Comparator} to be used for the sort order.
   * @param comp {@link Comparator}
   */
  public void setComparator(Comparator comp) {
    comparator = comp;
  }

  /**
   * Removes the given {@link Restrictor} from
   * the chain so it will not be applied.
   * @param restrictor {@link Restrictor}
   */
  public void remove(Restrictor restrictor) {
    restrictor_set.remove(restrictor);
    if (comparator == restrictor)
      comparator = null;
  }

  /**
   * Clears list of restictors in chain so that
   * no restrictors will be applied.
   */
  public void clear() {
    restrictor_set.clear();
  }

  /**
   * Compares objects for sort order.
   * @param obj1 {@link Object}
   * @param obj2 {@link Object}
   * @return <code>int</code>
   */
  public int compare(Object obj1, Object obj2) {
    if (comparator == null && !restrictor_set.isEmpty())
      comparator = ((Comparator)restrictor_set.iterator().next());
    return comparator.compare(obj1, obj2);
  }
}