/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Interface:   Concept.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;

import java.util.Comparator;
import java.util.List;

/**
 * Represents a UMLS Concept.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public interface Concept extends Comparable{

  /**
   * This removes all of the {@link Concept}'s components.
   */
  public void clear();

  /**
   * This indicates if the {@link Concept} is empty, meaning
   * that either the {@link SemanticType} was removed or
   * all of the {@link Atom}s were removed.
   * @return boolean if empty
   */
  public boolean isEmpty();

  //
  // Atoms
  //

  /**
   * Returns the {@link Atom}s.
   * @return an {@link Atom} <code>[]</code>
   */
  public Atom[] getAtoms();

  /**
   * Returns the {@link Atom}s as a {@link List}.
   * @return the {@link Atom}s as a {@link List}
   */
  public List getAtomsAsList();

  /**
   * Adds the specified {@link Atom}.
   * @param atom the {@link Atom} to be added
   */
  public void addAtom(Atom atom);

  /**
   * Removes the specified {@link Atom}.
   * @param atom the {@link Atom} to be removed
   */
  public void removeAtom(Atom atom);

  /**
   * Remove all {@link Atom}s.
   */
  public void clearAtoms();

  /**
   * Replace any {@link Atom}s with the
   * ones in the specified array.
   * @param atoms a {@link Atom} <code>[]</code>
   */
  public void setAtoms(Atom atoms[]);

  /**
   * Replace {@link Atom}s with the
   * ones in the specified {@link List}.
   * @param atoms a {@link List} of {@link Atom}s
   */
  public void setAtoms(List atoms);

  /**
   * Returns the preferred atom of the concept.
   * It has TS=P, STT=PF, and ISPREF=Y.
   * @return {@link Atom}
   */
  public Atom getPreferredAtom();

  /**
   * Returns the preferred atom of the concept.
   * Uses the given {@link Comparator} to determine the preferred atom.
   * @param comparator {@link Comparator}
   * @return {@link Atom}
   */
  public Atom getPreferredAtom(Comparator comparator);

  //
  // Locators
  //

  /**
   * Returns the {@link Locator}s.
   * @return a {@link Locator} <code>[]</code>
   */
  public Locator[] getLocators();

  /**
   * Returns the {@link Locator}s as a {@link List}.
   * @return the {@link Locator}s as a {@link List}
   */
  public List getLocatorsAsList();

  /**
   * Add the specified {@link Locator}.
   * @param loc the {@link Locator} to be added
   */
  public void addLocator(Locator loc);

  /**
   * Removes the specified {@link Locator}.
   * @param loc the {@link Locator} to be removed
   */
  public void removeLocator(Locator loc);

  /**
   * Remove all {@link Locator}s.
   */
  public void clearLocators();

  /**
   * Replace {@link Locator}s with the ones in the specified array.
   * @param locs a {@link Locator} <code>[]</code>
   */
  public void setLocators(Locator locs[]);

  /**
   * Replace {@link Locator}s with the ones in the specified {@link List}.
   * @param locs a {@link List} of {@link Locator}s
   */
  public void setLocators(List locs);


  //
  // Attributes
  //

  /**
   * Returns the {@link Attribute}s.
   * @return a {@link Attribute} <code>[]</code>
   */
  public Attribute[] getAttributes();

  /**
   * Returns the {@link Attribute}s as a {@link List}.
   * @return the {@link Attribute}s as a {@link List}
   */
  public List getAttributesAsList();

  /**
   * Adds the specified {@link Attribute}.
   * @param att the {@link Attribute} to be added
   */
  public void addAttribute(Attribute att);

  /**
   * Removes the specified {@link Attribute}.
   * @param att the {@link Attribute} to be removed
   */
  public void removeAttribute(Attribute att);

  /**
   * Remove all {@link Attribute}s.
   */
  public void clearAttributes();

  /**
   * Replace {@link Attribute}s with the ones in the specified array.
   * @param atts an {@link Attribute} <code>[]</code>
   */
  public void setAttributes(Attribute atts[]);

  /**
   * Replace {@link Attribute}s with the ones in the specified {@link List}.
   * @param atts a {@link List} of {@link Attribute}s
   */
  public void setAttributes(List atts);

  /**
   * Returns the attributes with aui that matches that of the given {@link Atom}.
   * @param atom {@link Atom}
   * @return {@link List} of {@link Attribute}s
   */
  public Attribute[] getAttributesForAtom(Atom atom);

  /**
   * Returns the attributes with aui that matches that of the given {@link Atom}
   * in a {@link List}.
   * @param atom {@link Atom}
   * @return {@link List} of {@link Attribute}s
   */
  public List getAttributesForAtomAsList(Atom atom);

  /**
   * Returns the attributes with aui that matches that of the given {@link Relationship}.
   * @param rel {@link Relationship}
   * @return {@link List} of {@link AtomAttribute}s
   */
  public AtomAttribute[] getAttributesForRelationship(AtomRelationship rel);

  /**
   * Returns the attributes with aui that matches that of the given {@link Relationship}
   * in a {@link List}.
   * @param rel {@link Relationship}
   * @return {@link List} of {@link Attribute}s
   */
  public List getAttributesForRelationshipAsList(AtomRelationship rel);

  //
  // Contexts
  //

  /**
   * Return the {@link Context}s.
   * @return a {@link Context} <code>[]</code>
   */
  public Context[] getContexts();

  /**
   * Returns a subset of this {@link Concept}'s {@link Context}s.
   * Restricts the return based on the given {@link Restrictor}.
   * @param restrictor {@link Restrictor}
   * @return Context[]
   */
  public Context[] getContexts(Restrictor restrictor);

  /**
   * Returns the {@link Context}s as a {@link List}.
   * @return the {@link Context}s as a {@link List}
   */
  public List getContextsAsList();

  /**
   * Adds the specified {@link Context}.
   * @param cxt the {@link Context} to be added
   */
  public void addContext(Context cxt);

  /**
   * Removes the specified {@link Context}.
   * @param cxt the {@link Context} to be removed
   */
  public void removeContext(Context cxt);

  /**
   * Remove all {@link Context}s.
   */
  public void clearContexts();

  /**
   * Replace {@link Context}s with the ones in the specified array.
   * @param cxts a {@link Context} <code>[]</code>
   */
  public void setContexts(Context cxts[]);

  /**
   * Replace {@link Context}s with the ones in the specified {@link List}.
   * @param cxts a {@link List} of {@link Context}s
   */
  public void setContexts(List cxts);

  //
  // Coocurrences
  //

  /**
   * Returns the {@link Coocurrence}s.
   * @return a {@link Coocurrence} <code>[]</code>
   */
  public Coocurrence[] getCoocurrences();

  /**
   * Returns a subset of this {@link Concept}'s {@link Coocurrence}s.
   * Restricts the return based on the given {@link Restrictor}.
   * @param restrictor {@link Restrictor}
   * @return Coocurrence[]
   */
  public Coocurrence[] getCoocurrences(Restrictor restrictor);

  /**
   * Returns the {@link Coocurrence}s as a {@link List}.
   * @return the {@link Coocurrence}s as a {@link List}
   */
  public List getCoocurrencesAsList();

  /**
   * Adds the specified {@link Coocurrence}.
   * @param coc the {@link Coocurrence} to be added
   */
  public void addCoocurrence(Coocurrence coc);

  /**
   * Removes the specified {@link Coocurrence}.
   * @param coc the {@link Coocurrence} to be removed
   */
  public void removeCoocurrence(Coocurrence coc);

  /**
   * Remove all {@link Coocurrence}s.
   */
  public void clearCoocurrences();

  /**
   * Replace {@link Coocurrence}s with the ones in the specified array.
   * @param cocs a {@link Coocurrence} <code>[]</code>
   */
  public void setCoocurrences(Coocurrence cocs[]);

  /**
   * Replace {@link Coocurrence}s with the ones in the specified {@link List}.
   * @param cocs a {@link List} of {@link Coocurrence}s
   */
  public void setCoocurrences(List cocs);

  //
  // Definitions
  //

  /**
   * Returns the {@link Definition}s.
   * @return a {@link Definition} <code>[]</code>
   */
  public Definition[] getDefinitions();

  /**
   * Returns a subset of this {@link Concept}'s {@link Definition}s.
   * Restricts the return based on the given {@link Restrictor}.
   * @param restrictor {@link Restrictor}
   * @return Definition[]
   */
  public Definition[] getDefinitions(Restrictor restrictor);

  /**
   * Returns the {@link Definition}s as a {@link List}.
   * @return the {@link Definition}s as a {@link List}
   */
  public List getDefinitionsAsList();

  /**
   * Adds the specified {@link Definition}.
   * @param def the {@link Definition} to be added
   */
  public void addDefinition(Definition def);

  /**
   * Remove the specified {@link Definition}.
   * @param def the {@link Definition} to be removed
   */
  public void removeDefinition(Definition def);

  /**
   * Remove all {@link Definition}s.
   */
  public void clearDefinitions();

  /**
   * Replace {@link Definition}s with the ones in the specified array.
   * @param defs a {@link Definition} <code>[]</code>
   */
  public void setDefinitions(Definition defs[]);

  /**
   * Replace {@link Definition}s with the ones in the specified {@link List}.
   * @param defs a {@link List} of {@link Definition}s
   */
  public void setDefinitions(List defs);

  //
  // Relationships
  //

  /**
   * Returns the {@link Relationship}s.
   * @return a {@link Relationship} <code>[]</code>
   */
  public Relationship[] getRelationships();

  /**
   * Returns a subset of this {@link Concept}'s {@link Relationship}s.
   * Restricts the return based on the given {@link Restrictor}.
   * @param restrictor {@link Restrictor}
   * @return Relationship[]
   */
  public Relationship[] getRelationships(Restrictor restrictor);

  /**
   * Returns the {@link Relationship}s as a {@link List}.
   * @return the {@link Relationship}s as a {@link List}
   */
  public List getRelationshipsAsList();

  /**
   * Adds the specified {@link Relationship}.
   * @param rel the {@link Relationship} to be added
   */
  public void addRelationship(Relationship rel);

  /**
   * Removes the specified {@link Relationship}.
   * @param rel the {@link Relationship} to be removed
   */
  public void removeRelationship(Relationship rel);

  /**
   * Remove all {@link Relationship}s.
   */
  public void clearRelationships();

  /**
   * Replace {@link Relationship}s with the ones in the specified array.
   * @param rels a {@link Relationship} <code>[]</code>
   */
  public void setRelationships(Relationship rels[]);

  /**
   * Replace {@link Relationship}s of the ones in the specified {@link List}.
   * @param rels a {@link List} of {@link Relationship}s
   */
  public void setRelationships(List rels);


  //
  // SemanticTypes
  //

   /**
   * Returns the {@link SemanticType}s.
   * @return a {@link SemanticType} <code>[]</code>
   */
  public SemanticType[] getSemanticTypes();

  /**
   * Returns the {@link SemanticType}s as a {@link List}.
   * @return the {@link SemanticType}s as a {@link List}
   */
  public List getSemanticTypesAsList();

  /**
   * Adds a {@link SemanticType}.
   * @param sty the {@link SemanticType} to be added
   */
  public void addSemanticType(SemanticType sty);

  /**
   * Removes the specified {@link SemanticType}.
   * @param sty the {@link SemanticType} to be removed
   */
  public void removeSemanticType(SemanticType sty);

  /**
   * Removes all {@link SemanticType}s.
   */
  public void clearSemanticTypes();

  /**
   * Replace {@link SemanticType}s with the ones in the specified array.
   * @param stys a {@link SemanticType} <code>[]</code>
   */
  public void setSemanticTypes(SemanticType stys[]);

  /**
   * Replace {@link SemanticType}s with the ones in the specified {@link List}.
   * @param stys a {@link List} of {@link SemanticType}s
   */
  public void setSemanticTypes(List stys);

  //
  // Component Histories
  //

   /**
   * Returns the {@link ComponentHistory}s.
   * @return a {@link ComponentHistory} <code>[]</code>
   */
  public ComponentHistory[] getComponentHistories();

  /**
   * Returns the {@link ComponentHistory}s as a {@link List}.
   * @return the {@link ComponentHistory}s as a {@link List}
   */
  public List getComponentHistoriesAsList() ;

  /**
   * Adds a {@link ComponentHistory}.
   * @param map the {@link ComponentHistory} to be added
   */
  public void addComponentHistory(ComponentHistory map);

  /**
   * Removes the specified {@link ComponentHistory}.
   * @param map the {@link ComponentHistory} to be removed
   */
  public void removeComponentHistory(ComponentHistory map);

  /**
   * Removes all {@link ComponentHistory}s.
   */
  public void clearComponentHistories();

  /**
   * Replace {@link ComponentHistory}s with the ones in the specified array.
   * @param maps a {@link ComponentHistory} <code>[]</code>
   */
  public void setComponentHistories(ComponentHistory maps[]);

  /**
   * Replace {@link ComponentHistory}s with the ones in the specified {@link List}.
   * @param maps a {@link List} of {@link ComponentHistory}s
   */
  public void setComponentHistories(List maps);

  //
  // Mappings
  //

   /**
   * Returns the {@link Mapping}s.
   * @return a {@link Mapping} <code>[]</code>
   */
  public Mapping[] getMappings();

  /**
   * Returns the {@link Mapping}s as a {@link List}.
   * @return the {@link Mapping}s as a {@link List}
   */
  public List getMappingsAsList();

  /**
   * Adds a {@link Mapping}.
   * @param map the {@link Mapping} to be added
   */
  public void addMapping(Mapping map);

  /**
   * Removes the specified {@link Mapping}.
   * @param map the {@link Mapping} to be removed
   */
  public void removeMapping(Mapping map);

  /**
   * Removes all {@link Mapping}s.
   */
  public void clearMappings();

  /**
   * Replace {@link Mapping}s with the ones in the specified array.
   * @param maps a {@link Mapping} <code>[]</code>
   */
  public void setMappings(Mapping maps[]);

  /**
   * Replace {@link Mapping}s with the ones in the specified {@link List}.
   * @param maps a {@link List} of {@link Mapping}s
   */
  public void setMappings(List maps);

  /**
   * Returns the concept unique identifier (CUI) for the {@link Concept}.
   * This is shorthand for {@link #getConceptIdentifier()}.
   * @return the CUI
   */
  public String getCui();

  /**
   * Sets the concept unique identifier (CUI) for the {@link Concept}.
   * This is shorthand for {@link #setConceptIdentifier(String)}.
   * @param cui the CUI
   */
  public void setCui(String cui);

  /**
   * Returns the concept unique identifier (CUI) for the {@link Concept}.
   * @return the CUI
   */
  public String getConceptIdentifier();

  /**
   * Sets the concept unique identifier (CUI) for the {@link Concept}.
   * @param cui the CUI
   */
  public void setConceptIdentifier(String cui);

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from key {@link Concept} fields
   */
  public String getKeyFields();

  //
  // UMLSKS View Methods
  //

  /**
   * Return the {@link ContextGroup}s.
   * @return a {@link ContextGroup} <code>[]</code>
   */
  public ContextGroup[] getContextGroups();

  /**
   * Returns the {@link ContextGroup}s as a {@link List}.
   * @return the {@link ContextGroup}s as a {@link List}
   */
  public List getContextGroupsAsList();

  /**
   * Returns the {@link CoocurrenceGroup}s.
   * @return a {@link CoocurrenceGroup} <code>[]</code>
   */
  public CoocurrenceGroup[] getCoocurrenceGroups();

  /**
   * Returns the {@link CoocurrenceGroup}s as a {@link List}.
   * @return the {@link CoocurrenceGroup}s as a {@link List}
   */
  public List getCoocurrenceGroupsAsList();

  /**
   * Returns the {@link TermAttribute}s.
   * @return a {@link TermAttribute} <code>[]</code>
   */
  public TermAttribute[] getTermAttributes();

  /**
   * Returns the {@link TermAttribute}s as a {@link List}.
   * @return the {@link TermAttribute}s as a {@link List}
   */
  public List getTermAttributesAsList();

  /**
   * Returns the {@link Term}s.
   * @return an {@link Term} <code>[]</code>
   */
  public Term[] getTerms();

  /**
   * Returns the {@link Term}s as a {@link List}.
   * @return the {@link Term}s as a {@link List}
   */
  public List getTermsAsList();

  /**
   * Returns a subset of this {@link Concept}'s {@link Atom}s.
   * Restricts the return based on the given {@link Restrictor}.
   * @param restrictor {@link Restrictor}
   * @return Atom[]
   */
  public Atom[] getAtoms(Restrictor restrictor);


  //
  // Inner classes
  //
  /**
   * Inner class compares {@link Concept}s based on cui.
   */
  public final static class ByCUIComparator implements Comparator {
    public ByCUIComparator() {
    }

    /**
     * Compares the specified field of each {@link String} object.
     * @param o1 the first {@link Object} to be compared.
     * @param o2 the second {@link Object} to be compared.
     * @return An <code>int</code> which might be negative integer, zero, or a positive integer
     *    as the first argument is less than, equal to, or greater than the second.
     */
    public int compare(Object o1, Object o2) {
      if (o1 == null || o2 == null ||
          !(o1 instanceof Concept) || !(o2 instanceof Concept))
         return 0;
      Concept s1 = (Concept) o1;
      Concept s2 = (Concept) o2;

      String cui1 = s1.getCui();
      String cui2 = s2.getCui();

      return cui1.compareTo(cui2);
    }

    /**
     * Returns the display name for the {@link Comparator}.
     * @return {@link String}
     */
    public String toString() {
      return "Cui Sort Order";
    }

    public boolean equals(Comparator comp) {
      return this.toString().equals(comp.toString());
    }
  }

  /**
   * Sorts the {@link Concept}s based on {@link String} order.
   */
  public final static class ByPreferredNameComparator implements Comparator {

    private Language lat;

    /**
     * Instantiates a comparator.
     * @param lat {@link Language} in which sort will be done
     */
    public ByPreferredNameComparator(Language lat) {
      this.lat = lat;
    }

    /**
     * Compares the specified field of each {@link String} object.
     * @param o1 the first {@link Object} to be compared.
     * @param o2 the second {@link Object} to be compared.
     * @return An <code>int</code> which might be negative integer, zero, or a positive integer
     *    as the first argument is less than, equal to, or greater than the second.
     */
    public int compare(Object o1, Object o2) {
      if (o1 == null || o2 == null ||
          !(o1 instanceof Concept) || !(o2 instanceof Concept))
         return 0;
      final Concept c1 = (Concept) o1;
      final Concept c2 = (Concept) o2;

      return getPreferredName(c1).compareTo(getPreferredName(c2));
    }

    /**
     * Looks up concept preferred name.
     * @param c the {@link Concept}
     * @return the preferred name of the {@link Concept}
     */
    public String getPreferredName(final Concept c) {
      final Atom[] atoms = c.getAtoms();
      for (int i=0; i<atoms.length; i++) {
        final Atom atom = atoms[i];
        if(atom.hasPreferredAtomIdentifier() &&
           atom.hasPreferredStringIdentifier() &&
           atom.hasPreferredTermIdentifier() &&
           atom.getLanguage().equals(lat)) {
          return atom.getStr().toLowerCase();
        }
      }
      return "";
    }

    public String toString() {
      return "String Sort Order";
    }

    public boolean equals(Comparator comp) {
      return this.toString().equals(comp.toString());
    }
  }  // end of StringSortOrderComparator

}