/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Interface:   CoocurrenceContext.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;

import java.util.List;

/**
 * Represents part of a {@link CoocurrenceGroup}, containing more
 * detailed information about how two co ocurring meanings (or identifiers)
 * are related in a particular source of biomedical literature.
 * <p>
 * Following is a UML Class diagram showing the neighborhood around {@link CoocurrenceContext}
 * <p><center><img src="CoocurrenceContext.gif"></center>
 * <p>
 * @see CoocurrenceGroup
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public interface CoocurrenceContext
  extends Comparable, ContentViewElement, SourceElement {

  /**
   * Returns the {@link CoocurrenceType} (COT).
   * @return the {@link CoocurrenceType}
   */
  public CoocurrenceType getCoocurrenceType();

  /**
   * Returns the {@link CoocurrenceType} (COT).
   * This is shorthand for {@link #getCoocurrenceType()}.
   * @return the {@link CoocurrenceType}
   */
  public CoocurrenceType getCot();

  /**
   * Sets the {@link CoocurrenceType} (COT).
   * @param cot the {@link CoocurrenceType}
   */
  public void setCoocurrenceType(CoocurrenceType cot);

  /**
   * Sets the {@link CoocurrenceType} (COT).
   * This is shorthand for {@link #setCoocurrenceType(CoocurrenceType)}.
   * @param cot the {@link CoocurrenceType}
   */
  public void setCot(CoocurrenceType cot);

  /**
   * Returns the CoocurrenceContext frequency (COF).
   * @return the CoocurrenceContext frequency
   */
  public int getCoocurrenceContextFrequency();

  /**
   * Returns the CoocurrenceContext frequency (COF).
   * This is shorthand for {@link #getCoocurrenceContextFrequency()}.
   * @return the CoocurrenceContext frequency
   */
  public int getCof();

  /**
   * Sets the CoocurrenceContext frequency (COF).
   * @param cof the CoocurrenceContext frequency
   */
  public void setCoocurrenceContextFrequency(int cof);

  /**
   * Sets the CoocurrenceContext frequency (COF).
   * This is shorthand for {@link #setCoocurrenceContextFrequency(int)}.
   * @param cof the CoocurrenceContext frequency
   */
  public void setCof(int cof);

  /**
   * Returns the {@link QualifierFrequency} list.
   * @return an array of {@link QualifierFrequency}s
   */
  public QualifierFrequency [] getQualifiers();

  /**
   * Returns the {@link QualifierFrequency} values as a {@link List}.
   * @return the {@link QualifierFrequency} values as a {@link List}
   */
  public List getQualifierAsList();

  /**
   * Replaces the {@link QualifierFrequency} values with those in the
   * specified array.
   * @param coa an array of {@link QualifierFrequency} values
   */
  public void setQualifiers(QualifierFrequency [] coa);

  /**
   * Replaces the {@link QualifierFrequency} values with those in the
   * specified {@link List}
   * @param coa a {@link List} of {@link QualifierFrequency} values
   */
  public void setQualifiers(List coa);

  /**
   * Adds the specified {@link QualifierFrequency}.
   * @param qual the {@link QualifierFrequency} to be added
   */
  public void addQualifier(QualifierFrequency qual);

  /**
   * Removes the specified {@link QualifierFrequency}.
   * @param qual the {@link QualifierFrequency} to be removed
   */
  public void removeQualifier(QualifierFrequency qual);

  /**
   * Removes all {@link QualifierFrequency}s.
   */
  public void clearQualifiers();

  /**
   * Returns the {@link CoocurrenceGroup} that holds this {@link CoocurrenceContext}.
   * @return the {@link CoocurrenceGroup} that holds this {@link CoocurrenceContext}
   */
  public CoocurrenceGroup getCoocurrenceGroup();

  /**
   * Sets the {@link CoocurrenceGroup} that holds this {@link CoocurrenceContext}.
   * @param group the {@link CoocurrenceGroup} that holds this {@link CoocurrenceContext}
   */
  public void setCoocurrenceGroup(CoocurrenceGroup group);

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from key {@link Atom} fields
   */
  public String getKeyFields();
}