/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Interface:   Locator.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;

/**
 * Represents information about a locator, which provides linkings between
 * Metathesaurus concepts and external databases.
 * <p>
 * Following is a UML Class diagram showing the neighborhood around {@link Locator}
 * <p><center><img src="Locator.gif"></center>
 * <p>
 * All of the data in the <code>MRLO</code> file is represented here.
 * <p>
 * @author Brian Carlsen, Deborah Shapiro
 */
public interface Locator
  extends AtomElement, AttributeElement, Comparable,
  ConceptElement, ContentViewElement, SourceElement {

  /**
   * Returns the frequency (FR).
   * @return the frequency
   * @see #getFrequencyMeaning()
   */
  public int getFrequency();

  /**
   * Returns the frequency (FR).  This is shorthand for {@link #getFrequency()}.
   * @return the frequency
   */
  public int getFr();

  /**
   * Sets the frequency (FR).
   * @param fr the frequency
   */
  public void setFrequency(int fr);

  /**
   * Set the frequency (FR).
   * This is shorthand for {@link #setFrequency(int)}.
   * @param fr the frequency
   */
  public void setFr(int fr);

  /**
   * Returns the meaning of the frequency (UN), if the frequency is set.
   * This is shorthand for {@link #getFrequencyMeaning()}.
   * @return the meaning of the frequency, if the frequency is set.
   */
  public String getUn();

    /**
   * Returns the meaning of the frequency (UN), if the frequency is set.
   * @return the meaning of the frequency, if the frequency is set.
   */
  public String getFrequencyMeaning();

  /**
   * Sets the meaning of the frequency (UN), if the frequency is set.
   * This is shorthand for {@link #setFrequencyMeaning(String)}.
   * @param un the meaning of the frequency, if the frequency is set.
   */
  public void setUn(String un);

  /**
   * Returns the meaning of the frequency (UN), if the frequency is set.
   * @param un the meaning of the frequency, if the frequency is set.
   */
  public void setFrequencyMeaning(String un);


  /**
   * Returns the actual name that occurs in the information source (SNA)
   * if not otherwise present in the Metathesaurus.
   * @return the actual name that occurs in the information source
   */
  public String getExternalName();

  /**
   * Returns the actual name that occurs in the information source (SNA)
   * if not otherwise present in the Metathesaurus. This is
   * shorthand for {@link #getExternalName()}.
   * @return the actual name that occurs in the information source
   */
  public String getSna();

  /**
   * Sets the actual name that occurs in the information source (SNA)
   * if not otherwise present in the Metathesaurus.
   * @param sna the actual name that occurs in the information source
   */
  public void setExternalName(String sna);

  /**
   * Sets the actual name that occurs in the information source (SNA)
   * if not otherwise present in the Metathesaurus. This is
   * shorthand for {@link #setExternalName(String)}.
   * @param sna the actual name that occurs in the information source
   */
  public void setSna(String sna);

  /**
   * Returns the record identifier (SOUI) corresponding to this {@link Locator}
   * in the actual information source.  This is shorthand for
   * {@link #getRecordIdentifier()}.
   * @return the record identifier
   */
  public String getSoui();

  /**
   * Returns the record identifier (SOUI) corresponding to this {@link Locator}
   * in the actual information source.
   * @return the record identifier
   */
  public String getRecordIdentifier();

  /**
   * Sets the record identifier (SOUI) corresponding to this {@link Locator}
   * in the actual information source.
   * @param soui the record identifier
   */
  public void setSoui(String soui);

  /**
   * Sets the record identifier (SOUI) corresponding to this {@link Locator}
   * in the actual information source.  This is shorthand for
   * {@link #setRecordIdentifier(String)}.
   * @param soui the record identifier
   */
  public void setRecordIdentifier(String soui);

  /**
   * Return the string identifier (SUI).
   * This is shorthand for {@link #getStringIdentifier()}.
   * @return the SUI
   */
  public String getSui();

  /**
   * Sets the string identifier (SUI).
   * This is shorthand for {@link #setStringIdentifier(String)}.
   * @param sui the SUI
   */
  public void setSui(String sui);

  /**
   * Return the string identifier (SUI).
   * @return the SUI
   */
  public String getStringIdentifier();

  /**
   * Sets the string identifier (SUI).
   * @param sui the SUI
   */
  public void setStringIdentifier(String sui);

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from key {@link Locator} fields
   */
  public String getKeyFields();
}