/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Interface:   MapObject.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;

/**
 * Represents information about the "to" or "from" end of a mapping.
 * For example, a SNOMEDCT->ICD9CM mapping might have a {@link MapObject}
 * representing a SNOMEDCT concept identifier and a {@link MapObject}
 * representing a list of ICD9CM codes.
 * <p>
 * Following is a UML Class diagram showing the neighborhood around {@link MapObject}
 * <p><center><img src="MapObject.gif"></center>
 * <p>
 * This is used in conjunction with {@link Mapping} to represent all
 * of the data in a row of the <code>MRMAP</code> file.
 * <p>
 * @author Brian Carlsen, Deborah Shapiro
 */
public interface MapObject extends SourceElement, Comparable {

  /**
   * Returns the source asserted unique identifier (FROMSID or TOSID).
   * @return the source asserted unique identifier
   */
  public String getSourceIdentifier();

  /**
   * Sets the source asserted unique identifier (FROMSID or TOSID).
   * @param sid source asserted unique identifier
   */
  public void setSourceIdentifier(String sid);

  /**
   * Returns the source asserted unique identifier (FROMSID or TOSID).
   * This method is shorthand for {@link #getSourceIdentifier()}.
   * @return the source asserted unique identifier
   */
  public String getSid();

  /**
   * Sets the source asserted unique identifier (FROMSID or TOSID).
   * This method is shorthand for {@link #setSourceIdentifier(String)}.
   * @param sid source asserted unique identifier
   */
  public void setSid(String sid);

  /**
   * Returns the unique identifier (FROMID or TOID).
   * @return the unique identifier
   */
  public String getIdentifier();

  /**
   * Sets the unique identifier (FROMID or TOID).
   * @param id unique identifier
   */
  public void setIdentifier(String id);

  /**
   * Returns the unique identifier (FROMID or TOID).
   * This method is shorthand for {@link #getIdentifier()}.
   * @return the unique identifier
   */
  public String getId();

  /**
   * Sets the unique identifier (FROMID or TOID).
   * This method is shorthand for {@link #setIdentifier(String)}.
   * @param id unique identifier
   */
  public void setId(String id);


  /**
   * Returns the expression (FROMEXPR or TOEXPR).
   * @return the expression
   */
  public String getExpression();

  /**
   * Sets the expression (FROMEXPR or TOEXPR).
   * @param expr expression
   */
  public void setExpression(String expr);

  /**
   * Returns the expression (FROMEXPR or TOEXPR).
   * This method is shorthand for {@link #getExpression()}.
   * @return the expression
   */
  public String getExpr();

  /**
   * Sets the expression (FROMEXPR or TOEXPR).
   * This method is shorthand for {@link #setExpression(String)}.
   * @param expr expression
   */
  public void setExpr(String expr);

  /**
   * Returns the type {@link SourceIdentifierType} (FROMTYPE or TOTYPE).
   * @return the type
   */
  public SourceIdentifierType getType();

  /**
   * Sets the type (FROMTYPE or TOTYPE).
   * @param type {@link SourceIdentifierType}
   */
  public void setType(SourceIdentifierType type);

  /**
   * Returns the machine processable rule for when mapping
   * is valid (FROMRULE or TORULE).
   * @return the rule
   */
  public String getRule();

  /**
   * Sets the machine processable rule for when mapping is
   * valid (FROMRULE or TORULE).
   * @param rule the rule
   */
  public void setRule(String rule);

  /**
   * Returns the restriction (FROMRES or TORES).
   * @return the restriction
   */
  public String getRestriction();

  /**
   * Sets the restriction (FROMRES or TORES).
   * @param res restriction
   */
  public void setRestriction(String res);

  /**
   * Returns the restriction (FROMRES or TORES).
   * This method is shorthand for {@link #getRestriction()}.
   * @return the restriction
   */
  public String getRes();

  /**
   * Sets the restriction (FROMRES or TORES).
   * This method is shorthand for {@link #setRestriction(String)}.
   * @param res restriction
   */
  public void setRes(String res);

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from key {@link Mapping} fields
   */
  public String getKeyFields();
}