/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Interface:   Mapping.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;

/**
 * Represents information about a mapping from one kind of
 * UMLS data to another.  This could be cross-source mappings
 * like those that appear in SNOMED, or it could be associated
 * expressions.
 * <p>
 * Following is a UML Class diagram showing the neighborhood around {@link Mapping}
 * <p><center><img src="Mapping.gif"></center>
 * <p>
 * All data in a row of the <code>MRMAP</code> file is represented by this
 * and by {@link MapObject}.
 * <p>
 * @author Brian Carlsen, Deborah Shapiro
 */
public interface Mapping
    extends ConceptElement, ContentViewElement, SourceElement, Comparable {

  /**
   * Returns the {@link AdditionalRelationshipLabel} (RELA).
   * @return the {@link AdditionalRelationshipLabel}
   */
  public AdditionalRelationshipLabel getAdditionalRelationshipLabel();

  /**
   * Returns the {@link AdditionalRelationshipLabel} (RELA).
   * This is shorthand for {@link #getAdditionalRelationshipLabel()}.
   * @return the {@link AdditionalRelationshipLabel}
   */
  public AdditionalRelationshipLabel getRela();

  /**
   * Sets the {@link AdditionalRelationshipLabel} (RELA).
   * @param rela the {@link AdditionalRelationshipLabel}
   */
  public void setAdditionalRelationshipLabel(AdditionalRelationshipLabel rela);

  /**
   * Sets the {@link AdditionalRelationshipLabel} (RELA).
   * This is shorthand for
   * {@link #setAdditionalRelationshipLabel(AdditionalRelationshipLabel)}.
   * @param rela the {@link AdditionalRelationshipLabel}
   */
  public void setRela(AdditionalRelationshipLabel rela);

  /**
   * Returns the subset identifier (SUBSETUI).
   * @return the subset identifier
   */
  public String getSubsetIdentifier();

  /**
   * Sets the subset identifier (SUBSETUI).
   * @param subset subset identifier
   */
  public void setSubsetIdentifier(String subset);

  /**
   * Returns the subset identifier (SUBSETUI).
   * This method is shorthand for {@link #getSubsetIdentifier()}.
   * @return the subset identifier
   */
  public String getSubsetId();

  /**
   * Sets the subset identifier (SUBSETUI).
   * This method is shorthand for {@link #setSubsetIdentifier(String)}.
   * @param subset subset identifier
   */
  public void setSubsetId(String subset);

  /**
   * Returns the rank (MAPRANK).
   * @return the rank
   */
  public String getRank();

  /**
   * Sets the rank (MAPRANK).
   * @param rank priority
   */
  public void setRank(String rank);

  /**
   * Returns the source asserted unique identifier (MAPSID).
   * @return the source asserted unique identifier
   */
  public String getSourceIdentifier();

  /**
   * Sets the source asserted unique identifier (MAPSID).
   * @param sid source asserted unique identifier
   */
  public void setSourceIdentifier(String sid);

  /**
   * Returns the source asserted unique identifier (MAPSID).
   * This method is shorthand for {@link #getSourceIdentifier()}.
   * @return the source asserted unique identifier
   */
  public String getSid();

  /**
   * Sets the source asserted unique identifier (MAPSID).
   * This method is shorthand for {@link #setSourceIdentifier(String)}.
   * @param sid source asserted unique identifier
   */
  public void setSid(String sid);

  /**
   * Returns the unique identifier (MAPID).
   * @return the unique identifier
   */
  public String getIdentifier();

  /**
   * Sets the unique identifier (MAPID).
   * @param id unique identifier
   */
  public void setIdentifier(String id);

  /**
   * Returns the unique identifier (MAPID).
   * This method is shorthand for {@link #getIdentifier()}.
   * @return the unique identifier
   */
  public String getId();

  /**
   * Sets the unique identifier (MAPID).
   * This method is shorthand for {@link #setIdentifier(String)}.
   * @param id unique identifier
   */
  public void setId(String id);

  /**
   * Returns the {@link RelationshipLabel} (REL).
   * @return the {@link RelationshipLabel}
   */
  public RelationshipLabel getRelationshipName();

  /**
   * Returns the {@link RelationshipLabel} (REL).
   * This is shorthand for {@link #getRelationshipName()}.
   * @return the {@link RelationshipLabel}
   */
  public RelationshipLabel getRel();

  /**
   * Sets the {@link RelationshipLabel} (REL).
   * @param rel the {@link RelationshipLabel}
   */
  public void setRelationshipName(RelationshipLabel rel);

  /**
   * Sets the {@link RelationshipLabel} (REL).
   * This is shorthand for {@link #setRelationshipName(RelationshipLabel)}.
   * @param rel the {@link RelationshipLabel}
   */
  public void setRel(RelationshipLabel rel);

  /**
   * Returns the machine processable rule for when to apply mapping (MAPRULE).
   * @return the map rule
   */
  public String getRule();

  /**
   * Sets the machine processable rule for when to apply mapping (MAPRULE).
   * @param rule {@link String}
   */
  public void setRule(String rule);

  /**
   * Returns the human readable rule for use of mapping (MAPRES).
   * @return the map restriction
   */
  public String getRestriction();

  /**
   * Sets the human readable rule for use of mapping (MAPRES).
   * @param res {@link String}
   */
  public void setRestriction(String res);

  /**
   * Returns the human readable rule for use of mapping (MAPRES).
   * This is shorthand for {@link #getRestriction()}.
   * @return the human readable rule for use of mapping
   */
  public String getRes();

  /**
   * Sets the human readable rule for use of mapping (MAPRES).
   * This is shorthand for {@link #setRestriction(String)}.
   * @param res the human readable rule for use of mapping
   */
  public void setRes(String res);

  /**
   * Returns the type of mapping (MAPTYPE).
   * @return the type of mapping
   */
  public String getType();

  /**
   * Sets the type of mapping (MAPTYPE).
   * @param type {@link String}
   */
  public void setType(String type);

  /**
   * Returns the mapped from {@link MapObject}.
   * @return the mapped from {@link MapObject}
   */
  public MapObject getFrom();

  /**
   * Sets the mapped from {@link MapObject}.
   * @param from the mapped from {@link MapObject}
   */
  public void setFrom(MapObject from);

  /**
   * Returns the mapped to {@link MapObject}.
   * @return the mapped to {@link MapObject}
   */
  public MapObject getTo();

  /**
   * Sets the mapped to {@link MapObject}.
   * @param to the mapped to {@link MapObject}
   */
  public void setTo(MapObject to);

  /**
   * Returns the {@link AttributeName} (MAPATN).
   * @return the {@link AttributeName}
   */
  public AttributeName getAttributeName();

  /**
   * Sets the {@link AttributeName} (MAPATN).
   * @param atn the {@link AttributeName}
   */
  public void setAttributeName(AttributeName atn);

  /**
   * Returns the attribute value (MAPATV).
   * @return the attribute value
   */
  public String getAttributeValue();

  /**
   * Sets the attribute value (MAPATV).
   * @param atv the attribute value
   */
  public void setAttributeValue(String atv);

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from key {@link Mapping} fields
   */
  public String getKeyFields();
}