/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Class:       MetaDescriptionInitializer.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;

import gov.nih.nlm.umls.meta.impl.MetaDescriptionImpl;
import gov.nih.nlm.umls.meta.impl.STYMetaDescriptionImpl;
import gov.nih.nlm.umls.meta.impl.SourceMetaDescriptionImpl;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;

/**
 * Provides an easy mechanism for initializing the system from MRDOC, MRSAB,
 * and SRDEF.  This is a convenience class to help users avoid the complexities
 * of initializing an application.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public final class MetaDescriptionInitializer   {
  /**
   * Instantiates an empty {@link MetaDescriptionInitializer}.
   */
  public MetaDescriptionInitializer() {};

  /**
   * Initializes all of the {@link MetaDescription}s at one time.
   * @param meta_dsc an array of {@link MetaDescription}s
   * @throws IllegalStateException if an attempt is made to call this
   * method more than once
   */
  public static synchronized void initialize(MetaDescription[] meta_dsc) {
    AttributeName.initialize(meta_dsc);
    CoocurrenceType.initialize(meta_dsc);
    Language.initialize(meta_dsc);
    AdditionalRelationshipLabel.initialize(meta_dsc);
    RelationshipLabel.initialize(meta_dsc);
    SourceIdentifierType.initialize(meta_dsc);
    StringType.initialize(meta_dsc);
    Suppressible.initialize(meta_dsc);
    TermStatus.initialize(meta_dsc);
    TermType.initialize(meta_dsc);
    TermTypeGroup.initialize(meta_dsc);
  }

  /**
   * Initializes all of the {@link STYMetaDescription}s at one time.
   * @param sty_meta_dsc an array of {@link STYMetaDescription}s
   * @throws IllegalStateException if an attempt is made to call this
   * method more than once
   */
  public static synchronized void initialize(STYMetaDescription[] sty_meta_dsc) {
    HierarchicalSemanticType.initialize(sty_meta_dsc);
  }

  /**
   * Initializes all of the {@link SourceMetaDescription}s at one time.
   * @param source_meta_dsc an array of {@link SourceMetaDescription}s
   * @throws IllegalStateException if an attempt is made to call this
   * method more than once
   */
  public static synchronized void initialize(SourceMetaDescription[] source_meta_dsc) {
    Source.initialize(source_meta_dsc);
  }

  /**
   * Initializes any or all of the following files: MRDOC, MRSAB, SRDEF.
   * BAckwards compatability.
   * @param mrdoc {@link File} MRDOC
   * @param mrsab {@link File} MRSAB
   * @param srdef {@link File} SRDEF
   * @throws IOException
   */
  public static synchronized void initialize(File mrdoc, File mrsab, File srdef)
    throws IOException{
    initialize(mrdoc,mrsab,srdef,null);
  }

  /**
   * Initializes any or all of the following files: MRDOC, MRSAB, SRDEF.
   * @param mrdoc {@link File} MRDOC
   * @param mrsab {@link File} MRSAB
   * @param srdef {@link File} SRDEF
   * @param sty_groups {@link File} SemGroups.txt
   * @throws IOException
   */
  public static synchronized void initialize(File mrdoc, File mrsab,
                                             File srdef, File sty_groups)
    throws IOException{

    String line = null;

    if (mrdoc != null) {
      ArrayList mdi = new ArrayList();
      BufferedReader in = new BufferedReader(new FileReader(mrdoc));

      while ((line = in.readLine()) != null) {
         mdi.add(new MetaDescriptionImpl(line));
      }
      initialize((MetaDescription[])mdi.toArray(new MetaDescription[0]));
    }

    if (sty_groups != null) {
      ArrayList mdi = new ArrayList();
      BufferedReader in = new BufferedReader(new FileReader(sty_groups));

      while ((line = in.readLine()) != null) {
         mdi.add(new MetaDescriptionImpl(line));
      }
      SemanticTypeGroup.initialize((MetaDescription[])mdi.toArray(new MetaDescription[0]));
    }

    if (mrsab != null) {
      ArrayList smd = new ArrayList();
      BufferedReader in = new BufferedReader(new FileReader(mrsab));

      while ((line = in.readLine()) != null) {
        smd.add(new SourceMetaDescriptionImpl(line));
      }
      initialize((SourceMetaDescription[])smd.toArray(new SourceMetaDescription[0]));
    }
    if (srdef != null) {
      ArrayList stymd = new ArrayList();
      BufferedReader in = new BufferedReader(new FileReader(srdef));

      while ((line = in.readLine()) != null) {
        stymd.add(new STYMetaDescriptionImpl(line));
      }
      initialize((STYMetaDescription[])stymd.toArray(new STYMetaDescription[0]));
    }
  }
}
