/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Interface:   Relationship.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;

import java.util.List;

/**
 * Represents a relationship between two {@link Concept}s or two
 * {@link Atom}s.
 * <p>
 * Following is a UML Class diagram showing the neighborhood around {@link Relationship}
 * <p><center><img src="Relationship.gif"></center>
 * <p>
 * All data in a row of the <code>MRREL</code> file is represented here.
 * <p>
 * @author Brian Carlsen, Deborah Shapiro
 */
public interface Relationship
  extends Comparable, ConceptElement, ContentViewElement,
  RelatedConceptElement, RelatedSourceIdentifierElement,
  SourceElement, SourceIdentifierElement, SuppressibleElement {

  /**
   * Returns the {@link RelationshipLabel} (REL).
   * @return the {@link RelationshipLabel}
   */
  public RelationshipLabel getRelationshipLabel();

  /**
   * Returns the {@link RelationshipLabel} (REL).
   * This is shorthand for {@link #getRelationshipLabel()}.
   * @return the {@link RelationshipLabel}
   */
  public RelationshipLabel getRel();

  /**
   * Sets the {@link RelationshipLabel} (REL).
   * @param rel the {@link RelationshipLabel}
   */
  public void setRelationshipLabel(RelationshipLabel rel);

  /**
   * Sets the {@link RelationshipLabel} (REL).
   * This is shorthand for {@link #setRelationshipLabel(RelationshipLabel)}.
   * @param rel the {@link RelationshipLabel}
   */
  public void setRel(RelationshipLabel rel);

  /**
   * Returns the {@link AdditionalRelationshipLabel} (RELA).
   * @return the {@link AdditionalRelationshipLabel}
   */
  public AdditionalRelationshipLabel getAdditionalRelationshipLabel();

  /**
   * Returns the {@link AdditionalRelationshipLabel} (RELA).
   * This is shorthand for {@link #getAdditionalRelationshipLabel()}.
   * @return the {@link AdditionalRelationshipLabel}
   */
  public AdditionalRelationshipLabel getRela();

  /**
   * Sets the {@link AdditionalRelationshipLabel} (RELA).
   * @param rela the {@link AdditionalRelationshipLabel}
   */
  public void setAdditionalRelationshipLabel(AdditionalRelationshipLabel rela);

  /**
   * Sets the {@link AdditionalRelationshipLabel} (RELA).
   * This is shorthand for
   * {@link #setAdditionalRelationshipLabel(AdditionalRelationshipLabel)}.
   * @param rela the {@link AdditionalRelationshipLabel}
   */
  public void setRela(AdditionalRelationshipLabel rela);

  /**
   * Returns the source asserted relationship identifier (SRUI).
   * @return the source asserted relationship identifier
   */
  public String getSourceRelationshipIdentifier();

  /**
   * Returns the source asserted relationship identifier (SRUI).
   * This is shorthand for {@link #getSourceRelationshipIdentifier()}.
   * @return the source asserted relationship identifier
   */
  public String getSrui();

  /**
   * Sets the source asserted relationship identifier (SRUI).
   * @param srui the source asserted relationship identifier
   */
  public void setSourceRelationshipIdentifier(String srui);

  /**
   * Sets the source asserted relationship identifier (SRUI).
   * This is shorthand for {@link #setSourceRelationshipIdentifier(String)}.
   * @param srui the source asserted relationship identifier
   */
  public void setSrui(String srui);

  /**
   * Returns the source of the relationship label (SL).
   * @return the source of the relationship label
   */
  public Source getSourceOfLabel();

  /**
   * Returns the source of the relationship label (SL).
   * This is shorthand for {@link #getSourceOfLabel()}.
   * @return the source of the relationship label
   */
  public Source getSl();

  /**
   * Sets the source of the relationship label (SL).
   * @param sl the source of the relationship label
   */
  public void setSourceOfLabel(Source sl);

  /**
   * Sets the source of the relationship label (SL).
   * This is shorthand for {@link #setSourceOfLabel(Source)}.
   * @param sl the source of the relationship label
   */
  public void setSl(Source sl);

  /**
   * Returns the relationship group (RG).
   * @return the relationship group
   */
  public String getRelationshipGroup();

  /**
   * Returns the relationship group (RG).
   * This is shorthand for {@link #getRelationshipGroup()}.
   * @return the relationship group
   */
  public String getRg();

  /**
   * Sets the relationship group (RG).
   * @param rg the relationship group
   */
  public void setRelationshipGroup(String rg);

  /**
   * Sets the relationship group (RG).
   * This is shorthand for {@link #getRelationshipGroup()}.
   * @param rg the relationship group
   */
  public void setRg(String rg);

  /**
   * Sets the directionality value (DIR).
   * @param dir the directionality value
   */
  public void setDir(String dir);

  /**
   * Returns the directionality value (DIR).
   * @return the directionality value
   */
  public String getDir();

  /**
   * Indicates the directionality of this {@link Relationship}.
   * @return <code>true</code> if directionality is in order read,
   *         <code>false</code> if directionality is reversed
   */
  public boolean isAssertedDirection();

  /**
   * Sets the flag indicating the directionality of the {@link Relationship}.
   * @param dir <code>true</code> if directionality is in order read,
   *         <code>false</code> if directionality is reversed
   */
  public void setIsAssertedDirection(boolean dir);

  /**
   * Shorthand for {@link #getRelationshipIdentifier()}.
   * @return the RUI
   */
  public String getRui();

  /**
   * Shorthand for {@link #setRelationshipIdentifier(String)}.
   * @param rui the RUI
   */
  public void setRui(String rui);

  /**
   * Returns the Relationship unique identifier (RUI) related to this
   * {@link Relationship}.
   * @return the RUI
   */
  public String getRelationshipIdentifier();

  /**
   * Sets the Relationship unique identifier (RUI) related to this
   * {@link Relationship}.
   * @param rui the RUI
   */
  public void setRelationshipIdentifier(String rui);

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from key {@link Relationship} fields
   */
  public String getKeyFields();

  //
  // RelationshipAttributes
  //

  /**
   * Returns the {@link RelationshipAttribute}s.
   * @return a {@link RelationshipAttribute} <code>[]</code>
   */
  public RelationshipAttribute[] getAttributes();

  /**
   * Returns the {@link RelationshipAttribute}s as a {@link List}.
   * @return the {@link RelationshipAttribute}s as a {@link List}
   */
  public List getAttributesAsList();

  /**
   * Adds the specified {@link RelationshipAttribute}.
   * @param att the {@link RelationshipAttribute} to be added
   */
  public void addAttribute(RelationshipAttribute att);

  /**
   * Removes the specified {@link RelationshipAttribute}.
   * @param att the {@link RelationshipAttribute} to be removed
   */
  public void removeAttribute(RelationshipAttribute att);

  /**
   * Remove all {@link RelationshipAttribute}s.
   */
  public void clearAttributes();

  /**
   * Replace {@link RelationshipAttribute}s with the ones in the specified array.
   * @param atts an {@link RelationshipAttribute} <code>[]</code>
   */
  public void setAttributes(RelationshipAttribute atts[]);

  /**
       * Replace {@link RelationshipAttribute}s with the ones in the specified {@link List}.
   * @param atts a {@link List} of {@link RelationshipAttribute}s
   */
  public void setAttributes(List atts);

  /**
   * Returns a subset of this {@link Relationship}'s {@link RelationshipAttribute}s.
   * Restricts the return based on the given {@link Restrictor}.
   * @param restrictor {@link Restrictor}
   * @return RelationshipAttribute[]
   */
  public RelationshipAttribute[] getAttributes(Restrictor restrictor);

}