/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Class:       SemanticTypeGroup.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Represents a grouping of {@link SemanticType}s. This class ensures that each unique
 * {@link SemanticType} group is represented by a singleton object.  In order for
 * this to work correctly, the class must be properly initialized and must be
 * initialized after Term Types is initialized.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public final class SemanticTypeGroup extends AbstractExpandedForm  {


  // Map by abbreviation to the actual classes
  private static final Map sty_groups = new HashMap();

  // stys in this group
  private Set stys = new HashSet();

  // Map by sty to the group(s) that sty belongs to
  private static Map sty_to_groups = new HashMap();

  // Abbreviation and description for this individual instance.
  private String abbrev;
  private String dsc;

  /**
   * Instantiates an empty {@link SemanticTypeGroup}.
   */
  public SemanticTypeGroup() {};

  /**
   * Instantiates an {@link SemanticTypeGroup} from the
   * specified {@link MetaDescription}.  Here {@link MetaDescription} is
   * being slightly abused because we are not reading this data from <tt>MRDOC.RRF</tt>,
   * instead we are reading it from <tt>SemGroups.txt</tt>.
   * @param meta_dsc a {@link MetaDescription} representing a term type
   */
  private SemanticTypeGroup(MetaDescription meta_dsc) {
    abbrev = meta_dsc.getKey();
    dsc = meta_dsc.getValue();
  }

  /**
   * Initializes the {@link Map} used to track all of the
   * {@link SemanticTypeGroup}s that exist at one time.  You may
   * pass all {@link MetaDescription}s as this method only
   * creates {@link SemanticTypeGroup}s for those with a type of "sty".
   * @param meta_dsc an array of {@link MetaDescription}s
   * @throws IllegalStateException if an attempt is made to call this
   * method more than once
   */
  public static synchronized void initialize(MetaDescription[] meta_dsc) {
    for (int i = 0; i < meta_dsc.length; i++) {
      SemanticTypeGroup sty_group = null;
      if (sty_groups.containsKey(meta_dsc[i].getKey())) {
        sty_group = (SemanticTypeGroup) sty_groups.get(meta_dsc[i].getKey());
      } else {
        sty_group = new SemanticTypeGroup(meta_dsc[i]);
        sty_groups.put(meta_dsc[i].getKey(), sty_group);
      }

      HierarchicalSemanticType sty = HierarchicalSemanticType.
          getHierarchicalSemanticType(meta_dsc[i].getExplain());
      sty_group.stys.add(sty);

      Set groups = (Set) sty_to_groups.get(sty);
      if (groups == null) {
        groups = new HashSet();
        sty_to_groups.put(sty, groups);
      }
      groups.add(sty_group);
    }
  }

  /**
   * Initializes the {@link Map} used to track all of the
   * {@link SemanticTypeGroup}s that exist at one time.  This
   * is implemented as a call to {@link #initialize(MetaDescription[])}.
   * @param meta_dsc an array of {@link MetaDescription}s
   */
  public static synchronized void initialize(List meta_dsc) {
    initialize((MetaDescription[])meta_dsc.toArray(new MetaDescription[0]));
  }

  /**
   * Returns the {@link SemanticTypeGroup} for the specified abbreviation.
   * This method will return <code>null</code> if the <code>initialize</code>
   * method has not been called or if the specified abbreviation
   * was not one of those used during initialization.
   * @param sty_group the term type abbreviation to look up
   * @return the {@link SemanticTypeGroup} for the specified abbreviation
   */
  public static SemanticTypeGroup getSemanticTypeGroup(String sty_group) {
    return (SemanticTypeGroup)sty_groups.get(sty_group);
  }

  /**
   * Return all {@link SemanticTypeGroup}s for which the given {@link SemanticType}
   * is a member.
   * @param sty {@link SemanticType}
   * @return an array of {@link SemanticTypeGroup}s
   */
  public static SemanticTypeGroup[] getSemanticTypeGroups(SemanticType sty) {
    return (SemanticTypeGroup[])((Set)sty_to_groups.get(sty)).toArray(new SemanticTypeGroup[0]);
  }

  /**
   * Return all {@link SemanticTypeGroup}s for which the given {@link SemanticType} is
   * a member as a {@link List}.
   * @param sty {@link SemanticType}
   * @return all {@link SemanticTypeGroup}s as a {@link List}
   */
  public static List getSemanticTypeGroupsAsList(SemanticType sty) {
    return new ArrayList(((Set)sty_to_groups.get(sty)));
  }

  /**
   * Return all {@link SemanticTypeGroup}s.
   * @return an array of {@link SemanticTypeGroup}s
   */
  public static SemanticTypeGroup[] getSemanticTypeGroups() {
    return (SemanticTypeGroup[])sty_groups.values().toArray(new SemanticTypeGroup[0]);
  }

  /**
   * Return all {@link SemanticTypeGroup}s as a {@link List}.
   * @return all {@link SemanticTypeGroup}s as a {@link List}
   */
  public static List getSemanticTypeGroupsAsList() {
    return new ArrayList(sty_groups.values());
  }

  /**
   * Return all {@link SemanticType}s in this {@link SemanticTypeGroup}.
   * @return an array of {@link SemanticTypeGroup}s
   */
  public SemanticType[] getSemanticTypes() {
    return (SemanticType[])stys.toArray(new SemanticType[0]);
  }

  /**
   * Return all {@link SemanticType}s in this {@link SemanticTypeGroup} as a {@link List}.
   * @return all {@link SemanticType}s as a {@link List}
   */
  public List getSemanticTypesAsList() {
    return new ArrayList(stys);
  }

  /**
   * Returns the abbreviation.
   * @return the abbreviation
   */
  public String getAbbreviation() {
    return abbrev;
  }

  /**
   * Returns the description
   * @return the description
   */
  public String getDescription() {
    return dsc;
  }

}
