/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Object:      Source.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

/**
 * Represents a UMLS source. This class ensures that each unique source is
 * represented by a singleton object.  In order for this to work correctly,
 * the class must be properly initialized.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public final class Source implements SourceMetaDescription {

  // Map by abbreviation to the actual classes
  private static final Map vsabs = new HashMap(200);
  private static final Map rsabs = new HashMap(200);

  // Abbreviation and description for this individual instance.
  private String cenc, cxty, curver, imeta, rcui, rmeta, rsab,
    scc, sf, slc, son, ssn, sver, vsab, vcui, srl, scit;

  private int cfr, tfr;
  private Language lat;
  private Date mend, mstart;
  private List atnl, ttyl;

  // This field is calculated at release time and when MMSYS runs,
  // so it is mutable.
  private String sabin;

  /**
   * Instantiates an empty {@link Source}.
   */
  public Source() {};

  /**
   * Instantiates an {@link Source} from the
   * specified {@link SourceMetaDescription}.
   * @param md a {@link SourceMetaDescription} representing an source
   */
  private Source(SourceMetaDescription md) {
    cenc = md.getCharacterEncoding();
    curver = md.getCurver();
    cxty = md.getContextType();
    imeta = md.getInsertMetaVersion();
    rcui = md.getRootConceptIdentifier();
    rmeta = md.getRemoveMetaVersion();
    rsab = md.getRootSourceAbbreviation();
    sabin = md.getSabin();
    scc = md.getContentContact();
    sf = md.getSourceFamily();
    slc = md.getLicenseContact();
    son = md.getOfficialName();
    srl = md.getRestrictionLevel();
    ssn = md.getShortName();
    sver = md.getVersion();
    vcui = md.getVersionedConceptIdentifier();
    vsab = md.getVersionedSourceAbbreviation();
    lat = md.getLanguage();
    cfr = md.getConceptFrequency();
    tfr = md.getTermFrequency();
    mend = md.getValidEndDate();
    mstart = md.getValidStartDate();
    atnl = md.getAttributeNamesAsList();
    ttyl = md.getTermTypesAsList();
    scit = md.getCitation();
  }

  /**
   * Initializes the {@link Map} used to track all of the
   * {@link Source}s that exist at one time.
   * @param meta_dsc an array of {@link SourceMetaDescription}s
   */
  public static synchronized void initialize(SourceMetaDescription[] meta_dsc) {
    for (int i=0; i<meta_dsc.length; i++) {
	if (vsabs.containsKey(meta_dsc[i].getVersionedSourceAbbreviation())) {
          Source smd = (Source)vsabs.get(meta_dsc[i].getVersionedSourceAbbreviation());
          smd.cenc = meta_dsc[i].getCharacterEncoding();
          smd.curver = meta_dsc[i].getCurver();
          smd.cxty = meta_dsc[i].getContextType();
          smd.imeta = meta_dsc[i].getInsertMetaVersion();
          smd.rcui = meta_dsc[i].getRootConceptIdentifier();
          smd.rmeta = meta_dsc[i].getRemoveMetaVersion();
          smd.rsab = meta_dsc[i].getRootSourceAbbreviation();
          smd.sabin = meta_dsc[i].getSabin();
          smd.scc = meta_dsc[i].getContentContact();
          smd.sf = meta_dsc[i].getSourceFamily();
          smd.slc = meta_dsc[i].getLicenseContact();
          smd.son = meta_dsc[i].getOfficialName();
          smd.srl = meta_dsc[i].getRestrictionLevel();
          smd.ssn = meta_dsc[i].getShortName();
          smd.sver = meta_dsc[i].getVersion();
          smd.vcui = meta_dsc[i].getVersionedConceptIdentifier();
          smd.lat = meta_dsc[i].getLanguage();
          smd.cfr = meta_dsc[i].getConceptFrequency();
          smd.tfr = meta_dsc[i].getTermFrequency();
          smd.mend = meta_dsc[i].getValidEndDate();
          smd.mstart = meta_dsc[i].getValidStartDate();
          smd.atnl = meta_dsc[i].getAttributeNamesAsList();
          smd.ttyl = meta_dsc[i].getTermTypesAsList();
          smd.scit = meta_dsc[i].getCitation();
        } else {
          Source src = new Source(meta_dsc[i]);
          vsabs.put(src.getVersionedSourceAbbreviation(), src);
          if (meta_dsc[i].inCurrentSubset())
            rsabs.put(src.getRootSourceAbbreviation(), src);
        }
    }
  }

  /**
   * Initializes the {@link Map} used to track all of the
   * {@link Source}s that exist at one time.  This
   * is implemented as a call to {@link #initialize(SourceMetaDescription[])}.
   * @param meta_dsc an array of {@link SourceMetaDescription}s
   * @throws IllegalStateException if an attempt is made to call this
   * method more than once
   */
  public static synchronized void initialize(List meta_dsc) {
    initialize((SourceMetaDescription[])meta_dsc.toArray(new SourceMetaDescription[0]));
  }

  /**
   * Returns the {@link Source} for the specified versioned source abbreviation.
   * This method will return <code>null</code> if the <code>initialize</code>
   * method has not been called or if the specified abbreviation
   * was not one of those used during initialization.
   * @param sab the specified versioned or versionless source abbreviation
   * @return the {@link Source} for the specified abbreviation
   */
  public static Source getSource(String sab) {

    Source s = (Source)rsabs.get(sab);
    if(s != null)
      return s;
    s = (Source)vsabs.get(sab);
    if(s != null)
      return s;
    return null;
  }

  /**
   * Returns all {@link Source}s.
   * @return an array of {@link Source}
   */
  public static Source[] getSources() {
    return (Source[])vsabs.values().toArray(new Source[0]);
  }

  /**
   * Returns all {@link Source}s as a {@link List}.
   * @return all {@link Source}s as a {@link List}
   */
  public static List getSourcesAsList() {
    return new ArrayList(vsabs.values());
  }


  /**
   * Returns the concept identifier for the versioned SRC concept (VCUI).
   * This is shorthand for {@link #getVersionedConceptIdentifier()}.
   * @return the concept identifier for the versioned SRC concept
   */
  public String getVcui() { return getVersionedConceptIdentifier(); }

  /**
   * Returns the concept identifier for the versioned SRC concept (VCUI).
   * @return the concept identifier for the versioned SRC concept
   */
  public String getVersionedConceptIdentifier() { return vcui; }

  /**
   * Returns the concept identifier for the root SRC concept (VCUI).
   * This is shorthand for {@link #getRootConceptIdentifier()}.
   * @return the concept identifier for the root SRC concept
   */
  public String getRcui() { return getRootConceptIdentifier(); }

  /**
   * Returns the concept identifier for the root SRC concept (VCUI).
   * @return the concept identifier for the root SRC concept
   */
  public String getRootConceptIdentifier() { return rcui; }

  /**
   * Returns the versioned source abbreviation (VSAB).
   * This is shorthand for (@ link #getVersionedSourceAbbreviation()}.
   * @return the versioned source abbreviation
   */
  public String getVsab() { return getVersionedSourceAbbreviation(); }

  /**
   * Returns the versioned source abbreviation (VSAB).
   * @return the versioned source abbreviation
   */
  public String getVersionedSourceAbbreviation() { return vsab; }

  /**
   * Returns the root source abbreviation (RSAB).
   * This is shorthand for {@link #getRootSourceAbbreviation()}.
   * @return the root source abbreviation
   */
  public String getRsab() { return getRootSourceAbbreviation(); }

  /**
   * Returns the root source abbreviation (RSAB).
   * @return the root source abbreviation
   */
  public String getRootSourceAbbreviation() { return rsab; }

  /**
   * Returns the source official name (SON).
   * This is shorthand for {@link #getOfficialName()}.
   * @return the source official name
   */
  public String getSon() { return getOfficialName(); }

  /**
   * Returns the source official name (SON).
   * @return the source official name
   */
  public String getOfficialName() { return son; }

  /**
   * Returns the source short name (SSN).
   * This is shorthand for {@link #getShortName()}.
   * @return the source short name
   */
  public String getSsn() { return getShortName(); }

  /**
   * Returns the source short name (SSN).
   * @return the source short name
   */
  public String getShortName() { return ssn; }

  /**
   * Returns the source family (SF).
   * This is shorthand for {@link #getSourceFamily()}.
   * @return the source family
   */
  public String getSf() { return getSourceFamily(); }

  /**
   * Returns the source family (SF).
   * @return the source family
   */
  public String getSourceFamily() { return sf; }

  /**
   * Returns source version (SVER).
   * This is shorthand for {@link #getVersion()}.
   * @return the source version
   */
  public String getSver() { return getVersion(); }

  /**
   * Returns source version (SVER).
   * @return the source version
   */
  public String getVersion() { return sver; }

  /**
   * Returns the date when this source becomes valid in the Metathesaurus (MSTART).
   * This is shorthand for {@link #getValidStartDate()}.
   * @return the date when this source becomes valid in the Metathesaurus
   */
  public Date getMstart() { return getValidStartDate(); } ;

  /**
   * Returns the date when this source becomes valid in the Metathesaurus (MSTART).
   * @return the date when this source becomes valid in the Metathesaurus
   */
  public Date getValidStartDate() { return mstart; }

  /**
   * Returns the date when this source is no longer valid in the Metathesaurus (MEND).
   * This is shorthand for {@link #getValidEndDate()}.
   * @return the date when this source is no longer valid in the Metathesaurus
   */
  public Date getMend() { return getValidEndDate(); }

  /**
   * Returns the date when this source is no longer valid in the Metathesaurus (MEND).
   * @return the date when this source is no longer valid in the Metathesaurus
   */
  public Date getValidEndDate() { return mend; }

  /**
   * Returns the version of the Metathesaurus in which this source
   * first appeared (IMETA).
   * This is shorthand for {@link #getInsertMetaVersion()}
   * @return the version of the Metathesaurus in which this source
   * first appeared
   */
  public String getImeta() { return getInsertMetaVersion(); }

  /**
   * Returns the version of the Metathesaurus in which this source
   * first appeared (IMETA).
   * @return the version of the Metathesaurus in which this source
   * first appeared
   */
  public String getInsertMetaVersion() { return imeta; }

  /**
   * Returns the first version of the Metathesaurus in which this source
   * no longer (RMETA).
   * This is shorthand for {@link #getRemoveMetaVersion()}
   * @return the first version of the Metathesaurus in which this source
   * no longer
   */
  public String getRmeta() { return getRemoveMetaVersion(); }

  /**
   * Returns the first version of the Metathesaurus in which this source
   * no longer (RMETA).
   * @return the first version of the Metathesaurus in which this source
   * no longer
   */
  public String getRemoveMetaVersion() { return rmeta; }

  /**
   * Returns the license contact information (SLC).
   * This is shorthand for {@link #getLicenseContact()}.
   * @return the license contact information
   */
  public String getSlc() { return getLicenseContact(); }

  /**
   * Returns the license contact information (SLC).
   * @return the license contact information
   */
  public String getLicenseContact() { return slc; }

  /**
   * Returns the citation information (SCIT).
   * This is shorthand for {@link #getCitation()}.
   * @return the citation information
   */
  public String getScit() { return getCitation(); }

  /**
   * Returns the citation information (SCIT).
   * @return the citation information
   */
  public String getCitation() { return scit; }

  /**
   * Returns the content contact information (SCC).
   * This is shorthand for {@link #getContentContact()}.
   * @return the content contact information
   */
  public String getContentContact() { return scc; }

  /**
   * Returns the content contact information (SCC).
   * @return the content contact information
   */
  public String getScc() { return scc; }

  /**
   * Returns the restriction level (SRL).
   * This is shorthand for {@link #getRestrictionLevel()}.
   * @return the restriction level
   */
  public String getSrl() { return getRestrictionLevel(); }

  /**
   * Returns the restriction level (SRL).
   * @return the restriction level
   */
  public String getRestrictionLevel() { return srl; }

  /**
   * Returns the frequency of terms (TFR).
   * This is shorthand for {@link #getTermFrequency()}
   * @return the frequency of terms
   */
  public int getTfr() { return getTermFrequency(); }

  /**
   * Returns the frequency of terms (TFR).
   * @return the frequency of terms
   */
  public int getTermFrequency() { return tfr; }

  /**
   * Returns the frequency of concepts (CFR).
   * This is shorthand for {@link #getConceptFrequency()}
   * @return the frequency of concepts
   */
  public int getCfr() { return getConceptFrequency(); }

  /**
   * Returns the frequency of concepts (CFR).
   * @return the frequency of concepts
   */
  public int getConceptFrequency() { return cfr; }

  /**
   * Returns the context type (CXTY).
   * This is shorthand for {@link #getContextType()}.
   * @return the context type
   */
  public String getCxty() { return getContextType(); }

  /**
   * Returns the context type (CXTY).
   * @return the context type
   */
  public String getContextType() { return cxty; }

  /**
   * Returns the {@link TermType}s.
   * @return an array of {@link TermType}s
   */
  public TermType[] getTermTypes() {
    return (TermType[]) ttyl.toArray(new TermType[0]);
  }

  /**
   * Returns the {@link TermType}s as a {@link List}.
   * @return the {@link TermType}s as a {@link List}
   */
  public List getTermTypesAsList() { return ttyl;}

  /**
   * Returns the term types as a comma (,) separated
   * list of term type abbreviations (TTYL).
   * This is shorthand for {@link #getTermTypesAsString()}.
   * @return the term type values
   */
  public String getTtyl() { return getTermTypesAsString(); }

  /**
   * Returns the term types as a comma (,) separated
   * list of term type abbreviations (TTYL).
   * @return the term type values
   */
  public String getTermTypesAsString() {
    StringBuffer sb = new StringBuffer(200);
    Iterator iter = ttyl.iterator();
    while (iter.hasNext()) {
      sb.append(((TermType)iter.next()).getAbbreviation());
      if (iter.hasNext()) sb.append(",");
    }
    return sb.toString();
  }
  /**
   * Get the {@link AttributeName}s for this {@link Source}.
   * Returns the {@link AttributeName}s.
   * @return an array of {@link AttributeName}s
   */
  public AttributeName[] getAttributeNames() {
    return (AttributeName[]) atnl.toArray ( new AttributeName[0]);
  }

  /**
   * Returns the {@link AttributeName}s as a {@link List}.
   * @return the {@link AttributeName}s as a {@link List}
   */
  public List getAttributeNamesAsList() { return atnl; }

  /**
   * Returns the attribute names as a comma (,) separated
   * list of attribute name abbreviations (ATNL).
   * This is shorthand for {@link #getAttributeNamesAsString()}.
   * @return the attribute name values
   */
  public String getAtnl() {return getAttributeNamesAsString(); }

  /**
   * Returns the attribute names as a comma (,) separated
   * list of attribute name abbreviations (ATNL).
   * @return the attribute name values
   */
  public String getAttributeNamesAsString() {
    StringBuffer sb = new StringBuffer(200);
    Iterator iter = atnl.iterator();
    while (iter.hasNext()) {
      sb.append(((AttributeName)iter.next()).getAbbreviation());
      if (iter.hasNext()) sb.append(",");
    }
    return sb.toString();
  }

  /**
   * Returns the {@link Language} (LAT).
   * This is shorthand for {@link #getLanguage()}.
   * @return the {@link Language}
   */
  public Language getLat() { return getLanguage(); }

  /**
   * Returns the {@link Language} (LAT).
   * @return the {@link Language}
   */
  public Language getLanguage() { return lat; }

  /**
   * Returns the ISO character encoding (CENC).
   * This is shorthand for {@link #getCharacterEncoding()}.
   * @return the character encoding
   */
  public String getCenc() { return getCharacterEncoding(); }

  /**
   * Returns the ISO character encoding (CENC).
   * @return the character encoding
   */
  public String getCharacterEncoding() { return cenc; }

  /**
   * Returns the "is current version" value (CURVER).
   * @return the "is current version" value
   */
  public String getCurver() { return curver; }

  /**
   * Indicates whether or not this represents the current version
   * of the source.
   * @return <code>true</code> if it does, <code>false</code> otherwise.
   */
  public boolean isCurrentVersion() { return curver != null && curver.equals("Y"); }

  /**
   * Returns the "in current subset" value (SABIN).
   * @return the "in current subset" value
   */
  public String getSabin() { return sabin; }

  /**
   * Sets whether or not this represents the current version
   * of the source.
   * @param sabin <code>boolean</code>
   */
  public void setInCurrentSubset(boolean sabin) {
    this.sabin = (sabin ? "Y" : "N");
  }

  /**
   * Sets the "in current subset" value (SABIN).
   * @param sabin "in current subset" value
   */
  public void setSabin(String sabin) {
    this.sabin = sabin;
  }

  /**
   * Indicates whether or not this source is represented
   * in the current subset.
   * @return <code>true</code> if it is, <code>false</code> otherwise.
   */
  public boolean inCurrentSubset() {
    return sabin != null && sabin.equals("Y");
  }

  /**
   * Return a hash code based on key fields.
   * @return a hash code based on key fields
   */
  public int hashCode() {
    return (getVersionedSourceAbbreviation()).hashCode();
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be an
   * {@link Source} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof Source)) return false;
    return  this == o;
  }

  /**
   * Compares this {@link Source} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof Source)
      return getVersionedSourceAbbreviation().compareTo(
        ((Source)o).getVersionedSourceAbbreviation());
    else
      return 0;
  }

  /**
   * Returns the list of {@link Source}s for a specified source family.
   * @param family a source family, e.g. <code>MSH</code>
   * @return the {@link List} of {@link Source}s for a specified source family.
   */
  public static List getSourcesForFamily(String family) {
    List list = new ArrayList();
    List sources = getSourcesAsList();
    for (int i = 0; i < sources.size(); i++) {
      if (((Source)sources.get(i)).getSourceFamily().equals(family))
	list.add(sources.get(i));
    }
    return list;
  }

  /**
   * Returns the set of {@link Source}s for a specified source family.
   * @param family a source family, e.g. <code>MSH</code>
   * @return the {@link Set} of {@link Source}s for a specified source family.
   */
  public static Set getSourcesForFamilyAsSet(String family) {
    return new TreeSet(getSourcesForFamily(family));
  }

  /**
   * Returns the versioned source abbreviation (VSAB).
   * @return the versioned source abbreviation
   */
  public String toString() {
    return getVersionedSourceAbbreviation();
  }
}
