/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Class:       SourceIdentifierType.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Represents information about a the type of a {@link MapObject}'s
 * expression. This class ensures that each
 * unique expression type is represented by a singleton object.  In order for
 * this to work correctly, the class must be properly initialized.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public final class SourceIdentifierType extends AbstractExpandedForm  {


  // Map by abbreviation to the actual classes
  private static final Map mets = new HashMap();

  // Abbreviation and description for this individual instance.
  private String abbrev;
  private String dsc;

  /**
   * Instantiates an empty {@link SourceIdentifierType}.
   */
  public SourceIdentifierType() {};

  /**
   * Instantiates an {@link SourceIdentifierType} from the
   * specified {@link MetaDescription}.
   * @param meta_dsc a {@link MetaDescription} representing a term type
   */
  private SourceIdentifierType(MetaDescription meta_dsc) {
    abbrev = meta_dsc.getValue();
    dsc = meta_dsc.getExplain();
  }

  /**
   * Initializes the {@link Map} used to track all of the
   * {@link SourceIdentifierType}s that exist at one time.  You may
   * pass all {@link MetaDescription}s as this method only
   * creates {@link SourceIdentifierType}s for those with a type of "TTY".
   * @param meta_dsc an array of {@link MetaDescription}s
   * @throws IllegalStateException if an attempt is made to call this
   * method more than once
   */
  public static synchronized void initialize(MetaDescription[] meta_dsc) {
    for (int i=0; i<meta_dsc.length; i++) {
      if ((meta_dsc[i].getKey().equals("STYPE1") ||
           meta_dsc[i].getKey().equals("STYPE2") ||
           meta_dsc[i].getKey().equals("STYPE") ||
           meta_dsc[i].getKey().equals("FROMTYPE") ||
           meta_dsc[i].getKey().equals("TOTYPE")) &&
          meta_dsc[i].getType().equals("expanded_form")) {
	if (mets.containsKey(meta_dsc[i].getValue())) {
	   SourceIdentifierType met = (SourceIdentifierType)mets.get(meta_dsc[i].getValue());
           met.dsc = meta_dsc[i].getExplain();
        } else {
          mets.put(meta_dsc[i].getValue(), new SourceIdentifierType(meta_dsc[i]));
        }
      }
    }
  }

  /**
   * Initializes the {@link Map} used to track all of the
   * {@link SourceIdentifierType}s that exist at one time.  This
   * is implemented as a call to {@link #initialize(MetaDescription[])}.
   * @param meta_dsc an array of {@link MetaDescription}s
   */
  public static synchronized void initialize(List meta_dsc) {
    initialize((MetaDescription[])meta_dsc.toArray(new MetaDescription[0]));
  }

  /**
   * Returns the {@link SourceIdentifierType} for the specified abbreviation.
   * This method will return <code>null</code> if the <code>initialize</code>
   * method has not been called or if the specified abbreviation
   * was not one of those used during initialization.
   * @param met the term type abbreviation to look up
   * @return the {@link SourceIdentifierType} for the specified abbreviation
   */
  public static SourceIdentifierType getSourceIdentifierType(String met) {
    return (SourceIdentifierType)mets.get(met);
  }

  /**
   * Return all {@link SourceIdentifierType}s.
   * @return an array of {@link SourceIdentifierType}s
   */
  public static SourceIdentifierType[] getSourceIdentifierTypes() {
    return (SourceIdentifierType[])mets.values().toArray(new SourceIdentifierType[0]);
  }

  /**
   * Return all {@link SourceIdentifierType}s as a {@link List}.
   * @return all {@link SourceIdentifierType}s as a {@link List}
   */
  public static List getSourceIdentifierTypesAsList() {
    return new ArrayList(mets.values());
  }

  /**
   * Returns the abbreviation.
   * @return the abbreviation
   */
  public String getAbbreviation() {
    return abbrev;
  }

  /**
   * Returns the description
   * @return the description
   */
  public String getDescription() {
    return dsc;
  }

}
