/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Interface:   SourceTermType.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;

/**
 * Represents a tuple consisting of a {@link Source} and a {@link TermType}.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class SourceTermType implements SuppressibleElement, Comparable {

  //
  // Fields
  //
  private Source source;
  private TermType tty;
  private Suppressible suppress;
  private int hashcode = -1;

  /**
   * Instantiate an empty {@link SourceTermType}.
   */
  public SourceTermType() {};

  /**
   * Returns the {@link Source}.
   * @return the {@link Source}
   */
  public Source getSource() {
    return source;
  }

  /**
   * Sets the {@link Source}.
   * @param source the {@link Source}
   */
  public void setSource(Source source) {
    this.source = source;
  }

  /**
   * Returns tbe {@link TermType}.
   * @return tbe {@link TermType}
   */
  public TermType getTermType() {
    return tty;
  }

  /**
   * Sets tbe {@link TermType}.
   * @param tty tbe {@link TermType}
   */
  public void setTermType(TermType tty) {
    this.tty = tty;
  }

  /**
   * Returns the suppressibility value (SUPPRESS).
   * @return the {@link Suppressible} value
   */
  public Suppressible getSuppress() {
    return suppress;
  }

  /**
   * Indicates whether or not this {@link SourceTermType} is suppressible.
   * @return <code>true</code> if suppressible,
   *         <code>false</code> if not suppressible
   */
  public boolean isSuppressible() {
    return (suppress != null && !suppress.getAbbreviation().equals("N"));
  }

  /**
   * Sets the suppressibility value (SUPPRESS).
   * @param suppress the {@link Suppressible} value
   */
  public void setSuppress(Suppressible suppress) {
      this.suppress = suppress;
  }

  /**
   * Indicates whether or not this {@link SourceTermType} is obsolete.
   * @return <code>true</code> if obsolete,
   *         <code>false</code> if not obsolete
   */
  public boolean isObsolete() {
    TermTypeGroup[] tty_groups = TermTypeGroup.getTermTypeGroups(getTermType());
    for (int i=0; i<tty_groups.length; i++) {
      if (tty_groups[i].getDescription().equals("obsolete"))
        return true;
    }
    return false;
  }

  /**
   * Return a hash code based on key fields.
   * @return a hash code based on key fields
   */
  public int hashCode() {
    if (hashcode == -1)
      hashcode = (getSource().getVsab() + getTermType().getAbbreviation()).hashCode();
    return hashcode;
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be an
   * {@link SourceTermType} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof SourceTermType)) return false;
    SourceTermType att = (SourceTermType)o;
    return att.getSource() == getSource() &&
           att.getTermType() == getTermType();
  }

  /**
   * Compares this {@link SourceTermType} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof SourceTermType) {
      SourceTermType source_tty = (SourceTermType)o;
      if (getSource() != source_tty.getSource())
        return getSource().compareTo(source_tty.getSource());
      return getTermType().compareTo(source_tty.getTermType());
    } else
      return 0;
  }
}
