/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Class:       TermStatus.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Represents an term status.  This class ensures that each unique
 * string type is represented by a singleton object.  In order for
 * this to work correctly, the class must be properly initialized.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public final class TermStatus extends AbstractExpandedForm {

  // Map by abbreviation to the actual classes
  private static final Map tss = new HashMap();

  // Abbreviation and description for this individual instance.
  private String abbrev;
  private String dsc;

  /**
   * Instantiates an empty {@link TermStatus}.
   */
  public TermStatus() {};

  /**
   * Instantiates an {@link TermStatus} from the
   * specified {@link MetaDescription}.
   * @param meta_dsc a {@link MetaDescription} representing a term status
   */
  private TermStatus(MetaDescription meta_dsc) {
    abbrev = meta_dsc.getValue();
    dsc = meta_dsc.getExplain();
  }

  /**
   * Initializes the {@link Map} used to track all of the
   * {@link TermStatus}es that exist at one time.  You may
   * pass all {@link MetaDescription}s as this method only
   * creates {@link TermStatus}s for those with a type of "TS".
   * @param meta_dsc an array of {@link MetaDescription}s
   */
  public static synchronized void initialize(MetaDescription[] meta_dsc) {
    for (int i=0; i<meta_dsc.length; i++) {
      if (meta_dsc[i].getKey().equals("TS")&&
          meta_dsc[i].getType().equals("expanded_form")) {
	if (tss.containsKey(meta_dsc[i].getValue())) {
	   TermStatus ts = (TermStatus)tss.get(meta_dsc[i].getValue());
           ts.dsc = meta_dsc[i].getExplain();
        } else {
          tss.put(meta_dsc[i].getValue(), new TermStatus(meta_dsc[i]));
        }
      }
    }
  }

  /**
   * Initializes the {@link Map} used to track all of the
   * {@link TermStatus}es that exist at one time.  This
   * is implemented as a call to {@link #initialize(MetaDescription[])}.
   * @param meta_dsc an array of {@link MetaDescription}s
   * @throws IllegalStateException if an attempt is made to call this
   * method more than once
   */
  public static synchronized void initialize(List meta_dsc) {
    initialize((MetaDescription[])meta_dsc.toArray(new MetaDescription[0]));
  }

  /**
   * Returns the {@link TermStatus} for the specified abbreviation.
   * This method will return <code>null</code> if the <code>initialize</code>
   * method has not been called or if the specified abbreviation
   * was not one of those used during initialization.
   * @param ts the term status abbreviation to look up
   * @return the {@link TermStatus} for the specified abbreviation
   */
  public static TermStatus getTermStatus(String ts) {
    return (TermStatus)tss.get(ts);
  }

  /**
   * Return all {@link TermStatus} values.
   * @return an array of {@link TermStatus}
   */
  public static TermStatus[] getTermStatuses() {
    return (TermStatus[])tss.values().toArray(new TermStatus[0]);
  }

  /**
   * Return all {@link TermStatus} values.
   * @return all {@link TermStatus} values
   */
  public static List getTermStatusAsList() {
    return new ArrayList(tss.values());
  }

  /**
   * Returns the abbreviation.
   * @return the abbreviation
   */
  public String getAbbreviation() {
    return abbrev;
  }

  /**
   * Returns the description
   * @return the description
   */
  public String getDescription() {
    return dsc;
  }

}