/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Class:       TermType.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Represents an Term Type. This class ensures that each unique
 * term type is represented by a singleton object.  In order for
 * this to work correctly, the class must be properly initialized.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public final class TermType extends AbstractExpandedForm  {


  // Map by abbreviation to the actual classes
  private static final Map ttys = new HashMap();

  // Abbreviation and description for this individual instance.
  private String abbrev;
  private String dsc;

  /**
   * Instantiates an empty {@link TermType}.
   */
  public TermType() {};

  /**
   * Instantiates an {@link TermType} from the
   * specified {@link MetaDescription}.
   * @param meta_dsc a {@link MetaDescription} representing a term type
   */
  private TermType(MetaDescription meta_dsc) {
    abbrev = meta_dsc.getValue();
    dsc = meta_dsc.getExplain();
  }

  /**
   * Initializes the {@link Map} used to track all of the
   * {@link TermType}s that exist at one time.  You may
   * pass all {@link MetaDescription}s as this method only
   * creates {@link TermType}s for those with a type of "TTY".
   * @param meta_dsc an array of {@link MetaDescription}s
   * @throws IllegalStateException if an attempt is made to call this
   * method more than once
   */
  public static synchronized void initialize(MetaDescription[] meta_dsc) {
    for (int i=0; i<meta_dsc.length; i++) {
      if (meta_dsc[i].getKey().equals("TTY") &&
          meta_dsc[i].getType().equals("expanded_form")) {
	if (ttys.containsKey(meta_dsc[i].getValue())) {
	   TermType tty = (TermType)ttys.get(meta_dsc[i].getValue());
           tty.dsc = meta_dsc[i].getExplain();
        } else {
          ttys.put(meta_dsc[i].getValue(), new TermType(meta_dsc[i]));
        }
      }
    }
  }

  /**
   * Initializes the {@link Map} used to track all of the
   * {@link TermType}s that exist at one time.  This
   * is implemented as a call to {@link #initialize(MetaDescription[])}.
   * @param meta_dsc an array of {@link MetaDescription}s
   */
  public static synchronized void initialize(List meta_dsc) {
    initialize((MetaDescription[])meta_dsc.toArray(new MetaDescription[0]));
  }

  /**
   * Returns the {@link TermType} for the specified abbreviation.
   * This method will return <code>null</code> if the <code>initialize</code>
   * method has not been called or if the specified abbreviation
   * was not one of those used during initialization.
   * @param tty the term type abbreviation to look up
   * @return the {@link TermType} for the specified abbreviation
   */
  public static TermType getTermType(String tty) {
    return (TermType)ttys.get(tty);
  }

  /**
   * Return all {@link TermType}s.
   * @return an array of {@link TermType}s
   */
  public static TermType[] getTermTypes() {
    return (TermType[])ttys.values().toArray(new TermType[0]);
  }

  /**
   * Return all {@link TermType}s as a {@link List}.
   * @return all {@link TermType}s as a {@link List}
   */
  public static List getTermTypesAsList() {
    return new ArrayList(ttys.values());
  }

  /**
   * Returns the abbreviation.
   * @return the abbreviation
   */
  public String getAbbreviation() {
    return abbrev;
  }

  /**
   * Returns the description
   * @return the description
   */
  public String getDescription() {
    return dsc;
  }

}
