/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta.impl
 * Class:       AtomImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.AtomAttribute;
import gov.nih.nlm.umls.meta.Context;
import gov.nih.nlm.umls.meta.Coocurrence;
import gov.nih.nlm.umls.meta.Definition;
import gov.nih.nlm.umls.meta.Locator;
import gov.nih.nlm.umls.meta.Relationship;
import gov.nih.nlm.umls.meta.Restrictor;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

/**
 * Default implementation of the {@link gov.nih.nlm.umls.meta.Atom}
 * interface.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class AtomImpl extends PlaceholderAtom  {

  //
  // Fields
  //

  private List definitions = null;
  private List locators = null;
  private List relationships = null;
  private List coocurrences = null;
  private List contexts = null;
  private List attributes = null;

  /**
   * Instantiates an empty {@link Atom}.
   */
  public AtomImpl() {};

  /**
   * Instantiates an {@link Atom} from the specified {@link Atom}.
   * @param atom {@link Atom} to be copied
   */
  public AtomImpl(Atom atom) {
    setTty(atom.getTty());
    setSaui(atom.getSaui());
    setAui(atom.getAui());
    setScui(atom.getScui());
    setSdui(atom.getSdui());
    setSuppress(atom.getSuppress());
    setSui(atom.getSui());
    setConcept(atom.getConcept());
    setSource(atom.getSource());
    setCvf(atom.getCvf());
    setLui(atom.getLui());
    setCode(atom.getCode());
    setTs(atom.getTs());
    setLat(atom.getLat());
    setStt(atom.getStt());
    setStr(atom.getStr());
    setIspref(atom.getIspref());
  }

  //
  // Definitions
  //

  /**
   * Returns the {@link Definition}s.
   * @return a {@link Definition} <code>[]</code>
   */
  public Definition[] getDefinitions() {
    return definitions == null ? new Definition[0] :
      (Definition[])definitions.toArray(new Definition[0]);
  }

  /**
   * Returns the {@link Definition}s as a {@link List}.
   * @return the {@link Definition}s as a {@link List}
   */
  public List getDefinitionsAsList() {
    return definitions == null ? new ArrayList(0) : new ArrayList(definitions);
  }

  /**
   * Adds the specified {@link Definition}.
   * @param def the {@link Definition} to be added
   */
  public void addDefinition(Definition def) {
    if (definitions == null ) definitions = new ArrayList();
    definitions.add(def);
  }

  /**
   * Remove the specified {@link Definition}.
   * @param def the {@link Definition} to be removed
   */
  public void removeDefinition(Definition def) {
    if (definitions != null) {
      definitions.remove(def);
    }
  }

  /**
   * Remove all {@link Definition}s.
   */
  public void clearDefinitions() {
    if (definitions != null) definitions = null;
  }

  /**
   * Replace {@link Definition}s with the ones in the specified array.
   * @param defs a {@link Definition} <code>[]</code>
   */
  public void setDefinitions(Definition defs[]) {
    definitions = new ArrayList();
    for (int i=0; i<defs.length; i++) {
      definitions.add(defs[i]);
    }
  }

  /**
   * Replace {@link Definition}s with the ones in the specified {@link List}.
   * @param defs a {@link List} of {@link Definition}s
   */
  public void setDefinitions(List defs) {
    definitions = new ArrayList(defs);
  }

  //
  // Locators
  //

  /**
   * Returns the {@link Locator}s.
   * @return a {@link Locator} <code>[]</code>
   */
  public Locator[] getLocators() {
     return locators == null ? new Locator[0] :
     (Locator[])locators.toArray(new Locator[0]);
  }

  /**
   * Returns the {@link Locator}s as a {@link List}.
   * @return the {@link Locator}s as a {@link List}
   */
  public List getLocatorsAsList() {
    return locators == null ? new ArrayList(0) : new ArrayList(locators);
  }

  /**
   * Add the specified {@link Locator}.
   * @param loc the {@link Locator} to be added
   */
  public void addLocator(Locator loc) {
    if (locators == null ) locators = new ArrayList(0);
    locators.add(loc);
  }

  /**
   * Removes the specified {@link Locator}.
   * @param loc the {@link Locator} to be removed
   */
  public void removeLocator(Locator loc) {
    if (locators != null) locators.remove(loc);
  }

  /**
   * Remove all {@link Locator}s.
   */
  public void clearLocators() {
    if (locators != null) locators = null;
  }

  /**
   * Replace {@link Locator}s with the ones in the specified array.
   * @param locs a {@link Locator} <code>[]</code>
   */
  public void setLocators(Locator locs[]) {
    locators = new ArrayList();
    for (int i=0; i<locs.length; i++) {
      locators.add(locs[i]);
    }
  }

  /**
   * Replace {@link Locator}s with the ones in the specified {@link List}.
   * @param locs a {@link List} of {@link Locator}s
   */
  public void setLocators(List locs) {
    locators = new ArrayList(locs);
  }

  //
  // Relationships
  //

  /**
   * Returns the {@link Relationship}s.
   * @return a {@link Relationship} <code>[]</code>
   */
  public Relationship[] getRelationships() {
    return relationships == null ? new Relationship[0] : (Relationship[])relationships.toArray(new Relationship[0]);  }

  /**
   * Returns the {@link Relationship}s as a {@link List}.
   * @return the {@link Relationship}s as a {@link List}
   */
  public List getRelationshipsAsList() {
    return relationships == null ? new ArrayList(0): new ArrayList(relationships);
  }

  /**
   * Adds the specified {@link Relationship}.
   * @param rel the {@link Relationship} to be added
   */
  public void addRelationship(Relationship rel) {
    if (relationships == null) relationships = new ArrayList();
    relationships.add(rel);
  }

  /**
   * Removes the specified {@link Relationship}.
   * @param rel the {@link Relationship} to be removed
   */
  public void removeRelationship(Relationship rel) {
    if (relationships != null) {
      int i = relationships.lastIndexOf(rel);
      if (i != -1)
        relationships.remove(i);
    }
  }

  /**
   * Remove all {@link Relationship}s.
   */
  public void clearRelationships() {
    if (relationships != null) relationships = null;
  }

  /**
   * Replace {@link Relationship}s with the ones in the specified array.
   * @param rels a {@link Relationship} <code>[]</code>
   */
  public void setRelationships(Relationship rels[]) {
    relationships = new ArrayList();
    for (int i=0; i<rels.length; i++) {
      relationships.add(rels[i]);
    }
  }

  /**
   * Replace {@link Relationship}s of the ones in the specified {@link List}.
   * @param rels a {@link List} of {@link Relationship}s
   */
  public void setRelationships(List rels) {
    relationships = new ArrayList(rels);
  }

  //
  // Coocurrences
  //

  /**
   * Returns the {@link Coocurrence}s.
   * @return a {@link Coocurrence} <code>[]</code>
   */
  public Coocurrence[] getCoocurrences() {
    return coocurrences == null ? new Coocurrence[0] :
      (Coocurrence[])coocurrences.toArray(new Coocurrence[0]);
  }


  /**
   * Returns the {@link Coocurrence}s as a {@link List}.
   * @return the {@link Coocurrence}s as a {@link List}
   */
  public List getCoocurrencesAsList() {
    if (coocurrences != null)
      return new ArrayList(coocurrences);
    else
      return new ArrayList(0);
  }

  /**
   * Adds the specified {@link Coocurrence}.
   * @param coc the {@link Coocurrence} to be added
   */
  public void addCoocurrence(Coocurrence coc) {
    if (coocurrences == null)
      coocurrences = new ArrayList();
    coocurrences.add(coc);
  }

  /**
   * Removes the specified {@link Coocurrence}.
   * @param coc the {@link Coocurrence} to be removed
   */
  public void removeCoocurrence(Coocurrence coc) {
    if (coocurrences != null) {
      int i = coocurrences.lastIndexOf(coc);
      if (i != -1)
        coocurrences.remove(i);
    }
  }

  /**
   * Remove all {@link Coocurrence}s.
   */
  public void clearCoocurrences() {
    if (coocurrences != null) coocurrences = null;
  }

  /**
   * Replace {@link Coocurrence}s with the ones in the specified array.
   * @param cocs a {@link Coocurrence} <code>[]</code>
   */
  public void setCoocurrences(Coocurrence cocs[]){
    coocurrences = new ArrayList();
    for (int i=0; i<cocs.length; i++) {
      coocurrences.add(cocs[i]);
    }
  }

  /**
   * Replace {@link Coocurrence}s with the ones in the specified {@link List}.
   * @param cocs a {@link List} of {@link Coocurrence}s
   */
  public void setCoocurrences(List cocs) {
    coocurrences = new ArrayList(cocs);
  }

  //
  // Contexts
  //

  /**
   * Return the {@link Context}s.
   * @return a {@link Context} <code>[]</code>
   */
  public Context[] getContexts() {
    return contexts == null ? new Context[0] :
      (Context[])contexts.toArray(new Context[0]);
  }

  /**
   * Returns the {@link Context}s as a {@link List}.
   * @return the {@link Context}s as a {@link List}
   */
  public List getContextsAsList() {
    return contexts == null ? new ArrayList(0) : new ArrayList(contexts);
  }

  /**
   * Adds the specified {@link Context}.
   * @param cxt the {@link Context} to be added
   */
  public void addContext(Context cxt) {
    if (contexts == null)
      contexts = new ArrayList();
    contexts.add(cxt);
  }

  /**
   * Removes the specified {@link Context}.
   * @param cxt the {@link Context} to be removed
   */
  public void removeContext(Context cxt) {
    if (contexts != null) {
      int i = contexts.lastIndexOf(cxt);
      if (i != -1)
        contexts.remove(i);
    }
  }

  /**
   * Remove all {@link Context}s.
   */
  public void clearContexts() {
    if (contexts != null) contexts = null;
  }

  /**
   * Replace {@link Context}s with the ones in the specified array.
   * @param cxts a {@link Context} <code>[]</code>
   */
  public void setContexts(Context cxts[]) {
    contexts = new ArrayList();
    for (int i=0; i<cxts.length; i++) {
      contexts.add(cxts[i]);
    }
  }

  /**
   * Replace {@link Context}s with the ones in the specified {@link List}.
   * @param cxts a {@link List} of {@link Context}s
   */
  public void setContexts(List cxts) {
    contexts = new ArrayList(cxts);
  }

  //
  // AtomAttributes
  //

  /**
   * Returns the {@link AtomAttribute}s.
   * @return a {@link AtomAttribute} <code>[]</code>
   */
  public AtomAttribute[] getAttributes() {
    return attributes == null ? new AtomAttribute[0] :
      (AtomAttribute[])attributes.toArray(new AtomAttribute[0]);
  }

  /**
   * Returns the {@link AtomAttribute}s as a {@link List}.
   * @return the {@link AtomAttribute}s as a {@link List}
   */
  public List getAttributesAsList() {
    if (attributes != null)
      return new ArrayList(attributes);
    else
      return new ArrayList(0);
  }

  /**
   * Adds the specified {@link AtomAttribute}.
   * @param att the {@link AtomAttribute} to be added
   */
  public void addAttribute(AtomAttribute att) {
    if (attributes == null)
      attributes = new ArrayList();
    attributes.add(att);
  }

  /**
   * Removes the specified {@link AtomAttribute}.
   * @param att the {@link AtomAttribute} to be removed
   */
  public void removeAttribute(AtomAttribute att) {
    if (attributes != null) {
      int i = attributes.lastIndexOf(att);
      if (i != -1)
        attributes.remove(i);
    }
  }

  /**
   * Remove all {@link AtomAttribute}s.
   */
  public void clearAttributes() {
    if (attributes != null) attributes = null;
  }

  /**
   * Replace {@link AtomAttribute}s with the ones in the specified array.
   * @param atts an {@link AtomAttribute} <code>[]</code>
   */
  public void setAttributes(AtomAttribute atts[]) {
    attributes = new ArrayList();
    for (int i=0; i<atts.length; i++) {
      attributes.add(atts[i]);
    }
  }

  /**
   * Replace {@link AtomAttribute}s with the ones in the specified {@link List}.
   * @param atts a {@link List} of {@link AtomAttribute}s
   */
  public void setAttributes(List atts) {
    attributes = new ArrayList(atts);
  }

  /**
   * Returns a subset of this {@link Atom}'s {@link Relationship}s.
   * Restricts the return based on the given {@link Restrictor}.
   * @param restrictor {@link Restrictor}
   * @return Relationship[]
   */
  public Relationship[] getRelationships(Restrictor restrictor) {
    List rels = getRelationshipsAsList();
    for (int i=rels.size()-1; i>=0; i--) {
      if (!restrictor.keep(rels.get(i)))
        rels.remove(i);
    }
    Collections.sort(rels, restrictor);
    return (Relationship[])rels.toArray(new Atom[0]);
  }

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from the cui, aui, sui,
   * versioned source, tty, and code.
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
      final StringBuffer sb1 = new StringBuffer(80);
      sb1.append(getConcept() != null ? getConcept().getCui() : "")
         .append(getSui())
         .append(getAui())
         .append(getSource() != null ? getSource().getVsab() : "")
         .append(getTty() != null ? getTty().getAbbreviation() : "")
         .append(getCode());
      key_fields = sb1.toString();
      key_fields_flag = true;
    }
    return key_fields;
  }

}
