/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta.impl
 * Class:       AtomRelationshipImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.AtomRelationship;
import gov.nih.nlm.umls.meta.Relationship;

/**
 * Implementation of the {@link gov.nih.nlm.umls.meta.Relationship}
 * interface.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class AtomRelationshipImpl extends RelationshipImpl implements AtomRelationship  {

  //
  // Fields
  //
  private Atom atom;
  private Atom related_atom;


  /**
   * Instantiates an empty {@link AtomRelationship}.
   */
  public AtomRelationshipImpl() {};

  /**
   * Instantiates an {@link AtomRelationship} from the specified {@link AtomRelationship}.
   * @param rel {@link AtomRelationship} to be copied
   */
  public AtomRelationshipImpl(AtomRelationship rel) {
    super(rel);
    setAtom(rel.getAtom());
    setRelatedAtom(rel.getRelatedAtom());
  }

  /**
   * Returns the related atom involved.
   * @return the related atom involved {@link Atom}
   */
  public Atom getRelatedAtom() {
    return related_atom;
  }

  /**
   * Sets the related atom involved.
   * @param related_atom the related atom involved
   */
  public void setRelatedAtom(Atom related_atom) {
    this.related_atom = related_atom;
    key_fields_flag = false;
  }

  /**
   * Returns the {@link Atom} this {@link Relationship} is connected to.
   * @return the {@link Atom} this {@link Relationship} is connected to
   */
  public Atom getAtom() {
    return atom;
  }

  /**
   * Sets the {@link Atom} this {@link Relationship} is connected to.
   * @param atom  the {@link Atom} this {@link Relationship} is connected to
   */
  public void setAtom(Atom atom) {
    this.atom = atom;
    key_fields_flag = false;
  }


  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from the cui1, rel,
   * aui1, versioned source, cui2, aui2, rela, suppress, rui, sl, and srui.
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
    final StringBuffer sb = new StringBuffer(100);
    sb.append(getConcept() != null ? getConcept().getCui() : "")
      .append(getAtom() != null ? getAtom().getAui() : "")
      .append(getRelatedConcept() != null ? getRelatedConcept().getCui() : "")
      .append(getRelatedAtom() != null ? getRelatedAtom().getAui() : "")
      .append(getSource() != null ? getSource().getVsab() : "")
      .append(getRel() != null ? getRel().getAbbreviation() : "")
      .append(getRela() != null ? getRela().getAbbreviation() : "")
      .append(getSuppress() != null ? getSuppress().getAbbreviation() : "")
      .append(getSl() != null ? getSl().getVsab() : "")
      .append(getRui())
      .append(getSrui());
      key_fields = sb.toString();
      key_fields_flag = true;
    }
    return key_fields;
  }

  /**
   * Return a hash code based on key fields.
   * @return a hash code based on key fields
   */
  public int hashCode() {
    return getKeyFields().hashCode();
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be an
   * {@link AtomRelationship} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof AtomRelationship)) return false;
    return getKeyFields().equals(((AtomRelationship)o).getKeyFields());
  }

  /**
   * Compares this {@link Relationship} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof Relationship)
      return getKeyFields().compareTo(((AtomRelationship)o).getKeyFields());
    else
      return 0;
  }


}
