/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta.impl
 * Class:       AttributeValueImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.AttributeName;
import gov.nih.nlm.umls.meta.AttributeValue;
import gov.nih.nlm.umls.meta.ContentView;
import gov.nih.nlm.umls.meta.Suppressible;

/**
 * Default implementation of the {@link gov.nih.nlm.umls.meta.AttributeValue}
 * interface.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class AttributeValueImpl implements AttributeValue {

  //
  // Fields
  //
  private AttributeName atn;
  private String atv;
  private ContentView cvf;
  private String satui;
  private String atui;
  private Suppressible suppress;

  // Used to compute key fields more efficiently
  private boolean key_fields_flag;
  private String key_fields = null;

  /**
   * Instantiates an empty {@link AttributeValue}.
   */
  public AttributeValueImpl() {};


  /**
   * Instantiates a {@link AttributeValue} from the specified {@link AttributeValue}.
   * @param att {@link AttributeValue} to be copied
   */
  public AttributeValueImpl(AttributeValue att) {
    setAtn(att.getAtn());
    setAtv(att.getAtv());
    setSuppress(att.getSuppress());
    setAtui(att.getAtui());
    setSatui(att.getSatui());
    setCvf(att.getCvf());
  }

  /**
   * Returns the {@link AttributeName} (ATN).
   * @return the {@link AttributeName}
   */
  public AttributeName getAttributeName() {
    return atn;
  }

  /**
   * Shorthand for {@link #getAttributeName()}.
   * @return the {@link AttributeName}
   */
  public AttributeName getAtn() {
    return getAttributeName();
  }

  /**
   * Sets the {@link AttributeName} (ATN).
   * @param atn the {@link AttributeName}
   */
  public void setAttributeName(AttributeName atn) {
    this.atn = atn;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setAttributeName(AttributeName)}.
   * @param atn the {@link AttributeName}
   */
  public void setAtn(AttributeName atn) {
    setAttributeName(atn);
  }

  /**
   * Returns the attribute value (ATV).
   * @return the attribute value
   */
  public String getAttributeValue() {
    return atv;
  }

  /**
   * Shorthand for {@link #getAttributeValue()}.
   * @return the attribute value
   */
  public String getAtv() {
    return getAttributeValue();
  }

  /**
   * Sets the attribute value (ATV).
   * @param atv the attribute value
   */
  public void setAttributeValue(String atv) {
    this.atv = atv;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setAttributeValue(String)}.
   * @param atv the attribute value
   */
  public void setAtv(String atv) {
    setAttributeValue(atv);
  }

  /**
   * Returns the source asserted attribute identifier (SATUI).
   * @return the SATUI
   */
  public String getSourceAttributeIdentifier() {
    return satui;
  }

  /**
   * Shorthand for {@link #getSourceAttributeIdentifier()}.
   * @return the SATUI
   */
  public String getSatui() {
    return getSourceAttributeIdentifier();
  }

  /**
   * Sets the source asserted attribute identifier (SATUI).
   * @param satui the SATUI
   */
  public void setSourceAttributeIdentifier(String satui) {
    this.satui = satui;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setSourceAttributeIdentifier(String)}.
   * @param satui the SATUI
   */
  public void setSatui(String satui) {
    setSourceAttributeIdentifier(satui);
  }

  /**
   * Returns the unique identifier (ATUI).
   * @return the ATUI
   */
  public String getAttributeIdentifier() {
    return atui;
  }

  /**
   * Returns the unique identifier (ATUI). This
   * is shorthand for {@link #getAttributeIdentifier()}.
   * @return the ATUI
   */
  public String getAtui() {
    return getAttributeIdentifier();
  }

  /**
   * Sets the unique identifier (ATUI).
   * @param atui the ATUI
   */
  public void setAttributeIdentifier(String atui) {
    this.atui = atui;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setAttributeIdentifier(String)}.
   * @param atui the ATUI
   */
  public void setAtui(String atui) {
    setAttributeIdentifier(atui);
  }

  /**
   * Returns the suppressibility value (SUPPRESS).
   * @return the {@link Suppressible} value
   */
  public Suppressible getSuppress() {
    return suppress;
  }

  /**
   * Indicates whether or not this {@link AttributeValue} is suppressible.
   * @return <code>true</code> if suppressible,
   *         <code>false</code> if not suppressible
   */
  public boolean isSuppressible() {
    return (suppress != null && !suppress.getAbbreviation().equals("N"));
  }

  /**
   * Sets the suppressibility value (SUPPRESS).
   * @param suppress the {@link Suppressible} value
   */
  public void setSuppress(Suppressible suppress) {
    this.suppress = suppress;
    key_fields_flag = false;
  }

  /**
   * Indicates whether or not this {@link AttributeValueImpl} is obsolete.
   * @return <code>true</code> if obsolete,
   *         <code>false</code> if not obsolete
   */
  public boolean isObsolete() {
    if (getSuppress().getAbbreviation().equals("O"))
      return true;
    return false;
  }

  /**
   * Returns tbe {@link ContentView} definition for this {@link AttributeValue} (CVF).
   * @return tbe {@link ContentView} definition
   */
  public ContentView getContentView() {
    return cvf;
  }

  /**
   * Sets tbe {@link ContentView} definition for this {@link AttributeValue} (CVF).
   * @param cvf the {@link ContentView} definition
   */
  public void setContentView(ContentView cvf) {
    this.cvf = cvf;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #getContentView()}.
   * @return tbe {@link ContentView} definition
   */
  public ContentView getCvf() {
    return getContentView();
  }

  /**
   * Shorthand for {@link #setContentView(ContentView)}.
   * @param cvf the {@link ContentView} definition
   */
  public void setCvf(ContentView cvf) {
    setContentView(cvf);
  }

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from the atn and atv.
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
      StringBuffer sb1 = new StringBuffer(100);
      sb1.append(getAtn())
         .append(getAtv());
      key_fields = sb1.toString();
      key_fields_flag = true;
    }
    return key_fields;
  }

  /**
   * Return a hash code based on key fields.
   * @return a hash code based on key fields
   */
  public int hashCode() {
    return getKeyFields().hashCode();
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be an
   * {@link AttributeValue} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof AttributeValue)) return false;
    return getKeyFields().equals(((AttributeValue)o).getKeyFields());
  }

  /**
   * Compares this {@link AttributeValue} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof AttributeValue)
      return getKeyFields().compareTo(((AttributeValue)o).getKeyFields());
    else
      return 0;
  }

}
