/**************************************************************************
 *
 * Package:     gov.nih.nlm.umls.meta.impl
 * Class:       ComponentHistoryImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.ComponentHistory;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.ContentView;
import gov.nih.nlm.umls.meta.Locator;
import gov.nih.nlm.umls.meta.Source;

/**
 * Default implementation of {@link gov.nih.nlm.umls.meta.ComponentHistory}.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class ComponentHistoryImpl implements ComponentHistory {

  //
  // Fields
  //
  private Concept concept;
  private Source source;
  private String changetype;
  private String changeval;
  private String changekey;
  private String reason;
  private String sver;
  private String ui;
  private ContentView content_view;

  // Used to compute key fields more efficiently
  private boolean key_fields_flag;
  private String key_fields = null;

  /**
   * Instantiates an empty {@link ComponentHistory}.
   */
  public ComponentHistoryImpl() {};

  /**
   * Instantiates a {@link ComponentHistory} from the
   * specified {@link ComponentHistory}.
   * @param co the {@link ComponentHistory} to be copied
   */
  public ComponentHistoryImpl(ComponentHistory co) {
    setConcept(co.getConcept());
    setSource(co.getSource());
    setUi(co.getUi());
    setSver(co.getSver());
    setChangeType(co.getChangeType());
    setChangeKey(co.getChangeKey());
    setChangeValue(co.getChangeValue());
    setReason(co.getReason());
  }

  /**
   * Returns the release date or version number of a source (SVER).
   * @return the release date or version number of a source
   */
  public String getVersion() { return sver; }

  /**
   * Shorthand for {@link #getVersion()}.
   * @return the result of <code>getVersion()</code>
   */
  public String getSver() { return getVersion(); }

  /**
   * Sets the release date or version number of a source (SVER).
   * @param sver the release date or version number of a source
   */
  public void setVersion(String sver) {
    this.sver = sver;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setVersion(String)}.
   * @param sver the release date or version number of a source
   */
  public void setSver(String sver) { setVersion(sver); }

  /**
   * Returns the change key which identifies what changed (CHANGEKEY).
   * @return the change key
   */
  public String getChangeKey() {
    return changekey;
  }

  /**
   * Sets the change key to identify what changed (CHANGEKEY).
   * @param changekey the change key
   */
  public void setChangeKey(String changekey) {
    this.changekey = changekey;
    key_fields_flag = false;
  }

  /**
   * Returns the change type which identifies what changed (CHANGETYPE).
   * @return the change type
   */
  public String getChangeType() {
    return changetype;
  }

  /**
   * Sets the change type to identify what changed (CHANGETYPE).
   * @param changetype the change type
   */
  public void setChangeType(String changetype) {
    this.changetype = changetype;
    key_fields_flag = false;
  }

  /**
   * Returns the new value for what changed (CHANGEVALUE).
   * @return the change value
   */
  public String getChangeValue() {
    return changeval;
  }

  /**
   * Sets the new value for what changed (CHANGEVALUE).
   * @param changeval the change value
   */
  public void setChangeValue(String changeval) {
    this.changeval = changeval;
    key_fields_flag = false;
  }

  /**
   * Returns the reason for the change (REASON).
   * @return the reason
   */
  public String getReason() {
    return reason;
  }

  /**
   * Sets the reason for the change (REASON).
   * @param reason reason for the changed
   */
  public void setReason(String reason) {
    this.reason = reason;
    key_fields_flag = false;
  }

  /**
   * Returns the {@link Concept} that holds this {@link ComponentHistory}.
   * @return the {@link Concept} that holds this {@link ComponentHistory}
   */
  public Concept getConcept() { return concept; }

  /**
   * Sets the {@link Concept} that holds this {@link ComponentHistory}.
   * @param concept the {@link Concept} that holds this {@link ComponentHistory}
   */
  public void setConcept(Concept concept) {
    this.concept = concept;
    key_fields_flag = false;
  }

  /**
   * Returns the source for this {@link ComponentHistory}.
   * @return {@link Source}
   */
  public Source getSource() { return source; }

  /**
   * Sets the source for this {@link ComponentHistory}.
   * @param source {@link Source}
   */
  public void setSource(Source source) {
    this.source = source;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #getIdentifier()}.
   * @return the result of <code>getIdentifier()</code>
   */
  public String getUi() { return getIdentifier(); }

  /**
   * Returns the unique identifier (UI).
   * @return the unique identifier
   */
  public String getIdentifier() { return ui; }

  /**
   * Sets the unique identifier (UI).
   * @param ui the new value
   */
  public void setIdentifier(String ui) {
    this.ui = ui;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setIdentifier(String)}.
   * @param ui the new value
   */
  public void setUi(String ui) {
    setIdentifier(ui);
  }

  /**
   * Returns tbe {@link ContentView} definition for this {@link Locator} (CVF).
   * @return tbe {@link ContentView} definition
   */
  public ContentView getContentView() {
    return content_view;
  }

  /**
   * Sets tbe {@link ContentView} definition for this {@link Locator} (CVF).
   * @param cvf the {@link ContentView} definition
   */
  public void setContentView(ContentView cvf) {
    this.content_view = cvf;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #getContentView()}.
   * @return tbe {@link ContentView} definition
   */
  public ContentView getCvf() {
    return getContentView();
  }

  /**
   * Shorthand for {@link #setContentView(ContentView)}.
   * @param cvf the {@link ContentView} definition
   */
  public void setCvf(ContentView cvf) {
    setContentView(cvf);
  }

  //
  // Overridden Object Methods
  //
  /**
   * Return a hash code based on key fields.
   * @return a hash code based on key fields
   */
  public int hashCode() {
    return (getKeyFields()).hashCode();
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be a
   * {@link ComponentHistory} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof ComponentHistory)) return false;
    return getKeyFields().equals(((ComponentHistory)o).getKeyFields());
  }

  /**
   * Compares this {@link ComponentHistory} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof ComponentHistory)
      return getKeyFields().compareTo(((ComponentHistory)o).getKeyFields());
    else
      return 0;
  }

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from the ui, change,
   * newval, and description.
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
      StringBuffer sb1 = new StringBuffer(100);
      sb1.append(getConcept() != null ? getConcept().getCui() : "")
         .append(getUi())
         .append(getSource() != null ? getSource().getVsab() : "")
         .append(getSver())
         .append(getChangeType())
         .append(getChangeKey())
         .append(getChangeValue())
         .append(getReason());
      key_fields = sb1.toString();
      key_fields_flag = true;
    }
    return key_fields;
  }
}
