/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta.impl
 * Class:       ContextGroupImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.ContextGroup;
import gov.nih.nlm.umls.meta.StringContext;
import gov.nih.nlm.umls.meta.StringInfo;

import java.util.ArrayList;
import java.util.List;

/**
 * Default implementation of the {@link gov.nih.nlm.umls.meta.ContextGroup}
 * interface.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class ContextGroupImpl implements ContextGroup {

  //
  // Fields
  //
  private StringInfo string_info;
  private Concept concept;
  private List string_contexts = new ArrayList();

  // Used to compute key fields more efficiently
  private boolean key_fields_flag;
  private String key_fields = null;

  /**
   * Instantiates an empty {@link ContextGroup}.
   */
  public ContextGroupImpl() {};

  /**
   * Instantiates a {@link ContextGroup} from the specified {@link ContextGroup}.
   * @param cg {@link ContextGroup} to be copied
   */
  public ContextGroupImpl(ContextGroup cg) {
    setStringInfo(cg.getStringInfo());
    setConcept(cg.getConcept());
    setStringContexts(cg.getStringContexts());
  }

  /**
   * Return the {@link StringInfo}.
   * @return the {@link StringInfo}
   */
  public StringInfo getStringInfo() {
    return string_info;
  }

  /**
   * Sets the {@link StringInfo}.
   * @param string_info the {@link StringInfo}
   */
  public void setStringInfo(StringInfo string_info) {
    this.string_info = string_info;
  }

  /**
   * Returns all {@link StringContext}s.
   * @return an array of {@link StringContext}s
   */
  public StringContext [] getStringContexts() {
    return (StringContext[])string_contexts.toArray(new StringContext[0]);
  }

  /**
   * Returns all {@link StringContext}s as a {@link List}
   * @return a {@link List} of {@link StringContext}s
   */
  public List getStringContextsAsList() {
    return new ArrayList(string_contexts);
  }

  /**
   * Adds the specified {@link StringContext}.
   * @param sc the {@link StringContext} to be added
   */
  public void addStringContext(StringContext sc) {
    string_contexts.add(sc);
  }

  /**
   * Removes the specified {@link StringContext}.
   * @param sc the {@link StringContext} to be removed
   */
  public void removeStringContext(StringContext sc) {
    string_contexts.remove(sc);
  }

  /**
   * Removes all {@link StringContext}s.
   */
  public void clearStringContexts() {
    string_contexts.clear();
  }

  /**
   * Replaces the {@link StringContext}s with the ones in the specified array.
   * @param scs a {@link StringContext}<code>[]</code>
   */
  public void setStringContexts(StringContext [] scs) {
    string_contexts.clear();
    for (int i=0; i<scs.length; i++) {
      string_contexts.add(scs[i]);
    }
  }

  /**
   * replaces the {@link StringContext}s with the ones in the specified {@link List}.
   * @param scs a {@link List} of {@link StringContext}s
   */
  public void setStringContexts(List scs) {
    string_contexts = new ArrayList(scs);
  }

  /**
   * Returns the {@link Concept} that holds this {@link ContextGroup}.
   * @return the {@link Concept} that holds this {@link ContextGroup}
   */
  public Concept getConcept() {
    return concept;
  }

  /**
   * Sets the {@link Concept} that holds this {@link ContextGroup}.
   * @param concept the {@link Concept} that holds this {@link ContextGroup}
   */
  public void setConcept(Concept concept) {
    this.concept = concept;
    key_fields_flag = false;
  }

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from the cui, aui, sui,
   * versioned source, tty, and code.
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
      StringBuffer sb1 = new StringBuffer(20);
      sb1.append(getConcept().getCui())
         .append(getStringInfo().getSui());
      key_fields = sb1.toString();
      key_fields_flag = true;
    }
    return key_fields;
  }
  /**
   * Return a hash code based on key fields.
   * @return a hash code based on key fields
   */
  public int hashCode() {
    return (getKeyFields()).hashCode();
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be a
   * {@link ContextGroup} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof ContextGroup)) return false;
    return getKeyFields().equals(((ContextGroup)o).getKeyFields());
  }

  /**
   * Compares this {@link ContextGroup} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof ContextGroup)
      return getKeyFields().compareTo(((ContextGroup)o).getKeyFields());
    else
      return 0;
  }

}
