/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta.impl
 * Class:       ContextImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.AdditionalRelationshipLabel;
import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.ContentView;
import gov.nih.nlm.umls.meta.Context;
import gov.nih.nlm.umls.meta.ContextMember;
import gov.nih.nlm.umls.meta.Source;

import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;

/**
 * Default implementation of the {@link gov.nih.nlm.umls.meta.Context} interface.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class ContextImpl implements Context {

  //
  // Fields
  //
  private int cxn = 0;
  private String paui;
  private List context_members = new ArrayList();
  private List path_to_root = new ArrayList();
  private Atom atom;
  private Concept concept;
  private ContentView cvf;
  private Source source;
  private AdditionalRelationshipLabel rela;
  private String hcd;

  // Used to compute key fields more efficiently
  private boolean key_fields_flag;
  private String key_fields = null;

  /**
   * Instantiates an empty {@link Context}.
   */
  public ContextImpl() {};

  /**
   * Instantiates a {@link Context} from the specified {@link Context}.
   * @param cxt {@link Context} to be copied
   */
  public ContextImpl(Context cxt) {
    setAtom(cxt.getAtom());
    //getAtom().setCode(cxt.getAtom().getCode());
    //getAtom().setSui(cxt.getAtom().getSui());
    setConcept(cxt.getConcept());
    setSource(cxt.getSource());
    setCvf(cxt.getCvf());
    setPaui(cxt.getPaui());
    setCxn(cxt.getCxn());
    setContextMembers(cxt.getContextMembers());
    setPtr(cxt.getPtr());
  }

  /**
   * Returns the {@link Atom} this {@link Context} is connected to.
   * @return the {@link Atom} this {@link Context} is connected to
   */
  public Atom getAtom() {
    return atom;
  }

  /**
   * Sets the {@link Atom} this {@link Context} is connected to.
   * @param atom  the {@link Atom} this {@link Context} is connected to
   */
  public void setAtom(Atom atom) {
    this.atom = atom;
    key_fields_flag = false;
  }

  /**
   * Returns the context number (CXN).
   * @return the <code>int</code> context number
   */
  public int getContextNumber() {
    return cxn;
  }

  /**
   * Sets the context number (CXN).
   * @param cxn the <code>int</code> context number
   */
  public void setContextNumber(int cxn) {
    this.cxn = cxn;
    key_fields_flag = false;
  }

  /**
   * This is a Shorthand for {@link #getContextNumber()}
   * @return cxn context number <code>int</code>
   */
  public int getCxn() {
    return getContextNumber();
  }

  /**
   * Shorthand for {@link #setContextNumber(int)}.
   * @param cxn <code>int</code> context number
   */
  public void setCxn(int cxn) {
    setContextNumber(cxn);
  }

  /**
   * Shorthand for {@link #getParentAtomIdentifier()}.
   * @return the PAUI
   */
  public String getPaui() {
    return getParentAtomIdentifier();
  }

  /**
   * Shorthand for {@link #setParentAtomIdentifier(String)}.
   * @param paui the PAUI
   */
  public void setPaui(String paui) {
    setParentAtomIdentifier(paui);
  }

  /**
   * Returns the parent atom identifier (PAUI).
   * @return the PAUI
   */
  public String getParentAtomIdentifier() {
    return paui;
  }

  /**
   * Sets the parent atom identifier (PAUI).
   * @param paui the PAUI
   */
  public void setParentAtomIdentifier(String paui) {
    this.paui = paui;
    key_fields_flag = false;
  }

  /**
   * Returns the path to the root (PTR) as a "." separated list of AUI values.
   * @return the path to the root
   */
  public String getPathToRootAsString() {
    StringBuffer sb = new StringBuffer(200);
    for (int i=0; i<path_to_root.size(); i++) {
      sb.append(path_to_root.get(i));
      if (i<=path_to_root.size() -2) sb.append(".");
    }
    return sb.toString();
  }

  /**
   * Returns the path to the root (PTR) as a {@link List} of AUI values.
   * @return the path to the root as a {@link List} of AUI values
   */
  public List getPathToRootAsList() {
    return new ArrayList(path_to_root);
  }

  /**
   * Returns the path to the root (PTR) as a {@link String}<code>[]</code>
   * of AUI values.
   * @return the path to the root as a {@link String}<code>[]</code>
   * of AUI values
   */
  public String [] getPathToRoot() {
    return (String[])path_to_root.toArray(new String[0]);
  }

  /**
   * Shorthand for {@link #getPathToRootAsString()}.
   * @return the path to the root
   */
  public String getPtr() {
    return getPathToRootAsString();
  }

  /**
   * Sets the path to the root (PTR) as a "." separated list of AUI values.
   * @param ptr the path to the root
   */
  public void setPathToRoot(String ptr) {
    StringTokenizer st = new StringTokenizer(ptr, ".");
     while (st.hasMoreTokens()) {
         path_to_root.add(st.nextToken());
     }
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setPathToRoot(String)}.
   * @param ptr the path to the root
   */
  public void setPtr(String ptr) {
    setPathToRoot(ptr);
  }

  /**
   * Returns all {@link ContextMember}s.  These are the ancestors,
   * children, and siblings.
   * @return a sorted array of {@link ContextMember}s
   */
  public ContextMember [] getContextMembers() {
    return (ContextMember[])context_members.toArray(new ContextMember[0]);
  }

  /**
   * Returns all {@link ContextMember}s as a {@link List}.  These are the ancestors,
   * children, and siblings.
   * @return a sorted {@link List} of {@link ContextMember}s
   */
  public List getContextMembersAsList() {
    return new ArrayList(context_members);
  }

  /**
   * Adds the specified {@link ContextMember}.
   * @param cm the {@link ContextMember} to be added
   */
  public void addContextMember(ContextMember cm) {
    context_members.add(cm);
  }

  /**
   * Removes the specified {@link ContextMember}.
   * @param cm the {@link ContextMember} to be removed
   */
  public void removeContextMember(ContextMember cm) {
    context_members.remove(cm);
  }

  /**
   * Removes all {@link ContextMember}s.
   */
  public void clearContextMembers() {
    context_members.clear();
  }

  /**
   * Replaces the {@link ContextMember}s with the ones in the specified array.
   * @param cms a {@link ContextMember}<code>[]</code>
   */
  public void setContextMembers(ContextMember [] cms) {
    context_members.clear();
    for (int i=0; i<cms.length; i++) {
      context_members.add(cms[i]);
    }
  }

  /**
   * Replaces the {@link ContextMember}s with the ones in the specified {@link List}.
   * @param cms a {@link List} of {@link ContextMember}s
   */
  public void setContextMembers(List cms) {
    context_members = new ArrayList(cms);
  }

  /**
   * Returns the {@link Concept} that holds this {@link Context}
   * @return the {@link Concept} that holds this {@link Context}
   */
  public Concept getConcept() {
    return concept;
  }

  /**
   * Sets the {@link Concept} that holds this {@link Context}
   * @param concept the {@link Concept} that holds this {@link Context}
   */
  public void setConcept(Concept concept) {
    this.concept = concept;
    key_fields_flag = false;
  }

  /**
   * Returns the source for this {@link Context}.
   * @return {@link Source}
   */
  public Source getSource() {
    return source;
  }

  /**
   * Sets the source for this {@link Context}.
   * @param source {@link Source}
   */
  public void setSource(Source source) {
    this.source = source;
    key_fields_flag = false;
  }

  /**
   * Returns tbe {@link ContentView} definition for this {@link Context} (CVF).
   * @return tbe {@link ContentView} definition
   */
  public ContentView getContentView() {
    return cvf;
  }

  /**
   * Sets tbe {@link ContentView} definition for this {@link Context} (CVF).
   * @param cvf the {@link ContentView} definition
   */
  public void setContentView(ContentView cvf) {
    this.cvf = cvf;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #getContentView()}.
   * @return tbe {@link ContentView} definition
   */
  public ContentView getCvf() {
    return getContentView();
  }

  /**
   * Shorthand for {@link #setContentView(ContentView)}.
   * @param cvf the {@link ContentView} definition
   */
  public void setCvf(ContentView cvf) {
    setContentView(cvf);
  }

  /**
   * Returns the hierarchical code (HCD).
   * @return the hierarchical code
   */
  public String getHierarchicalCode() {
    return hcd;
  }

  /**
   * Shorthand for {@link #getHierarchicalCode()}.
   * @return the hierarchical code
   */
  public String getHcd() {
    return getHierarchicalCode();
  }

  /**
   * Sets the hierarchical code (HCD).
   * @param hcd the hierarchical code
   */
  public void setHierarchicalCode(String hcd) {
    this.hcd = hcd;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setHierarchicalCode(String)}.
   * @param hcd the hierarchical code
   */
  public void setHcd(String hcd) {
    setHierarchicalCode(hcd);
  }

  /**
   * Returns the {@link AdditionalRelationshipLabel} (RELA) from the context member to its parent.
   * @return the {@link AdditionalRelationshipLabel}
   */
  public AdditionalRelationshipLabel getAdditionalRelationshipLabel() {
    return rela;
  }

  /**
   * Shorthand for {@link #getAdditionalRelationshipLabel()}.
   * @return the {@link AdditionalRelationshipLabel}
   */
  public AdditionalRelationshipLabel getRela() {
    return getAdditionalRelationshipLabel();
  }

  /**
   * Sets the {@link AdditionalRelationshipLabel} (RELA) from the context member to its parent.
   * @param rela the {@link AdditionalRelationshipLabel}
   */
  public void setAdditionalRelationshipLabel(AdditionalRelationshipLabel rela) {
    this.rela = rela;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setAdditionalRelationshipLabel(AdditionalRelationshipLabel)}.
   * @param rela the {@link AdditionalRelationshipLabel}
   */
  public void setRela(AdditionalRelationshipLabel rela) {
    setAdditionalRelationshipLabel(rela);
  }

  /**
   * Returns the {@link ContextMember} that has the context
   * member label (CXL) of "CCP" or self.
   * @return {@link ContextMember} with label "CCP"
   */
  public ContextMember getSelfMember() {
    ArrayList all_members = (ArrayList)getContextMembersAsList();
    for (int i=0; i<all_members.size(); i++) {
      if (((ContextMember)all_members.get(i)).getLabel().equals("CCP"))
        return (ContextMember)all_members.get(i);
    }
    return null;
  }

  /**
   * Returns only the {@link ContextMember}s that have the context
   * member label (CXL) of "CHD" or child.
   * @return {@link ContextMember}[] of members with label "CHD"
   */
  public ContextMember[] getChildMembers() {
    ArrayList all_members = (ArrayList)getContextMembersAsList();
    ArrayList members = new ArrayList();
    for (int i=0; i<all_members.size(); i++) {
      if (((ContextMember)all_members.get(i)).getLabel().equals("CHD"))
        members.add(all_members.get(i));
    }
    return (ContextMember[])members.toArray(new ContextMember[0]);
  }

  /**
   * Returns only the {@link ContextMember}s that have the context
   * member label (CXL) of "CHD" or child.
   * @return {@link ContextMember}[] of members with label "CHD"
   */
  public List getChildMembersAsList() {
    ArrayList all_members = (ArrayList)getContextMembersAsList();
    ArrayList members = new ArrayList();
    for (int i=0; i<all_members.size(); i++) {
      if (((ContextMember)all_members.get(i)).getLabel().equals("CHD"))
        members.add(all_members.get(i));
    }
    return members;
  }

  /**
   * Returns only the {@link ContextMember}s that have the context
   * member label (CXL) of "ANC" or self.
   * @return {@link ContextMember}[] of members with label "ANC"
   */
  public ContextMember[] getAncestorMembers() {
    ArrayList all_members = (ArrayList)getContextMembersAsList();
    ArrayList members = new ArrayList();
    for (int i=0; i<all_members.size(); i++) {
      if (((ContextMember)all_members.get(i)).getLabel().equals("ANC"))
        members.add(all_members.get(i));
    }
    return (ContextMember[])members.toArray(new ContextMember[0]);
  }

  /**
   * Returns only the {@link ContextMember}s that have the context
   * member label (CXL) of "ANC" or self.
   * @return {@link ContextMember}[] of members with label "ANC"
   */
  public List getAncestorMembersAsList() {
    ArrayList all_members = (ArrayList)getContextMembersAsList();
    ArrayList members = new ArrayList();
    for (int i=0; i<all_members.size(); i++) {
      if (((ContextMember)all_members.get(i)).getLabel().equals("ANC"))
        members.add(all_members.get(i));
    }
    return members;
  }

  /**
   * Returns only the {@link ContextMember}s that have the context
   * member label (CXL) of "SIB" or self.
   * @return {@link ContextMember}[] of members with label "SIB"
   */
  public ContextMember[] getSiblingMembers() {
    ArrayList all_members = (ArrayList)getContextMembersAsList();
    ArrayList members = new ArrayList();
    for (int i=0; i<all_members.size(); i++) {
      if (((ContextMember)all_members.get(i)).getLabel().equals("SIB"))
        members.add(all_members.get(i));
    }
    return (ContextMember[])members.toArray(new ContextMember[0]);
  }

  /**
   * Returns only the {@link ContextMember}s that have the context
   * member label (CXL) of "SIB" or self.
   * @return {@link ContextMember}[] of members with label "SIB"
   */
  public List getSiblingMembersAsList() {
    ArrayList all_members = (ArrayList)getContextMembersAsList();
    ArrayList members = new ArrayList();
    for (int i=0; i<all_members.size(); i++) {
      if (((ContextMember)all_members.get(i)).getLabel().equals("SIB"))
        members.add(all_members.get(i));
    }
    return members;
  }

  /**
   * Indicates if context is of the top level.
   * @return <code>boolean</code> true if top level
   *                              false if otherwise
   */
  public boolean isTopLevel() {
    return "".equals(getParentAtomIdentifier());
  }

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from the cui, aui, sui,
   * versioned source, cxn, and code.
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
      final StringBuffer sb1 = new StringBuffer(60);
      sb1.append(getConcept() != null ? getConcept().getCui() : "")
         .append(getAtom() != null ? getAtom().getSui() : "")
         .append(getAtom() != null ? getAtom().getAui() : "")
         .append(getSource() != null ? getSource().getVsab() : "")
         .append(getAtom() != null ? getAtom().getCode() : "")
         .append(getCxn());
      key_fields = sb1.toString();
      key_fields_flag = true;
    }
    return key_fields;
  }

  /**
   * Return a hash code based on key fields.
   * @return a hash code based on key fields
   */
  public int hashCode() {
    return getKeyFields().hashCode();
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be a
   * {@link Context} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof Context)) return false;
    return getKeyFields().equals(((Context)o).getKeyFields());
  }

  /**
   * Compares this {@link Source} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof Context)
      return getKeyFields().compareTo(((Context)o).getKeyFields());
    else
      return 0;
  }

}
