/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta.impl
 * Class:       ContextMemberImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.AdditionalRelationshipLabel;
import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.ContextMember;

/**
 * Default implementation of the {@link gov.nih.nlm.umls.meta.ContextMember}
 * interface.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class ContextMemberImpl implements ContextMember {

  //
  // Fields
  //
  private int rank = 0;
  private AdditionalRelationshipLabel rela;
  private Concept related_concept;
  private Atom related_atom;
  private String hcd;
  private String xc;
  private String cxl;

  // Used to compute key fields more efficiently
  protected boolean key_fields_flag;
  protected String key_fields = null;

  /**
   * Instantiates an empty {@link ContextMember}.
   */
  public ContextMemberImpl() {};

  /**
   * Instantiates a {@link ContextMember} from the specified {@link ContextMember}.
   * @param cm {@link ContextMember} to be copied
   */
  public ContextMemberImpl(ContextMember cm) {
    setRank(cm.getRank());
    setRela(cm.getRela());
    setRelatedConcept(cm.getRelatedConcept());
    setHcd(cm.getHcd());
    setXc(cm.getXc());
    setCxl(cm.getCxl());
  }

  /**
   * Returns the rank (RNK) if the member is an ancestor.
   * @return the <code>int</code> rank
   */
  public int getRank() {
    return rank;
  }

  /**
   * Sets the rank (RNK) if the member is an ancestor
   * @param rnk the <code>int</code> rank
   */
  public void setRank(int rnk) {
    this.rank = rnk;
    key_fields_flag = false;
  }

  /**
   * Returns the related concept of the ancestor, sibling, or child.
   * @return the related concept of the ancestor, sibling, or child
   */
  public Concept getRelatedConcept() {
    return related_concept;
  }

  /**
   * Sets the related concept of the ancestor, sibling, or child.
   * @param related_concept the related concept of the ancestor, sibling, or child
   */
  public void setRelatedConcept(Concept related_concept) {
    this.related_concept = related_concept;
    key_fields_flag = false;
  }


  /**
   * Returns the related Atom of the ancestor, sibling, or child.
   * @return the related Atom of the ancestor, sibling, or child
   */
  public Atom getRelatedAtom() {
    return related_atom;
  }

  /**
   * Sets the related atom of the ancestor, sibling, or child.
   * @param related_atom the related atom of the ancestor, sibling, or child
   */
  public void setRelatedAtom(Atom related_atom) {
    this.related_atom = related_atom;
    key_fields_flag = false;
  }

  /**
   * Returns the hierarchical code (HCD).
   * @return the hierarchical code
   */
  public String getHierarchicalCode() {
    return hcd;
  }

  /**
   * Shorthand for {@link #getHierarchicalCode()}.
   * @return the hierarchical code
   */
  public String getHcd() {
    return getHierarchicalCode();
  }

  /**
   * Sets the hierarchical code (HCD).
   * @param hcd the hierarchical code
   */
  public void setHierarchicalCode(String hcd) {
    this.hcd = hcd;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setHierarchicalCode(String)}.
   * @param hcd the hierarchical code
   */
  public void setHcd(String hcd) {
    setHierarchicalCode(hcd);
  }

  /**
   * Returns the {@link AdditionalRelationshipLabel} (RELA) from the context member to its parent.
   * @return the {@link AdditionalRelationshipLabel}
   */
  public AdditionalRelationshipLabel getAdditionalRelationshipLabel() {
    return rela;
  }

  /**
   * Shorthand for {@link #getAdditionalRelationshipLabel()}.
   * @return the {@link AdditionalRelationshipLabel}
   */
  public AdditionalRelationshipLabel getRela() {
    return getAdditionalRelationshipLabel();
  }

  /**
   * Sets the {@link AdditionalRelationshipLabel} (RELA) from the context member to its parent.
   * @param rela the {@link AdditionalRelationshipLabel}
   */
  public void setAdditionalRelationshipLabel(AdditionalRelationshipLabel rela) {
    this.rela = rela;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setAdditionalRelationshipLabel(AdditionalRelationshipLabel)}.
   * @param rela the {@link AdditionalRelationshipLabel}
   */
  public void setRela(AdditionalRelationshipLabel rela) {
    setAdditionalRelationshipLabel(rela);
  }

  /**
   * Returns the "has child flag" value (XC).
   * @return the "has child flag" value
   */
  public String getXc() {
    return (xc.equals("+") ? "+" : "");
  }

  /**
   * Sets the "has child flag" value (XC).
   * @param xc the "has child flag" value
   */
  public void setXc(String xc) {
    this.xc = xc;
    key_fields_flag = false;
  }

  /**
   * Indicates whether or not this context member has children.
   * @return <code>true</code> if {@link ContextMember} has a child,
   *         <code>false</code> if {@link ContextMember} does not have a child
   */
  public boolean hasChild() {
    return ((xc.equals("")) ? false : true);
  }

  /**
   * Sets the flag indicating whether or not this {@link ContextMember} has a child.
   * @param xc  <code>true</code> if {@link ContextMember} has a child,
   *            <code>false</code> if {@link ContextMember} does not have a child
   */
  public void setHasChild(boolean xc) {
    if (xc)
      this.xc = "+";
    else
      this.xc = "";
    key_fields_flag = false;
  }

  /**
   * Returns the label for this {@link ContextMember} (CXL).
   * @return the label for this {@link ContextMember}
   */
  public String getLabel() {
    return cxl;
  }

  /**
   * Shorthand for {@link #getLabel()}.
   * @return the label for this {@link ContextMember}
   */
  public String getCxl() {
    return getLabel();
  }

  /**
   * Sets the label for this {@link ContextMember}.
   * @param cxl context label
   */
  public void setLabel(String cxl) {
    this.cxl = cxl;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setLabel(String)}.
   * @param cxl the label for this {@link ContextMember}
   */
  public void setCxl(String cxl) {
    setLabel(cxl);
  }

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from the cxl, rank, str
   * cui2, and aui2.
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
      final StringBuffer sb1 = new StringBuffer(20);
      sb1.append(getCxl())
         .append(getRank())
         .append(getRelatedConcept() != null ? getRelatedConcept().getCui() : "");
      key_fields = sb1.toString();
      key_fields_flag = true;
    }
    return key_fields;
  }

  /**
   * Return a hash code based on key fields.
   * @return a hash code based on key fields
   */
  public int hashCode() {
    return getKeyFields().hashCode();
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be a
   * {@link ContextMember} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof ContextMember)) return false;
    return getKeyFields().equals(((ContextMember)o).getKeyFields());
  }

  /**
   * Compares this {@link ContextMember} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof ContextMember)
      return getKeyFields().compareTo(((ContextMember)o).getKeyFields());
    else
      return 0;
  }

}
