/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta.impl
 * Class:       CoocurrenceImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.ContentView;
import gov.nih.nlm.umls.meta.Coocurrence;
import gov.nih.nlm.umls.meta.CoocurrenceType;
import gov.nih.nlm.umls.meta.QualifierFrequency;
import gov.nih.nlm.umls.meta.Source;

import java.util.ArrayList;
import java.util.List;

/**
 * Default implementation of the {@link gov.nih.nlm.umls.meta.Coocurrence}
 * interface.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class CoocurrenceImpl implements Coocurrence {

  //
  // Fields
  //
  private Concept concept;
  private List qualifier_freq = null;
  private Concept related_concept;
  private Source source;
  private CoocurrenceType cot;
  private int cof;
  private String coa;
  private ContentView cvf;

  // Used to compute key fields more efficiently
  protected boolean key_fields_flag;
  protected String key_fields = null;

  /**
   * Instantiates an empty {@link Coocurrence}.
   */
  public CoocurrenceImpl() {};

  /**
   * Instantiates a {@link Coocurrence} from the specified {@link Coocurrence}.
   * @param co {@link Coocurrence} to be copied
   */
  public CoocurrenceImpl(Coocurrence co) {
    setQualifiers(co.getQualifiers());
    setRelatedConcept(co.getRelatedConcept());
    setConcept(co.getConcept());
    setSource(co.getSource());
    setCot(co.getCot());
    setCof(co.getCof());
    setCoa(co.getCoa());
    setCvf(co.getCvf());
  }

  /**
   * Returns the related concept involved.
   * @return the related concept involved
   */
  public Concept getRelatedConcept() {
    return related_concept;
  }

  /**
   * Sets the related concept involved.
   * @param related_concept the related concept involved
   */
  public void setRelatedConcept(Concept related_concept) {
    this.related_concept = related_concept;
    key_fields_flag = false;
  }

  /**
   * Returns the {@link CoocurrenceType} (COT).
   * @return the {@link CoocurrenceType}
   */
  public CoocurrenceType getCoocurrenceType() {
    return cot;
  }

  /**
   * Shorthand for {@link #getCoocurrenceType()}.
   * @return the {@link CoocurrenceType}
   */
  public CoocurrenceType getCot() {
    return getCoocurrenceType();
  }

  /**
   * Sets the {@link CoocurrenceType} (COT).
   * @param cot the {@link CoocurrenceType}
   */
  public void setCoocurrenceType(CoocurrenceType cot) {
    this.cot = cot;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setCoocurrenceType(CoocurrenceType)}.
   * @param cot the {@link CoocurrenceType}
   */
  public void setCot(CoocurrenceType cot) {
    setCoocurrenceType(cot);
  }

  /**
   * Returns the coocurrence frequency (COF).
   * @return the coocurrence frequency
   */
  public int getCoocurrenceFrequency() {
    return cof;
  }

  /**
   * Shorthand for {@link #getCoocurrenceFrequency()}.
   * @return the coocurrence frequency
   */
  public int getCof() {
    return getCoocurrenceFrequency();
  }

  /**
   * Sets the coocurrence frequency (COF).
   * @param cof the coocurrence frequency
   */
  public void setCoocurrenceFrequency(int cof) {
    this.cof = cof;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setCoocurrenceFrequency(int)}.
   * @param cof the coocurrence frequency
   */
  public void setCof(int cof) {
    setCoocurrenceFrequency(cof);
  }

  /**
   * Shorthand for {@link #getCoocurrenceAttribute()}.
   * @return the coocurrence attribute value
   */
  public String getCoa() {
    return getCoocurrenceAttribute();
  }

  /**
   * Returns the coocurrence attribute value (COA).
   * @return the coocurrence attribute value
   */
  public String getCoocurrenceAttribute() {
    return coa;
  }

  /**
   * Sets the coocurrence attribute value (COA). This is a comma separated
   * list of nqualifier=frequency pairs.
   * @param coa the coocuurence attribute value
   */
  public void setCoocurrenceAttribute(String coa) {
    this.coa = coa;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setCoocurrenceAttribute(String)}.
   * @param coa the coocuurence attribute value
   */
  public void setCoa(String coa) {
    setCoocurrenceAttribute(coa);
  }

  /**
   * Returns the {@link QualifierFrequency} list.
   * @return an array of {@link QualifierFrequency}s
   */
  public QualifierFrequency [] getQualifiers() {
    return qualifier_freq == null ?
      new QualifierFrequency[0] :
      (QualifierFrequency[])qualifier_freq.toArray(new QualifierFrequency[0]);
  }

  /**
   * Returns the {@link QualifierFrequency} values as a {@link List}.
   * @return the {@link QualifierFrequency} values as a {@link List}
   */
  public List getQualifierAsList() {
    return qualifier_freq == null ?
      new ArrayList() :
      new ArrayList(qualifier_freq);
  }

  /**
   * Replaces the {@link QualifierFrequency} values with those in the
   * specified array.
   * @param coa an array of {@link QualifierFrequency} values
   */
  public void setQualifiers(QualifierFrequency [] coa) {
    qualifier_freq = new ArrayList();
    for (int i=0; i<coa.length; i++) {
      qualifier_freq.add(coa[i]);
    }
  }

  /**
   * Replaces the {@link QualifierFrequency} values with those in the
   * specified {@link List}
   * @param coa a {@link List} of {@link QualifierFrequency} values
   */
  public void setQualifiers(List coa) {
    qualifier_freq = new ArrayList(coa);
  }

  /**
   * Returns the {@link Concept} that holds this {@link Coocurrence}.
   * @return the {@link Concept} that holds this {@link Coocurrence}
   */
  public Concept getConcept() {
    return concept;
  }

  /**
   * Sets the {@link Concept} that holds this {@link Coocurrence}.
   * @param concept the {@link Concept} that holds this {@link Coocurrence}
   */
  public void setConcept(Concept concept) {
    this.concept = concept;
    key_fields_flag = false;
  }


  /**
   * Returns the source for this {@link Coocurrence}.
   * @return {@link Source}
   */
  public Source getSource() {
    return source;
  }

  /**
   * Sets the source for this {@link Coocurrence}.
   * @param source {@link Source}
   */
  public void setSource(Source source) {
    this.source = source;
    key_fields_flag = false;
  }

  /**
   * Returns tbe {@link ContentView} definition for this {@link Coocurrence} (CVF).
   * @return tbe {@link ContentView} definition
   */
  public ContentView getContentView() {
    return cvf;
  }

  /**
   * Sets tbe {@link ContentView} definition for this {@link Coocurrence} (CVF).
   * @param cvf the {@link ContentView} definition
   */
  public void setContentView(ContentView cvf) {
    this.cvf = cvf;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #getContentView()}.
   * @return tbe {@link ContentView} definition
   */
  public ContentView getCvf() {
    return getContentView();
  }

  /**
   * Shorthand for {@link #setContentView(ContentView)}.
   * @param cvf the {@link ContentView} definition
   */
  public void setCvf(ContentView cvf) {
    setContentView(cvf);
  }

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from the cui1, aui1, aui2,
   * versioned source, cui2, cof and cot.
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
      final StringBuffer sb = new StringBuffer(70);
      sb.append(getConcept() != null ? getConcept().getCui() : "")
         .append(getRelatedConcept() != null ? getRelatedConcept().getCui() : "")
         .append(getSource() != null ? getSource().getVsab() : "")
         .append(getCot() != null ? getCot().getAbbreviation() : "")
         .append(getCoa())
         .append(getCof());
      key_fields = sb.toString();
      key_fields_flag = true;
    }
    return key_fields;
  }

  /**
   * Return a hash code based on key fields.
   * @return a hash code based on key fields
   */
  public int hashCode() {
    return (getKeyFields()).hashCode();
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be a
   * {@link Coocurrence} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof Coocurrence)) return false;
    return getKeyFields().equals(((Coocurrence)o).getKeyFields());
  }

  /**
   * Compares this {@link Coocurrence} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof Coocurrence)
      return getKeyFields().compareTo(((Coocurrence)o).getKeyFields());
    else
      return 0;
  }

}
