/**************************************************************************
 * Package:	gov.nih.nlm.umls.meta.impl
 * Class:	IndexEntryImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.IndexEntry;

/**
 * Default implementation of the {@link IndexEntry} interface.
 *
 * @author Deborah Shapiro, Brian Carlsen
 */
public class IndexEntryImpl implements IndexEntry {

  //
  // Private Fields
  //
  private String cui; //unique identifier for concept
  private String lui; //unique identifier for term
  private String sui; //unique identifier for string
  private String lat; //language of term
  private String value; // word or string

  // Used to compute key fields more efficiently
  private boolean key_fields_flag;
  private String key_fields = null;

  //
  // Constructor
  //

  /**
   * Instantiates an empty {@link IndexEntry}.
   */
  public IndexEntryImpl() {};

  /**
   * Instantiates an {@link IndexEntry} from the specified {@link IndexEntry}.
   * @param ie {@link IndexEntry} to be copied
   */
  public IndexEntryImpl(IndexEntry ie) {
    setCui(ie.getCui());
    setLat(ie.getLat());
    setLui(ie.getLui());
    setSui(ie.getSui());
    setValue(ie.getValue());
  }

  //
  // Methods
  //

  /**
   * Return a hash code based on all fields.
   * @return a hash code based on all fields
   */
  public int hashCode() {
    return (getKeyFields().hashCode());
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be an
   * {@link IndexEntry} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if(o == null || !(o instanceof IndexEntry)) return false;
    return getKeyFields().equals(((IndexEntry)o).getKeyFields());
  }

  /**
   * Returns a {@link String} representation of the index entry.
   * @return a {@link String} representation of the index entry
   */
  public String toString() {
    return "cui:" + getCui() + " lui:" + getLui() +
    " sui:" + getSui() + " lat:" + getLat() +
    " value:" + getValue();
  }

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from the cui, aui, sui,
   * versioned source, tty, and code.
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
      final StringBuffer sb1 = new StringBuffer(50);
      sb1.append(getCui())
         .append(getLui())
         .append(getSui())
         .append(getLat())
         .append(getValue());
      key_fields = sb1.toString();
      key_fields_flag = true;
    }
    return key_fields;
  }

  /**
   * Compares this {@link IndexEntry} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof IndexEntry)
      return getKeyFields().compareTo(((IndexEntry)o).getKeyFields());
    else
      return 0;
  }


  //
  // Accessor Methods
  //

  /**
   * Returns a {@link String} representation of the CUI.
   * @return a {@link String} representation of the CUI
   */
  public String getCui() {
    return getConceptIdentifier();
  }

  /**
   * Sets the CUI.
   * @param cui the {@link String} CUI value
   */
  public void setCui(String cui) {
    setConceptIdentifier(cui);
  }

  /**
   * Returns a {@link String} representation of the CUI.
   * @return a {@link String} representation of the CUI
   */
  public String getConceptIdentifier() {
    return cui;
  }

  /**
   * Sets the CUI.
   * @param cui the {@link String} CUI value
   */
  public void setConceptIdentifier(String cui) {
    this.cui = cui;
    key_fields_flag = false;
  }

  /**
   * Returns a {@link String} representation of the LUI.
   * @return a {@link String} representation of the LUI
   */
  public String getLui() {
    return getTermIdentifier();
  }

  /**
   * Sets the LUI.
   * @param lui the {@link String} LUI value
   */
  public void setLui(String lui ) {
    setTermIdentifier(lui);
  }

  /**
   * Returns the term identifier (LUI).
   * @return the LUI
   */
  public String getTermIdentifier() {
    return lui;
  }

  /**
   * Sets the term identifier (LUI).
   * @param lui the LUI
   */
  public void setTermIdentifier(String lui) {
    this.lui = lui;
    key_fields_flag = false;
  }

  /**
   * Returns a {@link String} representation of the SUI.
   * @return a {@link String} representation of the SUI
   */
  public String getSui() {
    return getStringIdentifier();
  }

  /**
   * Sets the SUI.
   * @param sui a {@link String} SUI value
   */
  public void setSui(String sui) {
    setStringIdentifier(sui);
  }

  /**
   * Return the string identifier (SUI).
   * @return the SUI
   */
  public String getStringIdentifier() {
    return sui;
  }

  /**
   * Sets the string identifier (SUI).
   * @param sui the SUI
   */
  public void setStringIdentifier(String sui) {
    this.sui = sui;
    key_fields_flag = false;
  }


  /**
   * Returns a {@link String} representation of the language.
   * @return a {@link String} representation of the language
   */
  public String getLat() {
    return getLanguage();
  }

  /**
   * Sets the language abbreviation.
   * @param lat the language abbreviation
   */
  public void setLat(String lat) {
    setLanguage(lat);
  }

  /**
   * Returns a {@link String} representation of the language.
   * @return a {@link String} representation of the language
   */
  public String getLanguage() {
    return lat;
  }

  /**
   * Sets the language abbreviation.
   * @param lat the language abbreviation
   */
  public void setLanguage(String lat) {
    this.lat = lat;
    key_fields_flag = false;
  }

  /**
   * Returns the {@link String} that is being indexed.
   * @return the {@link String} that is being indexed
   */
  public String getValue() {
    return value;
  }

  /**
   * Set the {@link String} being indexed
   * @param value the {@link String} being indexed
   */
  public void setValue(String value) {
    this.value= value;
    key_fields_flag = false;
  }

}

