/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta.impl
 * Class:       LocatorImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.ContentView;
import gov.nih.nlm.umls.meta.Locator;
import gov.nih.nlm.umls.meta.Source;

/**
 * Default implementation of the {@link gov.nih.nlm.umls.meta.Locator} interface.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class LocatorImpl implements Locator {

  //
  // Fields
  //
  private Concept concept;
  private Atom atom;
  private String atui;
  private String satui;
  private int fr = 0;
  private String un;
  private String sui;
  private String sna;
  private String soui;
  private Source source;
  private ContentView content_view;

  // Used to compute key fields more efficiently
  private boolean key_fields_flag;
  private String key_fields = null;

  /**
   * Instantiates an empty {@link Locator}.
   */
  public LocatorImpl() {};

  /**
   * Instantiates a {@link Locator} from the specified {@link Locator}.
   * @param loc {@link Locator} to be copied
   */
  public LocatorImpl(Locator loc) {
    setAtom(loc.getAtom());
    setFr(loc.getFr());
    setUn(loc.getUn());
    setSna(loc.getSna());
    setSoui(loc.getSoui());
    setAtui(loc.getAtui());
    setSatui(loc.getSatui());
    setConcept(loc.getConcept());
    setSource(loc.getSource());
    setCvf(loc.getCvf());
    getAtom().setSui(loc.getAtom().getSui());
  }

  /**
   * Returns the {@link Atom} related to this {@link Locator}.
   * @return the {@link Atom} related to this {@link Locator}
   */
  public Atom getAtom() {
    return atom;
  }

  /**
   * Sets the {@link Atom} related to this {@link Locator}.
   * @param atom the {@link Atom} related to this {@link Locator}
   */
  public void setAtom(Atom atom) {
    this.atom = atom;
    key_fields_flag = false;
  }

  /**
   * Returns the unique identifier (ATUI).
   * @return the ATUI
   */
  public String getAttributeIdentifier() {
    return atui;
  }

  /**
   * Shorthand for {@link #getAttributeIdentifier()}.
   * @return the ATUI
   */
  public String getAtui() {
    return getAttributeIdentifier();
  }

  /**
   * Sets the unique identifier (ATUI).
   * @param atui the ATUI
   */
  public void setAttributeIdentifier(String atui) {
    this.atui = atui;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setAttributeIdentifier(String)}.
   * @param atui the ATUI
   */
  public void setAtui(String atui) {
    setAttributeIdentifier(atui);
  }

  /**
   * Returns the source asserted attribute identifier (SATUI).
   * @return the SATUI
   */
  public String getSourceAttributeIdentifier() {
    return satui;
  }

  /**
   * Shorthand for {@link #getSourceAttributeIdentifier()}.
   * @return the SATUI
   */
  public String getSatui() {
    return getSourceAttributeIdentifier();
  }

  /**
   * Sets the source asserted attribute identifier (SATUI).
   * @param satui the SATUI
   */
  public void setSourceAttributeIdentifier(String satui) {
    this.satui = satui;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setSourceAttributeIdentifier(String)}.
   * @param satui the SATUI
   */
  public void setSatui(String satui) {
    setSourceAttributeIdentifier(satui);
  }

  /**
   * Returns the {@link Concept} that holds this {@link Locator}.
   * @return the {@link Concept} that holds this {@link Locator}
   */
  public Concept getConcept() {
    return concept;
  }

  /**
   * Sets the {@link Concept} that holds this {@link Locator}.
   * @param concept the {@link Concept} that holds this {@link Locator}
   */
  public void setConcept(Concept concept) {
    this.concept = concept;
    key_fields_flag = false;
  }

  /**
   * Returns the frequency (FR).
   * @return the frequency
   * @see #getFrequencyMeaning()
   */
  public int getFrequency() {
    return fr;
  }

  /**
   * Shorthand for {@link #getFrequency()}.
   * @return the frequency
   */
  public int getFr() {
    return getFrequency();
  }

  /**
   * Sets the frequency (FR).
   * @param fr the frequency
   */
  public void setFrequency(int fr) {
    this.fr = fr;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setFrequency(int)}.
   * @param fr the frequency
   */
  public void setFr(int fr) {
    setFrequency(fr);
  }

  /**
   * Shorthand for {@link #getFrequencyMeaning()}.
   * @return the meaning of the frequency, if the frequency is set.
   */
  public String getUn() {
    return getFrequencyMeaning();
  }

  /**
   * Returns the meaning of the frequency (UN), if the frequency is set.
   * @return the meaning of the frequency, if the frequency is set.
   */
  public String getFrequencyMeaning() {
    return un;
  }

  /**
   * Shorthand for {@link #setFrequencyMeaning(String)}.
   * @param un the meaning of the frequency, if the frequency is set.
   */
  public void setUn(String un) {
    setFrequencyMeaning(un);
  }

  /**
   * Returns the meaning of the frequency (UN), if the frequency is set.
   * @param un the meaning of the frequency, if the frequency is set.
   */
  public void setFrequencyMeaning(String un) {
    this.un = un;
    key_fields_flag = false;
  }

  /**
   * Returns the actual name that occurs in the information source (SNA)
   * if not otherwise present in the Metathesaurus.
   * @return the actual name that occurs in the information source
   */
  public String getExternalName() {
    return sna;
  }

  /**
   * Shorthand for {@link #getExternalName()}.
   * @return the actual name that occurs in the information source
   */
  public String getSna() {
    return getExternalName();
  }

  /**
   * Sets the actual name that occurs in the information source (SNA)
   * if not otherwise present in the Metathesaurus.
   * @param sna the actual name that occurs in the information source
   */
  public void setExternalName(String sna) {
    this.sna = sna;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setExternalName(String)}.
   * @param sna the actual name that occurs in the information source
   */
  public void setSna(String sna) {
    setExternalName(sna);
  }

  /**
   * Shorthand for {@link #getRecordIdentifier()}.
   * @return the record identifier
   */
  public String getSoui() {
    return getRecordIdentifier();
  }

  /**
   * Returns the record identifier (SOUI) corresponding to this {@link Locator}
   * in the actual information source.
   * @return the record identifier
   */
  public String getRecordIdentifier() {
    return soui;
  }

  /**
   * Shorthand for {@link #setRecordIdentifier(String)}.
   * @param soui the record identifier
   */
  public void setSoui(String soui) {
    setRecordIdentifier(soui);
  }

  /**
   * Sets the record identifier (SOUI) corresponding to this {@link Locator}
   * in the actual information source.
   * @param soui the record identifier
   */
  public void setRecordIdentifier(String soui) {
    this.soui = soui;
    key_fields_flag = false;
  }

  /**
   * Returns the source for this {@link Locator}.
   * @return {@link Source}
   */
  public Source getSource() {
    return source;
  }

  /**
   * Sets the source for this {@link Locator}.
   * @param source {@link Source}
   */
  public void setSource(Source source) {
    this.source = source;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #getStringIdentifier()}.
   * @return the SUI
   */
  public String getSui() {
    return getStringIdentifier();
  }

  /**
   * Shorthand for {@link #setStringIdentifier(String)}.
   * @param sui the SUI
   */
  public void setSui(String sui) {
    setStringIdentifier(sui);
  }

  /**
   * Return the string identifier (SUI).
   * @return the SUI
   */
  public String getStringIdentifier() {
    return sui;
  }

  /**
   * Sets the string identifier (SUI).
   * @param sui the SUI
   */
  public void setStringIdentifier(String sui) {
    this.sui = sui;
    key_fields_flag = false;
  }

  /**
   * Returns tbe {@link ContentView} definition for this {@link Locator} (CVF).
   * @return tbe {@link ContentView} definition
   */
  public ContentView getContentView() {
    return content_view;
  }

  /**
   * Sets tbe {@link ContentView} definition for this {@link Locator} (CVF).
   * @param cvf the {@link ContentView} definition
   */
  public void setContentView(ContentView cvf) {
    this.content_view = cvf;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #getContentView()}.
   * @return tbe {@link ContentView} definition
   */
  public ContentView getCvf() {
    return getContentView();
  }

  /**
   * Shorthand for {@link #setContentView(ContentView)}.
   * @param cvf the {@link ContentView} definition
   */
  public void setCvf(ContentView cvf) {
    setContentView(cvf);
  }


  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from the cui, aui,
   * versioned source, fr, un, sui, sna, and soui.
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
      StringBuffer sb1 = new StringBuffer(100);
      sb1.append(getConcept() != null ? getConcept().getCui() : "")
         .append(getAtom() != null ? getAtom().getAui() : "")
         .append(getSource() != null ? getSource().getVsab() : "")
         .append(getFr())
         .append(getUn())
         .append(getSui())
         .append(getSna())
         .append(getSoui());
      key_fields = sb1.toString();
      key_fields_flag = true;
    }
    return key_fields;
  }
  /**
   * Return a hash code based on key fields.
   * @return a hash code based on key fields
   */
  public int hashCode() {
    return getKeyFields().hashCode();
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be a
   * {@link Locator} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof Locator)) return false;
    return getKeyFields().equals(((Locator)o).getKeyFields());
  }

  /**
   * Compares this {@link Locator} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof Locator)
      return getKeyFields().compareTo(((Locator)o).getKeyFields());
    else
      return 0;
  }

}
