/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta.impl
 * Class:       MetaDescriptionImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.MetaDescription;
import gov.nih.nlm.util.FieldedStringTokenizer;

/**
 * Default implementation of the {@link gov.nih.nlm.umls.meta.MetaDescription}
 * interface.  The sole function of this class is to read a line from MRDOC.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class MetaDescriptionImpl implements MetaDescription {

  //
  // Fields
  //
  private String key;
  private String value;
  private String type;
  private String explain;

  // Used to compute key fields more efficiently
  protected boolean key_fields_flag;
  protected String key_fields = null;

  /**
   * Instantiates an empty {@link MetaDescriptionImpl}.
   */
  public MetaDescriptionImpl() {};

  /**
   * Constructor that takes a row from the MRDOC file.
   * @param mrdoc_row row from MRDOC file
   */
  public MetaDescriptionImpl(String mrdoc_row) {
    String [] tokens = FieldedStringTokenizer.split(mrdoc_row, "|", 5);
    /*type = tokens[0];
    key = tokens[1];
    subkey = tokens[2];
    value = tokens[3];*/
    key = tokens[0];
    value = tokens[1];
    type = tokens[2];
    explain = tokens[3];
  }

  /**
   * This is a copy constructor for {@link MetaDescription}.
   * @param md {@link MetaDescription} to be copied
   */
  public MetaDescriptionImpl(MetaDescription md) {
    key = md.getKey();
    value = md.getValue();
    type = md.getType();
    explain = md.getExplain();
  }

  /**
   * Returns the type.
   * @return the type
   */
  public String getType() {
    return type;
  }

  /**
   * Returns the key.
   * @return the key
   */
  public String getKey() {
    return key;
  }

  /**
   * Returns the explain field.
   * @return the explain field
   */
  public String getExplain() {
    return explain;
  }

  /**
   * Returns the more detailed description.
   * @return the description
   */
  public String getValue() {
    return value;
  }

  /**
   * Returns a {@link String} key which can be used by equals, compareTo, and
   * hashCode methods.
   *
   * @return a {@link String} composed from the key, value, and type.
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
      StringBuffer sb = new StringBuffer(50);
      sb.append(getKey()).append(getValue()).append(getType());
      key_fields = sb.toString();
      key_fields_flag = true;
    }
    return key_fields;
  }

  /**
   * Returns <code>true</code> if the specified object is equal to this object.
   * The specified object must be a {@link MetaDescription} with matching key
   * fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof MetaDescription)) return false;
    return getKeyFields().equals(((MetaDescription) o).getKeyFields());
  }

  /**
   * Return a hash code based on key fields.
   * @return a hash code based on key fields
   */
  public int hashCode() {
    return getKeyFields().hashCode();
  }

  /**
   * Compares this {@link MetaDescription} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the Object to be compared.
   * @return a negative integer, zero, or a positive integer as this object
   *         is less than, equal to, or greater than the specified object.
   * @throws ClassCastException if the specified object's type prevents it
   *                            from being compared to this Object.
   */
  public int compareTo(Object o) {
    if (o instanceof MetaDescription)
      return getKeyFields().compareTo(((MetaDescription) o).getKeyFields());
    else
      throw new ClassCastException("Object " + o + " not an instance of MetaDescription.");
  }
}
