/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta.impl
 * Class:       RelationshipImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.AdditionalRelationshipLabel;
import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.ContentView;
import gov.nih.nlm.umls.meta.Relationship;
import gov.nih.nlm.umls.meta.RelationshipAttribute;
import gov.nih.nlm.umls.meta.RelationshipLabel;
import gov.nih.nlm.umls.meta.Restrictor;
import gov.nih.nlm.umls.meta.Source;
import gov.nih.nlm.umls.meta.SourceIdentifierType;
import gov.nih.nlm.umls.meta.Suppressible;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

/**
 * Default implementation of the {@link gov.nih.nlm.umls.meta.Relationship}
 * interface.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class RelationshipImpl implements Relationship {

  //
  // Fields
  //
  private Atom atom;
  private Atom related_atom;
  private Concept related_concept;
  private RelationshipLabel rel;
  private AdditionalRelationshipLabel rela;
  private String srui;
  private Source sl;
  private String rg;
  private String rui;
  private Suppressible suppress;
  private Concept concept;
  private ContentView content_view;
  private Source source;
  private String dir;
  private List attributes;
  private SourceIdentifierType source_id_type;
  private SourceIdentifierType rel_source_id_type;

  // Used to compute key fields more efficiently
  protected boolean key_fields_flag;
  protected String key_fields = null;

  /**
   * Instantiates an empty {@link Relationship}.
   */
  public RelationshipImpl() {};

  /**
   * Instantiates a {@link Relationship} from the specified {@link Relationship}.
   * @param rel {@link Relationship} to be copied
   */
  public RelationshipImpl(Relationship rel) {
    setRelatedConcept(rel.getRelatedConcept());
    setRel(rel.getRel());
    setRela(rel.getRela());
    setSrui(rel.getSrui());
    setSl(rel.getSl());
    setRg(rel.getRg());
    setRui(rel.getRui());
    setSuppress(rel.getSuppress());
    setConcept(rel.getConcept());
    setSource(rel.getSource());
    setCvf(rel.getCvf());
    setSourceIdentifierType(rel.getSourceIdentifierType());
    setRelatedSourceIdentifierType(rel.getRelatedSourceIdentifierType());
  }

  /**
   * Returns the {@link SourceIdentifierType} to which this element is connected.
   * @return the {@link SourceIdentifierType} to which this element is connected
   */
  public SourceIdentifierType getSourceIdentifierType() {
    return source_id_type;
  }

  /**
   * Sets the {@link SourceIdentifierType} to which this element is connected.
   * @param sit the {@link SourceIdentifierType} to which this element is connected.
   */
  public void setSourceIdentifierType(SourceIdentifierType sit) {
    source_id_type = sit;
  }

  /**
   * Returns the {@link SourceIdentifierType} to which this element is connected.
   * @return the {@link SourceIdentifierType} to which this element is connected
   */
  public SourceIdentifierType getRelatedSourceIdentifierType() {
    return rel_source_id_type;
  }

  /**
   * Sets the {@link SourceIdentifierType} to which this element is connected.
   * @param sit the {@link SourceIdentifierType} to which this element is connected.
   */
  public void setRelatedSourceIdentifierType(SourceIdentifierType sit) {
    rel_source_id_type = sit;
  }

  /**
   * Returns the {@link RelationshipLabel} (REL).
   * @return the {@link RelationshipLabel}
   */
  public RelationshipLabel getRelationshipLabel() {
    return rel;
  }

  /**
   * Shorthand for {@link #getRelationshipLabel()}.
   * @return the {@link RelationshipLabel}
   */
  public RelationshipLabel getRel() {
    return getRelationshipLabel();
  }

  /**
   * Sets the {@link RelationshipLabel} (REL).
   * @param rel the {@link RelationshipLabel}
   */
  public void setRelationshipLabel(RelationshipLabel rel) {
    this.rel = rel;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setRelationshipLabel(RelationshipLabel)}.
   * @param rel the {@link RelationshipLabel}
   */
  public void setRel(RelationshipLabel rel) {
    setRelationshipLabel(rel);
  }

  /**
   * Returns the related concept involved.
   * @return the related concept involved
   */
  public Concept getRelatedConcept() {
    return related_concept;
  }

  /**
   * Sets the related concept involved.
   * @param related_concept the related concept involved
   */
  public void setRelatedConcept(Concept related_concept) {
    this.related_concept = related_concept;
    key_fields_flag = false;
  }

  /**
   * Returns the related atom involved.
   * @return the related atom involved
   */
  public Atom getRelatedAtom() {
    return related_atom;
  }

  /**
   * Sets the related atom involved.
   * @param related_atom the related atom involved
   */
  public void setRelatedAtom(Atom related_atom) {
    this.related_atom = related_atom;
    key_fields_flag = false;
  }

  /**
   * Returns the {@link AdditionalRelationshipLabel} (RELA).
   * @return the {@link AdditionalRelationshipLabel}
   */
  public AdditionalRelationshipLabel getAdditionalRelationshipLabel() {
    return rela;
  }

  /**
   * Shorthand for {@link #getAdditionalRelationshipLabel()}.
   * @return the {@link AdditionalRelationshipLabel}
   */
  public AdditionalRelationshipLabel getRela() {
    return getAdditionalRelationshipLabel();
  }

  /**
   * Sets the {@link AdditionalRelationshipLabel} (RELA).
   * @param rela the {@link AdditionalRelationshipLabel}
   */
  public void setAdditionalRelationshipLabel(AdditionalRelationshipLabel rela) {
    this.rela = rela;
    key_fields_flag = false;
  }

  /**
   * Shorthand for
   * {@link #setAdditionalRelationshipLabel(AdditionalRelationshipLabel)}.
   * @param rela the {@link AdditionalRelationshipLabel}
   */
  public void setRela(AdditionalRelationshipLabel rela) {
    setAdditionalRelationshipLabel(rela);
  }

  /**
   * Returns the source asserted relationship identifier (SRUI).
   * @return the source asserted relationship identifier
   */
  public String getSourceRelationshipIdentifier() {
    return srui;
  }

  /**
   * Shorthand for {@link #getSourceRelationshipIdentifier()}.
   * @return the source asserted relationship identifier
   */
  public String getSrui() {
    return getSourceRelationshipIdentifier();
  }

  /**
   * Sets the source asserted relationship identifier (SRUI).
   * @param srui the source asserted relationship identifier
   */
  public void setSourceRelationshipIdentifier(String srui) {
    this.srui = srui;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setSourceRelationshipIdentifier(String)}.
   * @param srui the source asserted relationship identifier
   */
  public void setSrui(String srui) {
    setSourceRelationshipIdentifier(srui);
  }

  /**
   * Returns the source of the relationship label (SL).
   * @return the source of the relationship label
   */
  public Source getSourceOfLabel() {
    return sl;
  }

  /**
   * Shorthand for {@link #getSourceOfLabel()}.
   * @return the source of the relationship label
   */
  public Source getSl() {
    return getSourceOfLabel();
  }

  /**
   * Sets the source of the relationship label (SL).
   * @param sl the source of the relationship label
   */
  public void setSourceOfLabel(Source sl) {
    this.sl = sl;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setSourceOfLabel(Source)}.
   * @param sl the source of the relationship label
   */
  public void setSl(Source sl) {
    setSourceOfLabel(sl);
  }

  /**
   * Returns the relationship group (RG).
   * @return the relationship group
   */
  public String getRelationshipGroup() {
    return rg;
  }

  /**
   * Shorthand for {@link #getRelationshipGroup()}.
   * @return the relationship group
   */
  public String getRg() {
    return getRelationshipGroup();
  }

  /**
   * Sets the relationship group (RG).
   * @param rg the relationship group
   */
  public void setRelationshipGroup(String rg) {
    this.rg = rg;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #getRelationshipGroup()}.
   * @param rg the relationship group
   */
  public void setRg(String rg) {
    setRelationshipGroup(rg);
  }

  /**
   * Returns the {@link Atom} this {@link Relationship} is connected to.
   * @return the {@link Atom} this {@link Relationship} is connected to
   */
  public Atom getAtom() {
    return atom;
  }

  /**
   * Sets the {@link Atom} this {@link Relationship} is connected to.
   * @param atom  the {@link Atom} this {@link Relationship} is connected to
   */
  public void setAtom(Atom atom) {
    this.atom = atom;
    key_fields_flag = false;
  }

  /**
   * Returns the source for this {@link Relationship}.
   * @return {@link Source}
   */
  public Source getSource() {
    return source;
  }

  /**
   * Sets the source for this {@link Relationship}.
   * @param source {@link Source}
   */
  public void setSource(Source source) {
    this.source = source;
    key_fields_flag = false;
  }

  /**
   * Returns the suppressibility value (SUPPRESS).
   * @return the {@link Suppressible} value
   */
  public Suppressible getSuppress() {
    return suppress;
  }

  /**
   * Indicates whether or not this {@link Atom} is suppressible.
   * @return <code>true</code> if suppressible,
   *         <code>false</code> if not suppressible
   */
  public boolean isSuppressible() {
    return ((suppress == null || suppress.getAbbreviation().equals("N")) ? false : true);
  }

  /**
   * Sets the suppressibility value (SUPPRESS).
   * @param suppress the {@link Suppressible} value
   */
  public void setSuppress(Suppressible suppress) {
    this.suppress = suppress;
    key_fields_flag = false;
  }

  /**
   * Sets the directionality value (DIR).
   * @param dir the directionality value
   */
  public void setDir(String dir) {
    this.dir = dir;
    key_fields_flag = false;
  }

  /**
   * Returns the directionality value (DIR).
   * @return the directionality value
   */
  public String getDir() {
    if (dir == null || dir.equals("") ) return "";
    return (dir.equals("Y") ? "Y" : "N");
  }

  /**
   * Indicates the directionality of this {@link Relationship}.
   * @return <code>true</code> if directionality is in order read,
   *         <code>false</code> if directionality is reversed
   */
  public boolean isAssertedDirection() {
    return ((dir == null || dir.equals("") || dir.equals("N")) ? false : true);
  }

  /**
   * Sets the flag indicating the directionality of the {@link Relationship}.
   * @param dir <code>true</code> if directionality is in order read,
   *         <code>false</code> if directionality is reversed
   */
  public void setIsAssertedDirection(boolean dir) {
    if (dir)
      this.dir = "Y";
    else
      this.dir = "N";
    key_fields_flag = false;
  }

  /**
   * Returns the {@link Concept} that holds this {@link Relationship}.
   * @return the {@link Concept} that holds this {@link Relationship}
   */
  public Concept getConcept() {
    return concept;
  }

  /**
   * Sets the {@link Concept} that holds this {@link Relationship}.
   * @param concept the {@link Concept} that holds this {@link Relationship}
   */
  public void setConcept(Concept concept) {
    this.concept = concept;
    key_fields_flag = false;
  }


  /**
   * Shorthand for {@link #getRelationshipIdentifier()}.
   * @return the RUI
   */
  public String getRui() {
    return getRelationshipIdentifier();
  }

  /**
   * Shorthand for {@link #setRelationshipIdentifier(String)}.
   * @param rui the RUI
   */
  public void setRui(String rui) {
    setRelationshipIdentifier(rui);
  }

  /**
   * Returns the Relationship unique identifier (RUI) related to this
   * {@link Relationship}.
   * @return the RUI
   */
  public String getRelationshipIdentifier() {
    return rui;
  }

  /**
   * Sets the Relationship unique identifier (RUI) related to this
   * {@link Relationship}.
   * @param rui the RUI
   */
  public void setRelationshipIdentifier(String rui) {
    this.rui = rui;
    key_fields_flag = false;
  }


  /**
   * Returns tbe {@link ContentView} definition for this {@link Relationship} (CVF).
   * @return tbe {@link ContentView} definition
   */
  public ContentView getContentView() {
    return content_view;
  }

  /**
   * Sets tbe {@link ContentView} definition for this {@link Relationship} (CVF).
   * @param cvf the {@link ContentView} definition
   */
  public void setContentView(ContentView cvf) {
    content_view = cvf;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #getContentView()}.
   * @return tbe {@link ContentView} definition
   */
  public ContentView getCvf() {
    return getContentView();
  }

  /**
   * Shorthand for {@link #setContentView(ContentView)}.
   * @param cvf the {@link ContentView} definition
   */
  public void setCvf(ContentView cvf) {
    setContentView(cvf);
  }

  //
  // RelationshipAttributes
  //

  /**
   * Returns the {@link RelationshipAttribute}s.
   * @return a {@link RelationshipAttribute} <code>[]</code>
   */
  public RelationshipAttribute[] getAttributes() {
    return (RelationshipAttribute[])getAttributesAsList().toArray(new RelationshipAttribute[0]);
  }

  /**
   * Returns the {@link RelationshipAttribute}s as a {@link List}.
   * @return the {@link RelationshipAttribute}s as a {@link List}
   */
  public List getAttributesAsList() {
    if (attributes != null)
      return new ArrayList(attributes);
    else
      return new ArrayList();
  }

  /**
   * Adds the specified {@link RelationshipAttribute}.
   * @param att the {@link RelationshipAttribute} to be added
   */
  public void addAttribute(RelationshipAttribute att) {
    if (attributes == null) attributes = new ArrayList();
    attributes.add(att);
  }

  /**
   * Removes the specified {@link RelationshipAttribute}.
   * @param att the {@link RelationshipAttribute} to be removed
   */
  public void removeAttribute(RelationshipAttribute att) {
    if (attributes != null)
      attributes.remove(att);
  }

  /**
   * Remove all {@link RelationshipAttribute}s.
   */
  public void clearAttributes() {
    if (attributes != null)
      attributes.clear();
  }

  /**
   * Replace {@link RelationshipAttribute}s with the ones in the specified array.
   * @param atts an {@link RelationshipAttribute} <code>[]</code>
   */
  public void setAttributes(RelationshipAttribute atts[]) {
    attributes = new ArrayList();
    for (int i = 0; i < atts.length; i++)
      attributes.add(atts[i]);
  }

  /**
       * Replace {@link RelationshipAttribute}s with the ones in the specified {@link List}.
   * @param atts a {@link List} of {@link RelationshipAttribute}s
   */
  public void setAttributes(List atts) {
    attributes = new ArrayList(atts);
  }

  /**
   * Returns a subset of this {@link Relationship}'s {@link RelationshipAttribute}s.
   * Restricts the return based on the given {@link Restrictor}.
   * @param restrictor {@link Restrictor}
   * @return RelationshipAttribute[]
   */
  public RelationshipAttribute[] getAttributes(Restrictor restrictor) {
    List atts = getAttributesAsList();
    for (int i=atts.size()-1; i>=0; i--) {
      if (!restrictor.keep(atts.get(i)))
        atts.remove(i);
    }
    Collections.sort(atts, restrictor);
    return (RelationshipAttribute[])atts.toArray(new RelationshipAttribute[0]);

  }

  /**
   * Indicates whether or not this {@link RelationshipImpl} is obsolete.
   * @return <code>true</code> if obsolete,
   *         <code>false</code> if not obsolete
   */
  public boolean isObsolete() {
    if (getSuppress().getAbbreviation().equals("O"))
      return true;
    return false;
  }

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from the cui1, rel,
   * versioned source, cui2, rela, suppress, rui, sl, and srui.
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
      final StringBuffer sb = new StringBuffer(80);
      sb.append(getConcept() != null ? getConcept().getCui() : "")
        .append(getRelatedConcept() != null ? getRelatedConcept().getCui() : "")
        .append(getSource() != null ? getSource().getVsab() : "")
        .append(getRel() != null ? getRel().getAbbreviation() : "")
        .append(getRela() != null ? getRela().getAbbreviation() : "")
        .append(getSuppress() != null ? getSuppress().getAbbreviation() : "")
        .append(getSl() != null ? getSl().getVsab() : "")
        .append(getRui())
        .append(getSrui());
      key_fields = sb.toString();
      key_fields_flag = true;
    }
    return key_fields;
  }

  /**
   * Return a hash code based on key fields.
   * @return a hash code based on key fields
   */
  public int hashCode() {
    return getKeyFields().hashCode();
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be an
   * {@link Relationship} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof Relationship)) return false;
    return getKeyFields().equals(((Relationship)o).getKeyFields());
  }

  /**
   * Compares this {@link Relationship} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof Relationship)
      return getKeyFields().compareTo(((Relationship)o).getKeyFields());
    else
      return 0;
  }


}
