/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta.impl
 * Class:       STYMetaDescriptionImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.STYMetaDescription;
import gov.nih.nlm.util.FieldedStringTokenizer;

import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;

/**
 * Default implementation of the {@link STYMetaDescription} interface.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class STYMetaDescriptionImpl implements STYMetaDescription {

  //
  // Fields
  //
  private String ui;
  private String sty;
  private String stn;
  private String def;
  private String un;
  private boolean nh;
  private String ex;

  /**
   * Instantiates an empty {@link STYMetaDescriptionImpl}.
   */
  public STYMetaDescriptionImpl() {};

  /**
   * Instantiates a {@link STYMetaDescription} from a row from the SRDEF file.
   * @param srdef_row row from SRDEF file
   */
  public STYMetaDescriptionImpl(String srdef_row) {
    String[] tokens = FieldedStringTokenizer.split(srdef_row, "|", 8);
    ui = tokens[0];
    sty = tokens[1];
    stn = tokens[2];
    def = tokens[3];
    ex = tokens[4];
    un = tokens[5];
    nh = (tokens[6].equals("Y")) ? true : false;
  }

  /**
   * Shorthand for {@link #getIdentifier()}.
   * @return the unique identifier
   */
  public String getUi() {
    return getIdentifier();
  }

  /**
   * Returns the unique identifier (UI).
   * @return the unique identifier
   */
  public String getIdentifier() {
    return ui;
  }

  /**
   * Shorthand for {@link #getName()}.
   * @return the name
   */
  public String getSty() {
    return getName();
  }

  /**
   * Returns the name (STY).
   * @return the name
   */
  public String getName() {
    return sty;
  }

  /**
   * Shorthand for {@link #getTreeNumber()}.
   * @return the tree number
   */
  public String getStn() {
    return getTreeNumber();
  }

  /**
   * Returns the tree number (STN).
   * @return the tree number
   */
  public String getTreeNumber() {
    return stn;
  }

  /**
   * Shorthand for {@link #getDefinition()}.
   * @return the definition
   */
  public String getDef() {
    return getDefinition();
  }

  /**
   * Returns the definition (DEF).
   * @return the definition
   */
  public String getDefinition() {
    return def;
  }

  /**
   * Shorthand for {@link #getExamplesAsString()}.
   * @return a value listing Metathesaurus examples
   */
  public String getEx() { return getExamplesAsString(); };

  /**
   * Returns a {@link String} array of of Metathesaurus strings (EX)
   * which are examples of this semantic type.
   * @return an {@link String} array of Metathesaurus examples
   */
  public String[] getExamples() {
    StringTokenizer st = new StringTokenizer(getEx(),";");
    String[] x = new String[st.countTokens()];
    int ct = 0;
    while (st.hasMoreTokens())
      x[ct++] = st.nextToken().trim();
    return x;
  };

  /**
   * Returns a semi-colon (;) separated list of Metathesaurus strings (EX)
   * which are examples of this semantic type.
   * @return a value listing Metathesaurus examples
   */
  public String getExamplesAsString() { return ex; };

  /**
   * Returns a {@link List} array of of Metathesaurus strings (EX)
   * which are examples of this semantic type.
   * @return an {@link List} of Metathesaurus examples
   */
  public List getExamplesAsList() {
    StringTokenizer st = new StringTokenizer(getEx(),";");
    List x = new ArrayList();
    while (st.hasMoreTokens())
      x.add(st.nextToken().trim());
    return x;
  };

  /**
   * Shorthand for {@link #getUsageNote()}.
   * @return the usage note
   */
  public String getUn() {
    return getUsageNote();
  }

  /**
   * Returns the usage note (UN).
   * @return the usage note
   */
  public String getUsageNote() {
    return un;
  }

  /**
   * Indicates whether or not this semantic type or its descendents allow
   * the non human flag.
   * @return <code>true</code> if allowed, <code>false</code> if not
   */
  public boolean allowsNonHumanFlag() {
    return nh;
  }

  /**
   * Returns the non human flag value (NH).
   * @return the non human flag value
   */
  public String getNH() {
    return ((nh) ? "true" : "false");
  }

  /**
   * Return a hash code based on key fields.
   * @return a hash code based on key fields
   */
  public int hashCode() {
    return (getUi()).hashCode();
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be a
   * {@link STYMetaDescription} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof STYMetaDescription)) return false;
    STYMetaDescription sty = (STYMetaDescription)o;
    return
      sty.getUi() == getUi();
  }

  /**
   * Compares this {@link STYMetaDescription} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof STYMetaDescription)
      return getUi().compareTo(
        ((STYMetaDescription)o).getUi());
    else
      return 0;
  }
}
