/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta.impl
 * Class:       SourceContextImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.ContentView;
import gov.nih.nlm.umls.meta.ContextMember;
import gov.nih.nlm.umls.meta.SourceContext;
import gov.nih.nlm.umls.meta.StringContext;

import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;

/**
 * Default implementation fo the {@link SourceContext} interface.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class SourceContextImpl implements SourceContext {

  //
  // Fields
  //
  private int cxn;
  private Atom atom;
  private String paui;
  private ContentView cvf;
  private StringContext string_context;
  private List path_to_root = new ArrayList();
  private List context_members = new ArrayList();


  // Used to compute key fields more efficiently
  private boolean key_fields_flag;
  private String key_fields = null;

  /**
   * Instantiates an empty {@link SourceContext}.
   */
  public SourceContextImpl() {};

  /**
   * Instantiates a {@link SourceContext} from the specified {@link SourceContext}.
   * @param sc {@link SourceContext} to be copied
   */
  public SourceContextImpl(SourceContext sc) {
    setAtom(sc.getAtom());
    setPaui(sc.getPaui());
    setCxn(sc.getCxn());
    setStringContext(sc.getStringContext());
    setPtr(sc.getPtr());
    setCvf(sc.getCvf());
    setContextMembers(sc.getContextMembers());
  }

  /**
   * Returns the {@link StringContext} that holds this {@link SourceContext}.
   * @return the {@link StringContext} that holds this {@link SourceContext}
   */
  public StringContext getStringContext() {
    return string_context;
  }

  /**
   * Sets the {@link StringContext} that holds this {@link SourceContext}.
   * @param sc the {@link StringContext} that holds this {@link SourceContext}
   */
  public void setStringContext(StringContext sc) {
    string_context = sc;
    key_fields_flag = false;
  }

  /**
   * Returns the context number (CXN).
   * @return the <code>int</code> context number
   */
  public int getContextNumber() {
    return cxn;
  }

  /**
   * Sets the context number (CXN).
   * @param cxn the <code>int</code> context number
   */
  public void setContextNumber(int cxn) {
    this.cxn = cxn;
    key_fields_flag = false;
  }

  /**
   * This is a shorthand for {@link #getContextNumber()}
   * @return cxn context number <code>int</code>
   */
  public int getCxn() {
    return getContextNumber();
  }

  /**
   * Shorthand for {@link #setContextNumber(int)}.
   * @param cxn <code>int</code> context number
   */
  public void setCxn(int cxn) {
    setContextNumber(cxn);
  }

  /**
   * Shorthand for {@link #getParentAtomIdentifier()}.
   * @return the PAUI
   */
  public String getPaui() {
    return getParentAtomIdentifier();
  }

  /**
   * Shorthand for {@link #setParentAtomIdentifier(String)}.
   * @param paui the PAUI
   */
  public void setPaui(String paui) {
    setParentAtomIdentifier(paui);
  }

  /**
   * Returns the parent atom identifier (PAUI).
   * @return the PAUI
   */
  public String getParentAtomIdentifier() {
    return paui;
  }

  /**
   * Sets the parent atom identifier (PAUI).
   * @param paui the PAUI
   */
  public void setParentAtomIdentifier(String paui) {
    this.paui = paui;
    key_fields_flag = false;
  }

  /**
   * Returns the path to the root (PTR) as a "." separated list of AUI values.
   * @return the path to the root
   */
  public String getPathToRootAsString() {
    StringBuffer sb = new StringBuffer(200);
    for (int i=0; i<path_to_root.size(); i++) {
      sb.append(path_to_root.get(i));
      if (i<=path_to_root.size() -2) sb.append(".");
    }
    return sb.toString();
  }

  /**
   * Returns the path to the root (PTR) as a {@link List} of AUI values.
   * @return the path to the root as a {@link List} of AUI values
   */
  public List getPathToRootAsList() {
    return new ArrayList(path_to_root);
  }

  /**
   * Returns the path to the root (PTR) as a {@link String}<code>[]</code>
   * of AUI values.
   * @return the path to the root as a {@link String}<code>[]</code>
   * of AUI values
   */
  public String [] getPathToRoot() {
    return (String[])path_to_root.toArray(new String[0]);
  }

  /**
   * Shorthand for {@link #getPathToRootAsString()}.
   * @return the path to the root
   */
  public String getPtr() {
    return getPathToRootAsString();
  }

  /**
   * Sets the path to the root (PTR) as a "." separated list of AUI values.
   * @param ptr the path to the root
   */
  public void setPathToRoot(String ptr) {
    StringTokenizer st = new StringTokenizer(ptr, ".");
     while (st.hasMoreTokens()) {
         path_to_root.add(st.nextToken());
     }
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setPathToRoot(String)}.
   * @param ptr the path to the root
   */
  public void setPtr(String ptr) {
    setPathToRoot(ptr);
  }

  /**
   * Returns the {@link ContextMember}s.  These are the ancestors,
   * children, and siblings.
   * @return a sorted array of {@link ContextMember}s
   */
  public ContextMember [] getContextMembers() {
    return (ContextMember[])context_members.toArray(new ContextMember[0]);
  }

  /**
   * Returns the {@link ContextMember}s.  These are the ancestors,
   * children, and siblings.
   * @return a sorted List of {@link ContextMember}s
   */
  public List getContextMembersAsList() {
    return new ArrayList(context_members);
  }

  /**
   * Adds the specified {@link ContextMember}.
   * @param cm the {@link ContextMember} to be added
   */
  public void addContextMember(ContextMember cm) {
    context_members.add(cm);
  }

  /**
   * Removes the specified {@link ContextMember}.
   * @param cm the {@link ContextMember} to be removed
   */
  public void removeContextMember(ContextMember cm) {
    context_members.remove(cm);
  }

  /**
   * Removes all {@link ContextMember}s.
   */
  public void clearContextMembers() {
    context_members.clear();
  }

  /**
   * Set the {@link ContextMember}s with the ones in the specified array.
   * @param cms a {@link ContextMember}<code>[]</code>
   */
  public void setContextMembers(ContextMember [] cms) {
    context_members.clear();
    for (int i=0; i<cms.length; i++) {
      context_members.add(cms[i]);
    }
  }

  /**
   * Set the {@link ContextMember}s with the ones in the specified {@link List}.
   * @param cms a {@link List} of {@link ContextMember}s
   */
  public void setContextMembers(List cms) {
    context_members = new ArrayList(cms);
  }

  /**
   * Returns the {@link Atom} this {@link SourceContext} is connected to.
   * @return the {@link Atom} this {@link SourceContext} is connected to
   */
  public Atom getAtom() {
    return atom;
  }

  /**
   * Sets the {@link Atom} this {@link SourceContext} is connected to.
   * @param atom  the {@link Atom} this {@link SourceContext} is connected to
   */
  public void setAtom(Atom atom) {
    this.atom = atom;
    key_fields_flag = false;
  }

  /**
   * Returns tbe {@link ContentView} definition for this {@link SourceContext} (CVF).
   * @return tbe {@link ContentView} definition
   */
  public ContentView getContentView() {
    return cvf;
  }

  /**
   * Sets tbe {@link ContentView} definition for this {@link SourceContext} (CVF).
   * @param cvf the {@link ContentView} definition
   */
  public void setContentView(ContentView cvf) {
    this.cvf = cvf;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #getContentView()}.
   * @return tbe {@link ContentView} definition
   */
  public ContentView getCvf() {
    return getContentView();
  }

  /**
   * Shorthand for {@link #setContentView(ContentView)}.
   * @param cvf the {@link ContentView} definition
   */
  public void setCvf(ContentView cvf) {
    setContentView(cvf);
  }


  /**
   * Returns only the {@link ContextMember}s that have the context
   * member label (CXL) of "CCP" or self.
   * @return {@link ContextMember}[] of members with label "CCP"
   */
  public ContextMember[] getSelfMembers() {
    ArrayList all_members = (ArrayList)getContextMembersAsList();
    ArrayList members = new ArrayList();
    for (int i=0; i<all_members.size(); i++) {
      if (((ContextMember)all_members.get(i)).getLabel().equals("CCP"))
        members.add(all_members.get(i));
    }
    return (ContextMember[])members.toArray(new ContextMember[0]);
  }

  /**
   * Returns only the {@link ContextMember}s that have the context
   * member label (CXL) of "CCP" or self.
   * @return {@link List} of members with label "CCP"
   */
  public List getSelfMembersAsList() {
    ArrayList all_members = (ArrayList)getContextMembersAsList();
    ArrayList members = new ArrayList();
    for (int i=0; i<all_members.size(); i++) {
      if (((ContextMember)all_members.get(i)).getLabel().equals("CCP"))
        members.add(all_members.get(i));
    }
    return members;
  }

  /**
   * Returns only the {@link ContextMember}s that have the context
   * member label (CXL) of "CHD" or child.
   * @return {@link ContextMember}[] of members with label "CHD"
   */
  public ContextMember[] getChildMembers() {
    ArrayList all_members = (ArrayList)getContextMembersAsList();
    ArrayList members = new ArrayList();
    for (int i=0; i<all_members.size(); i++) {
      if (((ContextMember)all_members.get(i)).getLabel().equals("CHD"))
        members.add(all_members.get(i));
    }
    return (ContextMember[])members.toArray(new ContextMember[0]);
  }

  /**
   * Returns only the {@link ContextMember}s that have the context
   * member label (CXL) of "CHD" or child.
   * @return {@link ContextMember}[] of members with label "CHD"
   */
  public List getChildMembersAsList() {
    ArrayList all_members = (ArrayList)getContextMembersAsList();
    ArrayList members = new ArrayList();
    for (int i=0; i<all_members.size(); i++) {
      if (((ContextMember)all_members.get(i)).getLabel().equals("CHD"))
        members.add(all_members.get(i));
    }
    return members;
  }

  /**
   * Returns only the {@link ContextMember}s that have the context
   * member label (CXL) of "ANC" or self.
   * @return {@link ContextMember}[] of members with label "ANC"
   */
  public ContextMember[] getAncestorMembers() {
    ArrayList all_members = (ArrayList)getContextMembersAsList();
    ArrayList members = new ArrayList();
    for (int i=0; i<all_members.size(); i++) {
      if (((ContextMember)all_members.get(i)).getLabel().equals("ANC"))
        members.add(all_members.get(i));
    }
    return (ContextMember[])members.toArray(new ContextMember[0]);
  }

  /**
   * Returns only the {@link ContextMember}s that have the context
   * member label (CXL) of "ANC" or self.
   * @return {@link ContextMember}[] of members with label "ANC"
   */
  public List getAncestorMembersAsList() {
    ArrayList all_members = (ArrayList)getContextMembersAsList();
    ArrayList members = new ArrayList();
    for (int i=0; i<all_members.size(); i++) {
      if (((ContextMember)all_members.get(i)).getLabel().equals("ANC"))
        members.add(all_members.get(i));
    }
    return members;
  }

  /**
   * Returns only the {@link ContextMember}s that have the context
   * member label (CXL) of "SIB" or self.
   * @return {@link ContextMember}[] of members with label "SIB"
   */
  public ContextMember[] getSiblingMembers() {
    ArrayList all_members = (ArrayList)getContextMembersAsList();
    ArrayList members = new ArrayList();
    for (int i=0; i<all_members.size(); i++) {
      if (((ContextMember)all_members.get(i)).getLabel().equals("SIB"))
        members.add(all_members.get(i));
    }
    return (ContextMember[])members.toArray(new ContextMember[0]);
  }

  /**
   * Returns only the {@link ContextMember}s that have the context
   * member label (CXL) of "SIB" or self.
   * @return {@link ContextMember}[] of members with label "SIB"
   */
  public List getSiblingMembersAsList() {
    ArrayList all_members = (ArrayList)getContextMembersAsList();
    ArrayList members = new ArrayList();
    for (int i=0; i<all_members.size(); i++) {
      if (((ContextMember)all_members.get(i)).getLabel().equals("SIB"))
        members.add(all_members.get(i));
    }
    return members;
  }

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from the cui and aui.
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
      StringBuffer sb1 = new StringBuffer(100);
      sb1.append(getAtom() != null ? getAtom().getAui() : "")
        .append(getCxn());
      key_fields = sb1.toString();
      key_fields_flag = true;
    }
    return key_fields;
  }

  /**
   * Return a hash code based on key fields.
   * @return a hash code based on key fields
   */
  public int hashCode() {
    return (getKeyFields()).hashCode();
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be an
   * {@link SourceContext} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof SourceContext)) return false;
    return getKeyFields().equals(((SourceContext)o).getKeyFields());
  }

  /**
   * Compares this {@link SourceContext} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof SourceContext)
      return getKeyFields().compareTo(((SourceContext)o).getKeyFields());
    else
      return 0;
  }

}
