/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta.impl
 * Class:       StringAttributeImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.AttributeContext;
import gov.nih.nlm.umls.meta.StringAttribute;
import gov.nih.nlm.umls.meta.StringInfo;
import gov.nih.nlm.umls.meta.TermAttribute;

import java.util.ArrayList;
import java.util.List;

/**
 * Default implementation of the {@link StringAttribute} interface.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class StringAttributeImpl implements StringAttribute {

  //
  // Fields
  //
  private TermAttribute term_attribute;
  private List attribute_contexts = new ArrayList();
  private StringInfo string_info;

  // Used to compute key fields more efficiently
  private boolean key_fields_flag;
  private String key_fields = null;


  /**
   * Instantiates an empty {@link StringAttribute}.
   */
  public StringAttributeImpl() {};

  /**
   * Instantiates a {@link StringAttribute} from the specified {@link StringAttribute}.
   * @param sa {@link StringAttribute} to be copied
   */
  public StringAttributeImpl(StringAttribute sa) {
    setTermAttribute(sa.getTermAttribute());
    setAttributeContexts(sa.getAttributeContexts());
    setStringInfo(sa.getStringInfo());
  }

  /**
   * Returns the {@link TermAttribute} that holds this {@link StringAttribute}.
   * @return the {@link TermAttribute} that holds this {@link StringAttribute}
   */
  public TermAttribute getTermAttribute() {
    return term_attribute;
  }

  /**
   * Sets the {@link TermAttribute} that holds this {@link StringAttribute}.
   * @param ta the {@link TermAttribute} that holds this {@link StringAttribute}
   */
  public void setTermAttribute(TermAttribute ta) {
    term_attribute = ta;
    key_fields_flag = false;
  }


  /**
   * Returns the {@link StringInfo}.
   * @return the {@link StringInfo}
   */
  public StringInfo getStringInfo() {
    return string_info;
  }

  /**
   * Sets the {@link StringInfo}.
   * @param si the {@link StringInfo}
   */
  public void setStringInfo(StringInfo si) {
    string_info = si;
    key_fields_flag = false;
  }

  /**
   * Returns all {@link AttributeContext}s.
   * @return a array of {@link AttributeContext}s
   */
  public AttributeContext [] getAttributeContexts() {
    return (AttributeContext[])attribute_contexts.toArray(new AttributeContext[0]);
  }

  /**
   * Returns all {@link AttributeContext}s as a {@link List}.
   * @return a {@link List} of {@link AttributeContext}s
   */
  public List getAttributeContextsAsList() {
    return new ArrayList(attribute_contexts);
  }

  /**
   * Adds the specified {@link AttributeContext}.
   * @param ac the {@link AttributeContext} to be added
   */
  public void addAttributeContext(AttributeContext ac) {
    attribute_contexts.add(ac);
  }

  /**
   * Removes the specified {@link AttributeContext}.
   * @param ac the {@link AttributeContext} to be removed
   */
  public void removeAttributeContext(AttributeContext ac) {
    attribute_contexts.remove(ac);
  }

  /**
   * Removes all {@link AttributeContext}s.
   */
  public void clearAttributeContexts() {
    attribute_contexts.clear();
  }

  /**
   * Replaces the {@link AttributeContext}s with the ones in the specified array.
   * @param acs a {@link AttributeContext}<code>[]</code>
   */
  public void setAttributeContexts(AttributeContext [] acs) {
    attribute_contexts.clear();
    for (int i=0; i<acs.length; i++ ) {
      attribute_contexts.add(acs[i]);
    }
  }

  /**
   * Replaces the {@link AttributeContext}s with the ones in the specified {@link List}.
   * @param acs a {@link List} of {@link AttributeContext}s
   */
  public void setAttributeContexts(List acs) {
    attribute_contexts = new ArrayList(acs);
  }

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from the cui, aui, sui,
   * versioned source, tty, and code.
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
      key_fields = getStringInfo().getSui();
      key_fields_flag = true;
    }
    return key_fields;
  }


  /**
   * Return a hash code based on key fields.
   * @return a hash code based on key fields
   */
  public int hashCode() {
    return getKeyFields().hashCode();
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be an
   * {@link StringAttribute} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof StringAttribute)) return false;
    return getKeyFields().equals(((StringAttribute)o).getKeyFields());
  }

  /**
   * Compares this {@link StringAttribute} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof StringAttribute)
      return getKeyFields().compareTo(((StringAttribute)o).getKeyFields());
    else
      return 0;
  }

}
