/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta.impl
 * Class:       StringInfoImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.StringInfo;
import gov.nih.nlm.umls.meta.StringType;
import gov.nih.nlm.umls.meta.Term;

import java.util.ArrayList;
import java.util.List;

/**
 * Default implementation of the {@link StringInfo} interface.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class StringInfoImpl implements StringInfo {

  //
  // Fields
  //
  private StringType stt;
  private String sui;
  private String str;
  private List atoms = new ArrayList();
  private Term term;

  // Used to compute key fields more efficiently
  private boolean key_fields_flag;
  private String key_fields = null;


  /**
   * Instantiates an empty {@link StringInfo}.
   */
  public StringInfoImpl() {};

  /**
   * Instantiates a {@link StringInfo} from the specified {@link StringInfo}.
   * @param si {@link StringInfo} to be copied
   */
  public StringInfoImpl(StringInfo si) {
    setStt(si.getStt());
    setSui(si.getSui());
    setStr(si.getStr());
    setAtoms(si.getAtoms());
  }

  /**
   * Returns the parent {@link Term}.
   * @return the parent {@link Term}
   */
  public Term getTerm() { return term; }

  /**
   * Sets the parent {@link Term}.
   * @param term the parent {@link Term}
   */
  public void setTerm(Term term) { this.term = term; }

  /**
   * Returns the {@link StringType} (STT).
   * @return the {@link StringType}
   */
  public StringType getStringType() {
    return stt;
  }

  /**
   * Shorthand for {@link #getStringType()}.
   * @return the {@link StringType}
   */
  public StringType getStt() {
    return getStringType();
  }

  /**
   * Sets the {@link StringType} (STT).
   * @param stt the STT
   */
  public void setStringType(StringType stt) {
    this.stt = stt;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setStringType(StringType)}.
   * @param stt the STT
   */
  public void setStt(StringType stt) {
    setStringType(stt);
  }

  /**
   * Shorthand for {@link #getStringIdentifier()}.
   * @return the SUI
   */
  public String getSui() {
    return getStringIdentifier();
  }

  /**
   * Shorthand for {@link #setStringIdentifier(String)}.
   * @param sui the SUI
   */
  public void setSui(String sui) {
    setStringIdentifier(sui);
  }

  /**
   * Return the string identifier (SUI).
   * @return the SUI
   */
  public String getStringIdentifier() {
    return sui;
  }

  /**
   * Sets the string identifier (SUI).
   * @param sui the SUI
   */
  public void setStringIdentifier(String sui) {
    this.sui = sui;
    key_fields_flag = false;
  }

  /**
   * Sets the name (STR).
   * @return the name
   */
  public String getName() {
    return str;
  }

  /**
   * Shorthand for {@link #getName()}.
   * @return the name
   */
  public String getStr() {
    return getName();
  }

  /**
   * Sets the name (STR).
   * @param str the name
   */
  public void setName(String str) {
    this.str = str;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setName(String)}.
   * @param str the name
   */
  public void setStr(String str) {
    setName(str);
  }

  /**
   * Returns all {@link Atom}s.
   * @return an array of {@link Atom}s
   */
  public Atom[] getAtoms() {
    return (Atom[])atoms.toArray(new Atom[0]);
  }

  /**
   * Returns all {@link Atom}s as a {@link List}.
   * @return all {@link Atom}s as a {@link List}
   */
  public List getAtomsAsList() {
    return new ArrayList(atoms);
  }

  /**
   * Adds the specified {@link Atom}.
   * @param atom the {@link Atom} to add
   */
  public void addAtom(Atom atom) {
    atoms.add(atom);
  }

  /**
   * Removes the specified {@link Atom}.
   * @param atom the {@link Atom} to remove
   */
  public void removeAtom(Atom atom) {
    atoms.remove(atom);
  }

  /**
   * Removes all {@link Atom}s.
   */
  public void clearAtoms() {
    atoms.clear();
  }

  /**
   * Replaces the {@link Atom}s with those in the specified array.
   * @param l_atoms an array of {@link Atom}s
   */
  public void setAtoms(Atom [] l_atoms) {
    atoms.clear();
    for (int i=0; i<l_atoms.length; i++) {
      atoms.add(l_atoms[i]);
    }
  }

  /**
   * Replaces the {@link Atom}s with those in the specified {@link List}
   * @param atoms a {@link List} of {@link Atom}s
   */
  public void setAtoms(List atoms) {
    this.atoms = new ArrayList(atoms);
  }

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from the sui.
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
      key_fields = getSui();
      key_fields_flag = true;
    }
    return key_fields;
  }

  /**
   * Return a hash code based on key fields.
   * @return a hash code based on key fields
   */
  public int hashCode() {
    return getKeyFields().hashCode();
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be an
   * {@link StringInfo} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof StringInfo)) return false;
    return getKeyFields().equals(((StringInfo)o).getKeyFields());
  }

  /**
   * Compares this {@link StringInfo} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof StringInfo)
      return getKeyFields().compareTo(((StringInfo)o).getKeyFields());
    else
      return 0;
  }

  /**
   * Indicates whether or not this has the preferred string identifier
   * of the {@link Term} it is part of.  In other words, does this {@link StringInfo} have
   * the same SUI as the preferred {@link StringInfo} of the {@link Term}.
   * @return <code>true</code> if so, <code>false</code> otherwise
   */
  public boolean hasPreferredStringIdentifier() {
    return getStt().getAbbreviation().equals("PF");
  }

  /**
   * Shorthand for {@link #hasPreferredStringIdentifier()}.
   * @return <code>true</code> if so, <code>false</code> otherwise
   */
  public boolean hasPreferredSui() {
    return hasPreferredStringIdentifier();
  }
}
